/*
 *	This file is part of qp42.
 *
 *	qp42 -- An Implementation of qp solver on 42 intervals.
 *	Copyright (C) 2012 by Janick Frasch, Hans Joachim Ferreau et al. 
 *	All rights reserved.
 *
 *	qp42 is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU Lesser General Public
 *	License as published by the Free Software Foundation; either
 *	version 2.1 of the License, or (at your option) any later version.
 *
 *	qp5 is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *	See the GNU Lesser General Public License for more details.
 *
 *	You should have received a copy of the GNU Lesser General Public
 *	License along with qp42; if not, write to the Free Software
 *	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */


/**
 *	\file examples/chainMass_M3_N50_1stStep.c
 *	\author Janick Frasch, Milan Vukov
 *	\version 1.0beta
 *	\date 2012
 *
 *	Dumped output from ACADO CodeGen chain of masses example (1st NMPC interval)
 */


#include <mpc/setup_mpc.h>
//#include "chainMass_M3_N50_1stStep_data.h"		/* defines global problem data */

#define INFTY 1.0e12


void setupData( double* H, double* g, double* C, double* c, double* dLow, double* dUpp, double* x0 );


int main( )
{
	int i;
	
	const unsigned int nI = 50;		/* number of control intervals */
	const unsigned int nX = 21;		/* number of states */
	const unsigned int nU = 3;		/* number of controls */
	const unsigned int nZ = nX+nU;	/* number of variables */
	unsigned int* nD = 0;  			/* number of constraints */
	

	double H[nI*nZ*nZ + nX*nX];
	for ( i = 0; i < nI*nZ*nZ + nX*nX; ++i )	H[i] = 0.0;		/* make sure H is initialized well */
	double g[nI*nZ + nX];
	double C[nI*nX*nZ];
	for ( i = 0; i < nI*nX*nZ; ++i )	C[i] = 0.0;		/* make sure C is initialized well */
	double c[nI*nX];
	double zLow[nI*nZ + nX];
	double zUpp[nI*nZ + nX];
	double x0[nX];
			
	
	setupData( H, g, C, c, zLow, zUpp, x0 );



	/** set up a new mpcDUNES problem */
	mpcProblem_t mpcProblem;
	qpOptions_t qpOptions = qpDUNES_setupDefaultOptions();
	
	/** set qpDUNES options */
	qpOptions.maxIter    = 100;
	qpOptions.printLevel = 3;
//	qpOptions.printIterationTiming = QPDUNES_TRUE;
//	qpOptions.logLevel = QPDUNES_LOG_ALL_DATA;
//	qpOptions.logLevel = QPDUNES_LOG_ITERATIONS;
	qpOptions.logLevel = QPDUNES_LOG_OFF;
	qpOptions.stationarityTolerance = 1.e-6;
	qpOptions.equalityTolerance     = 2.221e-16;
	qpOptions.nbrInitialGradientSteps = 0;
	qpOptions.regParam            = 1.e-4;
//	qpOptions.regParam            = 1.e10;
//	qpOptions.newtonHessDiagRegTolerance  = 1.e-4;
	qpOptions.newtonHessDiagRegTolerance  = 1.e-8;
//	qpOptions.lsType			= QP42_LS_GRADIENT_BISECTION_LS;
//	qpOptions.lsType			= QPDUNES_LS_GRID_LS;
//	qpOptions.lsType			= QP42_LS_ACCELERATED_GRID_LS;
//	qpOptions.lsType			= QP42_LS_BACKTRACKING_LS_WITH_AS_CHANGE;
	qpOptions.lsType			= QPDUNES_LS_ACCELERATED_GRADIENT_BISECTION_LS;
//	qpOptions.lsType			= QPDUNES_LS_BACKTRACKING_LS;
//	qpOptions.lineSearchNbrGridPoints   		= 30;
//	qpOptions.lineSearchNbrGridPoints   		= 5000;
	qpOptions.lineSearchStationarityTolerance = 1.e-3;
//	qpOptions.lineSearchStationarityTolerance = 1.e-6;
	qpOptions.lineSearchReductionFactor	= 0.1;
//	qpOptions.lineSearchMaxStepSize	= 10.;
	qpOptions.lineSearchMaxStepSize	= 1.;
	qpOptions.maxNumLineSearchIterations            = 15;
//	qpOptions.maxNumLineSearchIterations            = 25;
	qpOptions.maxNumLineSearchRefinementIterations  = 15;
//	qpOptions.maxNumLineSearchRefinementIterations  = 150;
//	qpOptions.regType            = QP42_REG_NORMALIZED_LEVENBERG_MARQUARDT;
//	qpOptions.regType            = QPDUNES_REG_LEVENBERG_MARQUARDT;
//	qpOptions.regType            = QPDUNES_REG_GRADIENT_STEP;
	qpOptions.regType            = QPDUNES_REG_SINGULAR_DIRECTIONS;

	mpcDUNES_setup( &mpcProblem, nI, nX, nU, nD, &(qpOptions) );

	
	/* setup */
	double t0 = getTime();
	mpcDUNES_initLtvSb(	&mpcProblem, H, g, C,c, zLow,zUpp, 0 );

	/* solve */
	double t1 = getTime();
	mpcDUNES_solve( &mpcProblem, x0 );
	double t2 = getTime();



	mpcDUNES_cleanup( &mpcProblem );

	printf( "Preparation time:       %.2lf ms\n", 1e3*(t1-t0) );
	printf( "Solution time:          %.2lf ms\n", 1e3*(t2-t1) );
	printf( "Total computation time: %.2lf ms\n", 1e3*(t2-t0) );

	return 0;
}



void setupData( double* H, double* g, double* C, double* c, double* dLow, double* dUpp, double* x0 ) {

//double H[29241];
//double g[1221];
//double C[22050];
//double c[1050];
//double dLow[1221];
//double dUpp[1221];
//double x0[21];

x0[0] = 0.000000000000e+00;
x0[1] = 0.000000000000e+00;
x0[2] = 0.000000000000e+00;
x0[3] = 0.000000000000e+00;
x0[4] = 0.000000000000e+00;
x0[5] = 0.000000000000e+00;
x0[6] = 0.000000000000e+00;
x0[7] = 0.000000000000e+00;
x0[8] = 0.000000000000e+00;
x0[9] = 0.000000000000e+00;
x0[10] = 0.000000000000e+00;
x0[11] = 0.000000000000e+00;
x0[12] = 0.000000000000e+00;
x0[13] = 0.000000000000e+00;
x0[14] = 0.000000000000e+00;
x0[15] = 0.000000000000e+00;
x0[16] = 0.000000000000e+00;
x0[17] = 0.000000000000e+00;
x0[18] = 0.000000000000e+00;
x0[19] = 0.000000000000e+00;
x0[20] = 0.000000000000e+00;
H[0] = 1.000000000000e+01;
H[25] = 1.000000000000e+01;
H[50] = 1.000000000000e+01;
H[75] = 1.000000000000e+01;
H[100] = 1.000000000000e+01;
H[125] = 1.000000000000e+01;
H[150] = 1.000000000000e+01;
H[175] = 1.000000000000e+01;
H[200] = 1.000000000000e+01;
H[225] = 1.000000000000e+00;
H[250] = 1.000000000000e+00;
H[275] = 1.000000000000e+00;
H[300] = 1.000000000000e+00;
H[325] = 1.000000000000e+00;
H[350] = 1.000000000000e+00;
H[375] = 1.000000000000e+00;
H[400] = 1.000000000000e+00;
H[425] = 1.000000000000e+00;
H[450] = 2.500000000000e+01;
H[475] = 2.500000000000e+01;
H[500] = 2.500000000000e+01;
H[525] = 1.000000000000e-02;
H[550] = 1.000000000000e-02;
H[575] = 1.000000000000e-02;
H[576] = 1.000000000000e+01;
H[601] = 1.000000000000e+01;
H[626] = 1.000000000000e+01;
H[651] = 1.000000000000e+01;
H[676] = 1.000000000000e+01;
H[701] = 1.000000000000e+01;
H[726] = 1.000000000000e+01;
H[751] = 1.000000000000e+01;
H[776] = 1.000000000000e+01;
H[801] = 1.000000000000e+00;
H[826] = 1.000000000000e+00;
H[851] = 1.000000000000e+00;
H[876] = 1.000000000000e+00;
H[901] = 1.000000000000e+00;
H[926] = 1.000000000000e+00;
H[951] = 1.000000000000e+00;
H[976] = 1.000000000000e+00;
H[1001] = 1.000000000000e+00;
H[1026] = 2.500000000000e+01;
H[1051] = 2.500000000000e+01;
H[1076] = 2.500000000000e+01;
H[1101] = 1.000000000000e-02;
H[1126] = 1.000000000000e-02;
H[1151] = 1.000000000000e-02;
H[1152] = 1.000000000000e+01;
H[1177] = 1.000000000000e+01;
H[1202] = 1.000000000000e+01;
H[1227] = 1.000000000000e+01;
H[1252] = 1.000000000000e+01;
H[1277] = 1.000000000000e+01;
H[1302] = 1.000000000000e+01;
H[1327] = 1.000000000000e+01;
H[1352] = 1.000000000000e+01;
H[1377] = 1.000000000000e+00;
H[1402] = 1.000000000000e+00;
H[1427] = 1.000000000000e+00;
H[1452] = 1.000000000000e+00;
H[1477] = 1.000000000000e+00;
H[1502] = 1.000000000000e+00;
H[1527] = 1.000000000000e+00;
H[1552] = 1.000000000000e+00;
H[1577] = 1.000000000000e+00;
H[1602] = 2.500000000000e+01;
H[1627] = 2.500000000000e+01;
H[1652] = 2.500000000000e+01;
H[1677] = 1.000000000000e-02;
H[1702] = 1.000000000000e-02;
H[1727] = 1.000000000000e-02;
H[1728] = 1.000000000000e+01;
H[1753] = 1.000000000000e+01;
H[1778] = 1.000000000000e+01;
H[1803] = 1.000000000000e+01;
H[1828] = 1.000000000000e+01;
H[1853] = 1.000000000000e+01;
H[1878] = 1.000000000000e+01;
H[1903] = 1.000000000000e+01;
H[1928] = 1.000000000000e+01;
H[1953] = 1.000000000000e+00;
H[1978] = 1.000000000000e+00;
H[2003] = 1.000000000000e+00;
H[2028] = 1.000000000000e+00;
H[2053] = 1.000000000000e+00;
H[2078] = 1.000000000000e+00;
H[2103] = 1.000000000000e+00;
H[2128] = 1.000000000000e+00;
H[2153] = 1.000000000000e+00;
H[2178] = 2.500000000000e+01;
H[2203] = 2.500000000000e+01;
H[2228] = 2.500000000000e+01;
H[2253] = 1.000000000000e-02;
H[2278] = 1.000000000000e-02;
H[2303] = 1.000000000000e-02;
H[2304] = 1.000000000000e+01;
H[2329] = 1.000000000000e+01;
H[2354] = 1.000000000000e+01;
H[2379] = 1.000000000000e+01;
H[2404] = 1.000000000000e+01;
H[2429] = 1.000000000000e+01;
H[2454] = 1.000000000000e+01;
H[2479] = 1.000000000000e+01;
H[2504] = 1.000000000000e+01;
H[2529] = 1.000000000000e+00;
H[2554] = 1.000000000000e+00;
H[2579] = 1.000000000000e+00;
H[2604] = 1.000000000000e+00;
H[2629] = 1.000000000000e+00;
H[2654] = 1.000000000000e+00;
H[2679] = 1.000000000000e+00;
H[2704] = 1.000000000000e+00;
H[2729] = 1.000000000000e+00;
H[2754] = 2.500000000000e+01;
H[2779] = 2.500000000000e+01;
H[2804] = 2.500000000000e+01;
H[2829] = 1.000000000000e-02;
H[2854] = 1.000000000000e-02;
H[2879] = 1.000000000000e-02;
H[2880] = 1.000000000000e+01;
H[2905] = 1.000000000000e+01;
H[2930] = 1.000000000000e+01;
H[2955] = 1.000000000000e+01;
H[2980] = 1.000000000000e+01;
H[3005] = 1.000000000000e+01;
H[3030] = 1.000000000000e+01;
H[3055] = 1.000000000000e+01;
H[3080] = 1.000000000000e+01;
H[3105] = 1.000000000000e+00;
H[3130] = 1.000000000000e+00;
H[3155] = 1.000000000000e+00;
H[3180] = 1.000000000000e+00;
H[3205] = 1.000000000000e+00;
H[3230] = 1.000000000000e+00;
H[3255] = 1.000000000000e+00;
H[3280] = 1.000000000000e+00;
H[3305] = 1.000000000000e+00;
H[3330] = 2.500000000000e+01;
H[3355] = 2.500000000000e+01;
H[3380] = 2.500000000000e+01;
H[3405] = 1.000000000000e-02;
H[3430] = 1.000000000000e-02;
H[3455] = 1.000000000000e-02;
H[3456] = 1.000000000000e+01;
H[3481] = 1.000000000000e+01;
H[3506] = 1.000000000000e+01;
H[3531] = 1.000000000000e+01;
H[3556] = 1.000000000000e+01;
H[3581] = 1.000000000000e+01;
H[3606] = 1.000000000000e+01;
H[3631] = 1.000000000000e+01;
H[3656] = 1.000000000000e+01;
H[3681] = 1.000000000000e+00;
H[3706] = 1.000000000000e+00;
H[3731] = 1.000000000000e+00;
H[3756] = 1.000000000000e+00;
H[3781] = 1.000000000000e+00;
H[3806] = 1.000000000000e+00;
H[3831] = 1.000000000000e+00;
H[3856] = 1.000000000000e+00;
H[3881] = 1.000000000000e+00;
H[3906] = 2.500000000000e+01;
H[3931] = 2.500000000000e+01;
H[3956] = 2.500000000000e+01;
H[3981] = 1.000000000000e-02;
H[4006] = 1.000000000000e-02;
H[4031] = 1.000000000000e-02;
H[4032] = 1.000000000000e+01;
H[4057] = 1.000000000000e+01;
H[4082] = 1.000000000000e+01;
H[4107] = 1.000000000000e+01;
H[4132] = 1.000000000000e+01;
H[4157] = 1.000000000000e+01;
H[4182] = 1.000000000000e+01;
H[4207] = 1.000000000000e+01;
H[4232] = 1.000000000000e+01;
H[4257] = 1.000000000000e+00;
H[4282] = 1.000000000000e+00;
H[4307] = 1.000000000000e+00;
H[4332] = 1.000000000000e+00;
H[4357] = 1.000000000000e+00;
H[4382] = 1.000000000000e+00;
H[4407] = 1.000000000000e+00;
H[4432] = 1.000000000000e+00;
H[4457] = 1.000000000000e+00;
H[4482] = 2.500000000000e+01;
H[4507] = 2.500000000000e+01;
H[4532] = 2.500000000000e+01;
H[4557] = 1.000000000000e-02;
H[4582] = 1.000000000000e-02;
H[4607] = 1.000000000000e-02;
H[4608] = 1.000000000000e+01;
H[4633] = 1.000000000000e+01;
H[4658] = 1.000000000000e+01;
H[4683] = 1.000000000000e+01;
H[4708] = 1.000000000000e+01;
H[4733] = 1.000000000000e+01;
H[4758] = 1.000000000000e+01;
H[4783] = 1.000000000000e+01;
H[4808] = 1.000000000000e+01;
H[4833] = 1.000000000000e+00;
H[4858] = 1.000000000000e+00;
H[4883] = 1.000000000000e+00;
H[4908] = 1.000000000000e+00;
H[4933] = 1.000000000000e+00;
H[4958] = 1.000000000000e+00;
H[4983] = 1.000000000000e+00;
H[5008] = 1.000000000000e+00;
H[5033] = 1.000000000000e+00;
H[5058] = 2.500000000000e+01;
H[5083] = 2.500000000000e+01;
H[5108] = 2.500000000000e+01;
H[5133] = 1.000000000000e-02;
H[5158] = 1.000000000000e-02;
H[5183] = 1.000000000000e-02;
H[5184] = 1.000000000000e+01;
H[5209] = 1.000000000000e+01;
H[5234] = 1.000000000000e+01;
H[5259] = 1.000000000000e+01;
H[5284] = 1.000000000000e+01;
H[5309] = 1.000000000000e+01;
H[5334] = 1.000000000000e+01;
H[5359] = 1.000000000000e+01;
H[5384] = 1.000000000000e+01;
H[5409] = 1.000000000000e+00;
H[5434] = 1.000000000000e+00;
H[5459] = 1.000000000000e+00;
H[5484] = 1.000000000000e+00;
H[5509] = 1.000000000000e+00;
H[5534] = 1.000000000000e+00;
H[5559] = 1.000000000000e+00;
H[5584] = 1.000000000000e+00;
H[5609] = 1.000000000000e+00;
H[5634] = 2.500000000000e+01;
H[5659] = 2.500000000000e+01;
H[5684] = 2.500000000000e+01;
H[5709] = 1.000000000000e-02;
H[5734] = 1.000000000000e-02;
H[5759] = 1.000000000000e-02;
H[5760] = 1.000000000000e+01;
H[5785] = 1.000000000000e+01;
H[5810] = 1.000000000000e+01;
H[5835] = 1.000000000000e+01;
H[5860] = 1.000000000000e+01;
H[5885] = 1.000000000000e+01;
H[5910] = 1.000000000000e+01;
H[5935] = 1.000000000000e+01;
H[5960] = 1.000000000000e+01;
H[5985] = 1.000000000000e+00;
H[6010] = 1.000000000000e+00;
H[6035] = 1.000000000000e+00;
H[6060] = 1.000000000000e+00;
H[6085] = 1.000000000000e+00;
H[6110] = 1.000000000000e+00;
H[6135] = 1.000000000000e+00;
H[6160] = 1.000000000000e+00;
H[6185] = 1.000000000000e+00;
H[6210] = 2.500000000000e+01;
H[6235] = 2.500000000000e+01;
H[6260] = 2.500000000000e+01;
H[6285] = 1.000000000000e-02;
H[6310] = 1.000000000000e-02;
H[6335] = 1.000000000000e-02;
H[6336] = 1.000000000000e+01;
H[6361] = 1.000000000000e+01;
H[6386] = 1.000000000000e+01;
H[6411] = 1.000000000000e+01;
H[6436] = 1.000000000000e+01;
H[6461] = 1.000000000000e+01;
H[6486] = 1.000000000000e+01;
H[6511] = 1.000000000000e+01;
H[6536] = 1.000000000000e+01;
H[6561] = 1.000000000000e+00;
H[6586] = 1.000000000000e+00;
H[6611] = 1.000000000000e+00;
H[6636] = 1.000000000000e+00;
H[6661] = 1.000000000000e+00;
H[6686] = 1.000000000000e+00;
H[6711] = 1.000000000000e+00;
H[6736] = 1.000000000000e+00;
H[6761] = 1.000000000000e+00;
H[6786] = 2.500000000000e+01;
H[6811] = 2.500000000000e+01;
H[6836] = 2.500000000000e+01;
H[6861] = 1.000000000000e-02;
H[6886] = 1.000000000000e-02;
H[6911] = 1.000000000000e-02;
H[6912] = 1.000000000000e+01;
H[6937] = 1.000000000000e+01;
H[6962] = 1.000000000000e+01;
H[6987] = 1.000000000000e+01;
H[7012] = 1.000000000000e+01;
H[7037] = 1.000000000000e+01;
H[7062] = 1.000000000000e+01;
H[7087] = 1.000000000000e+01;
H[7112] = 1.000000000000e+01;
H[7137] = 1.000000000000e+00;
H[7162] = 1.000000000000e+00;
H[7187] = 1.000000000000e+00;
H[7212] = 1.000000000000e+00;
H[7237] = 1.000000000000e+00;
H[7262] = 1.000000000000e+00;
H[7287] = 1.000000000000e+00;
H[7312] = 1.000000000000e+00;
H[7337] = 1.000000000000e+00;
H[7362] = 2.500000000000e+01;
H[7387] = 2.500000000000e+01;
H[7412] = 2.500000000000e+01;
H[7437] = 1.000000000000e-02;
H[7462] = 1.000000000000e-02;
H[7487] = 1.000000000000e-02;
H[7488] = 1.000000000000e+01;
H[7513] = 1.000000000000e+01;
H[7538] = 1.000000000000e+01;
H[7563] = 1.000000000000e+01;
H[7588] = 1.000000000000e+01;
H[7613] = 1.000000000000e+01;
H[7638] = 1.000000000000e+01;
H[7663] = 1.000000000000e+01;
H[7688] = 1.000000000000e+01;
H[7713] = 1.000000000000e+00;
H[7738] = 1.000000000000e+00;
H[7763] = 1.000000000000e+00;
H[7788] = 1.000000000000e+00;
H[7813] = 1.000000000000e+00;
H[7838] = 1.000000000000e+00;
H[7863] = 1.000000000000e+00;
H[7888] = 1.000000000000e+00;
H[7913] = 1.000000000000e+00;
H[7938] = 2.500000000000e+01;
H[7963] = 2.500000000000e+01;
H[7988] = 2.500000000000e+01;
H[8013] = 1.000000000000e-02;
H[8038] = 1.000000000000e-02;
H[8063] = 1.000000000000e-02;
H[8064] = 1.000000000000e+01;
H[8089] = 1.000000000000e+01;
H[8114] = 1.000000000000e+01;
H[8139] = 1.000000000000e+01;
H[8164] = 1.000000000000e+01;
H[8189] = 1.000000000000e+01;
H[8214] = 1.000000000000e+01;
H[8239] = 1.000000000000e+01;
H[8264] = 1.000000000000e+01;
H[8289] = 1.000000000000e+00;
H[8314] = 1.000000000000e+00;
H[8339] = 1.000000000000e+00;
H[8364] = 1.000000000000e+00;
H[8389] = 1.000000000000e+00;
H[8414] = 1.000000000000e+00;
H[8439] = 1.000000000000e+00;
H[8464] = 1.000000000000e+00;
H[8489] = 1.000000000000e+00;
H[8514] = 2.500000000000e+01;
H[8539] = 2.500000000000e+01;
H[8564] = 2.500000000000e+01;
H[8589] = 1.000000000000e-02;
H[8614] = 1.000000000000e-02;
H[8639] = 1.000000000000e-02;
H[8640] = 1.000000000000e+01;
H[8665] = 1.000000000000e+01;
H[8690] = 1.000000000000e+01;
H[8715] = 1.000000000000e+01;
H[8740] = 1.000000000000e+01;
H[8765] = 1.000000000000e+01;
H[8790] = 1.000000000000e+01;
H[8815] = 1.000000000000e+01;
H[8840] = 1.000000000000e+01;
H[8865] = 1.000000000000e+00;
H[8890] = 1.000000000000e+00;
H[8915] = 1.000000000000e+00;
H[8940] = 1.000000000000e+00;
H[8965] = 1.000000000000e+00;
H[8990] = 1.000000000000e+00;
H[9015] = 1.000000000000e+00;
H[9040] = 1.000000000000e+00;
H[9065] = 1.000000000000e+00;
H[9090] = 2.500000000000e+01;
H[9115] = 2.500000000000e+01;
H[9140] = 2.500000000000e+01;
H[9165] = 1.000000000000e-02;
H[9190] = 1.000000000000e-02;
H[9215] = 1.000000000000e-02;
H[9216] = 1.000000000000e+01;
H[9241] = 1.000000000000e+01;
H[9266] = 1.000000000000e+01;
H[9291] = 1.000000000000e+01;
H[9316] = 1.000000000000e+01;
H[9341] = 1.000000000000e+01;
H[9366] = 1.000000000000e+01;
H[9391] = 1.000000000000e+01;
H[9416] = 1.000000000000e+01;
H[9441] = 1.000000000000e+00;
H[9466] = 1.000000000000e+00;
H[9491] = 1.000000000000e+00;
H[9516] = 1.000000000000e+00;
H[9541] = 1.000000000000e+00;
H[9566] = 1.000000000000e+00;
H[9591] = 1.000000000000e+00;
H[9616] = 1.000000000000e+00;
H[9641] = 1.000000000000e+00;
H[9666] = 2.500000000000e+01;
H[9691] = 2.500000000000e+01;
H[9716] = 2.500000000000e+01;
H[9741] = 1.000000000000e-02;
H[9766] = 1.000000000000e-02;
H[9791] = 1.000000000000e-02;
H[9792] = 1.000000000000e+01;
H[9817] = 1.000000000000e+01;
H[9842] = 1.000000000000e+01;
H[9867] = 1.000000000000e+01;
H[9892] = 1.000000000000e+01;
H[9917] = 1.000000000000e+01;
H[9942] = 1.000000000000e+01;
H[9967] = 1.000000000000e+01;
H[9992] = 1.000000000000e+01;
H[10017] = 1.000000000000e+00;
H[10042] = 1.000000000000e+00;
H[10067] = 1.000000000000e+00;
H[10092] = 1.000000000000e+00;
H[10117] = 1.000000000000e+00;
H[10142] = 1.000000000000e+00;
H[10167] = 1.000000000000e+00;
H[10192] = 1.000000000000e+00;
H[10217] = 1.000000000000e+00;
H[10242] = 2.500000000000e+01;
H[10267] = 2.500000000000e+01;
H[10292] = 2.500000000000e+01;
H[10317] = 1.000000000000e-02;
H[10342] = 1.000000000000e-02;
H[10367] = 1.000000000000e-02;
H[10368] = 1.000000000000e+01;
H[10393] = 1.000000000000e+01;
H[10418] = 1.000000000000e+01;
H[10443] = 1.000000000000e+01;
H[10468] = 1.000000000000e+01;
H[10493] = 1.000000000000e+01;
H[10518] = 1.000000000000e+01;
H[10543] = 1.000000000000e+01;
H[10568] = 1.000000000000e+01;
H[10593] = 1.000000000000e+00;
H[10618] = 1.000000000000e+00;
H[10643] = 1.000000000000e+00;
H[10668] = 1.000000000000e+00;
H[10693] = 1.000000000000e+00;
H[10718] = 1.000000000000e+00;
H[10743] = 1.000000000000e+00;
H[10768] = 1.000000000000e+00;
H[10793] = 1.000000000000e+00;
H[10818] = 2.500000000000e+01;
H[10843] = 2.500000000000e+01;
H[10868] = 2.500000000000e+01;
H[10893] = 1.000000000000e-02;
H[10918] = 1.000000000000e-02;
H[10943] = 1.000000000000e-02;
H[10944] = 1.000000000000e+01;
H[10969] = 1.000000000000e+01;
H[10994] = 1.000000000000e+01;
H[11019] = 1.000000000000e+01;
H[11044] = 1.000000000000e+01;
H[11069] = 1.000000000000e+01;
H[11094] = 1.000000000000e+01;
H[11119] = 1.000000000000e+01;
H[11144] = 1.000000000000e+01;
H[11169] = 1.000000000000e+00;
H[11194] = 1.000000000000e+00;
H[11219] = 1.000000000000e+00;
H[11244] = 1.000000000000e+00;
H[11269] = 1.000000000000e+00;
H[11294] = 1.000000000000e+00;
H[11319] = 1.000000000000e+00;
H[11344] = 1.000000000000e+00;
H[11369] = 1.000000000000e+00;
H[11394] = 2.500000000000e+01;
H[11419] = 2.500000000000e+01;
H[11444] = 2.500000000000e+01;
H[11469] = 1.000000000000e-02;
H[11494] = 1.000000000000e-02;
H[11519] = 1.000000000000e-02;
H[11520] = 1.000000000000e+01;
H[11545] = 1.000000000000e+01;
H[11570] = 1.000000000000e+01;
H[11595] = 1.000000000000e+01;
H[11620] = 1.000000000000e+01;
H[11645] = 1.000000000000e+01;
H[11670] = 1.000000000000e+01;
H[11695] = 1.000000000000e+01;
H[11720] = 1.000000000000e+01;
H[11745] = 1.000000000000e+00;
H[11770] = 1.000000000000e+00;
H[11795] = 1.000000000000e+00;
H[11820] = 1.000000000000e+00;
H[11845] = 1.000000000000e+00;
H[11870] = 1.000000000000e+00;
H[11895] = 1.000000000000e+00;
H[11920] = 1.000000000000e+00;
H[11945] = 1.000000000000e+00;
H[11970] = 2.500000000000e+01;
H[11995] = 2.500000000000e+01;
H[12020] = 2.500000000000e+01;
H[12045] = 1.000000000000e-02;
H[12070] = 1.000000000000e-02;
H[12095] = 1.000000000000e-02;
H[12096] = 1.000000000000e+01;
H[12121] = 1.000000000000e+01;
H[12146] = 1.000000000000e+01;
H[12171] = 1.000000000000e+01;
H[12196] = 1.000000000000e+01;
H[12221] = 1.000000000000e+01;
H[12246] = 1.000000000000e+01;
H[12271] = 1.000000000000e+01;
H[12296] = 1.000000000000e+01;
H[12321] = 1.000000000000e+00;
H[12346] = 1.000000000000e+00;
H[12371] = 1.000000000000e+00;
H[12396] = 1.000000000000e+00;
H[12421] = 1.000000000000e+00;
H[12446] = 1.000000000000e+00;
H[12471] = 1.000000000000e+00;
H[12496] = 1.000000000000e+00;
H[12521] = 1.000000000000e+00;
H[12546] = 2.500000000000e+01;
H[12571] = 2.500000000000e+01;
H[12596] = 2.500000000000e+01;
H[12621] = 1.000000000000e-02;
H[12646] = 1.000000000000e-02;
H[12671] = 1.000000000000e-02;
H[12672] = 1.000000000000e+01;
H[12697] = 1.000000000000e+01;
H[12722] = 1.000000000000e+01;
H[12747] = 1.000000000000e+01;
H[12772] = 1.000000000000e+01;
H[12797] = 1.000000000000e+01;
H[12822] = 1.000000000000e+01;
H[12847] = 1.000000000000e+01;
H[12872] = 1.000000000000e+01;
H[12897] = 1.000000000000e+00;
H[12922] = 1.000000000000e+00;
H[12947] = 1.000000000000e+00;
H[12972] = 1.000000000000e+00;
H[12997] = 1.000000000000e+00;
H[13022] = 1.000000000000e+00;
H[13047] = 1.000000000000e+00;
H[13072] = 1.000000000000e+00;
H[13097] = 1.000000000000e+00;
H[13122] = 2.500000000000e+01;
H[13147] = 2.500000000000e+01;
H[13172] = 2.500000000000e+01;
H[13197] = 1.000000000000e-02;
H[13222] = 1.000000000000e-02;
H[13247] = 1.000000000000e-02;
H[13248] = 1.000000000000e+01;
H[13273] = 1.000000000000e+01;
H[13298] = 1.000000000000e+01;
H[13323] = 1.000000000000e+01;
H[13348] = 1.000000000000e+01;
H[13373] = 1.000000000000e+01;
H[13398] = 1.000000000000e+01;
H[13423] = 1.000000000000e+01;
H[13448] = 1.000000000000e+01;
H[13473] = 1.000000000000e+00;
H[13498] = 1.000000000000e+00;
H[13523] = 1.000000000000e+00;
H[13548] = 1.000000000000e+00;
H[13573] = 1.000000000000e+00;
H[13598] = 1.000000000000e+00;
H[13623] = 1.000000000000e+00;
H[13648] = 1.000000000000e+00;
H[13673] = 1.000000000000e+00;
H[13698] = 2.500000000000e+01;
H[13723] = 2.500000000000e+01;
H[13748] = 2.500000000000e+01;
H[13773] = 1.000000000000e-02;
H[13798] = 1.000000000000e-02;
H[13823] = 1.000000000000e-02;
H[13824] = 1.000000000000e+01;
H[13849] = 1.000000000000e+01;
H[13874] = 1.000000000000e+01;
H[13899] = 1.000000000000e+01;
H[13924] = 1.000000000000e+01;
H[13949] = 1.000000000000e+01;
H[13974] = 1.000000000000e+01;
H[13999] = 1.000000000000e+01;
H[14024] = 1.000000000000e+01;
H[14049] = 1.000000000000e+00;
H[14074] = 1.000000000000e+00;
H[14099] = 1.000000000000e+00;
H[14124] = 1.000000000000e+00;
H[14149] = 1.000000000000e+00;
H[14174] = 1.000000000000e+00;
H[14199] = 1.000000000000e+00;
H[14224] = 1.000000000000e+00;
H[14249] = 1.000000000000e+00;
H[14274] = 2.500000000000e+01;
H[14299] = 2.500000000000e+01;
H[14324] = 2.500000000000e+01;
H[14349] = 1.000000000000e-02;
H[14374] = 1.000000000000e-02;
H[14399] = 1.000000000000e-02;
H[14400] = 1.000000000000e+01;
H[14425] = 1.000000000000e+01;
H[14450] = 1.000000000000e+01;
H[14475] = 1.000000000000e+01;
H[14500] = 1.000000000000e+01;
H[14525] = 1.000000000000e+01;
H[14550] = 1.000000000000e+01;
H[14575] = 1.000000000000e+01;
H[14600] = 1.000000000000e+01;
H[14625] = 1.000000000000e+00;
H[14650] = 1.000000000000e+00;
H[14675] = 1.000000000000e+00;
H[14700] = 1.000000000000e+00;
H[14725] = 1.000000000000e+00;
H[14750] = 1.000000000000e+00;
H[14775] = 1.000000000000e+00;
H[14800] = 1.000000000000e+00;
H[14825] = 1.000000000000e+00;
H[14850] = 2.500000000000e+01;
H[14875] = 2.500000000000e+01;
H[14900] = 2.500000000000e+01;
H[14925] = 1.000000000000e-02;
H[14950] = 1.000000000000e-02;
H[14975] = 1.000000000000e-02;
H[14976] = 1.000000000000e+01;
H[15001] = 1.000000000000e+01;
H[15026] = 1.000000000000e+01;
H[15051] = 1.000000000000e+01;
H[15076] = 1.000000000000e+01;
H[15101] = 1.000000000000e+01;
H[15126] = 1.000000000000e+01;
H[15151] = 1.000000000000e+01;
H[15176] = 1.000000000000e+01;
H[15201] = 1.000000000000e+00;
H[15226] = 1.000000000000e+00;
H[15251] = 1.000000000000e+00;
H[15276] = 1.000000000000e+00;
H[15301] = 1.000000000000e+00;
H[15326] = 1.000000000000e+00;
H[15351] = 1.000000000000e+00;
H[15376] = 1.000000000000e+00;
H[15401] = 1.000000000000e+00;
H[15426] = 2.500000000000e+01;
H[15451] = 2.500000000000e+01;
H[15476] = 2.500000000000e+01;
H[15501] = 1.000000000000e-02;
H[15526] = 1.000000000000e-02;
H[15551] = 1.000000000000e-02;
H[15552] = 1.000000000000e+01;
H[15577] = 1.000000000000e+01;
H[15602] = 1.000000000000e+01;
H[15627] = 1.000000000000e+01;
H[15652] = 1.000000000000e+01;
H[15677] = 1.000000000000e+01;
H[15702] = 1.000000000000e+01;
H[15727] = 1.000000000000e+01;
H[15752] = 1.000000000000e+01;
H[15777] = 1.000000000000e+00;
H[15802] = 1.000000000000e+00;
H[15827] = 1.000000000000e+00;
H[15852] = 1.000000000000e+00;
H[15877] = 1.000000000000e+00;
H[15902] = 1.000000000000e+00;
H[15927] = 1.000000000000e+00;
H[15952] = 1.000000000000e+00;
H[15977] = 1.000000000000e+00;
H[16002] = 2.500000000000e+01;
H[16027] = 2.500000000000e+01;
H[16052] = 2.500000000000e+01;
H[16077] = 1.000000000000e-02;
H[16102] = 1.000000000000e-02;
H[16127] = 1.000000000000e-02;
H[16128] = 1.000000000000e+01;
H[16153] = 1.000000000000e+01;
H[16178] = 1.000000000000e+01;
H[16203] = 1.000000000000e+01;
H[16228] = 1.000000000000e+01;
H[16253] = 1.000000000000e+01;
H[16278] = 1.000000000000e+01;
H[16303] = 1.000000000000e+01;
H[16328] = 1.000000000000e+01;
H[16353] = 1.000000000000e+00;
H[16378] = 1.000000000000e+00;
H[16403] = 1.000000000000e+00;
H[16428] = 1.000000000000e+00;
H[16453] = 1.000000000000e+00;
H[16478] = 1.000000000000e+00;
H[16503] = 1.000000000000e+00;
H[16528] = 1.000000000000e+00;
H[16553] = 1.000000000000e+00;
H[16578] = 2.500000000000e+01;
H[16603] = 2.500000000000e+01;
H[16628] = 2.500000000000e+01;
H[16653] = 1.000000000000e-02;
H[16678] = 1.000000000000e-02;
H[16703] = 1.000000000000e-02;
H[16704] = 1.000000000000e+01;
H[16729] = 1.000000000000e+01;
H[16754] = 1.000000000000e+01;
H[16779] = 1.000000000000e+01;
H[16804] = 1.000000000000e+01;
H[16829] = 1.000000000000e+01;
H[16854] = 1.000000000000e+01;
H[16879] = 1.000000000000e+01;
H[16904] = 1.000000000000e+01;
H[16929] = 1.000000000000e+00;
H[16954] = 1.000000000000e+00;
H[16979] = 1.000000000000e+00;
H[17004] = 1.000000000000e+00;
H[17029] = 1.000000000000e+00;
H[17054] = 1.000000000000e+00;
H[17079] = 1.000000000000e+00;
H[17104] = 1.000000000000e+00;
H[17129] = 1.000000000000e+00;
H[17154] = 2.500000000000e+01;
H[17179] = 2.500000000000e+01;
H[17204] = 2.500000000000e+01;
H[17229] = 1.000000000000e-02;
H[17254] = 1.000000000000e-02;
H[17279] = 1.000000000000e-02;
H[17280] = 1.000000000000e+01;
H[17305] = 1.000000000000e+01;
H[17330] = 1.000000000000e+01;
H[17355] = 1.000000000000e+01;
H[17380] = 1.000000000000e+01;
H[17405] = 1.000000000000e+01;
H[17430] = 1.000000000000e+01;
H[17455] = 1.000000000000e+01;
H[17480] = 1.000000000000e+01;
H[17505] = 1.000000000000e+00;
H[17530] = 1.000000000000e+00;
H[17555] = 1.000000000000e+00;
H[17580] = 1.000000000000e+00;
H[17605] = 1.000000000000e+00;
H[17630] = 1.000000000000e+00;
H[17655] = 1.000000000000e+00;
H[17680] = 1.000000000000e+00;
H[17705] = 1.000000000000e+00;
H[17730] = 2.500000000000e+01;
H[17755] = 2.500000000000e+01;
H[17780] = 2.500000000000e+01;
H[17805] = 1.000000000000e-02;
H[17830] = 1.000000000000e-02;
H[17855] = 1.000000000000e-02;
H[17856] = 1.000000000000e+01;
H[17881] = 1.000000000000e+01;
H[17906] = 1.000000000000e+01;
H[17931] = 1.000000000000e+01;
H[17956] = 1.000000000000e+01;
H[17981] = 1.000000000000e+01;
H[18006] = 1.000000000000e+01;
H[18031] = 1.000000000000e+01;
H[18056] = 1.000000000000e+01;
H[18081] = 1.000000000000e+00;
H[18106] = 1.000000000000e+00;
H[18131] = 1.000000000000e+00;
H[18156] = 1.000000000000e+00;
H[18181] = 1.000000000000e+00;
H[18206] = 1.000000000000e+00;
H[18231] = 1.000000000000e+00;
H[18256] = 1.000000000000e+00;
H[18281] = 1.000000000000e+00;
H[18306] = 2.500000000000e+01;
H[18331] = 2.500000000000e+01;
H[18356] = 2.500000000000e+01;
H[18381] = 1.000000000000e-02;
H[18406] = 1.000000000000e-02;
H[18431] = 1.000000000000e-02;
H[18432] = 1.000000000000e+01;
H[18457] = 1.000000000000e+01;
H[18482] = 1.000000000000e+01;
H[18507] = 1.000000000000e+01;
H[18532] = 1.000000000000e+01;
H[18557] = 1.000000000000e+01;
H[18582] = 1.000000000000e+01;
H[18607] = 1.000000000000e+01;
H[18632] = 1.000000000000e+01;
H[18657] = 1.000000000000e+00;
H[18682] = 1.000000000000e+00;
H[18707] = 1.000000000000e+00;
H[18732] = 1.000000000000e+00;
H[18757] = 1.000000000000e+00;
H[18782] = 1.000000000000e+00;
H[18807] = 1.000000000000e+00;
H[18832] = 1.000000000000e+00;
H[18857] = 1.000000000000e+00;
H[18882] = 2.500000000000e+01;
H[18907] = 2.500000000000e+01;
H[18932] = 2.500000000000e+01;
H[18957] = 1.000000000000e-02;
H[18982] = 1.000000000000e-02;
H[19007] = 1.000000000000e-02;
H[19008] = 1.000000000000e+01;
H[19033] = 1.000000000000e+01;
H[19058] = 1.000000000000e+01;
H[19083] = 1.000000000000e+01;
H[19108] = 1.000000000000e+01;
H[19133] = 1.000000000000e+01;
H[19158] = 1.000000000000e+01;
H[19183] = 1.000000000000e+01;
H[19208] = 1.000000000000e+01;
H[19233] = 1.000000000000e+00;
H[19258] = 1.000000000000e+00;
H[19283] = 1.000000000000e+00;
H[19308] = 1.000000000000e+00;
H[19333] = 1.000000000000e+00;
H[19358] = 1.000000000000e+00;
H[19383] = 1.000000000000e+00;
H[19408] = 1.000000000000e+00;
H[19433] = 1.000000000000e+00;
H[19458] = 2.500000000000e+01;
H[19483] = 2.500000000000e+01;
H[19508] = 2.500000000000e+01;
H[19533] = 1.000000000000e-02;
H[19558] = 1.000000000000e-02;
H[19583] = 1.000000000000e-02;
H[19584] = 1.000000000000e+01;
H[19609] = 1.000000000000e+01;
H[19634] = 1.000000000000e+01;
H[19659] = 1.000000000000e+01;
H[19684] = 1.000000000000e+01;
H[19709] = 1.000000000000e+01;
H[19734] = 1.000000000000e+01;
H[19759] = 1.000000000000e+01;
H[19784] = 1.000000000000e+01;
H[19809] = 1.000000000000e+00;
H[19834] = 1.000000000000e+00;
H[19859] = 1.000000000000e+00;
H[19884] = 1.000000000000e+00;
H[19909] = 1.000000000000e+00;
H[19934] = 1.000000000000e+00;
H[19959] = 1.000000000000e+00;
H[19984] = 1.000000000000e+00;
H[20009] = 1.000000000000e+00;
H[20034] = 2.500000000000e+01;
H[20059] = 2.500000000000e+01;
H[20084] = 2.500000000000e+01;
H[20109] = 1.000000000000e-02;
H[20134] = 1.000000000000e-02;
H[20159] = 1.000000000000e-02;
H[20160] = 1.000000000000e+01;
H[20185] = 1.000000000000e+01;
H[20210] = 1.000000000000e+01;
H[20235] = 1.000000000000e+01;
H[20260] = 1.000000000000e+01;
H[20285] = 1.000000000000e+01;
H[20310] = 1.000000000000e+01;
H[20335] = 1.000000000000e+01;
H[20360] = 1.000000000000e+01;
H[20385] = 1.000000000000e+00;
H[20410] = 1.000000000000e+00;
H[20435] = 1.000000000000e+00;
H[20460] = 1.000000000000e+00;
H[20485] = 1.000000000000e+00;
H[20510] = 1.000000000000e+00;
H[20535] = 1.000000000000e+00;
H[20560] = 1.000000000000e+00;
H[20585] = 1.000000000000e+00;
H[20610] = 2.500000000000e+01;
H[20635] = 2.500000000000e+01;
H[20660] = 2.500000000000e+01;
H[20685] = 1.000000000000e-02;
H[20710] = 1.000000000000e-02;
H[20735] = 1.000000000000e-02;
H[20736] = 1.000000000000e+01;
H[20761] = 1.000000000000e+01;
H[20786] = 1.000000000000e+01;
H[20811] = 1.000000000000e+01;
H[20836] = 1.000000000000e+01;
H[20861] = 1.000000000000e+01;
H[20886] = 1.000000000000e+01;
H[20911] = 1.000000000000e+01;
H[20936] = 1.000000000000e+01;
H[20961] = 1.000000000000e+00;
H[20986] = 1.000000000000e+00;
H[21011] = 1.000000000000e+00;
H[21036] = 1.000000000000e+00;
H[21061] = 1.000000000000e+00;
H[21086] = 1.000000000000e+00;
H[21111] = 1.000000000000e+00;
H[21136] = 1.000000000000e+00;
H[21161] = 1.000000000000e+00;
H[21186] = 2.500000000000e+01;
H[21211] = 2.500000000000e+01;
H[21236] = 2.500000000000e+01;
H[21261] = 1.000000000000e-02;
H[21286] = 1.000000000000e-02;
H[21311] = 1.000000000000e-02;
H[21312] = 1.000000000000e+01;
H[21337] = 1.000000000000e+01;
H[21362] = 1.000000000000e+01;
H[21387] = 1.000000000000e+01;
H[21412] = 1.000000000000e+01;
H[21437] = 1.000000000000e+01;
H[21462] = 1.000000000000e+01;
H[21487] = 1.000000000000e+01;
H[21512] = 1.000000000000e+01;
H[21537] = 1.000000000000e+00;
H[21562] = 1.000000000000e+00;
H[21587] = 1.000000000000e+00;
H[21612] = 1.000000000000e+00;
H[21637] = 1.000000000000e+00;
H[21662] = 1.000000000000e+00;
H[21687] = 1.000000000000e+00;
H[21712] = 1.000000000000e+00;
H[21737] = 1.000000000000e+00;
H[21762] = 2.500000000000e+01;
H[21787] = 2.500000000000e+01;
H[21812] = 2.500000000000e+01;
H[21837] = 1.000000000000e-02;
H[21862] = 1.000000000000e-02;
H[21887] = 1.000000000000e-02;
H[21888] = 1.000000000000e+01;
H[21913] = 1.000000000000e+01;
H[21938] = 1.000000000000e+01;
H[21963] = 1.000000000000e+01;
H[21988] = 1.000000000000e+01;
H[22013] = 1.000000000000e+01;
H[22038] = 1.000000000000e+01;
H[22063] = 1.000000000000e+01;
H[22088] = 1.000000000000e+01;
H[22113] = 1.000000000000e+00;
H[22138] = 1.000000000000e+00;
H[22163] = 1.000000000000e+00;
H[22188] = 1.000000000000e+00;
H[22213] = 1.000000000000e+00;
H[22238] = 1.000000000000e+00;
H[22263] = 1.000000000000e+00;
H[22288] = 1.000000000000e+00;
H[22313] = 1.000000000000e+00;
H[22338] = 2.500000000000e+01;
H[22363] = 2.500000000000e+01;
H[22388] = 2.500000000000e+01;
H[22413] = 1.000000000000e-02;
H[22438] = 1.000000000000e-02;
H[22463] = 1.000000000000e-02;
H[22464] = 1.000000000000e+01;
H[22489] = 1.000000000000e+01;
H[22514] = 1.000000000000e+01;
H[22539] = 1.000000000000e+01;
H[22564] = 1.000000000000e+01;
H[22589] = 1.000000000000e+01;
H[22614] = 1.000000000000e+01;
H[22639] = 1.000000000000e+01;
H[22664] = 1.000000000000e+01;
H[22689] = 1.000000000000e+00;
H[22714] = 1.000000000000e+00;
H[22739] = 1.000000000000e+00;
H[22764] = 1.000000000000e+00;
H[22789] = 1.000000000000e+00;
H[22814] = 1.000000000000e+00;
H[22839] = 1.000000000000e+00;
H[22864] = 1.000000000000e+00;
H[22889] = 1.000000000000e+00;
H[22914] = 2.500000000000e+01;
H[22939] = 2.500000000000e+01;
H[22964] = 2.500000000000e+01;
H[22989] = 1.000000000000e-02;
H[23014] = 1.000000000000e-02;
H[23039] = 1.000000000000e-02;
H[23040] = 1.000000000000e+01;
H[23065] = 1.000000000000e+01;
H[23090] = 1.000000000000e+01;
H[23115] = 1.000000000000e+01;
H[23140] = 1.000000000000e+01;
H[23165] = 1.000000000000e+01;
H[23190] = 1.000000000000e+01;
H[23215] = 1.000000000000e+01;
H[23240] = 1.000000000000e+01;
H[23265] = 1.000000000000e+00;
H[23290] = 1.000000000000e+00;
H[23315] = 1.000000000000e+00;
H[23340] = 1.000000000000e+00;
H[23365] = 1.000000000000e+00;
H[23390] = 1.000000000000e+00;
H[23415] = 1.000000000000e+00;
H[23440] = 1.000000000000e+00;
H[23465] = 1.000000000000e+00;
H[23490] = 2.500000000000e+01;
H[23515] = 2.500000000000e+01;
H[23540] = 2.500000000000e+01;
H[23565] = 1.000000000000e-02;
H[23590] = 1.000000000000e-02;
H[23615] = 1.000000000000e-02;
H[23616] = 1.000000000000e+01;
H[23641] = 1.000000000000e+01;
H[23666] = 1.000000000000e+01;
H[23691] = 1.000000000000e+01;
H[23716] = 1.000000000000e+01;
H[23741] = 1.000000000000e+01;
H[23766] = 1.000000000000e+01;
H[23791] = 1.000000000000e+01;
H[23816] = 1.000000000000e+01;
H[23841] = 1.000000000000e+00;
H[23866] = 1.000000000000e+00;
H[23891] = 1.000000000000e+00;
H[23916] = 1.000000000000e+00;
H[23941] = 1.000000000000e+00;
H[23966] = 1.000000000000e+00;
H[23991] = 1.000000000000e+00;
H[24016] = 1.000000000000e+00;
H[24041] = 1.000000000000e+00;
H[24066] = 2.500000000000e+01;
H[24091] = 2.500000000000e+01;
H[24116] = 2.500000000000e+01;
H[24141] = 1.000000000000e-02;
H[24166] = 1.000000000000e-02;
H[24191] = 1.000000000000e-02;
H[24192] = 1.000000000000e+01;
H[24217] = 1.000000000000e+01;
H[24242] = 1.000000000000e+01;
H[24267] = 1.000000000000e+01;
H[24292] = 1.000000000000e+01;
H[24317] = 1.000000000000e+01;
H[24342] = 1.000000000000e+01;
H[24367] = 1.000000000000e+01;
H[24392] = 1.000000000000e+01;
H[24417] = 1.000000000000e+00;
H[24442] = 1.000000000000e+00;
H[24467] = 1.000000000000e+00;
H[24492] = 1.000000000000e+00;
H[24517] = 1.000000000000e+00;
H[24542] = 1.000000000000e+00;
H[24567] = 1.000000000000e+00;
H[24592] = 1.000000000000e+00;
H[24617] = 1.000000000000e+00;
H[24642] = 2.500000000000e+01;
H[24667] = 2.500000000000e+01;
H[24692] = 2.500000000000e+01;
H[24717] = 1.000000000000e-02;
H[24742] = 1.000000000000e-02;
H[24767] = 1.000000000000e-02;
H[24768] = 1.000000000000e+01;
H[24793] = 1.000000000000e+01;
H[24818] = 1.000000000000e+01;
H[24843] = 1.000000000000e+01;
H[24868] = 1.000000000000e+01;
H[24893] = 1.000000000000e+01;
H[24918] = 1.000000000000e+01;
H[24943] = 1.000000000000e+01;
H[24968] = 1.000000000000e+01;
H[24993] = 1.000000000000e+00;
H[25018] = 1.000000000000e+00;
H[25043] = 1.000000000000e+00;
H[25068] = 1.000000000000e+00;
H[25093] = 1.000000000000e+00;
H[25118] = 1.000000000000e+00;
H[25143] = 1.000000000000e+00;
H[25168] = 1.000000000000e+00;
H[25193] = 1.000000000000e+00;
H[25218] = 2.500000000000e+01;
H[25243] = 2.500000000000e+01;
H[25268] = 2.500000000000e+01;
H[25293] = 1.000000000000e-02;
H[25318] = 1.000000000000e-02;
H[25343] = 1.000000000000e-02;
H[25344] = 1.000000000000e+01;
H[25369] = 1.000000000000e+01;
H[25394] = 1.000000000000e+01;
H[25419] = 1.000000000000e+01;
H[25444] = 1.000000000000e+01;
H[25469] = 1.000000000000e+01;
H[25494] = 1.000000000000e+01;
H[25519] = 1.000000000000e+01;
H[25544] = 1.000000000000e+01;
H[25569] = 1.000000000000e+00;
H[25594] = 1.000000000000e+00;
H[25619] = 1.000000000000e+00;
H[25644] = 1.000000000000e+00;
H[25669] = 1.000000000000e+00;
H[25694] = 1.000000000000e+00;
H[25719] = 1.000000000000e+00;
H[25744] = 1.000000000000e+00;
H[25769] = 1.000000000000e+00;
H[25794] = 2.500000000000e+01;
H[25819] = 2.500000000000e+01;
H[25844] = 2.500000000000e+01;
H[25869] = 1.000000000000e-02;
H[25894] = 1.000000000000e-02;
H[25919] = 1.000000000000e-02;
H[25920] = 1.000000000000e+01;
H[25945] = 1.000000000000e+01;
H[25970] = 1.000000000000e+01;
H[25995] = 1.000000000000e+01;
H[26020] = 1.000000000000e+01;
H[26045] = 1.000000000000e+01;
H[26070] = 1.000000000000e+01;
H[26095] = 1.000000000000e+01;
H[26120] = 1.000000000000e+01;
H[26145] = 1.000000000000e+00;
H[26170] = 1.000000000000e+00;
H[26195] = 1.000000000000e+00;
H[26220] = 1.000000000000e+00;
H[26245] = 1.000000000000e+00;
H[26270] = 1.000000000000e+00;
H[26295] = 1.000000000000e+00;
H[26320] = 1.000000000000e+00;
H[26345] = 1.000000000000e+00;
H[26370] = 2.500000000000e+01;
H[26395] = 2.500000000000e+01;
H[26420] = 2.500000000000e+01;
H[26445] = 1.000000000000e-02;
H[26470] = 1.000000000000e-02;
H[26495] = 1.000000000000e-02;
H[26496] = 1.000000000000e+01;
H[26521] = 1.000000000000e+01;
H[26546] = 1.000000000000e+01;
H[26571] = 1.000000000000e+01;
H[26596] = 1.000000000000e+01;
H[26621] = 1.000000000000e+01;
H[26646] = 1.000000000000e+01;
H[26671] = 1.000000000000e+01;
H[26696] = 1.000000000000e+01;
H[26721] = 1.000000000000e+00;
H[26746] = 1.000000000000e+00;
H[26771] = 1.000000000000e+00;
H[26796] = 1.000000000000e+00;
H[26821] = 1.000000000000e+00;
H[26846] = 1.000000000000e+00;
H[26871] = 1.000000000000e+00;
H[26896] = 1.000000000000e+00;
H[26921] = 1.000000000000e+00;
H[26946] = 2.500000000000e+01;
H[26971] = 2.500000000000e+01;
H[26996] = 2.500000000000e+01;
H[27021] = 1.000000000000e-02;
H[27046] = 1.000000000000e-02;
H[27071] = 1.000000000000e-02;
H[27072] = 1.000000000000e+01;
H[27097] = 1.000000000000e+01;
H[27122] = 1.000000000000e+01;
H[27147] = 1.000000000000e+01;
H[27172] = 1.000000000000e+01;
H[27197] = 1.000000000000e+01;
H[27222] = 1.000000000000e+01;
H[27247] = 1.000000000000e+01;
H[27272] = 1.000000000000e+01;
H[27297] = 1.000000000000e+00;
H[27322] = 1.000000000000e+00;
H[27347] = 1.000000000000e+00;
H[27372] = 1.000000000000e+00;
H[27397] = 1.000000000000e+00;
H[27422] = 1.000000000000e+00;
H[27447] = 1.000000000000e+00;
H[27472] = 1.000000000000e+00;
H[27497] = 1.000000000000e+00;
H[27522] = 2.500000000000e+01;
H[27547] = 2.500000000000e+01;
H[27572] = 2.500000000000e+01;
H[27597] = 1.000000000000e-02;
H[27622] = 1.000000000000e-02;
H[27647] = 1.000000000000e-02;
H[27648] = 1.000000000000e+01;
H[27673] = 1.000000000000e+01;
H[27698] = 1.000000000000e+01;
H[27723] = 1.000000000000e+01;
H[27748] = 1.000000000000e+01;
H[27773] = 1.000000000000e+01;
H[27798] = 1.000000000000e+01;
H[27823] = 1.000000000000e+01;
H[27848] = 1.000000000000e+01;
H[27873] = 1.000000000000e+00;
H[27898] = 1.000000000000e+00;
H[27923] = 1.000000000000e+00;
H[27948] = 1.000000000000e+00;
H[27973] = 1.000000000000e+00;
H[27998] = 1.000000000000e+00;
H[28023] = 1.000000000000e+00;
H[28048] = 1.000000000000e+00;
H[28073] = 1.000000000000e+00;
H[28098] = 2.500000000000e+01;
H[28123] = 2.500000000000e+01;
H[28148] = 2.500000000000e+01;
H[28173] = 1.000000000000e-02;
H[28198] = 1.000000000000e-02;
H[28223] = 1.000000000000e-02;
H[28224] = 1.000000000000e+01;
H[28249] = 1.000000000000e+01;
H[28274] = 1.000000000000e+01;
H[28299] = 1.000000000000e+01;
H[28324] = 1.000000000000e+01;
H[28349] = 1.000000000000e+01;
H[28374] = 1.000000000000e+01;
H[28399] = 1.000000000000e+01;
H[28424] = 1.000000000000e+01;
H[28449] = 1.000000000000e+00;
H[28474] = 1.000000000000e+00;
H[28499] = 1.000000000000e+00;
H[28524] = 1.000000000000e+00;
H[28549] = 1.000000000000e+00;
H[28574] = 1.000000000000e+00;
H[28599] = 1.000000000000e+00;
H[28624] = 1.000000000000e+00;
H[28649] = 1.000000000000e+00;
H[28674] = 2.500000000000e+01;
H[28699] = 2.500000000000e+01;
H[28724] = 2.500000000000e+01;
H[28749] = 1.000000000000e-02;
H[28774] = 1.000000000000e-02;
H[28799] = 1.000000000000e-02;
H[28800] = 1.000000000000e+01;
H[28822] = 1.000000000000e+01;
H[28844] = 1.000000000000e+01;
H[28866] = 1.000000000000e+01;
H[28888] = 1.000000000000e+01;
H[28910] = 1.000000000000e+01;
H[28932] = 1.000000000000e+01;
H[28954] = 1.000000000000e+01;
H[28976] = 1.000000000000e+01;
H[28998] = 1.000000000000e+00;
H[29020] = 1.000000000000e+00;
H[29042] = 1.000000000000e+00;
H[29064] = 1.000000000000e+00;
H[29086] = 1.000000000000e+00;
H[29108] = 1.000000000000e+00;
H[29130] = 1.000000000000e+00;
H[29152] = 1.000000000000e+00;
H[29174] = 1.000000000000e+00;
H[29196] = 2.500000000000e+01;
H[29218] = 2.500000000000e+01;
H[29240] = 2.500000000000e+01;
g[0] = -3.630211624330e+00;
g[1] = 3.859849010350e+00;
g[2] = 3.723471502441e+00;
g[3] = -6.187215651625e+00;
g[4] = 6.286149249081e+00;
g[5] = 6.276883887492e+00;
g[6] = -8.138688838282e+00;
g[7] = 8.276744723341e+00;
g[8] = 8.299385791230e+00;
g[9] = -5.135719144816e-01;
g[10] = 5.458072534035e-01;
g[11] = 4.682888491301e-01;
g[12] = -6.608146406790e-01;
g[13] = 6.740599685878e-01;
g[14] = 6.282520023753e-01;
g[15] = -1.027945865057e+00;
g[16] = 1.010839910954e+00;
g[17] = 1.028964908821e+00;
g[18] = -2.499999999998e+01;
g[19] = 2.500000000000e+01;
g[20] = 2.500000000000e+01;
g[21] = 0.000000000000e+00;
g[22] = 0.000000000000e+00;
g[23] = 0.000000000000e+00;
g[24] = -3.897732071467e+00;
g[25] = 4.021858641525e+00;
g[26] = 3.867544997648e+00;
g[27] = -7.177719184606e+00;
g[28] = 7.402624587364e+00;
g[29] = 7.303001035391e+00;
g[30] = -9.661134878004e+00;
g[31] = 9.670690459981e+00;
g[32] = 9.658779945085e+00;
g[33] = 2.009760785474e-01;
g[34] = -2.934548502313e-01;
g[35] = -2.449185401864e-01;
g[36] = -3.111365788329e-01;
g[37] = 3.639185457782e-01;
g[38] = 3.411754405175e-01;
g[39] = -2.820290900372e-01;
g[40] = 2.057480136959e-01;
g[41] = 1.441070530980e-01;
g[42] = -2.499999999998e+01;
g[43] = 2.500000000000e+01;
g[44] = 2.500000000000e+01;
g[45] = 0.000000000000e+00;
g[46] = 0.000000000000e+00;
g[47] = 0.000000000000e+00;
g[48] = -3.324976069979e+00;
g[49] = 3.383250465375e+00;
g[50] = 3.320732801621e+00;
g[51] = -7.058605450005e+00;
g[52] = 7.154866340027e+00;
g[53] = 7.092251847558e+00;
g[54] = -9.052286796366e+00;
g[55] = 9.022002577152e+00;
g[56] = 8.875521277186e+00;
g[57] = 2.638915666757e-01;
g[58] = -2.306389204430e-01;
g[59] = -2.106853109952e-01;
g[60] = 5.279004614771e-01;
g[61] = -6.998924412679e-01;
g[62] = -6.524938868112e-01;
g[63] = 7.786662025730e-01;
g[64] = -7.424000060216e-01;
g[65] = -7.782947486217e-01;
g[66] = -2.499999999998e+01;
g[67] = 2.500000000000e+01;
g[68] = 2.500000000000e+01;
g[69] = 0.000000000000e+00;
g[70] = 0.000000000000e+00;
g[71] = 0.000000000000e+00;
g[72] = -2.839146539808e+00;
g[73] = 2.928021659371e+00;
g[74] = 2.852530896564e+00;
g[75] = -5.132625530195e+00;
g[76] = 4.969351753052e+00;
g[77] = 4.968052541436e+00;
g[78] = -7.301234111070e+00;
g[79] = 7.313186919335e+00;
g[80] = 7.259902661321e+00;
g[81] = 3.205439898946e-01;
g[82] = -3.736198596850e-01;
g[83] = -3.879652150722e-01;
g[84] = 1.232821871008e+00;
g[85] = -1.236577294497e+00;
g[86] = -1.238307203406e+00;
g[87] = 8.050546123188e-01;
g[88] = -8.346063240433e-01;
g[89] = -7.139883260030e-01;
g[90] = -2.499999999998e+01;
g[91] = 2.500000000000e+01;
g[92] = 2.500000000000e+01;
g[93] = 0.000000000000e+00;
g[94] = 0.000000000000e+00;
g[95] = 0.000000000000e+00;
g[96] = -1.716867301170e+00;
g[97] = 1.562154841108e+00;
g[98] = 1.506147842999e+00;
g[99] = -3.048003411594e+00;
g[100] = 3.176384833722e+00;
g[101] = 3.137592248092e+00;
g[102] = -6.189173265545e+00;
g[103] = 6.080010261015e+00;
g[104] = 6.246212625449e+00;
g[105] = 7.781583911296e-01;
g[106] = -8.929194800642e-01;
g[107] = -8.670684867205e-01;
g[108] = 6.613470339913e-01;
g[109] = -4.570155942885e-01;
g[110] = -4.567401762582e-01;
g[111] = 2.887215059961e-01;
g[112] = -3.261764585833e-01;
g[113] = -2.765410113218e-01;
g[114] = -2.499999999998e+01;
g[115] = 2.500000000000e+01;
g[116] = 2.500000000000e+01;
g[117] = 0.000000000000e+00;
g[118] = 0.000000000000e+00;
g[119] = 0.000000000000e+00;
g[120] = -3.726288018735e-01;
g[121] = 2.958270891046e-01;
g[122] = 3.067157037939e-01;
g[123] = -2.735254581451e+00;
g[124] = 2.994122245580e+00;
g[125] = 3.071628283900e+00;
g[126] = -6.105714730831e+00;
g[127] = 6.140348689535e+00;
g[128] = 6.219560620889e+00;
g[129] = 3.299944877970e-01;
g[130] = -1.296176812997e-01;
g[131] = -8.730431247729e-02;
g[132] = -2.783619345209e-01;
g[133] = 1.847259042950e-01;
g[134] = 2.841747992858e-01;
g[135] = -1.900484020515e-01;
g[136] = 3.540116046354e-01;
g[137] = 2.476438758663e-01;
g[138] = -2.499999999998e+01;
g[139] = 2.500000000000e+01;
g[140] = 2.500000000000e+01;
g[141] = 0.000000000000e+00;
g[142] = 0.000000000000e+00;
g[143] = 0.000000000000e+00;
g[144] = -1.025439287508e+00;
g[145] = 1.370744385981e+00;
g[146] = 1.473594927320e+00;
g[147] = -3.718230423611e+00;
g[148] = 3.795349806907e+00;
g[149] = 3.994959401842e+00;
g[150] = -6.835461030808e+00;
g[151] = 7.090465749857e+00;
g[152] = 7.045345170844e+00;
g[153] = -9.438500550931e-01;
g[154] = 1.064371281444e+00;
g[155] = 1.099982088467e+00;
g[156] = -6.456917688231e-01;
g[157] = 6.628397521951e-01;
g[158] = 6.544156009472e-01;
g[159] = -4.875516508107e-01;
g[160] = 4.770278656670e-01;
g[161] = 4.955748084095e-01;
g[162] = -2.499999999998e+01;
g[163] = 2.500000000000e+01;
g[164] = 2.500000000000e+01;
g[165] = 0.000000000000e+00;
g[166] = 0.000000000000e+00;
g[167] = 0.000000000000e+00;
g[168] = -3.254568694165e+00;
g[169] = 3.501310562953e+00;
g[170] = 3.572719385443e+00;
g[171] = -5.328478994098e+00;
g[172] = 5.726534265310e+00;
g[173] = 5.784497032984e+00;
g[174] = -7.812146898691e+00;
g[175] = 7.840651252917e+00;
g[176] = 7.953947523087e+00;
g[177] = -1.002750340022e+00;
g[178] = 8.296364720082e-01;
g[179] = 7.581186487182e-01;
g[180] = -9.566625067164e-01;
g[181] = 1.162262443969e+00;
g[182] = 1.065412580540e+00;
g[183] = -4.516126646195e-01;
g[184] = 3.198999106739e-01;
g[185] = 4.113351715949e-01;
g[186] = -2.499999999998e+01;
g[187] = 2.500000000000e+01;
g[188] = 2.500000000000e+01;
g[189] = 0.000000000000e+00;
g[190] = 0.000000000000e+00;
g[191] = 0.000000000000e+00;
g[192] = -4.251519193277e+00;
g[193] = 4.263772944133e+00;
g[194] = 4.146394032666e+00;
g[195] = -7.166570033272e+00;
g[196] = 7.548642470127e+00;
g[197] = 7.509694457683e+00;
g[198] = -8.646594933407e+00;
g[199] = 8.696978649239e+00;
g[200] = 8.830871270795e+00;
g[201] = 2.913229679389e-02;
g[202] = -3.292130111848e-02;
g[203] = -1.223491270502e-01;
g[204] = -6.998789463316e-01;
g[205] = 4.383348980876e-01;
g[206] = 4.455618801981e-01;
g[207] = -3.950999861909e-01;
g[208] = 5.461273562010e-01;
g[209] = 4.623245558323e-01;
g[210] = -2.499999999998e+01;
g[211] = 2.500000000000e+01;
g[212] = 2.500000000000e+01;
g[213] = 0.000000000000e+00;
g[214] = 0.000000000000e+00;
g[215] = 0.000000000000e+00;
g[216] = -3.632056062518e+00;
g[217] = 3.748651031161e+00;
g[218] = 3.569723472724e+00;
g[219] = -7.416317126570e+00;
g[220] = 7.177811641008e+00;
g[221] = 7.169932761302e+00;
g[222] = -9.271652914157e+00;
g[223] = 9.515050403425e+00;
g[224] = 9.395561891180e+00;
g[225] = 4.352165197363e-01;
g[226] = -4.067978901763e-01;
g[227] = -3.698391478653e-01;
g[228] = 5.048020205121e-01;
g[229] = -7.016563733480e-01;
g[230] = -7.098236542725e-01;
g[231] = -1.302650109286e-01;
g[232] = 6.031445173765e-02;
g[233] = -6.137759844253e-02;
g[234] = -2.499999999998e+01;
g[235] = 2.500000000000e+01;
g[236] = 2.500000000000e+01;
g[237] = 0.000000000000e+00;
g[238] = 0.000000000000e+00;
g[239] = 0.000000000000e+00;
g[240] = -2.811575384022e+00;
g[241] = 2.775965081530e+00;
g[242] = 2.725189582354e+00;
g[243] = -5.638587008894e+00;
g[244] = 5.509255426012e+00;
g[245] = 5.398651984440e+00;
g[246] = -8.649539284642e+00;
g[247] = 8.377531942148e+00;
g[248] = 8.183862834229e+00;
g[249] = 4.001727013728e-01;
g[250] = -5.427282589081e-01;
g[251] = -4.906479336080e-01;
g[252] = 1.053487290215e+00;
g[253] = -8.275947575334e-01;
g[254] = -9.001304092145e-01;
g[255] = 8.089249739665e-01;
g[256] = -1.135465509039e+00;
g[257] = -1.078972596880e+00;
g[258] = -2.499999999998e+01;
g[259] = 2.500000000000e+01;
g[260] = 2.500000000000e+01;
g[261] = 0.000000000000e+00;
g[262] = 0.000000000000e+00;
g[263] = 0.000000000000e+00;
g[264] = -1.898725960972e+00;
g[265] = 1.738803796863e+00;
g[266] = 1.658920044481e+00;
g[267] = -3.859782216988e+00;
g[268] = 3.958334354211e+00;
g[269] = 3.802216905950e+00;
g[270] = -6.500331890217e+00;
g[271] = 6.061049030665e+00;
g[272] = 6.073037629926e+00;
g[273] = 5.043219381652e-01;
g[274] = -4.484601874221e-01;
g[275] = -5.212543350635e-01;
g[276] = 6.808937663199e-01;
g[277] = -7.486356554285e-01;
g[278] = -7.015648684467e-01;
g[279] = 1.072456360662e+00;
g[280] = -8.472533358713e-01;
g[281] = -7.363790749916e-01;
g[282] = -2.499999999998e+01;
g[283] = 2.500000000000e+01;
g[284] = 2.500000000000e+01;
g[285] = 0.000000000000e+00;
g[286] = 0.000000000000e+00;
g[287] = 0.000000000000e+00;
g[288] = -1.034610338489e+00;
g[289] = 1.072092650693e+00;
g[290] = 9.069101665315e-01;
g[291] = -2.832292024103e+00;
g[292] = 2.729083664202e+00;
g[293] = 2.753691149977e+00;
g[294] = -5.443919668395e+00;
g[295] = 5.721510996785e+00;
g[296] = 5.852641071755e+00;
g[297] = 2.824374210860e-01;
g[298] = -2.135746616941e-01;
g[299] = -1.777680992352e-01;
g[300] = 3.376249281614e-01;
g[301] = -3.232573420693e-01;
g[302] = -2.245111063941e-01;
g[303] = -1.330665905825e-01;
g[304] = 4.485686418549e-01;
g[305] = 4.479932455813e-01;
g[306] = -2.499999999998e+01;
g[307] = 2.500000000000e+01;
g[308] = 2.500000000000e+01;
g[309] = 0.000000000000e+00;
g[310] = 0.000000000000e+00;
g[311] = 0.000000000000e+00;
g[312] = -9.667154844411e-01;
g[313] = 1.001972227457e+00;
g[314] = 1.091205619181e+00;
g[315] = -2.874003805412e+00;
g[316] = 3.273090547632e+00;
g[317] = 3.410887237688e+00;
g[318] = -6.608471916404e+00;
g[319] = 6.970927592871e+00;
g[320] = 7.028604236841e+00;
g[321] = -2.270667979123e-01;
g[322] = 2.178450566939e-01;
g[323] = 3.876892297539e-01;
g[324] = -4.920364377150e-01;
g[325] = 8.776793198908e-01;
g[326] = 8.950564731545e-01;
g[327] = -7.867123800453e-01;
g[328] = 5.868924824986e-01;
g[329] = 5.365844933502e-01;
g[330] = -2.499999999998e+01;
g[331] = 2.500000000000e+01;
g[332] = 2.500000000000e+01;
g[333] = 0.000000000000e+00;
g[334] = 0.000000000000e+00;
g[335] = 0.000000000000e+00;
g[336] = -1.944035227582e+00;
g[337] = 2.254177295463e+00;
g[338] = 2.481624299805e+00;
g[339] = -4.792033252990e+00;
g[340] = 5.516073459155e+00;
g[341] = 5.684628582034e+00;
g[342] = -7.785238808894e+00;
g[343] = 7.790979150333e+00;
g[344] = 7.806660659466e+00;
g[345] = -7.411601795175e-01;
g[346] = 1.023333874440e+00;
g[347] = 9.472431202027e-01;
g[348] = -1.253702509523e+00;
g[349] = 1.097788673318e+00;
g[350] = 1.118000219048e+00;
g[351] = -3.446958783921e-01;
g[352] = 3.006870402780e-01;
g[353] = 3.218508450654e-01;
g[354] = -2.499999999998e+01;
g[355] = 2.500000000000e+01;
g[356] = 2.500000000000e+01;
g[357] = 0.000000000000e+00;
g[358] = 0.000000000000e+00;
g[359] = 0.000000000000e+00;
g[360] = -3.694926531641e+00;
g[361] = 4.379137845165e+00;
g[362] = 4.277809950657e+00;
g[363] = -6.878020321708e+00;
g[364] = 6.973774762171e+00;
g[365] = 7.119986136101e+00;
g[366] = -8.322568226519e+00;
g[367] = 8.505606909239e+00;
g[368] = 8.633802564186e+00;
g[369] = -8.636482859129e-01;
g[370] = 7.998925018795e-01;
g[371] = 6.330913829923e-01;
g[372] = -6.436227299849e-01;
g[373] = 3.549595581909e-01;
g[374] = 2.892046664481e-01;
g[375] = -2.925254995402e-01;
g[376] = 4.154756205165e-01;
g[377] = 4.804808327721e-01;
g[378] = -2.499999999998e+01;
g[379] = 2.500000000000e+01;
g[380] = 2.500000000000e+01;
g[381] = 0.000000000000e+00;
g[382] = 0.000000000000e+00;
g[383] = 0.000000000000e+00;
g[384] = -4.594097530039e+00;
g[385] = 4.514858567287e+00;
g[386] = 4.334193174067e+00;
g[387] = -7.220460302901e+00;
g[388] = 7.237152239654e+00;
g[389] = 7.160255326223e+00;
g[390] = -8.988929803666e+00;
g[391] = 9.119700881791e+00;
g[392] = 9.235483825356e+00;
g[393] = 1.133843628718e-01;
g[394] = -6.683243243686e-01;
g[395] = -5.837761567421e-01;
g[396] = 2.152437650092e-01;
g[397] = -6.673170342129e-02;
g[398] = -1.860617062812e-01;
g[399] = -2.729502630840e-01;
g[400] = 9.565461988684e-02;
g[401] = -6.877127404687e-03;
g[402] = -2.499999999998e+01;
g[403] = 2.500000000000e+01;
g[404] = 2.500000000000e+01;
g[405] = 0.000000000000e+00;
g[406] = 0.000000000000e+00;
g[407] = 0.000000000000e+00;
g[408] = -3.294280707948e+00;
g[409] = 2.659645406140e+00;
g[410] = 2.652746081853e+00;
g[411] = -6.401095961009e+00;
g[412] = 6.416664430436e+00;
g[413] = 6.254868514407e+00;
g[414] = -8.889609576147e+00;
g[415] = 8.671372045598e+00;
g[416] = 8.463655795840e+00;
g[417] = 9.984341497015e-01;
g[418] = -8.684091511075e-01;
g[419] = -8.456260382931e-01;
g[420] = 5.797110569243e-01;
g[421] = -8.308691493857e-01;
g[422] = -7.880350401470e-01;
g[423] = 4.367890122074e-01;
g[424] = -5.352310388115e-01;
g[425] = -6.947375310355e-01;
g[426] = -2.499999999998e+01;
g[427] = 2.500000000000e+01;
g[428] = 2.500000000000e+01;
g[429] = 0.000000000000e+00;
g[430] = 0.000000000000e+00;
g[431] = 0.000000000000e+00;
g[432] = -1.609832733876e+00;
g[433] = 1.726162324133e+00;
g[434] = 1.589965045330e+00;
g[435] = -4.788807696393e+00;
g[436] = 4.165775349734e+00;
g[437] = 4.120366570561e+00;
g[438] = -7.445135388320e+00;
g[439] = 7.173531312226e+00;
g[440] = 6.920598635728e+00;
g[441] = 5.125683454938e-01;
g[442] = -1.205482249963e-01;
g[443] = -2.230509273726e-01;
g[444] = 1.019313939206e+00;
g[445] = -1.190993664671e+00;
g[446] = -1.164289141952e+00;
g[447] = 8.726351590145e-01;
g[448] = -8.829536196763e-01;
g[449] = -7.432272124591e-01;
g[450] = -2.499999999998e+01;
g[451] = 2.500000000000e+01;
g[452] = 2.500000000000e+01;
g[453] = 0.000000000000e+00;
g[454] = 0.000000000000e+00;
g[455] = 0.000000000000e+00;
g[456] = -1.287830634935e+00;
g[457] = 1.490850284975e+00;
g[458] = 1.320948855399e+00;
g[459] = -2.826568389159e+00;
g[460] = 2.705851940069e+00;
g[461] = 2.586206157749e+00;
g[462] = -5.979737670624e+00;
g[463] = 5.652092471259e+00;
g[464] = 5.821503833611e+00;
g[465] = -7.884666130430e-02;
g[466] = -2.229688972450e-01;
g[467] = -1.315599630734e-01;
g[468] = 7.457925191500e-01;
g[469] = -1.429337385594e-01;
g[470] = -2.076683878290e-01;
g[471] = 4.872127128602e-01;
g[472] = -4.675603749465e-01;
g[473] = -2.792432672667e-01;
g[474] = -2.499999999998e+01;
g[475] = 2.500000000000e+01;
g[476] = 2.500000000000e+01;
g[477] = 0.000000000000e+00;
g[478] = 0.000000000000e+00;
g[479] = 0.000000000000e+00;
g[480] = -1.351591407316e+00;
g[481] = 1.060674810867e+00;
g[482] = 1.105645195572e+00;
g[483] = -2.533595194416e+00;
g[484] = 3.320354108234e+00;
g[485] = 3.307968736108e+00;
g[486] = -5.729330000230e+00;
g[487] = 5.878581303118e+00;
g[488] = 6.083767394394e+00;
g[489] = 6.596540957165e-02;
g[490] = -5.273907988790e-02;
g[491] = 2.628351635178e-02;
g[492] = -5.012585025694e-01;
g[493] = 6.041708821553e-01;
g[494] = 7.902994496406e-01;
g[495] = -2.595908389641e-01;
g[496] = 7.116804099301e-01;
g[497] = 5.637391294426e-01;
g[498] = -2.499999999998e+01;
g[499] = 2.500000000000e+01;
g[500] = 2.500000000000e+01;
g[501] = 0.000000000000e+00;
g[502] = 0.000000000000e+00;
g[503] = 0.000000000000e+00;
g[504] = -1.408536142393e+00;
g[505] = 1.780351238226e+00;
g[506] = 1.891791567010e+00;
g[507] = -4.305274928748e+00;
g[508] = 4.703034753734e+00;
g[509] = 5.056402732480e+00;
g[510] = -6.958295769590e+00;
g[511] = 7.848742351340e+00;
g[512] = 7.753815507394e+00;
g[513] = -2.841379742100e-01;
g[514] = 7.735655353243e-01;
g[515] = 7.743149231870e-01;
g[516] = -1.045689007957e+00;
g[517] = 7.679800997473e-01;
g[518] = 8.530769935665e-01;
g[519] = -8.928541713149e-01;
g[520] = 9.821741546618e-01;
g[521] = 9.251740301074e-01;
g[522] = -2.499999999998e+01;
g[523] = 2.500000000000e+01;
g[524] = 2.500000000000e+01;
g[525] = 0.000000000000e+00;
g[526] = 0.000000000000e+00;
g[527] = 0.000000000000e+00;
g[528] = -2.775766877991e+00;
g[529] = 3.624416092051e+00;
g[530] = 3.697738920647e+00;
g[531] = -6.051133858428e+00;
g[532] = 6.448645112370e+00;
g[533] = 6.627775107735e+00;
g[534] = -8.697496954191e+00;
g[535] = 8.913477510110e+00;
g[536] = 9.000735274661e+00;
g[537] = -9.836999011600e-01;
g[538] = 8.723280032511e-01;
g[539] = 8.277469207232e-01;
g[540] = -6.618400727902e-01;
g[541] = 9.150113620970e-01;
g[542] = 7.047105628398e-01;
g[543] = -6.641992760407e-01;
g[544] = 4.207922567180e-02;
g[545] = 2.202994392014e-01;
g[546] = -2.499999999998e+01;
g[547] = 2.500000000000e+01;
g[548] = 2.500000000000e+01;
g[549] = 0.000000000000e+00;
g[550] = 0.000000000000e+00;
g[551] = 0.000000000000e+00;
g[552] = -4.418111146860e+00;
g[553] = 4.589063789091e+00;
g[554] = 4.509148082072e+00;
g[555] = -7.086904290824e+00;
g[556] = 7.678747903160e+00;
g[557] = 7.544325683326e+00;
g[558] = -9.136822393917e+00;
g[559] = 8.592566523857e+00;
g[560] = 8.833285093420e+00;
g[561] = -4.264655919601e-01;
g[562] = 1.167327757394e-02;
g[563] = -8.727231909005e-02;
g[564] = -3.742040086456e-01;
g[565] = 1.226360632788e-01;
g[566] = 7.632980165398e-02;
g[567] = 2.115377780157e-01;
g[568] = -1.651247249454e-01;
g[569] = -2.441942424179e-01;
g[570] = -2.499999999998e+01;
g[571] = 2.500000000000e+01;
g[572] = 2.500000000000e+01;
g[573] = 0.000000000000e+00;
g[574] = 0.000000000000e+00;
g[575] = 0.000000000000e+00;
g[576] = -4.105744439957e+00;
g[577] = 3.720378617879e+00;
g[578] = 3.508190685760e+00;
g[579] = -7.183080440550e+00;
g[580] = 6.668563741150e+00;
g[581] = 6.611427213798e+00;
g[582] = -8.396460565604e+00;
g[583] = 8.568078641543e+00;
g[584] = 8.431074926159e+00;
g[585] = 6.630230010448e-01;
g[586] = -7.999239631070e-01;
g[587] = -8.043714993752e-01;
g[588] = 4.063750769304e-01;
g[589] = -1.015126540719e+00;
g[590] = -9.523426726415e-01;
g[591] = 3.636999087880e-01;
g[592] = 4.069658527259e-02;
g[593] = -1.776944634673e-01;
g[594] = -2.499999999998e+01;
g[595] = 2.500000000000e+01;
g[596] = 2.500000000000e+01;
g[597] = 0.000000000000e+00;
g[598] = 0.000000000000e+00;
g[599] = 0.000000000000e+00;
g[600] = -2.427425235153e+00;
g[601] = 1.899388284532e+00;
g[602] = 1.800829234114e+00;
g[603] = -5.361800117175e+00;
g[604] = 4.548863484942e+00;
g[605] = 4.454602930285e+00;
g[606] = -7.922660887002e+00;
g[607] = 8.002838444591e+00;
g[608] = 7.701579623712e+00;
g[609] = 8.588134406253e-01;
g[610] = -8.652587436802e-01;
g[611] = -7.817971720538e-01;
g[612] = 1.261674113402e+00;
g[613] = -9.012964454222e-01;
g[614] = -9.987539353416e-01;
g[615] = 1.984193917202e-01;
g[616] = -7.137528612692e-01;
g[617] = -6.293773995003e-01;
g[618] = -2.499999999998e+01;
g[619] = 2.500000000000e+01;
g[620] = 2.500000000000e+01;
g[621] = 0.000000000000e+00;
g[622] = 0.000000000000e+00;
g[623] = 0.000000000000e+00;
g[624] = -1.075899162255e+00;
g[625] = 8.505498415426e-01;
g[626] = 8.207680587916e-01;
g[627] = -3.181199807746e+00;
g[628] = 3.315898172234e+00;
g[629] = 3.102789754115e+00;
g[630] = -7.042107680769e+00;
g[631] = 6.119722539733e+00;
g[632] = 6.153403539428e+00;
g[633] = 4.348813638618e-01;
g[634] = -1.010612090458e-01;
g[635] = -1.289450362556e-01;
g[636] = 7.018342085746e-01;
g[637] = -3.796578041566e-01;
g[638] = -3.379844321436e-01;
g[639] = 7.243902459228e-01;
g[640] = -8.958486750132e-01;
g[641] = -7.225967617203e-01;
g[642] = -2.499999999998e+01;
g[643] = 2.500000000000e+01;
g[644] = 2.500000000000e+01;
g[645] = 0.000000000000e+00;
g[646] = 0.000000000000e+00;
g[647] = 0.000000000000e+00;
g[648] = -8.386417027989e-01;
g[649] = 1.333700443716e+00;
g[650] = 1.230391794099e+00;
g[651] = -2.760473780265e+00;
g[652] = 2.921641124706e+00;
g[653] = 3.008266901137e+00;
g[654] = -5.603472734032e+00;
g[655] = 5.483285305989e+00;
g[656] = 5.669424823938e+00;
g[657] = -2.081995805505e-01;
g[658] = 4.386135192394e-01;
g[659] = 4.453913110984e-01;
g[660] = -1.876398286714e-01;
g[661] = 5.447981909577e-02;
g[662] = 2.491263377720e-01;
g[663] = 4.525525941177e-01;
g[664] = 3.565962078323e-01;
g[665] = 3.378117866890e-01;
g[666] = -2.499999999998e+01;
g[667] = 2.500000000000e+01;
g[668] = 2.500000000000e+01;
g[669] = 0.000000000000e+00;
g[670] = 0.000000000000e+00;
g[671] = 0.000000000000e+00;
g[672] = -1.720322719888e+00;
g[673] = 2.041372205670e+00;
g[674] = 2.143623773889e+00;
g[675] = -3.506645245187e+00;
g[676] = 3.971809593655e+00;
g[677] = 4.249081123839e+00;
g[678] = -6.008259419490e+00;
g[679] = 7.085347670973e+00;
g[680] = 7.186596774386e+00;
g[681] = -5.663515333388e-01;
g[682] = 2.391493813011e-01;
g[683] = 3.999309051994e-01;
g[684] = -5.385475993047e-01;
g[685] = 1.016797657643e+00;
g[686] = 9.933871645684e-01;
g[687] = -8.478515620196e-01;
g[688] = 1.010896613571e+00;
g[689] = 9.775768570787e-01;
g[690] = -2.499999999998e+01;
g[691] = 2.500000000000e+01;
g[692] = 2.500000000000e+01;
g[693] = 0.000000000000e+00;
g[694] = 0.000000000000e+00;
g[695] = 0.000000000000e+00;
g[696] = -2.670636670913e+00;
g[697] = 2.695501500680e+00;
g[698] = 2.926358761133e+00;
g[699] = -5.096335725669e+00;
g[700] = 6.356593641208e+00;
g[701] = 6.500419340448e+00;
g[702] = -8.145615958450e+00;
g[703] = 8.716721262188e+00;
g[704] = 8.792266549253e+00;
g[705] = -3.480440136164e-01;
g[706] = 5.438255232674e-01;
g[707] = 4.584155277793e-01;
g[708] = -1.048861972655e+00;
g[709] = 1.092977567934e+00;
g[710] = 1.030740920081e+00;
g[711] = -1.005847138137e+00;
g[712] = 5.264846192019e-01;
g[713] = 5.199873116398e-01;
g[714] = -2.499999999998e+01;
g[715] = 2.500000000000e+01;
g[716] = 2.500000000000e+01;
g[717] = 0.000000000000e+00;
g[718] = 0.000000000000e+00;
g[719] = 0.000000000000e+00;
g[720] = -3.329010699595e+00;
g[721] = 4.188047628250e+00;
g[722] = 4.032128205717e+00;
g[723] = -7.137745146403e+00;
g[724] = 7.486814266311e+00;
g[725] = 7.549885552539e+00;
g[726] = -9.281329430792e+00;
g[727] = 9.164810286873e+00;
g[728] = 9.201976126201e+00;
g[729] = -3.838206027556e-01;
g[730] = 7.494226733625e-01;
g[731] = 5.328630805258e-01;
g[732] = -7.581495256006e-01;
g[733] = 1.890757858705e-03;
g[734] = -4.746992029235e-02;
g[735] = -1.037645013760e-01;
g[736] = -5.202652234338e-02;
g[737] = -8.195899033850e-02;
g[738] = -2.499999999998e+01;
g[739] = 2.500000000000e+01;
g[740] = 2.500000000000e+01;
g[741] = 0.000000000000e+00;
g[742] = 0.000000000000e+00;
g[743] = 0.000000000000e+00;
g[744] = -4.128024498526e+00;
g[745] = 4.578993367663e+00;
g[746] = 4.272154326863e+00;
g[747] = -7.425011513589e+00;
g[748] = 6.823523007671e+00;
g[749] = 6.668935967859e+00;
g[750] = -8.903933388891e+00;
g[751] = 8.625490404897e+00;
g[752] = 8.585801902602e+00;
g[753] = -2.738333645783e-01;
g[754] = -5.228225519133e-01;
g[755] = -4.357144464634e-01;
g[756] = 4.819137712101e-01;
g[757] = -5.310805158719e-01;
g[758] = -6.846037164126e-01;
g[759] = 3.739132410910e-01;
g[760] = -4.444347489123e-01;
g[761] = -4.966452771731e-01;
g[762] = -2.499999999998e+01;
g[763] = 2.500000000000e+01;
g[764] = 2.500000000000e+01;
g[765] = 0.000000000000e+00;
g[766] = 0.000000000000e+00;
g[767] = 0.000000000000e+00;
g[768] = -3.737726023310e+00;
g[769] = 2.474076400724e+00;
g[770] = 2.443639982330e+00;
g[771] = -5.783921892875e+00;
g[772] = 5.498650210198e+00;
g[773] = 5.164727575823e+00;
g[774] = -8.025998037921e+00;
g[775] = 7.660814817360e+00;
g[776] = 7.479236768575e+00;
g[777] = 7.411323849534e-01;
g[778] = -1.269170774063e+00;
g[779] = -1.174176806507e+00;
g[780] = 9.596207699065e-01;
g[781] = -8.420766725503e-01;
g[782] = -8.109901806945e-01;
g[783] = 4.896239295182e-01;
g[784] = -4.548391779679e-01;
g[785] = -5.303644896927e-01;
g[786] = -2.499999999998e+01;
g[787] = 2.500000000000e+01;
g[788] = 2.500000000000e+01;
g[789] = 0.000000000000e+00;
g[790] = 0.000000000000e+00;
g[791] = 0.000000000000e+00;
g[792] = -1.581776619309e+00;
g[793] = 7.584375823923e-01;
g[794] = 6.741797505634e-01;
g[795] = -4.110107725129e+00;
g[796] = 3.507602524124e+00;
g[797] = 3.470132526059e+00;
g[798] = -7.004690821400e+00;
g[799] = 6.855584298081e+00;
g[800] = 6.652723448694e+00;
g[801] = 1.148548243314e+00;
g[802] = -3.060553572179e-01;
g[803] = -4.035591083099e-01;
g[804] = 6.949011273244e-01;
g[805] = -9.901012570290e-01;
g[806] = -7.914071363422e-01;
g[807] = 4.906908636880e-01;
g[808] = -3.913854978867e-01;
g[809] = -3.021152219730e-01;
g[810] = -2.499999999998e+01;
g[811] = 2.500000000000e+01;
g[812] = 2.500000000000e+01;
g[813] = 0.000000000000e+00;
g[814] = 0.000000000000e+00;
g[815] = 0.000000000000e+00;
g[816] = -2.849086403961e-01;
g[817] = 9.977596676384e-01;
g[818] = 9.027683783206e-01;
g[819] = -2.932137123659e+00;
g[820] = 2.495367349029e+00;
g[821] = 2.664618150315e+00;
g[822] = -6.269378075443e+00;
g[823] = 6.008059213761e+00;
g[824] = 6.174846764267e+00;
g[825] = 4.814778476545e-03;
g[826] = 3.785448815927e-01;
g[827] = 5.067866854146e-01;
g[828] = 4.452675911218e-01;
g[829] = 1.382925415487e-01;
g[830] = 1.265033677185e-01;
g[831] = 2.123278343242e-01;
g[832] = -3.827258040141e-01;
g[833] = -1.578626992183e-01;
g[834] = -2.499999999998e+01;
g[835] = 2.500000000000e+01;
g[836] = 2.500000000000e+01;
g[837] = 0.000000000000e+00;
g[838] = 0.000000000000e+00;
g[839] = 0.000000000000e+00;
g[840] = -1.300546847703e+00;
g[841] = 1.800712411346e+00;
g[842] = 2.064649670090e+00;
g[843] = -2.767111170738e+00;
g[844] = 3.869784810935e+00;
g[845] = 3.987046279349e+00;
g[846] = -6.178186019522e+00;
g[847] = 5.996791239478e+00;
g[848] = 6.373497521856e+00;
g[849] = -8.009486623177e-01;
g[850] = 4.155532464178e-01;
g[851] = 5.431673220859e-01;
g[852] = -3.975892123539e-01;
g[853] = 1.016618859387e+00;
g[854] = 1.054435777948e+00;
g[855] = -1.249117571034e-01;
g[856] = 5.227084783273e-01;
g[857] = 4.612842468043e-01;
g[858] = -2.499999999998e+01;
g[859] = 2.500000000000e+01;
g[860] = 2.500000000000e+01;
g[861] = 0.000000000000e+00;
g[862] = 0.000000000000e+00;
g[863] = 0.000000000000e+00;
g[864] = -2.615205519069e+00;
g[865] = 2.795915735152e+00;
g[866] = 2.969266037864e+00;
g[867] = -4.564940010913e+00;
g[868] = 5.720412158276e+00;
g[869] = 6.062189354565e+00;
g[870] = -6.883211940023e+00;
g[871] = 8.015850826710e+00;
g[872] = 8.061476364765e+00;
g[873] = -4.259225498513e-01;
g[874] = 5.754805349026e-01;
g[875] = 3.893823897898e-01;
g[876] = -1.244580618398e+00;
g[877] = 7.554042731967e-01;
g[878] = 8.597054192310e-01;
g[879] = -6.157516842577e-01;
g[880] = 1.261161128237e+00;
g[881] = 1.079169183992e+00;
g[882] = -2.499999999998e+01;
g[883] = 2.500000000000e+01;
g[884] = 2.500000000000e+01;
g[885] = 0.000000000000e+00;
g[886] = 0.000000000000e+00;
g[887] = 0.000000000000e+00;
g[888] = -3.207843535887e+00;
g[889] = 3.841125823101e+00;
g[890] = 3.686056833581e+00;
g[891] = -6.729616250663e+00;
g[892] = 7.011715382190e+00;
g[893] = 7.194255558247e+00;
g[894] = -8.565047786915e+00;
g[895] = 9.726872278596e+00;
g[896] = 9.653798654076e+00;
g[897] = -2.644934219676e-01;
g[898] = 3.864571217277e-01;
g[899] = 3.022960489354e-01;
g[900] = -7.107458637530e-01;
g[901] = 5.527618291366e-01;
g[902] = 2.797544452769e-01;
g[903] = -9.406958757501e-01;
g[904] = 2.224261217869e-01;
g[905] = 2.832172197168e-01;
g[906] = -2.499999999998e+01;
g[907] = 2.500000000000e+01;
g[908] = 2.500000000000e+01;
g[909] = 0.000000000000e+00;
g[910] = 0.000000000000e+00;
g[911] = 0.000000000000e+00;
g[912] = -3.790591377437e+00;
g[913] = 4.148663668234e+00;
g[914] = 3.956887462734e+00;
g[915] = -7.204636841953e+00;
g[916] = 7.549337416279e+00;
g[917] = 7.155849755847e+00;
g[918] = -9.773374981259e+00;
g[919] = 8.959873530427e+00;
g[920] = 8.985361663252e+00;
g[921] = -2.467850350885e-01;
g[922] = -8.994432692620e-02;
g[923] = -8.994680488397e-02;
g[924] = 1.486440419958e-01;
g[925] = -1.777368797267e-01;
g[926] = -3.605298209875e-01;
g[927] = -6.293601481055e-02;
g[928] = -7.697073433255e-01;
g[929] = -8.013499020805e-01;
g[930] = -2.499999999998e+01;
g[931] = 2.500000000000e+01;
g[932] = 2.500000000000e+01;
g[933] = 0.000000000000e+00;
g[934] = 0.000000000000e+00;
g[935] = 0.000000000000e+00;
g[936] = -3.758838730946e+00;
g[937] = 3.441261581976e+00;
g[938] = 3.179222336351e+00;
g[939] = -6.530018887733e+00;
g[940] = 6.009360209392e+00;
g[941] = 5.663789591260e+00;
g[942] = -8.637026100540e+00;
g[943] = 7.640220541765e+00;
g[944] = 7.410835788768e+00;
g[945] = 3.400791854228e-01;
g[946] = -6.359818070063e-01;
g[947] = -6.848954079301e-01;
g[948] = 5.063595904376e-01;
g[949] = -1.245312652196e+00;
g[950] = -1.063726960440e+00;
g[951] = 1.034483771476e+00;
g[952] = -4.305204934538e-01;
g[953] = -5.910274463249e-01;
g[954] = -2.499999999998e+01;
g[955] = 2.500000000000e+01;
g[956] = 2.500000000000e+01;
g[957] = 0.000000000000e+00;
g[958] = 0.000000000000e+00;
g[959] = 0.000000000000e+00;
g[960] = -2.543613706166e+00;
g[961] = 1.676394152020e+00;
g[962] = 1.466295719057e+00;
g[963] = -5.000094179990e+00;
g[964] = 3.635500773957e+00;
g[965] = 3.581869836946e+00;
g[966] = -6.775183955660e+00;
g[967] = 7.032848302410e+00;
g[968] = 6.747324254919e+00;
g[969] = 7.813615289530e-01;
g[970] = -1.015990968122e+00;
g[971] = -8.952458312168e-01;
g[972] = 1.030506931267e+00;
g[973] = -8.684700539516e-01;
g[974] = -8.197847924481e-01;
g[975] = 5.918801253829e-01;
g[976] = -3.023381759097e-01;
g[977] = -1.588920441654e-01;
g[978] = -2.499999999998e+01;
g[979] = 2.500000000000e+01;
g[980] = 2.500000000000e+01;
g[981] = 0.000000000000e+00;
g[982] = 0.000000000000e+00;
g[983] = 0.000000000000e+00;
g[984] = -1.059023943441e+00;
g[985] = 2.383177149579e-01;
g[986] = 3.428201070153e-01;
g[987] = -2.933313073662e+00;
g[988] = 2.898847405269e+00;
g[989] = 2.866691668829e+00;
g[990] = -6.428628516498e+00;
g[991] = 6.248697049761e+00;
g[992] = 6.421307971908e+00;
g[993] = 6.080821640175e-01;
g[994] = -1.971412857202e-01;
g[995] = -6.195411865930e-02;
g[996] = 7.971537287773e-01;
g[997] = 4.795084870549e-02;
g[998] = 1.042537279627e-01;
g[999] = -1.016207058502e-01;
g[1000] = -3.928967720702e-01;
g[1001] = -1.721964672869e-01;
g[1002] = -2.499999999998e+01;
g[1003] = 2.500000000000e+01;
g[1004] = 2.500000000000e+01;
g[1005] = 0.000000000000e+00;
g[1006] = 0.000000000000e+00;
g[1007] = 0.000000000000e+00;
g[1008] = -4.742574826103e-01;
g[1009] = 1.183731083811e+00;
g[1010] = 1.372239791330e+00;
g[1011] = -2.580788000816e+00;
g[1012] = 3.371182345050e+00;
g[1013] = 3.699959013257e+00;
g[1014] = -6.499153697773e+00;
g[1015] = 6.054814314061e+00;
g[1016] = 6.384629957035e+00;
g[1017] = -1.059896817363e-01;
g[1018] = 9.928435264095e-01;
g[1019] = 9.596786849920e-01;
g[1020] = -4.630385566875e-01;
g[1021] = 4.045119020075e-01;
g[1022] = 6.542967543487e-01;
g[1023] = 8.148859820885e-02;
g[1024] = 2.928718176093e-01;
g[1025] = 2.359851512618e-01;
g[1026] = -2.499999999998e+01;
g[1027] = 2.500000000000e+01;
g[1028] = 2.500000000000e+01;
g[1029] = 0.000000000000e+00;
g[1030] = 0.000000000000e+00;
g[1031] = 0.000000000000e+00;
g[1032] = -1.603728768442e+00;
g[1033] = 3.069197318183e+00;
g[1034] = 3.233835675743e+00;
g[1035] = -4.170434253297e+00;
g[1036] = 4.791766407178e+00;
g[1037] = 5.327553419221e+00;
g[1038] = -6.520580787356e+00;
g[1039] = 7.316488537625e+00;
g[1040] = 7.498679815040e+00;
g[1041] = -9.499517987469e-01;
g[1042] = 6.525313405613e-01;
g[1043] = 6.861295973916e-01;
g[1044] = -9.206195023456e-01;
g[1045] = 1.049605881228e+00;
g[1046] = 9.398012551204e-01;
g[1047] = -2.628407578541e-01;
g[1048] = 8.507567748741e-01;
g[1049] = 8.107391340304e-01;
g[1050] = -2.499999999998e+01;
g[1051] = 2.500000000000e+01;
g[1052] = 2.500000000000e+01;
g[1053] = 0.000000000000e+00;
g[1054] = 0.000000000000e+00;
g[1055] = 0.000000000000e+00;
g[1056] = -3.539777106970e+00;
g[1057] = 3.608979066378e+00;
g[1058] = 3.792013407661e+00;
g[1059] = -5.779033692436e+00;
g[1060] = 7.059166897590e+00;
g[1061] = 7.111374732516e+00;
g[1062] = -7.797465654972e+00;
g[1063] = 8.928456613990e+00;
g[1064] = 9.073897961042e+00;
g[1065] = -7.568284134919e-01;
g[1066] = 2.329106559231e-02;
g[1067] = -5.375242842713e-02;
g[1068] = -6.877786546406e-01;
g[1069] = 9.537848585574e-01;
g[1070] = 6.898745492229e-01;
g[1071] = -9.119418641930e-01;
g[1072] = 6.572080792821e-01;
g[1073] = 6.121429616355e-01;
g[1074] = -2.499999999998e+01;
g[1075] = 2.500000000000e+01;
g[1076] = 2.500000000000e+01;
g[1077] = 0.000000000000e+00;
g[1078] = 0.000000000000e+00;
g[1079] = 0.000000000000e+00;
g[1080] = -4.046253180559e+00;
g[1081] = 3.795454168729e+00;
g[1082] = 3.579327430727e+00;
g[1083] = -7.039439363871e+00;
g[1084] = 7.688428308749e+00;
g[1085] = 7.474344362594e+00;
g[1086] = -9.345982844576e+00;
g[1087] = 9.606435062045e+00;
g[1088] = 9.489117383876e+00;
g[1089] = 2.259734210283e-01;
g[1090] = 1.911128605343e-01;
g[1091] = -7.452543945985e-02;
g[1092] = -5.245164486260e-01;
g[1093] = -3.696155090982e-01;
g[1094] = -3.829983412264e-01;
g[1095] = -4.457989090994e-01;
g[1096] = -5.527369948294e-02;
g[1097] = -2.470561626138e-01;
g[1098] = -2.499999999998e+01;
g[1099] = 2.500000000000e+01;
g[1100] = 2.500000000000e+01;
g[1101] = 0.000000000000e+00;
g[1102] = 0.000000000000e+00;
g[1103] = 0.000000000000e+00;
g[1104] = -3.295256596526e+00;
g[1105] = 3.847487882668e+00;
g[1106] = 3.338266212530e+00;
g[1107] = -7.309856584408e+00;
g[1108] = 6.189632449356e+00;
g[1109] = 5.917518772387e+00;
g[1110] = -9.263869891191e+00;
g[1111] = 8.562649442909e+00;
g[1112] = 8.197590651240e+00;
g[1113] = 3.490509546281e-01;
g[1114] = -3.137576017722e-01;
g[1115] = -2.821824298687e-01;
g[1116] = 3.894328078782e-01;
g[1117] = -9.102245234379e-01;
g[1118] = -1.002937427469e+00;
g[1119] = 4.851160599753e-01;
g[1120] = -9.266211963899e-01;
g[1121] = -9.320320900617e-01;
g[1122] = -2.499999999998e+01;
g[1123] = 2.500000000000e+01;
g[1124] = 2.500000000000e+01;
g[1125] = 0.000000000000e+00;
g[1126] = 0.000000000000e+00;
g[1127] = 0.000000000000e+00;
g[1128] = -2.856784384702e+00;
g[1129] = 2.341327588019e+00;
g[1130] = 2.161462709674e+00;
g[1131] = -5.501856201367e+00;
g[1132] = 4.540737618958e+00;
g[1133] = 4.067196016702e+00;
g[1134] = -7.829101695745e+00;
g[1135] = 6.640359401262e+00;
g[1136] = 6.473057921416e+00;
g[1137] = 1.795407657609e-01;
g[1138] = -1.057842815215e+00;
g[1139] = -8.456521458346e-01;
g[1140] = 1.244228171663e+00;
g[1141] = -7.426104258410e-01;
g[1142] = -7.680939911958e-01;
g[1143] = 8.367156805105e-01;
g[1144] = -7.597047034488e-01;
g[1145] = -6.125991068327e-01;
g[1146] = -2.499999999998e+01;
g[1147] = 2.500000000000e+01;
g[1148] = 2.500000000000e+01;
g[1149] = 0.000000000000e+00;
g[1150] = 0.000000000000e+00;
g[1151] = 0.000000000000e+00;
g[1152] = -2.043986086753e+00;
g[1153] = 5.952692800220e-01;
g[1154] = 6.204542792323e-01;
g[1155] = -3.343248145992e+00;
g[1156] = 3.085629322454e+00;
g[1157] = 2.979441264253e+00;
g[1158] = -6.319299482493e+00;
g[1159] = 6.093675376126e+00;
g[1160] = 6.111862812430e+00;
g[1161] = 6.847752767491e-01;
g[1162] = -4.762372633180e-01;
g[1163] = -4.869628739767e-01;
g[1164] = 7.132841257052e-01;
g[1165] = -6.523626223147e-01;
g[1166] = -2.830660177912e-01;
g[1167] = 5.786161414378e-01;
g[1168] = 1.745020096096e-01;
g[1169] = 2.099869765753e-01;
g[1170] = -2.499999999998e+01;
g[1171] = 2.500000000000e+01;
g[1172] = 2.500000000000e+01;
g[1173] = 0.000000000000e+00;
g[1174] = 0.000000000000e+00;
g[1175] = 0.000000000000e+00;
g[1176] = -6.065684264550e-01;
g[1177] = 6.990658822914e-01;
g[1178] = 7.450808117918e-01;
g[1179] = -2.821913398570e+00;
g[1180] = 2.647276187230e+00;
g[1181] = 3.184223319045e+00;
g[1182] = -5.857819076734e+00;
g[1183] = 6.632033723223e+00;
g[1184] = 6.778658164452e+00;
g[1185] = 5.118141522936e-01;
g[1186] = 5.068498035718e-01;
g[1187] = 6.050418108684e-01;
g[1188] = -1.061346726038e-01;
g[1189] = 3.823150677188e-01;
g[1190] = 5.367970534682e-01;
g[1191] = -1.593601514446e-01;
g[1192] = 1.989947855755e-01;
g[1193] = 3.296342017567e-01;
g[1194] = -2.499999999998e+01;
g[1195] = 2.500000000000e+01;
g[1196] = 2.500000000000e+01;
g[1197] = 0.000000000000e+00;
g[1198] = 0.000000000000e+00;
g[1199] = 0.000000000000e+00;
g[1200] = -7.625106441041e-01;
g[1201] = 2.152616945075e+00;
g[1202] = 2.529168504587e+00;
g[1203] = -3.422972505453e+00;
g[1204] = 4.543507734679e+00;
g[1205] = 4.976908993638e+00;
g[1206] = -6.769825253339e+00;
g[1207] = 6.875107904621e+00;
g[1208] = 7.326974070739e+00;
g[1209] = -6.895350132249e-01;
g[1210] = 8.501277929934e-01;
g[1211] = 9.933872374727e-01;
g[1212] = -4.840867043377e-01;
g[1213] = 1.285412560611e+00;
g[1214] = 1.112376454157e+00;
g[1215] = -6.235106831614e-01;
g[1216] = 1.970242615125e-01;
g[1217] = 2.975350677762e-01;
g[1218] = -2.499999999998e+01;
g[1219] = 2.500000000000e+01;
g[1220] = 2.500000000000e+01;
C[0] = 7.455911171076e-02;
C[1] = 1.030505430404e-02;
C[2] = -1.466898732334e-03;
C[3] = 3.707886821251e-01;
C[4] = -3.573910349750e-06;
C[5] = -9.722355640771e-04;
C[6] = 5.083096611100e-02;
C[7] = -2.359837674902e-04;
C[8] = 3.169078848662e-05;
C[9] = 1.304872171714e-01;
C[10] = 9.813674390325e-04;
C[11] = -4.827387668933e-05;
C[12] = 3.039798986406e-02;
C[13] = -1.612395203388e-04;
C[14] = -1.262690840899e-04;
C[15] = 2.470431442715e-03;
C[16] = -2.158790094345e-05;
C[17] = 1.198897633770e-07;
C[18] = 2.453502843967e-03;
C[19] = -3.863334293522e-05;
C[20] = 6.827502478453e-06;
C[21] = -1.487526942362e-05;
C[22] = -2.086765209725e-06;
C[23] = -2.290437910367e-06;
C[24] = 1.043222204694e-02;
C[25] = 2.023574823573e-02;
C[26] = -7.054392309874e-03;
C[27] = -2.289534802498e-04;
C[28] = 3.976537832324e-01;
C[29] = 1.144115846724e-02;
C[30] = -2.159525261802e-04;
C[31] = 5.568728898901e-02;
C[32] = 3.621002947061e-03;
C[33] = 9.898452885592e-04;
C[34] = 1.261203159507e-01;
C[35] = -6.833895495526e-04;
C[36] = -1.739410869353e-04;
C[37] = 3.267928288353e-02;
C[38] = 1.042398825316e-03;
C[39] = -1.981936097187e-05;
C[40] = 2.713302826263e-03;
C[41] = 1.880007485397e-04;
C[42] = -3.343680613382e-05;
C[43] = 2.723669220148e-03;
C[44] = 2.608707632476e-04;
C[45] = -2.025041181146e-06;
C[46] = -1.340160022530e-05;
C[47] = 2.472934222668e-06;
C[48] = -1.441881837231e-03;
C[49] = -7.011491403351e-03;
C[50] = 7.137524615157e-02;
C[51] = -1.088987195894e-03;
C[52] = 1.177172049174e-02;
C[53] = 3.733977757814e-01;
C[54] = 5.537443960396e-05;
C[55] = 3.476920802338e-03;
C[56] = 5.430747088825e-02;
C[57] = -4.660608368248e-05;
C[58] = -6.805294891177e-04;
C[59] = 1.302074215524e-01;
C[60] = -1.330391153826e-04;
C[61] = 1.058140385798e-03;
C[62] = 3.073352585192e-02;
C[63] = 2.072867444289e-06;
C[64] = 1.793595697697e-04;
C[65] = 2.659423158199e-03;
C[66] = 1.294521296757e-05;
C[67] = 2.383084019576e-04;
C[68] = 2.826754539864e-03;
C[69] = -2.161393588513e-06;
C[70] = 2.162976857157e-06;
C[71] = -1.165294255601e-05;
C[72] = 3.706779923254e-01;
C[73] = -2.354398644366e-05;
C[74] = -9.091332315140e-04;
C[75] = 1.293719850189e-01;
C[76] = 2.262509455905e-04;
C[77] = -1.438276501046e-03;
C[78] = 3.853267390956e-01;
C[79] = 1.410040484900e-03;
C[80] = 1.637467421916e-03;
C[81] = 3.039061054408e-02;
C[82] = -1.602471206816e-04;
C[83] = -1.210488510906e-04;
C[84] = 1.332121751959e-01;
C[85] = 1.555716031206e-04;
C[86] = -1.257484958140e-05;
C[87] = 3.138761742983e-02;
C[88] = 8.927317963584e-05;
C[89] = 1.070413807446e-04;
C[90] = 5.861029983644e-02;
C[91] = 2.491799845067e-04;
C[92] = 7.065202087330e-04;
C[93] = 2.554028496248e-03;
C[94] = 8.990071975911e-06;
C[95] = 2.961532991337e-05;
C[96] = -5.435180013345e-06;
C[97] = 3.975265886306e-01;
C[98] = 1.198098322640e-02;
C[99] = 3.268286867737e-04;
C[100] = 9.286969222173e-02;
C[101] = -2.218382102274e-02;
C[102] = 1.353159548548e-03;
C[103] = 3.860985734972e-01;
C[104] = 6.656665650820e-03;
C[105] = -1.613636049830e-04;
C[106] = 3.267080324340e-02;
C[107] = 1.072091234776e-03;
C[108] = 1.622767858662e-04;
C[109] = 1.302477200509e-01;
C[110] = -1.975085071554e-03;
C[111] = 8.568510449439e-05;
C[112] = 3.166239775216e-02;
C[113] = 7.279661079873e-04;
C[114] = 2.589406054589e-04;
C[115] = 5.946183892217e-02;
C[116] = 2.347762565525e-03;
C[117] = 9.815059603751e-06;
C[118] = 2.605977089521e-03;
C[119] = 1.160661228896e-04;
C[120] = -8.666489443926e-04;
C[121] = 1.152591855749e-02;
C[122] = 3.733569210779e-01;
C[123] = -1.370667901740e-03;
C[124] = -2.210068973042e-02;
C[125] = 1.071201907516e-01;
C[126] = 1.566647840818e-03;
C[127] = 6.899854844947e-03;
C[128] = 3.984239289863e-01;
C[129] = -1.192299761109e-04;
C[130] = 1.048049497999e-03;
C[131] = 3.073080220502e-02;
C[132] = -8.067609627703e-06;
C[133] = -1.969542985399e-03;
C[134] = 1.312061506583e-01;
C[135] = 1.022111416867e-04;
C[136] = 7.405396428025e-04;
C[137] = 3.274142376945e-02;
C[138] = 7.200608279477e-04;
C[139] = 2.304514216840e-03;
C[140] = 6.463936341696e-02;
C[141] = 3.086187742277e-05;
C[142] = 1.129533503698e-04;
C[143] = 2.838490683577e-03;
C[144] = 4.999815457359e-02;
C[145] = -2.338203237152e-05;
C[146] = 2.097190020294e-06;
C[147] = 3.863078247784e-01;
C[148] = 1.061915074574e-03;
C[149] = 9.893975507316e-04;
C[150] = 5.405121587473e-02;
C[151] = -1.724331781532e-03;
C[152] = -5.791966350645e-03;
C[153] = 2.414910673554e-03;
C[154] = -6.981328051283e-06;
C[155] = -1.478949194622e-06;
C[156] = 3.145302314202e-02;
C[157] = 6.626880622951e-05;
C[158] = 6.372778901433e-05;
C[159] = 1.290156733533e-01;
C[160] = -1.216172559200e-04;
C[161] = -5.710290668913e-04;
C[162] = 5.073221633978e-01;
C[163] = 7.606131070757e-04;
C[164] = 4.793098283638e-03;
C[165] = 3.713449300733e-02;
C[166] = 6.300304816427e-05;
C[167] = 5.091161036594e-04;
C[168] = -3.364777551995e-05;
C[169] = 5.596939087530e-02;
C[170] = 3.638870487208e-03;
C[171] = 1.096228568891e-03;
C[172] = 3.863073944877e-01;
C[173] = 6.977135432545e-03;
C[174] = -1.731203780417e-03;
C[175] = 4.881481794421e-02;
C[176] = -1.343286218264e-02;
C[177] = -8.098834812100e-06;
C[178] = 2.732109618681e-03;
C[179] = 1.901562154277e-04;
C[180] = 6.835238523529e-05;
C[181] = 3.167631915153e-02;
C[182] = 7.456916819757e-04;
C[183] = -1.220753891790e-04;
C[184] = 1.286309711886e-01;
C[185] = -1.125230602859e-03;
C[186] = 7.523010530223e-04;
C[187] = 5.060524723970e-01;
C[188] = 2.646018895140e-03;
C[189] = 6.261354240457e-05;
C[190] = 3.697514534431e-02;
C[191] = 1.878993271165e-04;
C[192] = -5.900882549380e-06;
C[193] = 3.800845299880e-03;
C[194] = 5.338934424886e-02;
C[195] = 1.014462805221e-03;
C[196] = 6.707435629703e-03;
C[197] = 3.995358509001e-01;
C[198] = -5.804601895981e-03;
C[199] = -1.340700265955e-02;
C[200] = 1.822925555110e-02;
C[201] = -2.386221639024e-06;
C[202] = 1.999902387276e-04;
C[203] = 2.598214715574e-03;
C[204] = 6.550773963163e-05;
C[205] = 7.313507732462e-04;
C[206] = 3.281555189704e-02;
C[207] = -5.718714365804e-04;
C[208] = -1.123506634653e-03;
C[209] = 1.259379133350e-01;
C[210] = 4.795223506448e-03;
C[211] = 2.696173007955e-03;
C[212] = 5.263217975644e-01;
C[213] = 5.092293455002e-04;
C[214] = 1.901197186892e-04;
C[215] = 3.866329541151e-02;
C[216] = -7.038665398063e+00;
C[217] = 5.394588835269e-02;
C[218] = -8.991863476752e-03;
C[219] = 2.117796036768e+00;
C[220] = 1.410332333542e-02;
C[221] = -9.327333934408e-03;
C[222] = 7.952598361246e-01;
C[223] = -4.336212885321e-03;
C[224] = -1.063621086393e-03;
C[225] = 6.594087184734e-02;
C[226] = 1.296107628109e-02;
C[227] = -8.969740153511e-05;
C[228] = 3.783206210074e-01;
C[229] = -2.281615363953e-03;
C[230] = -2.139772840119e-03;
C[231] = 5.131738130580e-02;
C[232] = -4.174097797434e-04;
C[233] = -1.127316226573e-04;
C[234] = 7.841328106677e-02;
C[235] = -7.134807514153e-04;
C[236] = 3.488529275806e-04;
C[237] = 2.435283986852e-03;
C[238] = -3.760840160492e-05;
C[239] = -5.296948669957e-06;
C[240] = 5.648924321073e-02;
C[241] = -7.213946519711e+00;
C[242] = 3.641584204662e-03;
C[243] = 1.044366832111e-02;
C[244] = 2.112147598281e+00;
C[245] = 1.003237240190e-02;
C[246] = -4.154343639765e-03;
C[247] = 8.551901320986e-01;
C[248] = 4.981467902967e-02;
C[249] = 1.308824402399e-02;
C[250] = 2.124080279668e-02;
C[251] = -6.356326098693e-03;
C[252] = -2.485153872447e-03;
C[253] = 3.966993161230e-01;
C[254] = 1.010389622929e-02;
C[255] = -3.932007713052e-04;
C[256] = 5.568547697157e-02;
C[257] = 3.519938076650e-03;
C[258] = -5.891674485376e-04;
C[259] = 8.611678811748e-02;
C[260] = 7.187778649268e-03;
C[261] = -3.230888866798e-05;
C[262] = 2.739367682054e-03;
C[263] = 2.659855855471e-04;
C[264] = -8.491525574697e-03;
C[265] = 4.499602335122e-03;
C[266] = -7.040519423716e+00;
C[267] = -1.123471078681e-02;
C[268] = 1.404942477849e-02;
C[269] = 2.094145047114e+00;
C[270] = -7.726170940830e-04;
C[271] = 4.857621549230e-02;
C[272] = 8.430239780060e-01;
C[273] = -6.468050643233e-05;
C[274] = -6.313425192170e-03;
C[275] = 6.340522331488e-02;
C[276] = -2.260824509419e-03;
C[277] = 1.039746978768e-02;
C[278] = 3.808687928301e-01;
C[279] = -8.318398018494e-05;
C[280] = 3.402668187500e-03;
C[281] = 5.477654859635e-02;
C[282] = 4.946330113980e-04;
C[283] = 6.645610715289e-03;
C[284] = 8.867781887566e-02;
C[285] = 9.685827292035e-07;
C[286] = 2.434047316388e-04;
C[287] = 2.802321535260e-03;
C[288] = 2.115582240773e+00;
C[289] = 1.455185819723e-02;
C[290] = -7.637631499218e-03;
C[291] = -6.209152224061e+00;
C[292] = 5.030338974269e-03;
C[293] = -2.435565366465e-03;
C[294] = 2.144417358521e+00;
C[295] = 8.756972369331e-03;
C[296] = -5.054486674079e-04;
C[297] = 3.782099312076e-01;
C[298] = -2.279744378641e-03;
C[299] = -2.080970545039e-03;
C[300] = 1.209280372958e-01;
C[301] = 3.254062962168e-03;
C[302] = 1.046944684755e-03;
C[303] = 3.857137952967e-01;
C[304] = 8.546172937699e-04;
C[305] = 2.619954095843e-04;
C[306] = 9.930889965859e-01;
C[307] = 3.704038904705e-03;
C[308] = 1.142804101949e-02;
C[309] = 5.866555496181e-02;
C[310] = 2.009212329159e-04;
C[311] = 8.433041598087e-04;
C[312] = 1.406609794215e-02;
C[313] = 2.109603706244e+00;
C[314] = 1.823467947180e-02;
C[315] = 7.041893797933e-03;
C[316] = -6.288919777919e+00;
C[317] = -4.946645756236e-02;
C[318] = 7.832918805835e-03;
C[319] = 2.100780432076e+00;
C[320] = -1.972893083461e-02;
C[321] = -2.283476633617e-03;
C[322] = 3.965721215211e-01;
C[323] = 1.060673252235e-02;
C[324] = 3.354640703351e-03;
C[325] = 9.282339026908e-02;
C[326] = -2.058793381505e-02;
C[327] = 7.856556370714e-04;
C[328] = 3.872710173480e-01;
C[329] = 6.604317779580e-03;
C[330] = 3.858029154890e-03;
C[331] = 1.002593095883e+00;
C[332] = 3.454231376175e-02;
C[333] = 2.210835161211e-04;
C[334] = 5.933775532239e-02;
C[335] = 2.207519309311e-03;
C[336] = -7.215601540718e-03;
C[337] = 1.172757420687e-02;
C[338] = 2.093327953044e+00;
C[339] = -1.083393380355e-03;
C[340] = -4.780383171596e-02;
C[341] = -6.263245948804e+00;
C[342] = -1.835787680496e-03;
C[343] = -1.667119848377e-02;
C[344] = 2.122018914086e+00;
C[345] = -2.034186220435e-03;
C[346] = 1.018865631954e-02;
C[347] = 3.808279381266e-01;
C[348] = 1.114553284061e-03;
C[349] = -2.050480252273e-02;
C[350] = 9.945309249346e-02;
C[351] = 1.627723841885e-04;
C[352] = 6.825284974717e-03;
C[353] = 3.982431146464e-01;
C[354] = 1.171739157552e-02;
C[355] = 3.406625074493e-02;
C[356] = 1.087220749759e+00;
C[357] = 8.781607121187e-04;
C[358] = 2.155119100435e-03;
C[359] = 6.459514371485e-02;
C[360] = 7.786036053764e-01;
C[361] = -3.029337635906e-04;
C[362] = -1.838162085756e-03;
C[363] = 2.164039072177e+00;
C[364] = 2.008029326372e-03;
C[365] = -1.338079348222e-02;
C[366] = -7.093122371650e+00;
C[367] = -4.147955859397e-03;
C[368] = -3.155912595464e-02;
C[369] = 5.048456976839e-02;
C[370] = -2.006302774964e-04;
C[371] = -1.364612297686e-04;
C[372] = 3.866948809795e-01;
C[373] = 4.944111630982e-04;
C[374] = -4.144779058975e-04;
C[375] = 5.338742641520e-02;
C[376] = -9.868343937897e-04;
C[377] = -7.736768565936e-03;
C[378] = 4.074318638967e+00;
C[379] = 4.881272231489e-03;
C[380] = 4.662227849742e-02;
C[381] = 5.070441866910e-01;
C[382] = 7.896138955025e-04;
C[383] = 8.284089281164e-03;
C[384] = -2.894930459154e-04;
C[385] = 8.608321698242e-01;
C[386] = 5.181520918972e-02;
C[387] = 2.480734049167e-03;
C[388] = 2.104956851886e+00;
C[389] = -1.512558673181e-02;
C[390] = -4.285395837087e-03;
C[391] = -7.102304424125e+00;
C[392] = -3.532351909328e-02;
C[393] = -2.150737877732e-04;
C[394] = 5.596757885785e-02;
C[395] = 3.564617872371e-03;
C[396] = 5.408053777617e-04;
C[397] = 3.874798383385e-01;
C[398] = 6.902565562315e-03;
C[399] = -9.937063926742e-04;
C[400] = 5.031764421917e-02;
C[401] = -1.141803130639e-02;
C[402] = 4.720286097723e-03;
C[403] = 4.046613254861e+00;
C[404] = -5.786405256267e-03;
C[405] = 7.711330414351e-04;
C[406] = 5.033938835243e-01;
C[407] = 7.923260017855e-04;
C[408] = -1.815454507113e-03;
C[409] = 5.341152608605e-02;
C[410] = 8.246614452183e-01;
C[411] = -1.296554100131e-02;
C[412] = -1.871353125695e-02;
C[413] = 2.144257352362e+00;
C[414] = -3.181183686135e-02;
C[415] = -3.480632863137e-02;
C[416] = -7.222014502058e+00;
C[417] = -1.503232936933e-04;
C[418] = 3.699780429469e-03;
C[419] = 5.385842195696e-02;
C[420] = -3.610092071108e-04;
C[421] = 6.655087758463e-03;
C[422] = 3.993550365602e-01;
C[423] = -7.749404111271e-03;
C[424] = -1.139217178330e-02;
C[425] = 1.906137767773e-02;
C[426] = 4.660582797215e-02;
C[427] = -5.019605223339e-03;
C[428] = 4.171731650137e+00;
C[429] = 8.264719900545e-03;
C[430] = 8.513394681503e-04;
C[431] = 5.251299211072e-01;
C[432] = 0.000000000000e+00;
C[433] = 0.000000000000e+00;
C[434] = 0.000000000000e+00;
C[435] = 0.000000000000e+00;
C[436] = 0.000000000000e+00;
C[437] = 0.000000000000e+00;
C[438] = 0.000000000000e+00;
C[439] = 0.000000000000e+00;
C[440] = 0.000000000000e+00;
C[441] = 0.000000000000e+00;
C[442] = 0.000000000000e+00;
C[443] = 0.000000000000e+00;
C[444] = 0.000000000000e+00;
C[445] = 0.000000000000e+00;
C[446] = 0.000000000000e+00;
C[447] = 0.000000000000e+00;
C[448] = 0.000000000000e+00;
C[449] = 0.000000000000e+00;
C[450] = 1.000000000000e+00;
C[451] = 0.000000000000e+00;
C[452] = 0.000000000000e+00;
C[453] = 2.000000000000e-01;
C[454] = -0.000000000000e+00;
C[455] = -0.000000000000e+00;
C[456] = 0.000000000000e+00;
C[457] = 0.000000000000e+00;
C[458] = 0.000000000000e+00;
C[459] = 0.000000000000e+00;
C[460] = 0.000000000000e+00;
C[461] = 0.000000000000e+00;
C[462] = 0.000000000000e+00;
C[463] = 0.000000000000e+00;
C[464] = 0.000000000000e+00;
C[465] = 0.000000000000e+00;
C[466] = 0.000000000000e+00;
C[467] = 0.000000000000e+00;
C[468] = 0.000000000000e+00;
C[469] = 0.000000000000e+00;
C[470] = 0.000000000000e+00;
C[471] = 0.000000000000e+00;
C[472] = 0.000000000000e+00;
C[473] = 0.000000000000e+00;
C[474] = 0.000000000000e+00;
C[475] = 1.000000000000e+00;
C[476] = 0.000000000000e+00;
C[477] = -0.000000000000e+00;
C[478] = 2.000000000000e-01;
C[479] = -0.000000000000e+00;
C[480] = 0.000000000000e+00;
C[481] = 0.000000000000e+00;
C[482] = 0.000000000000e+00;
C[483] = 0.000000000000e+00;
C[484] = 0.000000000000e+00;
C[485] = 0.000000000000e+00;
C[486] = 0.000000000000e+00;
C[487] = 0.000000000000e+00;
C[488] = 0.000000000000e+00;
C[489] = 0.000000000000e+00;
C[490] = 0.000000000000e+00;
C[491] = 0.000000000000e+00;
C[492] = 0.000000000000e+00;
C[493] = 0.000000000000e+00;
C[494] = 0.000000000000e+00;
C[495] = 0.000000000000e+00;
C[496] = 0.000000000000e+00;
C[497] = 0.000000000000e+00;
C[498] = 0.000000000000e+00;
C[499] = 0.000000000000e+00;
C[500] = 1.000000000000e+00;
C[501] = -0.000000000000e+00;
C[502] = -0.000000000000e+00;
C[503] = 2.000000000000e-01;
C[504] = 6.182159955619e-02;
C[505] = 1.358673465671e-02;
C[506] = -5.926758121194e-05;
C[507] = 3.856012853359e-01;
C[508] = -3.179026086776e-03;
C[509] = -2.856608151728e-03;
C[510] = 5.252288152741e-02;
C[511] = -6.610905940219e-04;
C[512] = -4.532131797687e-04;
C[513] = 1.297085715106e-01;
C[514] = 1.146606235850e-03;
C[515] = 1.545963514764e-05;
C[516] = 3.148329982872e-02;
C[517] = -3.661538517804e-04;
C[518] = -2.601955448623e-04;
C[519] = 2.543015112424e-03;
C[520] = -3.790817813864e-05;
C[521] = -2.091898934832e-05;
C[522] = 2.506479660960e-03;
C[523] = -4.709124690226e-05;
C[524] = 4.313810423169e-05;
C[525] = -1.747113520744e-05;
C[526] = -1.154321516649e-06;
C[527] = 8.937553919969e-07;
C[528] = 1.358715507322e-02;
C[529] = 2.223429250074e-02;
C[530] = -2.094600326254e-03;
C[531] = -3.236391798170e-03;
C[532] = 3.967023490588e-01;
C[533] = 8.818494341009e-03;
C[534] = -6.372833925258e-04;
C[535] = 5.566180816134e-02;
C[536] = 3.248646217488e-03;
C[537] = 1.146634263618e-03;
C[538] = 1.262949744601e-01;
C[539] = -1.335420782201e-04;
C[540] = -3.694014510925e-04;
C[541] = 3.262281655354e-02;
C[542] = 8.128925975125e-04;
C[543] = -3.610418689080e-05;
C[544] = 2.711740090989e-03;
C[545] = 1.696190132477e-04;
C[546] = -3.944376248653e-05;
C[547] = 2.679586497099e-03;
C[548] = 2.063009715216e-04;
C[549] = -9.676653058815e-07;
C[550] = -1.567881144151e-05;
C[551] = 8.854489832203e-07;
C[552] = -4.161629051452e-05;
C[553] = -2.134539658842e-03;
C[554] = 5.874563443540e-02;
C[555] = -2.962507116397e-03;
C[556] = 9.091112895599e-03;
C[557] = 3.867653557393e-01;
C[558] = -4.167535944819e-04;
C[559] = 3.168262367539e-03;
C[560] = 5.588094727226e-02;
C[561] = 1.663638786080e-05;
C[562] = -1.362047003927e-04;
C[563] = 1.293560230866e-01;
C[564] = -2.660940121030e-04;
C[565] = 8.273433499542e-04;
C[566] = 3.167878717563e-02;
C[567] = -1.820341080551e-05;
C[568] = 1.641299406929e-04;
C[569] = 2.726103689625e-03;
C[570] = 5.438164377994e-05;
C[571] = 1.883027651974e-04;
C[572] = 2.847541825213e-03;
C[573] = 1.171726724637e-06;
C[574] = 5.217515488603e-07;
C[575] = -1.557416973897e-05;
C[576] = 3.854324121580e-01;
C[577] = -3.212175489393e-03;
C[578] = -3.069733071083e-03;
C[579] = 1.152254659298e-01;
C[580] = 4.080665807728e-03;
C[581] = 1.612088439155e-03;
C[582] = 3.834005008007e-01;
C[583] = 1.093748977435e-03;
C[584] = -3.866524374247e-05;
C[585] = 3.147204161686e-02;
C[586] = -3.677870305074e-04;
C[587] = -2.732424090820e-04;
C[588] = 1.322978782606e-01;
C[589] = 3.486278844136e-04;
C[590] = 1.007815336737e-04;
C[591] = 3.117662476890e-02;
C[592] = 1.108765195948e-04;
C[593] = 1.009976152607e-04;
C[594] = 5.821212959740e-02;
C[595] = 3.153307954692e-04;
C[596] = 1.652616408602e-03;
C[597] = 2.532928034896e-03;
C[598] = 1.542609529829e-05;
C[599] = 7.952490154687e-05;
C[600] = -3.176164175397e-03;
C[601] = 3.967806134185e-01;
C[602] = 8.677982522171e-03;
C[603] = 4.070614407401e-03;
C[604] = 9.401917676268e-02;
C[605] = -1.892766443115e-02;
C[606] = 1.035296133635e-03;
C[607] = 3.867692694857e-01;
C[608] = 7.683520760050e-03;
C[609] = -3.659630576884e-04;
C[610] = 3.262803417752e-02;
C[611] = 7.998013250591e-04;
C[612] = 3.479577910585e-04;
C[613] = 1.304374304260e-01;
C[614] = -1.753127504668e-03;
C[615] = 1.078672279719e-04;
C[616] = 3.157101173022e-02;
C[617] = 8.210692398774e-04;
C[618] = 3.538053122566e-04;
C[619] = 5.855172014707e-02;
C[620] = 2.065811302514e-03;
C[621] = 1.744049344989e-05;
C[622] = 2.556787146860e-03;
C[623] = 1.052767725778e-04;
C[624] = -2.973123022906e-03;
C[625] = 8.400775520127e-03;
C[626] = 3.867257888335e-01;
C[627] = 1.590502737480e-03;
C[628] = -1.890617193560e-02;
C[629] = 9.383870419663e-02;
C[630] = -1.233325473360e-04;
C[631] = 7.878432008930e-03;
C[632] = 3.973286988376e-01;
C[633] = -2.679632029408e-04;
C[634] = 7.850446761204e-04;
C[635] = 3.167614938190e-02;
C[636] = 9.934248689538e-05;
C[637] = -1.751694671631e-03;
C[638] = 1.302960135799e-01;
C[639] = 9.692534693297e-05;
C[640] = 8.313685093662e-04;
C[641] = 3.267075512008e-02;
C[642] = 1.708331668869e-03;
C[643] = 1.986447550591e-03;
C[644] = 6.416894990180e-02;
C[645] = 8.224978293110e-05;
C[646] = 1.007826510307e-04;
C[647] = 2.820225071948e-03;
C[648] = 5.203867328785e-02;
C[649] = -4.829515531991e-04;
C[650] = -1.688179603209e-04;
C[651] = 3.836183541223e-01;
C[652] = 9.667323955710e-04;
C[653] = 4.341001399029e-05;
C[654] = 5.371525405780e-02;
C[655] = -8.822714089450e-04;
C[656] = -9.149712396158e-03;
C[657] = 2.510734563120e-03;
C[658] = -2.581539760235e-05;
C[659] = -1.674368528107e-06;
C[660] = 3.119114832368e-02;
C[661] = 1.032963121009e-04;
C[662] = 1.080415176881e-04;
C[663] = 1.291118652014e-01;
C[664] = -1.271697829576e-04;
C[665] = -7.666054142474e-04;
C[666] = 5.082113271974e-01;
C[667] = 4.934463569096e-04;
C[668] = 9.249603228234e-03;
C[669] = 3.720345829210e-02;
C[670] = 5.010381122821e-05;
C[671] = 6.592712693976e-04;
C[672] = -5.141105938143e-04;
C[673] = 5.541739365993e-02;
C[674] = 3.259071612458e-03;
C[675] = 1.054797115438e-03;
C[676] = 3.865832965190e-01;
C[677] = 7.687244026797e-03;
C[678] = -8.839623543324e-04;
C[679] = 5.175592575630e-02;
C[680] = -1.086403890623e-02;
C[681] = -2.810951145813e-05;
C[682] = 2.695445790895e-03;
C[683] = 1.701838903542e-04;
C[684] = 1.082797287950e-04;
C[685] = 3.155861353244e-02;
C[686] = 8.186226438906e-04;
C[687] = -1.272825126501e-04;
C[688] = 1.288441635312e-01;
C[689] = -1.068524975343e-03;
C[690] = 4.446253184381e-04;
C[691] = 5.034485301231e-01;
C[692] = -2.364662976601e-04;
C[693] = 4.774207878001e-05;
C[694] = 3.691767925534e-02;
C[695] = 7.624566421521e-05;
C[696] = -2.082852440187e-04;
C[697] = 3.350309041113e-03;
C[698] = 5.562790486170e-02;
C[699] = 1.775219350262e-04;
C[700] = 7.484012259304e-03;
C[701] = 3.973635435398e-01;
C[702] = -9.157714654187e-03;
C[703] = -1.087823424334e-02;
C[704] = 1.988085414121e-02;
C[705] = -4.590460298320e-06;
C[706] = 1.763965348226e-04;
C[707] = 2.709234195588e-03;
C[708] = 1.154100938453e-04;
C[709] = 8.077686731610e-04;
C[710] = 3.267307810022e-02;
C[711] = -7.671388981160e-04;
C[712] = -1.069471331150e-03;
C[713] = 1.260274992528e-01;
C[714] = 9.172120303747e-03;
C[715] = -1.332289894224e-04;
C[716] = 5.244900829115e-01;
C[717] = 6.556079030788e-04;
C[718] = 8.139270144037e-05;
C[719] = 3.860463518371e-02;
C[720] = -7.043981539809e+00;
C[721] = 5.566336206082e-02;
C[722] = -1.081945613453e-02;
C[723] = 2.155969701737e+00;
C[724] = 8.635102475050e-03;
C[725] = -1.232311751498e-02;
C[726] = 8.152514044809e-01;
C[727] = -8.738139182775e-03;
C[728] = -7.561142080528e-03;
C[729] = 6.275106133501e-02;
C[730] = 1.305642100319e-02;
C[731] = -3.469530279048e-04;
C[732] = 3.877397336284e-01;
C[733] = -3.567423480035e-03;
C[734] = -3.207682203576e-03;
C[735] = 5.252611160661e-02;
C[736] = -6.306222475334e-04;
C[737] = -3.876001006495e-04;
C[738] = 8.019935137426e-02;
C[739] = -9.597513787296e-04;
C[740] = 1.406478947373e-03;
C[741] = 2.481497748616e-03;
C[742] = -3.790887082976e-05;
C[743] = 6.063410760318e-05;
C[744] = 5.567177039102e-02;
C[745] = -7.206264780642e+00;
C[746] = 2.894548007733e-02;
C[747] = 7.793865281243e-03;
C[748] = 2.115406353882e+00;
C[749] = 1.471833876048e-02;
C[750] = -8.472859005569e-03;
C[751] = 8.559327748365e-01;
C[752] = 4.591348332539e-02;
C[753] = 1.305684141970e-02;
C[754] = 2.324754465761e-02;
C[755] = 1.340144322810e-03;
C[756] = -3.623974102005e-03;
C[757] = 3.968071603907e-01;
C[758] = 8.281340586313e-03;
C[759] = -5.982136658858e-04;
C[760] = 5.555370714706e-02;
C[761] = 3.238520543699e-03;
C[762] = -7.636788382199e-04;
C[763] = 8.467500787596e-02;
C[764] = 5.868280541895e-03;
C[765] = -2.936028619702e-05;
C[766] = 2.660676412178e-03;
C[767] = 2.111187120053e-04;
C[768] = -1.046643032058e-02;
C[769] = 2.814669342557e-02;
C[770] = -7.056424555600e+00;
C[771] = -1.389302770210e-02;
C[772] = 1.859471782715e-02;
C[773] = 2.120575073560e+00;
C[774] = -7.175898868503e-03;
C[775] = 4.523426707125e-02;
C[776] = 8.607054731066e-01;
C[777] = -3.293017372073e-04;
C[778] = 1.300204990222e-03;
C[779] = 5.794652358970e-02;
C[780] = -3.321316957365e-03;
C[781] = 8.557691834154e-03;
C[782] = 3.881688703753e-01;
C[783] = -3.430626681414e-04;
C[784] = 3.160223999376e-03;
C[785] = 5.593714392383e-02;
C[786] = 1.671726246419e-03;
C[787] = 5.431140607289e-03;
C[788] = 8.995982265470e-02;
C[789] = 7.184284955016e-05;
C[790] = 1.922353168343e-04;
C[791] = 2.836625532113e-03;
C[792] = 2.152592238179e+00;
C[793] = 8.278191456796e-03;
C[794] = -1.603754679582e-02;
C[795] = -6.235107280711e+00;
C[796] = 1.040244196657e-02;
C[797] = 1.299346816217e-04;
C[798] = 2.121144985310e+00;
C[799] = 1.156997309430e-02;
C[800] = -9.220090443902e-03;
C[801] = 3.875708604505e-01;
C[802] = -3.599757793227e-03;
C[803] = -3.428542912051e-03;
C[804] = 1.150582805690e-01;
C[805] = 3.582037544873e-03;
C[806] = 4.050093125358e-04;
C[807] = 3.816142780386e-01;
C[808] = 1.882129000081e-03;
C[809] = 1.516007245619e-03;
C[810] = 9.837629893584e-01;
C[811] = 6.182414901889e-03;
C[812] = 2.753904223005e-02;
C[813] = 5.812252243975e-02;
C[814] = 3.723547022708e-04;
C[815] = 1.631901104695e-03;
C[816] = 8.692340702629e-03;
C[817] = 2.116971641075e+00;
C[818] = 1.033211035860e-02;
C[819] = 1.020141396005e-02;
C[820] = -6.284538273018e+00;
C[821] = -4.076580623256e-02;
C[822] = 1.082320862402e-02;
C[823] = 2.106318591909e+00;
C[824] = -6.041678045117e-03;
C[825] = -3.564561568656e-03;
C[826] = 3.968854247503e-01;
C[827] = 8.144561460726e-03;
C[828] = 3.571986144547e-03;
C[829] = 9.576021355526e-02;
C[830] = -1.884544191253e-02;
C[831] = 1.820741088049e-03;
C[832] = 3.850102135905e-01;
C[833] = 8.268582667575e-03;
C[834] = 6.679246173490e-03;
C[835] = 9.873626448835e-01;
C[836] = 3.181654289111e-02;
C[837] = 4.048119013689e-04;
C[838] = 5.854706439256e-02;
C[839] = 2.164163133971e-03;
C[840] = -1.465341493853e-02;
C[841] = 6.363962342851e-03;
C[842] = 2.119783735444e+00;
C[843] = -3.017793518872e-04;
C[844] = -4.033595632166e-02;
C[845] = -6.287148434094e+00;
C[846] = -1.022033352381e-02;
C[847] = -3.407729945425e-03;
C[848] = 2.109756509353e+00;
C[849] = -3.324197074754e-03;
C[850] = 7.863621765431e-03;
C[851] = 3.881293034694e-01;
C[852] = 3.834236108603e-04;
C[853] = -1.882394941699e-02;
C[854] = 9.293833846160e-02;
C[855] = 1.436064771517e-03;
C[856] = 8.458026854654e-03;
C[857] = 3.967484524215e-01;
C[858] = 2.821141245796e-02;
C[859] = 3.076852738594e-02;
C[860] = 1.078703357571e+00;
C[861] = 1.674368730491e-03;
C[862] = 2.088059866109e-03;
C[863] = 6.422119218418e-02;
C[864] = 8.055672396897e-01;
C[865] = -5.386222219034e-03;
C[866] = -2.217186185283e-03;
C[867] = 2.125502051742e+00;
C[868] = 9.451933862742e-03;
C[869] = -6.885482297281e-03;
C[870] = -7.095395551481e+00;
C[871] = -3.431652796354e-03;
C[872] = -4.309931824146e-02;
C[873] = 5.204190336705e-02;
C[874] = -4.438818265591e-04;
C[875] = -9.512703398041e-05;
C[876] = 3.818321313602e-01;
C[877] = 1.752177349985e-03;
C[878] = 1.602807332843e-03;
C[879] = 5.549567987072e-02;
C[880] = -1.890464751761e-03;
C[881] = -9.058351190897e-03;
C[882] = 4.086165473246e+00;
C[883] = 2.192722974829e-03;
C[884] = 5.178683148177e-02;
C[885] = 5.081872453312e-01;
C[886] = 6.843031729004e-04;
C[887] = 7.534021160703e-03;
C[888] = -5.798539178622e-03;
C[889] = 8.510444848083e-01;
C[890] = 4.705045196963e-02;
C[891] = 1.079093585435e-02;
C[892] = 2.102599132575e+00;
C[893] = -7.231489588100e-03;
C[894] = -3.465471704103e-03;
C[895] = -7.093337849613e+00;
C[896] = -2.853308714279e-02;
C[897] = -4.836422473258e-04;
C[898] = 5.530929264565e-02;
C[899] = 3.251033244296e-03;
C[900] = 1.843177138084e-03;
C[901] = 3.848242406238e-01;
C[902] = 8.266838872520e-03;
C[903] = -1.892155697148e-03;
C[904] = 5.223815489301e-02;
C[905] = -1.241339818919e-02;
C[906] = 1.433427235026e-03;
C[907] = 4.051017733798e+00;
C[908] = -1.507555003542e-02;
C[909] = 6.404257109028e-04;
C[910] = 5.048254494347e-01;
C[911] = 7.454163425568e-04;
C[912] = -2.662583365528e-03;
C[913] = 4.794673979789e-02;
C[914] = 8.556446248954e-01;
C[915] = -4.896346868529e-03;
C[916] = -1.003184806004e-02;
C[917] = 2.110453403396e+00;
C[918] = -4.325936340206e-02;
C[919] = -2.881699388491e-02;
C[920] = -7.219832070970e+00;
C[921] = -1.426721648995e-04;
C[922] = 3.340183367324e-03;
C[923] = 5.568410151327e-02;
C[924] = 1.732194424388e-03;
C[925] = 8.069074166829e-03;
C[926] = 3.967832971237e-01;
C[927] = -9.066353448927e-03;
C[928] = -1.242759352630e-02;
C[929] = 1.934699285004e-02;
C[930] = 5.061711606946e-02;
C[931] = -1.342916075670e-02;
C[932] = 4.169989353848e+00;
C[933] = 7.469670949646e-03;
C[934] = 8.462507876063e-04;
C[935] = 5.255319319065e-01;
C[936] = 0.000000000000e+00;
C[937] = 0.000000000000e+00;
C[938] = 0.000000000000e+00;
C[939] = 0.000000000000e+00;
C[940] = 0.000000000000e+00;
C[941] = 0.000000000000e+00;
C[942] = 0.000000000000e+00;
C[943] = 0.000000000000e+00;
C[944] = 0.000000000000e+00;
C[945] = 0.000000000000e+00;
C[946] = 0.000000000000e+00;
C[947] = 0.000000000000e+00;
C[948] = 0.000000000000e+00;
C[949] = 0.000000000000e+00;
C[950] = 0.000000000000e+00;
C[951] = 0.000000000000e+00;
C[952] = 0.000000000000e+00;
C[953] = 0.000000000000e+00;
C[954] = 1.000000000000e+00;
C[955] = 0.000000000000e+00;
C[956] = 0.000000000000e+00;
C[957] = 2.000000000000e-01;
C[958] = -0.000000000000e+00;
C[959] = -0.000000000000e+00;
C[960] = 0.000000000000e+00;
C[961] = 0.000000000000e+00;
C[962] = 0.000000000000e+00;
C[963] = 0.000000000000e+00;
C[964] = 0.000000000000e+00;
C[965] = 0.000000000000e+00;
C[966] = 0.000000000000e+00;
C[967] = 0.000000000000e+00;
C[968] = 0.000000000000e+00;
C[969] = 0.000000000000e+00;
C[970] = 0.000000000000e+00;
C[971] = 0.000000000000e+00;
C[972] = 0.000000000000e+00;
C[973] = 0.000000000000e+00;
C[974] = 0.000000000000e+00;
C[975] = 0.000000000000e+00;
C[976] = 0.000000000000e+00;
C[977] = 0.000000000000e+00;
C[978] = 0.000000000000e+00;
C[979] = 1.000000000000e+00;
C[980] = 0.000000000000e+00;
C[981] = -0.000000000000e+00;
C[982] = 2.000000000000e-01;
C[983] = -0.000000000000e+00;
C[984] = 0.000000000000e+00;
C[985] = 0.000000000000e+00;
C[986] = 0.000000000000e+00;
C[987] = 0.000000000000e+00;
C[988] = 0.000000000000e+00;
C[989] = 0.000000000000e+00;
C[990] = 0.000000000000e+00;
C[991] = 0.000000000000e+00;
C[992] = 0.000000000000e+00;
C[993] = 0.000000000000e+00;
C[994] = 0.000000000000e+00;
C[995] = 0.000000000000e+00;
C[996] = 0.000000000000e+00;
C[997] = 0.000000000000e+00;
C[998] = 0.000000000000e+00;
C[999] = 0.000000000000e+00;
C[1000] = 0.000000000000e+00;
C[1001] = 0.000000000000e+00;
C[1002] = 0.000000000000e+00;
C[1003] = 0.000000000000e+00;
C[1004] = 1.000000000000e+00;
C[1005] = -0.000000000000e+00;
C[1006] = -0.000000000000e+00;
C[1007] = 2.000000000000e-01;
C[1008] = 6.976237823532e-02;
C[1009] = 1.172357307160e-02;
C[1010] = -4.850763682856e-04;
C[1011] = 3.844852191425e-01;
C[1012] = -4.515778930404e-03;
C[1013] = -3.530087247153e-03;
C[1014] = 5.015773033666e-02;
C[1015] = -1.693851884531e-04;
C[1016] = 2.649406051235e-04;
C[1017] = 1.310490943219e-01;
C[1018] = 8.166483364376e-04;
C[1019] = -8.039734747161e-05;
C[1020] = 3.063308245628e-02;
C[1021] = -3.133676884073e-04;
C[1022] = -2.036523252827e-04;
C[1023] = 2.383820426454e-03;
C[1024] = -4.319197938783e-06;
C[1025] = 1.888744767791e-05;
C[1026] = 2.248776197457e-03;
C[1027] = 1.082209276453e-05;
C[1028] = 6.287481522772e-05;
C[1029] = -2.718272607703e-05;
C[1030] = 1.134995095612e-06;
C[1031] = 2.423471421814e-06;
C[1032] = 1.153430375823e-02;
C[1033] = 2.480103884499e-02;
C[1034] = 3.575501681373e-03;
C[1035] = -4.458251072817e-03;
C[1036] = 3.975086863043e-01;
C[1037] = 9.862623765826e-03;
C[1038] = -9.803040042404e-05;
C[1039] = 5.550184280636e-02;
C[1040] = 3.800140766939e-03;
C[1041] = 8.040303822129e-04;
C[1042] = 1.264528310791e-01;
C[1043] = 1.790895219002e-04;
C[1044] = -3.084744367318e-04;
C[1045] = 3.275223178483e-02;
C[1046] = 1.009711136992e-03;
C[1047] = -5.600493721356e-07;
C[1048] = 2.712047519725e-03;
C[1049] = 2.043085483004e-04;
C[1050] = 1.863764929401e-05;
C[1051] = 2.723494134595e-03;
C[1052] = 2.839454342529e-04;
C[1053] = 1.312613305959e-06;
C[1054] = -1.503756782086e-05;
C[1055] = 2.520617629696e-06;
C[1056] = -5.016196258597e-04;
C[1057] = 3.536353489705e-03;
C[1058] = 6.118924959698e-02;
C[1059] = -3.591431072989e-03;
C[1060] = 1.035289933037e-02;
C[1061] = 3.843459888010e-01;
C[1062] = 3.130008487946e-04;
C[1063] = 3.647763506173e-03;
C[1064] = 5.359108482617e-02;
C[1065] = -8.150023130988e-05;
C[1066] = 1.764796424556e-04;
C[1067] = 1.301661002194e-01;
C[1068] = -2.057478530593e-04;
C[1069] = 1.038579019996e-03;
C[1070] = 3.079757096301e-02;
C[1071] = 2.122082647694e-05;
C[1072] = 1.932578556640e-04;
C[1073] = 2.570774407721e-03;
C[1074] = 6.668775807146e-05;
C[1075] = 2.500265343985e-04;
C[1076] = 2.607983948296e-03;
C[1077] = 2.462443588420e-06;
C[1078] = 1.448921080487e-06;
C[1079] = -2.398863348110e-05;
C[1080] = 3.842290972071e-01;
C[1081] = -4.376678988591e-03;
C[1082] = -3.550478741099e-03;
C[1083] = 1.192716106274e-01;
C[1084] = 3.151784633462e-03;
C[1085] = -9.403162906817e-04;
C[1086] = 3.835212909735e-01;
C[1087] = 2.326502536634e-03;
C[1088] = 3.236201421369e-03;
C[1089] = 3.061600766059e-02;
C[1090] = -3.030362977834e-04;
C[1091] = -2.030176976000e-04;
C[1092] = 1.332033147859e-01;
C[1093] = 1.754993190181e-04;
C[1094] = -1.216525707802e-04;
C[1095] = 3.130064591281e-02;
C[1096] = 1.721502746993e-04;
C[1097] = 2.649625286316e-04;
C[1098] = 5.865739886575e-02;
C[1099] = 4.182106356858e-04;
C[1100] = 1.053015291352e-03;
C[1101] = 2.561048950961e-03;
C[1102] = 1.681827994253e-05;
C[1103] = 4.533608841728e-05;
C[1104] = -4.162975349868e-03;
C[1105] = 3.969857514784e-01;
C[1106] = 9.935799704705e-03;
C[1107] = 3.057526483940e-03;
C[1108] = 9.615982520165e-02;
C[1109] = -2.001779754253e-02;
C[1110] = 2.231857588025e-03;
C[1111] = 3.838294567946e-01;
C[1112] = 7.608093498435e-03;
C[1113] = -2.898474497049e-04;
C[1114] = 3.271736946310e-02;
C[1115] = 1.010772378285e-03;
C[1116] = 1.692154423833e-04;
C[1117] = 1.303773953381e-01;
C[1118] = -1.935702170494e-03;
C[1119] = 1.672590715885e-04;
C[1120] = 3.149856495612e-02;
C[1121] = 7.909141853770e-04;
C[1122] = 4.225787145433e-04;
C[1123] = 5.930766749087e-02;
C[1124] = 2.424892587388e-03;
C[1125] = 1.720930162633e-05;
C[1126] = 2.602650602597e-03;
C[1127] = 1.192404437873e-04;
C[1128] = -3.403078280291e-03;
C[1129] = 9.620095746385e-03;
C[1130] = 3.846000505272e-01;
C[1131] = -9.783863163799e-04;
C[1132] = -2.019506276749e-02;
C[1133] = 9.723249002822e-02;
C[1134] = 3.166410787448e-03;
C[1135] = 7.858863729335e-03;
C[1136] = 3.980918511983e-01;
C[1137] = -1.951850608252e-04;
C[1138] = 9.935426023627e-04;
C[1139] = 3.081450841142e-02;
C[1140] = -1.241905724934e-04;
C[1141] = -1.947519852159e-03;
C[1142] = 1.312242800206e-01;
C[1143] = 2.615800053111e-04;
C[1144] = 8.061743330012e-04;
C[1145] = 3.274137690185e-02;
C[1146] = 1.051461038119e-03;
C[1147] = 2.391189521334e-03;
C[1148] = 6.468466397993e-02;
C[1149] = 4.543384464686e-05;
C[1150] = 1.170340878462e-04;
C[1151] = 2.843021331953e-03;
C[1152] = 5.195636438397e-02;
C[1153] = -4.872004198415e-04;
C[1154] = 1.332966880976e-05;
C[1155] = 3.817115734833e-01;
C[1156] = 2.664521516838e-03;
C[1157] = 3.689371845480e-03;
C[1158] = 5.481639693633e-02;
C[1159] = -2.503768605501e-03;
C[1160] = -6.487656455649e-03;
C[1161] = 2.503729362942e-03;
C[1162] = -2.650471733330e-05;
C[1163] = 1.242747811282e-06;
C[1164] = 3.117999808012e-02;
C[1165] = 1.961033335093e-04;
C[1166] = 2.964440758466e-04;
C[1167] = 1.289895499633e-01;
C[1168] = -1.899228748052e-04;
C[1169] = -4.314817179071e-04;
C[1170] = 5.091538964147e-01;
C[1171] = 4.102838978090e-04;
C[1172] = 2.761109795883e-03;
C[1173] = 3.734473403422e-02;
C[1174] = 2.141148808207e-05;
C[1175] = 1.336662043776e-04;
C[1176] = -5.398192501364e-04;
C[1177] = 5.543329730617e-02;
C[1178] = 3.306702911423e-03;
C[1179] = 2.713005925103e-03;
C[1180] = 3.837374878805e-01;
C[1181] = 7.952364792651e-03;
C[1182] = -2.480545451523e-03;
C[1183] = 4.924239239906e-02;
C[1184] = -1.442540131266e-02;
C[1185] = -2.901480205100e-05;
C[1186] = 2.707477819713e-03;
C[1187] = 1.705204826807e-04;
C[1188] = 1.979171672639e-04;
C[1189] = 3.149243369518e-02;
C[1190] = 8.124077372223e-04;
C[1191] = -1.883746645400e-04;
C[1192] = 1.284249061217e-01;
C[1193] = -1.315085542928e-03;
C[1194] = 3.948338250480e-04;
C[1195] = 5.087701933654e-01;
C[1196] = 3.061676472128e-03;
C[1197] = 2.063823564000e-05;
C[1198] = 3.738797790827e-02;
C[1199] = 3.292505342109e-04;
C[1200] = -4.566756751297e-05;
C[1201] = 3.404742323413e-03;
C[1202] = 5.528546358870e-02;
C[1203] = 3.725583570643e-03;
C[1204] = 7.785775292192e-03;
C[1205] = 3.964900044390e-01;
C[1206] = -6.443334011702e-03;
C[1207] = -1.445897782584e-02;
C[1208] = 1.883450995114e-02;
C[1209] = -1.819763831185e-06;
C[1210] = 1.779486520654e-04;
C[1211] = 2.683732991890e-03;
C[1212] = 2.975880052499e-04;
C[1213] = 8.027596382942e-04;
C[1214] = 3.263458711790e-02;
C[1215] = -4.285268883107e-04;
C[1216] = -1.317323977140e-03;
C[1217] = 1.259633673307e-01;
C[1218] = 2.746192757986e-03;
C[1219] = 3.135625271643e-03;
C[1220] = 5.268095362616e-01;
C[1221] = 1.327130417370e-04;
C[1222] = 3.328570539690e-04;
C[1223] = 3.873563601026e-02;
C[1224] = -6.992765367324e+00;
C[1225] = 4.068143163330e-02;
C[1226] = -1.323562540569e-02;
C[1227] = 2.146142085148e+00;
C[1228] = -2.092770449626e-03;
C[1229] = -1.742993251229e-02;
C[1230] = 7.705530065150e-01;
C[1231] = 2.352680949702e-04;
C[1232] = 4.643039308771e-03;
C[1233] = 8.503515038045e-02;
C[1234] = 7.258399328246e-03;
C[1235] = -2.010760027384e-03;
C[1236] = 3.711671950145e-01;
C[1237] = -1.773491674206e-03;
C[1238] = -1.662618437032e-03;
C[1239] = 4.948842330920e-02;
C[1240] = -2.479225121891e-05;
C[1241] = 3.188082448074e-04;
C[1242] = 7.535226580706e-02;
C[1243] = 6.182994427669e-05;
C[1244] = 1.293933435189e-03;
C[1245] = 2.307458176893e-03;
C[1246] = -1.010149762383e-06;
C[1247] = 5.022490597346e-05;
C[1248] = 3.689604536589e-02;
C[1249] = -7.192165326234e+00;
C[1250] = 5.692060243463e-02;
C[1251] = -5.558082729284e-04;
C[1252] = 2.130400727539e+00;
C[1253] = 2.895739330591e-02;
C[1254] = 1.075473671686e-03;
C[1255] = 8.553348243201e-01;
C[1256] = 5.631121728281e-02;
C[1257] = 7.069130014875e-03;
C[1258] = 2.477789432042e-02;
C[1259] = 3.787946188382e-03;
C[1260] = -1.721100851938e-03;
C[1261] = 3.993883529529e-01;
C[1262] = 1.218124918793e-02;
C[1263] = 2.880431795456e-05;
C[1264] = 5.573858070943e-02;
C[1265] = 3.908311870081e-03;
C[1266] = 2.207745624527e-04;
C[1267] = 8.628461949099e-02;
C[1268] = 7.648166774835e-03;
C[1269] = 4.752635729229e-06;
C[1270] = 2.731028537881e-03;
C[1271] = 2.677484995959e-04;
C[1272] = -1.356649055718e-02;
C[1273] = 5.613763860126e-02;
C[1274] = -7.041126462747e+00;
C[1275] = -1.790514345853e-02;
C[1276] = 3.694209516303e-02;
C[1277] = 2.096682732860e+00;
C[1278] = 5.049746324233e-03;
C[1279] = 5.426342003423e-02;
C[1280] = 8.165900542333e-01;
C[1281] = -2.027303284958e-03;
C[1282] = 3.748797996714e-03;
C[1283] = 7.272432394096e-02;
C[1284] = -1.703551441469e-03;
C[1285] = 1.274245420519e-02;
C[1286] = 3.714689665585e-01;
C[1287] = 3.481750423647e-04;
C[1288] = 3.721268077806e-03;
C[1289] = 5.282853319797e-02;
C[1290] = 1.385985206590e-03;
C[1291] = 6.906549323982e-03;
C[1292] = 8.511550262614e-02;
C[1293] = 5.407986690488e-05;
C[1294] = 2.361114279836e-04;
C[1295] = 2.659371342814e-03;
C[1296] = 2.141019646440e+00;
C[1297] = 1.075633411591e-03;
C[1298] = -1.708609682074e-02;
C[1299] = -6.208653248660e+00;
C[1300] = 6.881726037393e-03;
C[1301] = -3.842751963509e-03;
C[1302] = 2.159768355641e+00;
C[1303] = 7.212518247064e-03;
C[1304] = 1.311730892351e-03;
C[1305] = 3.709110730791e-01;
C[1306] = -1.639528767712e-03;
C[1307] = -1.662599109579e-03;
C[1308] = 1.308844685136e-01;
C[1309] = 7.855640995529e-04;
C[1310] = -1.761608597721e-03;
C[1311] = 3.858729451619e-01;
C[1312] = 1.726171969978e-03;
C[1313] = 2.313625260510e-03;
C[1314] = 9.950903830423e-01;
C[1315] = 6.485826885285e-03;
C[1316] = 1.542598817993e-02;
C[1317] = 5.881626697111e-02;
C[1318] = 3.704297943416e-04;
C[1319] = 8.432962290975e-04;
C[1320] = 4.963301161097e-03;
C[1321] = 2.119942031020e+00;
C[1322] = 2.860010264967e-02;
C[1323] = 4.996563046947e-03;
C[1324] = -6.285636364741e+00;
C[1325] = -4.466240092111e-02;
C[1326] = 5.926395502168e-03;
C[1327] = 2.090808651262e+00;
C[1328] = -1.988535850429e-02;
C[1329] = -1.420688093670e-03;
C[1330] = 3.988654181269e-01;
C[1331] = 1.232535457953e-02;
C[1332] = 6.913059500306e-04;
C[1333] = 9.234565724514e-02;
C[1334] = -2.173282767042e-02;
C[1335] = 1.652638186826e-03;
C[1336] = 3.854984294405e-01;
C[1337] = 6.943726906739e-03;
C[1338] = 6.478785883987e-03;
C[1339] = 1.002625330082e+00;
C[1340] = 3.645095897382e-02;
C[1341] = 3.701013617824e-04;
C[1342] = 5.961257376437e-02;
C[1343] = 2.462425741755e-03;
C[1344] = -1.488975317505e-02;
C[1345] = 2.410683291708e-02;
C[1346] = 2.101763967384e+00;
C[1347] = -4.604152477474e-03;
C[1348] = -4.820770542034e-02;
C[1349] = -6.263483691972e+00;
C[1350] = 4.871201907512e-04;
C[1351] = -1.598306629339e-02;
C[1352] = 2.144142140299e+00;
C[1353] = -1.535609470170e-03;
C[1354] = 1.193872116849e-02;
C[1355] = 3.717230282847e-01;
C[1356] = -1.799678623420e-03;
C[1357] = -2.191009289538e-02;
C[1358] = 1.096875309182e-01;
C[1359] = 2.271984350835e-03;
C[1360] = 7.230252198903e-03;
C[1361] = 3.991445472722e-01;
C[1362] = 1.534369161483e-02;
C[1363] = 3.625069128187e-02;
C[1364] = 1.088401425259e+00;
C[1365] = 8.341372358263e-04;
C[1366] = 2.437516123748e-03;
C[1367] = 6.464281316846e-02;
C[1368] = 8.065256874612e-01;
C[1369] = -6.707926716663e-03;
C[1370] = -9.436772754642e-04;
C[1371] = 2.123574005836e+00;
C[1372] = 1.457967407842e-02;
C[1373] = 1.094634135139e-02;
C[1374] = -7.102257875295e+00;
C[1375] = -4.260926831119e-03;
C[1376] = -1.475094575755e-02;
C[1377] = 5.128705735651e-02;
C[1378] = -3.603657014629e-04;
C[1379] = 4.850386237989e-05;
C[1380] = 3.840632276716e-01;
C[1381] = 2.085302115638e-03;
C[1382] = 2.794945408867e-03;
C[1383] = 5.149074397567e-02;
C[1384] = -1.571009628018e-03;
C[1385] = -3.110600138348e-03;
C[1386] = 4.097665677499e+00;
C[1387] = -1.715483225149e-03;
C[1388] = 3.868505925816e-03;
C[1389] = 5.109214265971e-01;
C[1390] = -9.849503471225e-05;
C[1391] = 2.278298476940e-04;
C[1392] = -7.173413138694e-03;
C[1393] = 8.539639143164e-01;
C[1394] = 4.744220813923e-02;
C[1395] = 1.494258601645e-02;
C[1396] = 2.088969272981e+00;
C[1397] = -1.411304502705e-02;
C[1398] = -3.796463751549e-03;
C[1399] = -7.122414851367e+00;
C[1400] = -4.371646829147e-02;
C[1401] = -3.952263129021e-04;
C[1402] = 5.567003520924e-02;
C[1403] = 3.380207483056e-03;
C[1404] = 2.112675358447e-03;
C[1405] = 3.854064605264e-01;
C[1406] = 7.323753262220e-03;
C[1407] = -1.547786474040e-03;
C[1408] = 4.442807327726e-02;
C[1409] = -1.480193293639e-02;
C[1410] = -1.946119189692e-03;
C[1411] = 4.090546266264e+00;
C[1412] = 7.624887867037e-03;
C[1413] = -1.074886457428e-04;
C[1414] = 5.116627043664e-01;
C[1415] = 3.971265498807e-03;
C[1416] = -1.569124143959e-03;
C[1417] = 4.840324841229e-02;
C[1418] = 8.504776294838e-01;
C[1419] = 1.109937387783e-02;
C[1420] = -1.633172262914e-02;
C[1421] = 2.112105205113e+00;
C[1422] = -1.386449687861e-02;
C[1423] = -4.438799855501e-02;
C[1424] = -7.224089354081e+00;
C[1425] = 8.200072170860e-06;
C[1426] = 3.512913426556e-03;
C[1427] = 5.452291196050e-02;
C[1428] = 2.803007409785e-03;
C[1429] = 7.121408700496e-03;
C[1430] = 3.975427005129e-01;
C[1431] = -3.066277694401e-03;
C[1432] = -1.483550944956e-02;
C[1433] = 1.882687972492e-02;
C[1434] = 3.595572603345e-03;
C[1435] = 8.764053038480e-03;
C[1436] = 4.181066580852e+00;
C[1437] = 2.205461170864e-04;
C[1438] = 4.035393853966e-03;
C[1439] = 5.266551229633e-01;
C[1440] = 0.000000000000e+00;
C[1441] = 0.000000000000e+00;
C[1442] = 0.000000000000e+00;
C[1443] = 0.000000000000e+00;
C[1444] = 0.000000000000e+00;
C[1445] = 0.000000000000e+00;
C[1446] = 0.000000000000e+00;
C[1447] = 0.000000000000e+00;
C[1448] = 0.000000000000e+00;
C[1449] = 0.000000000000e+00;
C[1450] = 0.000000000000e+00;
C[1451] = 0.000000000000e+00;
C[1452] = 0.000000000000e+00;
C[1453] = 0.000000000000e+00;
C[1454] = 0.000000000000e+00;
C[1455] = 0.000000000000e+00;
C[1456] = 0.000000000000e+00;
C[1457] = 0.000000000000e+00;
C[1458] = 1.000000000000e+00;
C[1459] = 0.000000000000e+00;
C[1460] = 0.000000000000e+00;
C[1461] = 2.000000000000e-01;
C[1462] = -0.000000000000e+00;
C[1463] = -0.000000000000e+00;
C[1464] = 0.000000000000e+00;
C[1465] = 0.000000000000e+00;
C[1466] = 0.000000000000e+00;
C[1467] = 0.000000000000e+00;
C[1468] = 0.000000000000e+00;
C[1469] = 0.000000000000e+00;
C[1470] = 0.000000000000e+00;
C[1471] = 0.000000000000e+00;
C[1472] = 0.000000000000e+00;
C[1473] = 0.000000000000e+00;
C[1474] = 0.000000000000e+00;
C[1475] = 0.000000000000e+00;
C[1476] = 0.000000000000e+00;
C[1477] = 0.000000000000e+00;
C[1478] = 0.000000000000e+00;
C[1479] = 0.000000000000e+00;
C[1480] = 0.000000000000e+00;
C[1481] = 0.000000000000e+00;
C[1482] = 0.000000000000e+00;
C[1483] = 1.000000000000e+00;
C[1484] = 0.000000000000e+00;
C[1485] = -0.000000000000e+00;
C[1486] = 2.000000000000e-01;
C[1487] = -0.000000000000e+00;
C[1488] = 0.000000000000e+00;
C[1489] = 0.000000000000e+00;
C[1490] = 0.000000000000e+00;
C[1491] = 0.000000000000e+00;
C[1492] = 0.000000000000e+00;
C[1493] = 0.000000000000e+00;
C[1494] = 0.000000000000e+00;
C[1495] = 0.000000000000e+00;
C[1496] = 0.000000000000e+00;
C[1497] = 0.000000000000e+00;
C[1498] = 0.000000000000e+00;
C[1499] = 0.000000000000e+00;
C[1500] = 0.000000000000e+00;
C[1501] = 0.000000000000e+00;
C[1502] = 0.000000000000e+00;
C[1503] = 0.000000000000e+00;
C[1504] = 0.000000000000e+00;
C[1505] = 0.000000000000e+00;
C[1506] = 0.000000000000e+00;
C[1507] = 0.000000000000e+00;
C[1508] = 1.000000000000e+00;
C[1509] = -0.000000000000e+00;
C[1510] = -0.000000000000e+00;
C[1511] = 2.000000000000e-01;
C[1512] = 9.748595754964e-02;
C[1513] = -1.622695142384e-02;
C[1514] = -2.445699500617e-03;
C[1515] = 3.606091660603e-01;
C[1516] = 1.804105380041e-02;
C[1517] = 1.271019591058e-03;
C[1518] = 5.015094181610e-02;
C[1519] = 3.883245769211e-03;
C[1520] = 6.697490346782e-05;
C[1521] = 1.321666101258e-01;
C[1522] = -2.011567866556e-03;
C[1523] = 1.096931049836e-04;
C[1524] = 2.973275093347e-02;
C[1525] = 1.990969718367e-03;
C[1526] = -2.505708225601e-05;
C[1527] = 2.451305315520e-03;
C[1528] = 2.207161840910e-04;
C[1529] = -5.320795040086e-06;
C[1530] = 2.491584058007e-03;
C[1531] = 2.915500189642e-04;
C[1532] = -2.579454197514e-05;
C[1533] = -1.270008841081e-05;
C[1534] = 5.798264087748e-06;
C[1535] = -1.511341724359e-06;
C[1536] = -1.703597383824e-02;
C[1537] = 3.433940424595e-02;
C[1538] = 1.536478933499e-02;
C[1539] = 1.856702465100e-02;
C[1540] = 3.973165291669e-01;
C[1541] = 4.412835381689e-03;
C[1542] = 3.889145076232e-03;
C[1543] = 5.440633893709e-02;
C[1544] = 1.501733718457e-03;
C[1545] = -2.065502694182e-03;
C[1546] = 1.282253901726e-01;
C[1547] = 1.467812435097e-03;
C[1548] = 2.029705902197e-03;
C[1549] = 3.203483147389e-02;
C[1550] = 8.684656029239e-05;
C[1551] = 2.203249499130e-04;
C[1552] = 2.615761238625e-03;
C[1553] = 5.889739639761e-05;
C[1554] = 2.885092045772e-04;
C[1555] = 2.577043526715e-03;
C[1556] = 6.476660669812e-05;
C[1557] = 5.698626476306e-06;
C[1558] = -2.330857925607e-05;
C[1559] = -3.839058200096e-06;
C[1560] = -2.486737859328e-03;
C[1561] = 1.491471026989e-02;
C[1562] = 8.835036116082e-02;
C[1563] = 1.672380356055e-03;
C[1564] = 5.237955385532e-03;
C[1565] = 3.483176649605e-01;
C[1566] = -6.737220643847e-06;
C[1567] = 1.368956173830e-03;
C[1568] = 4.944824589952e-02;
C[1569] = 1.069572144028e-04;
C[1570] = 1.437807164090e-03;
C[1571] = 1.317501831937e-01;
C[1572] = -1.379159994555e-06;
C[1573] = 1.321358626962e-04;
C[1574] = 2.821483169433e-02;
C[1575] = -1.014268897333e-05;
C[1576] = 5.141560719038e-05;
C[1577] = 2.391558673190e-03;
C[1578] = -3.630275483650e-05;
C[1579] = 5.190833750150e-05;
C[1580] = 2.465671284169e-03;
C[1581] = -1.750813721899e-06;
C[1582] = -3.883440475961e-06;
C[1583] = -1.772402959923e-05;
C[1584] = 3.603156114141e-01;
C[1585] = 1.829575412531e-02;
C[1586] = 1.334069196236e-03;
C[1587] = 1.387562350802e-01;
C[1588] = -1.864634048665e-02;
C[1589] = -1.692683341386e-03;
C[1590] = 3.876835429414e-01;
C[1591] = -3.419256606104e-03;
C[1592] = 8.166448810413e-04;
C[1593] = 2.971318062372e-02;
C[1594] = 2.011621200485e-03;
C[1595] = -2.393323731581e-05;
C[1596] = 1.336888156053e-01;
C[1597] = -1.991596555359e-03;
C[1598] = 6.140948784820e-05;
C[1599] = 3.163285787877e-02;
C[1600] = -2.268500513071e-04;
C[1601] = 5.407806819268e-07;
C[1602] = 5.991157938489e-02;
C[1603] = -4.280076652568e-04;
C[1604] = -2.553226280322e-04;
C[1605] = 2.615142063148e-03;
C[1606] = -1.494641045660e-05;
C[1607] = -1.627668482402e-05;
C[1608] = 1.911018342692e-02;
C[1609] = 3.961146061146e-01;
C[1610] = 5.183537335527e-03;
C[1611] = -1.921463798402e-02;
C[1612] = 9.316231536671e-02;
C[1613] = -1.279747919349e-02;
C[1614] = -3.432752581483e-03;
C[1615] = 3.892632559903e-01;
C[1616] = 8.119252696485e-03;
C[1617] = 2.062245026801e-03;
C[1618] = 3.195470327041e-02;
C[1619] = 1.285079926959e-04;
C[1620] = -2.029483055184e-03;
C[1621] = 1.307762661287e-01;
C[1622] = -8.643330977048e-04;
C[1623] = -2.269285056874e-04;
C[1624] = 3.197893863537e-02;
C[1625] = 7.703679724127e-04;
C[1626] = -4.293181596113e-04;
C[1627] = 6.119437292613e-02;
C[1628] = 2.506001827449e-03;
C[1629] = -1.495495176453e-05;
C[1630] = 2.687408501385e-03;
C[1631] = 1.177778481193e-04;
C[1632] = 1.191194935184e-03;
C[1633] = 4.879878064002e-03;
C[1634] = 3.492438953719e-01;
C[1635] = -1.831466260628e-03;
C[1636] = -1.298908327116e-02;
C[1637] = 1.304978435676e-01;
C[1638] = 8.867247019878e-04;
C[1639] = 8.310699950843e-03;
C[1640] = 4.032701087150e-01;
C[1641] = -3.037872598097e-05;
C[1642] = 1.179827391133e-04;
C[1643] = 2.827658038842e-02;
C[1644] = 5.215729323206e-05;
C[1645] = -8.771067028824e-04;
C[1646] = 1.336015504102e-01;
C[1647] = 4.095834318854e-06;
C[1648] = 7.794990701910e-04;
C[1649] = 3.302023727617e-02;
C[1650] = -2.530085523603e-04;
C[1651] = 2.492040751593e-03;
C[1652] = 6.492600902695e-02;
C[1653] = -1.630571115809e-05;
C[1654] = 1.164277096642e-04;
C[1655] = 2.842456549370e-03;
C[1656] = 4.860194881782e-02;
C[1657] = 2.893750618679e-03;
C[1658] = 1.795078108702e-04;
C[1659] = 3.893458362997e-01;
C[1660] = -2.487008357718e-03;
C[1661] = 5.211785425368e-04;
C[1662] = 4.680046504964e-02;
C[1663] = -9.380440385398e-05;
C[1664] = 5.322807136421e-04;
C[1665] = 2.348039115634e-03;
C[1666] = 1.539653194094e-04;
C[1667] = 2.273646460945e-06;
C[1668] = 3.174367743599e-02;
C[1669] = -1.638788907698e-04;
C[1670] = -2.027390964455e-05;
C[1671] = 1.282748901293e-01;
C[1672] = 5.646528586952e-05;
C[1673] = 1.335399552359e-04;
C[1674] = 5.130612728004e-01;
C[1675] = -4.746670706937e-04;
C[1676] = -1.235936743573e-03;
C[1677] = 3.765027301020e-02;
C[1678] = -4.702611034841e-05;
C[1679] = -1.161046765106e-04;
C[1680] = 2.973800288749e-03;
C[1681] = 5.604273645732e-02;
C[1682] = 2.049761915662e-03;
C[1683] = -2.579592726644e-03;
C[1684] = 3.875245401291e-01;
C[1685] = 7.532911732429e-03;
C[1686] = -7.825030905943e-05;
C[1687] = 3.914828287516e-02;
C[1688] = -1.413086408832e-02;
C[1689] = 1.600864853935e-04;
C[1690] = 2.724854406640e-03;
C[1691] = 9.680265664581e-05;
C[1692] = -1.708724593430e-04;
C[1693] = 3.186302424463e-02;
C[1694] = 7.276465222968e-04;
C[1695] = 5.750222552249e-05;
C[1696] = 1.276278967837e-01;
C[1697] = -1.226607894214e-03;
C[1698] = -4.712113007493e-04;
C[1699] = 5.144996181623e-01;
C[1700] = 4.593402899532e-03;
C[1701] = -4.697441401863e-05;
C[1702] = 3.779518104123e-02;
C[1703] = 4.030840564494e-04;
C[1704] = 2.520637936153e-04;
C[1705] = 2.237801339194e-03;
C[1706] = 4.945686272587e-02;
C[1707] = 3.891599012732e-04;
C[1708] = 7.286280059496e-03;
C[1709] = 4.035021731140e-01;
C[1710] = 5.721811914412e-04;
C[1711] = -1.413624949121e-02;
C[1712] = 1.901171253607e-02;
C[1713] = 7.018464303081e-06;
C[1714] = 1.079685711135e-04;
C[1715] = 2.392133128280e-03;
C[1716] = -2.795821796928e-05;
C[1717] = 7.148364632801e-04;
C[1718] = 3.303570823611e-02;
C[1719] = 1.361999870892e-04;
C[1720] = -1.226966921073e-03;
C[1721] = 1.259450533071e-01;
C[1722] = -1.232273172140e-03;
C[1723] = 4.620211512994e-03;
C[1724] = 5.257404296429e-01;
C[1725] = -1.162037092478e-04;
C[1726] = 4.043417464330e-04;
C[1727] = 3.864897431274e-02;
C[1728] = -6.947763025316e+00;
C[1729] = -4.136549922551e-02;
C[1730] = 5.828283278664e-03;
C[1731] = 2.111003830469e+00;
C[1732] = 4.368696248373e-02;
C[1733] = 1.861952775480e-04;
C[1734] = 7.909661552330e-01;
C[1735] = 6.096032923693e-02;
C[1736] = -1.521636257667e-03;
C[1737] = 8.266204327778e-02;
C[1738] = -2.703177999377e-02;
C[1739] = 5.014434365962e-03;
C[1740] = 3.746815897582e-01;
C[1741] = 2.504718127467e-02;
C[1742] = -1.786356436300e-03;
C[1743] = 5.108991051040e-02;
C[1744] = 4.291951368250e-03;
C[1745] = -2.131369247315e-04;
C[1746] = 7.912566745228e-02;
C[1747] = 7.189566082282e-03;
C[1748] = -6.357457423318e-04;
C[1749] = 2.458250976518e-03;
C[1750] = 2.648461783504e-04;
C[1751] = -3.098889295572e-05;
C[1752] = -5.754594751355e-02;
C[1753] = -7.110409993757e+00;
C[1754] = 1.052063457217e-01;
C[1755] = 5.526553905500e-02;
C[1756] = 2.133939419501e+00;
C[1757] = 1.235877532352e-02;
C[1758] = 6.080417126489e-02;
C[1759] = 8.290029239592e-01;
C[1760] = 1.698152682241e-02;
C[1761] = -2.784080240818e-02;
C[1762] = 5.614106628822e-02;
C[1763] = 2.085518939337e-02;
C[1764] = 2.562103915471e-02;
C[1765] = 3.852420872603e-01;
C[1766] = -1.696304926835e-03;
C[1767] = 4.239167187169e-03;
C[1768] = 5.372121327277e-02;
C[1769] = 1.053671871235e-03;
C[1770] = 6.998141829863e-03;
C[1771] = 8.394849622910e-02;
C[1772] = 2.765375626427e-03;
C[1773] = 2.540602070257e-04;
C[1774] = 2.627809865353e-03;
C[1775] = 8.149587247595e-05;
C[1776] = 5.007516104452e-03;
C[1777] = 9.620476441966e-02;
C[1778] = -7.036103916482e+00;
C[1779] = 6.932504207748e-03;
C[1780] = 2.506043544330e-02;
C[1781] = 1.993903544010e+00;
C[1782] = -2.367351914709e-03;
C[1783] = 1.594214734161e-02;
C[1784] = 7.675313942711e-01;
C[1785] = 4.973396007252e-03;
C[1786] = 2.040511032827e-02;
C[1787] = 7.757970828149e-02;
C[1788] = -1.350566995369e-03;
C[1789] = -9.426817853885e-04;
C[1790] = 3.492146153304e-01;
C[1791] = -2.918583426319e-04;
C[1792] = 9.491010400285e-04;
C[1793] = 4.955455945780e-02;
C[1794] = -8.756514523229e-04;
C[1795] = 2.416510382725e-03;
C[1796] = 7.949474719112e-02;
C[1797] = -4.136306375828e-05;
C[1798] = 6.606252868143e-05;
C[1799] = 2.472413246725e-03;
C[1800] = 2.105132737544e+00;
C[1801] = 4.984012854129e-02;
C[1802] = 1.662810113724e-04;
C[1803] = -6.187387250027e+00;
C[1804] = -5.226213897957e-02;
C[1805] = 5.635091725969e-03;
C[1806] = 2.149201514662e+00;
C[1807] = -5.947199828794e-02;
C[1808] = 6.496088751710e-03;
C[1809] = 3.743880351120e-01;
C[1810] = 2.534976862903e-02;
C[1811] = -1.688878155188e-03;
C[1812] = 1.225275936916e-01;
C[1813] = -2.485916571830e-02;
C[1814] = 3.250284907001e-03;
C[1815] = 3.888078133975e-01;
C[1816] = -4.802656346080e-03;
C[1817] = -5.152516772115e-04;
C[1818] = 1.016207891265e+00;
C[1819] = -1.032146601806e-02;
C[1820] = -5.541421507151e-03;
C[1821] = 6.016033103047e-02;
C[1822] = -4.807980891942e-04;
C[1823] = -3.942558261212e-04;
C[1824] = 6.506955501392e-02;
C[1825] = 2.109900958456e+00;
C[1826] = 2.397207444322e-02;
C[1827] = -6.362808892695e-02;
C[1828] = -6.273038391350e+00;
C[1829] = -1.929334076735e-02;
C[1830] = -5.959866073926e-02;
C[1831] = 2.125999160616e+00;
C[1832] = 1.008233556262e-02;
C[1833] = 2.611631090118e-02;
C[1834] = 3.840401642080e-01;
C[1835] = -9.970998353930e-04;
C[1836] = -2.542746321567e-02;
C[1837] = 1.041604477841e-01;
C[1838] = -5.839273504315e-03;
C[1839] = -4.763813506960e-03;
C[1840] = 3.914692221378e-01;
C[1841] = 7.682604910281e-03;
C[1842] = -1.018931181775e-02;
C[1843] = 1.032981628419e+00;
C[1844] = 3.865713214964e-02;
C[1845] = -4.733542569152e-04;
C[1846] = 6.132514172441e-02;
C[1847] = 2.430286709247e-03;
C[1848] = -1.410297839940e-03;
C[1849] = 2.169962896979e-02;
C[1850] = 2.012428152238e+00;
C[1851] = 2.859433341127e-03;
C[1852] = -2.312542232062e-02;
C[1853] = -6.193683451813e+00;
C[1854] = 7.304145660770e-03;
C[1855] = 1.217500293988e-02;
C[1856] = 2.195869598610e+00;
C[1857] = -1.866181092174e-03;
C[1858] = -1.229262244522e-03;
C[1859] = 3.501408457418e-01;
C[1860] = 3.111501987759e-03;
C[1861] = -6.030877581979e-03;
C[1862] = 1.286209345156e-01;
C[1863] = -4.397136913764e-04;
C[1864] = 7.831685576165e-03;
C[1865] = 4.036032981821e-01;
C[1866] = -5.337505454767e-03;
C[1867] = 3.864703097736e-02;
C[1868] = 1.095516654057e+00;
C[1869] = -3.871771556125e-04;
C[1870] = 2.422755115895e-03;
C[1871] = 6.492883531326e-02;
C[1872] = 7.599862952674e-01;
C[1873] = 4.089628211383e-02;
C[1874] = 1.357548715572e-03;
C[1875] = 2.182447381828e+00;
C[1876] = -4.068377626396e-02;
C[1877] = -6.777528250901e-06;
C[1878] = -7.129938661425e+00;
C[1879] = -2.412983568506e-03;
C[1880] = 7.297485280393e-03;
C[1881] = 4.954091751212e-02;
C[1882] = 3.243772729616e-03;
C[1883] = -1.056133111179e-04;
C[1884] = 3.904701067558e-01;
C[1885] = -3.818069283195e-03;
C[1886] = -8.052598508275e-04;
C[1887] = 4.336809344176e-02;
C[1888] = 1.067429248276e-03;
C[1889] = 2.638317569035e-03;
C[1890] = 4.114594557888e+00;
C[1891] = -3.347961450353e-03;
C[1892] = -8.309134313055e-03;
C[1893] = 5.143053054661e-01;
C[1894] = -7.244797322707e-04;
C[1895] = -1.704067323345e-03;
C[1896] = 4.277141962769e-02;
C[1897] = 8.617308743637e-01;
C[1898] = 2.955826217824e-02;
C[1899] = -4.267872069874e-02;
C[1900] = 2.091224843392e+00;
C[1901] = -3.380761428402e-03;
C[1902] = -2.101901672615e-03;
C[1903] = -7.155382572048e+00;
C[1904] = -4.130106486233e-02;
C[1905] = 3.382505887789e-03;
C[1906] = 5.535761079299e-02;
C[1907] = 1.629906781860e-03;
C[1908] = -3.962992466619e-03;
C[1909] = 3.897305062767e-01;
C[1910] = 7.053897357751e-03;
C[1911] = 1.082983343070e-03;
C[1912] = 3.638414799585e-02;
C[1913] = -1.315649325441e-02;
C[1914] = -3.320484831389e-03;
C[1915] = 4.116652581861e+00;
C[1916] = 1.720958440099e-02;
C[1917] = -7.225197377756e-04;
C[1918] = 5.158581291864e-01;
C[1919] = 4.571203532164e-03;
C[1920] = 2.180141545283e-03;
C[1921] = 3.170287923716e-02;
C[1922] = 7.677037307981e-01;
C[1923] = -2.053610843652e-03;
C[1924] = -6.577117177159e-03;
C[1925] = 2.200510886590e+00;
C[1926] = 8.095494836375e-03;
C[1927] = -4.140877292001e-02;
C[1928] = -7.214110452562e+00;
C[1929] = -2.804803458400e-05;
C[1930] = 1.789739491973e-03;
C[1931] = 4.956317628415e-02;
C[1932] = -9.427366569796e-04;
C[1933] = 6.849632273292e-03;
C[1934] = 4.038353625811e-01;
C[1935] = 2.678218046835e-03;
C[1936] = -1.316187865730e-02;
C[1937] = 1.894865676788e-02;
C[1938] = -8.327875697456e-03;
C[1939] = 1.761346187073e-02;
C[1940] = 4.171248100926e+00;
C[1941] = -1.705026276912e-03;
C[1942] = 4.593714345730e-03;
C[1943] = 5.253606348712e-01;
C[1944] = 0.000000000000e+00;
C[1945] = 0.000000000000e+00;
C[1946] = 0.000000000000e+00;
C[1947] = 0.000000000000e+00;
C[1948] = 0.000000000000e+00;
C[1949] = 0.000000000000e+00;
C[1950] = 0.000000000000e+00;
C[1951] = 0.000000000000e+00;
C[1952] = 0.000000000000e+00;
C[1953] = 0.000000000000e+00;
C[1954] = 0.000000000000e+00;
C[1955] = 0.000000000000e+00;
C[1956] = 0.000000000000e+00;
C[1957] = 0.000000000000e+00;
C[1958] = 0.000000000000e+00;
C[1959] = 0.000000000000e+00;
C[1960] = 0.000000000000e+00;
C[1961] = 0.000000000000e+00;
C[1962] = 1.000000000000e+00;
C[1963] = 0.000000000000e+00;
C[1964] = 0.000000000000e+00;
C[1965] = 2.000000000000e-01;
C[1966] = -0.000000000000e+00;
C[1967] = -0.000000000000e+00;
C[1968] = 0.000000000000e+00;
C[1969] = 0.000000000000e+00;
C[1970] = 0.000000000000e+00;
C[1971] = 0.000000000000e+00;
C[1972] = 0.000000000000e+00;
C[1973] = 0.000000000000e+00;
C[1974] = 0.000000000000e+00;
C[1975] = 0.000000000000e+00;
C[1976] = 0.000000000000e+00;
C[1977] = 0.000000000000e+00;
C[1978] = 0.000000000000e+00;
C[1979] = 0.000000000000e+00;
C[1980] = 0.000000000000e+00;
C[1981] = 0.000000000000e+00;
C[1982] = 0.000000000000e+00;
C[1983] = 0.000000000000e+00;
C[1984] = 0.000000000000e+00;
C[1985] = 0.000000000000e+00;
C[1986] = 0.000000000000e+00;
C[1987] = 1.000000000000e+00;
C[1988] = 0.000000000000e+00;
C[1989] = -0.000000000000e+00;
C[1990] = 2.000000000000e-01;
C[1991] = -0.000000000000e+00;
C[1992] = 0.000000000000e+00;
C[1993] = 0.000000000000e+00;
C[1994] = 0.000000000000e+00;
C[1995] = 0.000000000000e+00;
C[1996] = 0.000000000000e+00;
C[1997] = 0.000000000000e+00;
C[1998] = 0.000000000000e+00;
C[1999] = 0.000000000000e+00;
C[2000] = 0.000000000000e+00;
C[2001] = 0.000000000000e+00;
C[2002] = 0.000000000000e+00;
C[2003] = 0.000000000000e+00;
C[2004] = 0.000000000000e+00;
C[2005] = 0.000000000000e+00;
C[2006] = 0.000000000000e+00;
C[2007] = 0.000000000000e+00;
C[2008] = 0.000000000000e+00;
C[2009] = 0.000000000000e+00;
C[2010] = 0.000000000000e+00;
C[2011] = 0.000000000000e+00;
C[2012] = 1.000000000000e+00;
C[2013] = -0.000000000000e+00;
C[2014] = -0.000000000000e+00;
C[2015] = 2.000000000000e-01;
C[2016] = 7.618468696842e-02;
C[2017] = -2.213744921265e-02;
C[2018] = 6.595017418714e-03;
C[2019] = 3.712053724310e-01;
C[2020] = 1.943410602660e-02;
C[2021] = 3.979987908148e-03;
C[2022] = 5.076142657330e-02;
C[2023] = 1.679196499583e-03;
C[2024] = 2.613752851991e-04;
C[2025] = 1.308231940046e-01;
C[2026] = -1.404609832004e-03;
C[2027] = 5.675724639245e-04;
C[2028] = 3.005813491910e-02;
C[2029] = 1.239425215030e-03;
C[2030] = 3.001967878580e-04;
C[2031] = 2.447375457379e-03;
C[2032] = 5.857085697290e-05;
C[2033] = 9.447765035722e-06;
C[2034] = 2.432846011648e-03;
C[2035] = 2.889047419559e-06;
C[2036] = -9.224720908575e-06;
C[2037] = -1.947207505711e-05;
C[2038] = -5.437974124317e-06;
C[2039] = -7.491638356921e-07;
C[2040] = -2.183719413550e-02;
C[2041] = 5.349023953149e-02;
C[2042] = 3.859618700283e-03;
C[2043] = 1.917912695179e-02;
C[2044] = 3.938065755966e-01;
C[2045] = 4.001407609126e-03;
C[2046] = 1.585562452956e-03;
C[2047] = 5.625563783534e-02;
C[2048] = 2.829970896829e-03;
C[2049] = -1.384592826861e-03;
C[2050] = 1.285075066926e-01;
C[2051] = -2.400744231806e-04;
C[2052] = 1.220627388492e-03;
C[2053] = 3.251627358845e-02;
C[2054] = 6.759169186975e-04;
C[2055] = 5.395111732480e-05;
C[2056] = 2.756271132062e-03;
C[2057] = 1.659291988537e-04;
C[2058] = -4.081713534807e-06;
C[2059] = 2.856729173033e-03;
C[2060] = 2.752136519608e-04;
C[2061] = -5.427022955550e-06;
C[2062] = -1.209972897660e-05;
C[2063] = 5.277913866026e-06;
C[2064] = 6.744744815256e-03;
C[2065] = 3.966229230308e-03;
C[2066] = 6.069901581027e-02;
C[2067] = 3.962552839266e-03;
C[2068] = 4.063316902404e-03;
C[2069] = 3.560239967476e-01;
C[2070] = 1.846473836356e-04;
C[2071] = 2.673355506500e-03;
C[2072] = 5.220964275235e-02;
C[2073] = 5.775542903606e-04;
C[2074] = -2.329670545122e-04;
C[2075] = 1.293189713256e-01;
C[2076] = 2.958897651512e-04;
C[2077] = 6.749396052652e-04;
C[2078] = 2.944354594293e-02;
C[2079] = 5.530009838941e-06;
C[2080] = 1.567709335083e-04;
C[2081] = 2.565980407809e-03;
C[2082] = -1.540357791485e-05;
C[2083] = 2.549459292799e-04;
C[2084] = 2.736046730872e-03;
C[2085] = -7.328234729001e-07;
C[2086] = 4.882486877346e-06;
C[2087] = -9.760863595299e-06;
C[2088] = 3.718600694910e-01;
C[2089] = 1.955223089656e-02;
C[2090] = 3.978532260947e-03;
C[2091] = 1.227602484506e-01;
C[2092] = -1.851734478669e-02;
C[2093] = -4.977515192314e-04;
C[2094] = 3.902234436141e-01;
C[2095] = -2.850880157291e-03;
C[2096] = -2.008742389900e-03;
C[2097] = 3.010178138976e-02;
C[2098] = 1.245500984810e-03;
C[2099] = 2.969550599299e-04;
C[2100] = 1.330318760390e-01;
C[2101] = -1.129617881430e-03;
C[2102] = -3.809248880200e-05;
C[2103] = 3.185523047980e-02;
C[2104] = -1.741454206286e-04;
C[2105] = -1.908380456068e-04;
C[2106] = 6.071246553789e-02;
C[2107] = -4.017153633201e-04;
C[2108] = -7.116895657496e-04;
C[2109] = 2.650851504504e-03;
C[2110] = -1.320627597259e-05;
C[2111] = -3.307499035325e-05;
C[2112] = 1.912572141809e-02;
C[2113] = 3.940043667039e-01;
C[2114] = 4.822363041875e-03;
C[2115] = -1.815862557216e-02;
C[2116] = 9.516766843658e-02;
C[2117] = -1.251843542746e-02;
C[2118] = -2.752736157880e-03;
C[2119] = 3.895919377572e-01;
C[2120] = 5.209259253951e-03;
C[2121] = 1.218866241130e-03;
C[2122] = 3.252945966226e-02;
C[2123] = 7.255426812299e-04;
C[2124] = -1.105703267128e-03;
C[2125] = 1.301481349177e-01;
C[2126] = -1.441963141874e-03;
C[2127] = -1.695169772711e-04;
C[2128] = 3.198857785969e-02;
C[2129] = 5.695514630276e-04;
C[2130] = -3.980271368847e-04;
C[2131] = 6.155910772115e-02;
C[2132] = 2.133000135705e-03;
C[2133] = -1.337936453117e-05;
C[2134] = 2.701709707652e-03;
C[2135] = 1.042726876208e-04;
C[2136] = 3.688152369715e-03;
C[2137] = 4.399864649555e-03;
C[2138] = 3.559785174366e-01;
C[2139] = -3.051257654795e-04;
C[2140] = -1.232192516283e-02;
C[2141] = 1.206351577351e-01;
C[2142] = -1.928092660737e-03;
C[2143] = 5.392266702784e-03;
C[2144] = 4.010917668931e-01;
C[2145] = 2.807410852958e-04;
C[2146] = 7.024807213927e-04;
C[2147] = 2.944051398887e-02;
C[2148] = -2.525077188520e-05;
C[2149] = -1.428862457565e-03;
C[2150] = 1.323113494145e-01;
C[2151] = -1.871833090588e-04;
C[2152] = 5.797300318884e-04;
C[2153] = 3.286968249420e-02;
C[2154] = -7.074077973151e-04;
C[2155] = 2.139033208640e-03;
C[2156] = 6.456997664101e-02;
C[2157] = -3.318044598469e-05;
C[2158] = 1.039103853076e-04;
C[2159] = 2.829116801534e-03;
C[2160] = 5.104621221213e-02;
C[2161] = 3.106492609469e-03;
C[2162] = 1.821707935420e-04;
C[2163] = 3.900765748802e-01;
C[2164] = -4.299110426914e-03;
C[2165] = -1.960879747815e-03;
C[2166] = 4.151704866824e-02;
C[2167] = 1.794899070484e-03;
C[2168] = 3.665265717000e-03;
C[2169] = 2.466361166635e-03;
C[2170] = 1.553464610923e-04;
C[2171] = 5.364903832697e-06;
C[2172] = 3.184543923087e-02;
C[2173] = -2.726085952067e-04;
C[2174] = -1.893691148640e-04;
C[2175] = 1.279423029384e-01;
C[2176] = 1.841265448565e-04;
C[2177] = 3.382581987968e-04;
C[2178] = 5.150276958518e-01;
C[2179] = -7.834974147848e-04;
C[2180] = -1.829994861404e-03;
C[2181] = 3.776662910104e-02;
C[2182] = -6.741500386795e-05;
C[2183] = -1.544184118330e-04;
C[2184] = 3.171762746769e-03;
C[2185] = 5.553357218257e-02;
C[2186] = 1.914553311133e-03;
C[2187] = -4.373551288285e-03;
C[2188] = 3.904140940436e-01;
C[2189] = 6.220897448601e-03;
C[2190] = 1.800582267826e-03;
C[2191] = 3.482713949625e-02;
C[2192] = -1.271081937384e-02;
C[2193] = 1.580752734520e-04;
C[2194] = 2.708133421877e-03;
C[2195] = 1.061841204838e-04;
C[2196] = -2.756568293615e-04;
C[2197] = 3.204338827878e-02;
C[2198] = 6.349720816095e-04;
C[2199] = 1.845054246793e-04;
C[2200] = 1.273514594637e-01;
C[2201] = -1.133968037836e-03;
C[2202] = -7.804211405602e-04;
C[2203] = 5.165979091025e-01;
C[2204] = 4.524904285970e-03;
C[2205] = -6.729542638889e-05;
C[2206] = 3.791244643141e-02;
C[2207] = 3.932750043767e-04;
C[2208] = 2.529200198644e-04;
C[2209] = 2.029548757175e-03;
C[2210] = 5.072765980943e-02;
C[2211] = -2.049582809232e-03;
C[2212] = 6.084669874893e-03;
C[2213] = 4.025348767193e-01;
C[2214] = 3.673322092376e-03;
C[2215] = -1.271333721381e-02;
C[2216] = 1.900409447338e-02;
C[2217] = 8.884080680074e-06;
C[2218] = 1.125677228768e-04;
C[2219] = 2.467181544948e-03;
C[2220] = -1.935607402289e-04;
C[2221] = 6.279121710904e-04;
C[2222] = 3.296588981594e-02;
C[2223] = 3.387952904885e-04;
C[2224] = -1.134135893834e-03;
C[2225] = 1.259603165196e-01;
C[2226] = -1.828997803171e-03;
C[2227] = 4.538578901536e-03;
C[2228] = 5.251879388660e-01;
C[2229] = -1.544289408150e-04;
C[2230] = 3.939451184306e-04;
C[2231] = 3.862226753962e-02;
C[2232] = -7.026819727087e+00;
C[2233] = -4.093289910253e-02;
C[2234] = 4.919861818576e-02;
C[2235] = 2.090704337817e+00;
C[2236] = 3.142328457451e-02;
C[2237] = 3.037817022775e-02;
C[2238] = 7.865471199048e-01;
C[2239] = 1.405792736316e-02;
C[2240] = 4.336124847199e-03;
C[2241] = 7.182454465165e-02;
C[2242] = -8.583522341884e-03;
C[2243] = 8.086248871007e-03;
C[2244] = 3.695558474921e-01;
C[2245] = 7.346489042804e-03;
C[2246] = 3.951880684718e-03;
C[2247] = 5.071748578088e-02;
C[2248] = 9.269378798138e-04;
C[2249] = 2.138365835503e-04;
C[2250] = 7.890848155797e-02;
C[2251] = 1.305707913140e-03;
C[2252] = -1.369964670117e-04;
C[2253] = 2.445685737274e-03;
C[2254] = 4.278444434644e-05;
C[2255] = -1.182352043521e-05;
C[2256] = -3.492779755946e-02;
C[2257] = -7.073264105417e+00;
C[2258] = 1.299224059039e-02;
C[2259] = 2.579870724325e-02;
C[2260] = 2.149541974508e+00;
C[2261] = 7.416314918840e-03;
C[2262] = 1.299281612585e-02;
C[2263] = 8.706950414017e-01;
C[2264] = 4.525589357722e-02;
C[2265] = -8.283267264731e-03;
C[2266] = 4.510895395927e-02;
C[2267] = -7.759420947860e-03;
C[2268] = 7.077702615856e-03;
C[2269] = 3.998876248419e-01;
C[2270] = 1.014417077289e-02;
C[2271] = 8.666243541009e-04;
C[2272] = 5.667543234842e-02;
C[2273] = 3.238865272175e-03;
C[2274] = 1.163961952325e-03;
C[2275] = 8.962854695782e-02;
C[2276] = 6.917215314051e-03;
C[2277] = 3.912247827251e-05;
C[2278] = 2.840043852364e-03;
C[2279] = 2.531958502880e-04;
C[2280] = 5.219316611660e-02;
C[2281] = 1.512445119089e-02;
C[2282] = -7.152275424663e+00;
C[2283] = 2.878631370721e-02;
C[2284] = 6.765469337151e-03;
C[2285] = 2.014579362795e+00;
C[2286] = 3.533335118461e-03;
C[2287] = 4.319197497700e-02;
C[2288] = 8.141567099738e-01;
C[2289] = 8.235976267549e-03;
C[2290] = -7.652810417835e-03;
C[2291] = 4.886204905660e-02;
C[2292] = 3.879656518132e-03;
C[2293] = 1.013952187187e-02;
C[2294] = 3.671370531938e-01;
C[2295] = 1.606829576760e-04;
C[2296] = 3.088594097010e-03;
C[2297] = 5.289308508778e-02;
C[2298] = -2.801189059086e-04;
C[2299] = 6.453864169767e-03;
C[2300] = 8.532530588965e-02;
C[2301] = -1.654535367629e-05;
C[2302] = 2.358484413392e-04;
C[2303] = 2.691739385568e-03;
C[2304] = 2.103798279016e+00;
C[2305] = 3.326078613872e-02;
C[2306] = 2.910590214082e-02;
C[2307] = -6.213311420644e+00;
C[2308] = -3.376673163712e-02;
C[2309] = 5.944733530235e-04;
C[2310] = 2.157686457776e+00;
C[2311] = -1.710839465466e-02;
C[2312] = -5.337893755286e-03;
C[2313] = 3.702105445520e-01;
C[2314] = 7.450806560630e-03;
C[2315] = 3.895635939812e-03;
C[2316] = 1.236565109532e-01;
C[2317] = -6.206555212881e-03;
C[2318] = -1.602581153563e-04;
C[2319] = 3.906779735889e-01;
C[2320] = -2.280144942589e-03;
C[2321] = -2.177491579032e-03;
C[2322] = 1.026786091084e+00;
C[2323] = -8.000661082225e-03;
C[2324] = -1.175789825653e-02;
C[2325] = 6.073100992638e-02;
C[2326] = -4.544006951205e-04;
C[2327] = -7.276678561405e-04;
C[2328] = 2.525559240438e-02;
C[2329] = 2.153497796653e+00;
C[2330] = 2.194639212657e-02;
C[2331] = -2.659234734655e-02;
C[2332] = -6.289297460631e+00;
C[2333] = -3.935942351976e-02;
C[2334] = -1.597248096285e-02;
C[2335] = 2.088033854492e+00;
C[2336] = -2.568796756130e-02;
C[2337] = 7.038104434297e-03;
C[2338] = 4.000854159491e-01;
C[2339] = 1.089856801134e-02;
C[2340] = -5.847835998352e-03;
C[2341] = 8.850853254192e-02;
C[2342] = -1.887928555647e-02;
C[2343] = -2.217893541409e-03;
C[2344] = 3.896559284981e-01;
C[2345] = 4.761042376162e-03;
C[2346] = -7.905240003963e-03;
C[2347] = 1.036787229254e+00;
C[2348] = 3.243668615569e-02;
C[2349] = -4.538263558430e-04;
C[2350] = 6.160505002432e-02;
C[2351] = 2.154732808284e-03;
C[2352] = 2.454145945909e-02;
C[2353] = 1.538545572742e-02;
C[2354] = 2.013669776577e+00;
C[2355] = 4.446988428061e-03;
C[2356] = -3.542921822704e-02;
C[2357] = -6.237495192484e+00;
C[2358] = -4.517708476102e-03;
C[2359] = -2.318369438054e-02;
C[2360] = 2.150670022907e+00;
C[2361] = 3.660045146285e-03;
C[2362] = 1.054262781332e-02;
C[2363] = 3.670915738829e-01;
C[2364] = 3.236763839551e-05;
C[2365] = -1.868277529183e-02;
C[2366] = 1.097588177231e-01;
C[2367] = -2.127683114538e-03;
C[2368] = 4.932964558220e-03;
C[2369] = 4.001609211267e-01;
C[2370] = -1.163067060725e-02;
C[2371] = 3.267126742074e-02;
C[2372] = 1.087628534168e+00;
C[2373] = -7.239153717975e-04;
C[2374] = 2.156435908674e-03;
C[2375] = 6.460013321100e-02;
C[2376] = 7.922428326814e-01;
C[2377] = 4.341141925611e-02;
C[2378] = 3.483803316588e-03;
C[2379] = 2.154749083097e+00;
C[2380] = -4.689996634353e-02;
C[2381] = -5.173450217674e-03;
C[2382] = -7.143097524507e+00;
C[2383] = 3.058446258814e-03;
C[2384] = 1.232019699926e-02;
C[2385] = 5.100227141971e-02;
C[2386] = 2.387554510614e-03;
C[2387] = 1.582063675824e-04;
C[2388] = 3.905311048550e-01;
C[2389] = -3.764267810443e-03;
C[2390] = -2.160470201617e-03;
C[2391] = 4.112250846526e-02;
C[2392] = 2.093422643014e-03;
C[2393] = 3.923930184011e-03;
C[2394] = 4.120627925041e+00;
C[2395] = -3.054667862585e-03;
C[2396] = -8.984324464902e-03;
C[2397] = 5.150323207844e-01;
C[2398] = -7.832262631931e-04;
C[2399] = -1.867037768788e-03;
C[2400] = 4.390925230687e-02;
C[2401] = 8.562537283463e-01;
C[2402] = 2.801593106966e-02;
C[2403] = -4.756181727453e-02;
C[2404] = 2.104476980219e+00;
C[2405] = -6.611079464195e-03;
C[2406] = 3.172110205665e-03;
C[2407] = -7.167863466738e+00;
C[2408] = -4.083659256142e-02;
C[2409] = 2.419504127000e-03;
C[2410] = 5.595336669565e-02;
C[2411] = 2.329791901643e-03;
C[2412] = -3.802816073583e-03;
C[2413] = 3.904780847844e-01;
C[2414] = 5.761595304037e-03;
C[2415] = 2.099105840357e-03;
C[2416] = 3.434448532890e-02;
C[2417] = -1.269182016696e-02;
C[2418] = -3.010364893921e-03;
C[2419] = 4.122904350260e+00;
C[2420] = 1.713728590602e-02;
C[2421] = -7.807995083592e-04;
C[2422] = 5.165403248044e-01;
C[2423] = 4.485979864285e-03;
C[2424] = 4.167019540505e-03;
C[2425] = 2.924745078414e-02;
C[2426] = 7.845170511153e-01;
C[2427] = -6.154702141936e-03;
C[2428] = -8.179755142453e-03;
C[2429] = 2.179532219430e+00;
C[2430] = 1.248132450676e-02;
C[2431] = -4.088694936078e-02;
C[2432] = -7.216442548991e+00;
C[2433] = 2.053813182156e-04;
C[2434] = 2.438443132521e-03;
C[2435] = 5.141110214486e-02;
C[2436] = -2.218331998365e-03;
C[2437] = 5.636452997105e-03;
C[2438] = 4.016040309528e-01;
C[2439] = 3.931986559386e-03;
C[2440] = -1.269433800693e-02;
C[2441] = 1.910606598561e-02;
C[2442] = -8.979268238215e-03;
C[2443] = 1.736518924702e-02;
C[2444] = 4.169486191792e+00;
C[2445] = -1.867329834058e-03;
C[2446] = 4.500674858960e-03;
C[2447] = 5.252155639260e-01;
C[2448] = 0.000000000000e+00;
C[2449] = 0.000000000000e+00;
C[2450] = 0.000000000000e+00;
C[2451] = 0.000000000000e+00;
C[2452] = 0.000000000000e+00;
C[2453] = 0.000000000000e+00;
C[2454] = 0.000000000000e+00;
C[2455] = 0.000000000000e+00;
C[2456] = 0.000000000000e+00;
C[2457] = 0.000000000000e+00;
C[2458] = 0.000000000000e+00;
C[2459] = 0.000000000000e+00;
C[2460] = 0.000000000000e+00;
C[2461] = 0.000000000000e+00;
C[2462] = 0.000000000000e+00;
C[2463] = 0.000000000000e+00;
C[2464] = 0.000000000000e+00;
C[2465] = 0.000000000000e+00;
C[2466] = 1.000000000000e+00;
C[2467] = 0.000000000000e+00;
C[2468] = 0.000000000000e+00;
C[2469] = 2.000000000000e-01;
C[2470] = -0.000000000000e+00;
C[2471] = -0.000000000000e+00;
C[2472] = 0.000000000000e+00;
C[2473] = 0.000000000000e+00;
C[2474] = 0.000000000000e+00;
C[2475] = 0.000000000000e+00;
C[2476] = 0.000000000000e+00;
C[2477] = 0.000000000000e+00;
C[2478] = 0.000000000000e+00;
C[2479] = 0.000000000000e+00;
C[2480] = 0.000000000000e+00;
C[2481] = 0.000000000000e+00;
C[2482] = 0.000000000000e+00;
C[2483] = 0.000000000000e+00;
C[2484] = 0.000000000000e+00;
C[2485] = 0.000000000000e+00;
C[2486] = 0.000000000000e+00;
C[2487] = 0.000000000000e+00;
C[2488] = 0.000000000000e+00;
C[2489] = 0.000000000000e+00;
C[2490] = 0.000000000000e+00;
C[2491] = 1.000000000000e+00;
C[2492] = 0.000000000000e+00;
C[2493] = -0.000000000000e+00;
C[2494] = 2.000000000000e-01;
C[2495] = -0.000000000000e+00;
C[2496] = 0.000000000000e+00;
C[2497] = 0.000000000000e+00;
C[2498] = 0.000000000000e+00;
C[2499] = 0.000000000000e+00;
C[2500] = 0.000000000000e+00;
C[2501] = 0.000000000000e+00;
C[2502] = 0.000000000000e+00;
C[2503] = 0.000000000000e+00;
C[2504] = 0.000000000000e+00;
C[2505] = 0.000000000000e+00;
C[2506] = 0.000000000000e+00;
C[2507] = 0.000000000000e+00;
C[2508] = 0.000000000000e+00;
C[2509] = 0.000000000000e+00;
C[2510] = 0.000000000000e+00;
C[2511] = 0.000000000000e+00;
C[2512] = 0.000000000000e+00;
C[2513] = 0.000000000000e+00;
C[2514] = 0.000000000000e+00;
C[2515] = 0.000000000000e+00;
C[2516] = 1.000000000000e+00;
C[2517] = -0.000000000000e+00;
C[2518] = -0.000000000000e+00;
C[2519] = 2.000000000000e-01;
C[2520] = 6.675784858528e-02;
C[2521] = -8.610570826372e-04;
C[2522] = 1.136664027319e-02;
C[2523] = 3.722791627001e-01;
C[2524] = -3.971621867062e-04;
C[2525] = 1.719024560383e-03;
C[2526] = 5.103793487463e-02;
C[2527] = -5.074292427221e-04;
C[2528] = -3.094163698257e-04;
C[2529] = 1.304320219118e-01;
C[2530] = -3.936718289971e-05;
C[2531] = 9.961833284550e-04;
C[2532] = 3.015818855555e-02;
C[2533] = -1.307442729330e-04;
C[2534] = 4.149505801556e-05;
C[2535] = 2.460904840400e-03;
C[2536] = -3.160913531349e-05;
C[2537] = -2.255958601849e-05;
C[2538] = 2.447409952130e-03;
C[2539] = -5.393992998072e-05;
C[2540] = -5.412577342309e-05;
C[2541] = -1.893812456255e-05;
C[2542] = -1.419244107007e-06;
C[2543] = -1.632959869793e-06;
C[2544] = -8.115361903848e-04;
C[2545] = 4.385778733802e-02;
C[2546] = -9.821780517203e-03;
C[2547] = -4.575127093343e-04;
C[2548] = 4.002316505994e-01;
C[2549] = 1.146918378206e-02;
C[2550] = -5.384535458460e-04;
C[2551] = 5.687743777589e-02;
C[2552] = 3.476895043422e-03;
C[2553] = -3.606579008288e-05;
C[2554] = 1.281076177792e-01;
C[2555] = -8.058190453153e-04;
C[2556] = -1.344337608823e-04;
C[2557] = 3.284653822367e-02;
C[2558] = 1.106992076720e-03;
C[2559] = -3.297817956944e-05;
C[2560] = 2.773475152370e-03;
C[2561] = 1.856222588655e-04;
C[2562] = -5.628390032759e-05;
C[2563] = 2.844431264087e-03;
C[2564] = 2.686072282237e-04;
C[2565] = -1.392425777364e-06;
C[2566] = -1.449027636872e-05;
C[2567] = 2.357021973530e-06;
C[2568] = 1.138517605248e-02;
C[2569] = -9.840122311150e-03;
C[2570] = 4.400923276708e-02;
C[2571] = 1.686238314314e-03;
C[2572] = 1.156036261462e-02;
C[2573] = 3.732462225994e-01;
C[2574] = -3.173030639662e-04;
C[2575] = 3.342353626117e-03;
C[2576] = 5.347581233138e-02;
C[2577] = 9.974190470739e-04;
C[2578] = -8.070418315784e-04;
C[2579] = 1.284689913092e-01;
C[2580] = 3.930289244029e-05;
C[2581] = 1.113081940677e-03;
C[2582] = 3.040990308829e-02;
C[2583] = -2.261749319829e-05;
C[2584] = 1.777828208617e-04;
C[2585] = 2.598618431523e-03;
C[2586] = -5.309695372452e-05;
C[2587] = 2.497363491483e-04;
C[2588] = 2.704903425686e-03;
C[2589] = -1.531059998623e-06;
C[2590] = 2.080099374989e-06;
C[2591] = -1.670079731887e-05;
C[2592] = 3.719500251344e-01;
C[2593] = -2.027047380675e-04;
C[2594] = 1.589256452500e-03;
C[2595] = 1.208874493530e-01;
C[2596] = 1.541606270010e-03;
C[2597] = 2.054815347882e-03;
C[2598] = 3.906342996408e-01;
C[2599] = -8.129142870377e-04;
C[2600] = -1.414009537095e-03;
C[2601] = 3.013624605117e-02;
C[2602] = -1.174465627979e-04;
C[2603] = 3.283743498601e-05;
C[2604] = 1.329199955805e-01;
C[2605] = 2.277137822725e-04;
C[2606] = 2.052861900486e-04;
C[2607] = 3.187418816431e-02;
C[2608] = -7.962971878907e-05;
C[2609] = -1.326464631188e-04;
C[2610] = 6.064386830121e-02;
C[2611] = -3.043829381463e-04;
C[2612] = -5.829908402043e-04;
C[2613] = 2.646499261490e-03;
C[2614] = -1.424164771592e-05;
C[2615] = -2.670370679670e-05;
C[2616] = -2.279674907905e-04;
C[2617] = 3.999010976630e-01;
C[2618] = 1.103928670536e-02;
C[2619] = 1.591550698439e-03;
C[2620] = 8.763711779867e-02;
C[2621] = -2.002831517309e-02;
C[2622] = -7.768176602373e-04;
C[2623] = 3.902515014677e-01;
C[2624] = 4.721365455638e-03;
C[2625] = -1.194646265387e-04;
C[2626] = 3.282450136124e-02;
C[2627] = 1.078343546726e-03;
C[2628] = 2.310434108345e-04;
C[2629] = 1.297430105476e-01;
C[2630] = -1.864007904608e-03;
C[2631] = -7.782785483069e-05;
C[2632] = 3.206916708334e-02;
C[2633] = 5.725224432701e-04;
C[2634] = -3.057218062644e-04;
C[2635] = 6.164030798135e-02;
C[2636] = 2.196001462806e-03;
C[2637] = -1.444338392115e-05;
C[2638] = 2.706934766331e-03;
C[2639] = 1.103855166571e-04;
C[2640] = 1.571074762627e-03;
C[2641] = 1.102933701719e-02;
C[2642] = 3.731254853086e-01;
C[2643] = 2.080123130801e-03;
C[2644] = -2.005826487006e-02;
C[2645] = 1.040862477952e-01;
C[2646] = -1.403046238053e-03;
C[2647] = 4.881108438768e-03;
C[2648] = 3.991993478528e-01;
C[2649] = 3.163173816514e-05;
C[2650] = 1.077668959062e-03;
C[2651] = 3.040185393557e-02;
C[2652] = 2.069733755766e-04;
C[2653] = -1.866004551072e-03;
C[2654] = 1.314075260029e-01;
C[2655] = -1.322573109913e-04;
C[2656] = 5.818175577762e-04;
C[2657] = 3.278398534710e-02;
C[2658] = -5.850348708574e-04;
C[2659] = 2.193264912366e-03;
C[2660] = 6.452570248235e-02;
C[2661] = -2.685399901847e-05;
C[2662] = 1.095646960735e-04;
C[2663] = 2.832544664862e-03;
C[2664] = 5.122959651010e-02;
C[2665] = -2.344777117393e-04;
C[2666] = -1.885941036595e-04;
C[2667] = 3.905189293750e-01;
C[2668] = -1.053940426148e-03;
C[2669] = -1.536939106494e-03;
C[2670] = 4.174011781600e-02;
C[2671] = 2.016207847779e-03;
C[2672] = 3.642116675891e-03;
C[2673] = 2.473682282765e-03;
C[2674] = -1.271312396233e-05;
C[2675] = -1.403689584450e-05;
C[2676] = 3.186649681326e-02;
C[2677] = -9.630270589137e-05;
C[2678] = -1.411835022207e-04;
C[2679] = 1.280214328582e-01;
C[2680] = 1.654647249949e-04;
C[2681] = 2.963414708773e-04;
C[2682] = 5.140858848913e-01;
C[2683] = -7.116624640124e-04;
C[2684] = -1.825423072082e-03;
C[2685] = 3.765503981864e-02;
C[2686] = -5.524737835401e-05;
C[2687] = -1.410418786447e-04;
C[2688] = -2.056386703062e-04;
C[2689] = 5.688115304799e-02;
C[2690] = 3.288267017643e-03;
C[2691] = -1.077855763469e-03;
C[2692] = 3.903918142660e-01;
C[2693] = 4.886618248788e-03;
C[2694] = 2.011634758308e-03;
C[2695] = 3.464324060861e-02;
C[2696] = -1.287572647871e-02;
C[2697] = -1.148976381910e-05;
C[2698] = 2.773722837176e-03;
C[2699] = 1.741770469635e-04;
C[2700] = -9.729248388450e-05;
C[2701] = 3.207852126989e-02;
C[2702] = 5.821848784441e-04;
C[2703] = 1.651598523635e-04;
C[2704] = 1.273803225286e-01;
C[2705] = -1.156235456427e-03;
C[2706] = -7.070887203176e-04;
C[2707] = 5.153474337025e-01;
C[2708] = 4.485732153358e-03;
C[2709] = -5.500695407872e-05;
C[2710] = 3.778053144629e-02;
C[2711] = 3.954191734429e-04;
C[2712] = -1.751118200947e-04;
C[2713] = 3.418861511135e-03;
C[2714] = 5.372142230896e-02;
C[2715] = -1.540792188751e-03;
C[2716] = 4.737840195934e-03;
C[2717] = 3.988671410507e-01;
C[2718] = 3.633116837206e-03;
C[2719] = -1.287557225329e-02;
C[2720] = 1.911389573089e-02;
C[2721] = -1.360594936976e-05;
C[2722] = 1.817533567130e-04;
C[2723] = 2.614992430029e-03;
C[2724] = -1.410986398959e-04;
C[2725] = 5.736207592899e-04;
C[2726] = 3.276183822696e-02;
C[2727] = 2.957414816317e-04;
C[2728] = -1.156225174732e-03;
C[2729] = 1.259787871485e-01;
C[2730] = -1.822050815818e-03;
C[2731] = 4.505694473288e-03;
C[2732] = 5.255812483311e-01;
C[2733] = -1.408220192055e-04;
C[2734] = 3.963932904454e-04;
C[2735] = 3.865901810637e-02;
C[2736] = -7.051607818105e+00;
C[2737] = -9.061724503063e-03;
C[2738] = 6.557936423928e-02;
C[2739] = 2.096902785093e+00;
C[2740] = -5.474837465433e-03;
C[2741] = 2.789384344281e-02;
C[2742] = 7.900383610530e-01;
C[2743] = -8.917415667554e-03;
C[2744] = -3.756511958784e-03;
C[2745] = 7.177540177660e-02;
C[2746] = -5.138945781436e-04;
C[2747] = 1.293516250495e-02;
C[2748] = 3.705684148952e-01;
C[2749] = -2.526314204345e-03;
C[2750] = 9.211495635294e-04;
C[2751] = 5.091327321540e-02;
C[2752] = -6.186509839126e-04;
C[2753] = -3.474412978861e-04;
C[2754] = 7.909398666546e-02;
C[2755] = -1.224393218739e-03;
C[2756] = -1.145254690091e-03;
C[2757] = 2.451482935404e-03;
C[2758] = -4.371164969554e-05;
C[2759] = -4.883523965466e-05;
C[2760] = -8.071306658014e-03;
C[2761] = -7.084650901126e+00;
C[2762] = -2.451621396103e-03;
C[2763] = -6.588978462461e-03;
C[2764] = 2.157376121607e+00;
C[2765] = 1.961460291473e-02;
C[2766] = -9.306460535952e-03;
C[2767] = 8.744689408392e-01;
C[2768] = 4.912750242542e-02;
C[2769] = -4.643736858912e-04;
C[2770] = 4.598450817141e-02;
C[2771] = -6.464155746825e-03;
C[2772] = -2.585491748721e-03;
C[2773] = 4.004140043357e-01;
C[2774] = 1.183168235837e-02;
C[2775] = -6.402154699040e-04;
C[2776] = 5.689123308482e-02;
C[2777] = 3.491372451488e-03;
C[2778] = -1.289055831033e-03;
C[2779] = 8.980414520475e-02;
C[2780] = 7.336139348167e-03;
C[2781] = -4.550184152883e-05;
C[2782] = 2.845788193186e-03;
C[2783] = 2.661515760214e-04;
C[2784] = 6.595007982495e-02;
C[2785] = -2.818457275030e-03;
C[2786] = -7.165386006905e+00;
C[2787] = 2.729105063283e-02;
C[2788] = 2.116373430080e-02;
C[2789] = 2.042141393782e+00;
C[2790] = -3.812800826055e-03;
C[2791] = 4.732826742220e-02;
C[2792] = 8.211732258488e-01;
C[2793] = 1.295369828424e-02;
C[2794] = -6.482497540772e-03;
C[2795] = 4.767819295622e-02;
C[2796] = 8.868956227419e-04;
C[2797] = 1.193849984921e-02;
C[2798] = 3.710543529883e-01;
C[2799] = -3.463305349675e-04;
C[2800] = 3.350491749773e-03;
C[2801] = 5.332330801486e-02;
C[2802] = -1.134779716687e-03;
C[2803] = 6.848984500232e-03;
C[2804] = 8.601819800075e-02;
C[2805] = -4.785049962762e-05;
C[2806] = 2.472224795350e-04;
C[2807] = 2.709880104873e-03;
C[2808] = 2.090320033778e+00;
C[2809] = -1.492819037126e-03;
C[2810] = 2.535141339655e-02;
C[2811] = -6.218654127344e+00;
C[2812] = 4.932805596747e-03;
C[2813] = 7.341164798305e-03;
C[2814] = 2.157972042798e+00;
C[2815] = 5.260769180274e-03;
C[2816] = 3.398429241610e-03;
C[2817] = 3.702392773294e-01;
C[2818] = -2.330683777454e-03;
C[2819] = 7.899137609277e-04;
C[2820] = 1.229355204348e-01;
C[2821] = 3.358171224700e-03;
C[2822] = 2.548755367283e-03;
C[2823] = 3.906758097433e-01;
C[2824] = -5.172759688693e-04;
C[2825] = -1.054618659725e-03;
C[2826] = 1.024820586870e+00;
C[2827] = -4.423810256968e-03;
C[2828] = -8.808527041942e-03;
C[2829] = 6.056266159560e-02;
C[2830] = -2.894840209147e-04;
C[2831] = -5.502522229228e-04;
C[2832] = -2.090943547119e-03;
C[2833] = 2.150765062879e+00;
C[2834] = 1.074221611554e-02;
C[2835] = 5.931694165321e-03;
C[2836] = -6.297821201220e+00;
C[2837] = -4.593252897376e-02;
C[2838] = 5.710467889904e-03;
C[2839] = 2.090731079171e+00;
C[2840] = -2.809174128070e-02;
C[2841] = -2.357119508429e-03;
C[2842] = 4.000834513993e-01;
C[2843] = 1.141742393995e-02;
C[2844] = 3.408115653129e-03;
C[2845] = 8.736834640619e-02;
C[2846] = -2.032399474963e-02;
C[2847] = -4.914158323529e-04;
C[2848] = 3.905606830766e-01;
C[2849] = 4.861792237817e-03;
C[2850] = -4.427474717312e-03;
C[2851] = 1.037112807902e+00;
C[2852] = 3.305488031709e-02;
C[2853] = -2.915195596123e-04;
C[2854] = 6.158243292981e-02;
C[2855] = 2.210618714992e-03;
C[2856] = 2.493484748768e-02;
C[2857] = 1.081766761738e-02;
C[2858] = 2.039726647964e+00;
C[2859] = 7.847320456682e-03;
C[2860] = -4.653152291304e-02;
C[2861] = -6.262218048779e+00;
C[2862] = 3.497935438492e-03;
C[2863] = -2.588900776229e-02;
C[2864] = 2.141323232495e+00;
C[2865] = 7.731997657731e-04;
C[2866] = 1.139183559350e-02;
C[2867] = 3.709336156974e-01;
C[2868] = 2.574063150202e-03;
C[2869] = -2.035394444660e-02;
C[2870] = 1.065830690115e-01;
C[2871] = -1.053535242020e-03;
C[2872] = 5.031145469275e-03;
C[2873] = 3.991245056548e-01;
C[2874] = -8.845880590650e-03;
C[2875] = 3.318162133316e-02;
C[2876] = 1.085839914710e+00;
C[2877] = -5.518538870999e-04;
C[2878] = 2.204376717457e-03;
C[2879] = 6.452244970897e-02;
C[2880] = 7.938715937623e-01;
C[2881] = -3.226943853817e-03;
C[2882] = -1.238621619920e-03;
C[2883] = 2.155664637483e+00;
C[2884] = 1.680125716993e-04;
C[2885] = 8.200780696585e-04;
C[2886] = -7.137206635368e+00;
C[2887] = 5.988148052201e-03;
C[2888] = 1.191856294555e-02;
C[2889] = 5.110493485086e-02;
C[2890] = -3.362396357973e-04;
C[2891] = -2.176215746608e-04;
C[2892] = 3.905604394775e-01;
C[2893] = -7.685385982631e-04;
C[2894] = -1.187428110462e-03;
C[2895] = 4.287476365766e-02;
C[2896] = 1.690772159110e-03;
C[2897] = 3.076283658464e-03;
C[2898] = 4.109929731878e+00;
C[2899] = -2.341375933626e-03;
C[2900] = -8.745670273368e-03;
C[2901] = 5.130490834370e-01;
C[2902] = -5.481455655808e-04;
C[2903] = -1.572622610733e-03;
C[2904] = -2.881604219235e-03;
C[2905] = 8.745432462811e-01;
C[2906] = 4.624653525273e-02;
C[2907] = -3.806034835437e-05;
C[2908] = 2.093537335138e+00;
C[2909] = -2.577881156191e-02;
C[2910] = 5.896686262781e-03;
C[2911] = -7.163024202076e+00;
C[2912] = -4.276595897010e-02;
C[2913] = -3.168604114967e-04;
C[2914] = 5.689494835692e-02;
C[2915] = 3.296405141299e-03;
C[2916] = -7.822174453007e-04;
C[2917] = 3.907009958749e-01;
C[2918] = 5.036655279294e-03;
C[2919] = 1.686199069639e-03;
C[2920] = 3.542826315801e-02;
C[2921] = -1.309999517269e-02;
C[2922] = -2.263483614465e-03;
C[2923] = 4.108677906645e+00;
C[2924] = 1.678282647864e-02;
C[2925] = -5.436276247370e-04;
C[2926] = 5.142383082545e-01;
C[2927] = 4.546295204960e-03;
C[2928] = -1.070420964163e-03;
C[2929] = 4.796683177969e-02;
C[2930] = 8.260854254005e-01;
C[2931] = 8.627762084746e-04;
C[2932] = -2.776224647477e-02;
C[2933] = 2.134679096453e+00;
C[2934] = 1.173856617185e-02;
C[2935] = -4.276287446168e-02;
C[2936] = -7.220400091890e+00;
C[2937] = -2.131367481551e-04;
C[2938] = 3.433338919201e-03;
C[2939] = 5.356891799245e-02;
C[2940] = -1.181401311382e-03;
C[2941] = 4.878266978113e-03;
C[2942] = 3.987922988527e-01;
C[2943] = 3.067283819779e-03;
C[2944] = -1.309984094727e-02;
C[2945] = 1.910184111369e-02;
C[2946] = -8.681050062371e-03;
C[2947] = 1.711402929633e-02;
C[2948] = 4.173996090980e+00;
C[2949] = -1.569751600906e-03;
C[2950] = 4.569806298421e-03;
C[2951] = 5.258418086797e-01;
C[2952] = 0.000000000000e+00;
C[2953] = 0.000000000000e+00;
C[2954] = 0.000000000000e+00;
C[2955] = 0.000000000000e+00;
C[2956] = 0.000000000000e+00;
C[2957] = 0.000000000000e+00;
C[2958] = 0.000000000000e+00;
C[2959] = 0.000000000000e+00;
C[2960] = 0.000000000000e+00;
C[2961] = 0.000000000000e+00;
C[2962] = 0.000000000000e+00;
C[2963] = 0.000000000000e+00;
C[2964] = 0.000000000000e+00;
C[2965] = 0.000000000000e+00;
C[2966] = 0.000000000000e+00;
C[2967] = 0.000000000000e+00;
C[2968] = 0.000000000000e+00;
C[2969] = 0.000000000000e+00;
C[2970] = 1.000000000000e+00;
C[2971] = 0.000000000000e+00;
C[2972] = 0.000000000000e+00;
C[2973] = 2.000000000000e-01;
C[2974] = -0.000000000000e+00;
C[2975] = -0.000000000000e+00;
C[2976] = 0.000000000000e+00;
C[2977] = 0.000000000000e+00;
C[2978] = 0.000000000000e+00;
C[2979] = 0.000000000000e+00;
C[2980] = 0.000000000000e+00;
C[2981] = 0.000000000000e+00;
C[2982] = 0.000000000000e+00;
C[2983] = 0.000000000000e+00;
C[2984] = 0.000000000000e+00;
C[2985] = 0.000000000000e+00;
C[2986] = 0.000000000000e+00;
C[2987] = 0.000000000000e+00;
C[2988] = 0.000000000000e+00;
C[2989] = 0.000000000000e+00;
C[2990] = 0.000000000000e+00;
C[2991] = 0.000000000000e+00;
C[2992] = 0.000000000000e+00;
C[2993] = 0.000000000000e+00;
C[2994] = 0.000000000000e+00;
C[2995] = 1.000000000000e+00;
C[2996] = 0.000000000000e+00;
C[2997] = -0.000000000000e+00;
C[2998] = 2.000000000000e-01;
C[2999] = -0.000000000000e+00;
C[3000] = 0.000000000000e+00;
C[3001] = 0.000000000000e+00;
C[3002] = 0.000000000000e+00;
C[3003] = 0.000000000000e+00;
C[3004] = 0.000000000000e+00;
C[3005] = 0.000000000000e+00;
C[3006] = 0.000000000000e+00;
C[3007] = 0.000000000000e+00;
C[3008] = 0.000000000000e+00;
C[3009] = 0.000000000000e+00;
C[3010] = 0.000000000000e+00;
C[3011] = 0.000000000000e+00;
C[3012] = 0.000000000000e+00;
C[3013] = 0.000000000000e+00;
C[3014] = 0.000000000000e+00;
C[3015] = 0.000000000000e+00;
C[3016] = 0.000000000000e+00;
C[3017] = 0.000000000000e+00;
C[3018] = 0.000000000000e+00;
C[3019] = 0.000000000000e+00;
C[3020] = 1.000000000000e+00;
C[3021] = -0.000000000000e+00;
C[3022] = -0.000000000000e+00;
C[3023] = 2.000000000000e-01;
C[3024] = 9.024027050868e-02;
C[3025] = -5.797852968967e-03;
C[3026] = 7.207973785482e-03;
C[3027] = 3.645613400035e-01;
C[3028] = 1.265533858576e-03;
C[3029] = 6.577489063965e-04;
C[3030] = 4.956985663771e-02;
C[3031] = 8.414570652570e-04;
C[3032] = 2.150115491476e-05;
C[3033] = 1.323334537114e-01;
C[3034] = -4.939390305262e-04;
C[3035] = 1.968408314463e-04;
C[3036] = 2.942503818374e-02;
C[3037] = 4.167700083555e-04;
C[3038] = 9.216454461592e-05;
C[3039] = 2.380525077397e-03;
C[3040] = 6.084995954783e-05;
C[3041] = 4.300634619998e-06;
C[3042] = 2.331112181798e-03;
C[3043] = 1.021840295552e-04;
C[3044] = 4.212445821878e-06;
C[3045] = -2.021590997219e-05;
C[3046] = 3.879647154118e-06;
C[3047] = 3.472103650716e-07;
C[3048] = -5.849342528807e-03;
C[3049] = 3.936247255664e-02;
C[3050] = 5.107493511253e-03;
C[3051] = 1.503664132429e-03;
C[3052] = 3.998559089210e-01;
C[3053] = 1.219063909833e-02;
C[3054] = 8.132017611876e-04;
C[3055] = 5.686788899385e-02;
C[3056] = 3.251658114760e-03;
C[3057] = -4.973716678489e-04;
C[3058] = 1.270334220855e-01;
C[3059] = 3.289159693565e-04;
C[3060] = 4.292319544970e-04;
C[3061] = 3.285405862900e-02;
C[3062] = 1.006752860819e-03;
C[3063] = 5.854379647665e-05;
C[3064] = 2.771744437263e-03;
C[3065] = 1.651137644616e-04;
C[3066] = 9.590909220153e-05;
C[3067] = 2.820951239177e-03;
C[3068] = 2.265267998138e-04;
C[3069] = 3.603983952365e-06;
C[3070] = -1.455112788500e-05;
C[3071] = 3.006366860341e-07;
C[3072] = 7.132599732164e-03;
C[3073] = 5.117755414155e-03;
C[3074] = 6.618923574323e-02;
C[3075] = 7.553830336537e-04;
C[3076] = 1.279096449063e-02;
C[3077] = 3.646027999449e-01;
C[3078] = -4.092230960332e-06;
C[3079] = 3.110688139366e-03;
C[3080] = 5.161759640377e-02;
C[3081] = 1.918158945585e-04;
C[3082] = 3.296000962167e-04;
C[3083] = 1.309153065865e-01;
C[3084] = 9.850605881110e-05;
C[3085] = 1.040620541948e-03;
C[3086] = 2.948369535929e-02;
C[3087] = 2.162428903827e-06;
C[3088] = 1.553567387103e-04;
C[3089] = 2.493783853581e-03;
C[3090] = -2.486546917455e-06;
C[3091] = 2.003272816655e-04;
C[3092] = 2.565952520565e-03;
C[3093] = 5.036659023541e-08;
C[3094] = -2.815882742842e-07;
C[3095] = -1.936221205017e-05;
C[3096] = 3.646846495270e-01;
C[3097] = 3.415525882192e-04;
C[3098] = 1.519819935390e-03;
C[3099] = 1.317232563230e-01;
C[3100] = -2.719158040537e-04;
C[3101] = 1.035460635238e-03;
C[3102] = 3.913701089170e-01;
C[3103] = -1.698746523876e-03;
C[3104] = -1.317912508072e-03;
C[3105] = 2.943325881864e-02;
C[3106] = 3.517578515496e-04;
C[3107] = 1.494685189268e-04;
C[3108] = 1.339686261504e-01;
C[3109] = -3.343615557297e-04;
C[3110] = -1.074321141332e-05;
C[3111] = 3.173810508658e-02;
C[3112] = -1.132575808981e-04;
C[3113] = -8.912706080635e-05;
C[3114] = 5.973245303764e-02;
C[3115] = -2.406218227070e-04;
C[3116] = -3.075200891973e-04;
C[3117] = 2.592290855132e-03;
C[3118] = -8.530369084307e-06;
C[3119] = -1.175471998988e-05;
C[3120] = 4.746703047846e-04;
C[3121] = 4.021241810082e-01;
C[3122] = 1.274851087800e-02;
C[3123] = -5.009967908023e-04;
C[3124] = 8.642998540061e-02;
C[3125] = -2.001510431995e-02;
C[3126] = -1.666200742903e-03;
C[3127] = 3.898896215809e-01;
C[3128] = 5.538155650793e-03;
C[3129] = 3.640457714361e-04;
C[3130] = 3.300527676814e-02;
C[3131] = 1.037790301107e-03;
C[3132] = -3.496336215129e-04;
C[3133] = 1.297782409910e-01;
C[3134] = -1.805293561711e-03;
C[3135] = -1.107629341720e-04;
C[3136] = 3.188724825763e-02;
C[3137] = 6.752834332105e-04;
C[3138] = -2.372867311645e-04;
C[3139] = 6.068275035494e-02;
C[3140] = 2.466675575240e-03;
C[3141] = -8.411122380722e-06;
C[3142] = 2.653778963110e-03;
C[3143] = 1.248141109014e-04;
C[3144] = 1.516608667670e-03;
C[3145] = 1.234849336843e-02;
C[3146] = 3.631132110470e-01;
C[3147] = 9.325803961151e-04;
C[3148] = -2.006391712959e-02;
C[3149] = 1.157742060449e-01;
C[3150] = -1.276875337057e-03;
C[3151] = 5.741965816446e-03;
C[3152] = 4.008969370253e-01;
C[3153] = 1.494218620342e-04;
C[3154] = 1.017276478826e-03;
C[3155] = 2.938438943276e-02;
C[3156] = -1.760189402151e-05;
C[3157] = -1.808547749021e-03;
C[3158] = 1.324926403886e-01;
C[3159] = -8.586968387519e-05;
C[3160] = 6.870367926847e-04;
C[3161] = 3.288338447950e-02;
C[3162] = -3.010722215335e-04;
C[3163] = 2.457764143134e-03;
C[3164] = 6.476213619375e-02;
C[3165] = -1.154135998713e-05;
C[3166] = 1.235362185717e-04;
C[3167] = 2.839914042460e-03;
C[3168] = 4.933224715200e-02;
C[3169] = -6.978688248457e-06;
C[3170] = -5.714610378536e-06;
C[3171] = 3.913328328347e-01;
C[3172] = -9.270922031887e-04;
C[3173] = -1.197346253154e-03;
C[3174] = 4.547845836740e-02;
C[3175] = 1.119456757065e-03;
C[3176] = 2.063978755270e-03;
C[3177] = 2.364684445017e-03;
C[3178] = 3.865099847582e-06;
C[3179] = 2.054270275946e-06;
C[3180] = 3.173562001442e-02;
C[3181] = -6.148903152445e-05;
C[3182] = -8.056774494836e-05;
C[3183] = 1.284701557168e-01;
C[3184] = 6.442780473060e-05;
C[3185] = 1.144137547229e-04;
C[3186] = 5.116619738952e-01;
C[3187] = -2.566755934337e-04;
C[3188] = -7.899371352897e-04;
C[3189] = 3.744919372361e-02;
C[3190] = -1.118922836133e-05;
C[3191] = -3.256892013936e-05;
C[3192] = 1.103084295081e-05;
C[3193] = 5.677829704679e-02;
C[3194] = 3.456774255736e-03;
C[3195] = -9.402969949514e-04;
C[3196] = 3.900055224636e-01;
C[3197] = 5.546347410585e-03;
C[3198] = 1.104048820278e-03;
C[3199] = 3.780319784998e-02;
C[3200] = -1.355146487531e-02;
C[3201] = 5.488211394080e-06;
C[3202] = 2.765771640791e-03;
C[3203] = 1.784291464683e-04;
C[3204] = -6.269427896983e-05;
C[3205] = 3.189497498315e-02;
C[3206] = 6.739955656272e-04;
C[3207] = 6.340060894482e-05;
C[3208] = 1.277993458674e-01;
C[3209] = -1.251209492604e-03;
C[3210] = -2.531918208823e-04;
C[3211] = 5.127673320851e-01;
C[3212] = 4.467144257979e-03;
C[3213] = -1.097051882106e-05;
C[3214] = 3.756438130617e-02;
C[3215] = 3.972112069611e-04;
C[3216] = 7.304462520363e-06;
C[3217] = 3.648428792421e-03;
C[3218] = 5.188614517433e-02;
C[3219] = -1.187554415925e-03;
C[3220] = 5.282778813388e-03;
C[3221] = 4.004421376516e-01;
C[3222] = 2.033338002343e-03;
C[3223] = -1.353313054813e-02;
C[3224] = 1.892203757132e-02;
C[3225] = 3.354188460372e-06;
C[3226] = 1.915651429724e-04;
C[3227] = 2.511687104952e-03;
C[3228] = -8.043652132985e-05;
C[3229] = 6.582583107168e-04;
C[3230] = 3.285306452125e-02;
C[3231] = 1.123710378611e-04;
C[3232] = -1.249987204126e-03;
C[3233] = 1.259422809440e-01;
C[3234] = -7.845194166558e-04;
C[3235] = 4.499420405854e-03;
C[3236] = 5.261984920778e-01;
C[3237] = -3.213601146165e-05;
C[3238] = 3.986934543029e-04;
C[3239] = 3.870351080097e-02;
C[3240] = -6.952857864848e+00;
C[3241] = -1.509101094919e-02;
C[3242] = 1.927851577223e-02;
C[3243] = 2.099900613297e+00;
C[3244] = 1.634754241154e-02;
C[3245] = 2.124102971581e-03;
C[3246] = 7.710370373137e-01;
C[3247] = 1.712104642223e-02;
C[3248] = 1.755751614719e-03;
C[3249] = 9.290494606200e-02;
C[3250] = -5.104158224783e-03;
C[3251] = -2.011047456180e-03;
C[3252] = 3.639739010655e-01;
C[3253] = 7.896081284706e-03;
C[3254] = 1.613435437785e-03;
C[3255] = 4.936391090257e-02;
C[3256] = 1.240136479746e-03;
C[3257] = 1.768294143350e-04;
C[3258] = 7.564565027627e-02;
C[3259] = 2.074580748249e-03;
C[3260] = 2.009266615690e-04;
C[3261] = 2.308246435163e-03;
C[3262] = 7.904068322996e-05;
C[3263] = 1.121010380808e-05;
C[3264] = -1.612080214601e-02;
C[3265] = -7.148067381070e+00;
C[3266] = 6.556552667195e-02;
C[3267] = 2.036376843061e-02;
C[3268] = 2.115226475774e+00;
C[3269] = 3.240008860165e-02;
C[3270] = 1.657098021532e-02;
C[3271] = 8.725865407064e-01;
C[3272] = 4.520324694647e-02;
C[3273] = -5.155647784624e-03;
C[3274] = 2.476814374806e-02;
C[3275] = 5.845281621642e-03;
C[3276] = 8.051189004814e-03;
C[3277] = 3.989964527397e-01;
C[3278] = 1.015704211486e-02;
C[3279] = 1.189012699555e-03;
C[3280] = 5.670957181613e-02;
C[3281] = 3.196423863458e-03;
C[3282] = 1.926551811471e-03;
C[3283] = 8.881450928916e-02;
C[3284] = 6.822985432580e-03;
C[3285] = 7.216714717295e-05;
C[3286] = 2.808525077252e-03;
C[3287] = 2.452508387141e-04;
C[3288] = 1.777103470588e-02;
C[3289] = 6.577076472999e-02;
C[3290] = -7.047215638700e+00;
C[3291] = 3.906200757930e-03;
C[3292] = 4.225626596983e-02;
C[3293] = 2.061710835980e+00;
C[3294] = 1.313256486205e-03;
C[3295] = 4.379473876095e-02;
C[3296] = 7.958320425922e-01;
C[3297] = -2.086421509497e-03;
C[3298] = 5.855543524544e-03;
C[3299] = 8.230252007317e-02;
C[3300] = 1.729963827009e-03;
C[3301] = 1.082437954212e-02;
C[3302] = 3.620219666736e-01;
C[3303] = 1.343504483111e-04;
C[3304] = 3.031457924361e-03;
C[3305] = 5.140306643085e-02;
C[3306] = 5.307578264181e-05;
C[3307] = 6.150959537439e-03;
C[3308] = 8.256415108948e-02;
C[3309] = 4.364746150347e-06;
C[3310] = 2.145532985665e-04;
C[3311] = 2.569035333873e-03;
C[3312] = 2.102366803766e+00;
C[3313] = -2.878462453585e-03;
C[3314] = 1.919493879265e-02;
C[3315] = -6.182811537018e+00;
C[3316] = -7.250548642700e-03;
C[3317] = 1.745373575175e-03;
C[3318] = 2.178687759970e+00;
C[3319] = -1.861241944779e-02;
C[3320] = -4.633712374051e-03;
C[3321] = 3.640972105890e-01;
C[3322] = 6.889077460605e-03;
C[3323] = 2.494400728745e-03;
C[3324] = 1.354618308837e-01;
C[3325] = -6.975045548969e-03;
C[3326] = -1.168260038200e-03;
C[3327] = 3.886254181184e-01;
C[3328] = -1.554416547631e-03;
C[3329] = -7.300901689423e-04;
C[3330] = 1.009242247402e+00;
C[3331] = -3.838656778851e-03;
C[3332] = -4.173891815007e-03;
C[3333] = 5.949265332492e-02;
C[3334] = -1.677148217361e-04;
C[3335] = -2.009662313606e-04;
C[3336] = 5.302713357008e-04;
C[3337] = 2.160591917517e+00;
C[3338] = 4.140719371732e-02;
C[3339] = -1.183216837767e-02;
C[3340] = -6.298869774547e+00;
C[3341] = -4.419400914758e-02;
C[3342] = -1.798220141676e-02;
C[3343] = 2.089087597864e+00;
C[3344] = -2.026954964078e-02;
C[3345] = 7.105217730914e-03;
C[3346] = 4.012647248268e-01;
C[3347] = 1.078192592949e-02;
C[3348] = -7.204126535718e-03;
C[3349] = 8.921018278206e-02;
C[3350] = -1.905330512733e-02;
C[3351] = -1.501501855084e-03;
C[3352] = 3.871664837236e-01;
C[3353] = 6.531084091562e-03;
C[3354] = -3.741596606534e-03;
C[3355] = 1.019354946271e+00;
C[3356] = 3.803186972853e-02;
C[3357] = -1.626990786919e-04;
C[3358] = 6.044501606615e-02;
C[3359] = 2.509421558924e-03;
C[3360] = 1.930129819705e-02;
C[3361] = 3.555717400360e-02;
C[3362] = 2.031919058022e+00;
C[3363] = -3.122312072805e-04;
C[3364] = -4.517026534052e-02;
C[3365] = -6.227214885080e+00;
C[3366] = -3.802675571927e-03;
C[3367] = -1.759213805658e-02;
C[3368] = 2.169648752510e+00;
C[3369] = 2.472295199059e-03;
C[3370] = 1.031489638496e-02;
C[3371] = 3.605323777757e-01;
C[3372] = -1.271140277323e-03;
C[3373] = -1.910211793698e-02;
C[3374] = 1.189309427222e-01;
C[3375] = -6.675203506083e-04;
C[3376] = 6.760754698188e-03;
C[3377] = 4.012624029407e-01;
C[3378] = -4.049136074467e-03;
C[3379] = 3.819859637654e-02;
C[3380] = 1.091367979796e+00;
C[3381] = -1.942461943288e-04;
C[3382] = 2.502396869304e-03;
C[3383] = 6.477211704670e-02;
C[3384] = 7.662848475995e-01;
C[3385] = 1.673712265988e-04;
C[3386] = 1.280808897841e-03;
C[3387] = 2.177942238323e+00;
C[3388] = -3.200030622477e-03;
C[3389] = -2.212093893878e-03;
C[3390] = -7.115935262207e+00;
C[3391] = 8.106692720548e-04;
C[3392] = 3.643166746251e-03;
C[3393] = 4.912630141687e-02;
C[3394] = 3.688322501187e-04;
C[3395] = 1.327280688929e-04;
C[3396] = 3.885881420360e-01;
C[3397] = -7.623933153701e-04;
C[3398] = -5.879912667058e-04;
C[3399] = 4.952897182191e-02;
C[3400] = 2.180079767552e-04;
C[3401] = 4.569605379616e-04;
C[3402] = 4.100009373364e+00;
C[3403] = -2.131253054766e-04;
C[3404] = -1.095949862981e-03;
C[3405] = 5.105594227083e-01;
C[3406] = 3.674605039544e-05;
C[3407] = 6.905125151926e-05;
C[3408] = 5.125219761047e-04;
C[3409] = 8.707947017650e-01;
C[3410] = 5.071646108834e-02;
C[3411] = -3.443428869299e-03;
C[3412] = 2.091405615518e+00;
C[3413] = -2.150450617380e-02;
C[3414] = 5.025105363183e-04;
C[3415] = -7.144635378963e+00;
C[3416] = -4.508338516549e-02;
C[3417] = 4.097102574398e-04;
C[3418] = 5.661997986907e-02;
C[3419] = 3.377544040730e-03;
C[3420] = -7.959670187066e-04;
C[3421] = 3.872823846063e-01;
C[3422] = 6.565136292327e-03;
C[3423] = 2.026000399683e-04;
C[3424] = 4.187469423380e-02;
C[3425] = -1.447217254541e-02;
C[3426] = -1.434981833407e-04;
C[3427] = 4.097333215348e+00;
C[3428] = 1.427244875885e-02;
C[3429] = 3.984134930663e-05;
C[3430] = 5.115862493979e-01;
C[3431] = 4.471806446591e-03;
C[3432] = 1.471817766831e-03;
C[3433] = 5.313866049970e-02;
C[3434] = 8.012030180034e-01;
C[3435] = -2.026550531103e-03;
C[3436] = -2.537708638889e-02;
C[3437] = 2.160552765037e+00;
C[3438] = 3.030351687712e-03;
C[3439] = -4.471669862185e-02;
C[3440] = -7.219705266533e+00;
C[3441] = 1.626327219406e-04;
C[3442] = 3.593194541120e-03;
C[3443] = 5.167161520141e-02;
C[3444] = -5.997320767949e-04;
C[3445] = 6.275707254157e-03;
C[3446] = 4.008076035671e-01;
C[3447] = 4.263197850346e-04;
C[3448] = -1.445383821823e-02;
C[3449] = 1.860989687308e-02;
C[3450] = -9.607981376248e-04;
C[3451] = 1.477022368422e-02;
C[3452] = 4.177598262881e+00;
C[3453] = 7.479074257633e-05;
C[3454] = 4.505236852260e-03;
C[3455] = 5.263777475018e-01;
C[3456] = 0.000000000000e+00;
C[3457] = 0.000000000000e+00;
C[3458] = 0.000000000000e+00;
C[3459] = 0.000000000000e+00;
C[3460] = 0.000000000000e+00;
C[3461] = 0.000000000000e+00;
C[3462] = 0.000000000000e+00;
C[3463] = 0.000000000000e+00;
C[3464] = 0.000000000000e+00;
C[3465] = 0.000000000000e+00;
C[3466] = 0.000000000000e+00;
C[3467] = 0.000000000000e+00;
C[3468] = 0.000000000000e+00;
C[3469] = 0.000000000000e+00;
C[3470] = 0.000000000000e+00;
C[3471] = 0.000000000000e+00;
C[3472] = 0.000000000000e+00;
C[3473] = 0.000000000000e+00;
C[3474] = 1.000000000000e+00;
C[3475] = 0.000000000000e+00;
C[3476] = 0.000000000000e+00;
C[3477] = 2.000000000000e-01;
C[3478] = -0.000000000000e+00;
C[3479] = -0.000000000000e+00;
C[3480] = 0.000000000000e+00;
C[3481] = 0.000000000000e+00;
C[3482] = 0.000000000000e+00;
C[3483] = 0.000000000000e+00;
C[3484] = 0.000000000000e+00;
C[3485] = 0.000000000000e+00;
C[3486] = 0.000000000000e+00;
C[3487] = 0.000000000000e+00;
C[3488] = 0.000000000000e+00;
C[3489] = 0.000000000000e+00;
C[3490] = 0.000000000000e+00;
C[3491] = 0.000000000000e+00;
C[3492] = 0.000000000000e+00;
C[3493] = 0.000000000000e+00;
C[3494] = 0.000000000000e+00;
C[3495] = 0.000000000000e+00;
C[3496] = 0.000000000000e+00;
C[3497] = 0.000000000000e+00;
C[3498] = 0.000000000000e+00;
C[3499] = 1.000000000000e+00;
C[3500] = 0.000000000000e+00;
C[3501] = -0.000000000000e+00;
C[3502] = 2.000000000000e-01;
C[3503] = -0.000000000000e+00;
C[3504] = 0.000000000000e+00;
C[3505] = 0.000000000000e+00;
C[3506] = 0.000000000000e+00;
C[3507] = 0.000000000000e+00;
C[3508] = 0.000000000000e+00;
C[3509] = 0.000000000000e+00;
C[3510] = 0.000000000000e+00;
C[3511] = 0.000000000000e+00;
C[3512] = 0.000000000000e+00;
C[3513] = 0.000000000000e+00;
C[3514] = 0.000000000000e+00;
C[3515] = 0.000000000000e+00;
C[3516] = 0.000000000000e+00;
C[3517] = 0.000000000000e+00;
C[3518] = 0.000000000000e+00;
C[3519] = 0.000000000000e+00;
C[3520] = 0.000000000000e+00;
C[3521] = 0.000000000000e+00;
C[3522] = 0.000000000000e+00;
C[3523] = 0.000000000000e+00;
C[3524] = 1.000000000000e+00;
C[3525] = -0.000000000000e+00;
C[3526] = -0.000000000000e+00;
C[3527] = 2.000000000000e-01;
C[3528] = 7.746024490278e-02;
C[3529] = 3.274978188829e-03;
C[3530] = -3.928759608126e-03;
C[3531] = 3.699353588446e-01;
C[3532] = 7.198924975493e-03;
C[3533] = 1.385803075541e-03;
C[3534] = 5.011382514047e-02;
C[3535] = 5.252473555505e-04;
C[3536] = 3.908648792373e-04;
C[3537] = 1.306342343487e-01;
C[3538] = 5.123566834407e-04;
C[3539] = -2.265674743869e-04;
C[3540] = 3.027790930039e-02;
C[3541] = 3.367940927457e-04;
C[3542] = 5.937963539742e-05;
C[3543] = 2.429566685596e-03;
C[3544] = 1.163159665002e-05;
C[3545] = 1.972473540944e-05;
C[3546] = 2.413318592034e-03;
C[3547] = -1.707681312112e-05;
C[3548] = 2.459202817592e-05;
C[3549] = -1.554750591552e-05;
C[3550] = -3.262219413160e-06;
C[3551] = -7.880781939044e-07;
C[3552] = 3.499012682360e-03;
C[3553] = 1.977877404754e-02;
C[3554] = -4.751853686457e-03;
C[3555] = 6.859731404867e-03;
C[3556] = 3.988231860869e-01;
C[3557] = 9.575391893876e-03;
C[3558] = 5.145335495232e-04;
C[3559] = 5.526615367831e-02;
C[3560] = 3.628933929862e-03;
C[3561] = 5.272923163428e-04;
C[3562] = 1.260752399139e-01;
C[3563] = -6.421210055788e-04;
C[3564] = 3.168233699606e-04;
C[3565] = 3.277497978560e-02;
C[3566] = 9.600716432102e-04;
C[3567] = 1.250159801996e-05;
C[3568] = 2.688259052451e-03;
C[3569] = 1.934181118226e-04;
C[3570] = -1.288984699791e-05;
C[3571] = 2.706311406925e-03;
C[3572] = 2.795675330964e-04;
C[3573] = -3.013976108020e-06;
C[3574] = -1.490661503438e-05;
C[3575] = 2.646428447955e-06;
C[3576] = -3.890234928919e-03;
C[3577] = -4.604471120318e-03;
C[3578] = 7.655804910743e-02;
C[3579] = 1.304144158810e-03;
C[3580] = 9.870145180707e-03;
C[3581] = 3.695279892566e-01;
C[3582] = 3.731882708998e-04;
C[3583] = 3.435530052446e-03;
C[3584] = 5.398985847045e-02;
C[3585] = -2.239991624397e-04;
C[3586] = -6.322955011695e-04;
C[3587] = 1.304909679386e-01;
C[3588] = 5.286143196853e-05;
C[3589] = 9.729768632717e-04;
C[3590] = 3.048569375066e-02;
C[3591] = 2.023671497788e-05;
C[3592] = 1.821000887907e-04;
C[3593] = 2.648487251007e-03;
C[3594] = 2.811191388736e-05;
C[3595] = 2.550652122157e-04;
C[3596] = 2.820233296437e-03;
C[3597] = -4.849650706804e-07;
C[3598] = 2.329976335496e-06;
C[3599] = -1.127641977027e-05;
C[3600] = 3.705141194045e-01;
C[3601] = 6.823554051184e-03;
C[3602] = 7.607385948858e-04;
C[3603] = 1.337068749759e-01;
C[3604] = -6.920623777435e-03;
C[3605] = -4.064929858929e-03;
C[3606] = 3.824079554259e-01;
C[3607] = 6.104094898467e-04;
C[3608] = 1.723896060298e-03;
C[3609] = 3.031649333772e-02;
C[3610] = 3.144115463817e-04;
C[3611] = 1.663439437360e-05;
C[3612] = 1.337574221234e-01;
C[3613] = -4.459995491526e-04;
C[3614] = -4.999015445069e-04;
C[3615] = 3.098637552762e-02;
C[3616] = 1.641274988996e-04;
C[3617] = 3.975456935078e-04;
C[3618] = 5.790148600546e-02;
C[3619] = 5.270351942737e-04;
C[3620] = 1.312927776412e-03;
C[3621] = 2.511193276970e-03;
C[3622] = 3.162425944816e-05;
C[3623] = 7.728220150162e-05;
C[3624] = 6.861479512585e-03;
C[3625] = 3.994378822865e-01;
C[3626] = 1.042970908986e-02;
C[3627] = -6.869156224938e-03;
C[3628] = 9.561259956575e-02;
C[3629] = -2.107604162114e-02;
C[3630] = 6.568487188539e-04;
C[3631] = 3.822892538544e-01;
C[3632] = 7.114738480354e-03;
C[3633] = 3.142977285519e-04;
C[3634] = 3.281595953224e-02;
C[3635] = 1.010281123882e-03;
C[3636] = -4.425683789861e-04;
C[3637] = 1.307689312393e-01;
C[3638] = -1.940169656430e-03;
C[3639] = 1.665586669179e-04;
C[3640] = 3.105476490598e-02;
C[3641] = 7.509621581133e-04;
C[3642] = 5.190142080203e-04;
C[3643] = 5.838671655492e-02;
C[3644] = 2.521903228921e-03;
C[3645] = 3.103932509937e-05;
C[3646] = 2.537875639840e-03;
C[3647] = 1.229201228552e-04;
C[3648] = 6.969135443402e-04;
C[3649] = 9.849832858091e-03;
C[3650] = 3.697812698749e-01;
C[3651] = -4.059184916501e-03;
C[3652] = -2.087497924864e-02;
C[3653] = 1.106575172394e-01;
C[3654] = 1.856815676147e-03;
C[3655] = 7.328121080272e-03;
C[3656] = 3.991139147286e-01;
C[3657] = 1.345366665070e-05;
C[3658] = 9.783677074912e-04;
C[3659] = 3.050257912521e-02;
C[3660] = -4.995185483451e-04;
C[3661] = -1.926765498263e-03;
C[3662] = 1.314550356585e-01;
C[3663] = 4.049451290490e-04;
C[3664] = 7.612176358380e-04;
C[3665] = 3.274734343833e-02;
C[3666] = 1.280791054695e-03;
C[3667] = 2.513071271636e-03;
C[3668] = 6.456107903899e-02;
C[3669] = 7.486197744111e-05;
C[3670] = 1.216717976923e-04;
C[3671] = 2.831897719568e-03;
C[3672] = 4.905968487800e-02;
C[3673] = 1.461543198750e-03;
C[3674] = 1.142447436743e-03;
C[3675] = 3.833554263669e-01;
C[3676] = -1.487380118967e-04;
C[3677] = 1.477402260416e-03;
C[3678] = 5.537526548444e-02;
C[3679] = -1.404232210896e-03;
C[3680] = -3.362163170605e-03;
C[3681] = 2.359290668098e-03;
C[3682] = 7.563557463507e-05;
C[3683] = 7.152065936742e-05;
C[3684] = 3.104954025702e-02;
C[3685] = 1.128528848679e-04;
C[3686] = 3.796509013335e-04;
C[3687] = 1.293259152351e-01;
C[3688] = -2.141125676699e-04;
C[3689] = -5.758383554140e-04;
C[3690] = 5.099897742948e-01;
C[3691] = 5.486639366439e-05;
C[3692] = 6.270429159050e-04;
C[3693] = 3.728670979155e-02;
C[3694] = 2.418288295207e-05;
C[3695] = 1.215632009228e-04;
C[3696] = 1.433538541117e-03;
C[3697] = 5.474214088331e-02;
C[3698] = 3.296800082914e-03;
C[3699] = -5.331323547171e-05;
C[3700] = 3.823348158048e-01;
C[3701] = 7.530564346706e-03;
C[3702] = -1.461217382862e-03;
C[3703] = 4.901330946084e-02;
C[3704] = -1.523971294188e-02;
C[3705] = 7.218434235446e-05;
C[3706] = 2.653324866118e-03;
C[3707] = 1.728514241552e-04;
C[3708] = 1.198793172117e-04;
C[3709] = 3.105780236934e-02;
C[3710] = 7.747138536003e-04;
C[3711] = -2.179115791343e-04;
C[3712] = 1.289300112556e-01;
C[3713] = -1.318768742332e-03;
C[3714] = 6.261306580697e-05;
C[3715] = 5.111817664140e-01;
C[3716] = 4.272684263669e-03;
C[3717] = 2.483122542699e-05;
C[3718] = 3.737723537386e-02;
C[3719] = 3.708812239965e-04;
C[3720] = 1.182986242617e-03;
C[3721] = 3.524326342592e-03;
C[3722] = 5.266646139356e-02;
C[3723] = 1.529453785467e-03;
C[3724] = 7.178272901787e-03;
C[3725] = 4.003681942986e-01;
C[3726] = -3.489124551712e-03;
C[3727] = -1.517420806823e-02;
C[3728] = 1.800465004528e-02;
C[3729] = 7.253282630142e-05;
C[3730] = 1.864442726713e-04;
C[3731] = 2.560260779214e-03;
C[3732] = 3.845828751857e-04;
C[3733] = 7.551977862088e-04;
C[3734] = 3.283096207632e-02;
C[3735] = -5.843024474878e-04;
C[3736] = -1.314401750756e-03;
C[3737] = 1.259009340265e-01;
C[3738] = 6.639543439808e-04;
C[3739] = 4.298860308809e-03;
C[3740] = 5.264966795466e-01;
C[3741] = 1.242314097398e-04;
C[3742] = 3.717969236812e-04;
C[3743] = 3.872415055647e-02;
C[3744] = -7.034049886049e+00;
C[3745] = 4.500893682768e-02;
C[3746] = -1.308395146004e-02;
C[3747] = 2.117619694852e+00;
C[3748] = 2.721195859329e-02;
C[3749] = -4.675865026560e-03;
C[3750] = 7.842026214609e-01;
C[3751] = 3.804732084470e-03;
C[3752] = 3.754872312063e-03;
C[3753] = 6.628778300154e-02;
C[3754] = 1.004742943301e-02;
C[3755] = -1.462011750545e-03;
C[3756] = 3.769903936733e-01;
C[3757] = 1.238605631338e-03;
C[3758] = -7.395645464302e-04;
C[3759] = 5.040630324802e-02;
C[3760] = 2.637971116035e-04;
C[3761] = 5.179951042156e-04;
C[3762] = 7.706230132335e-02;
C[3763] = 5.179708574108e-04;
C[3764] = 1.507960257214e-03;
C[3765] = 2.374032213452e-03;
C[3766] = 1.761086725796e-05;
C[3767] = 7.491511283039e-05;
C[3768] = 4.948962669830e-02;
C[3769] = -7.212921909351e+00;
C[3770] = 1.749934641698e-03;
C[3771] = 2.137317790770e-02;
C[3772] = 2.122159437467e+00;
C[3773] = 5.339265889154e-03;
C[3774] = 3.743812768450e-03;
C[3775] = 8.487949378954e-01;
C[3776] = 5.132393997075e-02;
C[3777] = 1.027146392654e-02;
C[3778] = 2.086456360655e-02;
C[3779] = -7.924776593626e-03;
C[3780] = 8.895554036303e-04;
C[3781] = 3.983586041893e-01;
C[3782] = 1.007347235101e-02;
C[3783] = 3.007558118482e-04;
C[3784] = 5.491068563208e-02;
C[3785] = 3.626155108883e-03;
C[3786] = 6.461610737068e-04;
C[3787] = 8.497234052402e-02;
C[3788] = 7.555551314142e-03;
C[3789] = 2.570438511134e-05;
C[3790] = 2.660401924904e-03;
C[3791] = 2.733948939337e-04;
C[3792] = -1.231345787589e-02;
C[3793] = 4.697585964480e-03;
C[3794] = -7.032117079183e+00;
C[3795] = -6.599229018814e-03;
C[3796] = 8.387056125874e-03;
C[3797] = 2.086441421079e+00;
C[3798] = 3.854674606777e-03;
C[3799] = 4.938266345326e-02;
C[3800] = 8.405050663915e-01;
C[3801] = -1.423487071338e-03;
C[3802] = -7.777394027487e-03;
C[3803] = 6.445350438567e-02;
C[3804] = -8.679993086681e-04;
C[3805] = 1.033505046284e-02;
C[3806] = 3.793664073680e-01;
C[3807] = 5.449440577240e-04;
C[3808] = 3.452405963741e-03;
C[3809] = 5.457242528349e-02;
C[3810] = 1.545568778148e-03;
C[3811] = 6.954197237172e-03;
C[3812] = 8.822335724653e-02;
C[3813] = 7.693785304799e-05;
C[3814] = 2.482681169663e-04;
C[3815] = 2.779411777846e-03;
C[3816] = 2.129194906051e+00;
C[3817] = 2.064963083404e-02;
C[3818] = -1.746734029729e-02;
C[3819] = -6.192294599515e+00;
C[3820] = -1.003523242150e-02;
C[3821] = -1.378650315861e-02;
C[3822] = 2.134836774908e+00;
C[3823] = 2.450670156911e-03;
C[3824] = 4.287420394105e-03;
C[3825] = 3.775691542332e-01;
C[3826] = 8.533780499474e-04;
C[3827] = -1.411404872592e-03;
C[3828] = 1.286219275255e-01;
C[3829] = -2.651226849551e-03;
C[3830] = -6.850286260974e-03;
C[3831] = 3.799578304340e-01;
C[3832] = 2.614277590166e-03;
C[3833] = 6.448031744522e-03;
C[3834] = 9.798202815976e-01;
C[3835] = 9.387603328870e-03;
C[3836] = 2.465817257890e-02;
C[3837] = 5.774738727564e-02;
C[3838] = 6.213071889002e-04;
C[3839] = 1.664387152611e-03;
C[3840] = 2.046304933514e-02;
C[3841] = 2.134453361459e+00;
C[3842] = 1.957833430892e-02;
C[3843] = -9.005881371561e-03;
C[3844] = -6.276392332318e+00;
C[3845] = -5.162027625918e-02;
C[3846] = 3.245542691312e-03;
C[3847] = 2.079728381513e+00;
C[3848] = -2.096394176913e-02;
C[3849] = 9.011601684307e-04;
C[3850] = 3.989733003889e-01;
C[3851] = 1.089461437199e-02;
C[3852] = -2.599759297054e-03;
C[3853] = 1.013398697328e-01;
C[3854] = -2.110168246724e-02;
C[3855] = 2.622165222008e-03;
C[3856] = 3.774697903495e-01;
C[3857] = 6.528086704448e-03;
C[3858] = 9.117875161542e-03;
C[3859] = 9.816973590292e-01;
C[3860] = 3.754620982591e-02;
C[3861] = 6.044166100342e-04;
C[3862] = 5.800794518143e-02;
C[3863] = 2.460164529700e-03;
C[3864] = -1.845365565058e-02;
C[3865] = 1.082808517346e-02;
C[3866] = 2.091507033445e+00;
C[3867] = -1.367160431005e-02;
C[3868] = -4.759902880924e-02;
C[3869] = -6.256519984051e+00;
C[3870] = 6.503170945824e-03;
C[3871] = -1.872526570008e-02;
C[3872] = 2.125113688847e+00;
C[3873] = -1.428454077631e-03;
C[3874] = 1.034791331522e-02;
C[3875] = 3.796196879863e-01;
C[3876] = -6.844541318547e-03;
C[3877] = -2.090062009474e-02;
C[3878] = 1.013418636612e-01;
C[3879] = 6.549160495977e-03;
C[3880] = 6.715361678627e-03;
C[3881] = 3.979252931510e-01;
C[3882] = 2.396744764858e-02;
C[3883] = 3.774116258339e-02;
C[3884] = 1.086046422385e+00;
C[3885] = 1.616072693881e-03;
C[3886] = 2.455294187774e-03;
C[3887] = 6.457043311133e-02;
C[3888] = 7.631198162114e-01;
C[3889] = 2.268400575298e-02;
C[3890] = 1.923985792364e-02;
C[3891] = 2.153786193728e+00;
C[3892] = -1.286619192370e-02;
C[3893] = -1.085097368800e-03;
C[3894] = -7.090290607925e+00;
C[3895] = -6.088016408317e-03;
C[3896] = -1.721185746885e-02;
C[3897] = 4.935216298554e-02;
C[3898] = 1.247765461075e-03;
C[3899] = 1.314203223567e-03;
C[3900] = 3.809053013750e-01;
C[3901] = 1.816578491257e-03;
C[3902] = 6.169747080245e-03;
C[3903] = 5.845699868940e-02;
C[3904] = -3.302881960379e-03;
C[3905] = -9.344381562634e-03;
C[3906] = 4.099676928703e+00;
C[3907] = -3.572291641687e-03;
C[3908] = -3.398505775757e-03;
C[3909] = 5.089968589307e-01;
C[3910] = 2.665570632339e-04;
C[3911] = 1.774459397422e-03;
C[3912] = 2.197055579580e-02;
C[3913] = 8.383146819954e-01;
C[3914] = 4.660806406262e-02;
C[3915] = -1.082378434946e-02;
C[3916] = 2.080639620520e+00;
C[3917] = -1.467640037139e-02;
C[3918] = -7.227719847639e-03;
C[3919] = -7.109100878599e+00;
C[3920] = -4.726184897892e-02;
C[3921] = 1.172088297170e-03;
C[3922] = 5.438667283708e-02;
C[3923] = 3.313675994210e-03;
C[3924] = 1.950554864848e-03;
C[3925] = 3.775153522998e-01;
C[3926] = 6.917804945062e-03;
C[3927] = -3.359867132345e-03;
C[3928] = 5.564685707876e-02;
C[3929] = -1.426461687881e-02;
C[3930] = -3.374732768411e-03;
C[3931] = 4.103060344189e+00;
C[3932] = 1.138829016215e-02;
C[3933] = 2.792969745766e-04;
C[3934] = 5.097122075187e-01;
C[3935] = 3.886868237961e-03;
C[3936] = 1.959729957966e-02;
C[3937] = 4.923178822535e-02;
C[3938] = 8.140371248535e-01;
C[3939] = 3.985748974711e-04;
C[3940] = -1.969325334047e-02;
C[3941] = 2.150199280246e+00;
C[3942] = -1.975108509098e-02;
C[3943] = -4.595175150592e-02;
C[3944] = -7.223700100101e+00;
C[3945] = 1.310116467595e-03;
C[3946] = 3.521547521614e-03;
C[3947] = 5.324902820659e-02;
C[3948] = 6.253589469690e-03;
C[3949] = 6.591621125881e-03;
C[3950] = 3.991795727210e-01;
C[3951] = -9.471342943740e-03;
C[3952] = -1.419911200516e-02;
C[3953] = 1.843402381130e-02;
C[3954] = -2.585838773696e-03;
C[3955] = 1.171581888262e-02;
C[3956] = 4.179276635730e+00;
C[3957] = 1.828757538612e-03;
C[3958] = 3.909256197630e-03;
C[3959] = 5.265763455176e-01;
C[3960] = 0.000000000000e+00;
C[3961] = 0.000000000000e+00;
C[3962] = 0.000000000000e+00;
C[3963] = 0.000000000000e+00;
C[3964] = 0.000000000000e+00;
C[3965] = 0.000000000000e+00;
C[3966] = 0.000000000000e+00;
C[3967] = 0.000000000000e+00;
C[3968] = 0.000000000000e+00;
C[3969] = 0.000000000000e+00;
C[3970] = 0.000000000000e+00;
C[3971] = 0.000000000000e+00;
C[3972] = 0.000000000000e+00;
C[3973] = 0.000000000000e+00;
C[3974] = 0.000000000000e+00;
C[3975] = 0.000000000000e+00;
C[3976] = 0.000000000000e+00;
C[3977] = 0.000000000000e+00;
C[3978] = 1.000000000000e+00;
C[3979] = 0.000000000000e+00;
C[3980] = 0.000000000000e+00;
C[3981] = 2.000000000000e-01;
C[3982] = -0.000000000000e+00;
C[3983] = -0.000000000000e+00;
C[3984] = 0.000000000000e+00;
C[3985] = 0.000000000000e+00;
C[3986] = 0.000000000000e+00;
C[3987] = 0.000000000000e+00;
C[3988] = 0.000000000000e+00;
C[3989] = 0.000000000000e+00;
C[3990] = 0.000000000000e+00;
C[3991] = 0.000000000000e+00;
C[3992] = 0.000000000000e+00;
C[3993] = 0.000000000000e+00;
C[3994] = 0.000000000000e+00;
C[3995] = 0.000000000000e+00;
C[3996] = 0.000000000000e+00;
C[3997] = 0.000000000000e+00;
C[3998] = 0.000000000000e+00;
C[3999] = 0.000000000000e+00;
C[4000] = 0.000000000000e+00;
C[4001] = 0.000000000000e+00;
C[4002] = 0.000000000000e+00;
C[4003] = 1.000000000000e+00;
C[4004] = 0.000000000000e+00;
C[4005] = -0.000000000000e+00;
C[4006] = 2.000000000000e-01;
C[4007] = -0.000000000000e+00;
C[4008] = 0.000000000000e+00;
C[4009] = 0.000000000000e+00;
C[4010] = 0.000000000000e+00;
C[4011] = 0.000000000000e+00;
C[4012] = 0.000000000000e+00;
C[4013] = 0.000000000000e+00;
C[4014] = 0.000000000000e+00;
C[4015] = 0.000000000000e+00;
C[4016] = 0.000000000000e+00;
C[4017] = 0.000000000000e+00;
C[4018] = 0.000000000000e+00;
C[4019] = 0.000000000000e+00;
C[4020] = 0.000000000000e+00;
C[4021] = 0.000000000000e+00;
C[4022] = 0.000000000000e+00;
C[4023] = 0.000000000000e+00;
C[4024] = 0.000000000000e+00;
C[4025] = 0.000000000000e+00;
C[4026] = 0.000000000000e+00;
C[4027] = 0.000000000000e+00;
C[4028] = 1.000000000000e+00;
C[4029] = -0.000000000000e+00;
C[4030] = -0.000000000000e+00;
C[4031] = 2.000000000000e-01;
C[4032] = 6.013004628700e-02;
C[4033] = 1.361550502263e-02;
C[4034] = 2.440464913889e-04;
C[4035] = 3.834278075146e-01;
C[4036] = -2.850622270829e-03;
C[4037] = -3.594244447866e-03;
C[4038] = 5.171227199694e-02;
C[4039] = -1.688446043804e-04;
C[4040] = 7.910530281516e-04;
C[4041] = 1.295564912058e-01;
C[4042] = 1.216658670250e-03;
C[4043] = 7.443219624055e-05;
C[4044] = 3.130355844753e-02;
C[4045] = -3.803850709934e-04;
C[4046] = -3.171875033939e-04;
C[4047] = 2.505079531072e-03;
C[4048] = -1.653079821447e-05;
C[4049] = 3.935767813705e-05;
C[4050] = 2.466396644143e-03;
C[4051] = -2.794642736207e-05;
C[4052] = 6.745604317508e-05;
C[4053] = -1.686697743049e-05;
C[4054] = -1.661324908542e-06;
C[4055] = -1.080668078189e-06;
C[4056] = 1.365739997785e-02;
C[4057] = 2.215048322570e-02;
C[4058] = -6.057653259759e-03;
C[4059] = -3.015582154906e-03;
C[4060] = 3.977120629970e-01;
C[4061] = 9.317530535546e-03;
C[4062] = -9.601108473354e-05;
C[4063] = 5.424015899913e-02;
C[4064] = 3.222486624087e-03;
C[4065] = 1.219451667264e-03;
C[4066] = 1.262840697351e-01;
C[4067] = -4.895926725298e-04;
C[4068] = -3.893259624562e-04;
C[4069] = 3.263219850841e-02;
C[4070] = 8.707145158647e-04;
C[4071] = -1.217664129414e-05;
C[4072] = 2.640128933784e-03;
C[4073] = 1.687686706089e-04;
C[4074] = -1.993963517683e-05;
C[4075] = 2.630317694348e-03;
C[4076] = 2.342026451021e-04;
C[4077] = -1.541545324109e-06;
C[4078] = -1.432850812019e-05;
C[4079] = 2.220247632606e-06;
C[4080] = 2.711025575759e-04;
C[4081] = -6.086804438905e-03;
C[4082] = 5.836399024307e-02;
C[4083] = -3.752908148518e-03;
C[4084] = 9.601731049008e-03;
C[4085] = 3.856413578830e-01;
C[4086] = 8.646573108075e-04;
C[4087] = 3.118037202228e-03;
C[4088] = 5.534818073798e-02;
C[4089] = 7.623593398636e-05;
C[4090] = -4.915360844729e-04;
C[4091] = 1.293637367105e-01;
C[4092] = -3.259177963267e-04;
C[4093] = 8.854273414930e-04;
C[4094] = 3.156230016767e-02;
C[4095] = 4.380139252700e-05;
C[4096] = 1.623630529707e-04;
C[4097] = 2.698698319794e-03;
C[4098] = 7.483578626371e-05;
C[4099] = 2.161934954700e-04;
C[4100] = 2.837575645694e-03;
C[4101] = -9.156480682748e-07;
C[4102] = 1.954126793095e-06;
C[4103] = -1.494221872071e-05;
C[4104] = 3.827730836151e-01;
C[4105] = -2.747366498668e-03;
C[4106] = -3.236545407446e-03;
C[4107] = 1.200120150981e-01;
C[4108] = 3.992525431911e-04;
C[4109] = -6.057307658333e-03;
C[4110] = 3.811148810235e-01;
C[4111] = 4.081963910214e-03;
C[4112] = 7.813255117385e-03;
C[4113] = 3.125991018756e-02;
C[4114] = -3.714449187070e-04;
C[4115] = -2.914936135885e-04;
C[4116] = 1.325268127418e-01;
C[4117] = 1.414483867469e-04;
C[4118] = -3.576315158349e-04;
C[4119] = 3.114151603001e-02;
C[4120] = 3.178503073016e-04;
C[4121] = 5.893273857160e-04;
C[4122] = 5.819274653774e-02;
C[4123] = 7.137861739553e-04;
C[4124] = 1.805587492582e-03;
C[4125] = 2.542894518501e-03;
C[4126] = 2.999644450396e-05;
C[4127] = 7.556070011806e-05;
C[4128] = -2.681801796479e-03;
C[4129] = 3.969905259567e-01;
C[4130] = 9.271242119315e-03;
C[4131] = 4.798800414691e-04;
C[4132] = 1.011720828391e-01;
C[4133] = -1.943723210058e-02;
C[4134] = 3.955948135621e-03;
C[4135] = 3.794615979790e-01;
C[4136] = 6.723399962705e-03;
C[4137] = -3.691303727034e-04;
C[4138] = 3.258409603906e-02;
C[4139] = 8.633947461801e-04;
C[4140] = 1.468235532988e-04;
C[4141] = 1.308028442185e-01;
C[4142] = -1.779719936853e-03;
C[4143] = 3.103679898863e-04;
C[4144] = 3.123782850691e-02;
C[4145] = 7.446546702551e-04;
C[4146] = 7.149667751334e-04;
C[4147] = 5.862232798803e-02;
C[4148] = 2.322353337563e-03;
C[4149] = 3.056036931544e-05;
C[4150] = 2.578176783987e-03;
C[4151] = 1.161075118823e-04;
C[4152] = -3.128821703190e-03;
C[4153] = 8.985319025873e-03;
C[4154] = 3.854469874177e-01;
C[4155] = -5.982678485210e-03;
C[4156] = -1.947164756361e-02;
C[4157] = 9.653050601955e-02;
C[4158] = 7.712618742541e-03;
C[4159] = 6.963052469801e-03;
C[4160] = 3.957460273011e-01;
C[4161] = -2.861593204155e-04;
C[4162] = 8.485670818865e-04;
C[4163] = 3.154934213665e-02;
C[4164] = -3.526562376267e-04;
C[4165] = -1.782014301055e-03;
C[4166] = 1.305185646259e-01;
C[4167] = 5.825673336131e-04;
C[4168] = 7.580955813721e-04;
C[4169] = 3.256780936833e-02;
C[4170] = 1.775695271157e-03;
C[4171] = 2.273095458220e-03;
C[4172] = 6.433631944826e-02;
C[4173] = 7.488304605328e-05;
C[4174] = 1.127092328658e-04;
C[4175] = 2.828948177768e-03;
C[4176] = 5.171152581413e-02;
C[4177] = -5.362387372282e-06;
C[4178] = 5.667941704471e-04;
C[4179] = 3.814417175717e-01;
C[4180] = 3.807923881591e-03;
C[4181] = 7.394874205873e-03;
C[4182] = 5.749684705536e-02;
C[4183] = -4.144515332566e-03;
C[4184] = -1.272063383958e-02;
C[4185] = 2.505029785551e-03;
C[4186] = -6.133394803394e-06;
C[4187] = 2.394384983631e-05;
C[4188] = 3.116330513322e-02;
C[4189] = 3.004997062844e-04;
C[4190] = 5.613843645019e-04;
C[4191] = 1.292557612629e-01;
C[4192] = -3.392500401588e-04;
C[4193] = -1.048042545805e-03;
C[4194] = 5.068958991837e-01;
C[4195] = 4.282876773415e-04;
C[4196] = 4.735382843119e-03;
C[4197] = 3.709132473781e-02;
C[4198] = 4.600702512226e-05;
C[4199] = 4.640102369339e-04;
C[4200] = -7.216296267294e-05;
C[4201] = 5.506192465495e-02;
C[4202] = 3.270430315278e-03;
C[4203] = 3.865309039006e-03;
C[4204] = 3.795795297787e-01;
C[4205] = 6.961048716191e-03;
C[4206] = -4.119034384668e-03;
C[4207] = 5.693640948639e-02;
C[4208] = -1.266871326847e-02;
C[4209] = -1.008535543398e-05;
C[4210] = 2.694913310839e-03;
C[4211] = 1.725225938407e-04;
C[4212] = 3.034066492211e-04;
C[4213] = 3.124569062689e-02;
C[4214] = 7.579619977981e-04;
C[4215] = -3.375513102990e-04;
C[4216] = 1.291141391659e-01;
C[4217] = -1.114991184136e-03;
C[4218] = 4.224536158810e-04;
C[4219] = 5.055908864841e-01;
C[4220] = 2.258227799068e-03;
C[4221] = 4.558713972025e-05;
C[4222] = 3.695785498529e-02;
C[4223] = 1.812151850631e-04;
C[4224] = 5.049010076705e-04;
C[4225] = 3.376171145144e-03;
C[4226] = 5.525617169568e-02;
C[4227] = 7.396897243483e-03;
C[4228] = 6.806582090270e-03;
C[4229] = 3.959275336361e-01;
C[4230] = -1.268284063106e-02;
C[4231] = -1.269950906251e-02;
C[4232] = 1.962381406648e-02;
C[4233] = 2.028087677165e-05;
C[4234] = 1.790143053461e-04;
C[4235] = 2.692564383640e-03;
C[4236] = 5.615701941225e-04;
C[4237] = 7.502001454261e-04;
C[4238] = 3.257990979066e-02;
C[4239] = -1.045522998570e-03;
C[4240] = -1.117044237072e-03;
C[4241] = 1.260372179739e-01;
C[4242] = 4.767329651230e-03;
C[4243] = 2.315598697212e-03;
C[4244] = 5.265538484870e-01;
C[4245] = 4.651863217583e-04;
C[4246] = 1.841122578419e-04;
C[4247] = 3.870422129488e-02;
C[4248] = -7.062135250008e+00;
C[4249] = 5.946096272243e-02;
C[4250] = -8.360509852774e-03;
C[4251] = 2.149442812238e+00;
C[4252] = 3.971758359702e-03;
C[4253] = -2.795710809642e-02;
C[4254] = 8.038926655460e-01;
C[4255] = -2.778176788896e-03;
C[4256] = 9.719934977431e-03;
C[4257] = 6.019076116176e-02;
C[4258] = 1.469569754893e-02;
C[4259] = 6.962487986329e-04;
C[4260] = 3.857213076187e-01;
C[4261] = -4.436021766942e-03;
C[4262] = -4.458171911549e-03;
C[4263] = 5.198191426663e-02;
C[4264] = -3.147559319330e-04;
C[4265] = 5.487121929245e-04;
C[4266] = 7.948229770123e-02;
C[4267] = -5.112626141771e-04;
C[4268] = 1.663988396685e-03;
C[4269] = 2.480244691742e-03;
C[4270] = -3.065739568991e-05;
C[4271] = 3.765602788429e-05;
C[4272] = 6.029886182687e-02;
C[4273] = -7.207800436705e+00;
C[4274] = 9.174401782170e-03;
C[4275] = 1.582575852641e-03;
C[4276] = 2.129157623583e+00;
C[4277] = 9.803632705549e-03;
C[4278] = -1.946358404622e-03;
C[4279] = 8.353518437215e-01;
C[4280] = 4.482637171016e-02;
C[4281] = 1.473759250415e-02;
C[4282] = 2.301088236235e-02;
C[4283] = -3.132290465074e-03;
C[4284] = -4.576026818629e-03;
C[4285] = 3.966352363966e-01;
C[4286] = 8.639598125559e-03;
C[4287] = -2.470627356704e-04;
C[4288] = 5.452646447660e-02;
C[4289] = 3.192539742455e-03;
C[4290] = -3.193663473911e-04;
C[4291] = 8.420844698768e-02;
C[4292] = 6.509167541864e-03;
C[4293] = -2.275890875662e-05;
C[4294] = 2.688955970704e-03;
C[4295] = 2.419471590973e-04;
C[4296] = -7.819388529034e-03;
C[4297] = 8.591378199246e-03;
C[4298] = -7.061013102153e+00;
C[4299] = -3.029920657196e-02;
C[4300] = 1.373783040505e-02;
C[4301] = 2.118592672498e+00;
C[4302] = 1.060127106435e-02;
C[4303] = 4.390751712019e-02;
C[4304] = 8.526223468254e-01;
C[4305] = 7.233048648199e-04;
C[4306] = -3.161441644220e-03;
C[4307] = 5.817653715660e-02;
C[4308] = -4.613273969122e-03;
C[4309] = 8.912508164624e-03;
C[4310] = 3.867793668152e-01;
C[4311] = 6.188342305690e-04;
C[4312] = 3.107189369548e-03;
C[4313] = 5.543660576548e-02;
C[4314] = 1.842909505103e-03;
C[4315] = 6.076325380210e-03;
C[4316] = 8.973707255901e-02;
C[4317] = 4.560618100531e-05;
C[4318] = 2.240147250227e-04;
C[4319] = 2.837166580333e-03;
C[4320] = 2.136348334250e+00;
C[4321] = 6.946888977415e-03;
C[4322] = -1.997195175050e-02;
C[4323] = -6.229004557865e+00;
C[4324] = 5.198489476710e-03;
C[4325] = -1.163131183682e-02;
C[4326] = 2.124742649783e+00;
C[4327] = 1.738282528541e-02;
C[4328] = 1.111587425518e-02;
C[4329] = 3.850665837192e-01;
C[4330] = -4.307811162390e-03;
C[4331] = -4.096911228050e-03;
C[4332] = 1.152572692139e-01;
C[4333] = 2.803156989621e-03;
C[4334] = -1.845867056445e-03;
C[4335] = 3.834738950783e-01;
C[4336] = 3.361678964443e-03;
C[4337] = 4.575564922647e-03;
C[4338] = 9.866649278185e-01;
C[4339] = 1.104379577255e-02;
C[4340] = 2.669795145735e-02;
C[4341] = 5.832144618047e-02;
C[4342] = 6.516238664351e-04;
C[4343] = 1.712197761811e-03;
C[4344] = 7.348167846706e-03;
C[4345] = 2.114726882776e+00;
C[4346] = 7.128051811178e-03;
C[4347] = 6.811039442271e-03;
C[4348] = -6.275652893863e+00;
C[4349] = -4.301313370532e-02;
C[4350] = 1.543227179898e-02;
C[4351] = 2.095537810460e+00;
C[4352] = -1.407308228720e-02;
C[4353] = -4.267201292592e-03;
C[4354] = 3.959136993563e-01;
C[4355] = 8.582019234930e-03;
C[4356] = 2.883784487899e-03;
C[4357] = 9.650794193995e-02;
C[4358] = -1.903618070451e-02;
C[4359] = 3.230888364028e-03;
C[4360] = 3.850368034565e-01;
C[4361] = 7.259215332438e-03;
C[4362] = 1.106437217748e-02;
C[4363] = 9.933550898254e-01;
C[4364] = 3.519568247571e-02;
C[4365] = 6.647747896642e-04;
C[4366] = 5.882430769338e-02;
C[4367] = 2.269938258179e-03;
C[4368] = -1.864865320290e-02;
C[4369] = 3.159402512081e-03;
C[4370] = 2.114705263192e+00;
C[4371] = -1.013872837434e-02;
C[4372] = -4.370144296590e-02;
C[4373] = -6.281973487967e+00;
C[4374] = 9.358155206598e-03;
C[4375] = -1.055981899967e-02;
C[4376] = 2.106553709378e+00;
C[4377] = -3.992749166873e-03;
C[4378] = 8.307386615886e-03;
C[4379] = 3.865849963499e-01;
C[4380] = -1.771237883321e-03;
C[4381] = -1.907059616754e-02;
C[4382] = 9.504255396641e-02;
C[4383] = 4.456690097938e-03;
C[4384] = 7.494203306460e-03;
C[4385] = 3.961075761357e-01;
C[4386] = 2.634452501261e-02;
C[4387] = 3.449842786225e-02;
C[4388] = 1.081813301729e+00;
C[4389] = 1.702964817780e-03;
C[4390] = 2.204062590945e-03;
C[4391] = 6.433377479373e-02;
C[4392] = 8.038777418897e-01;
C[4393] = -1.333844573971e-04;
C[4394] = 4.644008257141e-03;
C[4395] = 2.131279380747e+00;
C[4396] = 1.247178671839e-02;
C[4397] = 3.003264473237e-03;
C[4398] = -7.086961854836e+00;
C[4399] = -8.264311317639e-03;
C[4400] = -4.032837693177e-02;
C[4401] = 5.198116808381e-02;
C[4402] = -1.564140383092e-04;
C[4403] = 3.209710902086e-04;
C[4404] = 3.838007316265e-01;
C[4405] = 3.082864109999e-03;
C[4406] = 4.138945561270e-03;
C[4407] = 5.515425454691e-02;
C[4408] = -3.182945909154e-03;
C[4409] = -1.091617780247e-02;
C[4410] = 4.072918388333e+00;
C[4411] = -1.468024460498e-03;
C[4412] = 3.185888954419e-02;
C[4413] = 5.065941440950e-01;
C[4414] = 3.568416612135e-04;
C[4415] = 6.430069379185e-03;
C[4416] = -8.445439547477e-04;
C[4417] = 8.517871568379e-01;
C[4418] = 4.695537938119e-02;
C[4419] = 1.304972786126e-02;
C[4420] = 2.097896446454e+00;
C[4421] = -1.059989407188e-02;
C[4422] = -7.754692359675e-03;
C[4423] = -7.089907785235e+00;
C[4424] = -3.433624456611e-02;
C[4425] = -2.180742902256e-04;
C[4426] = 5.534823013243e-02;
C[4427] = 3.259582482598e-03;
C[4428] = 3.145024093236e-03;
C[4429] = 3.851547352562e-01;
C[4430] = 7.492199552850e-03;
C[4431] = -3.157464961256e-03;
C[4432] = 5.268585481670e-02;
C[4433] = -1.193045865001e-02;
C[4434] = -1.629936832038e-03;
C[4435] = 4.051454299889e+00;
C[4436] = -6.509547440649e-03;
C[4437] = 3.391680803053e-04;
C[4438] = 5.040089317654e-01;
C[4439] = 1.006857758989e-03;
C[4440] = 3.996894567810e-03;
C[4441] = 4.790006213130e-02;
C[4442] = 8.507821659792e-01;
C[4443] = 2.788716777140e-03;
C[4444] = -1.240943973591e-02;
C[4445] = 2.110183836078e+00;
C[4446] = -3.957251276134e-02;
C[4447] = -3.495216044675e-02;
C[4448] = -7.223229142324e+00;
C[4449] = 2.625601724434e-04;
C[4450] = 3.346224263512e-03;
C[4451] = 5.534459672318e-02;
C[4452] = 4.159207048745e-03;
C[4453] = 7.342397460003e-03;
C[4454] = 3.962890824707e-01;
C[4455] = -1.087838459394e-02;
C[4456] = -1.196125444404e-02;
C[4457] = 1.957193592331e-02;
C[4458] = 3.220341952665e-02;
C[4459] = -5.530455814517e-03;
C[4460] = 4.178649362839e+00;
C[4461] = 6.440627207641e-03;
C[4462] = 1.080765197148e-03;
C[4463] = 5.261460062694e-01;
C[4464] = 0.000000000000e+00;
C[4465] = 0.000000000000e+00;
C[4466] = 0.000000000000e+00;
C[4467] = 0.000000000000e+00;
C[4468] = 0.000000000000e+00;
C[4469] = 0.000000000000e+00;
C[4470] = 0.000000000000e+00;
C[4471] = 0.000000000000e+00;
C[4472] = 0.000000000000e+00;
C[4473] = 0.000000000000e+00;
C[4474] = 0.000000000000e+00;
C[4475] = 0.000000000000e+00;
C[4476] = 0.000000000000e+00;
C[4477] = 0.000000000000e+00;
C[4478] = 0.000000000000e+00;
C[4479] = 0.000000000000e+00;
C[4480] = 0.000000000000e+00;
C[4481] = 0.000000000000e+00;
C[4482] = 1.000000000000e+00;
C[4483] = 0.000000000000e+00;
C[4484] = 0.000000000000e+00;
C[4485] = 2.000000000000e-01;
C[4486] = -0.000000000000e+00;
C[4487] = -0.000000000000e+00;
C[4488] = 0.000000000000e+00;
C[4489] = 0.000000000000e+00;
C[4490] = 0.000000000000e+00;
C[4491] = 0.000000000000e+00;
C[4492] = 0.000000000000e+00;
C[4493] = 0.000000000000e+00;
C[4494] = 0.000000000000e+00;
C[4495] = 0.000000000000e+00;
C[4496] = 0.000000000000e+00;
C[4497] = 0.000000000000e+00;
C[4498] = 0.000000000000e+00;
C[4499] = 0.000000000000e+00;
C[4500] = 0.000000000000e+00;
C[4501] = 0.000000000000e+00;
C[4502] = 0.000000000000e+00;
C[4503] = 0.000000000000e+00;
C[4504] = 0.000000000000e+00;
C[4505] = 0.000000000000e+00;
C[4506] = 0.000000000000e+00;
C[4507] = 1.000000000000e+00;
C[4508] = 0.000000000000e+00;
C[4509] = -0.000000000000e+00;
C[4510] = 2.000000000000e-01;
C[4511] = -0.000000000000e+00;
C[4512] = 0.000000000000e+00;
C[4513] = 0.000000000000e+00;
C[4514] = 0.000000000000e+00;
C[4515] = 0.000000000000e+00;
C[4516] = 0.000000000000e+00;
C[4517] = 0.000000000000e+00;
C[4518] = 0.000000000000e+00;
C[4519] = 0.000000000000e+00;
C[4520] = 0.000000000000e+00;
C[4521] = 0.000000000000e+00;
C[4522] = 0.000000000000e+00;
C[4523] = 0.000000000000e+00;
C[4524] = 0.000000000000e+00;
C[4525] = 0.000000000000e+00;
C[4526] = 0.000000000000e+00;
C[4527] = 0.000000000000e+00;
C[4528] = 0.000000000000e+00;
C[4529] = 0.000000000000e+00;
C[4530] = 0.000000000000e+00;
C[4531] = 0.000000000000e+00;
C[4532] = 1.000000000000e+00;
C[4533] = -0.000000000000e+00;
C[4534] = -0.000000000000e+00;
C[4535] = 2.000000000000e-01;
C[4536] = 6.725435924408e-02;
C[4537] = 1.384761595171e-02;
C[4538] = 3.026536743996e-04;
C[4539] = 3.838235895804e-01;
C[4540] = -5.471889578361e-03;
C[4541] = -3.861336333463e-03;
C[4542] = 5.170013862538e-02;
C[4543] = -7.616182593781e-04;
C[4544] = -3.145162373993e-04;
C[4545] = 1.305962000986e-01;
C[4546] = 1.062119468192e-03;
C[4547] = 1.356526102819e-05;
C[4548] = 3.093824185012e-02;
C[4549] = -4.969722528202e-04;
C[4550] = -2.832320393094e-04;
C[4551] = 2.481258755174e-03;
C[4552] = -4.036813597262e-05;
C[4553] = -1.281470841822e-05;
C[4554] = 2.389792097187e-03;
C[4555] = -4.336086433906e-05;
C[4556] = 4.108621165941e-05;
C[4557] = -2.239504813577e-05;
C[4558] = -5.721273465314e-07;
C[4559] = 1.327693181902e-06;
C[4560] = 1.372919379737e-02;
C[4561] = 2.463228138330e-02;
C[4562] = 1.296127047956e-03;
C[4563] = -5.440800364217e-03;
C[4564] = 3.957781446157e-01;
C[4565] = 1.014705135345e-02;
C[4566] = -7.140414463941e-04;
C[4567] = 5.619048324709e-02;
C[4568] = 3.510104940351e-03;
C[4569] = 1.054224657903e-03;
C[4570] = 1.265308871308e-01;
C[4571] = 1.296214227730e-04;
C[4572] = -4.938106693984e-04;
C[4573] = 3.261270105181e-02;
C[4574] = 9.849964605774e-04;
C[4575] = -3.804655344223e-05;
C[4576] = 2.745904040830e-03;
C[4577] = 1.860644800643e-04;
C[4578] = -3.650936718042e-05;
C[4579] = 2.734515796571e-03;
C[4580] = 2.376931611781e-04;
C[4581] = -4.956569868953e-07;
C[4582] = -1.473767392409e-05;
C[4583] = 1.442737287802e-06;
C[4584] = 3.080659460451e-04;
C[4585] = 1.226444683785e-03;
C[4586] = 6.135730180098e-02;
C[4587] = -3.951412904621e-03;
C[4588] = 1.051969632035e-02;
C[4589] = 3.838563460992e-01;
C[4590] = -2.695503787305e-04;
C[4591] = 3.415990384923e-03;
C[4592] = 5.443907481105e-02;
C[4593] = 1.392607913788e-05;
C[4594] = 1.249759318282e-04;
C[4595] = 1.298405194717e-01;
C[4596] = -2.872831141114e-04;
C[4597] = 1.006242770945e-03;
C[4598] = 3.108499360413e-02;
C[4599] = -1.030866711537e-05;
C[4600] = 1.790672844605e-04;
C[4601] = 2.633250982582e-03;
C[4602] = 4.881531917452e-05;
C[4603] = 2.133498265564e-04;
C[4604] = 2.711492443763e-03;
C[4605] = 1.443748408337e-06;
C[4606] = 7.903913712700e-07;
C[4607] = -1.992893353140e-05;
C[4608] = 3.832410774100e-01;
C[4609] = -5.067000907434e-03;
C[4610] = -3.718240803928e-03;
C[4611] = 1.146924297433e-01;
C[4612] = 5.452745787834e-03;
C[4613] = 1.919305273426e-03;
C[4614] = 3.872244344556e-01;
C[4615] = 1.660662008015e-03;
C[4616] = 8.237385361941e-04;
C[4617] = 3.089940770542e-02;
C[4618] = -4.688907056128e-04;
C[4619] = -2.717383073985e-04;
C[4620] = 1.324741931954e-01;
C[4621] = 5.074937661257e-04;
C[4622] = 2.250688986292e-04;
C[4623] = 3.163782381833e-02;
C[4624] = 5.949325636836e-05;
C[4625] = 6.993142620856e-06;
C[4626] = 5.915326124586e-02;
C[4627] = 1.142336374887e-04;
C[4628] = 9.717597271158e-04;
C[4629] = 2.584972705718e-03;
C[4630] = -5.522169526216e-07;
C[4631] = 3.728186647391e-05;
C[4632] = -4.958117813822e-03;
C[4633] = 3.951282672913e-01;
C[4634] = 9.923131110327e-03;
C[4635] = 5.423422489399e-03;
C[4636] = 9.216266113999e-02;
C[4637] = -1.890361387603e-02;
C[4638] = 1.564157973473e-03;
C[4639] = 3.894090073559e-01;
C[4640] = 6.780295787829e-03;
C[4641] = -4.627208018509e-04;
C[4642] = 3.256937589684e-02;
C[4643] = 9.664717569432e-04;
C[4644] = 5.055388795634e-04;
C[4645] = 1.300535021364e-01;
C[4646] = -1.780621662533e-03;
C[4647] = 5.450918026142e-05;
C[4648] = 3.195849030081e-02;
C[4649] = 6.977818862281e-04;
C[4650] = 1.364688874932e-04;
C[4651] = 5.964079537431e-02;
C[4652] = 1.888682159725e-03;
C[4653] = 7.982076643374e-07;
C[4654] = 2.617160089908e-03;
C[4655] = 9.390768092872e-05;
C[4656] = -3.587336411945e-03;
C[4657] = 9.643298840187e-03;
C[4658] = 3.840684724970e-01;
C[4659] = 1.930767905549e-03;
C[4660] = -1.898871685255e-02;
C[4661] = 9.713661963070e-02;
C[4662] = 6.974845406203e-04;
C[4663] = 6.981048602002e-03;
C[4664] = 3.979129299445e-01;
C[4665] = -2.649653778749e-04;
C[4666] = 9.514129596934e-04;
C[4667] = 3.109913536398e-02;
C[4668] = 2.258330741041e-04;
C[4669] = -1.786295194301e-03;
C[4670] = 1.308925506827e-01;
C[4671] = 2.980687913802e-08;
C[4672] = 7.092183324646e-04;
C[4673] = 3.272947682778e-02;
C[4674] = 1.008928283699e-03;
C[4675] = 1.831070982889e-03;
C[4676] = 6.451404375971e-02;
C[4677] = 3.959551284972e-05;
C[4678] = 9.065110961206e-05;
C[4679] = 2.835863337677e-03;
C[4680] = 5.258309911212e-02;
C[4681] = -9.855604869402e-04;
C[4682] = -6.223465045097e-04;
C[4683] = 3.863488516583e-01;
C[4684] = 1.698103988312e-03;
C[4685] = 1.064591759684e-03;
C[4686] = 5.181400244653e-02;
C[4687] = -1.414271885179e-03;
C[4688] = -7.745668985535e-03;
C[4689] = 2.540122787623e-03;
C[4690] = -5.614782281196e-05;
C[4691] = -3.382840883399e-05;
C[4692] = 3.157945163185e-02;
C[4693] = 6.343891458403e-05;
C[4694] = 2.450362148340e-05;
C[4695] = 1.287399104118e-01;
C[4696] = -7.742237626932e-05;
C[4697] = -4.873665462674e-04;
C[4698] = 5.068195445013e-01;
C[4699] = 8.378042852072e-04;
C[4700] = 7.323219642710e-03;
C[4701] = 3.715642991790e-02;
C[4702] = 7.263328203509e-05;
C[4703] = 4.973873193302e-04;
C[4704] = -1.015557980034e-03;
C[4705] = 5.606854912188e-02;
C[4706] = 3.145892699317e-03;
C[4707] = 1.736180206907e-03;
C[4708] = 3.893256160783e-01;
C[4709] = 6.905526764352e-03;
C[4710] = -1.388274670345e-03;
C[4711] = 4.785903951833e-02;
C[4712] = -1.122160751918e-02;
C[4713] = -5.729745068303e-05;
C[4714] = 2.737775099149e-03;
C[4715] = 1.610607720867e-04;
C[4716] = 6.452780296111e-05;
C[4717] = 3.195293088230e-02;
C[4718] = 7.041835432880e-04;
C[4719] = -7.568922861376e-05;
C[4720] = 1.283089575982e-01;
C[4721] = -1.052964250765e-03;
C[4722] = 8.046485833687e-04;
C[4723] = 5.038823681654e-01;
C[4724] = 1.053070808143e-03;
C[4725] = 7.094938451762e-05;
C[4726] = 3.701244783494e-02;
C[4727] = 1.838631623165e-04;
C[4728] = -6.657025497940e-04;
C[4729] = 3.216603012120e-03;
C[4730] = 5.521447949343e-02;
C[4731] = 1.114581679330e-03;
C[4732] = 6.737848458367e-03;
C[4733] = 3.971196661745e-01;
C[4734] = -7.689474488311e-03;
C[4735] = -1.124126627097e-02;
C[4736] = 1.911823525836e-02;
C[4737] = -3.622712924453e-05;
C[4738] = 1.664976848489e-04;
C[4739] = 2.684944628074e-03;
C[4740] = 2.638268549655e-05;
C[4741] = 6.949520642640e-04;
C[4742] = 3.267659257645e-02;
C[4743] = -4.836202464525e-04;
C[4744] = -1.054274834218e-03;
C[4745] = 1.259575782950e-01;
C[4746] = 7.268098217555e-03;
C[4747] = 1.143851745461e-03;
C[4748] = 5.259541145150e-01;
C[4749] = 4.943180811958e-04;
C[4750] = 1.882726854924e-04;
C[4751] = 3.869733765859e-02;
C[4752] = -7.009137728537e+00;
C[4753] = 4.739810657955e-02;
C[4754] = -1.101092031165e-02;
C[4755] = 2.145108495889e+00;
C[4756] = -3.035817389526e-03;
C[4757] = -1.470241649838e-02;
C[4758] = 7.979237065639e-01;
C[4759] = -9.464706208539e-03;
C[4760] = -4.463650739606e-03;
C[4761] = 7.739379415924e-02;
C[4762] = 1.051517313266e-02;
C[4763] = -7.636057173973e-04;
C[4764] = 3.777543097578e-01;
C[4765] = -4.498686049164e-03;
C[4766] = -2.692132245444e-03;
C[4767] = 5.141427041127e-02;
C[4768] = -7.613714475289e-04;
C[4769] = -3.564167510704e-04;
C[4770] = 7.833386137771e-02;
C[4771] = -1.273316717619e-03;
C[4772] = 7.367243725960e-04;
C[4773] = 2.420392220747e-03;
C[4774] = -5.692412979245e-05;
C[4775] = 2.070872345438e-05;
C[4776] = 4.502966349283e-02;
C[4777] = -7.192842895688e+00;
C[4778] = 5.251391132636e-02;
C[4779] = -2.087343016911e-03;
C[4780] = 2.117751235921e+00;
C[4781] = 3.149499036390e-02;
C[4782] = -8.958435062705e-03;
C[4783] = 8.651200010170e-01;
C[4784] = 5.144958748828e-02;
C[4785] = 1.039675097832e-02;
C[4786] = 2.646260152019e-02;
C[4787] = 4.774001374536e-03;
C[4788] = -4.449969405950e-03;
C[4789] = 3.974426663736e-01;
C[4790] = 1.148462637794e-02;
C[4791] = -7.346249734794e-04;
C[4792] = 5.636161959557e-02;
C[4793] = 3.543323184077e-03;
C[4794] = -1.127137366368e-03;
C[4795] = 8.644142352630e-02;
C[4796] = 6.464699871629e-03;
C[4797] = -5.227169390179e-05;
C[4798] = 2.736044792926e-03;
C[4799] = 2.226200045369e-04;
C[4800] = -1.090267487874e-02;
C[4801] = 5.112026404293e-02;
C[4802] = -7.049914756208e+00;
C[4803] = -1.575693000164e-02;
C[4804] = 3.731330151325e-02;
C[4805] = 2.102151913337e+00;
C[4806] = -4.086888564537e-03;
C[4807] = 5.039528060371e-02;
C[4808] = 8.343576057936e-01;
C[4809] = -7.581934457518e-04;
C[4810] = 4.704319010365e-03;
C[4811] = 6.545877055223e-02;
C[4812] = -2.766987344152e-03;
C[4813] = 1.189385746275e-02;
C[4814] = 3.780576956718e-01;
C[4815] = -3.266809098058e-04;
C[4816] = 3.435008115998e-03;
C[4817] = 5.410825345428e-02;
C[4818] = 9.235018645775e-04;
C[4819] = 5.921081999684e-03;
C[4820] = 8.714405171182e-02;
C[4821] = 2.838359214573e-05;
C[4822] = 1.997722475935e-04;
C[4823] = 2.735444451500e-03;
C[4824] = 2.133458252481e+00;
C[4825] = 5.388646118756e-03;
C[4826] = -1.109348798777e-02;
C[4827] = -6.229119164132e+00;
C[4828] = 1.407294272089e-02;
C[4829] = 3.407819215007e-03;
C[4830] = 2.155315836225e+00;
C[4831] = 1.169221537447e-02;
C[4832] = -7.357980940909e-03;
C[4833] = 3.771717975873e-01;
C[4834] = -4.076169949167e-03;
C[4835] = -2.533815243459e-03;
C[4836] = 1.195168232228e-01;
C[4837] = 5.625551283261e-03;
C[4838] = 2.813081524200e-03;
C[4839] = 3.889279033936e-01;
C[4840] = 3.338568772802e-04;
C[4841] = -1.077460032126e-03;
C[4842] = 1.001811590571e+00;
C[4843] = 1.598658676686e-03;
C[4844] = 1.399674271302e-02;
C[4845] = 5.926902830658e-02;
C[4846] = 5.884813893029e-05;
C[4847] = 7.421803797377e-04;
C[4848] = 7.239617901267e-03;
C[4849] = 2.104753689432e+00;
C[4850] = 2.538199731285e-02;
C[4851] = 1.348647675219e-02;
C[4852] = -6.293292211576e+00;
C[4853] = -4.011628366848e-02;
C[4854] = 1.033201662674e-02;
C[4855] = 2.111349927917e+00;
C[4856] = -1.568543493127e-02;
C[4857] = -3.984914284625e-03;
C[4858] = 3.967927890492e-01;
C[4859] = 1.129729225273e-02;
C[4860] = 5.596227984827e-03;
C[4861] = 8.935456748224e-02;
C[4862] = -1.941160594470e-02;
C[4863] = 2.594754118179e-04;
C[4864] = 3.904896634288e-01;
C[4865] = 6.115497164828e-03;
C[4866] = 1.880621376026e-03;
C[4867] = 1.007026400919e+00;
C[4868] = 2.807723814303e-02;
C[4869] = 8.092950043892e-05;
C[4870] = 5.987278830191e-02;
C[4871] = 1.950122893052e-03;
C[4872] = -9.222418068015e-03;
C[4873] = 2.141994009872e-02;
C[4874] = 2.106394441292e+00;
C[4875] = 3.637071857467e-03;
C[4876] = -4.181834319888e-02;
C[4877] = -6.272354500291e+00;
C[4878] = -9.229195610105e-03;
C[4879] = -1.277312418283e-02;
C[4880] = 2.130788699325e+00;
C[4881] = -2.418132323926e-03;
C[4882] = 1.098087386468e-02;
C[4883] = 3.782698220696e-01;
C[4884] = 2.824544156323e-03;
C[4885] = -1.949670892122e-02;
C[4886] = 1.025574273382e-01;
C[4887] = -1.189799037356e-03;
C[4888] = 6.326602489219e-03;
C[4889] = 3.984647653095e-01;
C[4890] = 1.453879750962e-02;
C[4891] = 2.746381026709e-02;
C[4892] = 1.085414514102e+00;
C[4893] = 7.813532672991e-04;
C[4894] = 1.900582653076e-03;
C[4895] = 6.454283708601e-02;
C[4896] = 8.155829162987e-01;
C[4897] = -1.438881587363e-02;
C[4898] = -1.114281108012e-02;
C[4899] = 2.137804180279e+00;
C[4900] = 1.301093692352e-02;
C[4901] = -1.887051228825e-03;
C[4902] = -7.102132477440e+00;
C[4903] = -2.830296950110e-03;
C[4904] = -3.017108214734e-02;
C[4905] = 5.229723089801e-02;
C[4906] = -1.006144014026e-03;
C[4907] = -6.794770355850e-04;
C[4908] = 3.880523205963e-01;
C[4909] = 3.934214266569e-04;
C[4910] = -8.226918182925e-04;
C[4911] = 4.950870729420e-02;
C[4912] = -3.141963652705e-04;
C[4913] = -3.994335917228e-03;
C[4914] = 4.071468192905e+00;
C[4915] = 7.629811349127e-03;
C[4916] = 4.350753172873e-02;
C[4917] = 5.077646444252e-01;
C[4918] = 1.050097178005e-03;
C[4919] = 5.501857210584e-03;
C[4920] = -1.454350062166e-02;
C[4921] = 8.626813185128e-01;
C[4922] = 4.499332689157e-02;
C[4923] = 1.320257935229e-02;
C[4924] = 2.109682102364e+00;
C[4925] = -1.428356093582e-02;
C[4926] = -2.310352653442e-03;
C[4927] = -7.110552634952e+00;
C[4928] = -3.189190375080e-02;
C[4929] = -1.015311168185e-03;
C[4930] = 5.623968547036e-02;
C[4931] = 3.164910430391e-03;
C[4932] = 4.093750761714e-04;
C[4933] = 3.904062721511e-01;
C[4934] = 6.251080651570e-03;
C[4935] = -2.881991504371e-04;
C[4936] = 4.428327011482e-02;
C[4937] = -1.197035733011e-02;
C[4938] = 7.112485921314e-03;
C[4939] = 4.047937338690e+00;
C[4940] = -1.755803229559e-03;
C[4941] = 1.019246175788e-03;
C[4942] = 5.061898831516e-01;
C[4943] = 2.420752234164e-03;
C[4944] = -1.148737698750e-02;
C[4945] = 4.557954892365e-02;
C[4946] = 8.498656994411e-01;
C[4947] = -1.541118078202e-03;
C[4948] = -1.653438152051e-02;
C[4949] = 2.114923423926e+00;
C[4950] = -2.904719220287e-02;
C[4951] = -3.228507878661e-02;
C[4952] = -7.223962747513e+00;
C[4953] = -7.076030634651e-04;
C[4954] = 3.249821255846e-03;
C[4955] = 5.488365813666e-02;
C[4956] = -7.866168889905e-04;
C[4957] = 6.073049835366e-03;
C[4958] = 3.976715015395e-01;
C[4959] = -3.938141420004e-03;
C[4960] = -1.199001608190e-02;
C[4961] = 1.843581414117e-02;
C[4962] = 4.255624923787e-02;
C[4963] = -3.670714881015e-04;
C[4964] = 4.177479386356e+00;
C[4965] = 5.444795857638e-03;
C[4966] = 2.502700686988e-03;
C[4967] = 5.264739689403e-01;
C[4968] = 0.000000000000e+00;
C[4969] = 0.000000000000e+00;
C[4970] = 0.000000000000e+00;
C[4971] = 0.000000000000e+00;
C[4972] = 0.000000000000e+00;
C[4973] = 0.000000000000e+00;
C[4974] = 0.000000000000e+00;
C[4975] = 0.000000000000e+00;
C[4976] = 0.000000000000e+00;
C[4977] = 0.000000000000e+00;
C[4978] = 0.000000000000e+00;
C[4979] = 0.000000000000e+00;
C[4980] = 0.000000000000e+00;
C[4981] = 0.000000000000e+00;
C[4982] = 0.000000000000e+00;
C[4983] = 0.000000000000e+00;
C[4984] = 0.000000000000e+00;
C[4985] = 0.000000000000e+00;
C[4986] = 1.000000000000e+00;
C[4987] = 0.000000000000e+00;
C[4988] = 0.000000000000e+00;
C[4989] = 2.000000000000e-01;
C[4990] = -0.000000000000e+00;
C[4991] = -0.000000000000e+00;
C[4992] = 0.000000000000e+00;
C[4993] = 0.000000000000e+00;
C[4994] = 0.000000000000e+00;
C[4995] = 0.000000000000e+00;
C[4996] = 0.000000000000e+00;
C[4997] = 0.000000000000e+00;
C[4998] = 0.000000000000e+00;
C[4999] = 0.000000000000e+00;
C[5000] = 0.000000000000e+00;
C[5001] = 0.000000000000e+00;
C[5002] = 0.000000000000e+00;
C[5003] = 0.000000000000e+00;
C[5004] = 0.000000000000e+00;
C[5005] = 0.000000000000e+00;
C[5006] = 0.000000000000e+00;
C[5007] = 0.000000000000e+00;
C[5008] = 0.000000000000e+00;
C[5009] = 0.000000000000e+00;
C[5010] = 0.000000000000e+00;
C[5011] = 1.000000000000e+00;
C[5012] = 0.000000000000e+00;
C[5013] = -0.000000000000e+00;
C[5014] = 2.000000000000e-01;
C[5015] = -0.000000000000e+00;
C[5016] = 0.000000000000e+00;
C[5017] = 0.000000000000e+00;
C[5018] = 0.000000000000e+00;
C[5019] = 0.000000000000e+00;
C[5020] = 0.000000000000e+00;
C[5021] = 0.000000000000e+00;
C[5022] = 0.000000000000e+00;
C[5023] = 0.000000000000e+00;
C[5024] = 0.000000000000e+00;
C[5025] = 0.000000000000e+00;
C[5026] = 0.000000000000e+00;
C[5027] = 0.000000000000e+00;
C[5028] = 0.000000000000e+00;
C[5029] = 0.000000000000e+00;
C[5030] = 0.000000000000e+00;
C[5031] = 0.000000000000e+00;
C[5032] = 0.000000000000e+00;
C[5033] = 0.000000000000e+00;
C[5034] = 0.000000000000e+00;
C[5035] = 0.000000000000e+00;
C[5036] = 1.000000000000e+00;
C[5037] = -0.000000000000e+00;
C[5038] = -0.000000000000e+00;
C[5039] = 2.000000000000e-01;
C[5040] = 8.714608899921e-02;
C[5041] = 1.078912270110e-03;
C[5042] = -1.055394044059e-03;
C[5043] = 3.714635135386e-01;
C[5044] = 3.830120217240e-04;
C[5045] = 4.866977872092e-04;
C[5046] = 5.011914469585e-02;
C[5047] = 7.229088767076e-04;
C[5048] = 1.788381357544e-05;
C[5049] = 1.320006927331e-01;
C[5050] = -3.731007331244e-04;
C[5051] = -2.751536751475e-05;
C[5052] = 2.993768138148e-02;
C[5053] = 3.588831883272e-04;
C[5054] = 1.159924747074e-04;
C[5055] = 2.404821659559e-03;
C[5056] = 5.572736092466e-05;
C[5057] = 6.117012053624e-06;
C[5058] = 2.330585598582e-03;
C[5059] = 1.065700147029e-04;
C[5060] = 3.574080078488e-05;
C[5061] = -2.279579836563e-05;
C[5062] = 4.521307766946e-06;
C[5063] = 2.157379297737e-06;
C[5064] = 7.832368291486e-04;
C[5065] = 3.268529902019e-02;
C[5066] = 8.963703471497e-03;
C[5067] = 6.208232725742e-04;
C[5068] = 4.015363901847e-01;
C[5069] = 1.164984603157e-02;
C[5070] = 7.010103281308e-04;
C[5071] = 5.673173387343e-02;
C[5072] = 3.247220388149e-03;
C[5073] = -3.928124291885e-04;
C[5074] = 1.273851677442e-01;
C[5075] = 6.396276784681e-04;
C[5076] = 3.756106373302e-04;
C[5077] = 3.298524115593e-02;
C[5078] = 9.551423505897e-04;
C[5079] = 5.384721418083e-05;
C[5080] = 2.765402365535e-03;
C[5081] = 1.641612433203e-04;
C[5082] = 1.042262222450e-04;
C[5083] = 2.786091954548e-03;
C[5084] = 2.215391653907e-04;
C[5085] = 4.433470902057e-06;
C[5086] = -1.700858918342e-05;
C[5087] = 9.809668053352e-08;
C[5088] = -1.047050865536e-03;
C[5089] = 8.885843773251e-03;
C[5090] = 6.484126843184e-02;
C[5091] = 5.961855365851e-04;
C[5092] = 1.215389379513e-02;
C[5093] = 3.696850934137e-01;
C[5094] = -9.390130597377e-06;
C[5095] = 3.087884210881e-03;
C[5096] = 5.225926908145e-02;
C[5097] = -2.695915561325e-05;
C[5098] = 6.344370319184e-04;
C[5099] = 1.300118753568e-01;
C[5100] = 1.235111053019e-04;
C[5101] = 9.819439701044e-04;
C[5102] = 2.985238941239e-02;
C[5103] = 3.723059864040e-06;
C[5104] = 1.536816800415e-04;
C[5105] = 2.521986732841e-03;
C[5106] = 3.013991236661e-05;
C[5107] = 1.932156435438e-04;
C[5108] = 2.592927366092e-03;
C[5109] = 1.906681713361e-06;
C[5110] = -6.276804893373e-07;
C[5111] = -1.986899509810e-05;
C[5112] = 3.718198672453e-01;
C[5113] = 6.206417748552e-04;
C[5114] = 9.289566289737e-05;
C[5115] = 1.270276229745e-01;
C[5116] = 6.053751715094e-04;
C[5117] = 1.499776122316e-03;
C[5118] = 3.889888578847e-01;
C[5119] = -1.911377210818e-03;
C[5120] = -1.982430875381e-03;
C[5121] = 2.996143829526e-02;
C[5122] = 3.755985374823e-04;
C[5123] = 8.995844705606e-05;
C[5124] = 1.335921189864e-01;
C[5125] = -2.890934286869e-04;
C[5126] = 3.873610182086e-05;
C[5127] = 3.158764920774e-02;
C[5128] = -1.409832544789e-04;
C[5129] = -1.504389470847e-04;
C[5130] = 5.932332685631e-02;
C[5131] = -1.774329419578e-04;
C[5132] = -4.509701495204e-05;
C[5133] = 2.578531310718e-03;
C[5134] = -6.790219784277e-06;
C[5135] = -1.574692625957e-06;
C[5136] = 8.511191442993e-04;
C[5137] = 4.008805204520e-01;
C[5138] = 1.221060583513e-02;
C[5139] = 3.851916295765e-04;
C[5140] = 8.841493348043e-02;
C[5141] = -1.947316387318e-02;
C[5142] = -1.893231071339e-03;
C[5143] = 3.881643446020e-01;
C[5144] = 6.154684284850e-03;
C[5145] = 3.900903298322e-04;
C[5146] = 3.294151650709e-02;
C[5147] = 9.857247727709e-04;
C[5148] = -3.037723314824e-04;
C[5149] = 1.299449867268e-01;
C[5150] = -1.751821670705e-03;
C[5151] = -1.390792524854e-04;
C[5152] = 3.174421693529e-02;
C[5153] = 7.036142477195e-04;
C[5154] = -1.629599679756e-04;
C[5155] = 6.015070417067e-02;
C[5156] = 2.434223158056e-03;
C[5157] = -6.206074916323e-06;
C[5158] = 2.631599460209e-03;
C[5159] = 1.217387265499e-04;
C[5160] = 1.180238087344e-04;
C[5161] = 1.176621813574e-02;
C[5162] = 3.705951901116e-01;
C[5163] = 1.393545877831e-03;
C[5164] = -1.953832001793e-02;
C[5165] = 1.091353180807e-01;
C[5166] = -1.969198304243e-03;
C[5167] = 6.356821511883e-03;
C[5168] = 4.002498434643e-01;
C[5169] = 9.141420947573e-05;
C[5170] = 9.629004908675e-04;
C[5171] = 2.991306252558e-02;
C[5172] = 3.165408552182e-05;
C[5173] = -1.756165413688e-03;
C[5174] = 1.319818241701e-01;
C[5175] = -1.483529841512e-04;
C[5176] = 7.149999670555e-04;
C[5177] = 3.287283172814e-02;
C[5178] = -1.349720853929e-05;
C[5179] = 2.420740003637e-03;
C[5180] = 6.484880329024e-02;
C[5181] = -2.670460512123e-07;
C[5182] = 1.205471573380e-04;
C[5183] = 2.846690382239e-03;
C[5184] = 5.020828617793e-02;
C[5185] = -1.100303659596e-04;
C[5186] = -2.026072245474e-04;
C[5187] = 3.882940286666e-01;
C[5188] = -9.088447947889e-04;
C[5189] = -1.508344721310e-03;
C[5190] = 4.773283792649e-02;
C[5191] = 5.315651034085e-04;
C[5192] = 1.984993039359e-04;
C[5193] = 2.410764425030e-03;
C[5194] = -2.221654252004e-07;
C[5195] = -9.158079732631e-06;
C[5196] = 3.154132725987e-02;
C[5197] = -7.345350071538e-05;
C[5198] = -1.176294119556e-04;
C[5199] = 1.284938611396e-01;
C[5200] = 6.605700061564e-05;
C[5201] = 9.051937820834e-05;
C[5202] = 5.115503585555e-01;
C[5203] = 4.971686941494e-04;
C[5204] = 1.498821035467e-03;
C[5205] = 3.757446712500e-02;
C[5206] = 8.069066605668e-06;
C[5207] = 3.423929852109e-05;
C[5208] = -8.582910608151e-05;
C[5209] = 5.633677170590e-02;
C[5210] = 3.368061833316e-03;
C[5211] = -9.227794033113e-04;
C[5212] = 3.877134849912e-01;
C[5213] = 6.140666233568e-03;
C[5214] = 5.346860068172e-04;
C[5215] = 4.080661311878e-02;
C[5216] = -1.364828538567e-02;
C[5217] = 1.811495405389e-06;
C[5218] = 2.739071554366e-03;
C[5219] = 1.723601882038e-04;
C[5220] = -7.507673397847e-05;
C[5221] = 3.171415962790e-02;
C[5222] = 7.005896151679e-04;
C[5223] = 6.626506084289e-05;
C[5224] = 1.278661759566e-01;
C[5225] = -1.250578809313e-03;
C[5226] = 4.823256774664e-04;
C[5227] = 5.123759109774e-01;
C[5228] = 4.095459931570e-03;
C[5229] = 7.383962921598e-06;
C[5230] = 3.769284487412e-02;
C[5231] = 3.769588102103e-04;
C[5232] = -1.945090242861e-04;
C[5233] = 3.533482982433e-03;
C[5234] = 5.314291999091e-02;
C[5235] = -1.493368044114e-03;
C[5236] = 5.902255476742e-03;
C[5237] = 3.996097624892e-01;
C[5238] = 2.067495306249e-04;
C[5239] = -1.365233585247e-02;
C[5240] = 1.841333968589e-02;
C[5241] = -8.042510470477e-06;
C[5242] = 1.832454162726e-04;
C[5243] = 2.580896793472e-03;
C[5244] = -1.178347583335e-04;
C[5245] = 6.867856605123e-04;
C[5246] = 3.283015966313e-02;
C[5247] = 9.106939332094e-05;
C[5248] = -1.250848840433e-03;
C[5249] = 1.259201792104e-01;
C[5250] = 1.464386772669e-03;
C[5251] = 4.140740146829e-03;
C[5252] = 5.263258442821e-01;
C[5253] = 3.261999774055e-05;
C[5254] = 3.790431375538e-04;
C[5255] = 3.868927624681e-02;
C[5256] = -6.955308813618e+00;
C[5257] = -5.731417380117e-03;
C[5258] = 2.378726515011e-03;
C[5259] = 2.120061894732e+00;
C[5260] = 5.697713583893e-03;
C[5261] = 9.920649817169e-03;
C[5262] = 7.776944480177e-01;
C[5263] = 1.584121414900e-02;
C[5264] = 1.840755346590e-03;
C[5265] = 8.918051142185e-02;
C[5266] = -9.021579904927e-03;
C[5267] = 6.664495209386e-04;
C[5268] = 3.687463583087e-01;
C[5269] = 7.330535024605e-03;
C[5270] = 2.371637035258e-03;
C[5271] = 4.985329885719e-02;
C[5272] = 1.178139957456e-03;
C[5273] = 1.563758379825e-04;
C[5274] = 7.617306061138e-02;
C[5275] = 2.194458455097e-03;
C[5276] = 7.020036904438e-04;
C[5277] = 2.319764529059e-03;
C[5278] = 8.799035564523e-05;
C[5279] = 3.248408748359e-05;
C[5280] = -1.164492619934e-02;
C[5281] = -7.125089361090e+00;
C[5282] = 8.409456788123e-02;
C[5283] = 1.042515369887e-02;
C[5284] = 2.158246524672e+00;
C[5285] = 3.669426593905e-02;
C[5286] = 1.537541492347e-02;
C[5287] = 8.719142837907e-01;
C[5288] = 4.569089082880e-02;
C[5289] = -9.317255345887e-03;
C[5290] = 4.001209845823e-02;
C[5291] = 9.409201390531e-03;
C[5292] = 7.586339221382e-03;
C[5293] = 4.018329113345e-01;
C[5294] = 9.938076077545e-03;
C[5295] = 1.130614719020e-03;
C[5296] = 5.657227830080e-02;
C[5297] = 3.177484731607e-03;
C[5298] = 2.119790857696e-03;
C[5299] = 8.793478319670e-02;
C[5300] = 6.635878993452e-03;
C[5301] = 8.413446028861e-05;
C[5302] = 2.755240430672e-03;
C[5303] = 2.329656266351e-04;
C[5304] = 2.545590085462e-03;
C[5305] = 8.253737391630e-02;
C[5306] = -7.088416862202e+00;
C[5307] = 1.209501880036e-02;
C[5308] = 4.405722261831e-02;
C[5309] = 2.040706460507e+00;
C[5310] = 1.322713161098e-03;
C[5311] = 4.381653952091e-02;
C[5312] = 8.020744746861e-01;
C[5313] = 6.747926994611e-04;
C[5314] = 9.331341692285e-03;
C[5315] = 6.283511455815e-02;
C[5316] = 2.512607864914e-03;
C[5317] = 1.044103779484e-02;
C[5318] = 3.637131577671e-01;
C[5319] = 1.076265073014e-04;
C[5320] = 3.000244565488e-03;
C[5321] = 5.194664752426e-02;
C[5322] = 5.848405956693e-04;
C[5323] = 5.960138239088e-03;
C[5324] = 8.375962762402e-02;
C[5325] = 2.696576231689e-05;
C[5326] = 2.039003267604e-04;
C[5327] = 2.624859065246e-03;
C[5328] = 2.127188968865e+00;
C[5329] = 1.042152374449e-02;
C[5330] = 2.029221326600e-03;
C[5331] = -6.195506045957e+00;
C[5332] = -5.770753141022e-03;
C[5333] = 2.531416359471e-03;
C[5334] = 2.167169738849e+00;
C[5335] = -2.114044780413e-02;
C[5336] = -1.077556110523e-02;
C[5337] = 3.691027120153e-01;
C[5338] = 7.586157723663e-03;
C[5339] = 2.009317991226e-03;
C[5340] = 1.317810203562e-01;
C[5341] = -6.845355135293e-03;
C[5342] = -5.917030767629e-04;
C[5343] = 3.870104991207e-01;
C[5344] = -1.988118463200e-03;
C[5345] = -1.581697862905e-03;
C[5346] = 1.003884899376e+00;
C[5347] = -3.218788107633e-03;
C[5348] = -9.370350094496e-04;
C[5349] = 5.936219804003e-02;
C[5350] = -1.779491296182e-04;
C[5351] = -1.364828044604e-04;
C[5352] = 1.505985603540e-02;
C[5353] = 2.145129130018e+00;
C[5354] = 4.519146341826e-02;
C[5355] = -1.017442397968e-02;
C[5356] = -6.294489884974e+00;
C[5357] = -4.125727642441e-02;
C[5358] = -2.068123261495e-02;
C[5359] = 2.089344710748e+00;
C[5360] = -1.722946476520e-02;
C[5361] = 7.798642147180e-03;
C[5362] = 4.011770416018e-01;
C[5363] = 1.049774983484e-02;
C[5364] = -7.065538677226e-03;
C[5365] = 9.085214205729e-02;
C[5366] = -1.833547335080e-02;
C[5367] = -1.938974060516e-03;
C[5368] = 3.858593549500e-01;
C[5369] = 6.684826749157e-03;
C[5370] = -2.986019856125e-03;
C[5371] = 1.012774679964e+00;
C[5372] = 3.730560634535e-02;
C[5373] = -1.679282947051e-04;
C[5374] = 6.023498064449e-02;
C[5375] = 2.464051540643e-03;
C[5376] = 2.547170247673e-03;
C[5377] = 3.902170802240e-02;
C[5378] = 2.058908394464e+00;
C[5379] = 4.068114697579e-04;
C[5380] = -4.256039931935e-02;
C[5381] = -6.243867297602e+00;
C[5382] = -1.029644177087e-02;
C[5383] = -1.459076956472e-02;
C[5384] = 2.159496244409e+00;
C[5385] = 2.002963056783e-03;
C[5386] = 1.005444818171e-02;
C[5387] = 3.646232544649e-01;
C[5388] = -6.979333212486e-04;
C[5389] = -1.840062949555e-02;
C[5390] = 1.142433454814e-01;
C[5391] = -1.530672969194e-03;
C[5392] = 6.898408098808e-03;
C[5393] = 4.009598470674e-01;
C[5394] = -4.726942303214e-04;
C[5395] = 3.742914868662e-02;
C[5396] = 1.092030688105e+00;
C[5397] = -1.167429958752e-04;
C[5398] = 2.463295343230e-03;
C[5399] = 6.481638051667e-02;
C[5400] = 7.794772776591e-01;
C[5401] = -8.453989583378e-04;
C[5402] = -2.541628717903e-03;
C[5403] = 2.153273154488e+00;
C[5404] = -9.935070839521e-04;
C[5405] = -1.079370112205e-03;
C[5406] = -7.121109936102e+00;
C[5407] = -1.368806667015e-04;
C[5408] = -3.460588916128e-04;
C[5409] = 4.994244033926e-02;
C[5410] = 3.195740249299e-04;
C[5411] = -8.559058664869e-05;
C[5412] = 3.863156699027e-01;
C[5413] = -9.545877839656e-04;
C[5414] = -1.069819386261e-03;
C[5415] = 4.740372245952e-02;
C[5416] = 7.814898322604e-04;
C[5417] = 1.591567743020e-03;
C[5418] = 4.116313632953e+00;
C[5419] = 1.297317702794e-03;
C[5420] = 3.629005893005e-03;
C[5421] = 5.141419269084e-01;
C[5422] = -1.992467597798e-04;
C[5423] = -4.651403257927e-04;
C[5424] = -3.335455067809e-04;
C[5425] = 8.640150404401e-01;
C[5426] = 4.942009196959e-02;
C[5427] = -1.368491653997e-03;
C[5428] = 2.080327518532e+00;
C[5429] = -1.891387513102e-02;
C[5430] = -7.446259852698e-05;
C[5431] = -7.147459104739e+00;
C[5432] = -4.387068166827e-02;
C[5433] = 3.694019746669e-04;
C[5434] = 5.617731613327e-02;
C[5435] = 3.280422187922e-03;
C[5436] = -9.995206556936e-04;
C[5437] = 3.854084953392e-01;
C[5438] = 6.682252820493e-03;
C[5439] = 7.846107356692e-04;
C[5440] = 4.001963236352e-02;
C[5441] = -1.428935306741e-02;
C[5442] = 1.084251329478e-03;
C[5443] = 4.115786036693e+00;
C[5444] = 1.267692220683e-02;
C[5445] = -2.082167585553e-04;
C[5446] = 5.156107878455e-01;
C[5447] = 4.303235249719e-03;
C[5448] = -2.407101410641e-03;
C[5449] = 5.141614271449e-02;
C[5450] = 8.197474928753e-01;
C[5451] = -9.943044798987e-04;
C[5452] = -2.227804092735e-02;
C[5453] = 2.146694624908e+00;
C[5454] = -1.810543578330e-04;
C[5455] = -4.395169100421e-02;
C[5456] = -7.221022033825e+00;
C[5457] = -5.601699987899e-05;
C[5458] = 3.463747325891e-03;
C[5459] = 5.283029843372e-02;
C[5460] = -1.092635031639e-03;
C[5461] = 6.432397941050e-03;
C[5462] = 4.003197660924e-01;
C[5463] = 1.599817969709e-03;
C[5464] = -1.429340353421e-02;
C[5465] = 1.858877560023e-02;
C[5466] = 3.123268735472e-03;
C[5467] = 1.335672238291e-02;
C[5468] = 4.174818804365e+00;
C[5469] = -4.873932075895e-04;
C[5470] = 4.337854823916e-03;
C[5471] = 5.258192361464e-01;
C[5472] = 0.000000000000e+00;
C[5473] = 0.000000000000e+00;
C[5474] = 0.000000000000e+00;
C[5475] = 0.000000000000e+00;
C[5476] = 0.000000000000e+00;
C[5477] = 0.000000000000e+00;
C[5478] = 0.000000000000e+00;
C[5479] = 0.000000000000e+00;
C[5480] = 0.000000000000e+00;
C[5481] = 0.000000000000e+00;
C[5482] = 0.000000000000e+00;
C[5483] = 0.000000000000e+00;
C[5484] = 0.000000000000e+00;
C[5485] = 0.000000000000e+00;
C[5486] = 0.000000000000e+00;
C[5487] = 0.000000000000e+00;
C[5488] = 0.000000000000e+00;
C[5489] = 0.000000000000e+00;
C[5490] = 1.000000000000e+00;
C[5491] = 0.000000000000e+00;
C[5492] = 0.000000000000e+00;
C[5493] = 2.000000000000e-01;
C[5494] = -0.000000000000e+00;
C[5495] = -0.000000000000e+00;
C[5496] = 0.000000000000e+00;
C[5497] = 0.000000000000e+00;
C[5498] = 0.000000000000e+00;
C[5499] = 0.000000000000e+00;
C[5500] = 0.000000000000e+00;
C[5501] = 0.000000000000e+00;
C[5502] = 0.000000000000e+00;
C[5503] = 0.000000000000e+00;
C[5504] = 0.000000000000e+00;
C[5505] = 0.000000000000e+00;
C[5506] = 0.000000000000e+00;
C[5507] = 0.000000000000e+00;
C[5508] = 0.000000000000e+00;
C[5509] = 0.000000000000e+00;
C[5510] = 0.000000000000e+00;
C[5511] = 0.000000000000e+00;
C[5512] = 0.000000000000e+00;
C[5513] = 0.000000000000e+00;
C[5514] = 0.000000000000e+00;
C[5515] = 1.000000000000e+00;
C[5516] = 0.000000000000e+00;
C[5517] = -0.000000000000e+00;
C[5518] = 2.000000000000e-01;
C[5519] = -0.000000000000e+00;
C[5520] = 0.000000000000e+00;
C[5521] = 0.000000000000e+00;
C[5522] = 0.000000000000e+00;
C[5523] = 0.000000000000e+00;
C[5524] = 0.000000000000e+00;
C[5525] = 0.000000000000e+00;
C[5526] = 0.000000000000e+00;
C[5527] = 0.000000000000e+00;
C[5528] = 0.000000000000e+00;
C[5529] = 0.000000000000e+00;
C[5530] = 0.000000000000e+00;
C[5531] = 0.000000000000e+00;
C[5532] = 0.000000000000e+00;
C[5533] = 0.000000000000e+00;
C[5534] = 0.000000000000e+00;
C[5535] = 0.000000000000e+00;
C[5536] = 0.000000000000e+00;
C[5537] = 0.000000000000e+00;
C[5538] = 0.000000000000e+00;
C[5539] = 0.000000000000e+00;
C[5540] = 1.000000000000e+00;
C[5541] = -0.000000000000e+00;
C[5542] = -0.000000000000e+00;
C[5543] = 2.000000000000e-01;
C[5544] = 8.720100813715e-02;
C[5545] = -1.622522610535e-02;
C[5546] = 4.567399002715e-03;
C[5547] = 3.651494415257e-01;
C[5548] = 1.274624516075e-02;
C[5549] = 3.431750339772e-03;
C[5550] = 4.872801097340e-02;
C[5551] = 2.835142631627e-03;
C[5552] = 6.400527203112e-04;
C[5553] = 1.319280301409e-01;
C[5554] = -1.721423009796e-03;
C[5555] = 4.216558918984e-04;
C[5556] = 2.939386037645e-02;
C[5557] = 1.449523573229e-03;
C[5558] = 2.819271279013e-04;
C[5559] = 2.338108712897e-03;
C[5560] = 1.635197076371e-04;
C[5561] = 3.163981716428e-05;
C[5562] = 2.300003503160e-03;
C[5563] = 2.199517364370e-04;
C[5564] = 2.794006190976e-05;
C[5565] = -2.140242971154e-05;
C[5566] = 4.866520438217e-06;
C[5567] = 3.084046611491e-07;
C[5568] = -1.641822592014e-02;
C[5569] = 4.652993744234e-02;
C[5570] = 5.211277216668e-03;
C[5571] = 1.298297443972e-02;
C[5572] = 4.006807750755e-01;
C[5573] = 7.503633710682e-03;
C[5574] = 2.806200038787e-03;
C[5575] = 5.549674907230e-02;
C[5576] = 2.724694535682e-03;
C[5577] = -1.734289664115e-03;
C[5578] = 1.283867875103e-01;
C[5579] = 2.710155813788e-04;
C[5580] = 1.464891907537e-03;
C[5581] = 3.270284426471e-02;
C[5582] = 6.121406637953e-04;
C[5583] = 1.612153987029e-04;
C[5584] = 2.694938529603e-03;
C[5585] = 1.372748511118e-04;
C[5586] = 2.125178347750e-04;
C[5587] = 2.740592913481e-03;
C[5588] = 1.888082682581e-04;
C[5589] = 4.669474188352e-06;
C[5590] = -1.745578861621e-05;
C[5591] = 7.162032826712e-08;
C[5592] = 4.580170331930e-03;
C[5593] = 5.175223035564e-03;
C[5594] = 6.235725359032e-02;
C[5595] = 3.658707454297e-03;
C[5596] = 8.007498251078e-03;
C[5597] = 3.565487920467e-01;
C[5598] = 5.511353826055e-04;
C[5599] = 2.504564386027e-03;
C[5600] = 5.014281328965e-02;
C[5601] = 4.225073138461e-04;
C[5602] = 2.686119693052e-04;
C[5603] = 1.302475032291e-01;
C[5604] = 2.946196578668e-04;
C[5605] = 6.386606414477e-04;
C[5606] = 2.865558474980e-02;
C[5607] = 2.589708174041e-05;
C[5608] = 1.241675100983e-04;
C[5609] = 2.410532014069e-03;
C[5610] = 1.457091020229e-05;
C[5611] = 1.632619908754e-04;
C[5612] = 2.449381419410e-03;
C[5613] = -2.547424282437e-08;
C[5614] = -4.394672947058e-07;
C[5615] = -2.121555942274e-05;
C[5616] = 3.653206951178e-01;
C[5617] = 1.271640618601e-02;
C[5618] = 3.400496283482e-03;
C[5619] = 1.350962995929e-01;
C[5620] = -1.346512294458e-02;
C[5621] = -2.696995129825e-03;
C[5622] = 3.867445739590e-01;
C[5623] = -2.525424034694e-03;
C[5624] = 4.666530359521e-05;
C[5625] = 2.940527728260e-02;
C[5626] = 1.447120690623e-03;
C[5627] = 2.774055798125e-04;
C[5628] = 1.341071826406e-01;
C[5629] = -1.484199973058e-03;
C[5630] = -2.565960495180e-04;
C[5631] = 3.158784418815e-02;
C[5632] = -1.395618803222e-04;
C[5633] = 1.180679801083e-05;
C[5634] = 6.030085084795e-02;
C[5635] = -3.110660325725e-04;
C[5636] = -2.454376449552e-04;
C[5637] = 2.634225989783e-03;
C[5638] = -8.668467006922e-06;
C[5639] = -1.016934122039e-05;
C[5640] = 1.282169665748e-02;
C[5641] = 4.003384101219e-01;
C[5642] = 8.479865055697e-03;
C[5643] = -1.363628375467e-02;
C[5644] = 9.330547250439e-02;
C[5645] = -1.707187381396e-02;
C[5646] = -2.482210214569e-03;
C[5647] = 3.854150886319e-01;
C[5648] = 7.123727457338e-03;
C[5649] = 1.454553673011e-03;
C[5650] = 3.268001993447e-02;
C[5651] = 6.701517617556e-04;
C[5652] = -1.495610693731e-03;
C[5653] = 1.302941591474e-01;
C[5654] = -1.464589518955e-03;
C[5655] = -1.366254769950e-04;
C[5656] = 3.172530701555e-02;
C[5657] = 7.208549027477e-04;
C[5658] = -3.143655464206e-04;
C[5659] = 6.123772786045e-02;
C[5660] = 2.393061101357e-03;
C[5661] = -9.006564340042e-06;
C[5662] = 2.693588806975e-03;
C[5663] = 1.152737654150e-04;
C[5664] = 3.276904223480e-03;
C[5665] = 8.082974308133e-03;
C[5666] = 3.568533764801e-01;
C[5667] = -2.782688576976e-03;
C[5668] = -1.719801198965e-02;
C[5669] = 1.203110966638e-01;
C[5670] = 1.593367868097e-04;
C[5671] = 7.379429925061e-03;
C[5672] = 4.031515499128e-01;
C[5673] = 2.716040534818e-04;
C[5674] = 6.507633702921e-04;
C[5675] = 2.867589037870e-02;
C[5676] = -2.623089459947e-04;
C[5677] = -1.472998730668e-03;
C[5678] = 1.330900713984e-01;
C[5679] = 1.845827216589e-05;
C[5680] = 7.354069228925e-04;
C[5681] = 3.303295580536e-02;
C[5682] = -2.515488650866e-04;
C[5683] = 2.409085189492e-03;
C[5684] = 6.489508674809e-02;
C[5685] = -1.078510038206e-05;
C[5686] = 1.151982990566e-04;
C[5687] = 2.842848215345e-03;
C[5688] = 4.941923004187e-02;
C[5689] = 2.122361797702e-03;
C[5690] = 5.248807962573e-04;
C[5691] = 3.864212513734e-01;
C[5692] = -1.776775549223e-03;
C[5693] = 2.183406023039e-04;
C[5694] = 4.560307329725e-02;
C[5695] = 5.798937331834e-04;
C[5696] = 1.526956452660e-03;
C[5697] = 2.384189984129e-03;
C[5698] = 1.156261826305e-04;
C[5699] = 2.414677598386e-05;
C[5700] = 3.156628934911e-02;
C[5701] = -8.959649930526e-05;
C[5702] = 2.239185986550e-05;
C[5703] = 1.282067415191e-01;
C[5704] = 6.344383472991e-05;
C[5705] = 1.419351381453e-04;
C[5706] = 5.163224603170e-01;
C[5707] = -1.082792651939e-03;
C[5708] = -2.253570851269e-03;
C[5709] = 3.786244240351e-02;
C[5710] = -9.137134613279e-05;
C[5711] = -1.892602012406e-04;
C[5712] = 2.160837868363e-03;
C[5713] = 5.621418312361e-02;
C[5714] = 2.723606087528e-03;
C[5715] = -1.831213728825e-03;
C[5716] = 3.850406330908e-01;
C[5717] = 7.127507880305e-03;
C[5718] = 5.813799828779e-04;
C[5719] = 3.831661562961e-02;
C[5720] = -1.403587266488e-02;
C[5721] = 1.185660567528e-04;
C[5722] = 2.742767466357e-03;
C[5723] = 1.387702901983e-04;
C[5724] = -9.328119326426e-05;
C[5725] = 3.170034331281e-02;
C[5726] = 7.186121199088e-04;
C[5727] = 6.354291804287e-05;
C[5728] = 1.275757701995e-01;
C[5729] = -1.223681607199e-03;
C[5730] = -1.073187026842e-03;
C[5731] = 5.177484376223e-01;
C[5732] = 4.141286772835e-03;
C[5733] = -9.092689830404e-05;
C[5734] = 3.799505947650e-02;
C[5735] = 3.675591741818e-04;
C[5736] = 5.940002761224e-04;
C[5737] = 2.925814194251e-03;
C[5738] = 5.108679923404e-02;
C[5739] = 1.159388606707e-04;
C[5740] = 6.909652823070e-03;
C[5741] = 4.022607512972e-01;
C[5742] = 1.533450155926e-03;
C[5743] = -1.405142119325e-02;
C[5744] = 1.896592069859e-02;
C[5745] = 2.856965421836e-05;
C[5746] = 1.506828283497e-04;
C[5747] = 2.473464410361e-03;
C[5748] = 1.642503514919e-05;
C[5749] = 7.065832604631e-04;
C[5750] = 3.297356923099e-02;
C[5751] = 1.423680516964e-04;
C[5752] = -1.224718175757e-03;
C[5753] = 1.259577803851e-01;
C[5754] = -2.236877602684e-03;
C[5755] = 4.150213552647e-03;
C[5756] = 5.251811011130e-01;
C[5757] = -1.885229517808e-04;
C[5758] = 3.680427094624e-04;
C[5759] = 3.861383654083e-02;
C[5760] = -6.978106354072e+00;
C[5761] = -5.011940232160e-02;
C[5762] = 3.968878120838e-02;
C[5763] = 2.088729597835e+00;
C[5764] = 3.012882905579e-02;
C[5765] = 2.131551350192e-02;
C[5766] = 7.566804801867e-01;
C[5767] = 4.451032924369e-02;
C[5768] = 1.005958345412e-02;
C[5769] = 8.615250440946e-02;
C[5770] = -2.160886085027e-02;
C[5771] = 6.515894692218e-03;
C[5772] = 3.620334903898e-01;
C[5773] = 1.827796540646e-02;
C[5774] = 3.576211079617e-03;
C[5775] = 4.870993498648e-02;
C[5776] = 3.177838549607e-03;
C[5777] = 6.480317051860e-04;
C[5778] = 7.572249351471e-02;
C[5779] = 5.347201934553e-03;
C[5780] = 7.703765902556e-04;
C[5781] = 2.335982899634e-03;
C[5782] = 2.020470963519e-04;
C[5783] = 2.565644016138e-05;
C[5784] = -5.397939861734e-02;
C[5785] = -7.071533499486e+00;
C[5786] = 4.546243181125e-02;
C[5787] = 3.454798784058e-02;
C[5788] = 2.163972109934e+00;
C[5789] = 1.027412580478e-02;
C[5790] = 4.409030266547e-02;
C[5791] = 8.524854795754e-01;
C[5792] = 3.734861292071e-02;
C[5793] = -2.180186066506e-02;
C[5794] = 4.977024613108e-02;
C[5795] = 3.227845137956e-03;
C[5796] = 1.852920681582e-02;
C[5797] = 3.975267958964e-01;
C[5798] = 5.617234818027e-03;
C[5799] = 3.129515365940e-03;
C[5800] = 5.547353150856e-02;
C[5801] = 2.577749900790e-03;
C[5802] = 5.149520774683e-03;
C[5803] = 8.777534222192e-02;
C[5804] = 5.527447509870e-03;
C[5805] = 1.931204213693e-04;
C[5806] = 2.778063656694e-03;
C[5807] = 1.924054810013e-04;
C[5808] = 3.994420779269e-02;
C[5809] = 4.474134818917e-02;
C[5810] = -7.121411187136e+00;
C[5811] = 2.493813442885e-02;
C[5812] = 1.768975974560e-02;
C[5813] = 1.990983053573e+00;
C[5814] = 8.852446923341e-03;
C[5815] = 3.462652408892e-02;
C[5816] = 7.698226602177e-01;
C[5817] = 6.528666021433e-03;
C[5818] = 3.191790956853e-03;
C[5819] = 6.783206518195e-02;
C[5820] = 3.810157608024e-03;
C[5821] = 6.102487495788e-03;
C[5822] = 3.493966346654e-01;
C[5823] = 5.482178516426e-04;
C[5824] = 2.360234655700e-03;
C[5825] = 4.970399965437e-02;
C[5826] = 4.635656341863e-04;
C[5827] = 4.920827619010e-03;
C[5828] = 7.965617163646e-02;
C[5829] = 1.303941383473e-05;
C[5830] = 1.677222346137e-04;
C[5831] = 2.477400945219e-03;
C[5832] = 2.092154669678e+00;
C[5833] = 2.921662276652e-02;
C[5834] = 1.977391101255e-02;
C[5835] = -6.177155018821e+00;
C[5836] = -4.039945799681e-02;
C[5837] = -3.551151515521e-03;
C[5838] = 2.168281697887e+00;
C[5839] = -4.233049611998e-02;
C[5840] = -2.700145341556e-03;
C[5841] = 3.622047439819e-01;
C[5842] = 1.826263856212e-02;
C[5843] = 3.551946437209e-03;
C[5844] = 1.352370186301e-01;
C[5845] = -1.896922731974e-02;
C[5846] = -2.700233777287e-03;
C[5847] = 3.894566488012e-01;
C[5848] = -2.856443211125e-03;
C[5849] = 3.255215287614e-05;
C[5850] = 1.023222713320e+00;
C[5851] = -7.023538485440e-03;
C[5852] = -4.546269731653e-03;
C[5853] = 6.046659558347e-02;
C[5854] = -3.089149691500e-04;
C[5855] = -2.746606473956e-04;
C[5856] = 3.163785899042e-02;
C[5857] = 2.157124810862e+00;
C[5858] = 2.713709583798e-02;
C[5859] = -4.382267419864e-02;
C[5860] = -6.283566191714e+00;
C[5861] = -3.654594490112e-02;
C[5862] = -4.165222977932e-02;
C[5863] = 2.088515060296e+00;
C[5864] = -1.014012202596e-02;
C[5865] = 1.835341690319e-02;
C[5866] = 3.971844309428e-01;
C[5867] = 6.574854300407e-03;
C[5868] = -1.914038812983e-02;
C[5869] = 9.367329766221e-02;
C[5870] = -1.469835895044e-02;
C[5871] = -2.791969755504e-03;
C[5872] = 3.882005705184e-01;
C[5873] = 6.546004251128e-03;
C[5874] = -6.989177218871e-03;
C[5875] = 1.033523653261e+00;
C[5876] = 3.611903532986e-02;
C[5877] = -3.116441171364e-04;
C[5878] = 6.139074842174e-02;
C[5879] = 2.341324090411e-03;
C[5880] = 1.821859117607e-02;
C[5881] = 2.186093775380e-02;
C[5882] = 1.997074742242e+00;
C[5883] = -5.265020458546e-03;
C[5884] = -3.906870841498e-02;
C[5885] = -6.211318008423e+00;
C[5886] = -1.034865087716e-03;
C[5887] = -6.597123801969e-03;
C[5888] = 2.194983898405e+00;
C[5889] = 3.421364963325e-03;
C[5890] = 6.196575415478e-03;
C[5891] = 3.497012190988e-01;
C[5892] = -2.785927224438e-03;
C[5893] = -1.482449712614e-02;
C[5894] = 1.274024640765e-01;
C[5895] = 1.605055891585e-04;
C[5896] = 6.804926264407e-03;
C[5897] = 4.035428932163e-01;
C[5898] = -4.521367413392e-03;
C[5899] = 3.658589123895e-02;
C[5900] = 1.094029898955e+00;
C[5901] = -2.821545406409e-04;
C[5902] = 2.355234424995e-03;
C[5903] = 6.485164195938e-02;
C[5904] = 7.705048615561e-01;
C[5905] = 3.041353784376e-02;
C[5906] = 8.327355196377e-03;
C[5907] = 2.161815246176e+00;
C[5908] = -2.754353647239e-02;
C[5909] = 1.452112221686e-04;
C[5910] = -7.139408985586e+00;
C[5911] = 9.532368793840e-04;
C[5912] = 9.714602369685e-03;
C[5913] = 4.940115405495e-02;
C[5914] = 2.445677124855e-03;
C[5915] = 5.219632652945e-04;
C[5916] = 3.891333262157e-01;
C[5917] = -2.086535090157e-03;
C[5918] = 2.195094046527e-04;
C[5919] = 4.257115804574e-02;
C[5920] = 7.530275032718e-04;
C[5921] = 1.961551035981e-03;
C[5922] = 4.134865570912e+00;
C[5923] = -8.832341385619e-03;
C[5924] = -1.760012410855e-02;
C[5925] = 5.166984285568e-01;
C[5926] = -1.321021074478e-03;
C[5927] = -2.688222419546e-03;
C[5928] = 3.102423397842e-02;
C[5929] = 8.668341606016e-01;
C[5930] = 3.900735811894e-02;
C[5931] = -2.844629000261e-02;
C[5932] = 2.081025949473e+00;
C[5933] = -1.163556469708e-02;
C[5934] = 9.829618732742e-04;
C[5935] = -7.166780754859e+00;
C[5936] = -4.148609627262e-02;
C[5937] = 2.503533786343e-03;
C[5938] = 5.619096555987e-02;
C[5939] = 2.579276357201e-03;
C[5940] = -2.162232905257e-03;
C[5941] = 3.878261149773e-01;
C[5942] = 6.553004219651e-03;
C[5943] = 7.545137529663e-04;
C[5944] = 3.541340470356e-02;
C[5945] = -1.320868207340e-02;
C[5946] = -8.680943057311e-03;
C[5947] = 4.134712293526e+00;
C[5948] = 1.314555484692e-02;
C[5949] = -1.310300663995e-03;
C[5950] = 5.179263185363e-01;
C[5951] = 4.054514017566e-03;
C[5952] = 9.138534570341e-03;
C[5953] = 4.137100609209e-02;
C[5954] = 7.887023791055e-01;
C[5955] = -1.314674200046e-03;
C[5956] = -1.442161471134e-02;
C[5957] = 2.177167926093e+00;
C[5958] = 9.844476434999e-03;
C[5959] = -4.179706683994e-02;
C[5960] = -7.213714427135e+00;
C[5961] = 6.019792609972e-04;
C[5962] = 2.778869559360e-03;
C[5963] = 5.064798559876e-02;
C[5964] = 1.018257099516e-04;
C[5965] = 6.331929616860e-03;
C[5966] = 4.026520946007e-01;
C[5967] = 1.968044739247e-03;
C[5968] = -1.322423060176e-02;
C[5969] = 1.927539186041e-02;
C[5970] = -1.734943571545e-02;
C[5971] = 1.331477028785e-02;
C[5972] = 4.168462473669e+00;
C[5973] = -2.670380494104e-03;
C[5974] = 4.065316525082e-03;
C[5975] = 5.249947736761e-01;
C[5976] = 0.000000000000e+00;
C[5977] = 0.000000000000e+00;
C[5978] = 0.000000000000e+00;
C[5979] = 0.000000000000e+00;
C[5980] = 0.000000000000e+00;
C[5981] = 0.000000000000e+00;
C[5982] = 0.000000000000e+00;
C[5983] = 0.000000000000e+00;
C[5984] = 0.000000000000e+00;
C[5985] = 0.000000000000e+00;
C[5986] = 0.000000000000e+00;
C[5987] = 0.000000000000e+00;
C[5988] = 0.000000000000e+00;
C[5989] = 0.000000000000e+00;
C[5990] = 0.000000000000e+00;
C[5991] = 0.000000000000e+00;
C[5992] = 0.000000000000e+00;
C[5993] = 0.000000000000e+00;
C[5994] = 1.000000000000e+00;
C[5995] = 0.000000000000e+00;
C[5996] = 0.000000000000e+00;
C[5997] = 2.000000000000e-01;
C[5998] = -0.000000000000e+00;
C[5999] = -0.000000000000e+00;
C[6000] = 0.000000000000e+00;
C[6001] = 0.000000000000e+00;
C[6002] = 0.000000000000e+00;
C[6003] = 0.000000000000e+00;
C[6004] = 0.000000000000e+00;
C[6005] = 0.000000000000e+00;
C[6006] = 0.000000000000e+00;
C[6007] = 0.000000000000e+00;
C[6008] = 0.000000000000e+00;
C[6009] = 0.000000000000e+00;
C[6010] = 0.000000000000e+00;
C[6011] = 0.000000000000e+00;
C[6012] = 0.000000000000e+00;
C[6013] = 0.000000000000e+00;
C[6014] = 0.000000000000e+00;
C[6015] = 0.000000000000e+00;
C[6016] = 0.000000000000e+00;
C[6017] = 0.000000000000e+00;
C[6018] = 0.000000000000e+00;
C[6019] = 1.000000000000e+00;
C[6020] = 0.000000000000e+00;
C[6021] = -0.000000000000e+00;
C[6022] = 2.000000000000e-01;
C[6023] = -0.000000000000e+00;
C[6024] = 0.000000000000e+00;
C[6025] = 0.000000000000e+00;
C[6026] = 0.000000000000e+00;
C[6027] = 0.000000000000e+00;
C[6028] = 0.000000000000e+00;
C[6029] = 0.000000000000e+00;
C[6030] = 0.000000000000e+00;
C[6031] = 0.000000000000e+00;
C[6032] = 0.000000000000e+00;
C[6033] = 0.000000000000e+00;
C[6034] = 0.000000000000e+00;
C[6035] = 0.000000000000e+00;
C[6036] = 0.000000000000e+00;
C[6037] = 0.000000000000e+00;
C[6038] = 0.000000000000e+00;
C[6039] = 0.000000000000e+00;
C[6040] = 0.000000000000e+00;
C[6041] = 0.000000000000e+00;
C[6042] = 0.000000000000e+00;
C[6043] = 0.000000000000e+00;
C[6044] = 1.000000000000e+00;
C[6045] = -0.000000000000e+00;
C[6046] = -0.000000000000e+00;
C[6047] = 2.000000000000e-01;
C[6048] = 8.307861496708e-02;
C[6049] = -2.353691836138e-02;
C[6050] = 7.839346248028e-03;
C[6051] = 3.612250682773e-01;
C[6052] = 2.060241458303e-02;
C[6053] = 4.050638575425e-03;
C[6054] = 4.985863465352e-02;
C[6055] = 2.700937356050e-03;
C[6056] = 5.846365171765e-04;
C[6057] = 1.315022178965e-01;
C[6058] = -1.863601173785e-03;
C[6059] = 5.948792571040e-04;
C[6060] = 2.938067464111e-02;
C[6061] = 1.623686007835e-03;
C[6062] = 3.453124729246e-04;
C[6063] = 2.412713783571e-03;
C[6064] = 1.266593830108e-04;
C[6065] = 2.810913426753e-05;
C[6066] = 2.425387736313e-03;
C[6067] = 1.113414041939e-04;
C[6068] = 1.365882366534e-05;
C[6069] = -1.648478882165e-05;
C[6070] = -1.841575434194e-06;
C[6071] = -3.271359535109e-07;
C[6072] = -2.339151414423e-02;
C[6073] = 5.136576492431e-02;
C[6074] = 9.555019356663e-04;
C[6075] = 2.055511689854e-02;
C[6076] = 3.949424006315e-01;
C[6077] = 5.290744608488e-03;
C[6078] = 2.624092484187e-03;
C[6079] = 5.635925090162e-02;
C[6080] = 2.669565619572e-03;
C[6081] = -1.853907559308e-03;
C[6082] = 1.285662067711e-01;
C[6083] = -1.360976334856e-04;
C[6084] = 1.618856904162e-03;
C[6085] = 3.249365893821e-02;
C[6086] = 6.555808306800e-04;
C[6087] = 1.223949686457e-04;
C[6088] = 2.754219126563e-03;
C[6089] = 1.491768378868e-04;
C[6090] = 1.024135867511e-04;
C[6091] = 2.843865054362e-03;
C[6092] = 2.340264582233e-04;
C[6093] = -1.970128889028e-06;
C[6094] = -1.291788760313e-05;
C[6095] = 3.155772051516e-06;
C[6096] = 7.903836775973e-03;
C[6097] = 1.051820554311e-03;
C[6098] = 6.925096855482e-02;
C[6099] = 4.130114035614e-03;
C[6100] = 5.485186794846e-03;
C[6101] = 3.476607073314e-01;
C[6102] = 4.993854543985e-04;
C[6103] = 2.495704409992e-03;
C[6104] = 5.055318580065e-02;
C[6105] = 5.991786256337e-04;
C[6106] = -1.296763922426e-04;
C[6107] = 1.302365746006e-01;
C[6108] = 3.477485348121e-04;
C[6109] = 6.628570980584e-04;
C[6110] = 2.858756790444e-02;
C[6111] = 2.326907310902e-05;
C[6112] = 1.392613903257e-04;
C[6113] = 2.473864575004e-03;
C[6114] = 3.729210656650e-06;
C[6115] = 2.138818457672e-04;
C[6116] = 2.611304160106e-03;
C[6117] = -4.543535112975e-07;
C[6118] = 2.890489551150e-06;
C[6119] = -1.198291342572e-05;
C[6120] = 3.610084112357e-01;
C[6121] = 2.088439167045e-02;
C[6122] = 4.259060081124e-03;
C[6123] = 1.318981962280e-01;
C[6124] = -2.054587118688e-02;
C[6125] = -1.769139491229e-03;
C[6126] = 3.921022183357e-01;
C[6127] = -3.109491261316e-03;
C[6128] = -9.906172078688e-04;
C[6129] = 2.936623083834e-02;
C[6130] = 1.640808555622e-03;
C[6131] = 3.563449378462e-04;
C[6132] = 1.335913558863e-01;
C[6133] = -1.559718544611e-03;
C[6134] = -1.562970790214e-04;
C[6135] = 3.202810508067e-02;
C[6136] = -2.013882088157e-04;
C[6137] = -1.316564162914e-04;
C[6138] = 6.113733790306e-02;
C[6139] = -5.018520812655e-04;
C[6140] = -7.172781522785e-04;
C[6141] = 2.670452485281e-03;
C[6142] = -1.801203142280e-05;
C[6143] = -3.506213143351e-05;
C[6144] = 2.060533124536e-02;
C[6145] = 3.947325089951e-01;
C[6146] = 5.655764572331e-03;
C[6147] = -2.035711150639e-02;
C[6148] = 9.259921780053e-02;
C[6149] = -1.342197823484e-02;
C[6150] = -3.027049607254e-03;
C[6151] = 3.910350661008e-01;
C[6152] = 5.255446060287e-03;
C[6153] = 1.623880451990e-03;
C[6154] = 3.247966616245e-02;
C[6155] = 6.742289498908e-04;
C[6156] = -1.547134565912e-03;
C[6157] = 1.300177968964e-01;
C[6158] = -1.359724131883e-03;
C[6159] = -1.972967616911e-04;
C[6160] = 3.215682712319e-02;
C[6161] = 5.504854389921e-04;
C[6162] = -5.009883009156e-04;
C[6163] = 6.196346899906e-02;
C[6164] = 2.040980012626e-03;
C[6165] = -1.821867915060e-05;
C[6166] = 2.722664031995e-03;
C[6167] = 9.864726687220e-05;
C[6168] = 4.017073510523e-03;
C[6169] = 5.206403981349e-03;
C[6170] = 3.475027381818e-01;
C[6171] = -1.650515992817e-03;
C[6172] = -1.326930680471e-02;
C[6173] = 1.292556132849e-01;
C[6174] = -9.028198920636e-04;
C[6175] = 5.465545130354e-03;
C[6176] = 4.027480724054e-01;
C[6177] = 3.430748019312e-04;
C[6178] = 6.499581222041e-04;
C[6179] = 2.857703662779e-02;
C[6180] = -1.483888457939e-04;
C[6181] = -1.349546036541e-03;
C[6182] = 1.331757528475e-01;
C[6183] = -1.274846071474e-04;
C[6184] = 5.617532924904e-04;
C[6185] = 3.296398243275e-02;
C[6186] = -7.160749641693e-04;
C[6187] = 2.044463879820e-03;
C[6188] = 6.467585955383e-02;
C[6189] = -3.520404390713e-05;
C[6190] = 9.799158751451e-05;
C[6191] = 2.830693218069e-03;
C[6192] = 4.961743368773e-02;
C[6193] = 3.342049595606e-03;
C[6194] = 3.045248053040e-04;
C[6195] = 3.927492708222e-01;
C[6196] = -3.862944103681e-03;
C[6197] = -8.839478092179e-04;
C[6198] = 4.052993225707e-02;
C[6199] = 1.379603402797e-03;
C[6200] = 3.114126152651e-03;
C[6201] = 2.396633719185e-03;
C[6202] = 1.702587760737e-04;
C[6203] = 1.027836316938e-05;
C[6204] = 3.207124191310e-02;
C[6205] = -2.530230614529e-04;
C[6206] = -1.262264682911e-04;
C[6207] = 1.278507869585e-01;
C[6208] = 1.588349181680e-04;
C[6209] = 3.022657565013e-04;
C[6210] = 5.147939873833e-01;
C[6211] = -1.055972897349e-03;
C[6212] = -2.472187038479e-03;
C[6213] = 3.769866971021e-02;
C[6214] = -7.658444986410e-05;
C[6215] = -1.859589264502e-04;
C[6216] = 3.413811551282e-03;
C[6217] = 5.610072876692e-02;
C[6218] = 1.992307589953e-03;
C[6219] = -3.958343228916e-03;
C[6220] = 3.916298102393e-01;
C[6221] = 5.854520563550e-03;
C[6222] = 1.389699975539e-03;
C[6223] = 3.359820082593e-02;
C[6224] = -1.241103235015e-02;
C[6225] = 1.741843293596e-04;
C[6226] = 2.736984317583e-03;
C[6227] = 1.057016023231e-04;
C[6228] = -2.579783399891e-04;
C[6229] = 3.219647673242e-02;
C[6230] = 5.876849880368e-04;
C[6231] = 1.595080230175e-04;
C[6232] = 1.272578111546e-01;
C[6233] = -1.093337483778e-03;
C[6234] = -1.047597649963e-03;
C[6235] = 5.159923200494e-01;
C[6236] = 4.487064064944e-03;
C[6237] = -7.626309765529e-05;
C[6238] = 3.782242766356e-02;
C[6239] = 3.983814735781e-04;
C[6240] = 3.789950546709e-04;
C[6241] = 2.135180513683e-03;
C[6242] = 4.933187380590e-02;
C[6243] = -9.946544780930e-04;
C[6244] = 5.680044832649e-03;
C[6245] = 4.038903393738e-01;
C[6246] = 3.131040768736e-03;
C[6247] = -1.240680985581e-02;
C[6248] = 1.922567289293e-02;
C[6249] = 1.439970343383e-05;
C[6250] = 1.135511641609e-04;
C[6251] = 2.392443775355e-03;
C[6252] = -1.319255676397e-04;
C[6253] = 5.787920238163e-04;
C[6254] = 3.304013356398e-02;
C[6255] = 3.033933975736e-04;
C[6256] = -1.093055984155e-03;
C[6257] = 1.259586479484e-01;
C[6258] = -2.461552568079e-03;
C[6259] = 4.501293557255e-03;
C[6260] = 5.250943643065e-01;
C[6261] = -1.856107800111e-04;
C[6262] = 3.990551161037e-04;
C[6263] = 3.862483560279e-02;
C[6264] = -7.006685471096e+00;
C[6265] = -5.491427792201e-02;
C[6266] = 5.386333981402e-02;
C[6267] = 2.072966712782e+00;
C[6268] = 3.756501635209e-02;
C[6269] = 2.955380621136e-02;
C[6270] = 7.772312110088e-01;
C[6271] = 3.323583629606e-02;
C[6272] = 9.755133180796e-03;
C[6273] = 7.985337822291e-02;
C[6274] = -1.734758890339e-02;
C[6275] = 7.670625057637e-03;
C[6276] = 3.645139863832e-01;
C[6277] = 1.437661221791e-02;
C[6278] = 4.814676275524e-03;
C[6279] = 5.013065476323e-02;
C[6280] = 2.285597455646e-03;
C[6281] = 5.593412260587e-04;
C[6282] = 7.813598902303e-02;
C[6283] = 3.588805668450e-03;
C[6284] = 3.211889685332e-04;
C[6285] = 2.427852695397e-03;
C[6286] = 1.270133499470e-04;
C[6287] = 9.658350869689e-07;
C[6288] = -5.200619357900e-02;
C[6289] = -7.067941835394e+00;
C[6290] = 1.818662469708e-02;
C[6291] = 3.592802040535e-02;
C[6292] = 2.147432945911e+00;
C[6293] = 1.337899772189e-02;
C[6294] = 3.236246608931e-02;
C[6295] = 8.696252490490e-01;
C[6296] = 4.057899332011e-02;
C[6297] = -1.720218468624e-02;
C[6298] = 4.876224813858e-02;
C[6299] = -2.432906752692e-03;
C[6300] = 1.432698080235e-02;
C[6301] = 3.980230106479e-01;
C[6302] = 8.684691644193e-03;
C[6303] = 2.220623665155e-03;
C[6304] = 5.661397418800e-02;
C[6305] = 2.841108990876e-03;
C[6306] = 3.382464488233e-03;
C[6307] = 8.958895823909e-02;
C[6308] = 6.046532338976e-03;
C[6309] = 1.192440489462e-04;
C[6310] = 2.843977284559e-03;
C[6311] = 2.138347539347e-04;
C[6312] = 5.515315037293e-02;
C[6313] = 2.011299706997e-02;
C[6314] = -7.125405535790e+00;
C[6315] = 3.004776477141e-02;
C[6316] = 1.517262188954e-02;
C[6317] = 1.991418931197e+00;
C[6318] = 8.800780609582e-03;
C[6319] = 3.852643948533e-02;
C[6320] = 7.888783609875e-01;
C[6321] = 7.735115585582e-03;
C[6322] = -2.336588134048e-03;
C[6323] = 6.045348773696e-02;
C[6324] = 4.871022993893e-03;
C[6325] = 8.840322551453e-03;
C[6326] = 3.567992564538e-01;
C[6327] = 4.877611653891e-04;
C[6328] = 2.690153755531e-03;
C[6329] = 5.110889243369e-02;
C[6330] = 9.947072252225e-05;
C[6331] = 5.581938988039e-03;
C[6332] = 8.215650990136e-02;
C[6333] = -7.196076398839e-06;
C[6334] = 1.966406953059e-04;
C[6335] = 2.574288946763e-03;
C[6336] = 2.068633571951e+00;
C[6337] = 4.251351584340e-02;
C[6338] = 3.262668568162e-02;
C[6339] = -6.192694132575e+00;
C[6340] = -4.528826113484e-02;
C[6341] = -2.247125673554e-03;
C[6342] = 2.171787391825e+00;
C[6343] = -3.361998970056e-02;
C[6344] = -5.309637271597e-03;
C[6345] = 3.642973293416e-01;
C[6346] = 1.465625557425e-02;
C[6347] = 4.999969039404e-03;
C[6348] = 1.270826459934e-01;
C[6349] = -1.360479722084e-02;
C[6350] = -1.467069301298e-03;
C[6351] = 3.934613935615e-01;
C[6352] = -3.442556217570e-03;
C[6353] = -1.997832692807e-03;
C[6354] = 1.034766944328e+00;
C[6355] = -1.056044570847e-02;
C[6356] = -1.254074012456e-02;
C[6357] = 6.113740314738e-02;
C[6358] = -5.495329818575e-04;
C[6359] = -7.525034745678e-04;
C[6360] = 3.762334959864e-02;
C[6361] = 2.143235113182e+00;
C[6362] = 1.858417743924e-02;
C[6363] = -4.151306752514e-02;
C[6364] = -6.289958610132e+00;
C[6365] = -3.716711355763e-02;
C[6366] = -3.268917452798e-02;
C[6367] = 2.095882863479e+00;
C[6368] = -2.096006276133e-02;
C[6369] = 1.437952888024e-02;
C[6370] = 3.978131190115e-01;
C[6371] = 9.010900328938e-03;
C[6372] = -1.341603754035e-02;
C[6373] = 8.842614611843e-02;
C[6374] = -1.619942149662e-02;
C[6375] = -3.376795427243e-03;
C[6376] = 3.920252486410e-01;
C[6377] = 4.498458126254e-03;
C[6378] = -1.044776856533e-02;
C[6379] = 1.043264866305e+00;
C[6380] = 3.077283149638e-02;
C[6381] = -5.467856723541e-04;
C[6382] = 6.194450811139e-02;
C[6383] = 2.041054308193e-03;
C[6384] = 2.888250491332e-02;
C[6385] = 1.169218517912e-02;
C[6386] = 1.988259548204e+00;
C[6387] = 1.253442946900e-04;
C[6388] = -3.411368495506e-02;
C[6389] = -6.209231678618e+00;
C[6390] = -4.399374771208e-03;
C[6391] = -1.822579328705e-02;
C[6392] = 2.176730470104e+00;
C[6393] = 4.781111210622e-03;
C[6394] = 8.600351017055e-03;
C[6395] = 3.566412873042e-01;
C[6396] = -1.348445802886e-03;
C[6397] = -1.604675006649e-02;
C[6398] = 1.203106650900e-01;
C[6399] = -1.933873208017e-03;
C[6400] = 4.684126311383e-03;
C[6401] = 4.020276965793e-01;
C[6402] = -1.238985169064e-02;
C[6403] = 3.097190079161e-02;
C[6404] = 1.090664886457e+00;
C[6405] = -7.464906712686e-04;
C[6406] = 2.037902862818e-03;
C[6407] = 6.472182194685e-02;
C[6408] = 7.724071916929e-01;
C[6409] = 4.672160831768e-02;
C[6410] = 4.903567627691e-03;
C[6411] = 2.184728441555e+00;
C[6412] = -4.940706445652e-02;
C[6413] = -4.021933114294e-03;
C[6414] = -7.141152870074e+00;
C[6415] = 1.849067639621e-03;
C[6416] = 1.282925330309e-02;
C[6417] = 4.988945379743e-02;
C[6418] = 2.938580776573e-03;
C[6419] = 2.929005162946e-04;
C[6420] = 3.941084460480e-01;
C[6421] = -4.212689923670e-03;
C[6422] = -1.915001125171e-03;
C[6423] = 4.040894890758e-02;
C[6424] = 1.923731660785e-03;
C[6425] = 3.792110736275e-03;
C[6426] = 4.108984676635e+00;
C[6427] = -3.931594564850e-03;
C[6428] = -1.207598396027e-02;
C[6429] = 5.132577722644e-01;
C[6430] = -7.891545947185e-04;
C[6431] = -2.118092464367e-03;
C[6432] = 4.749332020069e-02;
C[6433] = 8.644548063550e-01;
C[6434] = 2.845850308454e-02;
C[6435] = -5.059702905257e-02;
C[6436] = 2.107777746248e+00;
C[6437] = -1.044628462311e-02;
C[6438] = 2.050999094463e-03;
C[6439] = -7.166700889080e+00;
C[6440] = -3.982092269680e-02;
C[6441] = 2.998471650878e-03;
C[6442] = 5.635545205330e-02;
C[6443] = 2.186756935491e-03;
C[6444] = -4.291408185170e-03;
C[6445] = 3.926199927795e-01;
C[6446] = 5.073101744579e-03;
C[6447] = 1.933828233527e-03;
C[6448] = 3.377796299467e-02;
C[6449] = -1.210903800708e-02;
C[6450] = -3.822426554936e-03;
C[6451] = 4.109429617411e+00;
C[6452] = 1.935099264432e-02;
C[6453] = -7.840377061293e-04;
C[6454] = 5.145420417098e-01;
C[6455] = 4.732731927894e-03;
C[6456] = 5.642303930684e-03;
C[6457] = 2.989129120234e-02;
C[6458] = 7.644521210926e-01;
C[6459] = -5.390382676081e-03;
C[6460] = -1.246808731408e-02;
C[6461] = 2.199575809471e+00;
C[6462] = 1.316754562479e-02;
C[6463] = -3.973647280981e-02;
C[6464] = -7.214174217691e+00;
C[6465] = 3.536997635531e-04;
C[6466] = 2.306723884988e-03;
C[6467] = 4.988758043894e-02;
C[6468] = -2.001869963031e-03;
C[6469] = 4.923056898617e-03;
C[6470] = 4.031699635477e-01;
C[6471] = 3.809025352360e-03;
C[6472] = -1.210481551273e-02;
C[6473] = 1.900233822808e-02;
C[6474] = -1.195382314957e-02;
C[6475] = 1.958837962076e-02;
C[6476] = 4.169995341816e+00;
C[6477] = -2.114334947903e-03;
C[6478] = 4.750015780019e-03;
C[6479] = 5.253857843620e-01;
C[6480] = 0.000000000000e+00;
C[6481] = 0.000000000000e+00;
C[6482] = 0.000000000000e+00;
C[6483] = 0.000000000000e+00;
C[6484] = 0.000000000000e+00;
C[6485] = 0.000000000000e+00;
C[6486] = 0.000000000000e+00;
C[6487] = 0.000000000000e+00;
C[6488] = 0.000000000000e+00;
C[6489] = 0.000000000000e+00;
C[6490] = 0.000000000000e+00;
C[6491] = 0.000000000000e+00;
C[6492] = 0.000000000000e+00;
C[6493] = 0.000000000000e+00;
C[6494] = 0.000000000000e+00;
C[6495] = 0.000000000000e+00;
C[6496] = 0.000000000000e+00;
C[6497] = 0.000000000000e+00;
C[6498] = 1.000000000000e+00;
C[6499] = 0.000000000000e+00;
C[6500] = 0.000000000000e+00;
C[6501] = 2.000000000000e-01;
C[6502] = -0.000000000000e+00;
C[6503] = -0.000000000000e+00;
C[6504] = 0.000000000000e+00;
C[6505] = 0.000000000000e+00;
C[6506] = 0.000000000000e+00;
C[6507] = 0.000000000000e+00;
C[6508] = 0.000000000000e+00;
C[6509] = 0.000000000000e+00;
C[6510] = 0.000000000000e+00;
C[6511] = 0.000000000000e+00;
C[6512] = 0.000000000000e+00;
C[6513] = 0.000000000000e+00;
C[6514] = 0.000000000000e+00;
C[6515] = 0.000000000000e+00;
C[6516] = 0.000000000000e+00;
C[6517] = 0.000000000000e+00;
C[6518] = 0.000000000000e+00;
C[6519] = 0.000000000000e+00;
C[6520] = 0.000000000000e+00;
C[6521] = 0.000000000000e+00;
C[6522] = 0.000000000000e+00;
C[6523] = 1.000000000000e+00;
C[6524] = 0.000000000000e+00;
C[6525] = -0.000000000000e+00;
C[6526] = 2.000000000000e-01;
C[6527] = -0.000000000000e+00;
C[6528] = 0.000000000000e+00;
C[6529] = 0.000000000000e+00;
C[6530] = 0.000000000000e+00;
C[6531] = 0.000000000000e+00;
C[6532] = 0.000000000000e+00;
C[6533] = 0.000000000000e+00;
C[6534] = 0.000000000000e+00;
C[6535] = 0.000000000000e+00;
C[6536] = 0.000000000000e+00;
C[6537] = 0.000000000000e+00;
C[6538] = 0.000000000000e+00;
C[6539] = 0.000000000000e+00;
C[6540] = 0.000000000000e+00;
C[6541] = 0.000000000000e+00;
C[6542] = 0.000000000000e+00;
C[6543] = 0.000000000000e+00;
C[6544] = 0.000000000000e+00;
C[6545] = 0.000000000000e+00;
C[6546] = 0.000000000000e+00;
C[6547] = 0.000000000000e+00;
C[6548] = 1.000000000000e+00;
C[6549] = -0.000000000000e+00;
C[6550] = -0.000000000000e+00;
C[6551] = 2.000000000000e-01;
C[6552] = 7.703057349141e-02;
C[6553] = -9.662284569165e-03;
C[6554] = 8.745827727916e-03;
C[6555] = 3.724036455820e-01;
C[6556] = 5.311710937554e-03;
C[6557] = 3.552007003618e-03;
C[6558] = 5.201783213028e-02;
C[6559] = -2.153571860313e-05;
C[6560] = -2.305933846389e-04;
C[6561] = 1.310558895496e-01;
C[6562] = -6.443177177556e-04;
C[6563] = 6.825191135266e-04;
C[6564] = 3.045508392927e-02;
C[6565] = 2.436549618828e-04;
C[6566] = 1.622535967758e-04;
C[6567] = 2.524028699155e-03;
C[6568] = -1.402956905368e-05;
C[6569] = -2.091354145600e-05;
C[6570] = 2.530086666865e-03;
C[6571] = -4.769114208169e-05;
C[6572] = -5.610858660033e-05;
C[6573] = -1.573893750923e-05;
C[6574] = -2.992900144229e-06;
C[6575] = -2.640344865407e-06;
C[6576] = -9.519629372182e-03;
C[6577] = 4.710533125798e-02;
C[6578] = -5.528965908799e-04;
C[6579] = 5.257635584276e-03;
C[6580] = 4.003509963294e-01;
C[6581] = 1.176607359880e-02;
C[6582] = -7.690409495052e-05;
C[6583] = 5.704497609529e-02;
C[6584] = 3.232220626003e-03;
C[6585] = -6.348073712901e-04;
C[6586] = 1.281538513347e-01;
C[6587] = 1.113639549845e-04;
C[6588] = 2.390929435923e-04;
C[6589] = 3.285451006526e-02;
C[6590] = 1.062390296404e-03;
C[6591] = -1.649578777517e-05;
C[6592] = 2.778512705947e-03;
C[6593] = 1.700628643104e-04;
C[6594] = -5.015640164629e-05;
C[6595] = 2.825474512361e-03;
C[6596] = 2.462424281396e-04;
C[6597] = -2.944901792176e-06;
C[6598] = -1.500050588178e-05;
C[6599] = 1.670966472217e-06;
C[6600] = 8.745314844806e-03;
C[6601] = -5.433268548666e-04;
C[6602] = 5.022318589703e-02;
C[6603] = 3.550578043184e-03;
C[6604] = 1.195431304641e-02;
C[6605] = 3.706181061179e-01;
C[6606] = -2.612661064142e-04;
C[6607] = 3.132530618753e-03;
C[6608] = 5.406679525340e-02;
C[6609] = 6.824849213193e-04;
C[6610] = 1.120019373854e-04;
C[6611] = 1.287588459472e-01;
C[6612] = 1.609890457227e-04;
C[6613] = 1.072782593450e-03;
C[6614] = 3.057882007695e-02;
C[6615] = -2.202523613513e-05;
C[6616] = 1.641915938311e-04;
C[6617] = 2.652282633089e-03;
C[6618] = -5.520398151351e-05;
C[6619] = 2.306526236087e-04;
C[6620] = 2.820284021943e-03;
C[6621] = -2.479580011000e-06;
C[6622] = 1.531886290850e-06;
C[6623] = -1.141484646010e-05;
C[6624] = 3.720862040902e-01;
C[6625] = 5.595688152000e-03;
C[6626] = 3.925485712454e-03;
C[6627] = 1.208845804307e-01;
C[6628] = -4.031705347342e-03;
C[6629] = 6.128027247615e-04;
C[6630] = 3.915688484677e-01;
C[6631] = -1.958809029158e-03;
C[6632] = -2.428659132369e-03;
C[6633] = 3.043392116315e-02;
C[6634] = 2.616297814405e-04;
C[6635] = 1.859828903407e-04;
C[6636] = 1.327462780525e-01;
C[6637] = -1.097179141420e-04;
C[6638] = 1.466328462222e-04;
C[6639] = 3.178920240600e-02;
C[6640] = -1.564294805598e-04;
C[6641] = -2.296840499943e-04;
C[6642] = 5.989981855217e-02;
C[6643] = -3.697807321323e-04;
C[6644] = -6.844564176565e-04;
C[6645] = 2.602162176614e-03;
C[6646] = -1.581519721925e-05;
C[6647] = -3.199936833153e-05;
C[6648] = 5.408800456983e-03;
C[6649] = 4.011147499800e-01;
C[6650] = 1.095961238735e-02;
C[6651] = -3.891312723203e-03;
C[6652] = 8.677186383821e-02;
C[6653] = -1.832914647407e-02;
C[6654] = -1.897092084281e-03;
C[6655] = 3.912247442456e-01;
C[6656] = 4.419935022957e-03;
C[6657] = 2.501275965114e-04;
C[6658] = 3.290542697530e-02;
C[6659] = 1.006469216179e-03;
C[6660] = -1.003584058661e-04;
C[6661] = 1.298046338546e-01;
C[6662] = -1.756655521584e-03;
C[6663] = -1.536785733847e-04;
C[6664] = 3.199568250197e-02;
C[6665] = 5.973450406472e-04;
C[6666] = -3.640204302224e-04;
C[6667] = 6.091775418563e-02;
C[6668] = 2.343929362570e-03;
C[6669] = -1.568701116835e-05;
C[6670] = 2.664021899072e-03;
C[6671] = 1.204464440525e-04;
C[6672] = 3.818183953121e-03;
C[6673] = 1.075922835925e-02;
C[6674] = 3.698142554116e-01;
C[6675] = 6.853736067704e-04;
C[6676] = -1.822126125069e-02;
C[6677] = 1.077125659737e-01;
C[6678] = -2.399136006819e-03;
C[6679] = 4.562026913811e-03;
C[6680] = 3.985525079327e-01;
C[6681] = 1.799987267427e-04;
C[6682] = 9.952672804338e-04;
C[6683] = 3.052523002986e-02;
C[6684] = 1.514709050228e-04;
C[6685] = -1.749463173359e-03;
C[6686] = 1.313174184388e-01;
C[6687] = -2.286446895905e-04;
C[6688] = 6.050334093465e-04;
C[6689] = 3.271008885269e-02;
C[6690] = -6.700078292851e-04;
C[6691] = 2.327066923039e-03;
C[6692] = 6.442852830995e-02;
C[6693] = -3.135237040314e-05;
C[6694] = 1.188845198135e-04;
C[6695] = 2.826764677074e-03;
C[6696] = 5.082447888396e-02;
C[6697] = 6.636055094779e-04;
C[6698] = 1.015906889412e-05;
C[6699] = 3.922943171983e-01;
C[6700] = -2.659939766976e-03;
C[6701] = -2.801556556716e-03;
C[6702] = 4.298991555140e-02;
C[6703] = 2.029251041776e-03;
C[6704] = 3.473436570589e-03;
C[6705] = 2.444471816067e-03;
C[6706] = 3.287151918672e-05;
C[6707] = -3.930224447913e-06;
C[6708] = 3.183756698804e-02;
C[6709] = -2.045350855644e-04;
C[6710] = -2.554727262502e-04;
C[6711] = 1.282786318471e-01;
C[6712] = 1.736568581669e-04;
C[6713] = 2.810602105535e-04;
C[6714] = 5.115568609759e-01;
C[6715] = -8.345848583426e-05;
C[6716] = -5.918543446558e-04;
C[6717] = 3.745588509527e-02;
C[6718] = -9.688692350103e-07;
C[6719] = -1.994370316651e-05;
C[6720] = 7.275936483970e-04;
C[6721] = 5.670590737162e-02;
C[6722] = 3.189398842896e-03;
C[6723] = -2.715051095114e-03;
C[6724] = 3.913588087086e-01;
C[6725] = 4.612481208106e-03;
C[6726] = 2.021741436792e-03;
C[6727] = 3.647735810620e-02;
C[6728] = -1.258435012836e-02;
C[6729] = 3.591238874632e-05;
C[6730] = 2.755908124369e-03;
C[6731] = 1.679828087740e-04;
C[6732] = -2.068456182902e-04;
C[6733] = 3.200462013284e-02;
C[6734] = 6.083970289662e-04;
C[6735] = 1.731562178347e-04;
C[6736] = 1.276818671453e-01;
C[6737] = -1.183510687110e-03;
C[6738] = -8.472633386730e-05;
C[6739] = 5.128136479070e-01;
C[6740] = 4.618859397602e-03;
C[6741] = -1.032756142062e-06;
C[6742] = 3.757562279524e-02;
C[6743] = 4.005082642242e-04;
C[6744] = 6.182825525831e-05;
C[6745] = 3.331136188528e-03;
C[6746] = 5.296944898150e-02;
C[6747] = -2.824139069815e-03;
C[6748] = 4.429049402323e-03;
C[6749] = 3.994317491825e-01;
C[6750] = 3.454139263039e-03;
C[6751] = -1.257365040719e-02;
C[6752] = 1.874706944035e-02;
C[6753] = -1.418765462852e-06;
C[6754] = 1.766572351454e-04;
C[6755] = 2.579126214962e-03;
C[6756] = -2.560493791574e-04;
C[6757] = 5.979526659383e-04;
C[6758] = 3.276870493601e-02;
C[6759] = 2.797737233836e-04;
C[6760] = -1.182797372365e-03;
C[6761] = 1.259584746881e-01;
C[6762] = -6.012502228971e-04;
C[6763] = 4.648748782211e-03;
C[6764] = 5.261706911909e-01;
C[6765] = -2.037088160326e-05;
C[6766] = 4.019634144587e-04;
C[6767] = 3.870522623289e-02;
C[6768] = -6.998174192447e+00;
C[6769] = -3.219330633724e-02;
C[6770] = 5.163168928725e-02;
C[6771] = 2.125862338337e+00;
C[6772] = -7.149154429171e-04;
C[6773] = 2.427555157459e-02;
C[6774] = 8.108832883236e-01;
C[6775] = -5.377948568563e-03;
C[6776] = -4.172650763845e-03;
C[6777] = 7.754227133738e-02;
C[6778] = -5.322738476101e-03;
C[6779] = 8.346496200730e-03;
C[6780] = 3.783163750620e-01;
C[6781] = -4.539095562825e-04;
C[6782] = 1.400160686705e-03;
C[6783] = 5.212384332562e-02;
C[6784] = -3.678757780414e-04;
C[6785] = -3.683802821050e-04;
C[6786] = 8.025796107965e-02;
C[6787] = -7.218763875623e-04;
C[6788] = -1.049812873510e-03;
C[6789] = 2.472843262237e-03;
C[6790] = -2.624578150115e-05;
C[6791] = -4.775134157793e-05;
C[6792] = -2.934020239759e-02;
C[6793] = -7.080434563507e+00;
C[6794] = 5.257729506639e-02;
C[6795] = -2.266892575591e-03;
C[6796] = 2.152407559829e+00;
C[6797] = 4.190867120643e-02;
C[6798] = -5.943967804646e-03;
C[6799] = 8.770218010808e-01;
C[6800] = 4.654107322129e-02;
C[6801] = -5.180083279119e-03;
C[6802] = 4.394272453562e-02;
C[6803] = 6.288518607000e-03;
C[6804] = -5.142186948952e-04;
C[6805] = 4.004876268479e-01;
C[6806] = 1.194002866319e-02;
C[6807] = -4.024396460532e-04;
C[6808] = 5.688037518782e-02;
C[6809] = 3.304746729132e-03;
C[6810] = -7.874071694650e-04;
C[6811] = 8.893963453789e-02;
C[6812] = 7.057834848967e-03;
C[6813] = -2.822138330840e-05;
C[6814] = 2.799234972759e-03;
C[6815] = 2.553569057298e-04;
C[6816] = 5.162143162505e-02;
C[6817] = 5.276868978666e-02;
C[6818] = -7.142091513004e+00;
C[6819] = 2.395747914592e-02;
C[6820] = 4.489387610904e-02;
C[6821] = 2.063636158456e+00;
C[6822] = -4.395750340145e-03;
C[6823] = 4.548766956744e-02;
C[6824] = 8.379829757189e-01;
C[6825] = 8.345983317620e-03;
C[6826] = 6.298088343014e-03;
C[6827] = 4.881429884674e-02;
C[6828] = 1.391337370754e-03;
C[6829] = 1.216269701016e-02;
C[6830] = 3.782660890828e-01;
C[6831] = -3.818135805287e-04;
C[6832] = 3.202768637080e-03;
C[6833] = 5.447860703234e-02;
C[6834] = -1.054070001078e-03;
C[6835] = 6.613413489620e-03;
C[6836] = 8.808682774866e-02;
C[6837] = -4.753860141073e-05;
C[6838] = 2.368923871103e-04;
C[6839] = 2.782322029442e-03;
C[6840] = 2.119513508501e+00;
C[6841] = 4.494158778887e-03;
C[6842] = 3.145563253131e-02;
C[6843] = -6.221862316601e+00;
C[6844] = -5.160908783700e-03;
C[6845] = 5.162509277038e-03;
C[6846] = 2.148010401477e+00;
C[6847] = -2.144390091223e-03;
C[6848] = -2.131482864858e-03;
C[6849] = 3.779989335702e-01;
C[6850] = -1.761661271713e-04;
C[6851] = 1.766245040024e-03;
C[6852] = 1.192501595134e-01;
C[6853] = 1.540084020302e-03;
C[6854] = 2.688211778822e-03;
C[6855] = 3.876577159943e-01;
C[6856] = -1.564820610104e-03;
C[6857] = -2.438947370990e-03;
C[6858] = 1.009875631413e+00;
C[6859] = -5.910429236817e-03;
C[6860] = -1.060532264738e-02;
C[6861] = 5.964364843621e-02;
C[6862] = -3.596412752742e-04;
C[6863] = -6.288382677472e-04;
C[6864] = 1.226874945665e-03;
C[6865] = 2.167682632840e+00;
C[6866] = 2.499986292772e-02;
C[6867] = -2.353056300922e-03;
C[6868] = -6.296111618199e+00;
C[6869] = -4.242383838116e-02;
C[6870] = -1.463044624489e-03;
C[6871] = 2.093975667157e+00;
C[6872] = -2.461709892130e-02;
C[6873] = -3.568200368534e-04;
C[6874] = 4.012513804985e-01;
C[6875] = 1.116799635110e-02;
C[6876] = 1.680476644441e-03;
C[6877] = 8.958003870639e-02;
C[6878] = -1.959625810856e-02;
C[6879] = -1.528298252610e-03;
C[6880] = 3.883319717586e-01;
C[6881] = 5.888863988357e-03;
C[6882] = -5.791341441949e-03;
C[6883] = 1.023096236350e+00;
C[6884] = 3.653956039818e-02;
C[6885] = -3.527740241736e-04;
C[6886] = 6.067485388454e-02;
C[6887] = 2.420315495709e-03;
C[6888] = 2.959909056465e-02;
C[6889] = 2.177176641548e-02;
C[6890] = 2.047559144330e+00;
C[6891] = 6.613926917216e-03;
C[6892] = -4.026613391367e-02;
C[6893] = -6.264538546342e+00;
C[6894] = -1.879637038738e-03;
C[6895] = -2.287318881333e-02;
C[6896] = 2.124858056802e+00;
C[6897] = 1.666337636209e-03;
C[6898] = 1.093318342364e-02;
C[6899] = 3.774622383765e-01;
C[6900] = 2.760782660830e-03;
C[6901] = -1.948837288519e-02;
C[6902] = 1.009998997135e-01;
C[6903] = -2.433207104204e-03;
C[6904] = 6.030580116996e-03;
C[6905] = 3.975995533245e-01;
C[6906] = -1.035845871745e-02;
C[6907] = 3.645655996689e-02;
C[6908] = 1.083915827362e+00;
C[6909] = -6.102625373569e-04;
C[6910] = 2.401816748510e-03;
C[6911] = 6.444665008501e-02;
C[6912] = 7.870162233973e-01;
C[6913] = 8.866224283922e-03;
C[6914] = 1.032753166022e-03;
C[6915] = 2.162519776090e+00;
C[6916] = -1.671999827840e-02;
C[6917] = -9.928048036664e-03;
C[6918] = -7.123209199388e+00;
C[6919] = 3.074087037530e-03;
C[6920] = 6.619058926707e-03;
C[6921] = 5.093049007930e-02;
C[6922] = 3.380699583752e-04;
C[6923] = -1.103884052203e-04;
C[6924] = 3.883831847250e-01;
C[6925] = -2.291145935306e-03;
C[6926] = -2.835627654101e-03;
C[6927] = 4.770210809751e-02;
C[6928] = 1.646322055736e-03;
C[6929] = 2.582605799746e-03;
C[6930] = 4.097386392156e+00;
C[6931] = 3.941799845230e-03;
C[6932] = 5.013502619469e-03;
C[6933] = 5.105806458755e-01;
C[6934] = 3.101429388257e-04;
C[6935] = 4.757334371590e-04;
C[6936] = 9.604638771439e-03;
C[6937] = 8.702404266074e-01;
C[6938] = 4.662503405030e-02;
C[6939] = -1.726923141034e-02;
C[6940] = 2.096656956417e+00;
C[6941] = -2.186410292742e-02;
C[6942] = 2.923894937846e-03;
C[6943] = -7.147516951031e+00;
C[6944] = -4.383632634279e-02;
C[6945] = 3.812535889587e-04;
C[6946] = 5.654130646416e-02;
C[6947] = 3.259636861223e-03;
C[6948] = -2.321062676060e-03;
C[6949] = 3.884660362216e-01;
C[6950] = 6.081034411292e-03;
C[6951] = 1.638812450752e-03;
C[6952] = 4.065885473146e-02;
C[6953] = -1.400471879629e-02;
C[6954] = 3.925054717211e-03;
C[6955] = 4.096244253262e+00;
C[6956] = 1.537058752166e-02;
C[6957] = 3.067642151219e-04;
C[6958] = 5.116828924666e-01;
C[6959] = 4.495778478606e-03;
C[6960] = 1.675782034098e-03;
C[6961] = 4.851938447179e-02;
C[6962] = 8.160360502811e-01;
C[6963] = -1.004108161378e-02;
C[6964] = -2.443481133398e-02;
C[6965] = 2.142442881798e+00;
C[6966] = 6.233112775714e-03;
C[6967] = -4.362233191931e-02;
C[6968] = -7.222875723975e+00;
C[6969] = -7.595864220786e-05;
C[6970] = 3.403662291657e-03;
C[6971] = 5.338126076044e-02;
C[6972] = -2.834427308437e-03;
C[6973] = 5.897978367723e-03;
C[6974] = 3.984787945743e-01;
C[6975] = 2.563308492196e-03;
C[6976] = -1.399401907512e-02;
C[6977] = 1.889737605712e-02;
C[6978] = 4.884039652515e-03;
C[6979] = 1.584923100021e-02;
C[6980] = 4.178672137254e+00;
C[6981] = 4.623437162399e-04;
C[6982] = 4.528990929876e-03;
C[6983] = 5.264771690437e-01;
C[6984] = 0.000000000000e+00;
C[6985] = 0.000000000000e+00;
C[6986] = 0.000000000000e+00;
C[6987] = 0.000000000000e+00;
C[6988] = 0.000000000000e+00;
C[6989] = 0.000000000000e+00;
C[6990] = 0.000000000000e+00;
C[6991] = 0.000000000000e+00;
C[6992] = 0.000000000000e+00;
C[6993] = 0.000000000000e+00;
C[6994] = 0.000000000000e+00;
C[6995] = 0.000000000000e+00;
C[6996] = 0.000000000000e+00;
C[6997] = 0.000000000000e+00;
C[6998] = 0.000000000000e+00;
C[6999] = 0.000000000000e+00;
C[7000] = 0.000000000000e+00;
C[7001] = 0.000000000000e+00;
C[7002] = 1.000000000000e+00;
C[7003] = 0.000000000000e+00;
C[7004] = 0.000000000000e+00;
C[7005] = 2.000000000000e-01;
C[7006] = -0.000000000000e+00;
C[7007] = -0.000000000000e+00;
C[7008] = 0.000000000000e+00;
C[7009] = 0.000000000000e+00;
C[7010] = 0.000000000000e+00;
C[7011] = 0.000000000000e+00;
C[7012] = 0.000000000000e+00;
C[7013] = 0.000000000000e+00;
C[7014] = 0.000000000000e+00;
C[7015] = 0.000000000000e+00;
C[7016] = 0.000000000000e+00;
C[7017] = 0.000000000000e+00;
C[7018] = 0.000000000000e+00;
C[7019] = 0.000000000000e+00;
C[7020] = 0.000000000000e+00;
C[7021] = 0.000000000000e+00;
C[7022] = 0.000000000000e+00;
C[7023] = 0.000000000000e+00;
C[7024] = 0.000000000000e+00;
C[7025] = 0.000000000000e+00;
C[7026] = 0.000000000000e+00;
C[7027] = 1.000000000000e+00;
C[7028] = 0.000000000000e+00;
C[7029] = -0.000000000000e+00;
C[7030] = 2.000000000000e-01;
C[7031] = -0.000000000000e+00;
C[7032] = 0.000000000000e+00;
C[7033] = 0.000000000000e+00;
C[7034] = 0.000000000000e+00;
C[7035] = 0.000000000000e+00;
C[7036] = 0.000000000000e+00;
C[7037] = 0.000000000000e+00;
C[7038] = 0.000000000000e+00;
C[7039] = 0.000000000000e+00;
C[7040] = 0.000000000000e+00;
C[7041] = 0.000000000000e+00;
C[7042] = 0.000000000000e+00;
C[7043] = 0.000000000000e+00;
C[7044] = 0.000000000000e+00;
C[7045] = 0.000000000000e+00;
C[7046] = 0.000000000000e+00;
C[7047] = 0.000000000000e+00;
C[7048] = 0.000000000000e+00;
C[7049] = 0.000000000000e+00;
C[7050] = 0.000000000000e+00;
C[7051] = 0.000000000000e+00;
C[7052] = 1.000000000000e+00;
C[7053] = -0.000000000000e+00;
C[7054] = -0.000000000000e+00;
C[7055] = 2.000000000000e-01;
C[7056] = 7.799225592183e-02;
C[7057] = 2.824475236821e-03;
C[7058] = 2.389643401288e-03;
C[7059] = 3.809287597780e-01;
C[7060] = -2.837838615217e-03;
C[7061] = -1.672643015271e-03;
C[7062] = 5.069812301953e-02;
C[7063] = -7.723317923518e-04;
C[7064] = -4.579839675382e-04;
C[7065] = 1.308661471001e-01;
C[7066] = 6.237265385437e-04;
C[7067] = 1.931879793043e-04;
C[7068] = 3.079890218007e-02;
C[7069] = -3.786720472889e-04;
C[7070] = -2.097184502689e-04;
C[7071] = 2.435589159077e-03;
C[7072] = -4.500189140411e-05;
C[7073] = -2.371447464190e-05;
C[7074] = 2.373623073533e-03;
C[7075] = -6.091289639156e-05;
C[7076] = -2.648006538070e-05;
C[7077] = -2.042819473738e-05;
C[7078] = -1.493647985200e-06;
C[7079] = -7.184045954940e-07;
C[7080] = 2.975879719942e-03;
C[7081] = 2.760649355815e-02;
C[7082] = 6.402899116217e-03;
C[7083] = -2.896420436978e-03;
C[7084] = 3.975231532159e-01;
C[7085] = 1.013727895410e-02;
C[7086] = -7.519911825473e-04;
C[7087] = 5.522919815966e-02;
C[7088] = 3.748857235071e-03;
C[7089] = 6.338201707518e-04;
C[7090] = 1.264424168038e-01;
C[7091] = -3.409265719209e-05;
C[7092] = -3.841093944928e-04;
C[7093] = 3.261882069874e-02;
C[7094] = 9.760780526574e-04;
C[7095] = -4.288209321782e-05;
C[7096] = 2.685487916079e-03;
C[7097] = 1.977813271364e-04;
C[7098] = -5.518159714116e-05;
C[7099] = 2.703701216882e-03;
C[7100] = 2.792183178646e-04;
C[7101] = -1.277877127557e-06;
C[7102] = -1.526182274078e-05;
C[7103] = 2.190623743789e-06;
C[7104] = 2.376356129686e-03;
C[7105] = 6.538734507165e-03;
C[7106] = 6.219178248002e-02;
C[7107] = -1.777461605112e-03;
C[7108] = 1.053458641544e-02;
C[7109] = 3.820814839741e-01;
C[7110] = -4.176697109913e-04;
C[7111] = 3.609753193329e-03;
C[7112] = 5.445547624209e-02;
C[7113] = 1.923021611975e-04;
C[7114] = -2.503696446228e-05;
C[7115] = 1.299464989402e-01;
C[7116] = -2.167173177006e-04;
C[7117] = 9.967047166080e-04;
C[7118] = 3.105450554721e-02;
C[7119] = -2.028880639817e-05;
C[7120] = 1.885051088339e-04;
C[7121] = 2.640954506891e-03;
C[7122] = -1.734135052249e-05;
C[7123] = 2.546704990317e-04;
C[7124] = 2.743801361126e-03;
C[7125] = -4.221239200975e-07;
C[7126] = 1.642073831120e-06;
C[7127] = -1.813358951979e-05;
C[7128] = 3.820047665727e-01;
C[7129] = -3.933488234415e-03;
C[7130] = -2.225470123572e-03;
C[7131] = 1.233941048400e-01;
C[7132] = 3.356263031190e-03;
C[7133] = 1.395606241890e-03;
C[7134] = 3.820069660447e-01;
C[7135] = 8.540279946306e-04;
C[7136] = 4.257203571057e-04;
C[7137] = 3.087063596638e-02;
C[7138] = -4.532472476553e-04;
C[7139] = -2.465845522646e-04;
C[7140] = 1.331982506856e-01;
C[7141] = 2.904409022087e-04;
C[7142] = -6.970404558466e-05;
C[7143] = 3.104534832178e-02;
C[7144] = 1.780330690417e-04;
C[7145] = 2.834678172727e-04;
C[7146] = 5.819877670782e-02;
C[7147] = 5.297070489120e-04;
C[7148] = 1.034616246672e-03;
C[7149] = 2.530765740087e-03;
C[7150] = 3.131364059480e-05;
C[7151] = 6.445270217881e-05;
C[7152] = -3.974217576146e-03;
C[7153] = 3.989904256483e-01;
C[7154] = 1.180913382773e-02;
C[7155] = 3.314542990966e-03;
C[7156] = 9.538159115745e-02;
C[7157] = -2.151740756797e-02;
C[7158] = 8.878885379126e-04;
C[7159] = 3.839564606260e-01;
C[7160] = 7.832233080825e-03;
C[7161] = -4.544306446842e-04;
C[7162] = 3.271663886089e-02;
C[7163] = 1.081674544094e-03;
C[7164] = 2.876595661937e-04;
C[7165] = 1.306843776838e-01;
C[7166] = -2.018799600733e-03;
C[7167] = 1.799274739060e-04;
C[7168] = 3.127810363432e-02;
C[7169] = 8.297817623236e-04;
C[7170] = 5.235850494284e-04;
C[7171] = 5.899428495327e-02;
C[7172] = 2.694422609055e-03;
C[7173] = 3.081020715855e-05;
C[7174] = 2.570832415746e-03;
C[7175] = 1.321131804944e-04;
C[7176] = -2.177120441676e-03;
C[7177] = 1.139820651765e-02;
C[7178] = 3.819777073283e-01;
C[7179] = 1.317740634001e-03;
C[7180] = -2.150295698450e-02;
C[7181] = 9.936081166133e-02;
C[7182] = 5.021645231537e-04;
C[7183] = 8.021000240715e-03;
C[7184] = 3.975796028346e-01;
C[7185] = -2.433502786959e-04;
C[7186] = 1.060139890227e-03;
C[7187] = 3.104758710416e-02;
C[7188] = -7.489508611060e-05;
C[7189] = -2.017836228502e-03;
C[7190] = 1.310033253496e-01;
C[7191] = 2.887240595098e-04;
C[7192] = 8.401354263987e-04;
C[7193] = 3.265685858519e-02;
C[7194] = 1.008011239066e-03;
C[7195] = 2.673262241875e-03;
C[7196] = 6.444090491743e-02;
C[7197] = 6.239051199562e-05;
C[7198] = 1.304071586808e-04;
C[7199] = 2.829300534171e-03;
C[7200] = 5.097458073651e-02;
C[7201] = -1.988130078397e-04;
C[7202] = 3.968673985970e-04;
C[7203] = 3.818493800148e-01;
C[7204] = 5.377751352030e-04;
C[7205] = 2.630528112047e-04;
C[7206] = 5.460938372487e-02;
C[7207] = -3.160458605383e-04;
C[7208] = -1.110603818081e-03;
C[7209] = 2.454019673542e-03;
C[7210] = -6.003548237309e-06;
C[7211] = 3.401366757439e-05;
C[7212] = 3.103484258645e-02;
C[7213] = 1.565865803920e-04;
C[7214] = 2.727832787132e-04;
C[7215] = 1.292198883822e-01;
C[7216] = -1.629168918870e-04;
C[7217] = -3.825170201399e-04;
C[7218] = 5.103300817381e-01;
C[7219] = -5.306839718812e-05;
C[7220] = 3.982932058575e-04;
C[7221] = 3.731439910381e-02;
C[7222] = 8.228240384316e-06;
C[7223] = 7.370472260660e-05;
C[7224] = -2.512871855580e-04;
C[7225] = 5.506828611338e-02;
C[7226] = 3.613476226401e-03;
C[7227] = 6.582992134433e-04;
C[7228] = 3.839686704317e-01;
C[7229] = 8.189613691772e-03;
C[7230] = -3.834929718754e-04;
C[7231] = 4.702737460336e-02;
C[7232] = -1.585387807733e-02;
C[7233] = -1.026558428452e-05;
C[7234] = 2.674760446327e-03;
C[7235] = 1.887533110386e-04;
C[7236] = 1.649844836292e-04;
C[7237] = 3.127891762137e-02;
C[7238] = 8.513763231358e-04;
C[7239] = -1.674133659761e-04;
C[7240] = 1.286648861331e-01;
C[7241] = -1.399580980728e-03;
C[7242] = -5.065426079312e-05;
C[7243] = 5.112644925431e-01;
C[7244] = 4.005008037215e-03;
C[7245] = 8.668008103428e-06;
C[7246] = 3.740330010798e-02;
C[7247] = 3.628295177071e-04;
C[7248] = 3.583481835765e-04;
C[7249] = 3.752641901894e-03;
C[7250] = 5.470939378246e-02;
C[7251] = 4.278712765568e-04;
C[7252] = 7.944464299780e-03;
C[7253] = 3.971993833215e-01;
C[7254] = -1.253501166549e-03;
C[7255] = -1.581263598701e-02;
C[7256] = 1.872120576715e-02;
C[7257] = 3.070766876575e-05;
C[7258] = 1.980336382580e-04;
C[7259] = 2.657882342916e-03;
C[7260] = 2.836112119028e-04;
C[7261] = 8.372638435873e-04;
C[7262] = 3.263151061765e-02;
C[7263] = -3.920435100378e-04;
C[7264] = -1.396831508039e-03;
C[7265] = 1.259820350930e-01;
C[7266] = 4.218243357540e-04;
C[7267] = 4.046559165526e-03;
C[7268] = 5.267910595334e-01;
C[7269] = 7.590555461567e-05;
C[7270] = 3.646434830246e-04;
C[7271] = 3.874382665432e-02;
C[7272] = -6.995897900602e+00;
C[7273] = 2.055611057459e-02;
C[7274] = 3.397422221648e-03;
C[7275] = 2.142191243882e+00;
C[7276] = 3.945965394286e-03;
C[7277] = -6.121827423367e-03;
C[7278] = 7.860787095289e-01;
C[7279] = -1.221177284957e-02;
C[7280] = -7.306602722358e-03;
C[7281] = 7.293749270614e-02;
C[7282] = 1.107093106673e-02;
C[7283] = 1.696181909040e-03;
C[7284] = 3.782207023470e-01;
C[7285] = -4.851330957919e-03;
C[7286] = -3.098006960466e-03;
C[7287] = 5.051106003626e-02;
C[7288] = -7.681426901040e-04;
C[7289] = -2.313016564187e-04;
C[7290] = 7.728524318787e-02;
C[7291] = -1.103555371313e-03;
C[7292] = 1.788156332520e-04;
C[7293] = 2.379207308661e-03;
C[7294] = -3.239444254980e-05;
C[7295] = 3.114585169349e-05;
C[7296] = 2.358420023701e-02;
C[7297] = -7.196601575708e+00;
C[7298] = 4.309150273170e-02;
C[7299] = 2.473013197916e-03;
C[7300] = 2.112044004972e+00;
C[7301] = 9.013423340151e-03;
C[7302] = -1.183017051845e-02;
C[7303] = 8.481166536099e-01;
C[7304] = 5.369108456297e-02;
C[7305] = 1.122233554985e-02;
C[7306] = 2.146840413708e-02;
C[7307] = -4.166706947630e-03;
C[7308] = -4.966589172164e-03;
C[7309] = 3.961452852344e-01;
C[7310] = 1.040548617624e-02;
C[7311] = -7.099359914697e-04;
C[7312] = 5.496803305382e-02;
C[7313] = 3.787436500797e-03;
C[7314] = -9.358464622141e-04;
C[7315] = 8.532351056924e-02;
C[7316] = 7.799252024627e-03;
C[7317] = -2.374985594001e-05;
C[7318] = 2.681988609594e-03;
C[7319] = 2.814688897491e-04;
C[7320] = 3.131676789621e-03;
C[7321] = 4.580821055065e-02;
C[7322] = -7.044836886401e+00;
C[7323] = -8.036423791038e-03;
C[7324] = 1.459661957993e-02;
C[7325] = 2.106547086348e+00;
C[7326] = -6.622996497215e-03;
C[7327] = 5.215852338319e-02;
C[7328] = 8.369077974927e-01;
C[7329] = 1.682894637439e-03;
C[7330] = -4.030871556683e-03;
C[7331] = 6.708240389741e-02;
C[7332] = -3.241982507790e-03;
C[7333] = 1.079746502615e-02;
C[7334] = 3.794105071780e-01;
C[7335] = -1.581271469236e-04;
C[7336] = 3.638824886584e-03;
C[7337] = 5.424210624204e-02;
C[7338] = 3.702962442833e-04;
C[7339] = 7.220048114740e-03;
C[7340] = 8.753567108854e-02;
C[7341] = 3.907180442620e-05;
C[7342] = 2.565294923530e-04;
C[7343] = 2.746226131037e-03;
C[7344] = 2.163711379776e+00;
C[7345] = -1.826834275083e-02;
C[7346] = -1.699659416024e-02;
C[7347] = -6.208140520020e+00;
C[7348] = 8.634607337001e-03;
C[7349] = -1.847705884682e-03;
C[7350] = 2.135012132945e+00;
C[7351] = 1.483848433444e-02;
C[7352] = 8.331229331335e-03;
C[7353] = 3.792967091417e-01;
C[7354] = -6.003656969602e-03;
C[7355] = -3.689991026250e-03;
C[7356] = 1.266948742046e-01;
C[7357] = 3.014288822042e-03;
C[7358] = -2.938822748525e-03;
C[7359] = 3.811816139106e-01;
C[7360] = 3.630622748103e-03;
C[7361] = 5.799768250254e-03;
C[7362] = 9.848137542422e-01;
C[7363] = 1.102180696456e-02;
C[7364] = 2.111642111003e-02;
C[7365] = 5.806766387509e-02;
C[7366] = 6.840981685257e-04;
C[7367] = 1.396168975884e-03;
C[7368] = -1.878161382428e-02;
C[7369] = 2.141389453620e+00;
C[7370] = 4.008756782575e-02;
C[7371] = 7.800206532503e-03;
C[7372] = -6.277806091736e+00;
C[7373] = -4.926065979329e-02;
C[7374] = 1.561446637318e-02;
C[7375] = 2.086848591140e+00;
C[7376] = -1.969323592720e-02;
C[7377] = -5.987709918847e-03;
C[7378] = 3.976125576668e-01;
C[7379] = 1.207201243844e-02;
C[7380] = 2.972568781818e-03;
C[7381] = 9.978555640181e-02;
C[7382] = -2.213781157176e-02;
C[7383] = 3.634920897515e-03;
C[7384] = 3.807329214051e-01;
C[7385] = 7.487930176820e-03;
C[7386] = 1.075637049195e-02;
C[7387] = 9.926690299519e-01;
C[7388] = 4.052939191716e-02;
C[7389] = 6.666784520362e-04;
C[7390] = 5.871448525063e-02;
C[7391] = 2.664668357266e-03;
C[7392] = -1.621137595147e-02;
C[7393] = 3.423197461118e-02;
C[7394] = 2.104471553433e+00;
C[7395] = -3.405018042463e-03;
C[7396] = -4.897164812386e-02;
C[7397] = -6.268598581079e+00;
C[7398] = 1.011092137351e-02;
C[7399] = -1.726119560212e-02;
C[7400] = 2.125145779192e+00;
C[7401] = -3.602484386871e-03;
C[7402] = 1.166641373979e-02;
C[7403] = 3.793067305323e-01;
C[7404] = -3.016688356414e-03;
C[7405] = -2.212336098829e-02;
C[7406] = 1.027991232588e-01;
C[7407] = 5.866674187960e-03;
C[7408] = 7.687166397470e-03;
C[7409] = 3.971072143110e-01;
C[7410] = 2.038239856016e-02;
C[7411] = 4.046894451984e-02;
C[7412] = 1.083526765504e+00;
C[7413] = 1.345219150545e-03;
C[7414] = 2.643768751688e-03;
C[7415] = 6.440290291975e-02;
C[7416] = 7.916078638685e-01;
C[7417] = -7.666070242962e-04;
C[7418] = 9.667745694551e-03;
C[7419] = 2.131860412349e+00;
C[7420] = 8.612198318991e-03;
C[7421] = 5.328687134530e-03;
C[7422] = -7.095800480393e+00;
C[7423] = -2.599788347845e-03;
C[7424] = -1.032570687552e-02;
C[7425] = 5.078751775324e-02;
C[7426] = -1.567578167621e-04;
C[7427] = 6.564099626647e-04;
C[7428] = 3.810240278808e-01;
C[7429] = 3.284807494806e-03;
C[7430] = 5.627562476011e-03;
C[7431] = 5.673501855872e-02;
C[7432] = -3.138093570836e-03;
C[7433] = -7.275749492575e-03;
C[7434] = 4.099368872643e+00;
C[7435] = -5.228264152447e-03;
C[7436] = -5.483434868278e-03;
C[7437] = 5.092443773537e-01;
C[7438] = 3.190473635898e-06;
C[7439] = 9.681898557237e-04;
C[7440] = -1.790880713697e-03;
C[7441] = 8.448984126842e-01;
C[7442] = 5.223298404462e-02;
C[7443] = 1.092390871069e-02;
C[7444] = 2.087092787255e+00;
C[7445] = -1.388892658098e-02;
C[7446] = -3.948730574587e-03;
C[7447] = -7.116436141246e+00;
C[7448] = -4.839640600092e-02;
C[7449] = -2.470980833101e-04;
C[7450] = 5.480712100754e-02;
C[7451] = 3.642547919655e-03;
C[7452] = 3.434893966916e-03;
C[7453] = 3.807451312108e-01;
C[7454] = 7.855779848527e-03;
C[7455] = -3.205540682174e-03;
C[7456] = 5.184127197504e-02;
C[7457] = -1.535042124051e-02;
C[7458] = -5.095397131995e-03;
C[7459] = 4.098830117775e+00;
C[7460] = 8.512790174254e-03;
C[7461] = 1.407639177591e-05;
C[7462] = 5.099433852858e-01;
C[7463] = 3.809324252722e-03;
C[7464] = 9.020040299936e-03;
C[7465] = 5.376677789944e-02;
C[7466] = 8.419861483001e-01;
C[7467] = 8.374247720358e-03;
C[7468] = -1.744861154809e-02;
C[7469] = 2.117541388931e+00;
C[7470] = -1.318365384488e-02;
C[7471] = -4.757156419435e-02;
C[7472] = -7.224489353109e+00;
C[7473] = 5.850304946960e-04;
C[7474] = 3.791221167621e-03;
C[7475] = 5.449602378241e-02;
C[7476] = 5.801919169705e-03;
C[7477] = 7.600161395776e-03;
C[7478] = 3.967269947979e-01;
C[7479] = -7.418646841043e-03;
C[7480] = -1.530917915018e-02;
C[7481] = 1.928687255208e-02;
C[7482] = -4.833561917590e-03;
C[7483] = 9.118549717317e-03;
C[7484] = 4.183338424868e+00;
C[7485] = 1.017250829480e-03;
C[7486] = 3.850096944357e-03;
C[7487] = 5.269337241478e-01;
C[7488] = 0.000000000000e+00;
C[7489] = 0.000000000000e+00;
C[7490] = 0.000000000000e+00;
C[7491] = 0.000000000000e+00;
C[7492] = 0.000000000000e+00;
C[7493] = 0.000000000000e+00;
C[7494] = 0.000000000000e+00;
C[7495] = 0.000000000000e+00;
C[7496] = 0.000000000000e+00;
C[7497] = 0.000000000000e+00;
C[7498] = 0.000000000000e+00;
C[7499] = 0.000000000000e+00;
C[7500] = 0.000000000000e+00;
C[7501] = 0.000000000000e+00;
C[7502] = 0.000000000000e+00;
C[7503] = 0.000000000000e+00;
C[7504] = 0.000000000000e+00;
C[7505] = 0.000000000000e+00;
C[7506] = 1.000000000000e+00;
C[7507] = 0.000000000000e+00;
C[7508] = 0.000000000000e+00;
C[7509] = 2.000000000000e-01;
C[7510] = -0.000000000000e+00;
C[7511] = -0.000000000000e+00;
C[7512] = 0.000000000000e+00;
C[7513] = 0.000000000000e+00;
C[7514] = 0.000000000000e+00;
C[7515] = 0.000000000000e+00;
C[7516] = 0.000000000000e+00;
C[7517] = 0.000000000000e+00;
C[7518] = 0.000000000000e+00;
C[7519] = 0.000000000000e+00;
C[7520] = 0.000000000000e+00;
C[7521] = 0.000000000000e+00;
C[7522] = 0.000000000000e+00;
C[7523] = 0.000000000000e+00;
C[7524] = 0.000000000000e+00;
C[7525] = 0.000000000000e+00;
C[7526] = 0.000000000000e+00;
C[7527] = 0.000000000000e+00;
C[7528] = 0.000000000000e+00;
C[7529] = 0.000000000000e+00;
C[7530] = 0.000000000000e+00;
C[7531] = 1.000000000000e+00;
C[7532] = 0.000000000000e+00;
C[7533] = -0.000000000000e+00;
C[7534] = 2.000000000000e-01;
C[7535] = -0.000000000000e+00;
C[7536] = 0.000000000000e+00;
C[7537] = 0.000000000000e+00;
C[7538] = 0.000000000000e+00;
C[7539] = 0.000000000000e+00;
C[7540] = 0.000000000000e+00;
C[7541] = 0.000000000000e+00;
C[7542] = 0.000000000000e+00;
C[7543] = 0.000000000000e+00;
C[7544] = 0.000000000000e+00;
C[7545] = 0.000000000000e+00;
C[7546] = 0.000000000000e+00;
C[7547] = 0.000000000000e+00;
C[7548] = 0.000000000000e+00;
C[7549] = 0.000000000000e+00;
C[7550] = 0.000000000000e+00;
C[7551] = 0.000000000000e+00;
C[7552] = 0.000000000000e+00;
C[7553] = 0.000000000000e+00;
C[7554] = 0.000000000000e+00;
C[7555] = 0.000000000000e+00;
C[7556] = 1.000000000000e+00;
C[7557] = -0.000000000000e+00;
C[7558] = -0.000000000000e+00;
C[7559] = 2.000000000000e-01;
C[7560] = 7.051181939262e-02;
C[7561] = 1.433121404203e-02;
C[7562] = 1.997963192304e-03;
C[7563] = 3.704690970273e-01;
C[7564] = -4.950970648742e-03;
C[7565] = -3.567033617857e-03;
C[7566] = 4.970015579673e-02;
C[7567] = 5.149945918605e-05;
C[7568] = 8.049666346501e-04;
C[7569] = 1.304867020703e-01;
C[7570] = 1.115304279132e-03;
C[7571] = 1.246234811136e-04;
C[7572] = 3.001159970984e-02;
C[7573] = -3.297796171960e-04;
C[7574] = -2.136343443767e-04;
C[7575] = 2.392794153305e-03;
C[7576] = 8.206560567558e-06;
C[7577] = 4.690464749333e-05;
C[7578] = 2.325515498871e-03;
C[7579] = 2.512495660496e-05;
C[7580] = 8.110801256211e-05;
C[7581] = -1.915375859397e-05;
C[7582] = 1.101599360624e-06;
C[7583] = 5.529543333946e-07;
C[7584] = 1.423079688636e-02;
C[7585] = 2.089380328795e-02;
C[7586] = -1.191308299688e-02;
C[7587] = -5.029852664415e-03;
C[7588] = 3.966509559107e-01;
C[7589] = 1.149416264405e-02;
C[7590] = 1.588373038424e-04;
C[7591] = 5.493128469474e-02;
C[7592] = 3.764489723312e-03;
C[7593] = 1.108609802088e-03;
C[7594] = 1.261303717146e-01;
C[7595] = -1.128670788181e-03;
C[7596] = -3.327370365090e-04;
C[7597] = 3.265888562121e-02;
C[7598] = 1.126414775389e-03;
C[7599] = 1.413529299584e-05;
C[7600] = 2.679850259371e-03;
C[7601] = 2.007573002653e-04;
C[7602] = 3.400952330152e-05;
C[7603] = 2.690648553610e-03;
C[7604] = 2.847254954068e-04;
C[7605] = 1.176276910650e-06;
C[7606] = -1.325498288500e-05;
C[7607] = 3.016823075752e-06;
C[7608] = 1.964707235036e-03;
C[7609] = -1.191854196698e-02;
C[7610] = 6.886650652643e-02;
C[7611] = -3.658239531287e-03;
C[7612] = 1.188023042909e-02;
C[7613] = 3.729130646009e-01;
C[7614] = 8.730283823817e-04;
C[7615] = 3.589066044549e-03;
C[7616] = 5.312241933166e-02;
C[7617] = 1.224064172957e-04;
C[7618] = -1.129034719521e-03;
C[7619] = 1.303220798709e-01;
C[7620] = -2.174471436137e-04;
C[7621] = 1.146950543170e-03;
C[7622] = 3.035864302545e-02;
C[7623] = 5.036866193070e-05;
C[7624] = 1.894440387402e-04;
C[7625] = 2.580421266104e-03;
C[7626] = 8.337358756657e-05;
C[7627] = 2.539564278777e-04;
C[7628] = 2.692965670094e-03;
C[7629] = 4.676517851643e-07;
C[7630] = 2.339073094841e-06;
C[7631] = -1.657579104991e-05;
C[7632] = 3.706589300620e-01;
C[7633] = -5.227430464701e-03;
C[7634] = -3.524390272480e-03;
C[7635] = 1.304202994981e-01;
C[7636] = 2.612419302329e-03;
C[7637] = -4.451277320229e-03;
C[7638] = 3.850950994676e-01;
C[7639] = 3.767462659040e-03;
C[7640] = 6.635591020784e-03;
C[7641] = 3.002425524549e-02;
C[7642] = -3.459088898614e-04;
C[7643] = -2.085238596932e-04;
C[7644] = 1.336811760329e-01;
C[7645] = 1.032766826970e-04;
C[7646] = -4.011367256495e-04;
C[7647] = 3.134779123950e-02;
C[7648] = 2.880525152353e-04;
C[7649] = 5.449588063578e-04;
C[7650] = 5.843648218229e-02;
C[7651] = 6.334934620779e-04;
C[7652] = 1.565280277687e-03;
C[7653] = 2.544916706551e-03;
C[7654] = 2.626380552389e-05;
C[7655] = 6.888271780266e-05;
C[7656] = -4.936535235237e-03;
C[7657] = 3.962747009973e-01;
C[7658] = 1.192034433488e-02;
C[7659] = 2.533902177275e-03;
C[7660] = 9.872406381643e-02;
C[7661] = -2.256446445910e-02;
C[7662] = 3.630152227407e-03;
C[7663] = 3.819894144767e-01;
C[7664] = 6.169094796189e-03;
C[7665] = -3.288172562956e-04;
C[7666] = 3.263380196032e-02;
C[7667] = 1.149624803556e-03;
C[7668] = 9.804220769337e-05;
C[7669] = 1.306550605842e-01;
C[7670] = -2.062529966664e-03;
C[7671] = 2.806687889278e-04;
C[7672] = 3.131006897470e-02;
C[7673] = 6.906299482247e-04;
C[7674] = 6.254212997958e-04;
C[7675] = 5.873342044251e-02;
C[7676] = 2.312943875673e-03;
C[7677] = 2.641187587832e-05;
C[7678] = 2.573408833065e-03;
C[7679] = 1.151606247483e-04;
C[7680] = -3.342809962812e-03;
C[7681] = 1.156606520388e-02;
C[7682] = 3.729231042568e-01;
C[7683] = -4.480596598143e-03;
C[7684] = -2.259235882148e-02;
C[7685] = 1.088352453331e-01;
C[7686] = 6.562965469929e-03;
C[7687] = 6.433563425349e-03;
C[7688] = 3.978753291966e-01;
C[7689] = -1.986861007070e-04;
C[7690] = 1.131208279378e-03;
C[7691] = 3.035931233584e-02;
C[7692] = -4.030913441771e-04;
C[7693] = -2.064389590822e-03;
C[7694] = 1.316651954263e-01;
C[7695] = 5.407063190823e-04;
C[7696] = 7.055178944424e-04;
C[7697] = 3.270827301698e-02;
C[7698] = 1.528385003859e-03;
C[7699] = 2.284278636382e-03;
C[7700] = 6.460349158788e-02;
C[7701] = 6.774458236654e-05;
C[7702] = 1.127563112069e-04;
C[7703] = 2.838238006278e-03;
C[7704] = 4.990335889761e-02;
C[7705] = -4.079191216526e-04;
C[7706] = 4.502873896553e-04;
C[7707] = 3.850926276640e-01;
C[7708] = 4.066997111599e-03;
C[7709] = 6.572677616426e-03;
C[7710] = 5.582249305759e-02;
C[7711] = -3.904666394153e-03;
C[7712] = -1.003464750998e-02;
C[7713] = 2.406341026697e-03;
C[7714] = -2.364846870384e-05;
C[7715] = 2.218592908227e-05;
C[7716] = 3.134762645260e-02;
C[7717] = 3.097917812073e-04;
C[7718] = 5.413537955154e-04;
C[7719] = 1.291823587470e-01;
C[7720] = -3.336331543361e-04;
C[7721] = -8.867509305558e-04;
C[7722] = 5.068600915353e-01;
C[7723] = 3.267625159615e-04;
C[7724] = 2.986888954240e-03;
C[7725] = 3.708340017273e-02;
C[7726] = 4.727625315773e-05;
C[7727] = 3.232939377689e-04;
C[7728] = -4.975152058320e-04;
C[7729] = 5.501768410281e-02;
C[7730] = 3.598407002778e-03;
C[7731] = 4.160822994421e-03;
C[7732] = 3.823116761183e-01;
C[7733] = 6.495995173619e-03;
C[7734] = -3.900079851417e-03;
C[7735] = 5.313967320761e-02;
C[7736] = -1.361490100347e-02;
C[7737] = -2.839441710031e-05;
C[7738] = 2.685610219908e-03;
C[7739] = 1.900667692888e-04;
C[7740] = 3.142765375941e-04;
C[7741] = 3.133155308413e-02;
C[7742] = 7.096800109937e-04;
C[7743] = -3.333273848204e-04;
C[7744] = 1.289431628910e-01;
C[7745] = -1.184713088235e-03;
C[7746] = 3.285490115983e-04;
C[7747] = 5.067085062097e-01;
C[7748] = 3.311318165491e-03;
C[7749] = 4.729956742519e-05;
C[7750] = 3.705343235684e-02;
C[7751] = 2.829905710274e-04;
C[7752] = 3.749132911489e-04;
C[7753] = 3.760469418951e-03;
C[7754] = 5.324607024910e-02;
C[7755] = 6.593472276584e-03;
C[7756] = 6.241194126607e-03;
C[7757] = 3.979435578606e-01;
C[7758] = -1.001463501625e-02;
C[7759] = -1.360496717256e-02;
C[7760] = 1.950483844158e-02;
C[7761] = 1.823442459325e-05;
C[7762] = 2.004892799746e-04;
C[7763] = 2.588664660601e-03;
C[7764] = 5.421508900779e-04;
C[7765] = 6.954365702525e-04;
C[7766] = 3.271282159458e-02;
C[7767] = -8.854167643073e-04;
C[7768] = -1.184050832841e-03;
C[7769] = 1.259868373866e-01;
C[7770] = 3.030655550844e-03;
C[7771] = 3.363094729018e-03;
C[7772] = 5.267787003367e-01;
C[7773] = 3.251798794923e-04;
C[7774] = 2.854154937074e-04;
C[7775] = 3.872756124341e-02;
C[7776] = -7.053596552303e+00;
C[7777] = 5.704891038125e-02;
C[7778] = -5.925545784002e-04;
C[7779] = 2.097877130377e+00;
C[7780] = 4.988239421809e-03;
C[7781] = -1.921947771946e-02;
C[7782] = 7.724095880636e-01;
C[7783] = 4.366078871849e-03;
C[7784] = 1.248436885608e-02;
C[7785] = 6.898245186031e-02;
C[7786] = 1.174465984747e-02;
C[7787] = 4.439174723178e-04;
C[7788] = 3.696846502062e-01;
C[7789] = -1.307260595888e-03;
C[7790] = -1.912210565958e-03;
C[7791] = 4.971743417391e-02;
C[7792] = 2.206987886157e-04;
C[7793] = 8.220664577883e-04;
C[7794] = 7.568085767753e-02;
C[7795] = 3.894882984263e-04;
C[7796] = 1.937644514267e-03;
C[7797] = 2.337106303845e-03;
C[7798] = 4.891853771014e-06;
C[7799] = 5.737657540841e-05;
C[7800] = 5.504056726778e-02;
C[7801] = -7.214145474655e+00;
C[7802] = -2.079536186041e-02;
C[7803] = 4.418406379307e-03;
C[7804] = 2.120531369265e+00;
C[7805] = 5.175896963076e-03;
C[7806] = 5.530480749037e-03;
C[7807] = 8.461830384140e-01;
C[7808] = 5.264243538814e-02;
C[7809] = 1.164424269180e-02;
C[7810] = 2.077059936000e-02;
C[7811] = -1.204850988079e-02;
C[7812] = -1.396186907240e-03;
C[7813] = 3.977700961601e-01;
C[7814] = 1.140316018121e-02;
C[7815] = 3.036484641784e-04;
C[7816] = 5.510672285630e-02;
C[7817] = 3.758811873083e-03;
C[7818] = 5.803850125591e-04;
C[7819] = 8.513674931573e-02;
C[7820] = 7.689346283023e-03;
C[7821] = 1.191527124539e-05;
C[7822] = 2.704858517336e-03;
C[7823] = 2.843523533929e-04;
C[7824] = -1.257673723759e-03;
C[7825] = -2.090454126238e-02;
C[7826] = -7.047476327558e+00;
C[7827] = -2.012079990947e-02;
C[7828] = 1.064046221889e-02;
C[7829] = 2.079305481795e+00;
C[7830] = 1.311866638893e-02;
C[7831] = 5.064124089753e-02;
C[7832] = 8.198617185901e-01;
C[7833] = 4.106615150498e-04;
C[7834] = -1.205396885089e-02;
C[7835] = 6.774333572056e-02;
C[7836] = -1.987473582015e-03;
C[7837] = 1.178151588417e-02;
C[7838] = 3.728835079131e-01;
C[7839] = 8.680089512524e-04;
C[7840] = 3.579011109499e-03;
C[7841] = 5.318467086892e-02;
C[7842] = 2.005136319494e-03;
C[7843] = 6.972122623589e-03;
C[7844] = 8.595377061200e-02;
C[7845] = 6.032019739341e-05;
C[7846] = 2.542121122878e-04;
C[7847] = 2.703108413242e-03;
C[7848] = 2.101673791070e+00;
C[7849] = 4.668503735861e-04;
C[7850] = -1.744381473332e-02;
C[7851] = -6.194471536896e+00;
C[7852] = 5.982330433228e-03;
C[7853] = -1.155271459358e-02;
C[7854] = 2.161347930541e+00;
C[7855] = 1.013078746774e-02;
C[7856] = 8.053586270683e-03;
C[7857] = 3.698744832409e-01;
C[7858] = -1.593764707526e-03;
C[7859] = -1.853624323207e-03;
C[7860] = 1.302384520336e-01;
C[7861] = -1.742359945258e-04;
C[7862] = -4.351411258304e-03;
C[7863] = 3.859497634240e-01;
C[7864] = 2.800093588643e-03;
C[7865] = 4.786451276352e-03;
C[7866] = 9.907193608232e-01;
C[7867] = 9.825140299674e-03;
C[7868] = 2.393842824685e-02;
C[7869] = 5.844903331762e-02;
C[7870] = 6.096130863285e-04;
C[7871] = 1.566101934556e-03;
C[7872] = 5.276947691913e-03;
C[7873] = 2.113006270998e+00;
C[7874] = 1.144274033475e-02;
C[7875] = 4.411987932141e-03;
C[7876] = -6.278592360173e+00;
C[7877] = -5.365168944379e-02;
C[7878] = 8.236231808788e-03;
C[7879] = 2.091939871375e+00;
C[7880] = -2.479893781092e-02;
C[7881] = -1.292825182382e-03;
C[7882] = 3.973938412467e-01;
C[7883] = 1.182162978997e-02;
C[7884] = -2.527531195802e-04;
C[7885] = 9.580432385544e-02;
C[7886] = -2.227221229122e-02;
C[7887] = 2.677920633133e-03;
C[7888] = 3.837820436561e-01;
C[7889] = 6.014274757716e-03;
C[7890] = 9.672706988169e-03;
C[7891] = 9.922762246004e-01;
C[7892] = 3.429020441766e-02;
C[7893] = 6.109681950403e-04;
C[7894] = 5.874783175131e-02;
C[7895] = 2.260070483776e-03;
C[7896] = -1.473500461854e-02;
C[7897] = 6.613948159676e-03;
C[7898] = 2.079506274913e+00;
C[7899] = -1.213910015187e-02;
C[7900] = -5.420957669137e-02;
C[7901] = -6.250437403911e+00;
C[7902] = 7.011522750703e-03;
C[7903] = -2.112747241984e-02;
C[7904] = 2.138116402033e+00;
C[7905] = -1.687986910912e-03;
C[7906] = 1.147506274104e-02;
C[7907] = 3.728935475690e-01;
C[7908] = -4.380730536219e-03;
C[7909] = -2.230010665360e-02;
C[7910] = 1.077728362647e-01;
C[7911] = 4.721833143929e-03;
C[7912] = 6.289106884933e-03;
C[7913] = 3.988539674998e-01;
C[7914] = 2.345438041971e-02;
C[7915] = 3.410918481640e-02;
C[7916] = 1.088133670830e+00;
C[7917] = 1.544803768965e-03;
C[7918] = 2.225218950556e-03;
C[7919] = 6.464255369481e-02;
C[7920] = 7.764736500812e-01;
C[7921] = -5.804647760864e-03;
C[7922] = 4.663846534406e-03;
C[7923] = 2.161298494469e+00;
C[7924] = 1.697312949264e-02;
C[7925] = 7.205765680644e-03;
C[7926] = -7.085266401038e+00;
C[7927] = -8.123548183378e-03;
C[7928] = -3.127431500887e-02;
C[7929] = 4.992063727479e-02;
C[7930] = -2.631079613167e-04;
C[7931] = 4.452679585260e-04;
C[7932] = 3.859472916204e-01;
C[7933] = 3.114765517325e-03;
C[7934] = 4.731545290426e-03;
C[7935] = 5.547358847904e-02;
C[7936] = -3.308037723883e-03;
C[7937] = -9.778654603820e-03;
C[7938] = 4.072209436624e+00;
C[7939] = -1.180654931815e-03;
C[7940] = 1.842550652864e-02;
C[7941] = 5.063488599070e-01;
C[7942] = 4.976434887813e-04;
C[7943] = 4.555432577110e-03;
C[7944] = -6.614214428513e-03;
C[7945] = 8.479110265753e-01;
C[7946] = 5.082806006210e-02;
C[7947] = 1.799799417536e-02;
C[7948] = 2.098385104206e+00;
C[7949] = -1.987883745444e-02;
C[7950] = -8.031817328669e-03;
C[7951] = -7.097255556751e+00;
C[7952] = -4.062320917841e-02;
C[7953] = -3.283158764024e-04;
C[7954] = 5.519312226436e-02;
C[7955] = 3.588352067728e-03;
C[7956] = 3.193453924024e-03;
C[7957] = 3.841043052976e-01;
C[7958] = 6.351538633203e-03;
C[7959] = -3.303451181148e-03;
C[7960] = 5.257321416235e-02;
C[7961] = -1.278426342948e-02;
C[7962] = -1.191843696474e-03;
C[7963] = 4.061830049298e+00;
C[7964] = 4.037222675087e-03;
C[7965] = 4.921709170620e-04;
C[7966] = 5.053023727504e-01;
C[7967] = 2.636205038268e-03;
C[7968] = 3.883301986058e-03;
C[7969] = 5.256202930093e-02;
C[7970] = 8.223347369390e-01;
C[7971] = 7.211211386698e-03;
C[7972] = -2.335695120256e-02;
C[7973] = 2.139480975315e+00;
C[7974] = -3.087406513431e-02;
C[7975] = -4.042453256008e-02;
C[7976] = -7.222045691700e+00;
C[7977] = 3.920131142870e-04;
C[7978] = 3.754791568722e-03;
C[7979] = 5.330832178636e-02;
C[7980] = 4.744332532153e-03;
C[7981] = 6.086374088134e-03;
C[7982] = 3.989221961639e-01;
C[7983] = -9.758642110092e-03;
C[7984] = -1.277432959856e-02;
C[7985] = 1.876219651085e-02;
C[7986] = 1.901058329070e-02;
C[7987] = 4.853596605015e-03;
C[7988] = 4.179720569998e+00;
C[7989] = 4.583295989776e-03;
C[7990] = 2.693796765259e-03;
C[7991] = 5.264840913641e-01;
C[7992] = 0.000000000000e+00;
C[7993] = 0.000000000000e+00;
C[7994] = 0.000000000000e+00;
C[7995] = 0.000000000000e+00;
C[7996] = 0.000000000000e+00;
C[7997] = 0.000000000000e+00;
C[7998] = 0.000000000000e+00;
C[7999] = 0.000000000000e+00;
C[8000] = 0.000000000000e+00;
C[8001] = 0.000000000000e+00;
C[8002] = 0.000000000000e+00;
C[8003] = 0.000000000000e+00;
C[8004] = 0.000000000000e+00;
C[8005] = 0.000000000000e+00;
C[8006] = 0.000000000000e+00;
C[8007] = 0.000000000000e+00;
C[8008] = 0.000000000000e+00;
C[8009] = 0.000000000000e+00;
C[8010] = 1.000000000000e+00;
C[8011] = 0.000000000000e+00;
C[8012] = 0.000000000000e+00;
C[8013] = 2.000000000000e-01;
C[8014] = -0.000000000000e+00;
C[8015] = -0.000000000000e+00;
C[8016] = 0.000000000000e+00;
C[8017] = 0.000000000000e+00;
C[8018] = 0.000000000000e+00;
C[8019] = 0.000000000000e+00;
C[8020] = 0.000000000000e+00;
C[8021] = 0.000000000000e+00;
C[8022] = 0.000000000000e+00;
C[8023] = 0.000000000000e+00;
C[8024] = 0.000000000000e+00;
C[8025] = 0.000000000000e+00;
C[8026] = 0.000000000000e+00;
C[8027] = 0.000000000000e+00;
C[8028] = 0.000000000000e+00;
C[8029] = 0.000000000000e+00;
C[8030] = 0.000000000000e+00;
C[8031] = 0.000000000000e+00;
C[8032] = 0.000000000000e+00;
C[8033] = 0.000000000000e+00;
C[8034] = 0.000000000000e+00;
C[8035] = 1.000000000000e+00;
C[8036] = 0.000000000000e+00;
C[8037] = -0.000000000000e+00;
C[8038] = 2.000000000000e-01;
C[8039] = -0.000000000000e+00;
C[8040] = 0.000000000000e+00;
C[8041] = 0.000000000000e+00;
C[8042] = 0.000000000000e+00;
C[8043] = 0.000000000000e+00;
C[8044] = 0.000000000000e+00;
C[8045] = 0.000000000000e+00;
C[8046] = 0.000000000000e+00;
C[8047] = 0.000000000000e+00;
C[8048] = 0.000000000000e+00;
C[8049] = 0.000000000000e+00;
C[8050] = 0.000000000000e+00;
C[8051] = 0.000000000000e+00;
C[8052] = 0.000000000000e+00;
C[8053] = 0.000000000000e+00;
C[8054] = 0.000000000000e+00;
C[8055] = 0.000000000000e+00;
C[8056] = 0.000000000000e+00;
C[8057] = 0.000000000000e+00;
C[8058] = 0.000000000000e+00;
C[8059] = 0.000000000000e+00;
C[8060] = 1.000000000000e+00;
C[8061] = -0.000000000000e+00;
C[8062] = -0.000000000000e+00;
C[8063] = 2.000000000000e-01;
C[8064] = 6.462130758407e-02;
C[8065] = 1.171363659653e-02;
C[8066] = -2.104362530121e-03;
C[8067] = 3.777966377585e-01;
C[8068] = 5.853136660641e-04;
C[8069] = -1.309646161951e-03;
C[8070] = 5.159766462126e-02;
C[8071] = 5.446829986951e-05;
C[8072] = 2.297310826681e-04;
C[8073] = 1.300209425810e-01;
C[8074] = 9.950048931769e-04;
C[8075] = -2.488731785197e-04;
C[8076] = 3.092846685349e-02;
C[8077] = -5.724018391621e-05;
C[8078] = -1.127219377441e-04;
C[8079] = 2.504689111137e-03;
C[8080] = -3.795274443345e-06;
C[8081] = 1.178866709511e-05;
C[8082] = 2.473908297214e-03;
C[8083] = -1.168356263822e-05;
C[8084] = 5.357377687680e-05;
C[8085] = -1.636489120044e-05;
C[8086] = -1.408133754936e-06;
C[8087] = 2.386247934375e-07;
C[8088] = 1.173715376996e-02;
C[8089] = 2.275252640447e-02;
C[8090] = -6.174556585119e-03;
C[8091] = 4.221259969965e-04;
C[8092] = 3.986079025325e-01;
C[8093] = 1.089950765995e-02;
C[8094] = 7.974502602979e-05;
C[8095] = 5.567118819394e-02;
C[8096] = 3.470982680905e-03;
C[8097] = 9.965727047384e-04;
C[8098] = 1.264409350991e-01;
C[8099] = -2.770410001127e-04;
C[8100] = -6.556174434258e-05;
C[8101] = 3.277962680557e-02;
C[8102] = 9.750002097923e-04;
C[8103] = -2.020028392483e-06;
C[8104] = 2.714456199476e-03;
C[8105] = 1.800433885858e-04;
C[8106] = -5.110314480647e-06;
C[8107] = 2.699344063776e-03;
C[8108] = 2.353130732378e-04;
C[8109] = -1.253850314948e-06;
C[8110] = -1.535541773796e-05;
C[8111] = 1.107857693390e-06;
C[8112] = -2.061280720384e-03;
C[8113] = -6.271811530228e-03;
C[8114] = 6.307361754826e-02;
C[8115] = -1.396968021376e-03;
C[8116] = 1.121755133646e-02;
C[8117] = 3.800248765458e-01;
C[8118] = 2.493855779361e-04;
C[8119] = 3.347048242336e-03;
C[8120] = 5.517449891934e-02;
C[8121] = -2.460010578705e-04;
C[8122] = -2.835246631200e-04;
C[8123] = 1.296284386928e-01;
C[8124] = -1.174320259373e-04;
C[8125] = 9.916663885020e-04;
C[8126] = 3.122426361511e-02;
C[8127] = 1.320263313538e-05;
C[8128] = 1.725001003061e-04;
C[8129] = 2.698467227666e-03;
C[8130] = 5.961463429123e-05;
C[8131] = 2.137700788712e-04;
C[8132] = 2.850575537932e-03;
C[8133] = 4.109933109420e-07;
C[8134] = 7.477507875458e-07;
C[8135] = -1.342538125894e-05;
C[8136] = 3.768899726766e-01;
C[8137] = 6.968961412257e-04;
C[8138] = -8.129747642121e-04;
C[8139] = 1.232115323731e-01;
C[8140] = -6.205318598482e-04;
C[8141] = -2.476772250893e-03;
C[8142] = 3.840552826588e-01;
C[8143] = 1.526326772637e-03;
C[8144] = 2.055732341196e-03;
C[8145] = 3.086802251470e-02;
C[8146] = -4.724373472730e-05;
C[8147] = -7.849914212636e-05;
C[8148] = 1.327936518488e-01;
C[8149] = 2.986126207096e-05;
C[8150] = -1.107968933752e-04;
C[8151] = 3.127976440926e-02;
C[8152] = 9.713409626357e-05;
C[8153] = 1.377392993434e-04;
C[8154] = 5.833266838794e-02;
C[8155] = 2.859668976300e-04;
C[8156] = 1.108509643592e-03;
C[8157] = 2.541928260132e-03;
C[8158] = 1.045477830559e-05;
C[8159] = 4.581303105595e-05;
C[8160] = 7.813250156828e-04;
C[8161] = 3.978958214612e-01;
C[8162] = 1.030753866409e-02;
C[8163] = -5.621535227667e-04;
C[8164] = 9.414215846481e-02;
C[8165] = -2.065594025868e-02;
C[8166] = 1.450263632365e-03;
C[8167] = 3.850793387479e-01;
C[8168] = 6.884598700311e-03;
C[8169] = -4.417276060829e-05;
C[8170] = 3.273215473415e-02;
C[8171] = 9.309988770103e-04;
C[8172] = 3.375315120973e-05;
C[8173] = 1.303096213532e-01;
C[8174] = -1.857907834310e-03;
C[8175] = 9.278929952050e-05;
C[8176] = 3.156545200799e-02;
C[8177] = 7.570822642108e-04;
C[8178] = 3.054570614755e-04;
C[8179] = 5.897375759598e-02;
C[8180] = 2.261562025606e-03;
C[8181] = 1.166579682075e-05;
C[8182] = 2.584927411765e-03;
C[8183] = 1.137853384389e-04;
C[8184] = -7.553788381402e-04;
C[8185] = 9.979296802434e-03;
C[8186] = 3.798930061453e-01;
C[8187] = -2.421257693886e-03;
C[8188] = -2.062877564506e-02;
C[8189] = 1.001161553254e-01;
C[8190] = 1.964524158815e-03;
C[8191] = 7.121215879469e-03;
C[8192] = 3.979328501684e-01;
C[8193] = -7.577078282336e-05;
C[8194] = 9.136528192911e-04;
C[8195] = 3.121547225508e-02;
C[8196] = -1.070959229081e-04;
C[8197] = -1.856096860068e-03;
C[8198] = 1.307099869825e-01;
C[8199] = 1.321739013159e-04;
C[8200] = 7.695671196612e-04;
C[8201] = 3.271356691590e-02;
C[8202] = 1.137861771554e-03;
C[8203] = 2.205873032617e-03;
C[8204] = 6.451009942478e-02;
C[8205] = 4.765122823939e-05;
C[8206] = 1.102619433069e-04;
C[8207] = 2.834296332244e-03;
C[8208] = 5.089459627709e-02;
C[8209] = 1.053227300989e-04;
C[8210] = 1.413307776351e-04;
C[8211] = 3.845593702142e-01;
C[8212] = 1.327163623381e-03;
C[8213] = 1.989951845662e-03;
C[8214] = 5.483994006001e-02;
C[8215] = -2.274915827429e-03;
C[8216] = -8.356456426196e-03;
C[8217] = 2.457817888193e-03;
C[8218] = -3.148481212117e-07;
C[8219] = 5.998979781981e-06;
C[8220] = 3.131337024628e-02;
C[8221] = 8.458263225484e-05;
C[8222] = 1.338690804391e-04;
C[8223] = 1.290679985235e-01;
C[8224] = -1.609604507603e-04;
C[8225] = -6.016888927884e-04;
C[8226] = 5.072905307180e-01;
C[8227] = 9.262689598623e-04;
C[8228] = 6.219391957862e-03;
C[8229] = 3.717539510264e-02;
C[8230] = 7.772172042470e-05;
C[8231] = 4.636507151139e-04;
C[8232] = 9.280220971640e-05;
C[8233] = 5.572914589636e-02;
C[8234] = 3.435978936419e-03;
C[8235] = 1.352926935857e-03;
C[8236] = 3.849155202201e-01;
C[8237] = 6.894520927512e-03;
C[8238] = -2.251608697134e-03;
C[8239] = 5.128888792464e-02;
C[8240] = -1.255743293727e-02;
C[8241] = -1.239680453552e-06;
C[8242] = 2.718320046304e-03;
C[8243] = 1.784288132450e-04;
C[8244] = 8.557410714489e-05;
C[8245] = 3.155453077281e-02;
C[8246] = 7.544541228640e-04;
C[8247] = -1.594066420740e-04;
C[8248] = 1.286794685776e-01;
C[8249] = -1.161527649622e-03;
C[8250] = 8.995990034894e-04;
C[8251] = 5.052246152622e-01;
C[8252] = 2.027428433899e-03;
C[8253] = 7.634714141279e-05;
C[8254] = 3.705990084268e-02;
C[8255] = 2.227769029151e-04;
C[8256] = 1.317686634841e-04;
C[8257] = 3.574787584649e-03;
C[8258] = 5.452898714223e-02;
C[8259] = 1.995382815667e-03;
C[8260] = 6.674882606400e-03;
C[8261] = 3.984983410539e-01;
C[8262] = -8.306852261644e-03;
C[8263] = -1.258019872578e-02;
C[8264] = 1.830486465258e-02;
C[8265] = 5.257839149511e-06;
C[8266] = 1.869637155020e-04;
C[8267] = 2.655433109192e-03;
C[8268] = 1.337159976414e-04;
C[8269] = 7.431011912834e-04;
C[8270] = 3.275126630826e-02;
C[8271] = -5.983819484850e-04;
C[8272] = -1.163045368857e-03;
C[8273] = 1.259109172167e-01;
C[8274] = 6.181102050694e-03;
C[8275] = 2.105680395261e-03;
C[8276] = 5.260662254540e-01;
C[8277] = 4.614639738950e-04;
C[8278] = 2.266214367072e-04;
C[8279] = 3.869648661732e-02;
C[8280] = -7.048399402119e+00;
C[8281] = 6.030634977171e-02;
C[8282] = -2.070173849285e-02;
C[8283] = 2.141755791864e+00;
C[8284] = 1.704571285338e-02;
C[8285] = -1.806318662432e-02;
C[8286] = 8.045378495209e-01;
C[8287] = 8.273727487784e-04;
C[8288] = 1.800784984709e-03;
C[8289] = 6.590425056055e-02;
C[8290] = 1.221472364292e-02;
C[8291] = -4.238856336550e-03;
C[8292] = 3.832541970749e-01;
C[8293] = -5.283993709566e-04;
C[8294] = -1.964249259948e-03;
C[8295] = 5.189761812149e-02;
C[8296] = -6.243763553556e-05;
C[8297] = 9.344045182008e-05;
C[8298] = 7.914143613716e-02;
C[8299] = -9.471348285223e-05;
C[8300] = 1.321754660034e-03;
C[8301] = 2.451942607496e-03;
C[8302] = -1.600017037233e-05;
C[8303] = 4.070604326809e-05;
C[8304] = 6.077669324019e-02;
C[8305] = -7.191825388093e+00;
C[8306] = 2.623570054614e-02;
C[8307] = 1.471702482802e-02;
C[8308] = 2.131916630890e+00;
C[8309] = 2.395953093847e-02;
C[8310] = 1.040993258319e-03;
C[8311] = 8.565435680918e-01;
C[8312] = 4.836198602870e-02;
C[8313] = 1.223824081635e-02;
C[8314] = 2.661114970533e-02;
C[8315] = 2.350364970904e-03;
C[8316] = -6.606358472374e-04;
C[8317] = 3.988750486145e-01;
C[8318] = 9.891061645952e-03;
C[8319] = -4.222014196845e-05;
C[8320] = 5.574015956917e-02;
C[8321] = 3.388619127304e-03;
C[8322] = 4.620054456621e-05;
C[8323] = 8.559166247909e-02;
C[8324] = 6.618209873506e-03;
C[8325] = -1.021744690573e-05;
C[8326] = 2.705414922667e-03;
C[8327] = 2.363823539284e-04;
C[8328] = -1.984010229811e-02;
C[8329] = 2.429060164396e-02;
C[8330] = -7.058291414771e+00;
C[8331] = -1.935265026065e-02;
C[8332] = 2.848293323687e-02;
C[8333] = 2.105707073632e+00;
C[8334] = 1.983289500133e-03;
C[8335] = 4.724083509231e-02;
C[8336] = 8.530694063998e-01;
C[8337] = -4.195774526813e-03;
C[8338] = 2.253110025795e-03;
C[8339] = 5.894239532290e-02;
C[8340] = -2.047184762359e-03;
C[8341] = 1.021130109191e-02;
C[8342] = 3.847969105318e-01;
C[8343] = 1.108681180134e-04;
C[8344] = 3.281986334634e-03;
C[8345] = 5.545096880159e-02;
C[8346] = 1.460097916855e-03;
C[8347] = 6.099185459057e-03;
C[8348] = 8.962395770291e-02;
C[8349] = 4.671622002912e-05;
C[8350] = 2.147490452191e-04;
C[8351] = 2.829922962091e-03;
C[8352] = 2.123622490227e+00;
C[8353] = 2.021242771260e-02;
C[8354] = -7.672785117374e-03;
C[8355] = -6.221738575484e+00;
C[8356] = 2.390979677849e-03;
C[8357] = -5.126570712080e-03;
C[8358] = 2.134130433093e+00;
C[8359] = 4.782920158414e-03;
C[8360] = -5.700721939489e-03;
C[8361] = 3.823475319930e-01;
C[8362] = -3.858657030082e-04;
C[8363] = -1.463191505195e-03;
C[8364] = 1.178789329039e-01;
C[8365] = 1.396615192151e-03;
C[8366] = -6.243214508053e-05;
C[8367] = 3.839833969176e-01;
C[8368] = 6.774165509354e-04;
C[8369] = 2.988890078627e-04;
C[8370] = 9.880170582662e-01;
C[8371] = 3.990093519782e-03;
C[8372] = 1.615575812759e-02;
C[8373] = 5.842593981839e-02;
C[8374] = 2.316699969626e-04;
C[8375] = 9.513912815427e-04;
C[8376] = 2.096593984576e-02;
C[8377] = 2.117675009462e+00;
C[8378] = 1.028267978938e-02;
C[8379] = 3.558546419480e-03;
C[8380] = -6.286974511507e+00;
C[8381] = -4.440103862183e-02;
C[8382] = 3.638177293564e-03;
C[8383] = 2.099869866910e+00;
C[8384] = -1.415535982436e-02;
C[8385] = -3.323880213378e-04;
C[8386] = 3.981629675431e-01;
C[8387] = 9.301288419534e-03;
C[8388] = 1.454993529233e-03;
C[8389] = 9.291863449922e-02;
C[8390] = -1.945323079776e-02;
C[8391] = 6.070341746206e-04;
C[8392] = 3.861664196967e-01;
C[8393] = 7.243450564311e-03;
C[8394] = 4.256669002669e-03;
C[8395] = 9.961466282400e-01;
C[8396] = 3.435230317027e-02;
C[8397] = 2.528914942582e-04;
C[8398] = 5.913127197689e-02;
C[8399] = 2.292611808249e-03;
C[8400] = -6.977840148099e-03;
C[8401] = 5.555313788110e-03;
C[8402] = 2.103069665623e+00;
C[8403] = -4.016279571938e-03;
C[8404] = -4.385774634933e-02;
C[8405] = -6.277034065748e+00;
C[8406] = -7.262078145109e-03;
C[8407] = -1.104897920130e-02;
C[8408] = 2.113962515183e+00;
C[8409] = -1.409981936137e-03;
C[8410] = 8.970850788434e-03;
C[8411] = 3.846650401313e-01;
C[8412] = -6.917588073408e-06;
C[8413] = -1.942606618414e-02;
C[8414] = 9.561464660844e-02;
C[8415] = 2.032912201121e-04;
C[8416] = 7.457962916349e-03;
C[8417] = 3.974578194569e-01;
C[8418] = 1.663358796133e-02;
C[8419] = 3.370907898467e-02;
C[8420] = 1.084613756518e+00;
C[8421] = 9.867650403701e-04;
C[8422] = 2.238857793606e-03;
C[8423] = 6.452884457740e-02;
C[8424] = 7.904764826375e-01;
C[8425] = 1.552547339701e-03;
C[8426] = -1.778065058869e-04;
C[8427] = 2.144212184200e+00;
C[8428] = 1.176177113866e-03;
C[8429] = -6.753524408155e-03;
C[8430] = -7.094329627688e+00;
C[8431] = -6.152597012012e-03;
C[8432] = -3.167248393883e-02;
C[8433] = 5.119454977732e-02;
C[8434] = -1.664243789938e-05;
C[8435] = 2.813317712416e-06;
C[8436] = 3.844874844730e-01;
C[8437] = 4.839341656356e-04;
C[8438] = 2.287189069598e-04;
C[8439] = 5.356472189689e-02;
C[8440] = -1.346697278480e-03;
C[8441] = -5.772285637310e-03;
C[8442] = 4.081396915434e+00;
C[8443] = 5.919513245922e-03;
C[8444] = 3.827131349820e-02;
C[8445] = 5.082819354179e-01;
C[8446] = 9.731258898848e-04;
C[8447] = 5.537297447051e-03;
C[8448] = 1.594050945716e-03;
C[8449] = 8.577027221402e-01;
C[8450] = 4.901944897397e-02;
C[8451] = 1.314923422812e-03;
C[8452] = 2.096593496355e+00;
C[8453] = -1.558287824044e-02;
C[8454] = -5.686454406121e-03;
C[8455] = -7.103796168895e+00;
C[8456] = -3.655985387662e-02;
C[8457] = -2.410372568866e-05;
C[8458] = 5.579811727159e-02;
C[8459] = 3.370917028717e-03;
C[8460] = 5.040167141553e-04;
C[8461] = 3.860026011690e-01;
C[8462] = 7.231267964392e-03;
C[8463] = -1.323390148186e-03;
C[8464] = 4.871407236767e-02;
C[8465] = -1.311803224312e-02;
C[8466] = 5.482430018732e-03;
C[8467] = 4.060079291391e+00;
C[8468] = -1.626544898954e-03;
C[8469] = 9.453515124867e-04;
C[8470] = 5.066409556722e-01;
C[8471] = 2.321313045479e-03;
C[8472] = -1.584633989713e-04;
C[8473] = 5.043808410358e-02;
C[8474] = 8.401591708575e-01;
C[8475] = -6.907712450062e-03;
C[8476] = -1.834968170258e-02;
C[8477] = 2.125272332892e+00;
C[8478] = -3.068040064780e-02;
C[8479] = -3.701516964690e-02;
C[8480] = -7.224834530203e+00;
C[8481] = -4.521967363961e-06;
C[8482] = 3.492424031048e-03;
C[8483] = 5.480545702448e-02;
C[8484] = 2.385394823340e-04;
C[8485] = 7.033734470400e-03;
C[8486] = 3.980233103424e-01;
C[8487] = -5.722681472759e-03;
C[8488] = -1.314079803164e-02;
C[8489] = 1.825784433413e-02;
C[8490] = 3.758111095924e-02;
C[8491] = -3.927579015351e-04;
C[8492] = 4.176270032397e+00;
C[8493] = 5.492654559532e-03;
C[8494] = 2.397891322519e-03;
C[8495] = 5.262640715422e-01;
C[8496] = 0.000000000000e+00;
C[8497] = 0.000000000000e+00;
C[8498] = 0.000000000000e+00;
C[8499] = 0.000000000000e+00;
C[8500] = 0.000000000000e+00;
C[8501] = 0.000000000000e+00;
C[8502] = 0.000000000000e+00;
C[8503] = 0.000000000000e+00;
C[8504] = 0.000000000000e+00;
C[8505] = 0.000000000000e+00;
C[8506] = 0.000000000000e+00;
C[8507] = 0.000000000000e+00;
C[8508] = 0.000000000000e+00;
C[8509] = 0.000000000000e+00;
C[8510] = 0.000000000000e+00;
C[8511] = 0.000000000000e+00;
C[8512] = 0.000000000000e+00;
C[8513] = 0.000000000000e+00;
C[8514] = 1.000000000000e+00;
C[8515] = 0.000000000000e+00;
C[8516] = 0.000000000000e+00;
C[8517] = 2.000000000000e-01;
C[8518] = -0.000000000000e+00;
C[8519] = -0.000000000000e+00;
C[8520] = 0.000000000000e+00;
C[8521] = 0.000000000000e+00;
C[8522] = 0.000000000000e+00;
C[8523] = 0.000000000000e+00;
C[8524] = 0.000000000000e+00;
C[8525] = 0.000000000000e+00;
C[8526] = 0.000000000000e+00;
C[8527] = 0.000000000000e+00;
C[8528] = 0.000000000000e+00;
C[8529] = 0.000000000000e+00;
C[8530] = 0.000000000000e+00;
C[8531] = 0.000000000000e+00;
C[8532] = 0.000000000000e+00;
C[8533] = 0.000000000000e+00;
C[8534] = 0.000000000000e+00;
C[8535] = 0.000000000000e+00;
C[8536] = 0.000000000000e+00;
C[8537] = 0.000000000000e+00;
C[8538] = 0.000000000000e+00;
C[8539] = 1.000000000000e+00;
C[8540] = 0.000000000000e+00;
C[8541] = -0.000000000000e+00;
C[8542] = 2.000000000000e-01;
C[8543] = -0.000000000000e+00;
C[8544] = 0.000000000000e+00;
C[8545] = 0.000000000000e+00;
C[8546] = 0.000000000000e+00;
C[8547] = 0.000000000000e+00;
C[8548] = 0.000000000000e+00;
C[8549] = 0.000000000000e+00;
C[8550] = 0.000000000000e+00;
C[8551] = 0.000000000000e+00;
C[8552] = 0.000000000000e+00;
C[8553] = 0.000000000000e+00;
C[8554] = 0.000000000000e+00;
C[8555] = 0.000000000000e+00;
C[8556] = 0.000000000000e+00;
C[8557] = 0.000000000000e+00;
C[8558] = 0.000000000000e+00;
C[8559] = 0.000000000000e+00;
C[8560] = 0.000000000000e+00;
C[8561] = 0.000000000000e+00;
C[8562] = 0.000000000000e+00;
C[8563] = 0.000000000000e+00;
C[8564] = 1.000000000000e+00;
C[8565] = -0.000000000000e+00;
C[8566] = -0.000000000000e+00;
C[8567] = 2.000000000000e-01;
C[8568] = 7.333900798553e-02;
C[8569] = 6.362047153685e-03;
C[8570] = -6.010556591766e-04;
C[8571] = 3.848786393481e-01;
C[8572] = -3.839269078305e-03;
C[8573] = -1.558871121848e-03;
C[8574] = 5.161439389046e-02;
C[8575] = -9.515777819764e-04;
C[8576] = -3.917158969830e-04;
C[8577] = 1.308146376828e-01;
C[8578] = 4.809337538862e-04;
C[8579] = 2.760273761874e-04;
C[8580] = 3.103820854629e-02;
C[8581] = -4.790520795244e-04;
C[8582] = -1.825849133125e-04;
C[8583] = 2.478746361851e-03;
C[8584] = -5.693102703078e-05;
C[8585] = -2.300567705913e-05;
C[8586] = 2.402637181610e-03;
C[8587] = -7.440209412287e-05;
C[8588] = -8.375749022958e-06;
C[8589] = -2.250151616666e-05;
C[8590] = -1.775194205613e-06;
C[8591] = 1.320423183131e-07;
C[8592] = 6.331793885090e-03;
C[8593] = 3.787787515571e-02;
C[8594] = 9.910217947446e-03;
C[8595] = -3.898197676111e-03;
C[8596] = 4.001059493986e-01;
C[8597] = 1.017610738349e-02;
C[8598] = -9.288856286221e-04;
C[8599] = 5.611777010490e-02;
C[8600] = 3.515082648402e-03;
C[8601] = 4.789168693132e-04;
C[8602] = 1.277944512613e-01;
C[8603] = 5.504435068064e-04;
C[8604] = -4.825684975439e-04;
C[8605] = 3.280129738197e-02;
C[8606] = 9.552542365383e-04;
C[8607] = -5.537960168734e-05;
C[8608] = 2.733327214088e-03;
C[8609] = 1.844632089005e-04;
C[8610] = -6.867176556651e-05;
C[8611] = 2.737431667324e-03;
C[8612] = 2.516373248261e-04;
C[8613] = -1.579081280064e-06;
C[8614] = -1.684288168692e-05;
C[8615] = 1.407261118437e-06;
C[8616] = -5.322277253969e-04;
C[8617] = 9.903388994638e-03;
C[8618] = 4.831667022501e-02;
C[8619] = -1.628278532670e-03;
C[8620] = 1.041439436398e-02;
C[8621] = 3.831099238665e-01;
C[8622] = -3.536817361256e-04;
C[8623] = 3.384614194885e-03;
C[8624] = 5.444247504754e-02;
C[8625] = 2.806159051060e-04;
C[8626] = 5.499882432859e-04;
C[8627] = 1.285734827553e-01;
C[8628] = -1.863553457980e-04;
C[8629] = 9.666320668157e-04;
C[8630] = 3.106827557173e-02;
C[8631] = -2.045241923280e-05;
C[8632] = 1.762303426394e-04;
C[8633] = 2.636140035131e-03;
C[8634] = 3.536291668137e-07;
C[8635] = 2.285702707152e-04;
C[8636] = 2.729754820560e-03;
C[8637] = 4.132793604381e-07;
C[8638] = 8.819999986882e-07;
C[8639] = -1.898666280408e-05;
C[8640] = 3.848992119356e-01;
C[8641] = -3.199652999310e-03;
C[8642] = -2.374164509556e-03;
C[8643] = 1.157086279206e-01;
C[8644] = 3.645746132494e-03;
C[8645] = 1.667010263417e-03;
C[8646] = 3.852644227109e-01;
C[8647] = 9.079628339077e-04;
C[8648] = 1.806080957141e-04;
C[8649] = 3.103958005213e-02;
C[8650] = -4.359988524238e-04;
C[8651] = -2.360810775903e-04;
C[8652] = 1.324959303574e-01;
C[8653] = 4.593731902983e-04;
C[8654] = 2.123517559687e-04;
C[8655] = 3.150325728994e-02;
C[8656] = 4.893203743447e-05;
C[8657] = 5.998641941000e-06;
C[8658] = 5.928324281646e-02;
C[8659] = 1.696110587644e-04;
C[8660] = 4.564667297258e-04;
C[8661] = 2.592753334133e-03;
C[8662] = 5.425303714786e-06;
C[8663] = 1.872198591318e-05;
C[8664] = -3.211571197433e-03;
C[8665] = 3.991340729382e-01;
C[8666] = 1.076159750982e-02;
C[8667] = 3.684206656989e-03;
C[8668] = 9.172833588709e-02;
C[8669] = -1.896709337182e-02;
C[8670] = 8.642921396082e-04;
C[8671] = 3.875003963498e-01;
C[8672] = 7.009770087307e-03;
C[8673] = -4.372055541329e-04;
C[8674] = 3.273650561794e-02;
C[8675] = 9.897789432052e-04;
C[8676] = 4.619372252647e-04;
C[8677] = 1.300752176732e-01;
C[8678] = -1.790098096608e-03;
C[8679] = 4.638884031548e-05;
C[8680] = 3.185147814128e-02;
C[8681] = 7.271134206730e-04;
C[8682] = 1.790338051449e-04;
C[8683] = 6.025962030049e-02;
C[8684] = 2.318700963145e-03;
C[8685] = 5.934437415530e-06;
C[8686] = 2.648016219369e-03;
C[8687] = 1.134219762461e-04;
C[8688] = -2.321759594670e-03;
C[8689] = 1.051060404440e-02;
C[8690] = 3.840803432546e-01;
C[8691] = 1.708463120532e-03;
C[8692] = -1.904239744979e-02;
C[8693] = 9.567177250533e-02;
C[8694] = 1.103164162828e-04;
C[8695] = 7.205602291581e-03;
C[8696] = 3.976222805806e-01;
C[8697] = -2.334441448339e-04;
C[8698] = 9.775540139325e-04;
C[8699] = 3.113297019760e-02;
C[8700] = 2.151152797764e-04;
C[8701] = -1.795118368473e-03;
C[8702] = 1.308052552195e-01;
C[8703] = 1.923703751462e-06;
C[8704] = 7.383675189668e-04;
C[8705] = 3.270374022858e-02;
C[8706] = 4.716131291516e-04;
C[8707] = 2.283886570639e-03;
C[8708] = 6.460140831012e-02;
C[8709] = 1.955529353270e-05;
C[8710] = 1.112004855535e-04;
C[8711] = 2.839435727087e-03;
C[8712] = 5.252397672524e-02;
C[8713] = -5.670684785109e-04;
C[8714] = -3.516633948767e-04;
C[8715] = 3.844593499173e-01;
C[8716] = 6.214401472345e-04;
C[8717] = 2.410019453391e-04;
C[8718] = 5.083028937570e-02;
C[8719] = -6.997154054240e-04;
C[8720] = -2.826735162552e-03;
C[8721] = 2.539385217504e-03;
C[8722] = -3.125845834659e-05;
C[8723] = -2.031786314954e-05;
C[8724] = 3.144958577037e-02;
C[8725] = 3.019870749056e-05;
C[8726] = 1.063607235521e-05;
C[8727] = 1.286257958815e-01;
C[8728] = -3.509936737283e-05;
C[8729] = -1.448546449311e-04;
C[8730] = 5.098176887705e-01;
C[8731] = 7.235268272991e-04;
C[8732] = 2.915944083113e-03;
C[8733] = 3.740364627332e-02;
C[8734] = 3.943309954943e-05;
C[8735] = 1.514789513542e-04;
C[8736] = -5.860950189153e-04;
C[8737] = 5.634153747573e-02;
C[8738] = 3.227048031900e-03;
C[8739] = 6.466633685213e-04;
C[8740] = 3.871395534938e-01;
C[8741] = 7.229108727246e-03;
C[8742] = -6.863518179874e-04;
C[8743] = 4.418304894972e-02;
C[8744] = -1.386883018977e-02;
C[8745] = -3.256550949337e-05;
C[8746] = 2.748245038811e-03;
C[8747] = 1.657259317738e-04;
C[8748] = 3.151207307538e-05;
C[8749] = 3.182742195088e-02;
C[8750] = 7.399346146778e-04;
C[8751] = -3.420846154373e-05;
C[8752] = 1.279845128215e-01;
C[8753] = -1.242549854456e-03;
C[8754] = 7.056559897082e-04;
C[8755] = 5.095697085074e-01;
C[8756] = 3.379967624815e-03;
C[8757] = 3.855913996150e-05;
C[8758] = 3.745096584923e-02;
C[8759] = 3.370155948657e-04;
C[8760] = -3.798292148161e-04;
C[8761] = 3.317737146599e-03;
C[8762] = 5.535473629209e-02;
C[8763] = 2.801653092087e-04;
C[8764] = 7.087374242972e-03;
C[8765] = 3.968364809049e-01;
C[8766] = -2.801149288395e-03;
C[8767] = -1.389472680161e-02;
C[8768] = 1.839321595570e-02;
C[8769] = -2.221323158133e-05;
C[8770] = 1.713068421137e-04;
C[8771] = 2.696957451435e-03;
C[8772] = 1.263578950730e-05;
C[8773] = 7.322870310506e-04;
C[8774] = 3.265135358353e-02;
C[8775] = -1.431489199872e-04;
C[8776] = -1.244276295246e-03;
C[8777] = 1.259360447948e-01;
C[8778] = 2.882962493574e-03;
C[8779] = 3.442548037530e-03;
C[8780] = 5.267521346009e-01;
C[8781] = 1.497805364168e-04;
C[8782] = 3.400800149988e-04;
C[8783] = 3.873494057668e-02;
C[8784] = -6.990283568859e+00;
C[8785] = 1.129892406609e-02;
C[8786] = 8.312599072662e-03;
C[8787] = 2.151994347489e+00;
C[8788] = -1.607112560251e-02;
C[8789] = 4.925386960888e-03;
C[8790] = 7.974055987665e-01;
C[8791] = -1.564360684939e-02;
C[8792] = -6.236988587713e-03;
C[8793] = 7.693484107961e-02;
C[8794] = 3.999805105579e-03;
C[8795] = 6.698719099028e-03;
C[8796] = 3.793586789395e-01;
C[8797] = -6.603948258726e-03;
C[8798] = -2.018764171283e-03;
C[8799] = 5.142410087593e-02;
C[8800] = -1.108122294245e-03;
C[8801] = -4.802346559914e-04;
C[8802] = 7.886851302290e-02;
C[8803] = -1.716036859424e-03;
C[8804] = -2.627075213296e-04;
C[8805] = 2.439966102930e-03;
C[8806] = -6.308971195461e-05;
C[8807] = -1.126482051388e-05;
C[8808] = 1.069385869419e-02;
C[8809] = -7.108014937445e+00;
C[8810] = 7.862750721347e-02;
C[8811] = -1.704964878896e-02;
C[8812] = 2.160565678047e+00;
C[8813] = 3.221571390729e-02;
C[8814] = -1.532054259118e-02;
C[8815] = 8.635986504066e-01;
C[8816] = 5.159754944721e-02;
C[8817] = 3.969551836985e-03;
C[8818] = 4.414348757306e-02;
C[8819] = 6.338209685311e-03;
C[8820] = -6.656967301763e-03;
C[8821] = 3.996337512904e-01;
C[8822] = 1.118819049473e-02;
C[8823] = -1.079470194499e-03;
C[8824] = 5.619660666091e-02;
C[8825] = 3.548236325489e-03;
C[8826] = -1.583929635680e-03;
C[8827] = 8.719208524981e-02;
C[8828] = 7.023722126024e-03;
C[8829] = -5.738415623838e-05;
C[8830] = 2.752394805205e-03;
C[8831] = 2.425788932865e-04;
C[8832] = 9.689157748255e-03;
C[8833] = 7.849092815731e-02;
C[8834] = -7.131062173847e+00;
C[8835] = 3.854350503277e-03;
C[8836] = 3.549888983432e-02;
C[8837] = 2.067579116413e+00;
C[8838] = -5.731932207456e-03;
C[8839] = 4.986497452209e-02;
C[8840] = 8.322574253134e-01;
C[8841] = 6.767547032808e-03;
C[8842] = 6.331380732502e-03;
C[8843] = 4.774883995821e-02;
C[8844] = -2.090500342033e-03;
C[8845] = 1.140057400991e-02;
C[8846] = 3.764176684583e-01;
C[8847] = -4.368226819445e-04;
C[8848] = 3.411802238381e-03;
C[8849] = 5.406824109083e-02;
C[8850] = -7.017895635980e-05;
C[8851] = 6.490336475638e-03;
C[8852] = 8.747645927461e-02;
C[8853] = -3.041669124971e-06;
C[8854] = 2.218189570393e-04;
C[8855] = 2.759488389802e-03;
C[8856] = 2.152405799239e+00;
C[8857] = -3.078755252937e-03;
C[8858] = -1.106336903443e-02;
C[8859] = -6.228440981947e+00;
C[8860] = 1.089564386145e-02;
C[8861] = 4.224771493786e-03;
C[8862] = 2.144121809615e+00;
C[8863] = 1.356586223378e-02;
C[8864] = 1.079491965570e-03;
C[8865] = 3.793792515270e-01;
C[8866] = -5.958422624962e-03;
C[8867] = -2.836386318918e-03;
C[8868] = 1.189805804301e-01;
C[8869] = 6.280914105735e-03;
C[8870] = 2.875493907914e-03;
C[8871] = 3.874738194645e-01;
C[8872] = 8.879689085713e-04;
C[8873] = 4.451661197544e-05;
C[8874] = 1.004595480847e+00;
C[8875] = 3.063873632143e-03;
C[8876] = 6.909135881960e-03;
C[8877] = 5.947550586136e-02;
C[8878] = 1.229454763867e-04;
C[8879] = 3.120327755002e-04;
C[8880] = -3.517167985062e-03;
C[8881] = 2.141128148838e+00;
C[8882] = 4.244295275116e-02;
C[8883] = 1.166485435135e-02;
C[8884] = -6.290171286442e+00;
C[8885] = -3.969767542273e-02;
C[8886] = 1.291142230670e-02;
C[8887] = 2.100315808677e+00;
C[8888] = -1.893356552632e-02;
C[8889] = -5.976250377853e-03;
C[8890] = 3.986618748299e-01;
C[8891] = 1.174777715575e-02;
C[8892] = 6.319374630230e-03;
C[8893] = 9.043126514843e-02;
C[8894] = -1.950648161452e-02;
C[8895] = 8.432954896900e-04;
C[8896] = 3.894299394355e-01;
C[8897] = 6.296128676286e-03;
C[8898] = 3.142256654036e-03;
C[8899] = 1.017600331937e+00;
C[8900] = 3.472224664862e-02;
C[8901] = 1.279298936721e-04;
C[8902] = 6.052851447304e-02;
C[8903] = 2.325376972635e-03;
C[8904] = -1.033238249555e-02;
C[8905] = 3.890564712555e-02;
C[8906] = 2.086987504174e+00;
C[8907] = 5.053828636100e-03;
C[8908] = -4.120375698208e-02;
C[8909] = -6.277535928791e+00;
C[8910] = 3.180295043413e-05;
C[8911] = -1.609716537653e-02;
C[8912] = 2.126601773693e+00;
C[8913] = -2.781652644105e-03;
C[8914] = 1.152268715565e-02;
C[8915] = 3.773880878463e-01;
C[8916] = 2.916946765030e-03;
C[8917] = -1.958178569249e-02;
C[8918] = 1.019309366312e-01;
C[8919] = -2.616030153918e-05;
C[8920] = 6.496439207874e-03;
C[8921] = 3.979640423452e-01;
C[8922] = 7.035927762579e-03;
C[8923] = 3.440800963413e-02;
C[8924] = 1.086287928144e+00;
C[8925] = 3.201845640490e-04;
C[8926] = 2.295426575758e-03;
C[8927] = 6.456995975460e-02;
C[8928] = 8.155972554622e-01;
C[8929] = -8.084199588953e-03;
C[8930] = -5.691565382478e-03;
C[8931] = 2.128020353744e+00;
C[8932] = 8.054382459223e-03;
C[8933] = 2.645513531559e-03;
C[8934] = -7.115071207943e+00;
C[8935] = -1.141378587648e-03;
C[8936] = -9.055012287949e-03;
C[8937] = 5.233368371071e-02;
C[8938] = -7.176530443882e-04;
C[8939] = -4.348043406956e-04;
C[8940] = 3.866687466709e-01;
C[8941] = 6.004434973162e-04;
C[8942] = 1.045252275171e-04;
C[8943] = 4.734754772383e-02;
C[8944] = -7.390931873036e-05;
C[8945] = -6.654965972135e-04;
C[8946] = 4.095765367625e+00;
C[8947] = 3.396529971663e-03;
C[8948] = 1.190159071217e-02;
C[8949] = 5.113376749901e-01;
C[8950] = 2.984977465401e-04;
C[8951] = 9.739459201190e-04;
C[8952] = -8.333951588164e-03;
C[8953] = 8.680739978233e-01;
C[8954] = 4.671365126239e-02;
C[8955] = 8.339872926047e-03;
C[8956] = 2.093098951558e+00;
C[8957] = -1.562703666325e-02;
C[8958] = -8.741068389162e-04;
C[8959] = -7.135560647748e+00;
C[8960] = -4.184633004912e-02;
C[8961] = -7.426395311834e-04;
C[8962] = 5.642037403175e-02;
C[8963] = 3.254236075396e-03;
C[8964] = 6.266694431849e-04;
C[8965] = 3.890690965796e-01;
C[8966] = 6.519945643538e-03;
C[8967] = -6.054573129376e-05;
C[8968] = 3.980316429657e-02;
C[8969] = -1.378055443991e-02;
C[8970] = 3.127586804582e-03;
C[8971] = 4.087527300595e+00;
C[8972] = 9.888815270604e-03;
C[8973] = 2.851376812954e-04;
C[8974] = 5.119466382813e-01;
C[8975] = 3.964201536681e-03;
C[8976] = -5.999254944374e-03;
C[8977] = 4.765063941116e-02;
C[8978] = 8.505026502043e-01;
C[8979] = 3.070636235462e-03;
C[8980] = -1.738148241302e-02;
C[8981] = 2.110885780179e+00;
C[8982] = -8.543294804800e-03;
C[8983] = -4.236426228598e-02;
C[8984] = -7.226018724340e+00;
C[8985] = -4.683479738245e-04;
C[8986] = 3.350890823686e-03;
C[8987] = 5.498050233538e-02;
C[8988] = 1.440738254700e-04;
C[8989] = 6.373732831951e-03;
C[8990] = 3.971782426696e-01;
C[8991] = -6.399107230560e-04;
C[8992] = -1.380645105175e-02;
C[8993] = 1.859309831848e-02;
C[8994] = 1.138557188851e-02;
C[8995] = 1.085864798746e-02;
C[8996] = 4.180686236952e+00;
C[8997] = 9.482966874960e-04;
C[8998] = 4.020892699323e-03;
C[8999] = 5.266569105120e-01;
C[9000] = 0.000000000000e+00;
C[9001] = 0.000000000000e+00;
C[9002] = 0.000000000000e+00;
C[9003] = 0.000000000000e+00;
C[9004] = 0.000000000000e+00;
C[9005] = 0.000000000000e+00;
C[9006] = 0.000000000000e+00;
C[9007] = 0.000000000000e+00;
C[9008] = 0.000000000000e+00;
C[9009] = 0.000000000000e+00;
C[9010] = 0.000000000000e+00;
C[9011] = 0.000000000000e+00;
C[9012] = 0.000000000000e+00;
C[9013] = 0.000000000000e+00;
C[9014] = 0.000000000000e+00;
C[9015] = 0.000000000000e+00;
C[9016] = 0.000000000000e+00;
C[9017] = 0.000000000000e+00;
C[9018] = 1.000000000000e+00;
C[9019] = 0.000000000000e+00;
C[9020] = 0.000000000000e+00;
C[9021] = 2.000000000000e-01;
C[9022] = -0.000000000000e+00;
C[9023] = -0.000000000000e+00;
C[9024] = 0.000000000000e+00;
C[9025] = 0.000000000000e+00;
C[9026] = 0.000000000000e+00;
C[9027] = 0.000000000000e+00;
C[9028] = 0.000000000000e+00;
C[9029] = 0.000000000000e+00;
C[9030] = 0.000000000000e+00;
C[9031] = 0.000000000000e+00;
C[9032] = 0.000000000000e+00;
C[9033] = 0.000000000000e+00;
C[9034] = 0.000000000000e+00;
C[9035] = 0.000000000000e+00;
C[9036] = 0.000000000000e+00;
C[9037] = 0.000000000000e+00;
C[9038] = 0.000000000000e+00;
C[9039] = 0.000000000000e+00;
C[9040] = 0.000000000000e+00;
C[9041] = 0.000000000000e+00;
C[9042] = 0.000000000000e+00;
C[9043] = 1.000000000000e+00;
C[9044] = 0.000000000000e+00;
C[9045] = -0.000000000000e+00;
C[9046] = 2.000000000000e-01;
C[9047] = -0.000000000000e+00;
C[9048] = 0.000000000000e+00;
C[9049] = 0.000000000000e+00;
C[9050] = 0.000000000000e+00;
C[9051] = 0.000000000000e+00;
C[9052] = 0.000000000000e+00;
C[9053] = 0.000000000000e+00;
C[9054] = 0.000000000000e+00;
C[9055] = 0.000000000000e+00;
C[9056] = 0.000000000000e+00;
C[9057] = 0.000000000000e+00;
C[9058] = 0.000000000000e+00;
C[9059] = 0.000000000000e+00;
C[9060] = 0.000000000000e+00;
C[9061] = 0.000000000000e+00;
C[9062] = 0.000000000000e+00;
C[9063] = 0.000000000000e+00;
C[9064] = 0.000000000000e+00;
C[9065] = 0.000000000000e+00;
C[9066] = 0.000000000000e+00;
C[9067] = 0.000000000000e+00;
C[9068] = 1.000000000000e+00;
C[9069] = -0.000000000000e+00;
C[9070] = -0.000000000000e+00;
C[9071] = 2.000000000000e-01;
C[9072] = 9.247227117764e-02;
C[9073] = 1.682748803974e-03;
C[9074] = 1.066615565861e-02;
C[9075] = 3.581248661422e-01;
C[9076] = -6.285624569752e-03;
C[9077] = -1.160492248360e-03;
C[9078] = 4.766815274685e-02;
C[9079] = 2.530453724296e-03;
C[9080] = -6.937768430872e-04;
C[9081] = 1.328009982960e-01;
C[9082] = -1.227245021222e-03;
C[9083] = 1.061431036816e-03;
C[9084] = 2.850665044478e-02;
C[9085] = 8.585079466304e-04;
C[9086] = -2.934111547245e-04;
C[9087] = 2.266552798778e-03;
C[9088] = 2.099188073640e-04;
C[9089] = -4.527137879242e-05;
C[9090] = 2.168688780022e-03;
C[9091] = 4.034288502346e-04;
C[9092] = -7.017816864850e-05;
C[9093] = -2.573944032892e-05;
C[9094] = 1.811683513888e-05;
C[9095] = -1.685066411181e-06;
C[9096] = 6.915074234295e-04;
C[9097] = 4.324027839177e-02;
C[9098] = 4.345482870172e-03;
C[9099] = -5.230306074159e-03;
C[9100] = 4.016811136223e-01;
C[9101] = 9.340285119749e-03;
C[9102] = 2.606927968920e-03;
C[9103] = 5.565959363828e-02;
C[9104] = 6.909854727574e-04;
C[9105] = -1.293327779925e-03;
C[9106] = 1.284816049487e-01;
C[9107] = 1.108776160772e-03;
C[9108] = 9.288311560667e-04;
C[9109] = 3.246462932112e-02;
C[9110] = -5.058265335150e-05;
C[9111] = 2.115964408014e-04;
C[9112] = 2.680475854021e-03;
C[9113] = -1.590984486240e-05;
C[9114] = 3.996299897453e-04;
C[9115] = 2.652883028071e-03;
C[9116] = -9.412454365292e-05;
C[9117] = 1.765797993358e-05;
C[9118] = -2.310104449418e-05;
C[9119] = -1.144652559209e-05;
C[9120] = 1.072191048157e-02;
C[9121] = 3.931562579811e-03;
C[9122] = 6.752052777095e-02;
C[9123] = -1.000807748508e-03;
C[9124] = 1.036232000120e-02;
C[9125] = 3.538286952619e-01;
C[9126] = -7.070483092388e-04;
C[9127] = 5.775835079180e-04;
C[9128] = 4.806190806753e-02;
C[9129] = 1.065148025014e-03;
C[9130] = 1.081181474748e-03;
C[9131] = 1.312983612589e-01;
C[9132] = -2.816808846365e-04;
C[9133] = 1.171771812630e-05;
C[9134] = 2.783239099515e-02;
C[9135] = -4.755417334397e-05;
C[9136] = -2.301259051021e-05;
C[9137] = 2.268784211056e-03;
C[9138] = -7.878231666472e-05;
C[9139] = -1.075434945623e-04;
C[9140] = 2.210517000275e-03;
C[9141] = -2.156415774851e-06;
C[9142] = -1.155062485261e-05;
C[9143] = -2.960180472583e-05;
C[9144] = 3.582530360855e-01;
C[9145] = -6.570601253353e-03;
C[9146] = -9.529815399145e-04;
C[9147] = 1.373946602290e-01;
C[9148] = 6.120259184528e-03;
C[9149] = 2.930532851061e-03;
C[9150] = 3.929336558291e-01;
C[9151] = -2.866411492255e-03;
C[9152] = 3.038999927433e-04;
C[9153] = 2.851519510767e-02;
C[9154] = 8.394781441205e-04;
C[9155] = -2.784924707303e-04;
C[9156] = 1.346726260854e-01;
C[9157] = -8.227671877457e-04;
C[9158] = 4.284838308261e-04;
C[9159] = 3.200233556925e-02;
C[9160] = -2.574521212762e-04;
C[9161] = -2.898893213733e-05;
C[9162] = 6.064362464252e-02;
C[9163] = -5.741740905254e-04;
C[9164] = -2.627836767860e-04;
C[9165] = 2.642751847959e-03;
C[9166] = -2.696551089059e-05;
C[9167] = -1.576200467267e-05;
C[9168] = -5.634902802482e-03;
C[9169] = 4.015439720198e-01;
C[9170] = 1.024636751541e-02;
C[9171] = 5.116849197258e-03;
C[9172] = 8.678948421604e-02;
C[9173] = -1.749969034895e-02;
C[9174] = -2.945031586888e-03;
C[9175] = 3.909741675090e-01;
C[9176] = 8.039188488585e-03;
C[9177] = 9.018893977818e-04;
C[9178] = 3.245548654763e-02;
C[9179] = 3.987552406624e-06;
C[9180] = -8.896611868970e-04;
C[9181] = 1.301816850359e-01;
C[9182] = -7.024996538843e-04;
C[9183] = -2.593860713843e-04;
C[9184] = 3.209916361163e-02;
C[9185] = 7.864239168526e-04;
C[9186] = -5.718447184848e-04;
C[9187] = 6.150588830711e-02;
C[9188] = 2.542095739345e-03;
C[9189] = -2.655462185206e-05;
C[9190] = 2.699652159557e-03;
C[9191] = 1.210785626585e-04;
C[9192] = -9.511533320166e-04;
C[9193] = 1.028523210271e-02;
C[9194] = 3.540031845002e-01;
C[9195] = 2.736075162848e-03;
C[9196] = -1.815483122917e-02;
C[9197] = 1.240362594955e-01;
C[9198] = 3.035909995710e-04;
C[9199] = 8.186244321798e-03;
C[9200] = 4.045201529391e-01;
C[9201] = -2.794552269682e-04;
C[9202] = 1.241381217891e-05;
C[9203] = 2.784402361104e-02;
C[9204] = 4.155199849452e-04;
C[9205] = -7.461757125657e-04;
C[9206] = 1.339536726668e-01;
C[9207] = -2.742573302734e-05;
C[9208] = 7.944072976500e-04;
C[9209] = 3.314213555748e-02;
C[9210] = -2.535452463425e-04;
C[9211] = 2.529843825069e-03;
C[9212] = 6.520399600373e-02;
C[9213] = -1.519988720359e-05;
C[9214] = 1.199840121312e-04;
C[9215] = 2.856098070313e-03;
C[9216] = 4.882897674571e-02;
C[9217] = -1.333154245717e-03;
C[9218] = -3.713102289305e-04;
C[9219] = 3.917096975460e-01;
C[9220] = 9.610420584523e-04;
C[9221] = -2.221327233399e-05;
C[9222] = 4.364677607895e-02;
C[9223] = 7.288810918464e-04;
C[9224] = 1.410869719417e-03;
C[9225] = 2.343941065369e-03;
C[9226] = -5.107570684105e-05;
C[9227] = -2.517163465675e-05;
C[9228] = 3.192073835038e-02;
C[9229] = 1.018838305035e-06;
C[9230] = -4.914601782101e-05;
C[9231] = 1.280476223555e-01;
C[9232] = 9.196478204204e-05;
C[9233] = 1.780540302092e-04;
C[9234] = 5.136267001679e-01;
C[9235] = -3.629886442613e-04;
C[9236] = -9.317661682963e-04;
C[9237] = 3.770830093087e-02;
C[9238] = -4.496585137763e-05;
C[9239] = -1.030625834710e-04;
C[9240] = -1.327561581164e-03;
C[9241] = 5.700476639987e-02;
C[9242] = 3.042167295585e-03;
C[9243] = 9.417109355710e-04;
C[9244] = 3.894209083861e-01;
C[9245] = 5.710210949880e-03;
C[9246] = 7.302574917813e-04;
C[9247] = 3.582049209954e-02;
C[9248] = -1.325623774059e-02;
C[9249] = -4.728221299998e-05;
C[9250] = 2.770154038127e-03;
C[9251] = 1.412929953342e-04;
C[9252] = -3.577292754512e-06;
C[9253] = 3.199561300344e-02;
C[9254] = 6.293384061888e-04;
C[9255] = 9.205654203770e-05;
C[9256] = 1.273957189902e-01;
C[9257] = -1.159526560866e-03;
C[9258] = -3.636666876619e-04;
C[9259] = 5.150604356897e-01;
C[9260] = 4.477848828433e-03;
C[9261] = -4.503604737646e-05;
C[9262] = 3.785473204027e-02;
C[9263] = 3.915981491923e-04;
C[9264] = -4.045415167186e-04;
C[9265] = 3.247712739665e-03;
C[9266] = 5.069395086566e-02;
C[9267] = -3.467976883773e-06;
C[9268] = 5.453780960278e-03;
C[9269] = 4.020668582288e-01;
C[9270] = 1.428045194217e-03;
C[9271] = -1.325899376392e-02;
C[9272] = 1.898124123382e-02;
C[9273] = -2.598902370117e-05;
C[9274] = 1.545386395981e-04;
C[9275] = 2.444253730931e-03;
C[9276] = -4.948013011246e-05;
C[9277] = 6.140634149655e-04;
C[9278] = 3.297858257679e-02;
C[9279] = 1.791990618625e-04;
C[9280] = -1.159710295754e-03;
C[9281] = 1.259461109041e-01;
C[9282] = -9.360915016536e-04;
C[9283] = 4.504450135945e-03;
C[9284] = 5.258114946650e-01;
C[9285] = -1.033710889959e-04;
C[9286] = 3.927970109301e-04;
C[9287] = 3.865106929059e-02;
C[9288] = -6.957446098000e+00;
C[9289] = -3.465859594606e-02;
C[9290] = 6.260767879614e-02;
C[9291] = 2.060106464201e+00;
C[9292] = 7.992166611727e-03;
C[9293] = 2.244625495567e-03;
C[9294] = 7.371587004243e-01;
C[9295] = 5.986425322210e-02;
C[9296] = -1.139588846418e-02;
C[9297] = 9.971795487596e-02;
C[9298] = -2.886863608500e-02;
C[9299] = 1.475489527178e-02;
C[9300] = 3.493912970146e-01;
C[9301] = 2.420550186557e-02;
C[9302] = -4.488309507644e-03;
C[9303] = 4.719375792330e-02;
C[9304] = 4.372867535971e-03;
C[9305] = -8.749648549170e-04;
C[9306] = 7.252644800804e-02;
C[9307] = 7.592974147419e-03;
C[9308] = -1.671247029070e-03;
C[9309] = 2.204794678358e-03;
C[9310] = 2.951312848930e-04;
C[9311] = -6.535718042855e-05;
C[9312] = -5.448342355696e-02;
C[9313] = -7.082855386226e+00;
C[9314] = 7.345028710390e-02;
C[9315] = 2.889118933015e-02;
C[9316] = 2.155725824919e+00;
C[9317] = -2.564344931287e-03;
C[9318] = 6.042598206786e-02;
C[9319] = 8.488508782859e-01;
C[9320] = -5.723203677081e-03;
C[9321] = -2.985987746554e-02;
C[9322] = 5.420146150084e-02;
C[9323] = 2.258875272207e-02;
C[9324] = 2.528624342444e-02;
C[9325] = 3.912808139273e-01;
C[9326] = -1.020791218788e-02;
C[9327] = 4.340038913420e-03;
C[9328] = 5.498837291506e-02;
C[9329] = -5.384868595082e-04;
C[9330] = 7.401621161923e-03;
C[9331] = 8.600931440593e-02;
C[9332] = -1.518634563171e-05;
C[9333] = 2.812969499086e-04;
C[9334] = 2.689609921018e-03;
C[9335] = -1.997212865686e-05;
C[9336] = 6.372277525544e-02;
C[9337] = 6.517188129669e-02;
C[9338] = -7.084328684807e+00;
C[9339] = 5.924714731889e-03;
C[9340] = 1.569429270475e-02;
C[9341] = 1.973786153445e+00;
C[9342] = -1.194137225265e-02;
C[9343] = -6.736464809970e-03;
C[9344] = 7.301327081570e-01;
C[9345] = 1.481065009475e-02;
C[9346] = 2.217483243171e-02;
C[9347] = 8.615811841970e-02;
C[9348] = -4.221886892565e-03;
C[9349] = -9.120330286961e-03;
C[9350] = 3.344630312836e-01;
C[9351] = -9.282139477149e-04;
C[9352] = -6.654050428979e-04;
C[9353] = 4.686663763009e-02;
C[9354] = -1.831160499319e-03;
C[9355] = -4.479441519288e-04;
C[9356] = 7.469609558959e-02;
C[9357] = -7.266506063167e-05;
C[9358] = -4.068451168752e-05;
C[9359] = 2.290951157818e-03;
C[9360] = 2.062669863066e+00;
C[9361] = 2.085285746275e-03;
C[9362] = 6.881238903757e-03;
C[9363] = -6.161877336502e+00;
C[9364] = -1.434420389867e-02;
C[9365] = 2.070454181770e-02;
C[9366] = 2.209890128409e+00;
C[9367] = -6.119586558039e-02;
C[9368] = 1.277210460703e-02;
C[9369] = 3.495194669578e-01;
C[9370] = 2.394594824525e-02;
C[9371] = -4.174060683971e-03;
C[9372] = 1.452660390459e-01;
C[9373] = -2.386635319511e-02;
C[9374] = 6.954627323902e-03;
C[9375] = 3.940310780221e-01;
C[9376] = -5.092981162687e-03;
C[9377] = -1.346609848967e-04;
C[9378] = 1.028338489805e+00;
C[9379] = -1.088995722842e-02;
C[9380] = -4.440241888001e-03;
C[9381] = 6.074239449490e-02;
C[9382] = -5.175448195382e-04;
C[9383] = -3.548533628562e-04;
C[9384] = 2.100660195713e-02;
C[9385] = 2.152982992870e+00;
C[9386] = 1.337524298884e-02;
C[9387] = -3.441240364406e-02;
C[9388] = -6.296050654709e+00;
C[9389] = -1.330642122485e-02;
C[9390] = -6.177585694962e-02;
C[9391] = 2.115254289936e+00;
C[9392] = 2.885944280012e-02;
C[9393] = 2.485622363284e-02;
C[9394] = 3.911436723249e-01;
C[9395] = -9.236282772756e-03;
C[9396] = -2.486976318238e-02;
C[9397] = 9.710750618738e-02;
C[9398] = 2.344222229534e-03;
C[9399] = -5.059055755660e-03;
C[9400] = 3.915083043451e-01;
C[9401] = 8.751450192374e-03;
C[9402] = -1.071077143028e-02;
C[9403] = 1.036135167377e+00;
C[9404] = 3.966034441741e-02;
C[9405] = -5.053898653104e-04;
C[9406] = 6.156469383095e-02;
C[9407] = 2.433173977827e-03;
C[9408] = 6.431403822437e-03;
C[9409] = 1.633459472783e-02;
C[9410] = 1.977275938211e+00;
C[9411] = 1.681538805344e-02;
C[9412] = -2.640923882926e-02;
C[9413] = -6.185657498998e+00;
C[9414] = 1.324949013293e-02;
C[9415] = 3.067839608415e-02;
C[9416] = 2.231959868729e+00;
C[9417] = -4.278970591300e-03;
C[9418] = -9.262965204923e-03;
C[9419] = 3.346375205219e-01;
C[9420] = 6.760169635689e-03;
C[9421] = 1.689081349313e-03;
C[9422] = 1.426600272413e-01;
C[9423] = -7.740774189108e-05;
C[9424] = 8.921783579506e-03;
C[9425] = 4.063021178890e-01;
C[9426] = -4.238422127096e-03;
C[9427] = 3.978220427945e-02;
C[9428] = 1.100646700328e+00;
C[9429] = -3.466452862892e-04;
C[9430] = 2.435180687008e-03;
C[9431] = 6.513179911269e-02;
C[9432] = 7.603751804016e-01;
C[9433] = -1.837566222488e-02;
C[9434] = -5.226610646489e-03;
C[9435] = 2.185410962747e+00;
C[9436] = 1.634561595720e-02;
C[9437] = 6.733404694827e-03;
C[9438] = -7.134014118840e+00;
C[9439] = 1.154671253383e-03;
C[9440] = 8.484280282687e-03;
C[9441] = 4.835458192216e-02;
C[9442] = 3.999566987828e-04;
C[9443] = -5.924758674066e-04;
C[9444] = 3.928071197390e-01;
C[9445] = -1.152982110319e-03;
C[9446] = -4.032120137961e-04;
C[9447] = 4.170472977529e-02;
C[9448] = 1.187851432532e-03;
C[9449] = 2.721554728546e-03;
C[9450] = 4.117787644404e+00;
C[9451] = -3.119803633788e-03;
C[9452] = -7.015045567815e-03;
C[9453] = 5.150261993733e-01;
C[9454] = -7.339318720222e-04;
C[9455] = -1.602454102357e-03;
C[9456] = -1.729605288710e-02;
C[9457] = 8.757543335177e-01;
C[9458] = 4.255521094336e-02;
C[9459] = 1.496658297612e-02;
C[9460] = 2.084189107478e+00;
C[9461] = -1.884227135420e-02;
C[9462] = 1.182199252081e-03;
C[9463] = -7.163818002808e+00;
C[9464] = -3.896152120769e-02;
C[9465] = 5.148522305108e-04;
C[9466] = 5.633354567665e-02;
C[9467] = 1.799178744769e-03;
C[9468] = -1.284858734862e-03;
C[9469] = 3.899550452221e-01;
C[9470] = 6.445750207589e-03;
C[9471] = 1.189227832466e-03;
C[9472] = 3.450602085125e-02;
C[9473] = -1.253264689834e-02;
C[9474] = -3.133835849577e-03;
C[9475] = 4.119741247705e+00;
C[9476] = 1.719678123769e-02;
C[9477] = -7.332141549657e-04;
C[9478] = 5.166060364539e-01;
C[9479] = 4.466750189160e-03;
C[9480] = -5.611181936808e-03;
C[9481] = 4.541134166108e-02;
C[9482] = 7.827735641196e-01;
C[9483] = 6.624745214491e-03;
C[9484] = -2.284870776601e-02;
C[9485] = 2.182893974523e+00;
C[9486] = 8.827789778671e-03;
C[9487] = -3.901664167419e-02;
C[9488] = -7.213290976257e+00;
C[9489] = -5.857295285484e-04;
C[9490] = 2.018240407400e-03;
C[9491] = 4.949868042822e-02;
C[9492] = -4.420289545237e-04;
C[9493] = 6.166042664067e-03;
C[9494] = 4.038488231787e-01;
C[9495] = 2.738730203345e-03;
C[9496] = -1.253540292167e-02;
C[9497] = 1.905491176347e-02;
C[9498] = -7.106946197768e-03;
C[9499] = 1.758667872154e-02;
C[9500] = 4.171548799195e+00;
C[9501] = -1.605625370539e-03;
C[9502] = 4.487748796491e-03;
C[9503] = 5.253520213942e-01;
C[9504] = 0.000000000000e+00;
C[9505] = 0.000000000000e+00;
C[9506] = 0.000000000000e+00;
C[9507] = 0.000000000000e+00;
C[9508] = 0.000000000000e+00;
C[9509] = 0.000000000000e+00;
C[9510] = 0.000000000000e+00;
C[9511] = 0.000000000000e+00;
C[9512] = 0.000000000000e+00;
C[9513] = 0.000000000000e+00;
C[9514] = 0.000000000000e+00;
C[9515] = 0.000000000000e+00;
C[9516] = 0.000000000000e+00;
C[9517] = 0.000000000000e+00;
C[9518] = 0.000000000000e+00;
C[9519] = 0.000000000000e+00;
C[9520] = 0.000000000000e+00;
C[9521] = 0.000000000000e+00;
C[9522] = 1.000000000000e+00;
C[9523] = 0.000000000000e+00;
C[9524] = 0.000000000000e+00;
C[9525] = 2.000000000000e-01;
C[9526] = -0.000000000000e+00;
C[9527] = -0.000000000000e+00;
C[9528] = 0.000000000000e+00;
C[9529] = 0.000000000000e+00;
C[9530] = 0.000000000000e+00;
C[9531] = 0.000000000000e+00;
C[9532] = 0.000000000000e+00;
C[9533] = 0.000000000000e+00;
C[9534] = 0.000000000000e+00;
C[9535] = 0.000000000000e+00;
C[9536] = 0.000000000000e+00;
C[9537] = 0.000000000000e+00;
C[9538] = 0.000000000000e+00;
C[9539] = 0.000000000000e+00;
C[9540] = 0.000000000000e+00;
C[9541] = 0.000000000000e+00;
C[9542] = 0.000000000000e+00;
C[9543] = 0.000000000000e+00;
C[9544] = 0.000000000000e+00;
C[9545] = 0.000000000000e+00;
C[9546] = 0.000000000000e+00;
C[9547] = 1.000000000000e+00;
C[9548] = 0.000000000000e+00;
C[9549] = -0.000000000000e+00;
C[9550] = 2.000000000000e-01;
C[9551] = -0.000000000000e+00;
C[9552] = 0.000000000000e+00;
C[9553] = 0.000000000000e+00;
C[9554] = 0.000000000000e+00;
C[9555] = 0.000000000000e+00;
C[9556] = 0.000000000000e+00;
C[9557] = 0.000000000000e+00;
C[9558] = 0.000000000000e+00;
C[9559] = 0.000000000000e+00;
C[9560] = 0.000000000000e+00;
C[9561] = 0.000000000000e+00;
C[9562] = 0.000000000000e+00;
C[9563] = 0.000000000000e+00;
C[9564] = 0.000000000000e+00;
C[9565] = 0.000000000000e+00;
C[9566] = 0.000000000000e+00;
C[9567] = 0.000000000000e+00;
C[9568] = 0.000000000000e+00;
C[9569] = 0.000000000000e+00;
C[9570] = 0.000000000000e+00;
C[9571] = 0.000000000000e+00;
C[9572] = 1.000000000000e+00;
C[9573] = -0.000000000000e+00;
C[9574] = -0.000000000000e+00;
C[9575] = 2.000000000000e-01;
C[9576] = 6.999546239140e-02;
C[9577] = -2.891057549165e-02;
C[9578] = 1.464660100946e-02;
C[9579] = 3.793880005830e-01;
C[9580] = 2.532370608394e-02;
C[9581] = -4.343996942290e-03;
C[9582] = 5.242928129735e-02;
C[9583] = 3.441824012961e-03;
C[9584] = -2.281730530731e-04;
C[9585] = 1.302952665153e-01;
C[9586] = -2.368024203798e-03;
C[9587] = 9.090902667848e-04;
C[9588] = 3.096129338774e-02;
C[9589] = 2.096559737869e-03;
C[9590] = -1.160053991637e-04;
C[9591] = 2.542146579865e-03;
C[9592] = 1.684789293593e-04;
C[9593] = 3.336502537680e-06;
C[9594] = 2.564861160752e-03;
C[9595] = 1.687452331935e-04;
C[9596] = 2.605493587149e-05;
C[9597] = -1.703013134188e-05;
C[9598] = -6.450079615973e-07;
C[9599] = 3.431025414910e-06;
C[9600] = -2.874960672603e-02;
C[9601] = 6.780476554677e-02;
C[9602] = 1.640375341623e-02;
C[9603] = 2.520263140936e-02;
C[9604] = 3.795220384999e-01;
C[9605] = -5.644052646474e-03;
C[9606] = 3.381867063578e-03;
C[9607] = 5.415494669967e-02;
C[9608] = 1.234249075497e-03;
C[9609] = -2.357292952757e-03;
C[9610] = 1.297774500711e-01;
C[9611] = 9.821799135957e-04;
C[9612] = 2.088689165860e-03;
C[9613] = 3.137167038129e-02;
C[9614] = -1.638904478617e-04;
C[9615] = 1.648857461230e-04;
C[9616] = 2.654169080101e-03;
C[9617] = 8.473931184214e-05;
C[9618] = 1.610035896679e-04;
C[9619] = 2.758300384162e-03;
C[9620] = 1.790450848092e-04;
C[9621] = -8.513451818061e-07;
C[9622] = -1.141922072425e-05;
C[9623] = 4.936192935335e-06;
C[9624] = 1.479591915316e-02;
C[9625] = 1.660136399341e-02;
C[9626] = 6.769911624102e-02;
C[9627] = -4.251248698214e-03;
C[9628] = -5.575611829061e-03;
C[9629] = 3.505571522762e-01;
C[9630] = -2.951159763373e-04;
C[9631] = 1.147983541234e-03;
C[9632] = 5.114195573831e-02;
C[9633] = 9.190448096982e-04;
C[9634] = 9.953539520744e-04;
C[9635] = 1.299789076928e-01;
C[9636] = -1.149159008042e-04;
C[9637] = -1.646860195491e-04;
C[9638] = 2.887990910499e-02;
C[9639] = -1.444768580637e-06;
C[9640] = 7.903823365859e-05;
C[9641] = 2.506204233383e-03;
C[9642] = 1.421134382355e-05;
C[9643] = 1.658926755453e-04;
C[9644] = 2.653868437369e-03;
C[9645] = 3.006392529018e-06;
C[9646] = 4.528884979923e-06;
C[9647] = -1.115608142156e-05;
C[9648] = 3.794987410337e-01;
C[9649] = 2.539145010518e-02;
C[9650] = -4.133624080773e-03;
C[9651] = 1.155482883687e-01;
C[9652] = -2.467455884790e-02;
C[9653] = 7.795077186606e-03;
C[9654] = 3.881587895369e-01;
C[9655] = -4.544877140763e-03;
C[9656] = -1.551542844639e-03;
C[9657] = 3.096867608445e-02;
C[9658] = 2.101277078915e-03;
C[9659] = -1.070742596415e-04;
C[9660] = 1.322698162506e-01;
C[9661] = -2.000697795898e-03;
C[9662] = 3.783759870182e-04;
C[9663] = 3.166481125923e-02;
C[9664] = -2.761545731481e-04;
C[9665] = -1.917283547355e-04;
C[9666] = 6.046227275761e-02;
C[9667] = -5.649861114843e-04;
C[9668] = -7.765879060919e-04;
C[9669] = 2.638278256397e-03;
C[9670] = -1.766870679386e-05;
C[9671] = -3.880886160022e-05;
C[9672] = 2.521280505392e-02;
C[9673] = 3.794660925511e-01;
C[9674] = -5.272413760234e-03;
C[9675] = -2.453863125535e-02;
C[9676] = 1.104037331141e-01;
C[9677] = -2.106371197299e-03;
C[9678] = -4.483517092591e-03;
C[9679] = 3.908776135459e-01;
C[9680] = 6.791091139790e-03;
C[9681] = 2.089166335868e-03;
C[9682] = 3.136794065138e-02;
C[9683] = -1.444728149606e-04;
C[9684] = -1.991635956395e-03;
C[9685] = 1.314374916983e-01;
C[9686] = -5.729112038354e-04;
C[9687] = -2.724779067757e-04;
C[9688] = 3.195717550341e-02;
C[9689] = 6.622108904144e-04;
C[9690] = -5.604930229038e-04;
C[9691] = 6.146635679850e-02;
C[9692] = 2.238430574740e-03;
C[9693] = -1.760700969192e-05;
C[9694] = 2.687417714805e-03;
C[9695] = 1.059522314357e-04;
C[9696] = -4.598530317839e-03;
C[9697] = -5.811578268754e-03;
C[9698] = 3.506254405465e-01;
C[9699] = 8.013994772957e-03;
C[9700] = -1.853765380808e-03;
C[9701] = 1.257971696825e-01;
C[9702] = -1.465906196654e-03;
C[9703] = 6.892193069607e-03;
C[9704] = 4.025511037678e-01;
C[9705] = -1.329742908669e-04;
C[9706] = -1.750588226804e-04;
C[9707] = 2.888446165634e-02;
C[9708] = 3.929704927748e-04;
C[9709] = -5.560708160693e-04;
C[9710] = 1.328367866759e-01;
C[9711] = -1.866700110769e-04;
C[9712] = 6.681729924072e-04;
C[9713] = 3.296709695568e-02;
C[9714] = -7.701800306443e-04;
C[9715] = 2.244697613560e-03;
C[9716] = 6.470882344218e-02;
C[9717] = -3.872142603733e-05;
C[9718] = 1.060033670272e-04;
C[9719] = 2.833482759171e-03;
C[9720] = 5.205503605307e-02;
C[9721] = 4.048391107997e-03;
C[9722] = -1.245205470456e-03;
C[9723] = 3.884041433466e-01;
C[9724] = -5.159753547769e-03;
C[9725] = -5.385852905015e-04;
C[9726] = 4.082139749671e-02;
C[9727] = 1.783997340158e-03;
C[9728] = 3.867690239065e-03;
C[9729] = 2.517196896913e-03;
C[9730] = 2.093206824176e-04;
C[9731] = -6.478406818855e-05;
C[9732] = 3.168116817988e-02;
C[9733] = -3.175603371208e-04;
C[9734] = -1.248486173334e-04;
C[9735] = 1.279613739982e-01;
C[9736] = 1.872375761340e-04;
C[9737] = 3.588542539650e-04;
C[9738] = 5.163119860888e-01;
C[9739] = -9.290661105563e-04;
C[9740] = -1.943930233544e-03;
C[9741] = 3.785772647380e-02;
C[9742] = -8.066011245480e-05;
C[9743] = -1.689433057972e-04;
C[9744] = 4.071321133011e-03;
C[9745] = 5.272837529634e-02;
C[9746] = 1.783462372292e-04;
C[9747] = -5.187625597692e-03;
C[9748] = 3.921641041415e-01;
C[9749] = 7.920091746762e-03;
C[9750] = 1.787927853467e-03;
C[9751] = 3.466858301066e-02;
C[9752] = -1.236425416813e-02;
C[9753] = 2.104454040293e-04;
C[9754] = 2.559064319878e-03;
C[9755] = 1.439574672492e-05;
C[9756] = -3.190044702767e-04;
C[9757] = 3.204294154312e-02;
C[9758] = 7.366995708842e-04;
C[9759] = 1.874996103546e-04;
C[9760] = 1.274077348795e-01;
C[9761] = -1.130795132466e-03;
C[9762] = -9.258062070256e-04;
C[9763] = 5.179610856219e-01;
C[9764] = 4.319999599445e-03;
C[9765] = -8.051561696922e-05;
C[9766] = 3.800561088004e-02;
C[9767] = 3.788711141075e-04;
C[9768] = -1.234270801589e-03;
C[9769] = 2.008640665825e-04;
C[9770] = 4.984681183996e-02;
C[9771] = -5.750933593466e-04;
C[9772] = 7.876226703011e-03;
C[9773] = 4.038385942091e-01;
C[9774] = 3.873626689187e-03;
C[9775] = -1.235880838422e-02;
C[9776] = 1.891860241978e-02;
C[9777] = -6.373668069670e-05;
C[9778] = 1.584697791454e-05;
C[9779] = 2.419861306827e-03;
C[9780] = -1.266317223826e-04;
C[9781] = 7.345532612958e-04;
C[9782] = 3.305292965177e-02;
C[9783] = 3.592500173065e-04;
C[9784] = -1.130432080205e-03;
C[9785] = 1.259435538018e-01;
C[9786] = -1.939949258260e-03;
C[9787] = 4.326486928875e-03;
C[9788] = 5.249163573489e-01;
C[9789] = -1.687703194357e-04;
C[9790] = 3.791647317463e-04;
C[9791] = 3.860020424199e-02;
C[9792] = -7.019835695425e+00;
C[9793] = -6.741699578466e-02;
C[9794] = 5.155991355603e-02;
C[9795] = 2.124809095448e+00;
C[9796] = 5.042018361999e-02;
C[9797] = 1.875427782897e-02;
C[9798] = 8.131070821654e-01;
C[9799] = 4.459465070443e-02;
C[9800] = 3.614212369222e-03;
C[9801] = 6.792082155343e-02;
C[9802] = -2.289071871474e-02;
C[9803] = 7.022304611204e-03;
C[9804] = 3.815287203518e-01;
C[9805] = 1.993186723810e-02;
C[9806] = 3.189329689099e-03;
C[9807] = 5.249604069621e-02;
C[9808] = 3.110563661204e-03;
C[9809] = 2.004846789363e-04;
C[9810] = 8.196981673385e-02;
C[9811] = 5.068920683759e-03;
C[9812] = -7.951266033138e-05;
C[9813] = 2.548492380470e-03;
C[9814] = 1.845466444717e-04;
C[9815] = -5.395935131292e-06;
C[9816] = -6.419762047226e-02;
C[9817] = -7.036489624280e+00;
C[9818] = 5.464986278968e-02;
C[9819] = 4.801457883787e-02;
C[9820] = 2.126817845030e+00;
C[9821] = -1.212977722525e-04;
C[9822] = 4.370110363312e-02;
C[9823] = 8.439103492841e-01;
C[9824] = 2.464579449316e-02;
C[9825] = -2.272974994912e-02;
C[9826] = 5.864492759064e-02;
C[9827] = 6.949039234060e-03;
C[9828] = 1.983506339564e-02;
C[9829] = 3.896805768117e-01;
C[9830] = 2.302315296997e-03;
C[9831] = 3.051429449230e-03;
C[9832] = 5.471360105351e-02;
C[9833] = 1.759128457919e-03;
C[9834] = 4.916708484502e-03;
C[9835] = 8.633103809132e-02;
C[9836] = 4.191755930422e-03;
C[9837] = 1.793899045375e-04;
C[9838] = 2.711036545234e-03;
C[9839] = 1.502820873790e-04;
C[9840] = 5.454627643005e-02;
C[9841] = 5.860207433330e-02;
C[9842] = -7.130800822697e+00;
C[9843] = 1.889553157577e-02;
C[9844] = -4.223458927336e-04;
C[9845] = 1.999338040120e+00;
C[9846] = 2.701510573711e-03;
C[9847] = 2.342713162516e-02;
C[9848] = 7.980636623242e-01;
C[9849] = 7.171622754905e-03;
C[9850] = 7.146649811241e-03;
C[9851] = 5.649231610257e-02;
C[9852] = 3.251840992994e-03;
C[9853] = 2.354801068172e-03;
C[9854] = 3.603529952551e-01;
C[9855] = 1.296248115990e-04;
C[9856] = 1.671377472280e-03;
C[9857] = 5.175465845152e-02;
C[9858] = -2.881733197111e-04;
C[9859] = 3.945747903016e-03;
C[9860] = 8.326969947734e-02;
C[9861] = -1.370250069069e-05;
C[9862] = 1.413511531587e-04;
C[9863] = 2.616102965681e-03;
C[9864] = 2.127023904462e+00;
C[9865] = 5.179095275436e-02;
C[9866] = 2.124802392459e-02;
C[9867] = -6.231159218393e+00;
C[9868] = -5.736725293540e-02;
C[9869] = 1.134367807893e-02;
C[9870] = 2.128104125305e+00;
C[9871] = -4.748533353091e-02;
C[9872] = -4.916023002988e-03;
C[9873] = 3.816394608025e-01;
C[9874] = 2.002388209146e-02;
C[9875] = 3.369465610434e-03;
C[9876] = 1.144374220829e-01;
C[9877] = -1.911857140157e-02;
C[9878] = 6.016602878026e-04;
C[9879] = 3.870928286046e-01;
C[9880] = -4.358152709902e-03;
C[9881] = -2.285275955171e-03;
C[9882] = 1.020748810617e+00;
C[9883] = -1.174120875913e-02;
C[9884] = -1.241908363715e-02;
C[9885] = 6.041093385915e-02;
C[9886] = -5.953286878298e-04;
C[9887] = -7.769014726124e-04;
C[9888] = 4.820216301957e-02;
C[9889] = 2.125698926055e+00;
C[9890] = 5.641615483799e-03;
C[9891] = -5.464870108436e-02;
C[9892] = -6.254393120240e+00;
C[9893] = -1.716374223193e-02;
C[9894] = -4.654975474865e-02;
C[9895] = 2.109520900865e+00;
C[9896] = -3.181125793624e-03;
C[9897] = 1.982096620808e-02;
C[9898] = 3.896246308630e-01;
C[9899] = 2.657999136999e-03;
C[9900] = -1.898264380902e-02;
C[9901] = 1.013865595017e-01;
C[9902] = -1.041190042271e-02;
C[9903] = -4.294770177891e-03;
C[9904] = 3.893294534453e-01;
C[9905] = 6.602763469997e-03;
C[9906] = -1.163608889332e-02;
C[9907] = 1.035850611978e+00;
C[9908] = 3.563449679231e-02;
C[9909] = -5.933955001943e-04;
C[9910] = 6.143866859984e-02;
C[9911] = 2.288493413542e-03;
C[9912] = 1.366361031800e-02;
C[9913] = -3.471810217854e-03;
C[9914] = 2.000703805525e+00;
C[9915] = 1.572202980594e-02;
C[9916] = -1.211162590210e-02;
C[9917] = -6.222066831165e+00;
C[9918] = -3.620874621123e-03;
C[9919] = -1.665413521354e-03;
C[9920] = 2.168582071786e+00;
C[9921] = 2.934796313550e-03;
C[9922] = 2.134789674717e-03;
C[9923] = 3.604212835253e-01;
C[9924] = 8.205778741529e-04;
C[9925] = -1.015929460622e-02;
C[9926] = 1.161341084247e-01;
C[9927] = -2.199778697574e-03;
C[9928] = 6.702872102831e-03;
C[9929] = 4.017921369453e-01;
C[9930] = -1.227588198820e-02;
C[9931] = 3.578546882590e-02;
C[9932] = 1.090654361587e+00;
C[9933] = -7.744484195477e-04;
C[9934] = 2.290763289032e-03;
C[9935] = 6.473470697788e-02;
C[9936] = 8.056221772798e-01;
C[9937] = 5.703158452150e-02;
C[9938] = -1.630027930866e-02;
C[9939] = 2.133011201499e+00;
C[9940] = -6.007448385219e-02;
C[9941] = 1.492554350193e-02;
C[9942] = -7.148343931613e+00;
C[9943] = 1.318986797364e-03;
C[9944] = 1.443257136444e-02;
C[9945] = 5.212179545193e-02;
C[9946] = 3.717953493648e-03;
C[9947] = -8.204646825197e-04;
C[9948] = 3.873381824143e-01;
C[9949] = -4.971006633068e-03;
C[9950] = -1.272457791421e-03;
C[9951] = 4.184982035963e-02;
C[9952] = 2.049996806782e-03;
C[9953] = 4.273545081239e-03;
C[9954] = 4.131982253086e+00;
C[9955] = -4.649002802086e-03;
C[9956] = -1.015938789015e-02;
C[9957] = 5.162622687861e-01;
C[9958] = -1.000715334899e-03;
C[9959] = -2.116667062772e-03;
C[9960] = 5.718459310544e-02;
C[9961] = 8.153789212174e-01;
C[9962] = 4.034385545062e-03;
C[9963] = -6.034030266949e-02;
C[9964] = 2.135250712777e+00;
C[9965] = 1.889256002174e-02;
C[9966] = 1.397597063539e-03;
C[9967] = -7.166734377567e+00;
C[9968] = -3.774637688026e-02;
C[9969] = 3.740060781255e-03;
C[9970] = 5.328702965017e-02;
C[9971] = 7.017401682749e-04;
C[9972] = -5.000901166831e-03;
C[9973] = 3.906159440409e-01;
C[9974] = 7.730770779986e-03;
C[9975] = 2.053927320091e-03;
C[9976] = 3.570382274114e-02;
C[9977] = -1.276625834524e-02;
C[9978] = -4.599858075471e-03;
C[9979] = 4.135523673849e+00;
C[9980] = 1.545458273011e-02;
C[9981] = -9.974102043233e-04;
C[9982] = 5.178154331138e-01;
C[9983] = 4.308493737624e-03;
C[9984] = -1.650774260109e-02;
C[9985] = 3.978094314877e-03;
C[9986] = 7.721607843573e-01;
C[9987] = 1.461296670287e-02;
C[9988] = 1.852158547081e-02;
C[9989] = 2.194331880612e+00;
C[9990] = 1.455130036687e-02;
C[9991] = -3.763746120198e-02;
C[9992] = -7.214084731836e+00;
C[9993] = -8.056130695795e-04;
C[9994] = 7.257434490048e-04;
C[9995] = 5.045951455317e-02;
C[9996] = -1.308826469879e-03;
C[9997] = 7.687899033219e-03;
C[9998] = 4.030796273866e-01;
C[9999] = 4.279481531360e-03;
C[10000] = -1.276081256132e-02;
C[10001] = 1.901349149339e-02;
C[10002] = -1.009996147867e-02;
C[10003] = 1.555244882074e-02;
C[10004] = 4.166700837818e+00;
C[10005] = -2.112543861026e-03;
C[10006] = 4.314583385530e-03;
C[10007] = 5.248678289465e-01;
C[10008] = 0.000000000000e+00;
C[10009] = 0.000000000000e+00;
C[10010] = 0.000000000000e+00;
C[10011] = 0.000000000000e+00;
C[10012] = 0.000000000000e+00;
C[10013] = 0.000000000000e+00;
C[10014] = 0.000000000000e+00;
C[10015] = 0.000000000000e+00;
C[10016] = 0.000000000000e+00;
C[10017] = 0.000000000000e+00;
C[10018] = 0.000000000000e+00;
C[10019] = 0.000000000000e+00;
C[10020] = 0.000000000000e+00;
C[10021] = 0.000000000000e+00;
C[10022] = 0.000000000000e+00;
C[10023] = 0.000000000000e+00;
C[10024] = 0.000000000000e+00;
C[10025] = 0.000000000000e+00;
C[10026] = 1.000000000000e+00;
C[10027] = 0.000000000000e+00;
C[10028] = 0.000000000000e+00;
C[10029] = 2.000000000000e-01;
C[10030] = -0.000000000000e+00;
C[10031] = -0.000000000000e+00;
C[10032] = 0.000000000000e+00;
C[10033] = 0.000000000000e+00;
C[10034] = 0.000000000000e+00;
C[10035] = 0.000000000000e+00;
C[10036] = 0.000000000000e+00;
C[10037] = 0.000000000000e+00;
C[10038] = 0.000000000000e+00;
C[10039] = 0.000000000000e+00;
C[10040] = 0.000000000000e+00;
C[10041] = 0.000000000000e+00;
C[10042] = 0.000000000000e+00;
C[10043] = 0.000000000000e+00;
C[10044] = 0.000000000000e+00;
C[10045] = 0.000000000000e+00;
C[10046] = 0.000000000000e+00;
C[10047] = 0.000000000000e+00;
C[10048] = 0.000000000000e+00;
C[10049] = 0.000000000000e+00;
C[10050] = 0.000000000000e+00;
C[10051] = 1.000000000000e+00;
C[10052] = 0.000000000000e+00;
C[10053] = -0.000000000000e+00;
C[10054] = 2.000000000000e-01;
C[10055] = -0.000000000000e+00;
C[10056] = 0.000000000000e+00;
C[10057] = 0.000000000000e+00;
C[10058] = 0.000000000000e+00;
C[10059] = 0.000000000000e+00;
C[10060] = 0.000000000000e+00;
C[10061] = 0.000000000000e+00;
C[10062] = 0.000000000000e+00;
C[10063] = 0.000000000000e+00;
C[10064] = 0.000000000000e+00;
C[10065] = 0.000000000000e+00;
C[10066] = 0.000000000000e+00;
C[10067] = 0.000000000000e+00;
C[10068] = 0.000000000000e+00;
C[10069] = 0.000000000000e+00;
C[10070] = 0.000000000000e+00;
C[10071] = 0.000000000000e+00;
C[10072] = 0.000000000000e+00;
C[10073] = 0.000000000000e+00;
C[10074] = 0.000000000000e+00;
C[10075] = 0.000000000000e+00;
C[10076] = 1.000000000000e+00;
C[10077] = -0.000000000000e+00;
C[10078] = -0.000000000000e+00;
C[10079] = 2.000000000000e-01;
C[10080] = 7.303437068356e-02;
C[10081] = -1.398913675512e-02;
C[10082] = 5.711526464867e-03;
C[10083] = 3.768545500168e-01;
C[10084] = 1.045887585278e-02;
C[10085] = 5.259109883775e-03;
C[10086] = 5.133967337615e-02;
C[10087] = 5.055418753155e-04;
C[10088] = 1.091414366243e-04;
C[10089] = 1.307845513279e-01;
C[10090] = -8.693614983227e-04;
C[10091] = 5.890371779111e-04;
C[10092] = 3.056964715919e-02;
C[10093] = 5.338366010428e-04;
C[10094] = 2.705107658236e-04;
C[10095] = 2.477010961903e-03;
C[10096] = 3.594467614326e-06;
C[10097] = -6.075669722072e-06;
C[10098] = 2.472496774780e-03;
C[10099] = -5.069427987224e-05;
C[10100] = -4.847953715049e-05;
C[10101] = -1.807976138415e-05;
C[10102] = -4.779966222325e-06;
C[10103] = -2.658435141303e-06;
C[10104] = -1.377950922592e-02;
C[10105] = 5.036369372795e-02;
C[10106] = -1.430884782152e-04;
C[10107] = 1.029422656591e-02;
C[10108] = 3.983208385035e-01;
C[10109] = 9.685484679880e-03;
C[10110] = 4.442300579732e-04;
C[10111] = 5.621568529630e-02;
C[10112] = 3.495223557001e-03;
C[10113] = -8.553863297094e-04;
C[10114] = 1.284684864197e-01;
C[10115] = -1.045464984165e-04;
C[10116] = 5.215471748727e-04;
C[10117] = 3.270788570796e-02;
C[10118] = 1.015968431323e-03;
C[10119] = 1.220522352745e-06;
C[10120] = 2.741878082169e-03;
C[10121] = 1.904414391648e-04;
C[10122] = -5.359778919662e-05;
C[10123] = 2.819629082522e-03;
C[10124] = 2.819227770089e-04;
C[10125] = -4.633364559417e-06;
C[10126] = -1.338526975970e-05;
C[10127] = 3.244881228866e-06;
C[10128] = 5.787671810880e-03;
C[10129] = -1.297087421751e-04;
C[10130] = 4.765759896844e-02;
C[10131] = 5.230555898927e-03;
C[10132] = 9.835446336830e-03;
C[10133] = 3.702691873517e-01;
C[10134] = 5.919319307288e-05;
C[10135] = 3.355962752101e-03;
C[10136] = 5.396509391137e-02;
C[10137] = 5.941135343120e-04;
C[10138] = -1.036545160138e-04;
C[10139] = 1.284872759895e-01;
C[10140] = 2.664414026831e-04;
C[10141] = 1.023021990632e-03;
C[10142] = 3.051366022084e-02;
C[10143] = -7.828640867386e-06;
C[10144] = 1.826232199703e-04;
C[10145] = 2.644926395562e-03;
C[10146] = -4.945697829898e-05;
C[10147] = 2.654799147458e-04;
C[10148] = 2.801966243067e-03;
C[10149] = -2.421095887759e-06;
C[10150] = 3.093488960543e-06;
C[10151] = -1.202006014457e-05;
C[10152] = 3.766457986117e-01;
C[10153] = 1.072380531005e-02;
C[10154] = 5.196545833842e-03;
C[10155] = 1.203038055376e-01;
C[10156] = -9.985995946655e-03;
C[10157] = -2.106103847948e-03;
C[10158] = 3.874950749929e-01;
C[10159] = -1.510820876037e-03;
C[10160] = -1.522500822431e-03;
C[10161] = 3.055573039885e-02;
C[10162] = 5.501857578152e-04;
C[10163] = 2.641740650107e-04;
C[10164] = 1.327869200659e-01;
C[10165] = -4.957773921688e-04;
C[10166] = -9.478653697222e-05;
C[10167] = 3.162682563477e-02;
C[10168] = -6.977945725860e-05;
C[10169] = -9.254407515598e-05;
C[10170] = 6.022040037259e-02;
C[10171] = -2.258983884362e-04;
C[10172] = -4.976689578096e-04;
C[10173] = 2.628329655700e-03;
C[10174] = -5.800484045318e-06;
C[10175] = -1.917249497142e-05;
C[10176] = 1.044725155648e-02;
C[10177] = 3.982699647186e-01;
C[10178] = 9.387658970795e-03;
C[10179] = -9.752229045091e-03;
C[10180] = 9.254376888014e-02;
C[10181] = -1.822802786111e-02;
C[10182] = -1.427454572885e-03;
C[10183] = 3.881761272216e-01;
C[10184] = 5.669268758390e-03;
C[10185] = 5.330616479560e-04;
C[10186] = 3.270449412230e-02;
C[10187] = 9.931694995629e-04;
C[10188] = -4.801929320645e-04;
C[10189] = 1.300728658237e-01;
C[10190] = -1.816139783841e-03;
C[10191] = -6.619295313341e-05;
C[10192] = 3.191013975944e-02;
C[10193] = 6.536480348949e-04;
C[10194] = -2.320680114494e-04;
C[10195] = 6.136936002906e-02;
C[10196] = 2.352262780630e-03;
C[10197] = -6.452372648298e-06;
C[10198] = 2.695443818595e-03;
C[10199] = 1.174805873702e-04;
C[10200] = 5.005944435683e-03;
C[10201] = 9.183782796130e-03;
C[10202] = 3.698662880121e-01;
C[10203] = -1.968072410968e-03;
C[10204] = -1.815430172212e-02;
C[10205] = 1.072729626878e-01;
C[10206] = -1.441851302696e-03;
C[10207] = 5.853006384691e-03;
C[10208] = 3.988884670099e-01;
C[10209] = 2.536330692842e-04;
C[10210] = 9.825216390728e-04;
C[10211] = 3.048680026488e-02;
C[10212] = -8.558444117358e-05;
C[10213] = -1.811224707908e-03;
C[10214] = 1.313398663370e-01;
C[10215] = -8.914171667709e-05;
C[10216] = 6.639387803103e-04;
C[10217] = 3.272678661143e-02;
C[10218] = -5.084972150699e-04;
C[10219] = 2.348805390764e-03;
C[10220] = 6.434060067519e-02;
C[10221] = -2.010090833291e-05;
C[10222] = 1.164522574377e-04;
C[10223] = 2.822109974843e-03;
C[10224] = 5.145576699348e-02;
C[10225] = 1.761654121964e-03;
C[10226] = 5.861907091258e-04;
C[10227] = 3.876906106595e-01;
C[10228] = -2.684074577691e-03;
C[10229] = -1.889282296413e-03;
C[10230] = 4.404142604100e-02;
C[10231] = 1.833225418660e-03;
C[10232] = 3.775660972419e-03;
C[10233] = 2.484750536392e-03;
C[10234] = 8.904879328545e-05;
C[10235] = 2.730452686948e-05;
C[10236] = 3.163986134588e-02;
C[10237] = -1.499676201204e-04;
C[10238] = -1.189704495916e-04;
C[10239] = 1.282676142157e-01;
C[10240] = 1.337926531813e-04;
C[10241] = 2.761826430226e-04;
C[10242] = 5.144406457764e-01;
C[10243] = -1.015904875448e-03;
C[10244] = -2.439690687553e-03;
C[10245] = 3.762493253539e-02;
C[10246] = -7.215266695959e-05;
C[10247] = -1.848034588742e-04;
C[10248] = 1.819161117776e-03;
C[10249] = 5.619934080448e-02;
C[10250] = 2.998872486300e-03;
C[10251] = -2.739817171772e-03;
C[10252] = 3.887127179455e-01;
C[10253] = 6.174126122879e-03;
C[10254] = 1.822804083739e-03;
C[10255] = 3.702769278927e-02;
C[10256] = -1.360311308317e-02;
C[10257] = 9.116908377838e-05;
C[10258] = 2.740788449380e-03;
C[10259] = 1.588172022502e-04;
C[10260] = -1.517125436409e-04;
C[10261] = 3.194591247436e-02;
C[10262] = 6.853467628562e-04;
C[10263] = 1.330978975199e-04;
C[10264] = 1.276052747084e-01;
C[10265] = -1.232113017227e-03;
C[10266] = -1.006111781844e-03;
C[10267] = 5.154015954497e-01;
C[10268] = 4.287606351211e-03;
C[10269] = -7.162482077459e-05;
C[10270] = 3.772266535165e-02;
C[10271] = 3.835490129528e-04;
C[10272] = 6.608993218880e-04;
C[10273] = 3.133165956043e-03;
C[10274] = 5.302078228769e-02;
C[10275] = -1.958978063893e-03;
C[10276] = 6.027398409366e-03;
C[10277] = 3.995909452165e-01;
C[10278] = 3.754780173031e-03;
C[10279] = -1.360811493710e-02;
C[10280] = 1.921292616303e-02;
C[10281] = 3.070818929550e-05;
C[10282] = 1.663042657675e-04;
C[10283] = 2.581972287317e-03;
C[10284] = -1.216425579201e-04;
C[10285] = 6.775233449600e-04;
C[10286] = 3.277361849187e-02;
C[10287] = 2.747905897300e-04;
C[10288] = -1.232446474156e-03;
C[10289] = 1.259947215267e-01;
C[10290] = -2.424479173410e-03;
C[10291] = 4.303721929932e-03;
C[10292] = 5.254900532317e-01;
C[10293] = -1.839203248576e-04;
C[10294] = 3.844314749524e-04;
C[10295] = 3.866294083987e-02;
C[10296] = -7.011466160956e+00;
C[10297] = -3.499612603751e-02;
C[10298] = 4.911778629976e-02;
C[10299] = 2.127778928862e+00;
C[10300] = 6.667247945591e-03;
C[10301] = 2.845738639210e-02;
C[10302] = 7.961638417839e-01;
C[10303] = -8.809785489892e-04;
C[10304] = -2.495297555727e-04;
C[10305] = 7.522557847722e-02;
C[10306] = -5.661292437436e-03;
C[10307] = 9.369117962010e-03;
C[10308] = 3.763207379560e-01;
C[10309] = 6.584645990125e-04;
C[10310] = 1.963876042807e-03;
C[10311] = 5.131683318661e-02;
C[10312] = -5.451134547465e-05;
C[10313] = -6.337085301606e-05;
C[10314] = 7.981226411552e-02;
C[10315] = -3.050392028540e-04;
C[10316] = -6.666486085015e-04;
C[10317] = 2.478454890892e-03;
C[10318] = -1.016932602677e-05;
C[10319] = -2.719174330092e-05;
C[10320] = -3.080357545353e-02;
C[10321] = -7.068294910387e+00;
C[10322] = 3.727647763384e-02;
C[10323] = 2.891646140124e-03;
C[10324] = 2.161721256982e+00;
C[10325] = 2.992760054433e-02;
C[10326] = -1.442797559147e-03;
C[10327] = 8.667886486950e-01;
C[10328] = 5.214941506271e-02;
C[10329] = -5.451664908236e-03;
C[10330] = 4.778637397303e-02;
C[10331] = 5.417169308206e-04;
C[10332] = 4.726862555976e-04;
C[10333] = 3.999346681929e-01;
C[10334] = 1.246182335495e-02;
C[10335] = -7.686476019796e-05;
C[10336] = 5.640233809613e-02;
C[10337] = 3.669547030155e-03;
C[10338] = -3.753355252204e-04;
C[10339] = 8.911796368941e-02;
C[10340] = 7.571030577549e-03;
C[10341] = -1.114542630274e-05;
C[10342] = 2.829622130322e-03;
C[10343] = 2.714582273889e-04;
C[10344] = 5.064069322001e-02;
C[10345] = 3.754407235464e-02;
C[10346] = -7.161527760064e+00;
C[10347] = 2.726285629092e-02;
C[10348] = 3.186145559620e-02;
C[10349] = 2.051094534522e+00;
C[10350] = -5.500524457265e-04;
C[10351] = 5.040410138527e-02;
C[10352] = 8.352537100303e-01;
C[10353] = 9.445263308023e-03;
C[10354] = 5.550966668608e-04;
C[10355] = 4.465273681752e-02;
C[10356] = 1.904459219100e-03;
C[10357] = 1.259982806021e-02;
C[10358] = 3.766093529202e-01;
C[10359] = -7.677033478471e-05;
C[10360] = 3.537744984389e-03;
C[10361] = 5.430224412734e-02;
C[10362] = -7.199047945644e-04;
C[10363] = 7.144073589697e-03;
C[10364] = 8.765523408523e-02;
C[10365] = -2.842790165977e-05;
C[10366] = 2.554545015220e-04;
C[10367] = 2.767967817041e-03;
C[10368] = 2.123603900760e+00;
C[10369] = 1.148322102286e-02;
C[10370] = 2.658265498922e-02;
C[10371] = -6.221526467784e+00;
C[10372] = -1.536277273305e-02;
C[10373] = -8.063508128531e-04;
C[10374] = 2.142309303890e+00;
C[10375] = -1.543019592605e-03;
C[10376] = 1.575203089360e-03;
C[10377] = 3.761119865509e-01;
C[10378] = 9.022649997343e-04;
C[10379] = 1.870449154015e-03;
C[10380] = 1.206661645957e-01;
C[10381] = -8.758144450695e-04;
C[10382] = 2.026268666397e-04;
C[10383] = 3.879272365174e-01;
C[10384] = -3.973016738865e-04;
C[10385] = -3.688595552190e-04;
C[10386] = 1.017747394213e+00;
C[10387] = -3.734786874730e-03;
C[10388] = -7.343061533641e-03;
C[10389] = 6.009107808671e-02;
C[10390] = -2.104321165745e-04;
C[10391] = -4.364817897354e-04;
C[10392] = 6.434762019543e-03;
C[10393] = 2.160703781284e+00;
C[10394] = 2.290570827550e-02;
C[10395] = -1.068743470178e-02;
C[10396] = -6.289853909361e+00;
C[10397] = -4.473637801404e-02;
C[10398] = -5.948896063802e-04;
C[10399] = 2.088901901420e+00;
C[10400] = -2.604268278945e-02;
C[10401] = 6.468403027101e-04;
C[10402] = 3.998837944080e-01;
C[10403] = 1.215204069417e-02;
C[10404] = -6.420475435059e-04;
C[10405] = 8.958995363565e-02;
C[10406] = -2.105261763697e-02;
C[10407] = -3.549371318102e-04;
C[10408] = 3.894656583031e-01;
C[10409] = 5.786889637203e-03;
C[10410] = -3.824393025606e-03;
C[10411] = 1.033510985549e+00;
C[10412] = 3.577896729136e-02;
C[10413] = -2.200141644966e-04;
C[10414] = 6.128652218908e-02;
C[10415] = 2.384331969418e-03;
C[10416] = 2.339407743026e-02;
C[10417] = 1.989356286933e-02;
C[10418] = 2.043036547731e+00;
C[10419] = 1.954277926738e-03;
C[10420] = -4.326185523414e-02;
C[10421] = -6.264277767741e+00;
C[10422] = 2.415489966033e-03;
C[10423] = -2.348791214074e-02;
C[10424] = 2.128364844888e+00;
C[10425] = 1.710710594715e-03;
C[10426] = 1.196012147120e-02;
C[10427] = 3.762064535806e-01;
C[10428] = 3.406583036192e-04;
C[10429] = -2.097889149798e-02;
C[10430] = 1.019058462034e-01;
C[10431] = -3.290940310165e-04;
C[10432] = 5.966948324863e-03;
C[10433] = 3.977724142801e-01;
C[10434] = -7.492967804530e-03;
C[10435] = 3.582668108401e-02;
C[10436] = 1.082424547108e+00;
C[10437] = -4.485325824856e-04;
C[10438] = 2.369678783229e-03;
C[10439] = 6.435996022681e-02;
C[10440] = 7.984857141305e-01;
C[10441] = 2.490568372066e-02;
C[10442] = 9.989897875332e-03;
C[10443] = 2.146220017223e+00;
C[10444] = -2.572728970249e-02;
C[10445] = -6.533129908320e-03;
C[10446] = -7.131517502556e+00;
C[10447] = 4.184645346758e-03;
C[10448] = 1.262813282448e-02;
C[10449] = 5.143292680394e-02;
C[10450] = 1.240559303793e-03;
C[10451] = 4.502271812682e-04;
C[10452] = 3.881227721841e-01;
C[10453] = -1.611557136616e-03;
C[10454] = -7.765250247341e-04;
C[10455] = 4.587635810260e-02;
C[10456] = 1.118130223110e-03;
C[10457] = 2.582907143202e-03;
C[10458] = 4.110475839977e+00;
C[10459] = -5.144511143384e-03;
C[10460] = -1.458614562708e-02;
C[10461] = 5.122026621457e-01;
C[10462] = -7.581805791019e-04;
C[10463] = -2.190716282166e-03;
C[10464] = 2.539140630023e-02;
C[10465] = 8.664617588585e-01;
C[10466] = 4.326229606924e-02;
C[10467] = -2.612294550729e-02;
C[10468] = 2.099633715898e+00;
C[10469] = -1.706551737697e-02;
C[10470] = 3.976218648331e-03;
C[10471] = -7.155519701463e+00;
C[10472] = -4.374717261954e-02;
C[10473] = 1.259107896986e-03;
C[10474] = 5.638599360431e-02;
C[10475] = 3.180654718587e-03;
C[10476] = -1.626297969621e-03;
C[10477] = 3.900022490270e-01;
C[10478] = 6.288068063051e-03;
C[10479] = 1.107708888189e-03;
C[10480] = 3.804315461427e-02;
C[10481] = -1.395729272065e-02;
C[10482] = -4.973149281432e-03;
C[10483] = 4.105069679656e+00;
C[10484] = 1.374326487920e-02;
C[10485] = -7.474509182900e-04;
C[10486] = 5.129017955768e-01;
C[10487] = 4.313192112698e-03;
C[10488] = 1.078562794970e-02;
C[10489] = 4.490826304354e-02;
C[10490] = 8.163674775567e-01;
C[10491] = -7.154341739888e-03;
C[10492] = -1.888008976993e-02;
C[10493] = 2.142414409019e+00;
C[10494] = 1.221051683671e-02;
C[10495] = -4.384720969806e-02;
C[10496] = -7.220137905791e+00;
C[10497] = 4.883870322476e-04;
C[10498] = 3.307489429197e-03;
C[10499] = 5.335793250366e-02;
C[10500] = -8.053367966814e-04;
C[10501] = 6.145019288179e-03;
C[10502] = 3.984748924867e-01;
C[10503] = 2.562026343813e-03;
C[10504] = -1.396229457458e-02;
C[10505] = 1.933897731843e-02;
C[10506] = -1.431637716810e-02;
C[10507] = 1.404532702964e-02;
C[10508] = 4.175454387078e+00;
C[10509] = -2.174722608037e-03;
C[10510] = 4.339270527974e-03;
C[10511] = 5.260717130882e-01;
C[10512] = 0.000000000000e+00;
C[10513] = 0.000000000000e+00;
C[10514] = 0.000000000000e+00;
C[10515] = 0.000000000000e+00;
C[10516] = 0.000000000000e+00;
C[10517] = 0.000000000000e+00;
C[10518] = 0.000000000000e+00;
C[10519] = 0.000000000000e+00;
C[10520] = 0.000000000000e+00;
C[10521] = 0.000000000000e+00;
C[10522] = 0.000000000000e+00;
C[10523] = 0.000000000000e+00;
C[10524] = 0.000000000000e+00;
C[10525] = 0.000000000000e+00;
C[10526] = 0.000000000000e+00;
C[10527] = 0.000000000000e+00;
C[10528] = 0.000000000000e+00;
C[10529] = 0.000000000000e+00;
C[10530] = 1.000000000000e+00;
C[10531] = 0.000000000000e+00;
C[10532] = 0.000000000000e+00;
C[10533] = 2.000000000000e-01;
C[10534] = -0.000000000000e+00;
C[10535] = -0.000000000000e+00;
C[10536] = 0.000000000000e+00;
C[10537] = 0.000000000000e+00;
C[10538] = 0.000000000000e+00;
C[10539] = 0.000000000000e+00;
C[10540] = 0.000000000000e+00;
C[10541] = 0.000000000000e+00;
C[10542] = 0.000000000000e+00;
C[10543] = 0.000000000000e+00;
C[10544] = 0.000000000000e+00;
C[10545] = 0.000000000000e+00;
C[10546] = 0.000000000000e+00;
C[10547] = 0.000000000000e+00;
C[10548] = 0.000000000000e+00;
C[10549] = 0.000000000000e+00;
C[10550] = 0.000000000000e+00;
C[10551] = 0.000000000000e+00;
C[10552] = 0.000000000000e+00;
C[10553] = 0.000000000000e+00;
C[10554] = 0.000000000000e+00;
C[10555] = 1.000000000000e+00;
C[10556] = 0.000000000000e+00;
C[10557] = -0.000000000000e+00;
C[10558] = 2.000000000000e-01;
C[10559] = -0.000000000000e+00;
C[10560] = 0.000000000000e+00;
C[10561] = 0.000000000000e+00;
C[10562] = 0.000000000000e+00;
C[10563] = 0.000000000000e+00;
C[10564] = 0.000000000000e+00;
C[10565] = 0.000000000000e+00;
C[10566] = 0.000000000000e+00;
C[10567] = 0.000000000000e+00;
C[10568] = 0.000000000000e+00;
C[10569] = 0.000000000000e+00;
C[10570] = 0.000000000000e+00;
C[10571] = 0.000000000000e+00;
C[10572] = 0.000000000000e+00;
C[10573] = 0.000000000000e+00;
C[10574] = 0.000000000000e+00;
C[10575] = 0.000000000000e+00;
C[10576] = 0.000000000000e+00;
C[10577] = 0.000000000000e+00;
C[10578] = 0.000000000000e+00;
C[10579] = 0.000000000000e+00;
C[10580] = 1.000000000000e+00;
C[10581] = -0.000000000000e+00;
C[10582] = -0.000000000000e+00;
C[10583] = 2.000000000000e-01;
C[10584] = 7.753251151053e-02;
C[10585] = -1.056366928704e-03;
C[10586] = 9.115334720828e-03;
C[10587] = 3.785797510017e-01;
C[10588] = -5.163158673200e-03;
C[10589] = -2.085764266791e-03;
C[10590] = 5.184935706757e-02;
C[10591] = -1.053330121297e-03;
C[10592] = -5.712929641361e-04;
C[10593] = 1.310580193366e-01;
C[10594] = 2.329134362713e-04;
C[10595] = 5.854313798808e-04;
C[10596] = 3.074642504114e-02;
C[10597] = -5.364303099895e-04;
C[10598] = -2.567789616412e-04;
C[10599] = 2.501782837006e-03;
C[10600] = -5.852169205009e-05;
C[10601] = -3.303986733924e-05;
C[10602] = 2.459656405909e-03;
C[10603] = -7.815728506826e-05;
C[10604] = -6.050337604022e-05;
C[10605] = -1.820254143988e-05;
C[10606] = -1.583379365019e-06;
C[10607] = -2.482798137884e-06;
C[10608] = -9.525121404213e-04;
C[10609] = 3.909291185669e-02;
C[10610] = 6.231769631794e-03;
C[10611] = -5.176563352762e-03;
C[10612] = 3.959053706577e-01;
C[10613] = 1.236692127743e-02;
C[10614] = -1.022555113487e-03;
C[10615] = 5.629958589288e-02;
C[10616] = 3.516520446469e-03;
C[10617] = 2.398370888235e-04;
C[10618] = 1.271275228257e-01;
C[10619] = 2.640059497240e-04;
C[10620] = -5.374474917970e-04;
C[10621] = 3.250317641818e-02;
C[10622] = 1.083002963425e-03;
C[10623] = -5.658178178836e-05;
C[10624] = 2.742080211518e-03;
C[10625] = 1.813745675272e-04;
C[10626] = -7.550014805172e-05;
C[10627] = 2.766296863738e-03;
C[10628] = 2.507456027397e-04;
C[10629] = -1.588949747474e-06;
C[10630] = -1.399042258044e-05;
C[10631] = 1.417135151038e-06;
C[10632] = 9.115844151870e-03;
C[10633] = 6.263885804052e-03;
C[10634] = 5.501345966624e-02;
C[10635] = -2.172703977960e-03;
C[10636] = 1.269945141126e-02;
C[10637] = 3.804063076323e-01;
C[10638] = -5.348546249372e-04;
C[10639] = 3.441817373223e-03;
C[10640] = 5.438425404038e-02;
C[10641] = 5.854653419502e-04;
C[10642] = 2.661470278745e-04;
C[10643] = 1.296258901848e-01;
C[10644] = -2.623329537276e-04;
C[10645] = 1.102735698958e-03;
C[10646] = 3.100668694999e-02;
C[10647] = -3.048877552086e-05;
C[10648] = 1.759877517385e-04;
C[10649] = 2.643768007163e-03;
C[10650] = -5.402816439144e-05;
C[10651] = 2.314199044248e-04;
C[10652] = 2.763060332422e-03;
C[10653] = -2.348588858864e-06;
C[10654] = 1.059228865382e-06;
C[10655] = -1.672442899720e-05;
C[10656] = 3.786494323304e-01;
C[10657] = -5.678093537115e-03;
C[10658] = -1.496459116217e-03;
C[10659] = 1.184803235029e-01;
C[10660] = 4.932376669555e-03;
C[10661] = 2.689283226976e-03;
C[10662] = 3.887225195415e-01;
C[10663] = 9.356303023629e-04;
C[10664] = 4.351466364753e-04;
C[10665] = 3.075107046306e-02;
C[10666] = -5.708828374205e-04;
C[10667] = -2.172499629447e-04;
C[10668] = 1.326396819387e-01;
C[10669] = 5.417812330191e-04;
C[10670] = 2.781283161478e-04;
C[10671] = 3.164350797161e-02;
C[10672] = 4.844715273796e-05;
C[10673] = 1.600302814044e-05;
C[10674] = 5.926716951096e-02;
C[10675] = 3.347206088561e-05;
C[10676] = -7.350197210859e-05;
C[10677] = 2.580788750206e-03;
C[10678] = -4.403113031482e-07;
C[10679] = -4.041245247385e-06;
C[10680] = -5.732087806856e-03;
C[10681] = 3.978128434440e-01;
C[10682] = 1.302976753508e-02;
C[10683] = 4.977253554828e-03;
C[10684] = 8.987483287769e-02;
C[10685] = -2.061319731906e-02;
C[10686] = 9.100896346703e-04;
C[10687] = 3.911460040579e-01;
C[10688] = 5.905213203934e-03;
C[10689] = -5.743589189000e-04;
C[10690] = 3.263034127060e-02;
C[10691] = 1.124756773880e-03;
C[10692] = 5.447730253706e-04;
C[10693] = 1.300385818446e-01;
C[10694] = -1.903562481127e-03;
C[10695] = 4.666841059081e-05;
C[10696] = 3.201929121640e-02;
C[10697] = 6.853366549080e-04;
C[10698] = 2.917724895946e-05;
C[10699] = 6.042441743765e-02;
C[10700] = 2.390189641151e-03;
C[10701] = -4.378571665254e-07;
C[10702] = 2.644202362978e-03;
C[10703] = 1.197893623425e-04;
C[10704] = -1.475859382184e-03;
C[10705] = 1.281877972284e-02;
C[10706] = 3.791602370311e-01;
C[10707] = 2.720072357474e-03;
C[10708] = -2.061513128380e-02;
C[10709] = 1.009129911491e-01;
C[10710] = 3.956601127918e-04;
C[10711] = 6.060416033838e-03;
C[10712] = 3.973600016021e-01;
C[10713] = -2.161186360008e-04;
C[10714] = 1.113126859786e-03;
C[10715] = 3.092361557657e-02;
C[10716] = 2.801809248477e-04;
C[10717] = -1.903691412110e-03;
C[10718] = 1.310386686658e-01;
C[10719] = 1.322675291219e-05;
C[10720] = 6.941176212361e-04;
C[10721] = 3.267293129208e-02;
C[10722] = -8.064621417306e-05;
C[10723] = 2.350349141721e-03;
C[10724] = 6.456389894977e-02;
C[10725] = -3.993720374404e-06;
C[10726] = 1.170978832018e-04;
C[10727] = 2.837152686047e-03;
C[10728] = 5.166836027950e-02;
C[10729] = -8.725034339815e-04;
C[10730] = -4.111750081086e-04;
C[10731] = 3.889455541018e-01;
C[10732] = 6.050879864887e-04;
C[10733] = -2.524113092879e-05;
C[10734] = 4.899959553433e-02;
C[10735] = 2.466599047330e-04;
C[10736] = 3.337745008579e-04;
C[10737] = 2.489716384468e-03;
C[10738] = -4.657833648802e-05;
C[10739] = -2.224346773229e-05;
C[10740] = 3.165837694230e-02;
C[10741] = 2.633496737870e-05;
C[10742] = -1.483333000251e-05;
C[10743] = 1.287189166615e-01;
C[10744] = -3.620285237040e-06;
C[10745] = -7.773119416174e-05;
C[10746] = 5.080607797510e-01;
C[10747] = 1.691951174572e-05;
C[10748] = 1.928239773423e-04;
C[10749] = 3.715151619495e-02;
C[10750] = 2.216038200864e-05;
C[10751] = 1.178266470726e-04;
C[10752] = -8.952011863669e-04;
C[10753] = 5.637323588467e-02;
C[10754] = 3.609158024518e-03;
C[10755] = 6.357969407411e-04;
C[10756] = 3.914158474777e-01;
C[10757] = 6.090633745258e-03;
C[10758] = 2.341940798857e-04;
C[10759] = 4.136393417450e-02;
C[10760] = -1.348709307705e-02;
C[10761] = -4.797976305474e-05;
C[10762] = 2.746990210971e-03;
C[10763] = 1.871437951583e-04;
C[10764] = 2.845826196318e-05;
C[10765] = 3.203728077772e-02;
C[10766] = 6.961321353307e-04;
C[10767] = -4.451340226863e-06;
C[10768] = 1.280688628108e-01;
C[10769] = -1.199184662462e-03;
C[10770] = 2.218006120404e-05;
C[10771] = 5.081850418713e-01;
C[10772] = 3.716046112503e-03;
C[10773] = 2.239886774493e-05;
C[10774] = 3.717020859066e-02;
C[10775] = 3.154508775129e-04;
C[10776] = -4.425707950077e-04;
C[10777] = 3.714329413189e-03;
C[10778] = 5.442737652994e-02;
C[10779] = 2.126924270004e-05;
C[10780] = 5.899204040190e-03;
C[10781] = 3.971880626479e-01;
C[10782] = 3.056756832356e-04;
C[10783] = -1.346199317599e-02;
C[10784] = 1.862292367213e-02;
C[10785] = -2.445838939735e-05;
C[10786] = 1.945618319752e-04;
C[10787] = 2.646642839800e-03;
C[10788] = -1.158879811125e-05;
C[10789] = 6.849360439917e-04;
C[10790] = 3.266146869513e-02;
C[10791] = -7.960444866989e-05;
C[10792] = -1.197511335725e-03;
C[10793] = 1.259353214285e-01;
C[10794] = 2.105740496362e-04;
C[10795] = 3.768893318646e-03;
C[10796] = 5.270887236842e-01;
C[10797] = 1.188122733134e-04;
C[10798] = 3.179283775464e-04;
C[10799] = 3.876661592788e-02;
C[10800] = -6.985709886255e+00;
C[10801] = -9.806246791364e-03;
C[10802] = 2.831863035765e-02;
C[10803] = 2.139770225817e+00;
C[10804] = -1.402665188636e-02;
C[10805] = -3.394536803239e-03;
C[10806] = 8.050320752803e-01;
C[10807] = -1.655270287463e-02;
C[10808] = -8.398285230736e-03;
C[10809] = 7.791388280539e-02;
C[10810] = 5.095504130801e-03;
C[10811] = 4.481371377644e-03;
C[10812] = 3.790233253185e-01;
C[10813] = -6.690200696900e-03;
C[10814] = -3.072815791380e-03;
C[10815] = 5.179477242855e-02;
C[10816] = -1.167907599703e-03;
C[10817] = -6.064826749883e-04;
C[10818] = 7.922012166457e-02;
C[10819] = -2.020299409851e-03;
C[10820] = -1.293678571256e-03;
C[10821] = 2.448154316767e-03;
C[10822] = -7.608126585998e-05;
C[10823] = -5.681451240569e-05;
C[10824] = -7.729151025700e-03;
C[10825] = -7.158303366930e+00;
C[10826] = 6.584994513548e-02;
C[10827] = -1.433672052100e-02;
C[10828] = 2.103512115669e+00;
C[10829] = 3.111431743938e-02;
C[10830] = -1.606722534854e-02;
C[10831] = 8.647670059482e-01;
C[10832] = 4.995787178859e-02;
C[10833] = 5.199358919084e-03;
C[10834] = 2.623732019486e-02;
C[10835] = 3.449051095804e-03;
C[10836] = -6.695344579361e-03;
C[10837] = 3.950281230127e-01;
C[10838] = 1.151804530217e-02;
C[10839] = -1.130211652348e-03;
C[10840] = 5.621770206462e-02;
C[10841] = 3.507432012220e-03;
C[10842] = -1.946225248960e-03;
C[10843] = 8.732154467379e-02;
C[10844] = 7.249750914824e-03;
C[10845] = -7.333803080778e-05;
C[10846] = 2.766976836292e-03;
C[10847] = 2.614392889190e-04;
C[10848] = 2.832881897848e-02;
C[10849] = 6.649226858064e-02;
C[10850] = -7.069003632097e+00;
C[10851] = -4.895872204335e-03;
C[10852] = 3.664325324112e-02;
C[10853] = 2.106556442306e+00;
C[10854] = -7.913696431512e-03;
C[10855] = 4.932760206839e-02;
C[10856] = 8.370181142392e-01;
C[10857] = 4.481880808685e-03;
C[10858] = 3.481167268062e-03;
C[10859] = 6.623743522245e-02;
C[10860] = -3.174931622967e-03;
C[10861] = 1.190825149192e-02;
C[10862] = 3.797491706541e-01;
C[10863] = -5.612741086379e-04;
C[10864] = 3.422224683206e-03;
C[10865] = 5.430668788201e-02;
C[10866] = -1.136936662162e-03;
C[10867] = 6.747490012220e-03;
C[10868] = 8.780040568028e-02;
C[10869] = -5.005708739476e-05;
C[10870] = 2.393872801896e-04;
C[10871] = 2.761160781481e-03;
C[10872] = 2.141163852392e+00;
C[10873] = -2.436732420805e-02;
C[10874] = 8.629025030522e-03;
C[10875] = -6.223367011746e+00;
C[10876] = 1.236206089631e-02;
C[10877] = 6.713904908543e-03;
C[10878] = 2.149858421596e+00;
C[10879] = 1.619297036120e-02;
C[10880] = 8.500359830626e-03;
C[10881] = 3.790930066472e-01;
C[10882] = -7.196874763714e-03;
C[10883] = -2.498686761224e-03;
C[10884] = 1.194221811538e-01;
C[10885] = 6.742344345672e-03;
C[10886] = 3.331139800050e-03;
C[10887] = 3.875458795172e-01;
C[10888] = 1.012899655936e-03;
C[10889] = 3.136454504500e-04;
C[10890] = 1.001128546535e+00;
C[10891] = 1.394120751545e-03;
C[10892] = -1.407642990321e-05;
C[10893] = 5.908904972910e-02;
C[10894] = 6.321298913170e-05;
C[10895] = 9.666346784364e-05;
C[10896] = -2.540523455949e-02;
C[10897] = 2.141661571395e+00;
C[10898] = 4.324957571754e-02;
C[10899] = 1.325959860176e-02;
C[10900] = -6.291998181274e+00;
C[10901] = -4.392549545980e-02;
C[10902] = 1.577345471380e-02;
C[10903] = 2.106854166670e+00;
C[10904] = -2.228792615728e-02;
C[10905] = -7.259129830557e-03;
C[10906] = 3.969355957990e-01;
C[10907] = 1.223856761574e-02;
C[10908] = 6.787221230945e-03;
C[10909] = 9.143025859675e-02;
C[10910] = -2.038706332238e-02;
C[10911] = 9.731750086722e-04;
C[10912] = 3.897867125212e-01;
C[10913] = 6.313655095825e-03;
C[10914] = 1.323978975894e-03;
C[10915] = 1.016134964313e+00;
C[10916] = 3.608527968205e-02;
C[10917] = 6.538147888987e-05;
C[10918] = 6.018254611527e-02;
C[10919] = 2.365526864011e-03;
C[10920] = 8.803560888900e-03;
C[10921] = 4.015148634757e-02;
C[10922] = 2.081635030282e+00;
C[10923] = 7.329687518521e-03;
C[10924] = -4.396417475453e-02;
C[10925] = -6.268795846284e+00;
C[10926] = 7.874430384472e-03;
C[10927] = -2.021710434817e-02;
C[10928] = 2.124682048249e+00;
C[10929] = -2.462910906773e-03;
C[10930] = 1.196990374758e-02;
C[10931] = 3.785031000529e-01;
C[10932] = 3.361928930548e-03;
C[10933] = -2.038899728712e-02;
C[10934] = 1.019406590804e-01;
C[10935] = 2.535365565100e-04;
C[10936] = 6.484202004812e-03;
C[10937] = 3.976867811380e-01;
C[10938] = -1.372793212453e-04;
C[10939] = 3.566868579071e-02;
C[10940] = 1.086264349868e+00;
C[10941] = 1.002807888282e-04;
C[10942] = 2.321295200961e-03;
C[10943] = 6.457891308245e-02;
C[10944] = 8.014121395189e-01;
C[10945] = -1.306619175844e-02;
C[10946] = -5.440104094940e-03;
C[10947] = 2.154319112802e+00;
C[10948] = 9.673421750170e-03;
C[10949] = -5.435944899394e-04;
C[10950] = -7.102040344823e+00;
C[10951] = 7.967041762968e-04;
C[10952] = -4.909534178834e-03;
C[10953] = 5.161377564047e-02;
C[10954] = -9.801599728432e-04;
C[10955] = -4.375944918093e-04;
C[10956] = 3.877689140775e-01;
C[10957] = 6.681733604905e-04;
C[10958] = -1.673646872106e-04;
C[10959] = 5.190938137434e-02;
C[10960] = -2.701072726698e-04;
C[10961] = -2.225066935222e-03;
C[10962] = 4.070912129234e+00;
C[10963] = 2.826595455292e-03;
C[10964] = 1.326691255577e-02;
C[10965] = 5.063755180457e-01;
C[10966] = 5.893285172573e-04;
C[10967] = 2.930456952852e-03;
C[10968] = -1.339012417603e-02;
C[10969] = 8.662400057840e-01;
C[10970] = 5.267441509430e-02;
C[10971] = 1.019630312877e-02;
C[10972] = 2.112251035066e+00;
C[10973] = -1.961275011978e-02;
C[10974] = 5.473876793499e-04;
C[10975] = -7.122320835138e+00;
C[10976] = -4.133116805746e-02;
C[10977] = -1.009778664773e-03;
C[10978] = 5.629135205641e-02;
C[10979] = 3.589565334502e-03;
C[10980] = 7.130662943144e-04;
C[10981] = 3.900565559410e-01;
C[10982] = 6.514419716232e-03;
C[10983] = -2.825730975171e-04;
C[10984] = 4.519193303150e-02;
C[10985] = -1.324364474233e-02;
C[10986] = 2.892397420009e-03;
C[10987] = 4.058410096207e+00;
C[10988] = 6.642517001402e-03;
C[10989] = 5.878450332559e-04;
C[10990] = 5.058118951721e-01;
C[10991] = 3.079568577891e-03;
C[10992] = -5.823841848169e-03;
C[10993] = 5.391405112299e-02;
C[10994] = 8.378805640304e-01;
C[10995] = 2.228119551200e-04;
C[10996] = -2.240810943215e-02;
C[10997] = 2.121243269163e+00;
C[10998] = -5.471510531279e-03;
C[10999] = -4.082917003616e-02;
C[11000] = -7.226962568366e+00;
C[11001] = -4.777605058600e-04;
C[11002] = 3.705240978940e-03;
C[11003] = 5.434981037157e-02;
C[11004] = -1.002319433252e-04;
C[11005] = 6.307645932081e-03;
C[11006] = 3.975148421838e-01;
C[11007] = -2.253165752845e-03;
C[11008] = -1.321854484127e-02;
C[11009] = 1.828600033945e-02;
C[11010] = 1.356003106412e-02;
C[11011] = 7.458647560371e-03;
C[11012] = 4.183865454997e+00;
C[11013] = 2.937042697955e-03;
C[11014] = 3.137857328284e-03;
C[11015] = 5.271775061832e-01;
C[11016] = 0.000000000000e+00;
C[11017] = 0.000000000000e+00;
C[11018] = 0.000000000000e+00;
C[11019] = 0.000000000000e+00;
C[11020] = 0.000000000000e+00;
C[11021] = 0.000000000000e+00;
C[11022] = 0.000000000000e+00;
C[11023] = 0.000000000000e+00;
C[11024] = 0.000000000000e+00;
C[11025] = 0.000000000000e+00;
C[11026] = 0.000000000000e+00;
C[11027] = 0.000000000000e+00;
C[11028] = 0.000000000000e+00;
C[11029] = 0.000000000000e+00;
C[11030] = 0.000000000000e+00;
C[11031] = 0.000000000000e+00;
C[11032] = 0.000000000000e+00;
C[11033] = 0.000000000000e+00;
C[11034] = 1.000000000000e+00;
C[11035] = 0.000000000000e+00;
C[11036] = 0.000000000000e+00;
C[11037] = 2.000000000000e-01;
C[11038] = -0.000000000000e+00;
C[11039] = -0.000000000000e+00;
C[11040] = 0.000000000000e+00;
C[11041] = 0.000000000000e+00;
C[11042] = 0.000000000000e+00;
C[11043] = 0.000000000000e+00;
C[11044] = 0.000000000000e+00;
C[11045] = 0.000000000000e+00;
C[11046] = 0.000000000000e+00;
C[11047] = 0.000000000000e+00;
C[11048] = 0.000000000000e+00;
C[11049] = 0.000000000000e+00;
C[11050] = 0.000000000000e+00;
C[11051] = 0.000000000000e+00;
C[11052] = 0.000000000000e+00;
C[11053] = 0.000000000000e+00;
C[11054] = 0.000000000000e+00;
C[11055] = 0.000000000000e+00;
C[11056] = 0.000000000000e+00;
C[11057] = 0.000000000000e+00;
C[11058] = 0.000000000000e+00;
C[11059] = 1.000000000000e+00;
C[11060] = 0.000000000000e+00;
C[11061] = -0.000000000000e+00;
C[11062] = 2.000000000000e-01;
C[11063] = -0.000000000000e+00;
C[11064] = 0.000000000000e+00;
C[11065] = 0.000000000000e+00;
C[11066] = 0.000000000000e+00;
C[11067] = 0.000000000000e+00;
C[11068] = 0.000000000000e+00;
C[11069] = 0.000000000000e+00;
C[11070] = 0.000000000000e+00;
C[11071] = 0.000000000000e+00;
C[11072] = 0.000000000000e+00;
C[11073] = 0.000000000000e+00;
C[11074] = 0.000000000000e+00;
C[11075] = 0.000000000000e+00;
C[11076] = 0.000000000000e+00;
C[11077] = 0.000000000000e+00;
C[11078] = 0.000000000000e+00;
C[11079] = 0.000000000000e+00;
C[11080] = 0.000000000000e+00;
C[11081] = 0.000000000000e+00;
C[11082] = 0.000000000000e+00;
C[11083] = 0.000000000000e+00;
C[11084] = 1.000000000000e+00;
C[11085] = -0.000000000000e+00;
C[11086] = -0.000000000000e+00;
C[11087] = 2.000000000000e-01;
C[11088] = 7.191726856667e-02;
C[11089] = 1.108265450597e-02;
C[11090] = 1.044572507574e-03;
C[11091] = 3.764264125771e-01;
C[11092] = -3.468885582367e-03;
C[11093] = -2.029225229960e-03;
C[11094] = 5.073170017364e-02;
C[11095] = -3.943663091298e-04;
C[11096] = -2.865806980465e-04;
C[11097] = 1.303588089905e-01;
C[11098] = 9.241971835089e-04;
C[11099] = 6.818057775928e-05;
C[11100] = 3.057287737511e-02;
C[11101] = -2.346179007910e-04;
C[11102] = -1.296211245689e-04;
C[11103] = 2.443766488203e-03;
C[11104] = -1.519389327622e-05;
C[11105] = -8.601608482253e-06;
C[11106] = 2.374235466359e-03;
C[11107] = -9.868458825169e-07;
C[11108] = 3.260614186371e-05;
C[11109] = -1.996329517213e-05;
C[11110] = 8.122139852337e-07;
C[11111] = 8.203917927915e-07;
C[11112] = 1.105519479192e-02;
C[11113] = 1.893475809650e-02;
C[11114] = -6.376786383532e-03;
C[11115] = -3.445610281879e-03;
C[11116] = 3.969957075163e-01;
C[11117] = 9.690896587302e-03;
C[11118] = -3.859164391859e-04;
C[11119] = 5.566884018602e-02;
C[11120] = 3.548447276079e-03;
C[11121] = 9.223665359054e-04;
C[11122] = 1.259799540470e-01;
C[11123] = -7.679096733497e-04;
C[11124] = -2.333429831627e-04;
C[11125] = 3.270926691868e-02;
C[11126] = 9.453630663320e-04;
C[11127] = -1.448149191264e-05;
C[11128] = 2.711739070900e-03;
C[11129] = 1.885673854416e-04;
C[11130] = 3.916328096646e-06;
C[11131] = 2.700372647877e-03;
C[11132] = 2.608058916363e-04;
C[11133] = 8.825407066372e-07;
C[11134] = -1.547040089369e-05;
C[11135] = 2.188558719998e-06;
C[11136] = 1.017506251888e-03;
C[11137] = -6.321151911505e-03;
C[11138] = 6.833415734965e-02;
C[11139] = -2.073150542139e-03;
C[11140] = 1.004927045378e-02;
C[11141] = 3.778648222049e-01;
C[11142] = -2.693215433230e-04;
C[11143] = 3.401504070040e-03;
C[11144] = 5.432851586277e-02;
C[11145] = 6.637616071354e-05;
C[11146] = -7.642007085479e-04;
C[11147] = 1.301007629728e-01;
C[11148] = -1.315112033943e-04;
C[11149] = 9.640124528781e-04;
C[11150] = 3.084189389761e-02;
C[11151] = -7.499242108011e-06;
C[11152] = 1.783921635165e-04;
C[11153] = 2.644868974660e-03;
C[11154] = 3.795899232620e-05;
C[11155] = 2.318870053133e-04;
C[11156] = 2.769628960351e-03;
C[11157] = 8.501973646543e-07;
C[11158] = 1.430841049682e-06;
C[11159] = -1.555975414292e-05;
C[11160] = 3.777912857387e-01;
C[11161] = -4.007322027410e-03;
C[11162] = -2.474055830625e-03;
C[11163] = 1.251946513469e-01;
C[11164] = 4.890792571112e-03;
C[11165] = 2.142710707864e-03;
C[11166] = 3.840012446766e-01;
C[11167] = 1.986211906257e-04;
C[11168] = -6.476502926971e-04;
C[11169] = 3.066386891922e-02;
C[11170] = -2.707904328647e-04;
C[11171] = -1.582382226267e-04;
C[11172] = 1.334555339886e-01;
C[11173] = 2.699316961091e-04;
C[11174] = 3.241717489639e-05;
C[11175] = 3.098364828337e-02;
C[11176] = 4.222013028302e-05;
C[11177] = 7.093173786412e-05;
C[11178] = 5.742801018124e-02;
C[11179] = 2.629136528994e-04;
C[11180] = 1.052738402161e-03;
C[11181] = 2.481298699553e-03;
C[11182] = 1.344105411367e-05;
C[11183] = 5.514342489510e-05;
C[11184] = -3.876722687713e-03;
C[11185] = 3.978103207574e-01;
C[11186] = 1.051968458889e-02;
C[11187] = 4.781757149609e-03;
C[11188] = 9.467379597071e-02;
C[11189] = -2.113117139839e-02;
C[11190] = 1.806687731083e-04;
C[11191] = 3.850249557585e-01;
C[11192] = 7.116066766427e-03;
C[11193] = -2.618070411473e-04;
C[11194] = 3.276357446809e-02;
C[11195] = 9.953733952193e-04;
C[11196] = 2.626626680089e-04;
C[11197] = 1.307232131041e-01;
C[11198] = -1.932732157917e-03;
C[11199] = 4.167437696357e-05;
C[11200] = 3.116259594512e-02;
C[11201] = 7.605390467494e-04;
C[11202] = 2.828498216772e-04;
C[11203] = 5.783805169328e-02;
C[11204] = 2.349155325345e-03;
C[11205] = 1.448910257723e-05;
C[11206] = 2.506171389630e-03;
C[11207] = 1.161265873078e-04;
C[11208] = -2.359183968894e-03;
C[11209] = 1.010767999277e-02;
C[11210] = 3.780799638090e-01;
C[11211] = 2.042417784343e-03;
C[11212] = -2.103290306525e-02;
C[11213] = 1.022716114492e-01;
C[11214] = -6.800546777163e-04;
C[11215] = 7.281619420803e-03;
C[11216] = 3.987150254168e-01;
C[11217] = -1.516183738311e-04;
C[11218] = 9.731486266965e-04;
C[11219] = 3.085623667121e-02;
C[11220] = 2.573097999501e-05;
C[11221] = -1.926180935708e-03;
C[11222] = 1.310008284984e-01;
C[11223] = 7.021094806036e-05;
C[11224] = 7.688815714755e-04;
C[11225] = 3.284001292517e-02;
C[11226] = 1.099121752739e-03;
C[11227] = 2.314053152532e-03;
C[11228] = 6.486240548929e-02;
C[11229] = 5.759681911603e-05;
C[11230] = 1.140719988706e-04;
C[11231] = 2.854421244269e-03;
C[11232] = 5.006324505120e-02;
C[11233] = -6.518791840068e-04;
C[11234] = -1.402538093922e-04;
C[11235] = 3.838387926925e-01;
C[11236] = 4.815948844081e-04;
C[11237] = -5.684196901164e-04;
C[11238] = 5.542740378190e-02;
C[11239] = -8.467888091398e-04;
C[11240] = -4.858964715749e-03;
C[11241] = 2.399202813374e-03;
C[11242] = -3.221234156737e-05;
C[11243] = 1.105273487378e-06;
C[11244] = 3.097281815110e-02;
C[11245] = 6.173611771689e-05;
C[11246] = 7.765328056702e-05;
C[11247] = 1.294081136611e-01;
C[11248] = -1.229617666939e-04;
C[11249] = -5.262653657470e-04;
C[11250] = 5.084437877822e-01;
C[11251] = 1.079405079924e-03;
C[11252] = 5.541398061036e-03;
C[11253] = 3.724430721707e-02;
C[11254] = 9.167970604768e-05;
C[11255] = 4.456164456473e-04;
C[11256] = -6.774885487864e-04;
C[11257] = 5.457531858860e-02;
C[11258] = 3.412514899855e-03;
C[11259] = 5.608127453398e-04;
C[11260] = 3.849564410816e-01;
C[11261] = 7.331499689170e-03;
C[11262] = -8.727501945921e-04;
C[11263] = 5.051703509614e-02;
C[11264] = -1.377838852013e-02;
C[11265] = -3.406870925333e-05;
C[11266] = 2.638837631071e-03;
C[11267] = 1.791262188375e-04;
C[11268] = 6.636623393063e-05;
C[11269] = 3.115802829999e-02;
C[11270] = 7.722069226999e-04;
C[11271] = -1.246925257241e-04;
C[11272] = 1.290496235145e-01;
C[11273] = -1.239631010157e-03;
C[11274] = 1.054951294287e-03;
C[11275] = 5.072318102450e-01;
C[11276] = 2.881664808139e-03;
C[11277] = 9.062977878197e-05;
C[11278] = 3.717286490762e-02;
C[11279] = 2.882809160342e-04;
C[11280] = -1.795300299710e-04;
C[11281] = 3.581063826763e-03;
C[11282] = 5.443146388020e-02;
C[11283] = -4.234476802588e-04;
C[11284] = 6.993734304639e-03;
C[11285] = 3.989394941657e-01;
C[11286] = -4.911833498846e-03;
C[11287] = -1.370851264079e-02;
C[11288] = 1.831561150785e-02;
C[11289] = -1.464897277223e-06;
C[11290] = 1.907418221538e-04;
C[11291] = 2.651732175821e-03;
C[11292] = 8.587857869334e-05;
C[11293] = 7.523835492969e-04;
C[11294] = 3.285497750844e-02;
C[11295] = -5.297899512868e-04;
C[11296] = -1.234972618202e-03;
C[11297] = 1.258573995645e-01;
C[11298] = 5.494953786848e-03;
C[11299] = 2.945090308062e-03;
C[11300] = 5.257559676416e-01;
C[11301] = 4.435420638811e-04;
C[11302] = 2.909111478956e-04;
C[11303] = 3.865220455972e-02;
C[11304] = -7.038300174753e+00;
C[11305] = 4.743636524452e-02;
C[11306] = -1.877660676304e-03;
C[11307] = 2.119543607241e+00;
C[11308] = 1.341587446298e-02;
C[11309] = -2.908242337349e-03;
C[11310] = 7.878894743636e-01;
C[11311] = -2.485046245904e-03;
C[11312] = -3.305758921928e-03;
C[11313] = 6.603889959251e-02;
C[11314] = 1.055456027583e-02;
C[11315] = 1.848395839625e-04;
C[11316] = 3.767889997952e-01;
C[11317] = -4.650265043684e-04;
C[11318] = -8.670663302189e-04;
C[11319] = 5.044677617385e-02;
C[11320] = -1.695553860071e-04;
C[11321] = -7.015184370200e-05;
C[11322] = 7.635073998975e-02;
C[11323] = -8.152921191957e-05;
C[11324] = 1.109896024664e-03;
C[11325] = 2.323760105650e-03;
C[11326] = -4.135872905113e-06;
C[11327] = 4.954823105623e-05;
C[11328] = 4.688717096346e-02;
C[11329] = -7.220454912580e+00;
C[11330] = -5.140550289368e-03;
C[11331] = 1.387705058433e-02;
C[11332] = 2.113056631680e+00;
C[11333] = 1.548242299989e-04;
C[11334] = -2.552307167710e-03;
C[11335] = 8.558217744734e-01;
C[11336] = 4.959270208114e-02;
C[11337] = 1.052710056177e-02;
C[11338] = 1.930066200535e-02;
C[11339] = -9.305565612077e-03;
C[11340] = -4.833499242631e-04;
C[11341] = 3.983316107164e-01;
C[11342] = 9.365614156493e-03;
C[11343] = -1.625214585213e-04;
C[11344] = 5.525621899773e-02;
C[11345] = 3.539952755056e-03;
C[11346] = 4.333296381565e-05;
C[11347] = 8.436199726135e-02;
C[11348] = 7.140812599873e-03;
C[11349] = 1.388587210796e-06;
C[11350] = 2.620814303347e-03;
C[11351] = 2.610973721363e-04;
C[11352] = -2.418985790025e-03;
C[11353] = -4.027860848820e-03;
C[11354] = -7.042377199796e+00;
C[11355] = -3.328148903929e-03;
C[11356] = 4.976026054899e-03;
C[11357] = 2.093013647409e+00;
C[11358] = -3.291382799663e-03;
C[11359] = 4.796444956743e-02;
C[11360] = 8.380472442804e-01;
C[11361] = 1.577733282764e-04;
C[11362] = -9.249931140050e-03;
C[11363] = 6.418928878710e-02;
C[11364] = -9.125852155043e-04;
C[11365] = 9.719518385818e-03;
C[11366] = 3.785790127758e-01;
C[11367] = -5.706624083779e-05;
C[11368] = 3.367639448057e-03;
C[11369] = 5.447299592159e-02;
C[11370] = 1.257303213266e-03;
C[11371] = 6.459023913002e-03;
C[11372] = 8.809354652421e-02;
C[11373] = 5.597361761482e-05;
C[11374] = 2.308093654073e-04;
C[11375] = 2.783730186174e-03;
C[11376] = 2.146841070474e+00;
C[11377] = 2.642815673720e-03;
C[11378] = -1.134625467365e-02;
C[11379] = -6.201847667982e+00;
C[11380] = 9.644255454050e-03;
C[11381] = 2.970235580246e-04;
C[11382] = 2.139747801522e+00;
C[11383] = 2.354811014965e-03;
C[11384] = -7.532023349841e-03;
C[11385] = 3.781538729568e-01;
C[11386] = -1.045061669794e-03;
C[11387] = -1.313490503990e-03;
C[11388] = 1.304595172265e-01;
C[11389] = 1.114446574172e-03;
C[11390] = -1.507996613908e-03;
C[11391] = 3.782674237642e-01;
C[11392] = 9.234993342704e-04;
C[11393] = 1.638884364136e-03;
C[11394] = 9.697674675362e-01;
C[11395] = 5.305699693270e-03;
C[11396] = 1.902991382116e-02;
C[11397] = 5.717814434367e-02;
C[11398] = 3.350370307858e-04;
C[11399] = 1.214589723929e-03;
C[11400] = 5.259132356072e-03;
C[11401] = 2.129348896503e+00;
C[11402] = 1.438430875725e-02;
C[11403] = 7.463547023976e-03;
C[11404] = -6.277776685317e+00;
C[11405] = -4.972253025371e-02;
C[11406] = 2.336943101145e-03;
C[11407] = 2.089309691579e+00;
C[11408] = -1.722176303497e-02;
C[11409] = -8.728636097140e-04;
C[11410] = 3.991462239576e-01;
C[11411] = 1.018993252094e-02;
C[11412] = 1.005411152669e-03;
C[11413] = 1.012720204240e-01;
C[11414] = -2.080555099916e-02;
C[11415] = 9.089236766727e-04;
C[11416] = 3.774914376902e-01;
C[11417] = 6.905819145634e-03;
C[11418] = 5.521890800573e-03;
C[11419] = 9.717294837041e-01;
C[11420] = 3.522531181372e-02;
C[11421] = 3.511114871539e-04;
C[11422] = 5.748984754131e-02;
C[11423] = 2.344199842355e-03;
C[11424] = -9.507417116029e-03;
C[11425] = 8.490492339352e-03;
C[11426] = 2.097316479490e+00;
C[11427] = -1.708834912388e-03;
C[11428] = -4.775716359097e-02;
C[11429] = -6.268576860084e+00;
C[11430] = -7.588471768049e-03;
C[11431] = -1.550421720926e-02;
C[11432] = 2.129126448924e+00;
C[11433] = -1.197025069153e-03;
C[11434] = 9.782397561961e-03;
C[11435] = 3.787941543799e-01;
C[11436] = -1.608289537429e-03;
C[11437] = -2.070728266602e-02;
C[11438] = 9.983983451273e-02;
C[11439] = 1.615587183587e-03;
C[11440] = 7.079024284211e-03;
C[11441] = 4.001018192646e-01;
C[11442] = 1.963126256729e-02;
C[11443] = 3.510047494371e-02;
C[11444] = 1.091792326030e+00;
C[11445] = 1.256798496470e-03;
C[11446] = 2.328308899544e-03;
C[11447] = 6.492194059996e-02;
C[11448] = 7.745203719149e-01;
C[11449] = -7.871562064128e-03;
C[11450] = -7.100281210463e-04;
C[11451] = 2.136498761841e+00;
C[11452] = 8.355465327141e-03;
C[11453] = -5.355772016052e-03;
C[11454] = -7.086443941791e+00;
C[11455] = -5.301208790851e-03;
C[11456] = -2.898468253766e-02;
C[11457] = 4.977832105141e-02;
C[11458] = -4.284842033422e-04;
C[11459] = 7.200149309306e-05;
C[11460] = 3.781049717802e-01;
C[11461] = 1.209849787973e-03;
C[11462] = 1.727222171187e-03;
C[11463] = 6.030527332043e-02;
C[11464] = -1.974514239279e-03;
C[11465] = -7.616277696066e-03;
C[11466] = 4.101880963583e+00;
C[11467] = 6.487603469221e-03;
C[11468] = 3.453189913782e-02;
C[11469] = 5.095684517540e-01;
C[11470] = 1.227484590511e-03;
C[11471] = 5.786473548868e-03;
C[11472] = -8.147491039036e-03;
C[11473] = 8.339513425249e-01;
C[11474] = 4.818466616374e-02;
C[11475] = 9.598642109248e-03;
C[11476] = 2.087939398040e+00;
C[11477] = -1.450661184193e-02;
C[11478] = -5.820436499897e-03;
C[11479] = -7.097335003275e+00;
C[11480] = -4.225366422032e-02;
C[11481] = -4.526776256637e-04;
C[11482] = 5.416269740030e-02;
C[11483] = 3.378650277872e-03;
C[11484] = 1.285690888985e-03;
C[11485] = 3.774229230133e-01;
C[11486] = 7.128904552578e-03;
C[11487] = -2.000475624732e-03;
C[11488] = 5.731976830854e-02;
C[11489] = -1.380785067126e-02;
C[11490] = 6.140514298521e-03;
C[11491] = 4.088056899089e+00;
C[11492] = 4.358800428297e-03;
C[11493] = 1.206493510261e-03;
C[11494] = 5.083374346535e-01;
C[11495] = 3.151330737105e-03;
C[11496] = -1.164745560419e-03;
C[11497] = 5.024503309482e-02;
C[11498] = 8.401062046289e-01;
C[11499] = -3.047971101076e-03;
C[11500] = -1.966841227073e-02;
C[11501] = 2.133615823903e+00;
C[11502] = -3.004205819959e-02;
C[11503] = -4.085614663365e-02;
C[11504] = -7.223334112716e+00;
C[11505] = 3.689882437346e-05;
C[11506] = 3.572569305740e-03;
C[11507] = 5.457594393902e-02;
C[11508] = 1.863086976574e-03;
C[11509] = 6.783486683846e-03;
C[11510] = 4.003262880136e-01;
C[11511] = -7.669146479163e-03;
C[11512] = -1.373797479193e-02;
C[11513] = 1.727677226719e-02;
C[11514] = 3.387377237241e-02;
C[11515] = 5.224218012663e-03;
C[11516] = 4.168128062822e+00;
C[11517] = 5.743520109443e-03;
C[11518] = 3.197245250922e-03;
C[11519] = 5.252626089486e-01;
C[11520] = 0.000000000000e+00;
C[11521] = 0.000000000000e+00;
C[11522] = 0.000000000000e+00;
C[11523] = 0.000000000000e+00;
C[11524] = 0.000000000000e+00;
C[11525] = 0.000000000000e+00;
C[11526] = 0.000000000000e+00;
C[11527] = 0.000000000000e+00;
C[11528] = 0.000000000000e+00;
C[11529] = 0.000000000000e+00;
C[11530] = 0.000000000000e+00;
C[11531] = 0.000000000000e+00;
C[11532] = 0.000000000000e+00;
C[11533] = 0.000000000000e+00;
C[11534] = 0.000000000000e+00;
C[11535] = 0.000000000000e+00;
C[11536] = 0.000000000000e+00;
C[11537] = 0.000000000000e+00;
C[11538] = 1.000000000000e+00;
C[11539] = 0.000000000000e+00;
C[11540] = 0.000000000000e+00;
C[11541] = 2.000000000000e-01;
C[11542] = -0.000000000000e+00;
C[11543] = -0.000000000000e+00;
C[11544] = 0.000000000000e+00;
C[11545] = 0.000000000000e+00;
C[11546] = 0.000000000000e+00;
C[11547] = 0.000000000000e+00;
C[11548] = 0.000000000000e+00;
C[11549] = 0.000000000000e+00;
C[11550] = 0.000000000000e+00;
C[11551] = 0.000000000000e+00;
C[11552] = 0.000000000000e+00;
C[11553] = 0.000000000000e+00;
C[11554] = 0.000000000000e+00;
C[11555] = 0.000000000000e+00;
C[11556] = 0.000000000000e+00;
C[11557] = 0.000000000000e+00;
C[11558] = 0.000000000000e+00;
C[11559] = 0.000000000000e+00;
C[11560] = 0.000000000000e+00;
C[11561] = 0.000000000000e+00;
C[11562] = 0.000000000000e+00;
C[11563] = 1.000000000000e+00;
C[11564] = 0.000000000000e+00;
C[11565] = -0.000000000000e+00;
C[11566] = 2.000000000000e-01;
C[11567] = -0.000000000000e+00;
C[11568] = 0.000000000000e+00;
C[11569] = 0.000000000000e+00;
C[11570] = 0.000000000000e+00;
C[11571] = 0.000000000000e+00;
C[11572] = 0.000000000000e+00;
C[11573] = 0.000000000000e+00;
C[11574] = 0.000000000000e+00;
C[11575] = 0.000000000000e+00;
C[11576] = 0.000000000000e+00;
C[11577] = 0.000000000000e+00;
C[11578] = 0.000000000000e+00;
C[11579] = 0.000000000000e+00;
C[11580] = 0.000000000000e+00;
C[11581] = 0.000000000000e+00;
C[11582] = 0.000000000000e+00;
C[11583] = 0.000000000000e+00;
C[11584] = 0.000000000000e+00;
C[11585] = 0.000000000000e+00;
C[11586] = 0.000000000000e+00;
C[11587] = 0.000000000000e+00;
C[11588] = 1.000000000000e+00;
C[11589] = -0.000000000000e+00;
C[11590] = -0.000000000000e+00;
C[11591] = 2.000000000000e-01;
C[11592] = 6.203881647585e-02;
C[11593] = 1.141415191787e-02;
C[11594] = -4.829781879940e-04;
C[11595] = 3.787122012240e-01;
C[11596] = 4.806890046797e-06;
C[11597] = -1.950792209805e-03;
C[11598] = 4.985765622006e-02;
C[11599] = 8.239249890057e-05;
C[11600] = 9.638426177788e-04;
C[11601] = 1.298437041079e-01;
C[11602] = 1.054078293401e-03;
C[11603] = -4.018649392375e-05;
C[11604] = 3.074974519353e-02;
C[11605] = -1.288233265205e-04;
C[11606] = -1.647334013952e-04;
C[11607] = 2.405691178366e-03;
C[11608] = -2.268513625167e-06;
C[11609] = 4.982797619792e-05;
C[11610] = 2.357388403794e-03;
C[11611] = -7.435252226826e-06;
C[11612] = 9.035236001083e-05;
C[11613] = -1.835087974349e-05;
C[11614] = -1.242671886210e-06;
C[11615] = 7.285259871295e-07;
C[11616] = 1.146497852719e-02;
C[11617] = 2.122859899845e-02;
C[11618] = -9.137832743107e-03;
C[11619] = -2.158009773170e-04;
C[11620] = 3.998592136963e-01;
C[11621] = 1.001797668471e-02;
C[11622] = 1.471961827792e-04;
C[11623] = 5.360707252476e-02;
C[11624] = 3.436148298200e-03;
C[11625] = 1.057466734022e-03;
C[11626] = 1.262265478776e-01;
C[11627] = -7.826770078043e-04;
C[11628] = -1.411470872675e-04;
C[11629] = 3.279843242467e-02;
C[11630] = 9.971437949961e-04;
C[11631] = 2.087286535002e-06;
C[11632] = 2.608560889512e-03;
C[11633] = 1.838507574058e-04;
C[11634] = 2.461686330764e-06;
C[11635] = 2.604606848933e-03;
C[11636] = 2.609104510062e-04;
C[11637] = -9.266534913760e-07;
C[11638] = -1.531795421590e-05;
C[11639] = 2.494230483387e-06;
C[11640] = -4.594906453391e-04;
C[11641] = -9.163008305645e-03;
C[11642] = 6.127147727474e-02;
C[11643] = -2.082556425812e-03;
C[11644] = 1.037839261317e-02;
C[11645] = 3.803090834945e-01;
C[11646] = 1.014557736042e-03;
C[11647] = 3.242728858728e-03;
C[11648] = 5.477419724505e-02;
C[11649] = -3.862065774676e-05;
C[11650] = -7.843553786402e-04;
C[11651] = 1.297154969351e-01;
C[11652] = -1.724803499348e-04;
C[11653] = 1.016027579884e-03;
C[11654] = 3.105123309815e-02;
C[11655] = 5.336571296232e-05;
C[11656] = 1.726833035611e-04;
C[11657] = 2.665162841962e-03;
C[11658] = 9.839965244109e-05;
C[11659] = 2.367666804584e-04;
C[11660] = 2.783912561194e-03;
C[11661] = 1.045801561244e-06;
C[11662] = 2.076426477305e-06;
C[11663] = -1.650165280027e-05;
C[11664] = 3.778062786792e-01;
C[11665] = -3.553931215031e-04;
C[11666] = -1.918698597886e-03;
C[11667] = 1.297474436034e-01;
C[11668] = -1.171138590230e-03;
C[11669] = -7.206585169482e-03;
C[11670] = 3.772786530920e-01;
C[11671] = 2.653868405913e-03;
C[11672] = 7.108907326276e-03;
C[11673] = 3.068935035721e-02;
C[11674] = -1.504532302133e-04;
C[11675] = -1.615564947397e-04;
C[11676] = 1.333702021985e-01;
C[11677] = -9.122587064962e-05;
C[11678] = -6.314176330407e-04;
C[11679] = 3.091463009426e-02;
C[11680] = 2.942168851258e-04;
C[11681] = 6.961944669376e-04;
C[11682] = 5.811784760058e-02;
C[11683] = 8.361205944885e-04;
C[11684] = 2.218923223483e-03;
C[11685] = 2.546063642076e-03;
C[11686] = 4.220480662724e-05;
C[11687] = 1.055704599870e-04;
C[11688] = -2.159325009570e-04;
C[11689] = 3.989262970708e-01;
C[11690] = 9.961739691560e-03;
C[11691] = -1.140512088491e-03;
C[11692] = 1.048711796682e-01;
C[11693] = -1.989374215398e-02;
C[11694] = 2.572856623864e-03;
C[11695] = 3.741021078940e-01;
C[11696] = 5.835936164993e-03;
C[11697] = -1.435392859207e-04;
C[11698] = 3.273623798298e-02;
C[11699] = 9.882507184437e-04;
C[11700] = -8.918410386701e-05;
C[11701] = 1.311095149502e-01;
C[11702] = -1.866722800737e-03;
C[11703] = 2.895013904027e-04;
C[11704] = 3.079175442481e-02;
C[11705] = 6.729048347292e-04;
C[11706] = 8.295652054701e-04;
C[11707] = 5.806720858112e-02;
C[11708] = 2.286196492627e-03;
C[11709] = 4.207773700453e-05;
C[11710] = 2.554037913995e-03;
C[11711] = 1.150745331975e-04;
C[11712] = -1.834351846408e-03;
C[11713] = 9.611017509241e-03;
C[11714] = 3.804164843034e-01;
C[11715] = -7.236844646160e-03;
C[11716] = -1.997596939702e-02;
C[11717] = 9.931965468070e-02;
C[11718] = 7.135635434661e-03;
C[11719] = 6.158539859335e-03;
C[11720] = 3.987160409542e-01;
C[11721] = -1.569707105020e-04;
C[11722] = 9.700131832982e-04;
C[11723] = 3.105839315208e-02;
C[11724] = -6.334349314859e-04;
C[11725] = -1.872204616940e-03;
C[11726] = 1.309416652058e-01;
C[11727] = 6.982297300343e-04;
C[11728] = 6.913958239055e-04;
C[11729] = 3.268257256723e-02;
C[11730] = 2.190962529993e-03;
C[11731] = 2.277498290916e-03;
C[11732] = 6.432243907391e-02;
C[11733] = 1.038783246389e-04;
C[11734] = 1.133168888225e-04;
C[11735] = 2.819068914674e-03;
C[11736] = 5.054689402390e-02;
C[11737] = 4.901542592811e-04;
C[11738] = 1.178174954100e-03;
C[11739] = 3.774095176514e-01;
C[11740] = 2.429851288428e-03;
C[11741] = 7.119808616375e-03;
C[11742] = 6.053319331532e-02;
C[11743] = -3.444494312305e-03;
C[11744] = -1.181365420402e-02;
C[11745] = 2.451640365289e-03;
C[11746] = 2.495115830180e-05;
C[11747] = 6.427352749957e-05;
C[11748] = 3.092335439823e-02;
C[11749] = 2.799677013736e-04;
C[11750] = 6.971746088152e-04;
C[11751] = 1.293972778336e-01;
C[11752] = -3.531591453266e-04;
C[11753] = -1.027372530585e-03;
C[11754] = 5.090883723989e-01;
C[11755] = 5.578281034004e-04;
C[11756] = 3.438207893246e-03;
C[11757] = 3.724274388808e-02;
C[11758] = 4.734263596384e-05;
C[11759] = 2.655298305224e-04;
C[11760] = 4.147447364936e-04;
C[11761] = 5.446638274375e-02;
C[11762] = 3.228888425485e-03;
C[11763] = 2.516747542586e-03;
C[11764] = 3.740506676375e-01;
C[11765] = 6.118615201476e-03;
C[11766] = -3.436493479926e-03;
C[11767] = 5.961847618469e-02;
C[11768] = -1.309830250645e-02;
C[11769] = 1.988830221437e-05;
C[11770] = 2.665848237445e-03;
C[11771] = 1.717606080116e-04;
C[11772] = 2.850754942374e-04;
C[11773] = 3.078832507437e-02;
C[11774] = 6.887341800482e-04;
C[11775] = -3.526257565013e-04;
C[11776] = 1.292960485435e-01;
C[11777] = -1.186119718281e-03;
C[11778] = 5.542015909153e-04;
C[11779] = 5.090599176619e-01;
C[11780] = 3.550540574712e-03;
C[11781] = 4.717677275746e-05;
C[11782] = 3.726164779832e-02;
C[11783] = 3.218144586318e-04;
C[11784] = 1.139600338868e-03;
C[11785] = 3.400925147287e-03;
C[11786] = 5.445075873675e-02;
C[11787] = 7.206697170979e-03;
C[11788] = 5.972212495795e-03;
C[11789] = 3.986097902026e-01;
C[11790] = -1.186704456574e-02;
C[11791] = -1.317849109323e-02;
C[11792] = 1.806832804277e-02;
C[11793] = 6.154515760385e-05;
C[11794] = 1.815025473449e-04;
C[11795] = 2.643600274743e-03;
C[11796] = 7.027137899178e-04;
C[11797] = 6.819899234494e-04;
C[11798] = 3.267548918379e-02;
C[11799] = -1.030931888033e-03;
C[11800] = -1.191465624066e-03;
C[11801] = 1.259720136275e-01;
C[11802] = 3.452416807540e-03;
C[11803] = 3.581925849978e-03;
C[11804] = 5.262858749391e-01;
C[11805] = 2.665418426883e-04;
C[11806] = 3.237670642799e-04;
C[11807] = 3.872445667858e-02;
C[11808] = -7.067791280413e+00;
C[11809] = 5.991873053321e-02;
C[11810] = -9.460693623366e-03;
C[11811] = 2.137126000767e+00;
C[11812] = 1.491257263414e-02;
C[11813] = -2.328420606245e-02;
C[11814] = 7.753041482534e-01;
C[11815] = 9.174855650663e-04;
C[11816] = 1.242502063723e-02;
C[11817] = 6.364918032033e-02;
C[11818] = 1.367893818880e-02;
C[11819] = -9.436354178205e-04;
C[11820] = 3.784557951829e-01;
C[11821] = -1.660681708515e-03;
C[11822] = -2.857342767796e-03;
C[11823] = 5.017285649878e-02;
C[11824] = 4.418701213151e-05;
C[11825] = 9.196375223577e-04;
C[11826] = 7.710043036893e-02;
C[11827] = 2.916573017777e-04;
C[11828] = 2.621222880180e-03;
C[11829] = 2.410552755490e-03;
C[11830] = 8.465519618001e-06;
C[11831] = 9.897070051746e-05;
C[11832] = 6.093526271966e-02;
C[11833] = -7.204654464527e+00;
C[11834] = -4.345003220010e-03;
C[11835] = 1.152409618485e-02;
C[11836] = 2.145775911088e+00;
C[11837] = 1.074656012652e-02;
C[11838] = 1.684391548517e-03;
C[11839] = 8.263550572762e-01;
C[11840] = 4.848885898677e-02;
C[11841] = 1.372976479812e-02;
C[11842] = 2.299206091854e-02;
C[11843] = -6.813766148493e-03;
C[11844] = -1.885398403773e-03;
C[11845] = 3.988501899945e-01;
C[11846] = 1.050568349849e-02;
C[11847] = 1.211616766559e-04;
C[11848] = 5.394091915689e-02;
C[11849] = 3.461716398729e-03;
C[11850] = 5.249884875636e-04;
C[11851] = 8.368884161447e-02;
C[11852] = 7.119620959730e-03;
C[11853] = 1.939304253349e-05;
C[11854] = 2.667531699007e-03;
C[11855] = 2.613654930649e-04;
C[11856] = -8.990942770268e-03;
C[11857] = -4.848514470779e-03;
C[11858] = -7.059014801129e+00;
C[11859] = -2.540913430126e-02;
C[11860] = 1.577011432868e-02;
C[11861] = 2.098394127250e+00;
C[11862] = 1.317961868624e-02;
C[11863] = 4.636514941513e-02;
C[11864] = 8.434572403005e-01;
C[11865] = -9.201478751656e-04;
C[11866] = -6.838941711032e-03;
C[11867] = 6.243747468589e-02;
C[11868] = -3.006373583659e-03;
C[11869] = 1.085461563256e-02;
C[11870] = 3.806443205548e-01;
C[11871] = 9.859368416403e-04;
C[11872] = 3.297333695755e-03;
C[11873] = 5.465172944480e-02;
C[11874] = 2.778603920049e-03;
C[11875] = 6.560173435453e-03;
C[11876] = 8.804363782077e-02;
C[11877] = 1.057807222955e-04;
C[11878] = 2.386987633473e-04;
C[11879] = 2.759092089822e-03;
C[11880] = 2.119007549871e+00;
C[11881] = 8.732253301126e-03;
C[11882] = -2.213197774272e-02;
C[11883] = -6.204797361648e+00;
C[11884] = -4.339394054967e-05;
C[11885] = -1.768123202438e-02;
C[11886] = 2.130046414944e+00;
C[11887] = 1.057711458180e-02;
C[11888] = 9.000855174527e-03;
C[11889] = 3.775498726381e-01;
C[11890] = -2.024990547959e-03;
C[11891] = -2.842515755732e-03;
C[11892] = 1.240034429226e-01;
C[11893] = -6.460629182191e-04;
C[11894] = -6.610602785184e-03;
C[11895] = 3.830606646427e-01;
C[11896] = 3.889903314282e-03;
C[11897] = 7.447635639013e-03;
C[11898] = 9.888791511940e-01;
C[11899] = 1.419099238811e-02;
C[11900] = 3.506134613466e-02;
C[11901] = 5.842066581079e-02;
C[11902] = 8.742474088160e-04;
C[11903] = 2.165868711006e-03;
C[11904] = 1.049778481407e-02;
C[11905] = 2.127117578579e+00;
C[11906] = 7.437055896465e-03;
C[11907] = 5.691360942351e-04;
C[11908] = -6.267383215094e+00;
C[11909] = -4.688758368013e-02;
C[11910] = 9.458065478164e-03;
C[11911] = 2.077936974860e+00;
C[11912] = -2.189818660088e-02;
C[11913] = -1.881421099519e-03;
C[11914] = 3.979172733691e-01;
C[11915] = 1.043796271095e-02;
C[11916] = -6.154364164799e-04;
C[11917] = 9.949357166191e-02;
C[11918] = -2.059396185378e-02;
C[11919] = 3.800040437588e-03;
C[11920] = 3.802449655073e-01;
C[11921] = 6.207706866467e-03;
C[11922] = 1.398305458625e-02;
C[11923] = 9.856896802098e-01;
C[11924] = 3.485870193473e-02;
C[11925] = 8.640822054946e-04;
C[11926] = 5.841266893909e-02;
C[11927] = 2.314006352554e-03;
C[11928] = -2.095539879450e-02;
C[11929] = 2.607376617133e-03;
C[11930] = 2.100542143428e+00;
C[11931] = -1.828642155795e-02;
C[11932] = -4.853212854093e-02;
C[11933] = -6.270599801818e+00;
C[11934] = 9.813825608495e-03;
C[11935] = -1.708346218096e-02;
C[11936] = 2.122748499228e+00;
C[11937] = -2.740902404399e-03;
C[11938] = 1.009872432302e-02;
C[11939] = 3.807517213637e-01;
C[11940] = -6.640862261862e-03;
C[11941] = -2.067618909682e-02;
C[11942] = 1.014736626464e-01;
C[11943] = 7.477938600495e-03;
C[11944] = 6.517170920758e-03;
C[11945] = 3.964165600884e-01;
C[11946] = 3.441635825953e-02;
C[11947] = 3.479893150952e-02;
C[11948] = 1.080352078158e+00;
C[11949] = 2.117957420649e-03;
C[11950] = 2.287750691618e-03;
C[11951] = 6.422664422916e-02;
C[11952] = 7.890889043303e-01;
C[11953] = 8.543553078555e-03;
C[11954] = 1.645196304741e-02;
C[11955] = 2.132663706134e+00;
C[11956] = 6.597958769447e-03;
C[11957] = 9.497289242767e-03;
C[11958] = -7.088335893444e+00;
C[11959] = -9.816080894865e-03;
C[11960] = -3.285486945908e-02;
C[11961] = 5.086209430263e-02;
C[11962] = 4.641197531578e-04;
C[11963] = 1.149554059699e-03;
C[11964] = 3.831915292022e-01;
C[11965] = 3.657035102153e-03;
C[11966] = 7.462111782209e-03;
C[11967] = 5.485663712777e-02;
C[11968] = -4.270427043092e-03;
C[11969] = -1.094191782552e-02;
C[11970] = 4.089136973505e+00;
C[11971] = -3.881800696392e-03;
C[11972] = 4.936595568639e-03;
C[11973] = 5.086858628798e-01;
C[11974] = 2.019216571513e-04;
C[11975] = 2.263394294437e-03;
C[11976] = 7.564530316927e-03;
C[11977] = 8.435412616559e-01;
C[11978] = 4.608834075026e-02;
C[11979] = 7.834697315256e-03;
C[11980] = 2.076908169729e+00;
C[11981] = -1.788195533815e-02;
C[11982] = -9.656064247284e-03;
C[11983] = -7.097106318991e+00;
C[11984] = -4.148748617706e-02;
C[11985] = 3.765392497245e-04;
C[11986] = 5.480022937588e-02;
C[11987] = 3.283493262512e-03;
C[11988] = 3.752782450954e-03;
C[11989] = 3.801935252508e-01;
C[11990] = 6.477246262898e-03;
C[11991] = -4.262426210712e-03;
C[11992] = 5.316207342008e-02;
C[11993] = -1.341665379991e-02;
C[11994] = -3.939134872242e-03;
C[11995] = 4.088759396775e+00;
C[11996] = 8.073540126342e-03;
C[11997] = 2.005556007000e-04;
C[11998] = 5.090675282804e-01;
C[11999] = 3.450817614190e-03;
C[12000] = 1.594017505901e-02;
C[12001] = 4.778439596851e-02;
C[12002] = 8.369884701346e-01;
C[12003] = 1.095665206860e-02;
C[12004] = -1.917265998484e-02;
C[12005] = 2.120623484197e+00;
C[12006] = -3.392267669333e-02;
C[12007] = -4.309125791261e-02;
C[12008] = -7.222971601264e+00;
C[12009] = 1.095395243447e-03;
C[12010] = 3.426493247816e-03;
C[12011] = 5.432829093651e-02;
C[12012] = 7.545425483716e-03;
C[12013] = 6.343983197270e-03;
C[12014] = 3.963103093368e-01;
C[12015] = -1.099530818723e-02;
C[12016] = -1.349684238668e-02;
C[12017] = 1.984050442313e-02;
C[12018] = 5.293347820767e-03;
C[12019] = 8.717511389441e-03;
C[12020] = 4.183029607890e+00;
C[12021] = 2.298407921860e-03;
C[12022] = 3.498234462045e-03;
C[12023] = 5.269248759567e-01;
C[12024] = 0.000000000000e+00;
C[12025] = 0.000000000000e+00;
C[12026] = 0.000000000000e+00;
C[12027] = 0.000000000000e+00;
C[12028] = 0.000000000000e+00;
C[12029] = 0.000000000000e+00;
C[12030] = 0.000000000000e+00;
C[12031] = 0.000000000000e+00;
C[12032] = 0.000000000000e+00;
C[12033] = 0.000000000000e+00;
C[12034] = 0.000000000000e+00;
C[12035] = 0.000000000000e+00;
C[12036] = 0.000000000000e+00;
C[12037] = 0.000000000000e+00;
C[12038] = 0.000000000000e+00;
C[12039] = 0.000000000000e+00;
C[12040] = 0.000000000000e+00;
C[12041] = 0.000000000000e+00;
C[12042] = 1.000000000000e+00;
C[12043] = 0.000000000000e+00;
C[12044] = 0.000000000000e+00;
C[12045] = 2.000000000000e-01;
C[12046] = -0.000000000000e+00;
C[12047] = -0.000000000000e+00;
C[12048] = 0.000000000000e+00;
C[12049] = 0.000000000000e+00;
C[12050] = 0.000000000000e+00;
C[12051] = 0.000000000000e+00;
C[12052] = 0.000000000000e+00;
C[12053] = 0.000000000000e+00;
C[12054] = 0.000000000000e+00;
C[12055] = 0.000000000000e+00;
C[12056] = 0.000000000000e+00;
C[12057] = 0.000000000000e+00;
C[12058] = 0.000000000000e+00;
C[12059] = 0.000000000000e+00;
C[12060] = 0.000000000000e+00;
C[12061] = 0.000000000000e+00;
C[12062] = 0.000000000000e+00;
C[12063] = 0.000000000000e+00;
C[12064] = 0.000000000000e+00;
C[12065] = 0.000000000000e+00;
C[12066] = 0.000000000000e+00;
C[12067] = 1.000000000000e+00;
C[12068] = 0.000000000000e+00;
C[12069] = -0.000000000000e+00;
C[12070] = 2.000000000000e-01;
C[12071] = -0.000000000000e+00;
C[12072] = 0.000000000000e+00;
C[12073] = 0.000000000000e+00;
C[12074] = 0.000000000000e+00;
C[12075] = 0.000000000000e+00;
C[12076] = 0.000000000000e+00;
C[12077] = 0.000000000000e+00;
C[12078] = 0.000000000000e+00;
C[12079] = 0.000000000000e+00;
C[12080] = 0.000000000000e+00;
C[12081] = 0.000000000000e+00;
C[12082] = 0.000000000000e+00;
C[12083] = 0.000000000000e+00;
C[12084] = 0.000000000000e+00;
C[12085] = 0.000000000000e+00;
C[12086] = 0.000000000000e+00;
C[12087] = 0.000000000000e+00;
C[12088] = 0.000000000000e+00;
C[12089] = 0.000000000000e+00;
C[12090] = 0.000000000000e+00;
C[12091] = 0.000000000000e+00;
C[12092] = 1.000000000000e+00;
C[12093] = -0.000000000000e+00;
C[12094] = -0.000000000000e+00;
C[12095] = 2.000000000000e-01;
C[12096] = 7.219195967715e-02;
C[12097] = 1.423120687994e-02;
C[12098] = -2.191446884109e-03;
C[12099] = 3.775009606865e-01;
C[12100] = -4.836976485659e-03;
C[12101] = -3.601562147169e-03;
C[12102] = 5.113050573089e-02;
C[12103] = -4.424277532419e-04;
C[12104] = 2.806095774599e-04;
C[12105] = 1.309983835094e-01;
C[12106] = 1.017521698907e-03;
C[12107] = -1.090759543759e-04;
C[12108] = 3.052632136453e-02;
C[12109] = -4.882417846434e-04;
C[12110] = -2.715231932439e-04;
C[12111] = 2.461112830064e-03;
C[12112] = -2.754836429800e-05;
C[12113] = 1.484490800750e-05;
C[12114] = 2.393793578211e-03;
C[12115] = -3.692665219220e-05;
C[12116] = 3.380634134902e-05;
C[12117] = -2.065361342326e-05;
C[12118] = -1.089022727731e-06;
C[12119] = 4.864613394610e-07;
C[12120] = 1.414514596631e-02;
C[12121] = 2.830295367470e-02;
C[12122] = 1.036850447610e-03;
C[12123] = -4.939199663554e-03;
C[12124] = 3.979457916328e-01;
C[12125] = 1.234635567194e-02;
C[12126] = -3.559531597786e-04;
C[12127] = 5.605918980464e-02;
C[12128] = 3.638847871252e-03;
C[12129] = 1.011784304664e-03;
C[12130] = 1.270985011146e-01;
C[12131] = 1.861494108777e-04;
C[12132] = -4.922763860826e-04;
C[12133] = 3.272743974747e-02;
C[12134] = 1.123578803196e-03;
C[12135] = -2.287478041220e-05;
C[12136] = 2.737455979765e-03;
C[12137] = 1.895800529448e-04;
C[12138] = -2.953246018889e-05;
C[12139] = 2.754827001011e-03;
C[12140] = 2.566792650592e-04;
C[12141] = -9.666663671354e-07;
C[12142] = -1.457235469325e-05;
C[12143] = 1.395540064667e-06;
C[12144] = -2.149603952049e-03;
C[12145] = 9.482278742724e-04;
C[12146] = 6.338685836808e-02;
C[12147] = -3.713845076672e-03;
C[12148] = 1.271800265921e-02;
C[12149] = 3.780051396171e-01;
C[12150] = 3.416663514415e-04;
C[12151] = 3.518697685525e-03;
C[12152] = 5.359777935928e-02;
C[12153] = -1.062864255719e-04;
C[12154] = 1.802412393218e-04;
C[12155] = 1.296756425142e-01;
C[12156] = -2.768089242946e-04;
C[12157] = 1.143096342329e-03;
C[12158] = 3.067850988546e-02;
C[12159] = 1.821413110232e-05;
C[12160] = 1.816104375603e-04;
C[12161] = 2.597202291614e-03;
C[12162] = 3.893646517054e-05;
C[12163] = 2.317771442764e-04;
C[12164] = 2.696540326619e-03;
C[12165] = 5.913244034530e-07;
C[12166] = 8.331277191121e-07;
C[12167] = -1.828443593814e-05;
C[12168] = 3.764870897578e-01;
C[12169] = -4.024709691728e-03;
C[12170] = -3.464793320281e-03;
C[12171] = 1.206383007968e-01;
C[12172] = 2.453476531961e-03;
C[12173] = -2.121251428630e-03;
C[12174] = 3.880627626844e-01;
C[12175] = 3.592108004090e-03;
C[12176] = 4.576920375452e-03;
C[12177] = 3.045872996929e-02;
C[12178] = -4.313103879608e-04;
C[12179] = -2.602054738685e-04;
C[12180] = 1.328162456853e-01;
C[12181] = 3.156607360776e-04;
C[12182] = -5.544040756467e-05;
C[12183] = 3.172893436112e-02;
C[12184] = 2.155903290587e-04;
C[12185] = 2.776270727445e-04;
C[12186] = 5.953123522956e-02;
C[12187] = 3.820100290437e-04;
C[12188] = 7.468873127548e-04;
C[12189] = 2.603154831264e-03;
C[12190] = 1.087036939867e-05;
C[12191] = 2.500285553037e-05;
C[12192] = -3.897718515486e-03;
C[12193] = 3.962320152396e-01;
C[12194] = 1.203730469741e-02;
C[12195] = 2.497866259066e-03;
C[12196] = 9.268905442687e-02;
C[12197] = -2.051440768699e-02;
C[12198] = 3.473667257392e-03;
C[12199] = 3.876211028573e-01;
C[12200] = 6.029614176086e-03;
C[12201] = -4.256245866319e-04;
C[12202] = 3.261318798792e-02;
C[12203] = 1.097716478209e-03;
C[12204] = 3.186200512180e-04;
C[12205] = 1.300287225659e-01;
C[12206] = -1.868141249752e-03;
C[12207] = 2.089602315552e-04;
C[12208] = 3.192984201049e-02;
C[12209] = 6.456845115346e-04;
C[12210] = 3.768835539255e-04;
C[12211] = 6.034319114766e-02;
C[12212] = 2.154088177004e-03;
C[12213] = 1.100877691910e-05;
C[12214] = 2.656229410465e-03;
C[12215] = 1.060797580564e-04;
C[12216] = -3.337431866691e-03;
C[12217] = 1.170747387667e-02;
C[12218] = 3.789127602088e-01;
C[12219] = -2.069057265741e-03;
C[12220] = -2.059179855143e-02;
C[12221] = 1.030859773048e-01;
C[12222] = 4.485385123725e-03;
C[12223] = 6.235862931986e-03;
C[12224] = 3.973152406519e-01;
C[12225] = -2.539145078787e-04;
C[12226] = 1.080986683512e-03;
C[12227] = 3.073901792491e-02;
C[12228] = -5.196079670541e-05;
C[12229] = -1.873300640715e-03;
C[12230] = 1.312945318841e-01;
C[12231] = 2.718428587654e-04;
C[12232] = 6.573895894698e-04;
C[12233] = 3.268975808788e-02;
C[12234] = 7.279408835882e-04;
C[12235] = 2.117395086499e-03;
C[12236] = 6.464537750869e-02;
C[12237] = 2.473503234500e-05;
C[12238] = 1.036125213425e-04;
C[12239] = 2.842199250489e-03;
C[12240] = 5.164211654433e-02;
C[12241] = -4.994667682468e-04;
C[12242] = -7.409705582529e-05;
C[12243] = 3.876568183947e-01;
C[12244] = 3.486152474865e-03;
C[12245] = 4.581871658445e-03;
C[12246] = 5.036678192410e-02;
C[12247] = -3.107678110705e-03;
C[12248] = -6.351005873876e-03;
C[12249] = 2.495220217627e-03;
C[12250] = -3.244235431008e-05;
C[12251] = -9.503429382138e-06;
C[12252] = 3.170187140847e-02;
C[12253] = 2.097925793867e-04;
C[12254] = 2.782752944134e-04;
C[12255] = 1.285755683042e-01;
C[12256] = -1.995373014067e-04;
C[12257] = -3.999962846899e-04;
C[12258] = 5.079222278584e-01;
C[12259] = 2.544277180999e-04;
C[12260] = 1.830070420156e-03;
C[12261] = 3.724160311901e-02;
C[12262] = 2.638485163569e-05;
C[12263] = 1.319712744495e-04;
C[12264] = -5.528277764488e-04;
C[12265] = 5.627590609706e-02;
C[12266] = 3.384645198257e-03;
C[12267] = 3.516528340384e-03;
C[12268] = 3.875975337500e-01;
C[12269] = 6.142493277156e-03;
C[12270] = -3.074393952363e-03;
C[12271] = 4.537361152516e-02;
C[12272] = -1.304218736159e-02;
C[12273] = -3.490836584512e-05;
C[12274] = 2.751903732592e-03;
C[12275] = 1.726736050758e-04;
C[12276] = 2.105516848116e-04;
C[12277] = 3.192827073667e-02;
C[12278] = 6.511649458145e-04;
C[12279] = -1.973183575171e-04;
C[12280] = 1.280526126464e-01;
C[12281] = -1.149612453420e-03;
C[12282] = 2.514099189402e-04;
C[12283] = 5.078674508545e-01;
C[12284] = 3.393033360936e-03;
C[12285] = 2.608716222261e-05;
C[12286] = 3.727475945315e-02;
C[12287] = 3.215019307373e-04;
C[12288] = -1.236725269861e-04;
C[12289] = 3.474875326491e-03;
C[12290] = 5.427237693588e-02;
C[12291] = 4.566617078050e-03;
C[12292] = 5.982484129653e-03;
C[12293] = 3.969291142483e-01;
C[12294] = -6.287486508068e-03;
C[12295] = -1.306228811403e-02;
C[12296] = 1.906083778773e-02;
C[12297] = -1.210723228890e-05;
C[12298] = 1.786485499608e-04;
C[12299] = 2.642175463387e-03;
C[12300] = 2.769401862510e-04;
C[12301] = 6.425425084391e-04;
C[12302] = 3.266401632764e-02;
C[12303] = -3.957616603027e-04;
C[12304] = -1.150952503583e-03;
C[12305] = 1.259461166651e-01;
C[12306] = 1.839420822385e-03;
C[12307] = 3.455221664868e-03;
C[12308] = 5.271933018591e-01;
C[12309] = 1.318614489326e-04;
C[12310] = 3.245196867694e-04;
C[12311] = 3.876724943156e-02;
C[12312] = -6.999922127033e+00;
C[12313] = 4.634689660844e-02;
C[12314] = -1.844654263344e-02;
C[12315] = 2.135695582326e+00;
C[12316] = -1.064949454640e-02;
C[12317] = -2.162132472423e-02;
C[12318] = 7.924805638733e-01;
C[12319] = -6.125191145045e-03;
C[12320] = 3.544006403721e-03;
C[12321] = 8.111423537607e-02;
C[12322] = 9.180333234207e-03;
C[12323] = -1.205894068285e-03;
C[12324] = 3.750672151163e-01;
C[12325] = -5.324739520245e-03;
C[12326] = -3.182279939165e-03;
C[12327] = 5.109432539976e-02;
C[12328] = -5.536856167900e-04;
C[12329] = 1.177949747895e-04;
C[12330] = 7.814115277841e-02;
C[12331] = -1.008605117025e-03;
C[12332] = 4.393475017122e-04;
C[12333] = 2.418282009118e-03;
C[12334] = -4.894031950258e-05;
C[12335] = 2.790037527601e-06;
C[12336] = 4.462567833569e-02;
C[12337] = -7.154787173272e+00;
C[12338] = 6.166414261475e-02;
C[12339] = -1.173703888211e-02;
C[12340] = 2.145043458988e+00;
C[12341] = 4.134288679437e-02;
C[12342] = -5.044052463557e-03;
C[12343] = 8.632633205795e-01;
C[12344] = 5.242423399421e-02;
C[12345] = 9.094272320570e-03;
C[12346] = 3.668125706543e-02;
C[12347] = 6.767539753148e-03;
C[12348] = -5.383967326295e-03;
C[12349] = 3.990346417846e-01;
C[12350] = 1.266845224653e-02;
C[12351] = -4.856504821921e-04;
C[12352] = 5.631305659342e-02;
C[12353] = 3.593044605672e-03;
C[12354] = -8.603512224025e-04;
C[12355] = 8.741922661356e-02;
C[12356] = 7.098245767339e-03;
C[12357] = -4.398993335813e-05;
C[12358] = 2.779115946059e-03;
C[12359] = 2.452009692060e-04;
C[12360] = -1.760968399224e-02;
C[12361] = 5.989169114799e-02;
C[12362] = -7.076593037850e+00;
C[12363] = -2.300063489929e-02;
C[12364] = 4.670879531009e-02;
C[12365] = 2.074418514727e+00;
C[12366] = 4.212099668309e-03;
C[12367] = 5.105696782154e-02;
C[12368] = 8.234198009431e-01;
C[12369] = -1.164051136225e-03;
C[12370] = 6.678917179810e-03;
C[12371] = 5.835312272684e-02;
C[12372] = -3.272691817043e-03;
C[12373] = 1.303094455410e-02;
C[12374] = 3.741621659377e-01;
C[12375] = 1.648351127365e-04;
C[12376] = 3.472624758028e-03;
C[12377] = 5.346578572808e-02;
C[12378] = 5.607097268609e-04;
C[12379] = 6.534119129467e-03;
C[12380] = 8.653499226433e-02;
C[12381] = 8.073886450065e-06;
C[12382] = 2.224685004696e-04;
C[12383] = 2.726369236373e-03;
C[12384] = 2.115418163751e+00;
C[12385] = 6.552760554423e-03;
C[12386] = -1.801959977146e-02;
C[12387] = -6.219165976030e+00;
C[12388] = 9.210584131651e-03;
C[12389] = -3.243960861900e-03;
C[12390] = 2.160675811900e+00;
C[12391] = 1.620008147726e-02;
C[12392] = 7.904889408721e-03;
C[12393] = 3.740533441876e-01;
C[12394] = -4.469477354469e-03;
C[12395] = -3.023640060651e-03;
C[12396] = 1.210755478397e-01;
C[12397] = 4.488573707404e-03;
C[12398] = 8.135772741731e-04;
C[12399] = 3.904256639013e-01;
C[12400] = 1.774510787665e-03;
C[12401] = 1.499179274219e-03;
C[12402] = 1.008979357038e+00;
C[12403] = 5.145945136122e-03;
C[12404] = 9.357116357695e-03;
C[12405] = 5.970185243829e-02;
C[12406] = 2.699044712751e-04;
C[12407] = 5.171213180478e-04;
C[12408] = 8.135664857058e-03;
C[12409] = 2.110767931123e+00;
C[12410] = 3.309483607413e-02;
C[12411] = 1.009837867377e-02;
C[12412] = -6.292657207716e+00;
C[12413] = -4.322335300273e-02;
C[12414] = 1.442410614436e-02;
C[12415] = 2.103270234626e+00;
C[12416] = -2.258471822242e-02;
C[12417] = -4.385481550072e-03;
C[12418] = 3.973208653914e-01;
C[12419] = 1.235024659230e-02;
C[12420] = 4.532963434510e-03;
C[12421] = 8.837491637708e-02;
C[12422] = -1.975577112972e-02;
C[12423] = 1.671306189160e-03;
C[12424] = 3.911858966251e-01;
C[12425] = 5.362718157634e-03;
C[12426] = 5.058417945092e-03;
C[12427] = 1.019737892926e+00;
C[12428] = 3.183920060156e-02;
C[12429] = 2.700830979259e-04;
C[12430] = 6.062344376996e-02;
C[12431] = 2.133266534219e-03;
C[12432] = -1.633871911466e-02;
C[12433] = 2.856525088914e-02;
C[12434] = 2.092570926560e+00;
C[12435] = -2.200077604124e-03;
C[12436] = -4.477117029159e-02;
C[12437] = -6.262525323549e+00;
C[12438] = 6.374463752659e-03;
C[12439] = -1.966560759980e-02;
C[12440] = 2.132899783094e+00;
C[12441] = -2.918149658686e-03;
C[12442] = 1.202957045127e-02;
C[12443] = 3.750697865294e-01;
C[12444] = 8.657714370620e-04;
C[12445] = -1.983316199416e-02;
C[12446] = 1.053029721408e-01;
C[12447] = 1.408598480000e-03;
C[12448] = 5.569765132480e-03;
C[12449] = 3.985445311774e-01;
C[12450] = 9.159977162958e-03;
C[12451] = 3.150276165404e-02;
C[12452] = 1.088471241966e+00;
C[12453] = 5.103187558992e-04;
C[12454] = 2.097990224160e-03;
C[12455] = 6.468726457054e-02;
C[12456] = 8.027127801421e-01;
C[12457] = -7.914324632921e-03;
C[12458] = -4.103168477027e-03;
C[12459] = 2.152556926105e+00;
C[12460] = 1.467381049382e-02;
C[12461] = 8.304194447057e-03;
C[12462] = -7.108535178605e+00;
C[12463] = -3.822218461225e-03;
C[12464] = -1.298966141532e-02;
C[12465] = 5.160593621320e-02;
C[12466] = -6.291640906602e-04;
C[12467] = -2.509282945304e-04;
C[12468] = 3.900197196116e-01;
C[12469] = 1.683791406633e-03;
C[12470] = 1.505085014720e-03;
C[12471] = 4.724223891882e-02;
C[12472] = -1.126787762062e-03;
C[12473] = -2.493724125674e-03;
C[12474] = 4.076379462040e+00;
C[12475] = 4.134900716716e-04;
C[12476] = 8.128489311144e-03;
C[12477] = 5.087443793634e-01;
C[12478] = 2.078193858147e-04;
C[12479] = 1.213348997300e-03;
C[12480] = -8.333191609181e-03;
C[12481] = 8.675976464277e-01;
C[12482] = 4.837591807618e-02;
C[12483] = 1.468848820313e-02;
C[12484] = 2.102798852480e+00;
C[12485] = -2.153300069640e-02;
C[12486] = -3.156535294370e-03;
C[12487] = -7.127685981461e+00;
C[12488] = -3.972105472248e-02;
C[12489] = -6.640856399968e-04;
C[12490] = 5.652977288583e-02;
C[12491] = 3.338572270760e-03;
C[12492] = 1.698931123959e-03;
C[12493] = 3.911623275178e-01;
C[12494] = 5.476395477650e-03;
C[12495] = -1.093503603719e-03;
C[12496] = 4.049957597120e-02;
C[12497] = -1.260483227246e-02;
C[12498] = 3.250332979011e-04;
C[12499] = 4.067257804612e+00;
C[12500] = 1.001488533339e-02;
C[12501] = 2.016652541994e-04;
C[12502] = 5.089172767504e-01;
C[12503] = 3.660491993344e-03;
C[12504] = -4.541635685200e-03;
C[12505] = 4.914478309899e-02;
C[12506] = 8.369117524751e-01;
C[12507] = 7.698823460676e-03;
C[12508] = -2.352731915109e-02;
C[12509] = 2.125177255022e+00;
C[12510] = -1.171927409916e-02;
C[12511] = -4.012306977126e-02;
C[12512] = -7.226263494708e+00;
C[12513] = -2.864871296566e-04;
C[12514] = 3.429072060911e-03;
C[12515] = 5.414038330468e-02;
C[12516] = 1.488875976816e-03;
C[12517] = 5.315588111200e-03;
C[12518] = 3.981584047738e-01;
C[12519] = -2.430204759865e-03;
C[12520] = -1.262493302490e-02;
C[12521] = 1.811059586665e-02;
C[12522] = 8.100371440878e-03;
C[12523] = 1.098337096437e-02;
C[12524] = 4.183273992343e+00;
C[12525] = 1.210226448870e-03;
C[12526] = 3.720518822482e-03;
C[12527] = 5.270927606078e-01;
C[12528] = 0.000000000000e+00;
C[12529] = 0.000000000000e+00;
C[12530] = 0.000000000000e+00;
C[12531] = 0.000000000000e+00;
C[12532] = 0.000000000000e+00;
C[12533] = 0.000000000000e+00;
C[12534] = 0.000000000000e+00;
C[12535] = 0.000000000000e+00;
C[12536] = 0.000000000000e+00;
C[12537] = 0.000000000000e+00;
C[12538] = 0.000000000000e+00;
C[12539] = 0.000000000000e+00;
C[12540] = 0.000000000000e+00;
C[12541] = 0.000000000000e+00;
C[12542] = 0.000000000000e+00;
C[12543] = 0.000000000000e+00;
C[12544] = 0.000000000000e+00;
C[12545] = 0.000000000000e+00;
C[12546] = 1.000000000000e+00;
C[12547] = 0.000000000000e+00;
C[12548] = 0.000000000000e+00;
C[12549] = 2.000000000000e-01;
C[12550] = -0.000000000000e+00;
C[12551] = -0.000000000000e+00;
C[12552] = 0.000000000000e+00;
C[12553] = 0.000000000000e+00;
C[12554] = 0.000000000000e+00;
C[12555] = 0.000000000000e+00;
C[12556] = 0.000000000000e+00;
C[12557] = 0.000000000000e+00;
C[12558] = 0.000000000000e+00;
C[12559] = 0.000000000000e+00;
C[12560] = 0.000000000000e+00;
C[12561] = 0.000000000000e+00;
C[12562] = 0.000000000000e+00;
C[12563] = 0.000000000000e+00;
C[12564] = 0.000000000000e+00;
C[12565] = 0.000000000000e+00;
C[12566] = 0.000000000000e+00;
C[12567] = 0.000000000000e+00;
C[12568] = 0.000000000000e+00;
C[12569] = 0.000000000000e+00;
C[12570] = 0.000000000000e+00;
C[12571] = 1.000000000000e+00;
C[12572] = 0.000000000000e+00;
C[12573] = -0.000000000000e+00;
C[12574] = 2.000000000000e-01;
C[12575] = -0.000000000000e+00;
C[12576] = 0.000000000000e+00;
C[12577] = 0.000000000000e+00;
C[12578] = 0.000000000000e+00;
C[12579] = 0.000000000000e+00;
C[12580] = 0.000000000000e+00;
C[12581] = 0.000000000000e+00;
C[12582] = 0.000000000000e+00;
C[12583] = 0.000000000000e+00;
C[12584] = 0.000000000000e+00;
C[12585] = 0.000000000000e+00;
C[12586] = 0.000000000000e+00;
C[12587] = 0.000000000000e+00;
C[12588] = 0.000000000000e+00;
C[12589] = 0.000000000000e+00;
C[12590] = 0.000000000000e+00;
C[12591] = 0.000000000000e+00;
C[12592] = 0.000000000000e+00;
C[12593] = 0.000000000000e+00;
C[12594] = 0.000000000000e+00;
C[12595] = 0.000000000000e+00;
C[12596] = 1.000000000000e+00;
C[12597] = -0.000000000000e+00;
C[12598] = -0.000000000000e+00;
C[12599] = 2.000000000000e-01;
C[12600] = 8.266161281419e-02;
C[12601] = 1.130907868292e-04;
C[12602] = 4.840644279432e-03;
C[12603] = 3.709405320722e-01;
C[12604] = -1.950524297909e-03;
C[12605] = 4.273153889431e-04;
C[12606] = 5.080736290274e-02;
C[12607] = 2.323401854968e-04;
C[12608] = 1.076638682276e-05;
C[12609] = 1.314031752301e-01;
C[12610] = -2.486960003139e-04;
C[12611] = 4.832207012530e-04;
C[12612] = 2.999139607056e-02;
C[12613] = 8.559357514119e-05;
C[12614] = 8.103337341053e-05;
C[12615] = 2.446261179045e-03;
C[12616] = 2.618220416998e-05;
C[12617] = 2.514066612536e-06;
C[12618] = 2.395082094215e-03;
C[12619] = 6.125591683691e-05;
C[12620] = 1.396593618284e-05;
C[12621] = -2.102745332196e-05;
C[12622] = 3.260310091203e-06;
C[12623] = 8.534866640658e-07;
C[12624] = -5.611818261444e-05;
C[12625] = 4.512228360797e-02;
C[12626] = 2.709605225833e-03;
C[12627] = -1.777542807063e-03;
C[12628] = 4.017301648103e-01;
C[12629] = 1.087660570570e-02;
C[12630] = 2.222499091770e-04;
C[12631] = 5.732338954936e-02;
C[12632] = 3.051980110671e-03;
C[12633] = -2.599765982768e-04;
C[12634] = 1.281017191550e-01;
C[12635] = -6.778584771596e-05;
C[12636] = 9.736451523283e-05;
C[12637] = 3.300555819433e-02;
C[12638] = 9.354745277125e-04;
C[12639] = 2.486270966281e-05;
C[12640] = 2.796565566311e-03;
C[12641] = 1.564316112679e-04;
C[12642] = 5.845881242715e-05;
C[12643] = 2.831276424495e-03;
C[12644] = 2.179154103750e-04;
C[12645] = 3.107521724519e-06;
C[12646] = -1.627219416705e-05;
C[12647] = 5.392109476961e-07;
C[12648] = 4.850796981149e-03;
C[12649] = 2.695036287651e-03;
C[12650] = 4.899652490545e-02;
C[12651] = 4.960927617413e-04;
C[12652] = 1.109081618577e-02;
C[12653] = 3.692074883865e-01;
C[12654] = 1.143689778099e-07;
C[12655] = 2.912468544153e-03;
C[12656] = 5.272149465143e-02;
C[12657] = 4.838975480342e-04;
C[12658] = -6.875711026149e-05;
C[12659] = 1.288053979442e-01;
C[12660] = 8.566134446710e-05;
C[12661] = 9.462597893314e-04;
C[12662] = 3.000713393192e-02;
C[12663] = 1.258574285797e-06;
C[12664] = 1.480826659111e-04;
C[12665] = 2.557059385185e-03;
C[12666] = 1.045523057757e-05;
C[12667] = 1.956152651716e-04;
C[12668] = 2.658429115011e-03;
C[12669] = 6.655953079118e-07;
C[12670] = 1.164855676483e-07;
C[12671] = -1.722196965568e-05;
C[12672] = 3.715369554451e-01;
C[12673] = -1.806382623118e-03;
C[12674] = 1.661871867029e-04;
C[12675] = 1.228011275202e-01;
C[12676] = 2.533099472489e-03;
C[12677] = 1.796983105563e-03;
C[12678] = 3.918684244087e-01;
C[12679] = -1.101078012865e-03;
C[12680] = -1.221177660892e-03;
C[12681] = 3.003115762875e-02;
C[12682] = 9.544186082921e-05;
C[12683] = 6.366763946454e-05;
C[12684] = 1.330588151792e-01;
C[12685] = 2.399420775925e-05;
C[12686] = 1.639416940828e-04;
C[12687] = 3.196285109042e-02;
C[12688] = -1.483829787310e-04;
C[12689] = -1.909172663987e-04;
C[12690] = 6.030673853209e-02;
C[12691] = -3.368749775424e-04;
C[12692] = -4.177034372141e-04;
C[12693] = 2.631678524055e-03;
C[12694] = -1.922227540789e-05;
C[12695] = -2.545392030855e-05;
C[12696] = -1.689547617882e-03;
C[12697] = 4.017854247009e-01;
C[12698] = 1.182215217534e-02;
C[12699] = 2.397010589225e-03;
C[12700] = 8.572581513267e-02;
C[12701] = -1.854401052085e-02;
C[12702] = -1.113874416776e-03;
C[12703] = 3.910520605660e-01;
C[12704] = 4.954996197308e-03;
C[12705] = 1.029920204763e-04;
C[12706] = 3.300924218703e-02;
C[12707] = 9.950155219697e-04;
C[12708] = 1.492161554167e-05;
C[12709] = 1.295933367079e-01;
C[12710] = -1.648780753837e-03;
C[12711] = -1.485276498376e-04;
C[12712] = 3.206876927012e-02;
C[12713] = 5.612314973315e-04;
C[12714] = -3.264483258116e-04;
C[12715] = 6.100453987535e-02;
C[12716] = 2.110053775611e-03;
C[12717] = -1.860309061266e-05;
C[12718] = 2.675238726172e-03;
C[12719] = 1.042921644533e-04;
C[12720] = 2.090282826369e-04;
C[12721] = 1.162241059521e-02;
C[12722] = 3.695305198393e-01;
C[12723] = 1.752863395427e-03;
C[12724] = -1.860038636639e-02;
C[12725] = 1.077660710472e-01;
C[12726] = -1.255831997098e-03;
C[12727] = 5.120274506609e-03;
C[12728] = 4.000024869532e-01;
C[12729] = 6.648089965678e-05;
C[12730] = 9.851948536792e-04;
C[12731] = 3.002866936210e-02;
C[12732] = 1.610003800737e-04;
C[12733] = -1.652539143539e-03;
C[12734] = 1.317612529979e-01;
C[12735] = -1.924777977790e-04;
C[12736] = 5.703955711762e-04;
C[12737] = 3.284774610004e-02;
C[12738] = -3.924560391636e-04;
C[12739] = 2.097308290981e-03;
C[12740] = 6.479594240203e-02;
C[12741] = -2.401115967734e-05;
C[12742] = 1.031665317767e-04;
C[12743] = 2.843979124536e-03;
C[12744] = 5.105890057326e-02;
C[12745] = -6.176984265792e-04;
C[12746] = -4.663402375279e-04;
C[12747] = 3.914511274468e-01;
C[12748] = -3.665627026137e-04;
C[12749] = -9.682675522032e-04;
C[12750] = 4.404962825575e-02;
C[12751] = 5.810685463641e-04;
C[12752] = 5.805601607454e-04;
C[12753] = 2.463030357079e-03;
C[12754] = -3.113384605427e-05;
C[12755] = -2.983839948125e-05;
C[12756] = 3.193503129297e-02;
C[12757] = -9.870686889353e-05;
C[12758] = -1.733068347861e-04;
C[12759] = 1.280811456048e-01;
C[12760] = 1.132976896578e-04;
C[12761] = 1.693964892223e-04;
C[12762] = 5.111515410021e-01;
C[12763] = 4.251368347007e-04;
C[12764] = 9.063836730285e-04;
C[12765] = 3.754176663716e-02;
C[12766] = 1.683464502368e-05;
C[12767] = 3.272213172087e-05;
C[12768] = -5.887199584687e-04;
C[12769] = 5.705372512171e-02;
C[12770] = 3.123284967966e-03;
C[12771] = -4.081449759506e-04;
C[12772] = 3.908377374687e-01;
C[12773] = 4.937116547131e-03;
C[12774] = 5.985239585501e-04;
C[12775] = 3.705010901083e-02;
C[12776] = -1.257954756342e-02;
C[12777] = -2.855513876106e-05;
C[12778] = 2.778587937801e-03;
C[12779] = 1.621370941652e-04;
C[12780] = -1.021874429367e-04;
C[12781] = 3.205448106363e-02;
C[12782] = 5.581850405443e-04;
C[12783] = 1.144613838035e-04;
C[12784] = 1.275000675806e-01;
C[12785] = -1.115733490997e-03;
C[12786] = 4.158236339337e-04;
C[12787] = 5.123597718690e-01;
C[12788] = 4.455051521820e-03;
C[12789] = 1.627153282751e-05;
C[12790] = 3.768192869168e-02;
C[12791] = 3.970531359225e-04;
C[12792] = -4.677235126100e-04;
C[12793] = 3.259708070781e-03;
C[12794] = 5.314858447561e-02;
C[12795] = -9.808185403373e-04;
C[12796] = 4.754387250091e-03;
C[12797] = 3.996881534777e-01;
C[12798] = 6.168990231007e-04;
C[12799] = -1.258184420931e-02;
C[12800] = 1.812935828333e-02;
C[12801] = -2.938526001631e-05;
C[12802] = 1.702801419419e-04;
C[12803] = 2.585532040130e-03;
C[12804] = -1.748933250284e-04;
C[12805] = 5.478575675171e-04;
C[12806] = 3.282679053500e-02;
C[12807] = 1.718190800460e-04;
C[12808] = -1.115886600723e-03;
C[12809] = 1.259096626359e-01;
C[12810] = 8.807576845574e-04;
C[12811] = 4.494353110068e-03;
C[12812] = 5.263860051943e-01;
C[12813] = 3.122966667843e-05;
C[12814] = 3.988723657170e-04;
C[12815] = 3.869522262911e-02;
C[12816] = -6.992753031943e+00;
C[12817] = -1.095066861370e-02;
C[12818] = 3.455401719995e-02;
C[12819] = 2.100408649517e+00;
C[12820] = -1.447103061149e-03;
C[12821] = 1.930701322614e-02;
C[12822] = 7.879037011609e-01;
C[12823] = 7.580660222890e-03;
C[12824] = 2.225152611122e-03;
C[12825] = 7.921835632481e-02;
C[12826] = -5.873458357169e-03;
C[12827] = 7.130755126197e-03;
C[12828] = 3.691134404311e-01;
C[12829] = 3.652914489999e-03;
C[12830] = 2.407723925575e-03;
C[12831] = 5.072861997906e-02;
C[12832] = 5.229945314402e-04;
C[12833] = 2.836181969186e-05;
C[12834] = 7.801072300187e-02;
C[12835] = 9.069595080175e-04;
C[12836] = -8.943832187574e-06;
C[12837] = 2.403291726219e-03;
C[12838] = 3.264511320516e-05;
C[12839] = -1.148103850887e-05;
C[12840] = -1.433484800257e-02;
C[12841] = -7.078154621447e+00;
C[12842] = 3.823781418648e-02;
C[12843] = 1.941097372672e-03;
C[12844] = 2.162090942963e+00;
C[12845] = 1.875968768522e-02;
C[12846] = 7.281464546750e-03;
C[12847] = 8.817097717571e-01;
C[12848] = 4.233666245826e-02;
C[12849] = -6.042667326612e-03;
C[12850] = 4.503512472800e-02;
C[12851] = -1.520990880448e-03;
C[12852] = 3.844626331213e-03;
C[12853] = 4.025388152373e-01;
C[12854] = 9.332132324061e-03;
C[12855] = 4.804524752771e-04;
C[12856] = 5.729868508478e-02;
C[12857] = 2.968338931940e-03;
C[12858] = 8.148607479078e-04;
C[12859] = 8.941683680055e-02;
C[12860] = 6.330881776548e-03;
C[12861] = 2.751579030322e-05;
C[12862] = 2.814586046412e-03;
C[12863] = 2.244408421885e-04;
C[12864] = 3.475707123430e-02;
C[12865] = 3.794643542282e-02;
C[12866] = -7.161124704600e+00;
C[12867] = 2.073415040336e-02;
C[12868] = 2.169308283141e-02;
C[12869] = 2.027018655903e+00;
C[12870] = 1.890996532044e-03;
C[12871] = 4.056121487708e-02;
C[12872] = 8.097556781526e-01;
C[12873] = 7.140907827915e-03;
C[12874] = -1.535559818631e-03;
C[12875] = 4.970312033895e-02;
C[12876] = 2.489442296145e-03;
C[12877] = 9.524363786147e-03;
C[12878] = 3.661774444630e-01;
C[12879] = -4.361316627496e-06;
C[12880] = 2.823537907144e-03;
C[12881] = 5.254576820383e-02;
C[12882] = -7.664412687722e-05;
C[12883] = 5.774751161542e-03;
C[12884] = 8.490354920181e-02;
C[12885] = -1.437983004161e-05;
C[12886] = 2.023293712573e-04;
C[12887] = 2.672630584809e-03;
C[12888] = 2.112337116974e+00;
C[12889] = 1.364301051587e-03;
C[12890] = 1.413603890259e-02;
C[12891] = -6.212354632534e+00;
C[12892] = 1.209901125373e-03;
C[12893] = 5.811868888338e-03;
C[12894] = 2.170124137837e+00;
C[12895] = -1.192259559651e-02;
C[12896] = -6.800100784592e-03;
C[12897] = 3.697098638039e-01;
C[12898] = 3.815786515159e-03;
C[12899] = 2.159536721107e-03;
C[12900] = 1.242182005614e-01;
C[12901] = -2.017918203367e-03;
C[12902] = 1.854602131937e-03;
C[12903] = 3.919244409350e-01;
C[12904] = -2.430443589218e-03;
C[12905] = -2.970388012538e-03;
C[12906] = 1.020551916706e+00;
C[12907] = -6.291642693104e-03;
C[12908] = -8.078084670881e-03;
C[12909] = 6.042670954556e-02;
C[12910] = -3.930661658852e-04;
C[12911] = -5.568898744557e-04;
C[12912] = 3.772430539387e-03;
C[12913] = 2.163196140774e+00;
C[12914] = 3.631980262290e-02;
C[12915] = -1.511876539899e-03;
C[12916] = -6.301296782172e+00;
C[12917] = -3.939589654240e-02;
C[12918] = -1.205185037304e-02;
C[12919] = 2.093312937660e+00;
C[12920] = -2.190082834297e-02;
C[12921] = 3.913891170025e-03;
C[12922] = 4.025940751279e-01;
C[12923] = 1.025569977572e-02;
C[12924] = -2.154007086630e-03;
C[12925] = 8.605446688018e-02;
C[12926] = -1.707678960722e-02;
C[12927] = -2.411031366793e-03;
C[12928] = 3.898060796479e-01;
C[12929] = 4.742708877386e-03;
C[12930] = -6.058348563663e-03;
C[12931] = 1.026728939648e+00;
C[12932] = 3.163372041025e-02;
C[12933] = -3.782247365750e-04;
C[12934] = 6.110418964034e-02;
C[12935] = 2.102063772709e-03;
C[12936] = 1.494127110002e-02;
C[12937] = 3.367578547523e-02;
C[12938] = 2.033479284959e+00;
C[12939] = 4.929474685618e-03;
C[12940] = -4.052341345315e-02;
C[12941] = -6.252898146670e+00;
C[12942] = -7.307951819965e-03;
C[12943] = -1.976303557145e-02;
C[12944] = 2.150964673644e+00;
C[12945] = 2.189436819269e-03;
C[12946] = 1.007793721356e-02;
C[12947] = 3.665004759158e-01;
C[12948] = 1.810482421801e-03;
C[12949] = -1.713316545276e-02;
C[12950] = 1.105991124656e-01;
C[12951] = -2.985065931516e-03;
C[12952] = 4.909048558992e-03;
C[12953] = 4.002453021319e-01;
C[12954] = -7.613721110693e-03;
C[12955] = 3.170791875599e-02;
C[12956] = 1.090650384332e+00;
C[12957] = -5.284189213278e-04;
C[12958] = 2.097698371983e-03;
C[12959] = 6.475577243208e-02;
C[12960] = 7.929344545712e-01;
C[12961] = -9.517502168373e-03;
C[12962] = -7.438095598070e-03;
C[12963] = 2.161778198600e+00;
C[12964] = 2.894383910195e-03;
C[12965] = -1.556662922074e-03;
C[12966] = -7.130366913169e+00;
C[12967] = 1.384086567551e-03;
C[12968] = 1.868820261814e-03;
C[12969] = 5.098015764957e-02;
C[12970] = -3.594958604791e-04;
C[12971] = -4.708159231332e-04;
C[12972] = 3.915071439731e-01;
C[12973] = -1.663719652631e-03;
C[12974] = -2.697501486621e-03;
C[12975] = 4.221548962876e-02;
C[12976] = 1.782976155357e-03;
C[12977] = 2.944070595274e-03;
C[12978] = 4.101384988254e+00;
C[12979] = 5.022606663041e-03;
C[12980] = 8.425559782356e-03;
C[12981] = 5.130280737340e-01;
C[12982] = 2.068700176586e-04;
C[12983] = 2.504240117193e-04;
C[12984] = -8.840542656421e-03;
C[12985] = 8.763164832041e-01;
C[12986] = 4.477754335333e-02;
C[12987] = 1.936065141787e-03;
C[12988] = 2.089026475715e+00;
C[12989] = -2.342619476101e-02;
C[12990] = 1.733194811270e-03;
C[12991] = -7.156042841309e+00;
C[12992] = -4.140072731303e-02;
C[12993] = -2.980656125253e-04;
C[12994] = 5.702902065714e-02;
C[12995] = 3.034354330957e-03;
C[12996] = -1.737510552303e-03;
C[12997] = 3.895917565507e-01;
C[12998] = 4.725890599514e-03;
C[12999] = 1.800431567543e-03;
C[13000] = 3.588171984815e-02;
C[13001] = -1.249593629268e-02;
C[13002] = 4.852626915818e-03;
C[13003] = 4.105071965827e+00;
C[13004] = 1.845273158845e-02;
C[13005] = 1.959860759632e-04;
C[13006] = 5.147850404838e-01;
C[13007] = 4.691669769894e-03;
C[13008] = -7.344645377533e-03;
C[13009] = 4.649122166047e-02;
C[13010] = 8.182974746363e-01;
C[13011] = -1.992918373507e-03;
C[13012] = -2.591300728729e-02;
C[13013] = 2.144678004133e+00;
C[13014] = 2.595597508922e-03;
C[13015] = -4.144666023084e-02;
C[13016] = -7.222367616936e+00;
C[13017] = -4.501280797409e-04;
C[13018] = 3.176066892051e-03;
C[13019] = 5.297285802802e-02;
C[13020] = -2.730028891983e-03;
C[13021] = 4.542099930170e-03;
C[13022] = 3.999309686564e-01;
C[13023] = 2.980409457630e-03;
C[13024] = -1.249823293857e-02;
C[13025] = 1.857271997200e-02;
C[13026] = 7.953854785405e-03;
C[13027] = 1.904054211288e-02;
C[13028] = 4.175534710207e+00;
C[13029] = 2.214565762524e-04;
C[13030] = 4.723107184167e-03;
C[13031] = 5.259092030433e-01;
C[13032] = 0.000000000000e+00;
C[13033] = 0.000000000000e+00;
C[13034] = 0.000000000000e+00;
C[13035] = 0.000000000000e+00;
C[13036] = 0.000000000000e+00;
C[13037] = 0.000000000000e+00;
C[13038] = 0.000000000000e+00;
C[13039] = 0.000000000000e+00;
C[13040] = 0.000000000000e+00;
C[13041] = 0.000000000000e+00;
C[13042] = 0.000000000000e+00;
C[13043] = 0.000000000000e+00;
C[13044] = 0.000000000000e+00;
C[13045] = 0.000000000000e+00;
C[13046] = 0.000000000000e+00;
C[13047] = 0.000000000000e+00;
C[13048] = 0.000000000000e+00;
C[13049] = 0.000000000000e+00;
C[13050] = 1.000000000000e+00;
C[13051] = 0.000000000000e+00;
C[13052] = 0.000000000000e+00;
C[13053] = 2.000000000000e-01;
C[13054] = -0.000000000000e+00;
C[13055] = -0.000000000000e+00;
C[13056] = 0.000000000000e+00;
C[13057] = 0.000000000000e+00;
C[13058] = 0.000000000000e+00;
C[13059] = 0.000000000000e+00;
C[13060] = 0.000000000000e+00;
C[13061] = 0.000000000000e+00;
C[13062] = 0.000000000000e+00;
C[13063] = 0.000000000000e+00;
C[13064] = 0.000000000000e+00;
C[13065] = 0.000000000000e+00;
C[13066] = 0.000000000000e+00;
C[13067] = 0.000000000000e+00;
C[13068] = 0.000000000000e+00;
C[13069] = 0.000000000000e+00;
C[13070] = 0.000000000000e+00;
C[13071] = 0.000000000000e+00;
C[13072] = 0.000000000000e+00;
C[13073] = 0.000000000000e+00;
C[13074] = 0.000000000000e+00;
C[13075] = 1.000000000000e+00;
C[13076] = 0.000000000000e+00;
C[13077] = -0.000000000000e+00;
C[13078] = 2.000000000000e-01;
C[13079] = -0.000000000000e+00;
C[13080] = 0.000000000000e+00;
C[13081] = 0.000000000000e+00;
C[13082] = 0.000000000000e+00;
C[13083] = 0.000000000000e+00;
C[13084] = 0.000000000000e+00;
C[13085] = 0.000000000000e+00;
C[13086] = 0.000000000000e+00;
C[13087] = 0.000000000000e+00;
C[13088] = 0.000000000000e+00;
C[13089] = 0.000000000000e+00;
C[13090] = 0.000000000000e+00;
C[13091] = 0.000000000000e+00;
C[13092] = 0.000000000000e+00;
C[13093] = 0.000000000000e+00;
C[13094] = 0.000000000000e+00;
C[13095] = 0.000000000000e+00;
C[13096] = 0.000000000000e+00;
C[13097] = 0.000000000000e+00;
C[13098] = 0.000000000000e+00;
C[13099] = 0.000000000000e+00;
C[13100] = 1.000000000000e+00;
C[13101] = -0.000000000000e+00;
C[13102] = -0.000000000000e+00;
C[13103] = 2.000000000000e-01;
C[13104] = 7.831064075868e-02;
C[13105] = -1.161448367637e-02;
C[13106] = 8.672211723918e-03;
C[13107] = 3.658065474886e-01;
C[13108] = 9.150679380483e-03;
C[13109] = 3.863758165974e-03;
C[13110] = 4.924973491727e-02;
C[13111] = 2.033718487059e-03;
C[13112] = -3.162603543911e-05;
C[13113] = 1.315078847062e-01;
C[13114] = -1.414520166225e-03;
C[13115] = 7.031206961604e-04;
C[13116] = 2.934158264030e-02;
C[13117] = 1.169088555165e-03;
C[13118] = 2.822402696276e-04;
C[13119] = 2.355513261396e-03;
C[13120] = 1.250862341112e-04;
C[13121] = -3.520120146587e-06;
C[13122] = 2.297618978932e-03;
C[13123] = 1.837005578751e-04;
C[13124] = -9.933313347353e-06;
C[13125] = -2.336685091303e-05;
C[13126] = 5.254129366367e-06;
C[13127] = 3.597936673174e-08;
C[13128] = -1.177043695006e-02;
C[13129] = 4.505263425528e-02;
C[13130] = -2.364375731201e-03;
C[13131] = 9.436445256114e-03;
C[13132] = 4.018182478651e-01;
C[13133] = 8.021547975460e-03;
C[13134] = 1.970304740223e-03;
C[13135] = 5.631760491582e-02;
C[13136] = 2.475779987257e-03;
C[13137] = -1.424917051137e-03;
C[13138] = 1.282018771111e-01;
C[13139] = -1.166778565551e-04;
C[13140] = 1.186977442033e-03;
C[13141] = 3.285542088768e-02;
C[13142] = 6.493552104327e-04;
C[13143] = 1.209761932133e-04;
C[13144] = 2.737234193532e-03;
C[13145] = 1.252253545975e-04;
C[13146] = 1.754535420671e-04;
C[13147] = 2.787682150703e-03;
C[13148] = 1.787495753597e-04;
C[13149] = 5.052005722777e-06;
C[13150] = -1.732551710954e-05;
C[13151] = 1.830280843223e-07;
C[13152] = 8.635124933522e-03;
C[13153] = -2.442556426200e-03;
C[13154] = 5.595196072776e-02;
C[13155] = 4.031150574622e-03;
C[13156] = 8.420006236597e-03;
C[13157] = 3.608236974810e-01;
C[13158] = -8.708123668340e-05;
C[13159] = 2.328541060096e-03;
C[13160] = 5.035491617087e-02;
C[13161] = 7.006482434674e-04;
C[13162] = -1.218899028884e-04;
C[13163] = 1.299940615699e-01;
C[13164] = 2.928035557641e-04;
C[13165] = 6.736026999159e-04;
C[13166] = 2.887321068142e-02;
C[13167] = -7.243822488885e-06;
C[13168] = 1.158568817383e-04;
C[13169] = 2.412321317215e-03;
C[13170] = -1.849277215619e-05;
C[13171] = 1.561562798279e-04;
C[13172] = 2.432742329517e-03;
C[13173] = -2.125100838002e-07;
C[13174] = -3.542245630654e-07;
C[13175] = -2.315387798952e-05;
C[13176] = 3.661247394016e-01;
C[13177] = 9.202665684830e-03;
C[13178] = 3.725971912948e-03;
C[13179] = 1.288863768330e-01;
C[13180] = -7.387687369911e-03;
C[13181] = 1.598612563631e-03;
C[13182] = 3.912002806811e-01;
C[13183] = -4.242651286561e-03;
C[13184] = -3.659005018629e-03;
C[13185] = 2.936279543450e-02;
C[13186] = 1.171392137281e-03;
C[13187] = 2.724583116525e-04;
C[13188] = 1.339172863351e-01;
C[13189] = -1.026614004432e-03;
C[13190] = 7.367053638281e-05;
C[13191] = 3.179875562114e-02;
C[13192] = -2.851101639989e-04;
C[13193] = -2.731869619391e-04;
C[13194] = 6.044976995408e-02;
C[13195] = -5.528712669309e-04;
C[13196] = -6.753272290785e-04;
C[13197] = 2.629432347867e-03;
C[13198] = -1.960762241910e-05;
C[13199] = -2.760505260745e-05;
C[13200] = 9.264311303316e-03;
C[13201] = 4.019599642398e-01;
C[13202] = 8.039868228170e-03;
C[13203] = -7.587708938183e-03;
C[13204] = 8.886986094945e-02;
C[13205] = -1.622661901387e-02;
C[13206] = -4.164947534540e-03;
C[13207] = 3.878725039677e-01;
C[13208] = 5.591936526899e-03;
C[13209] = 1.176664016688e-03;
C[13210] = 3.286486864600e-02;
C[13211] = 6.482601660208e-04;
C[13212] = -1.039948775650e-03;
C[13213] = 1.300184393881e-01;
C[13214] = -1.422383236582e-03;
C[13215] = -2.800358428313e-04;
C[13216] = 3.180556086457e-02;
C[13217] = 6.481665277159e-04;
C[13218] = -5.405633942361e-04;
C[13219] = 6.122111800743e-02;
C[13220] = 2.345603818032e-03;
C[13221] = -1.933675885703e-05;
C[13222] = 2.684006676979e-03;
C[13223] = 1.170797722987e-04;
C[13224] = 3.688565686174e-03;
C[13225] = 7.870276413201e-03;
C[13226] = 3.607824160602e-01;
C[13227] = 1.494630721653e-03;
C[13228] = -1.635594348972e-02;
C[13229] = 1.163467625005e-01;
C[13230] = -3.583315861210e-03;
C[13231] = 5.762298918825e-03;
C[13232] = 4.025805951957e-01;
C[13233] = 2.705607709743e-04;
C[13234] = 6.392704396154e-04;
C[13235] = 2.887045858670e-02;
C[13236] = 6.673841358426e-05;
C[13237] = -1.431004868305e-03;
C[13238] = 1.329005082615e-01;
C[13239] = -2.678680571519e-04;
C[13240] = 6.584458400745e-04;
C[13241] = 3.300980326521e-02;
C[13242] = -6.564979353839e-04;
C[13243] = 2.358947490698e-03;
C[13244] = 6.490034947165e-02;
C[13245] = -2.709706003384e-05;
C[13246] = 1.171426255543e-04;
C[13247] = 2.844908966579e-03;
C[13248] = 4.985525911936e-02;
C[13249] = 1.198637493477e-03;
C[13250] = 2.619605768226e-05;
C[13251] = 3.902491847591e-01;
C[13252] = -3.245296800991e-03;
C[13253] = -3.460169554544e-03;
C[13254] = 4.178587344582e-02;
C[13255] = 2.364893690861e-03;
C[13256] = 4.448265006494e-03;
C[13257] = 2.395881541535e-03;
C[13258] = 6.953171009696e-05;
C[13259] = 3.079971354929e-07;
C[13260] = 3.173534922634e-02;
C[13261] = -2.187257939280e-04;
C[13262] = -2.596583033742e-04;
C[13263] = 1.280399867459e-01;
C[13264] = 2.028298699485e-04;
C[13265] = 3.735543217670e-04;
C[13266] = 5.158285157276e-01;
C[13267] = -4.198629055887e-04;
C[13268] = -9.523063979982e-04;
C[13269] = 3.784900347412e-02;
C[13270] = -5.497852308767e-05;
C[13271] = -1.149133349517e-04;
C[13272] = 1.266526910570e-03;
C[13273] = 5.653081633060e-02;
C[13274] = 2.706213827158e-03;
C[13275] = -3.299863734255e-03;
C[13276] = 3.874757269566e-01;
C[13277] = 5.490973092883e-03;
C[13278] = 2.351525961254e-03;
C[13279] = 3.560088974121e-02;
C[13280] = -1.270530956501e-02;
C[13281] = 7.394012901203e-05;
C[13282] = 2.751448287850e-03;
C[13283] = 1.410350662091e-04;
C[13284] = -2.222576605118e-04;
C[13285] = 3.177910906382e-02;
C[13286] = 6.403574516783e-04;
C[13287] = 2.019386879747e-04;
C[13288] = 1.274728025715e-01;
C[13289] = -1.162354943658e-03;
C[13290] = -4.250796367578e-04;
C[13291] = 5.177194134796e-01;
C[13292] = 4.378778415762e-03;
C[13293] = -5.514112994959e-05;
C[13294] = 3.801225851234e-02;
C[13295] = 3.785892447999e-04;
C[13296] = 8.577528737057e-05;
C[13297] = 2.884085514851e-03;
C[13298] = 5.178187100456e-02;
C[13299] = -3.488225583780e-03;
C[13300] = 5.297349915032e-03;
C[13301] = 4.011955022215e-01;
C[13302] = 4.424708030607e-03;
C[13303] = -1.271157165863e-02;
C[13304] = 1.911468843584e-02;
C[13305] = 4.306634707391e-06;
C[13306] = 1.524457231039e-04;
C[13307] = 2.507451639460e-03;
C[13308] = -2.618016662826e-04;
C[13309] = 6.285274202581e-04;
C[13310] = 3.291746373360e-02;
C[13311] = 3.719838567079e-04;
C[13312] = -1.162772416566e-03;
C[13313] = 1.259686330113e-01;
C[13314] = -9.649373789141e-04;
C[13315] = 4.388301849957e-03;
C[13316] = 5.253335452920e-01;
C[13317] = -1.153484451840e-04;
C[13318] = 3.790336302717e-04;
C[13319] = 3.862340072543e-02;
C[13320] = -7.009659215605e+00;
C[13321] = -4.139716360973e-02;
C[13322] = 5.868708872857e-02;
C[13323] = 2.074331281720e+00;
C[13324] = 2.550608055280e-02;
C[13325] = 3.299433770170e-02;
C[13326] = 7.607647282774e-01;
C[13327] = 3.418058209104e-02;
C[13328] = 8.644355854049e-04;
C[13329] = 8.453643899229e-02;
C[13330] = -1.948771498001e-02;
C[13331] = 9.290312517730e-03;
C[13332] = 3.594199214031e-01;
C[13333] = 1.596937440062e-02;
C[13334] = 3.922272246759e-03;
C[13335] = 4.878002026317e-02;
C[13336] = 2.490981142406e-03;
C[13337] = 5.798644580426e-05;
C[13338] = 7.544639055427e-02;
C[13339] = 4.291723592465e-03;
C[13340] = -3.798954006660e-05;
C[13341] = 2.304516987563e-03;
C[13342] = 1.672753994851e-04;
C[13343] = 1.580052885825e-06;
C[13344] = -4.451622908343e-02;
C[13345] = -7.076881166084e+00;
C[13346] = 2.085729916168e-02;
C[13347] = 3.065305006945e-02;
C[13348] = 2.161356291667e+00;
C[13349] = 1.203931421529e-02;
C[13350] = 3.333602998275e-02;
C[13351] = 8.643688220808e-01;
C[13352] = 3.314377683059e-02;
C[13353] = -1.964366825369e-02;
C[13354] = 4.719283022813e-02;
C[13355] = 1.395244979210e-03;
C[13356] = 1.626871373208e-02;
C[13357] = 3.994278281246e-01;
C[13358] = 5.717753326794e-03;
C[13359] = 2.414398707312e-03;
C[13360] = 5.612273756513e-02;
C[13361] = 2.374414069265e-03;
C[13362] = 4.080657467216e-03;
C[13363] = 8.857746245458e-02;
C[13364] = 5.377785418705e-03;
C[13365] = 1.579160079595e-04;
C[13366] = 2.788663612504e-03;
C[13367] = 1.958049053978e-04;
C[13368] = 5.794535292065e-02;
C[13369] = 1.929368526168e-02;
C[13370] = -7.130380610976e+00;
C[13371] = 3.591822038961e-02;
C[13372] = 1.880505201365e-02;
C[13373] = 2.000339544775e+00;
C[13374] = 1.415778461819e-04;
C[13375] = 3.132974008238e-02;
C[13376] = 7.697997351763e-01;
C[13377] = 9.253225727334e-03;
C[13378] = 1.317064284211e-03;
C[13379] = 6.857056327197e-02;
C[13380] = 4.117742346076e-03;
C[13381] = 6.179005226488e-03;
C[13382] = 3.499827253254e-01;
C[13383] = -1.059759189362e-05;
C[13384] = 2.203136835848e-03;
C[13385] = 4.968714412973e-02;
C[13386] = -2.528660663557e-04;
C[13387] = 4.798742920679e-03;
C[13388] = 7.962867573956e-02;
C[13389] = -8.208798955407e-06;
C[13390] = 1.704169720960e-04;
C[13391] = 2.474882296866e-03;
C[13392] = 2.080695119980e+00;
C[13393] = 2.597745864375e-02;
C[13394] = 2.981464715613e-02;
C[13395] = -6.185398567569e+00;
C[13396] = -2.717662974067e-02;
C[13397] = 5.925825457062e-03;
C[13398] = 2.179698423039e+00;
C[13399] = -4.148216195956e-02;
C[13400] = -9.559885888228e-03;
C[13401] = 3.597381133161e-01;
C[13402] = 1.603493416080e-02;
C[13403] = 3.812563684402e-03;
C[13404] = 1.370994453638e-01;
C[13405] = -1.502304189086e-02;
C[13406] = 2.351925805034e-04;
C[13407] = 3.897710279703e-01;
C[13408] = -3.910766939952e-03;
C[13409] = -2.385133467753e-03;
C[13410] = 1.022218914248e+00;
C[13411] = -9.796588195561e-03;
C[13412] = -1.007401537230e-02;
C[13413] = 6.037556215569e-02;
C[13414] = -5.076125116418e-04;
C[13415] = -6.400280303586e-04;
C[13416] = 2.777871900947e-02;
C[13417] = 2.164190619162e+00;
C[13418] = 1.120229184511e-02;
C[13419] = -3.117706110610e-02;
C[13420] = -6.291682277669e+00;
C[13421] = -3.606742961813e-02;
C[13422] = -4.025745901582e-02;
C[13423] = 2.086340685367e+00;
C[13424] = -1.181734758712e-02;
C[13425] = 1.608300632346e-02;
C[13426] = 3.995695444994e-01;
C[13427] = 5.798867218061e-03;
C[13428] = -1.522306345913e-02;
C[13429] = 9.205344163550e-02;
C[13430] = -1.466864918426e-02;
C[13431] = -3.813049678624e-03;
C[13432] = 3.871958333444e-01;
C[13433] = 6.485375317656e-03;
C[13434] = -9.551999463964e-03;
C[13435] = 1.030433572912e+00;
C[13436] = 3.645073330046e-02;
C[13437] = -4.989433493438e-04;
C[13438] = 6.120518286613e-02;
C[13439] = 2.365708533657e-03;
C[13440] = 2.949048810569e-02;
C[13441] = 9.013882970107e-03;
C[13442] = 1.999513916358e+00;
C[13443] = 3.846188617497e-03;
C[13444] = -3.865391913515e-02;
C[13445] = -6.218319877097e+00;
C[13446] = -8.260987478193e-03;
C[13447] = -9.483150032143e-03;
C[13448] = 2.193904529396e+00;
C[13449] = 3.747079766959e-03;
C[13450] = 5.566481764534e-03;
C[13451] = 3.499414439046e-01;
C[13452] = 1.312107385251e-04;
C[13453] = -1.479797366011e-02;
C[13454] = 1.271087442560e-01;
C[13455] = -2.283729011485e-03;
C[13456] = 6.683965938541e-03;
C[13457] = 4.034140744103e-01;
C[13458] = -9.765945032343e-03;
C[13459] = 3.690561140209e-02;
C[13460] = 1.094041924324e+00;
C[13461] = -6.293836428791e-04;
C[13462] = 2.382854392140e-03;
C[13463] = 6.485328498777e-02;
C[13464] = 7.728752123193e-01;
C[13465] = 1.790268504784e-02;
C[13466] = 2.407123733495e-03;
C[13467] = 2.160676504600e+00;
C[13468] = -2.186444434484e-02;
C[13469] = -5.798061344875e-03;
C[13470] = -7.141534318823e+00;
C[13471] = 3.038276576202e-03;
C[13472] = 1.136842684541e-02;
C[13473] = 4.938554446527e-02;
C[13474] = 1.642731460566e-03;
C[13475] = 1.026797024720e-04;
C[13476] = 3.888199320483e-01;
C[13477] = -2.893398945075e-03;
C[13478] = -2.160582704819e-03;
C[13479] = 4.291157355057e-02;
C[13480] = 1.889799569987e-03;
C[13481] = 3.772864226578e-03;
C[13482] = 4.134544747261e+00;
C[13483] = -2.716866561980e-03;
C[13484] = -6.045241678472e-03;
C[13485] = 5.166570932172e-01;
C[13486] = -8.024635490195e-04;
C[13487] = -1.651820414345e-03;
C[13488] = 1.883675056127e-02;
C[13489] = 8.686330503764e-01;
C[13490] = 3.888319542362e-02;
C[13491] = -2.262641091344e-02;
C[13492] = 2.078405145144e+00;
C[13493] = -1.490966655099e-02;
C[13494] = 2.770921984069e-03;
C[13495] = -7.169574111036e+00;
C[13496] = -4.098108441812e-02;
C[13497] = 1.723789565918e-03;
C[13498] = 5.633594897991e-02;
C[13499] = 2.580809602911e-03;
C[13500] = -2.967979387646e-03;
C[13501] = 3.867990563332e-01;
C[13502] = 6.412640112599e-03;
C[13503] = 1.876431840380e-03;
C[13504] = 3.588355428630e-02;
C[13505] = -1.327594896722e-02;
C[13506] = -2.762342034762e-03;
C[13507] = 4.138078098195e+00;
C[13508] = 1.413136847265e-02;
C[13509] = -8.027364109004e-04;
C[13510] = 5.183285042742e-01;
C[13511] = 4.181513491139e-03;
C[13512] = 3.212462041599e-03;
C[13513] = 4.130988738248e-02;
C[13514] = 7.983388318499e-01;
C[13515] = -6.144297191254e-03;
C[13516] = -1.770907982445e-02;
C[13517] = 2.166202669913e+00;
C[13518] = 1.089728732766e-02;
C[13519] = -4.110632629047e-02;
C[13520] = -7.214611425750e+00;
C[13521] = 1.753877686139e-04;
C[13522] = 2.782719596859e-03;
C[13523] = 5.111409896342e-02;
C[13524] = -2.214354032904e-03;
C[13525] = 6.190788705789e-03;
C[13526] = 4.020289814361e-01;
C[13527] = 3.749307250690e-03;
C[13528] = -1.328221106084e-02;
C[13529] = 1.928387673970e-02;
C[13530] = -6.173670778512e-03;
C[13531] = 1.429129223813e-02;
C[13532] = 4.169476010700e+00;
C[13533] = -1.654826220276e-03;
C[13534] = 4.190489745984e-03;
C[13535] = 5.251119476653e-01;
C[13536] = 0.000000000000e+00;
C[13537] = 0.000000000000e+00;
C[13538] = 0.000000000000e+00;
C[13539] = 0.000000000000e+00;
C[13540] = 0.000000000000e+00;
C[13541] = 0.000000000000e+00;
C[13542] = 0.000000000000e+00;
C[13543] = 0.000000000000e+00;
C[13544] = 0.000000000000e+00;
C[13545] = 0.000000000000e+00;
C[13546] = 0.000000000000e+00;
C[13547] = 0.000000000000e+00;
C[13548] = 0.000000000000e+00;
C[13549] = 0.000000000000e+00;
C[13550] = 0.000000000000e+00;
C[13551] = 0.000000000000e+00;
C[13552] = 0.000000000000e+00;
C[13553] = 0.000000000000e+00;
C[13554] = 1.000000000000e+00;
C[13555] = 0.000000000000e+00;
C[13556] = 0.000000000000e+00;
C[13557] = 2.000000000000e-01;
C[13558] = -0.000000000000e+00;
C[13559] = -0.000000000000e+00;
C[13560] = 0.000000000000e+00;
C[13561] = 0.000000000000e+00;
C[13562] = 0.000000000000e+00;
C[13563] = 0.000000000000e+00;
C[13564] = 0.000000000000e+00;
C[13565] = 0.000000000000e+00;
C[13566] = 0.000000000000e+00;
C[13567] = 0.000000000000e+00;
C[13568] = 0.000000000000e+00;
C[13569] = 0.000000000000e+00;
C[13570] = 0.000000000000e+00;
C[13571] = 0.000000000000e+00;
C[13572] = 0.000000000000e+00;
C[13573] = 0.000000000000e+00;
C[13574] = 0.000000000000e+00;
C[13575] = 0.000000000000e+00;
C[13576] = 0.000000000000e+00;
C[13577] = 0.000000000000e+00;
C[13578] = 0.000000000000e+00;
C[13579] = 1.000000000000e+00;
C[13580] = 0.000000000000e+00;
C[13581] = -0.000000000000e+00;
C[13582] = 2.000000000000e-01;
C[13583] = -0.000000000000e+00;
C[13584] = 0.000000000000e+00;
C[13585] = 0.000000000000e+00;
C[13586] = 0.000000000000e+00;
C[13587] = 0.000000000000e+00;
C[13588] = 0.000000000000e+00;
C[13589] = 0.000000000000e+00;
C[13590] = 0.000000000000e+00;
C[13591] = 0.000000000000e+00;
C[13592] = 0.000000000000e+00;
C[13593] = 0.000000000000e+00;
C[13594] = 0.000000000000e+00;
C[13595] = 0.000000000000e+00;
C[13596] = 0.000000000000e+00;
C[13597] = 0.000000000000e+00;
C[13598] = 0.000000000000e+00;
C[13599] = 0.000000000000e+00;
C[13600] = 0.000000000000e+00;
C[13601] = 0.000000000000e+00;
C[13602] = 0.000000000000e+00;
C[13603] = 0.000000000000e+00;
C[13604] = 1.000000000000e+00;
C[13605] = -0.000000000000e+00;
C[13606] = -0.000000000000e+00;
C[13607] = 2.000000000000e-01;
C[13608] = 9.372684715392e-02;
C[13609] = -2.911740612151e-02;
C[13610] = 9.711119482655e-03;
C[13611] = 3.545831101010e-01;
C[13612] = 2.313834225850e-02;
C[13613] = 2.276993745584e-03;
C[13614] = 4.851032280317e-02;
C[13615] = 3.544226612063e-03;
C[13616] = 5.559547777378e-04;
C[13617] = 1.325090326774e-01;
C[13618] = -2.506489151996e-03;
C[13619] = 6.116570861821e-04;
C[13620] = 2.885004656098e-02;
C[13621] = 2.008408444968e-03;
C[13622] = 2.400980940801e-04;
C[13623] = 2.346975822465e-03;
C[13624] = 1.797357326488e-04;
C[13625] = 3.060693772087e-05;
C[13626] = 2.360250807929e-03;
C[13627] = 1.935818784194e-04;
C[13628] = 2.887071075580e-05;
C[13629] = -1.620746620672e-05;
C[13630] = 6.709119967130e-07;
C[13631] = 8.985588099476e-07;
C[13632] = -2.904529323284e-02;
C[13633] = 5.079978918859e-02;
C[13634] = 9.434151903619e-03;
C[13635] = 2.324788569379e-02;
C[13636] = 3.945374809058e-01;
C[13637] = 3.061957727513e-03;
C[13638] = 3.476115627884e-03;
C[13639] = 5.557599588531e-02;
C[13640] = 2.491786566169e-03;
C[13641] = -2.501681626084e-03;
C[13642] = 1.285268923733e-01;
C[13643] = 7.264484179345e-04;
C[13644] = 2.013089940557e-03;
C[13645] = 3.234813934595e-02;
C[13646] = 4.534105170535e-04;
C[13647] = 1.756961045978e-04;
C[13648] = 2.708093085598e-03;
C[13649] = 1.396663496959e-04;
C[13650] = 1.840605224427e-04;
C[13651] = 2.782901605856e-03;
C[13652] = 2.208865578724e-04;
C[13653] = 4.978943906217e-07;
C[13654] = -1.430830214399e-05;
C[13655] = 3.054026174345e-06;
C[13656] = 9.734855527411e-03;
C[13657] = 9.574491721098e-03;
C[13658] = 8.362630499518e-02;
C[13659] = 2.494190615258e-03;
C[13660] = 3.531204855978e-03;
C[13661] = 3.393638039236e-01;
C[13662] = 4.528848649233e-04;
C[13663] = 2.275861445367e-03;
C[13664] = 4.889501672251e-02;
C[13665] = 6.132394891658e-04;
C[13666] = 7.358044057665e-04;
C[13667] = 1.316701025721e-01;
C[13668] = 2.502699036966e-04;
C[13669] = 4.753315910794e-04;
C[13670] = 2.775254690519e-02;
C[13671] = 2.441263284665e-05;
C[13672] = 1.268909906645e-04;
C[13673] = 2.382636410861e-03;
C[13674] = 1.556740245801e-05;
C[13675] = 1.961926315035e-04;
C[13676] = 2.492207969701e-03;
C[13677] = 6.326885051317e-07;
C[13678] = 2.616497052598e-06;
C[13679] = -1.372743106975e-05;
C[13680] = 3.538958971743e-01;
C[13681] = 2.333524417486e-02;
C[13682] = 2.695208093755e-03;
C[13683] = 1.427508341892e-01;
C[13684] = -2.402412782067e-02;
C[13685] = -9.492596997909e-04;
C[13686] = 3.901339013889e-01;
C[13687] = -3.505224784198e-03;
C[13688] = -3.150421350572e-04;
C[13689] = 2.880423236586e-02;
C[13690] = 2.018913839295e-03;
C[13691] = 2.636710689297e-04;
C[13692] = 1.344843080192e-01;
C[13693] = -2.055572886046e-03;
C[13694] = -2.014945999379e-04;
C[13695] = 3.178050008826e-02;
C[13696] = -1.636963565809e-04;
C[13697] = -4.776040684323e-06;
C[13698] = 6.061448754110e-02;
C[13699] = -3.356081953670e-04;
C[13700] = -3.612123125512e-04;
C[13701] = 2.641174062910e-03;
C[13702] = -5.531956269226e-06;
C[13703] = -1.411066452067e-05;
C[13704] = 2.317643871125e-02;
C[13705] = 3.946497664806e-01;
C[13706] = 3.200043774873e-03;
C[13707] = -2.401622062581e-02;
C[13708] = 9.536802096569e-02;
C[13709] = -1.154287634850e-02;
C[13710] = -3.413453770685e-03;
C[13711] = 3.889220151525e-01;
C[13712] = 6.793564488713e-03;
C[13713] = 2.010948208484e-03;
C[13714] = 3.235562505094e-02;
C[13715] = 4.532541856724e-04;
C[13716] = -2.055045739723e-03;
C[13717] = 1.303731084396e-01;
C[13718] = -1.233236187589e-03;
C[13719] = -1.586305255331e-04;
C[13720] = 3.196014920530e-02;
C[13721] = 6.894724357711e-04;
C[13722] = -3.401705240073e-04;
C[13723] = 6.163263319077e-02;
C[13724] = 2.316439697121e-03;
C[13725] = -6.106179412740e-06;
C[13726] = 2.706359684900e-03;
C[13727] = 1.115942180704e-04;
C[13728] = 2.328967882402e-03;
C[13729] = 2.420643905132e-03;
C[13730] = 3.389309836582e-01;
C[13731] = -8.571210270965e-04;
C[13732] = -1.132475558527e-02;
C[13733] = 1.389758486877e-01;
C[13734] = -1.627422003722e-04;
C[13735] = 7.073370297894e-03;
C[13736] = 4.040646711451e-01;
C[13737] = 2.435630365346e-04;
C[13738] = 4.106562622281e-04;
C[13739] = 2.772369222082e-02;
C[13740] = -1.953520217583e-04;
C[13741] = -1.218694803374e-03;
C[13742] = 1.341021034556e-01;
C[13743] = 3.490005712624e-06;
C[13744] = 7.044707513970e-04;
C[13745] = 3.301581470092e-02;
C[13746] = -3.697192014321e-04;
C[13747] = 2.326017525862e-03;
C[13748] = 6.469142192770e-02;
C[13749] = -1.504272239510e-05;
C[13750] = 1.110465380627e-04;
C[13751] = 2.827060274815e-03;
C[13752] = 4.788796416864e-02;
C[13753] = 3.940795211145e-03;
C[13754] = 3.140285405984e-04;
C[13755] = 3.910134588713e-01;
C[13756] = -3.818038955288e-03;
C[13757] = 1.104161603901e-04;
C[13758] = 4.295352332621e-02;
C[13759] = 1.016475208209e-03;
C[13760] = 2.606122898894e-03;
C[13761] = 2.305485246829e-03;
C[13762] = 2.066747434819e-04;
C[13763] = 1.515554455832e-05;
C[13764] = 3.183913725376e-02;
C[13765] = -1.856028711734e-04;
C[13766] = 2.170056309678e-05;
C[13767] = 1.280981210065e-01;
C[13768] = 8.582529021684e-05;
C[13769] = 2.063478806401e-04;
C[13770] = 5.159347618255e-01;
C[13771] = -1.398827462442e-03;
C[13772] = -2.969880531002e-03;
C[13773] = 3.777316924948e-02;
C[13774] = -1.082691759879e-04;
C[13775] = -2.426886518642e-04;
C[13776] = 4.000331533507e-03;
C[13777] = 5.567820438735e-02;
C[13778] = 1.665075564130e-03;
C[13779] = -3.896386404278e-03;
C[13780] = 3.892421631552e-01;
C[13781] = 7.555962226824e-03;
C[13782] = 1.012895060260e-03;
C[13783] = 3.550338983252e-02;
C[13784] = -1.339496618023e-02;
C[13785] = 2.101427274118e-04;
C[13786] = 2.714906985735e-03;
C[13787] = 8.617193191533e-05;
C[13788] = -1.897737979195e-04;
C[13789] = 3.198149240547e-02;
C[13790] = 7.366435466590e-04;
C[13791] = 8.558661368692e-05;
C[13792] = 1.274466478575e-01;
C[13793] = -1.201918075701e-03;
C[13794] = -1.385692224941e-03;
C[13795] = 5.169252056248e-01;
C[13796] = 4.144658392819e-03;
C[13797] = -1.076138154120e-04;
C[13798] = 3.787225795907e-02;
C[13799] = 3.756359131136e-04;
C[13800] = 4.109068720313e-04;
C[13801] = 1.872211782832e-03;
C[13802] = 4.777326105988e-02;
C[13803] = -1.593048446174e-05;
C[13804] = 7.307813560887e-03;
C[13805] = 4.050714848198e-01;
C[13806] = 2.599938274674e-03;
C[13807] = -1.339556793824e-02;
C[13808] = 1.978098035878e-02;
C[13809] = 2.093707734043e-05;
C[13810] = 9.836136414003e-05;
C[13811] = 2.307852700019e-03;
C[13812] = 1.516473602204e-05;
C[13813] = 7.237557072493e-04;
C[13814] = 3.308293561256e-02;
C[13815] = 2.059355723588e-04;
C[13816] = -1.201958192902e-03;
C[13817] = 1.260133403299e-01;
C[13818] = -2.948299999895e-03;
C[13819] = 4.156456807257e-03;
C[13820] = 5.250410539755e-01;
C[13821] = -2.415948485310e-04;
C[13822] = 3.762310932513e-04;
C[13823] = 3.861155675533e-02;
C[13824] = -6.956248561268e+00;
C[13825] = -8.289511894556e-02;
C[13826] = 5.063200886026e-02;
C[13827] = 2.072318207618e+00;
C[13828] = 4.277245346414e-02;
C[13829] = 2.142886197140e-02;
C[13830] = 7.598823243598e-01;
C[13831] = 4.765521288740e-02;
C[13832] = 1.016163967970e-02;
C[13833] = 9.270390230917e-02;
C[13834] = -2.651468055189e-02;
C[13835] = 5.905401513256e-03;
C[13836] = 3.603432973175e-01;
C[13837] = 1.989402235604e-02;
C[13838] = 3.970883646384e-03;
C[13839] = 4.888056239193e-02;
C[13840] = 3.379537207222e-03;
C[13841] = 6.852773200905e-04;
C[13842] = 7.604213189907e-02;
C[13843] = 5.595614937407e-03;
C[13844] = 7.549237017848e-04;
C[13845] = 2.347436672196e-03;
C[13846] = 2.092339658194e-04;
C[13847] = 2.849463474367e-05;
C[13848] = -8.145286117216e-02;
C[13849] = -7.073272622025e+00;
C[13850] = 6.510455490280e-02;
C[13851] = 4.390975529570e-02;
C[13852] = 2.139623567774e+00;
C[13853] = 2.328852394030e-02;
C[13854] = 4.683342074460e-02;
C[13855] = 8.562595213173e-01;
C[13856] = 3.934408348246e-02;
C[13857] = -2.644256766322e-02;
C[13858] = 4.818655014202e-02;
C[13859] = 9.481900054419e-03;
C[13860] = 1.997034963076e-02;
C[13861] = 3.951043057230e-01;
C[13862] = 7.345049602115e-03;
C[13863] = 3.314927708133e-03;
C[13864] = 5.570390341516e-02;
C[13865] = 2.717954398799e-03;
C[13866] = 5.370414848786e-03;
C[13867] = 8.818821098063e-02;
C[13868] = 5.734524798182e-03;
C[13869] = 2.003802819258e-04;
C[13870] = 2.795699794913e-03;
C[13871] = 1.991596430403e-04;
C[13872] = 5.110672975538e-02;
C[13873] = 6.791135125238e-02;
C[13874] = -7.060942658710e+00;
C[13875] = 2.422085425028e-02;
C[13876] = 2.929969209421e-02;
C[13877] = 1.979672841983e+00;
C[13878] = 9.030422161641e-03;
C[13879] = 3.698637871256e-02;
C[13880] = 7.648762003230e-01;
C[13881] = 5.929137558012e-03;
C[13882] = 9.622239871898e-03;
C[13883] = 7.904623586715e-02;
C[13884] = 4.192936901525e-03;
C[13885] = 7.818352744856e-03;
C[13886] = 3.475210502238e-01;
C[13887] = 5.911104946775e-04;
C[13888] = 2.524121655014e-03;
C[13889] = 4.937553962870e-02;
C[13890] = 4.522765262505e-04;
C[13891] = 5.168124475106e-03;
C[13892] = 7.905134468312e-02;
C[13893] = 1.601824546325e-05;
C[13894] = 1.761068079441e-04;
C[13895] = 2.456538243622e-03;
C[13896] = 2.058573949084e+00;
C[13897] = 4.565692491716e-02;
C[13898] = 2.824120382021e-02;
C[13899] = -6.160403163509e+00;
C[13900] = -6.095553680390e-02;
C[13901] = -3.349146430592e-03;
C[13902] = 2.176633402386e+00;
C[13903] = -4.631270607729e-02;
C[13904] = -3.152566260289e-03;
C[13905] = 3.596560843908e-01;
C[13906] = 2.005770811183e-02;
C[13907] = 4.393954380022e-03;
C[13908] = 1.362417819608e-01;
C[13909] = -2.114839662223e-02;
C[13910] = -3.273195836776e-03;
C[13911] = 3.908781796962e-01;
C[13912] = -2.986603292599e-03;
C[13913] = 1.090541208833e-04;
C[13914] = 1.027039087614e+00;
C[13915] = -7.793144545494e-03;
C[13916] = -5.912876371311e-03;
C[13917] = 6.060087327839e-02;
C[13918] = -3.299693125633e-04;
C[13919] = -3.465666800422e-04;
C[13920] = 4.353438251905e-02;
C[13921] = 2.141869279271e+00;
C[13922] = 2.267647047212e-02;
C[13923] = -6.079739290675e-02;
C[13924] = -6.279854555231e+00;
C[13925] = -3.397027937635e-02;
C[13926] = -4.505549546760e-02;
C[13927] = 2.096667558464e+00;
C[13928] = -1.260017245353e-02;
C[13929] = 1.993211880879e-02;
C[13930] = 3.952165912979e-01;
C[13931] = 7.487191663751e-03;
C[13932] = -2.114048942737e-02;
C[13933] = 9.343717747692e-02;
C[13934] = -1.531382577396e-02;
C[13935] = -2.897718166019e-03;
C[13936] = 3.901655221852e-01;
C[13937] = 6.327289087198e-03;
C[13938] = -7.788378665254e-03;
C[13939] = 1.038373475937e+00;
C[13940] = 3.563016091615e-02;
C[13941] = -3.367292443422e-04;
C[13942] = 6.159543898428e-02;
C[13943] = 2.331219681671e-03;
C[13944] = 2.246834470776e-02;
C[13945] = 1.046224749268e-02;
C[13946] = 1.971016436673e+00;
C[13947] = -1.506372976705e-03;
C[13948] = -2.960786411180e-02;
C[13949] = -6.176261202196e+00;
C[13950] = -1.059568508081e-03;
C[13951] = -8.894645499321e-03;
C[13952] = 2.201018606086e+00;
C[13953] = 4.022857783202e-03;
C[13954] = 6.703735779734e-03;
C[13955] = 3.470882299583e-01;
C[13956] = -3.181057164082e-03;
C[13957] = -1.509570501073e-02;
C[13958] = 1.313154736122e-01;
C[13959] = 2.529572010067e-04;
C[13960] = 6.583932339735e-03;
C[13961] = 4.033218011705e-01;
C[13962] = -5.941667942616e-03;
C[13963] = 3.596683358106e-02;
C[13964] = 1.092005242277e+00;
C[13965] = -3.588051269850e-04;
C[13966] = 2.333261318218e-03;
C[13967] = 6.471497220703e-02;
C[13968] = 7.474351516690e-01;
C[13969] = 5.612701240981e-02;
C[13970] = 6.253295675143e-03;
C[13971] = 2.194224552034e+00;
C[13972] = -5.314719915967e-02;
C[13973] = 4.403598707166e-03;
C[13974] = -7.136914912517e+00;
C[13975] = 4.993414307271e-04;
C[13976] = 1.329723330848e-02;
C[13977] = 4.825820375740e-02;
C[13978] = 3.779607291394e-03;
C[13979] = 4.522541703526e-04;
C[13980] = 3.917577371786e-01;
C[13981] = -3.302303350622e-03;
C[13982] = 5.261155617691e-04;
C[13983] = 4.321456930169e-02;
C[13984] = 7.321266421936e-04;
C[13985] = 2.405255393066e-03;
C[13986] = 4.122922953212e+00;
C[13987] = -1.031196597654e-02;
C[13988] = -2.244010811243e-02;
C[13989] = 5.145066517683e-01;
C[13990] = -1.436698517783e-03;
C[13991] = -3.336427829518e-03;
C[13992] = 5.677731131629e-02;
C[13993] = 8.583036913582e-01;
C[13994] = 2.477066108782e-02;
C[13995] = -5.413593847888e-02;
C[13996] = 2.103070518518e+00;
C[13997] = 7.571930792760e-04;
C[13998] = 4.277384717514e-04;
C[13999] = -7.163985084014e+00;
C[14000] = -4.070002094800e-02;
C[14001] = 3.835642128666e-03;
C[14002] = 5.580611191720e-02;
C[14003] = 1.913335773777e-03;
C[14004] = -3.377764912679e-03;
C[14005] = 3.904856701879e-01;
C[14006] = 7.066524268665e-03;
C[14007] = 7.285464942448e-04;
C[14008] = 3.583056171755e-02;
C[14009] = -1.335693064967e-02;
C[14010] = -1.009535356377e-02;
C[14011] = 4.118540480686e+00;
C[14012] = 1.420922507513e-02;
C[14013] = -1.421885911873e-03;
C[14014] = 5.152282489313e-01;
C[14015] = 4.298790518105e-03;
C[14016] = 7.260681565567e-03;
C[14017] = 2.695258781571e-02;
C[14018] = 7.424410870705e-01;
C[14019] = 2.829666751620e-03;
C[14020] = -2.315191010052e-03;
C[14021] = 2.221154879579e+00;
C[14022] = 1.317354082406e-02;
C[14023] = -4.071205610816e-02;
C[14024] = -7.209790022136e+00;
C[14025] = 5.402294143840e-04;
C[14026] = 2.098379615462e-03;
C[14027] = 4.825378396608e-02;
C[14028] = 4.081657714788e-04;
C[14029] = 6.841538159372e-03;
C[14030] = 4.043286148451e-01;
C[14031] = 2.399070768846e-03;
C[14032] = -1.335753240767e-02;
C[14033] = 1.983315809707e-02;
C[14034] = -2.208609176398e-02;
C[14035] = 1.442050727306e-02;
C[14036] = 4.169846830319e+00;
C[14037] = -3.312442005886e-03;
C[14038] = 4.315438786660e-03;
C[14039] = 5.251652421459e-01;
C[14040] = 0.000000000000e+00;
C[14041] = 0.000000000000e+00;
C[14042] = 0.000000000000e+00;
C[14043] = 0.000000000000e+00;
C[14044] = 0.000000000000e+00;
C[14045] = 0.000000000000e+00;
C[14046] = 0.000000000000e+00;
C[14047] = 0.000000000000e+00;
C[14048] = 0.000000000000e+00;
C[14049] = 0.000000000000e+00;
C[14050] = 0.000000000000e+00;
C[14051] = 0.000000000000e+00;
C[14052] = 0.000000000000e+00;
C[14053] = 0.000000000000e+00;
C[14054] = 0.000000000000e+00;
C[14055] = 0.000000000000e+00;
C[14056] = 0.000000000000e+00;
C[14057] = 0.000000000000e+00;
C[14058] = 1.000000000000e+00;
C[14059] = 0.000000000000e+00;
C[14060] = 0.000000000000e+00;
C[14061] = 2.000000000000e-01;
C[14062] = -0.000000000000e+00;
C[14063] = -0.000000000000e+00;
C[14064] = 0.000000000000e+00;
C[14065] = 0.000000000000e+00;
C[14066] = 0.000000000000e+00;
C[14067] = 0.000000000000e+00;
C[14068] = 0.000000000000e+00;
C[14069] = 0.000000000000e+00;
C[14070] = 0.000000000000e+00;
C[14071] = 0.000000000000e+00;
C[14072] = 0.000000000000e+00;
C[14073] = 0.000000000000e+00;
C[14074] = 0.000000000000e+00;
C[14075] = 0.000000000000e+00;
C[14076] = 0.000000000000e+00;
C[14077] = 0.000000000000e+00;
C[14078] = 0.000000000000e+00;
C[14079] = 0.000000000000e+00;
C[14080] = 0.000000000000e+00;
C[14081] = 0.000000000000e+00;
C[14082] = 0.000000000000e+00;
C[14083] = 1.000000000000e+00;
C[14084] = 0.000000000000e+00;
C[14085] = -0.000000000000e+00;
C[14086] = 2.000000000000e-01;
C[14087] = -0.000000000000e+00;
C[14088] = 0.000000000000e+00;
C[14089] = 0.000000000000e+00;
C[14090] = 0.000000000000e+00;
C[14091] = 0.000000000000e+00;
C[14092] = 0.000000000000e+00;
C[14093] = 0.000000000000e+00;
C[14094] = 0.000000000000e+00;
C[14095] = 0.000000000000e+00;
C[14096] = 0.000000000000e+00;
C[14097] = 0.000000000000e+00;
C[14098] = 0.000000000000e+00;
C[14099] = 0.000000000000e+00;
C[14100] = 0.000000000000e+00;
C[14101] = 0.000000000000e+00;
C[14102] = 0.000000000000e+00;
C[14103] = 0.000000000000e+00;
C[14104] = 0.000000000000e+00;
C[14105] = 0.000000000000e+00;
C[14106] = 0.000000000000e+00;
C[14107] = 0.000000000000e+00;
C[14108] = 1.000000000000e+00;
C[14109] = -0.000000000000e+00;
C[14110] = -0.000000000000e+00;
C[14111] = 2.000000000000e-01;
C[14112] = 8.685376982899e-02;
C[14113] = -1.364080669610e-02;
C[14114] = 5.692677908075e-04;
C[14115] = 3.731272453139e-01;
C[14116] = 1.005720415574e-02;
C[14117] = 3.528332146067e-03;
C[14118] = 5.181568308591e-02;
C[14119] = 8.691342769930e-04;
C[14120] = 3.187577915091e-04;
C[14121] = 1.315439189058e-01;
C[14122] = -7.509035284345e-04;
C[14123] = -5.514189410596e-05;
C[14124] = 3.058495118483e-02;
C[14125] = 6.145605438596e-04;
C[14126] = 2.192388010808e-04;
C[14127] = 2.518017524847e-03;
C[14128] = 2.837335287635e-05;
C[14129] = 9.928892457566e-06;
C[14130] = 2.534604854098e-03;
C[14131] = -9.744668943964e-06;
C[14132] = -2.569575045873e-05;
C[14133] = -1.385370236652e-05;
C[14134] = -3.561702017460e-06;
C[14135] = -2.464742678787e-06;
C[14136] = -1.340210484775e-02;
C[14137] = 4.012325784456e-02;
C[14138] = 9.770967504623e-03;
C[14139] = 9.913087290620e-03;
C[14140] = 3.973145144224e-01;
C[14141] = 1.176476622261e-02;
C[14142] = 8.254745708345e-04;
C[14143] = 5.652925465628e-02;
C[14144] = 3.361454507351e-03;
C[14145] = -7.349900718778e-04;
C[14146] = 1.274338147547e-01;
C[14147] = 6.978937008907e-04;
C[14148] = 6.047622141008e-04;
C[14149] = 3.267957422854e-02;
C[14150] = 1.026378780221e-03;
C[14151] = 2.654023003360e-05;
C[14152] = 2.757209836311e-03;
C[14153] = 1.737772075617e-04;
C[14154] = -1.341705579984e-05;
C[14155] = 2.811786543379e-03;
C[14156] = 2.439991733142e-04;
C[14157] = -3.576051404873e-06;
C[14158] = -1.287274878797e-05;
C[14159] = 1.215189435686e-06;
C[14160] = 5.573329250140e-04;
C[14161] = 9.821475985925e-03;
C[14162] = 6.471312956171e-02;
C[14163] = 3.569945923622e-03;
C[14164] = 1.210865884179e-02;
C[14165] = 3.691314925189e-01;
C[14166] = 2.783425474758e-04;
C[14167] = 3.253419540673e-03;
C[14168] = 5.422387208763e-02;
C[14169] = -5.593755182552e-05;
C[14170] = 7.012609329774e-04;
C[14171] = 1.297355116514e-01;
C[14172] = 2.196310416213e-04;
C[14173] = 1.042733894221e-03;
C[14174] = 3.060330255711e-02;
C[14175] = 8.021031321663e-06;
C[14176] = 1.672687735037e-04;
C[14177] = 2.669702224185e-03;
C[14178] = -2.945335411953e-05;
C[14179] = 2.282749402178e-04;
C[14180] = 2.857839765412e-03;
C[14181] = -2.428780652145e-06;
C[14182] = 1.141604032220e-06;
C[14183] = -9.597222533836e-06;
C[14184] = 3.730913123372e-01;
C[14185] = 9.800160359000e-03;
C[14186] = 4.210655284013e-03;
C[14187] = 1.242476118501e-01;
C[14188] = -9.954836561168e-03;
C[14189] = -2.928532197101e-03;
C[14190] = 3.886355997335e-01;
C[14191] = -9.908563134998e-04;
C[14192] = -3.712359615680e-04;
C[14193] = 3.058255565305e-02;
C[14194] = 5.972337519928e-04;
C[14195] = 2.623449989807e-04;
C[14196] = 1.328293758572e-01;
C[14197] = -5.574460058543e-04;
C[14198] = -1.355588602134e-04;
C[14199] = 3.161035287984e-02;
C[14200] = -7.481740382935e-05;
C[14201] = -7.980120950835e-05;
C[14202] = 5.954489206322e-02;
C[14203] = -2.660382742101e-04;
C[14204] = -5.940061314981e-04;
C[14205] = 2.590672621265e-03;
C[14206] = -1.102053205106e-05;
C[14207] = -2.913471376686e-05;
C[14208] = 9.620226017332e-03;
C[14209] = 3.982585382966e-01;
C[14210] = 1.214141562580e-02;
C[14211] = -9.798739518983e-03;
C[14212] = 9.002419352025e-02;
C[14213] = -1.908334002804e-02;
C[14214] = -9.475791051100e-04;
C[14215] = 3.900176893582e-01;
C[14216] = 5.693036719099e-03;
C[14217] = 5.854286679659e-04;
C[14218] = 3.274250915349e-02;
C[14219] = 1.044917679821e-03;
C[14220] = -5.470395363753e-04;
C[14221] = 1.299960232460e-01;
C[14222] = -1.787260387307e-03;
C[14223] = -7.331109039354e-05;
C[14224] = 3.192373716482e-02;
C[14225] = 6.693598127566e-04;
C[14226] = -2.724460939238e-04;
C[14227] = 6.064830797425e-02;
C[14228] = 2.424476616083e-03;
C[14229] = -1.137084706196e-05;
C[14230] = 2.653581922674e-03;
C[14231] = 1.214952580008e-04;
C[14232] = 4.023945515168e-03;
C[14233] = 1.168798052523e-02;
C[14234] = 3.683099547970e-01;
C[14235] = -2.808687611584e-03;
C[14236] = -1.893791288357e-02;
C[14237] = 1.108943474656e-01;
C[14238] = -3.258238169442e-04;
C[14239] = 5.862943386629e-03;
C[14240] = 3.984022403333e-01;
C[14241] = 2.522796923542e-04;
C[14242] = 1.021259733728e-03;
C[14243] = 3.054853337565e-02;
C[14244] = -1.275692211789e-04;
C[14245] = -1.777565244342e-03;
C[14246] = 1.313577234701e-01;
C[14247] = -7.854291895378e-05;
C[14248] = 6.780840185423e-04;
C[14249] = 3.270284333958e-02;
C[14250] = -6.010883080164e-04;
C[14251] = 2.400904746290e-03;
C[14252] = 6.440529400222e-02;
C[14253] = -2.939103216780e-05;
C[14254] = 1.194344555531e-04;
C[14255] = 2.826242746283e-03;
C[14256] = 5.061100381047e-02;
C[14257] = 1.574864407793e-03;
C[14258] = 4.260053240262e-04;
C[14259] = 3.899757164916e-01;
C[14260] = -1.930685153133e-03;
C[14261] = -8.563073242806e-04;
C[14262] = 4.661515580813e-02;
C[14263] = 8.602576263435e-04;
C[14264] = 2.396328086686e-03;
C[14265] = 2.437705573150e-03;
C[14266] = 7.649955249751e-05;
C[14267] = 1.786521642502e-05;
C[14268] = 3.169969399705e-02;
C[14269] = -1.388514935951e-04;
C[14270] = -1.139084861095e-04;
C[14271] = 1.285859065880e-01;
C[14272] = 8.256342683485e-05;
C[14273] = 1.682855677171e-04;
C[14274] = 5.105387359170e-01;
C[14275] = -6.158437895806e-04;
C[14276] = -1.932523540107e-03;
C[14277] = 3.729447674455e-02;
C[14278] = -2.148111730391e-05;
C[14279] = -6.964916514880e-05;
C[14280] = 1.631968593352e-03;
C[14281] = 5.625500355449e-02;
C[14282] = 3.299990607631e-03;
C[14283] = -2.003750381053e-03;
C[14284] = 3.905686725800e-01;
C[14285] = 5.945711416382e-03;
C[14286] = 8.609299032882e-04;
C[14287] = 3.936142312246e-02;
C[14288] = -1.363423417090e-02;
C[14289] = 7.922897396694e-05;
C[14290] = 2.738926429525e-03;
C[14291] = 1.703735113010e-04;
C[14292] = -1.423436749996e-04;
C[14293] = 3.196046937961e-02;
C[14294] = 6.836018871925e-04;
C[14295] = 8.260824529783e-05;
C[14296] = 1.279542787317e-01;
C[14297] = -1.229309526248e-03;
C[14298] = -6.029457828676e-04;
C[14299] = 5.111335680613e-01;
C[14300] = 4.337371984795e-03;
C[14301] = -2.093105289559e-05;
C[14302] = 3.736528978095e-02;
C[14303] = 3.730705287591e-04;
C[14304] = 4.900500425373e-04;
C[14305] = 3.440034179528e-03;
C[14306] = 5.251839510125e-02;
C[14307] = -9.465087185755e-04;
C[14308] = 5.727130998418e-03;
C[14309] = 3.998731218153e-01;
C[14310] = 2.390862659796e-03;
C[14311] = -1.360596329638e-02;
C[14312] = 1.888435903413e-02;
C[14313] = 2.134837585945e-05;
C[14314] = 1.790158523735e-04;
C[14315] = 2.556003758426e-03;
C[14316] = -1.181527266422e-04;
C[14317] = 6.716327647112e-04;
C[14318] = 3.280090210505e-02;
C[14319] = 1.679212059244e-04;
C[14320] = -1.227424801281e-03;
C[14321] = 1.259308745883e-01;
C[14322] = -1.908601971076e-03;
C[14323] = 4.368839883543e-03;
C[14324] = 5.261523308854e-01;
C[14325] = -6.859957335974e-05;
C[14326] = 3.745392439264e-04;
C[14327] = 3.872382946739e-02;
C[14328] = -6.960156355790e+00;
C[14329] = -2.409662943485e-02;
C[14330] = 6.366676002777e-03;
C[14331] = 2.143476041657e+00;
C[14332] = 1.623119164865e-02;
C[14333] = 5.636051661078e-03;
C[14334] = 8.107667079086e-01;
C[14335] = 6.559144228485e-03;
C[14336] = 2.917123198754e-03;
C[14337] = 8.001418456748e-02;
C[14338] = -2.904138552129e-03;
C[14339] = -1.251638894673e-03;
C[14340] = 3.814582035009e-01;
C[14341] = 3.170426785229e-03;
C[14342] = 1.397385109322e-03;
C[14343] = 5.219422151311e-02;
C[14344] = 3.898871678859e-04;
C[14345] = 1.313923545686e-04;
C[14346] = 8.041416162423e-02;
C[14347] = 3.557221536520e-04;
C[14348] = -4.392903465943e-04;
C[14349] = 2.492541542971e-03;
C[14350] = 4.880228296862e-06;
C[14351] = -2.748397915644e-05;
C[14352] = -1.932259246784e-02;
C[14353] = -7.132590916465e+00;
C[14354] = 8.983329619428e-02;
C[14355] = 1.298050526368e-02;
C[14356] = 2.128385115264e+00;
C[14357] = 4.333348172315e-02;
C[14358] = 6.142245221579e-03;
C[14359] = 8.697992008697e-01;
C[14360] = 4.873193892983e-02;
C[14361] = -2.665436703778e-03;
C[14362] = 3.304697615089e-02;
C[14363] = 1.019262908658e-02;
C[14364] = 3.009881698435e-03;
C[14365] = 3.984839857908e-01;
C[14366] = 1.162611293088e-02;
C[14367] = 3.646096569289e-04;
C[14368] = 5.651013842732e-02;
C[14369] = 3.389078796598e-03;
C[14370] = 2.611154954662e-04;
C[14371] = 8.833419260315e-02;
C[14372] = 7.066827468867e-03;
C[14373] = 1.561261148279e-06;
C[14374] = 2.797635691210e-03;
C[14375] = 2.498323119425e-04;
C[14376] = 6.127978686909e-03;
C[14377] = 9.084346582031e-02;
C[14378] = -7.078344289879e+00;
C[14379] = 5.586806837622e-03;
C[14380] = 4.772941462921e-02;
C[14381] = 2.086535256808e+00;
C[14382] = 2.655657918952e-03;
C[14383] = 4.783817519725e-02;
C[14384] = 8.454323772825e-01;
C[14385] = -1.263573760466e-03;
C[14386] = 1.024313756788e-02;
C[14387] = 5.810748414194e-02;
C[14388] = 1.408698484776e-03;
C[14389] = 1.193889894591e-02;
C[14390] = 3.817630631699e-01;
C[14391] = 1.083428657377e-04;
C[14392] = 3.296691298205e-03;
C[14393] = 5.496383855276e-02;
C[14394] = -5.408524620823e-04;
C[14395] = 6.646407174664e-03;
C[14396] = 8.883075173564e-02;
C[14397] = -3.145188527411e-05;
C[14398] = 2.324358247507e-04;
C[14399] = 2.804126584021e-03;
C[14400] = 2.142757382124e+00;
C[14401] = 1.072196663128e-02;
C[14402] = 1.840099404543e-02;
C[14403] = -6.217853481112e+00;
C[14404] = -1.575517051704e-02;
C[14405] = -3.221330043045e-03;
C[14406] = 2.139870933935e+00;
C[14407] = -5.897698542254e-03;
C[14408] = 2.593953428435e-04;
C[14409] = 3.814222705242e-01;
C[14410] = 2.896954766815e-03;
C[14411] = 2.049407845166e-03;
C[14412] = 1.178163398862e-01;
C[14413] = -2.192395909659e-03;
C[14414] = 4.254785180328e-05;
C[14415] = 3.866762257118e-01;
C[14416] = -1.236303021755e-03;
C[14417] = -1.399252791412e-03;
C[14418] = 1.005180154878e+00;
C[14419] = -5.242556950432e-03;
C[14420] = -9.974729802764e-03;
C[14421] = 5.929039658341e-02;
C[14422] = -2.789868177193e-04;
C[14423] = -5.228273981126e-04;
C[14424] = 7.491628880545e-03;
C[14425] = 2.147265592747e+00;
C[14426] = 4.838455030946e-02;
C[14427] = -1.263322967334e-02;
C[14428] = -6.292937207053e+00;
C[14429] = -4.243553369424e-02;
C[14430] = -5.584414814158e-03;
C[14431] = 2.095532548946e+00;
C[14432] = -2.173944945226e-02;
C[14433] = 2.733448646824e-03;
C[14434] = 3.994280096650e-01;
C[14435] = 1.197165572992e-02;
C[14436] = -2.036298867474e-03;
C[14437] = 9.036712426269e-02;
C[14438] = -1.944242758288e-02;
C[14439] = -1.218639039712e-03;
C[14440] = 3.883428849419e-01;
C[14441] = 6.256227665656e-03;
C[14442] = -5.260944247148e-03;
C[14443] = 1.019271178896e+00;
C[14444] = 3.705310190938e-02;
C[14445] = -2.780844546056e-04;
C[14446] = 6.034741133234e-02;
C[14447] = 2.433793918308e-03;
C[14448] = 1.554831904310e-02;
C[14449] = 4.179776777541e-02;
C[14450] = 2.070104502370e+00;
C[14451] = -8.244383327030e-04;
C[14452] = -3.952699080476e-02;
C[14453] = -6.261574825344e+00;
C[14454] = 4.634043759347e-04;
C[14455] = -1.974619337719e-02;
C[14456] = 2.119864967004e+00;
C[14457] = 1.892998478423e-03;
C[14458] = 1.154932723349e-02;
C[14459] = 3.809415254480e-01;
C[14460] = 1.623924373204e-04;
C[14461] = -1.929700043841e-02;
C[14462] = 9.882180024600e-02;
C[14463] = -1.390968772848e-03;
C[14464] = 6.409913435655e-03;
C[14465] = 3.974692514193e-01;
C[14466] = -9.946328162659e-03;
C[14467] = 3.684346555421e-02;
C[14468] = 1.084005369759e+00;
C[14469] = -5.148719248614e-04;
C[14470] = 2.402230162092e-03;
C[14471] = 6.449655915637e-02;
C[14472] = 7.866731223997e-01;
C[14473] = 2.113004196075e-02;
C[14474] = 5.608913449959e-03;
C[14475] = 2.166673269098e+00;
C[14476] = -2.524653577462e-02;
C[14477] = -1.014626577079e-02;
C[14478] = -7.110563509243e+00;
C[14479] = 1.199219036524e-03;
C[14480] = 6.327811671444e-03;
C[14481] = 5.098954223767e-02;
C[14482] = 1.113999493888e-03;
C[14483] = 2.560056422881e-04;
C[14484] = 3.880163424699e-01;
C[14485] = -2.201745087735e-03;
C[14486] = -1.921452280184e-03;
C[14487] = 5.106540866817e-02;
C[14488] = 8.711829384368e-04;
C[14489] = 1.389415570122e-03;
C[14490] = 4.082812732470e+00;
C[14491] = 7.487318557932e-04;
C[14492] = -9.316046500439e-04;
C[14493] = 5.075897756484e-01;
C[14494] = 1.657210870801e-04;
C[14495] = 3.254965480034e-04;
C[14496] = 2.181583055566e-02;
C[14497] = 8.643141788339e-01;
C[14498] = 4.876959653642e-02;
C[14499] = -2.615557989332e-02;
C[14500] = 2.106552213382e+00;
C[14501] = -1.809083278213e-02;
C[14502] = 1.212664575418e-03;
C[14503] = -7.133101153645e+00;
C[14504] = -4.387852881965e-02;
C[14505] = 1.152721484245e-03;
C[14506] = 5.623588732553e-02;
C[14507] = 3.343262365163e-03;
C[14508] = -2.249197089308e-03;
C[14509] = 3.888938681637e-01;
C[14510] = 6.492681465408e-03;
C[14511] = 8.718552153815e-04;
C[14512] = 4.418407460176e-02;
C[14513] = -1.386776798388e-02;
C[14514] = 9.210352478359e-04;
C[14515] = 4.076536468515e+00;
C[14516] = 1.224811890974e-02;
C[14517] = 1.711122252521e-04;
C[14518] = 5.079942528014e-01;
C[14519] = 3.981131401032e-03;
C[14520] = 6.342968219319e-03;
C[14521] = 5.030353237337e-02;
C[14522] = 8.113228375549e-01;
C[14523] = -1.124605979731e-02;
C[14524] = -2.105756386588e-02;
C[14525] = 2.149282596645e+00;
C[14526] = 6.218503133640e-03;
C[14527] = -4.331311132939e-02;
C[14528] = -7.223932267412e+00;
C[14529] = 3.026846055968e-04;
C[14530] = 3.467658468775e-03;
C[14531] = 5.325836156638e-02;
C[14532] = -1.974525548419e-03;
C[14533] = 6.290321944975e-03;
C[14534] = 3.989401329013e-01;
C[14535] = 1.383950143231e-03;
C[14536] = -1.383949710936e-02;
C[14537] = 1.813764823709e-02;
C[14538] = -5.879483097859e-04;
C[14539] = 1.274494833352e-02;
C[14540] = 4.180103975517e+00;
C[14541] = 3.341779215696e-04;
C[14542] = 4.020750959400e-03;
C[14543] = 5.269607954231e-01;
C[14544] = 0.000000000000e+00;
C[14545] = 0.000000000000e+00;
C[14546] = 0.000000000000e+00;
C[14547] = 0.000000000000e+00;
C[14548] = 0.000000000000e+00;
C[14549] = 0.000000000000e+00;
C[14550] = 0.000000000000e+00;
C[14551] = 0.000000000000e+00;
C[14552] = 0.000000000000e+00;
C[14553] = 0.000000000000e+00;
C[14554] = 0.000000000000e+00;
C[14555] = 0.000000000000e+00;
C[14556] = 0.000000000000e+00;
C[14557] = 0.000000000000e+00;
C[14558] = 0.000000000000e+00;
C[14559] = 0.000000000000e+00;
C[14560] = 0.000000000000e+00;
C[14561] = 0.000000000000e+00;
C[14562] = 1.000000000000e+00;
C[14563] = 0.000000000000e+00;
C[14564] = 0.000000000000e+00;
C[14565] = 2.000000000000e-01;
C[14566] = -0.000000000000e+00;
C[14567] = -0.000000000000e+00;
C[14568] = 0.000000000000e+00;
C[14569] = 0.000000000000e+00;
C[14570] = 0.000000000000e+00;
C[14571] = 0.000000000000e+00;
C[14572] = 0.000000000000e+00;
C[14573] = 0.000000000000e+00;
C[14574] = 0.000000000000e+00;
C[14575] = 0.000000000000e+00;
C[14576] = 0.000000000000e+00;
C[14577] = 0.000000000000e+00;
C[14578] = 0.000000000000e+00;
C[14579] = 0.000000000000e+00;
C[14580] = 0.000000000000e+00;
C[14581] = 0.000000000000e+00;
C[14582] = 0.000000000000e+00;
C[14583] = 0.000000000000e+00;
C[14584] = 0.000000000000e+00;
C[14585] = 0.000000000000e+00;
C[14586] = 0.000000000000e+00;
C[14587] = 1.000000000000e+00;
C[14588] = 0.000000000000e+00;
C[14589] = -0.000000000000e+00;
C[14590] = 2.000000000000e-01;
C[14591] = -0.000000000000e+00;
C[14592] = 0.000000000000e+00;
C[14593] = 0.000000000000e+00;
C[14594] = 0.000000000000e+00;
C[14595] = 0.000000000000e+00;
C[14596] = 0.000000000000e+00;
C[14597] = 0.000000000000e+00;
C[14598] = 0.000000000000e+00;
C[14599] = 0.000000000000e+00;
C[14600] = 0.000000000000e+00;
C[14601] = 0.000000000000e+00;
C[14602] = 0.000000000000e+00;
C[14603] = 0.000000000000e+00;
C[14604] = 0.000000000000e+00;
C[14605] = 0.000000000000e+00;
C[14606] = 0.000000000000e+00;
C[14607] = 0.000000000000e+00;
C[14608] = 0.000000000000e+00;
C[14609] = 0.000000000000e+00;
C[14610] = 0.000000000000e+00;
C[14611] = 0.000000000000e+00;
C[14612] = 1.000000000000e+00;
C[14613] = -0.000000000000e+00;
C[14614] = -0.000000000000e+00;
C[14615] = 2.000000000000e-01;
C[14616] = 7.203741446649e-02;
C[14617] = 5.170504681936e-03;
C[14618] = -1.294524725328e-03;
C[14619] = 3.853705904146e-01;
C[14620] = -2.075378740634e-04;
C[14621] = -5.091023433732e-04;
C[14622] = 5.217037468811e-02;
C[14623] = -7.101279093369e-04;
C[14624] = -6.869630495941e-04;
C[14625] = 1.302759864749e-01;
C[14626] = 6.574154616597e-04;
C[14627] = 8.223963739253e-06;
C[14628] = 3.136411446304e-02;
C[14629] = -2.106915622343e-04;
C[14630] = -1.409203383630e-04;
C[14631] = 2.518512286703e-03;
C[14632] = -4.602469813076e-05;
C[14633] = -3.779533557185e-05;
C[14634] = 2.454705106760e-03;
C[14635] = -6.730940148693e-05;
C[14636] = -3.424569701025e-05;
C[14637] = -1.923480656648e-05;
C[14638] = -2.367744488821e-06;
C[14639] = -2.091716655448e-06;
C[14640] = 5.283544228078e-03;
C[14641] = 2.439220555261e-02;
C[14642] = 5.427078652176e-03;
C[14643] = -2.863575918691e-04;
C[14644] = 3.974920025573e-01;
C[14645] = 9.213414947411e-03;
C[14646] = -7.188983454496e-04;
C[14647] = 5.554187414296e-02;
C[14648] = 3.343923541750e-03;
C[14649] = 6.649514314026e-04;
C[14650] = 1.262975806497e-01;
C[14651] = 7.126367355618e-05;
C[14652] = -2.162347002444e-04;
C[14653] = 3.263027880147e-02;
C[14654] = 8.465986573887e-04;
C[14655] = -4.564480199527e-05;
C[14656] = 2.701275021951e-03;
C[14657] = 1.747725267310e-04;
C[14658] = -6.276711736928e-05;
C[14659] = 2.684595019546e-03;
C[14660] = 2.407638941016e-04;
C[14661] = -2.226584026282e-06;
C[14662] = -1.554732834207e-05;
C[14663] = 1.931678606817e-06;
C[14664] = -1.291630620872e-03;
C[14665] = 5.517873499721e-03;
C[14666] = 5.937021126658e-02;
C[14667] = -5.903579116594e-04;
C[14668] = 9.513388585296e-03;
C[14669] = 3.873647748623e-01;
C[14670] = -6.756977907879e-04;
C[14671] = 3.255891532393e-03;
C[14672] = 5.566533518985e-02;
C[14673] = 8.416904036367e-06;
C[14674] = 7.731666339249e-05;
C[14675] = 1.294676449753e-01;
C[14676] = -1.465490365856e-04;
C[14677] = 8.620377265337e-04;
C[14678] = 3.164985131080e-02;
C[14679] = -3.591229353458e-05;
C[14680] = 1.686075241820e-04;
C[14681] = 2.710924963532e-03;
C[14682] = -2.397436742594e-05;
C[14683] = 2.206063698683e-04;
C[14684] = 2.846149512402e-03;
C[14685] = -1.769225706265e-06;
C[14686] = 1.491709573149e-06;
C[14687] = -1.555338629448e-05;
C[14688] = 3.863981594493e-01;
C[14689] = -5.412491667688e-04;
C[14690] = -9.016797159828e-04;
C[14691] = 1.167236309903e-01;
C[14692] = 2.156064902955e-03;
C[14693] = 2.302832130903e-03;
C[14694] = 3.832715444038e-01;
C[14695] = -5.131078190564e-04;
C[14696] = -1.887298575197e-03;
C[14697] = 3.143261906536e-02;
C[14698] = -2.332274719044e-04;
C[14699] = -1.673038235405e-04;
C[14700] = 1.325444254616e-01;
C[14701] = 2.903701278555e-04;
C[14702] = 1.971562832730e-04;
C[14703] = 3.108333218695e-02;
C[14704] = 2.298668357491e-06;
C[14705] = -5.260604896149e-05;
C[14706] = 5.771336905232e-02;
C[14707] = 1.652598448991e-04;
C[14708] = 5.816676823560e-04;
C[14709] = 2.505325984063e-03;
C[14710] = 1.049077524485e-05;
C[14711] = 3.345397466989e-05;
C[14712] = -6.129310713157e-04;
C[14713] = 3.985030872433e-01;
C[14714] = 1.028491740649e-02;
C[14715] = 2.187000853974e-03;
C[14716] = 9.362975078159e-02;
C[14717] = -1.955985698038e-02;
C[14718] = -5.032359894467e-04;
C[14719] = 3.859819417251e-01;
C[14720] = 7.510785073277e-03;
C[14721] = -2.377177753845e-04;
C[14722] = 3.269768444720e-02;
C[14723] = 9.134729812796e-04;
C[14724] = 2.924325245900e-04;
C[14725] = 1.305141802225e-01;
C[14726] = -1.848762711471e-03;
C[14727] = 2.463336856941e-06;
C[14728] = 3.145604462179e-02;
C[14729] = 8.346717346052e-04;
C[14730] = 1.847152667627e-04;
C[14731] = 5.864369788701e-02;
C[14732] = 2.558925438498e-03;
C[14733] = 1.154619064972e-05;
C[14734] = 2.556655726462e-03;
C[14735] = 1.284451542722e-04;
C[14736] = -8.972390477009e-04;
C[14737] = 9.949930507241e-03;
C[14738] = 3.873219027785e-01;
C[14739] = 2.305243269430e-03;
C[14740] = -1.951966956396e-02;
C[14741] = 9.388614768695e-02;
C[14742] = -1.898054714454e-03;
C[14743] = 7.686456313589e-03;
C[14744] = 3.966429654893e-01;
C[14745] = -1.667961186515e-04;
C[14746] = 8.956996947107e-04;
C[14747] = 3.164699317188e-02;
C[14748] = 1.973170258415e-04;
C[14749] = -1.846083550376e-03;
C[14750] = 1.303437156062e-01;
C[14751] = -5.338690047288e-05;
C[14752] = 8.441610839917e-04;
C[14753] = 3.265320892214e-02;
C[14754] = 6.173630184381e-04;
C[14755] = 2.498922195165e-03;
C[14756] = 6.456900122508e-02;
C[14757] = 3.525698271732e-05;
C[14758] = 1.247774347996e-04;
C[14759] = 2.840154768883e-03;
C[14760] = 5.176046227851e-02;
C[14761] = -1.581991137212e-04;
C[14762] = -1.718546895479e-04;
C[14763] = 3.832503168282e-01;
C[14764] = -9.599947771607e-04;
C[14765] = -2.397503902680e-03;
C[14766] = 5.558807390337e-02;
C[14767] = 2.971402463591e-04;
C[14768] = -2.209951030192e-03;
C[14769] = 2.491184792730e-03;
C[14770] = -8.264853213373e-06;
C[14771] = -2.322753451916e-06;
C[14772] = 3.108191701524e-02;
C[14773] = -2.798724899066e-05;
C[14774] = -8.668351302128e-05;
C[14775] = 1.293102598016e-01;
C[14776] = -3.439376160159e-05;
C[14777] = -3.512934970563e-04;
C[14778] = 5.071137386910e-01;
C[14779] = 8.489402778385e-04;
C[14780] = 4.754354679350e-03;
C[14781] = 3.713911315709e-02;
C[14782] = 6.961287636446e-05;
C[14783] = 4.388620326654e-04;
C[14784] = -1.550113359747e-04;
C[14785] = 5.539649735922e-02;
C[14786] = 3.473025084625e-03;
C[14787] = -9.066225907847e-04;
C[14788] = 3.858730657559e-01;
C[14789] = 7.807961490975e-03;
C[14790] = 2.635933885721e-04;
C[14791] = 5.010792146474e-02;
C[14792] = -1.374597185356e-02;
C[14793] = -9.016926573283e-06;
C[14794] = 2.691583236368e-03;
C[14795] = 1.830830943307e-04;
C[14796] = -2.393564975773e-05;
C[14797] = 3.144878622384e-02;
C[14798] = 8.522614291507e-04;
C[14799] = -3.663021878739e-05;
C[14800] = 1.288493970739e-01;
C[14801] = -1.250697358865e-03;
C[14802] = 8.284512145293e-04;
C[14803] = 5.058951536936e-01;
C[14804] = 2.395155778830e-03;
C[14805] = 6.866329860759e-05;
C[14806] = 3.703029357270e-02;
C[14807] = 2.167572598821e-04;
C[14808] = -1.698530175786e-04;
C[14809] = 3.579470620296e-03;
C[14810] = 5.574324281779e-02;
C[14811] = -2.282113271850e-03;
C[14812] = 7.585079819003e-03;
C[14813] = 3.967102121498e-01;
C[14814] = -2.292852430423e-03;
C[14815] = -1.372587528907e-02;
C[14816] = 1.822039575604e-02;
C[14817] = -3.321333437483e-06;
C[14818] = 1.904756653007e-04;
C[14819] = 2.716118805394e-03;
C[14820] = -7.892702873836e-05;
C[14821] = 8.396247176536e-04;
C[14822] = 3.265769203283e-02;
C[14823] = -3.568202570717e-04;
C[14824] = -1.249357587899e-03;
C[14825] = 1.259332610691e-01;
C[14826] = 4.723908059660e-03;
C[14827] = 2.469465253706e-03;
C[14828] = 5.266954653886e-01;
C[14829] = 4.378000539712e-04;
C[14830] = 2.203422509358e-04;
C[14831] = 3.870763635282e-02;
C[14832] = -7.011281314037e+00;
C[14833] = 3.068386512169e-02;
C[14834] = -6.290487773997e-03;
C[14835] = 2.155689714027e+00;
C[14836] = 9.854901339155e-03;
C[14837] = -1.370310590044e-03;
C[14838] = 8.097003400656e-01;
C[14839] = -1.217893823919e-02;
C[14840] = -1.175301063456e-02;
C[14841] = 6.606356076217e-02;
C[14842] = 1.006107540809e-02;
C[14843] = 1.040602849999e-03;
C[14844] = 3.859668694597e-01;
C[14845] = -3.484431013763e-03;
C[14846] = -2.530037364666e-03;
C[14847] = 5.204325624713e-02;
C[14848] = -8.372564114368e-04;
C[14849] = -6.801076934753e-04;
C[14850] = 7.895088093340e-02;
C[14851] = -1.233655778654e-03;
C[14852] = -2.820275588587e-04;
C[14853] = 2.426552822350e-03;
C[14854] = -4.473725107806e-05;
C[14855] = -6.506292020600e-06;
C[14856] = 3.294465604454e-02;
C[14857] = -7.205317057487e+00;
C[14858] = 4.381754942900e-02;
C[14859] = 8.178770713687e-03;
C[14860] = 2.110336054905e+00;
C[14861] = 1.024693119330e-02;
C[14862] = -1.218203408378e-02;
C[14863] = 8.528752271493e-01;
C[14864] = 4.742624976107e-02;
C[14865] = 1.017411495423e-02;
C[14866] = 2.120493694291e-02;
C[14867] = -1.050370269938e-03;
C[14868] = -3.578048890406e-03;
C[14869] = 3.961396720278e-01;
C[14870] = 8.647192133194e-03;
C[14871] = -8.192335561316e-04;
C[14872] = 5.529351584476e-02;
C[14873] = 3.391792669790e-03;
C[14874] = -1.103261056144e-03;
C[14875] = 8.479482157513e-02;
C[14876] = 6.917902227935e-03;
C[14877] = -3.840184604746e-05;
C[14878] = 2.664334895860e-03;
C[14879] = 2.567845588501e-04;
C[14880] = -6.232605684863e-03;
C[14881] = 4.563344637989e-02;
C[14882] = -7.050526784499e+00;
C[14883] = -2.971870074212e-03;
C[14884] = 1.435942530187e-02;
C[14885] = 2.121993233408e+00;
C[14886] = -1.147994907946e-02;
C[14887] = 4.662981472302e-02;
C[14888] = 8.562181775155e-01;
C[14889] = 1.043496954456e-03;
C[14890] = -9.595754223931e-04;
C[14891] = 5.994756927246e-02;
C[14892] = -2.642521426442e-03;
C[14893] = 8.941837239232e-03;
C[14894] = 3.870841762196e-01;
C[14895] = -6.399303181212e-04;
C[14896] = 3.297291116857e-03;
C[14897] = 5.566034512446e-02;
C[14898] = -4.908846198916e-05;
C[14899] = 6.417380026077e-03;
C[14900] = 9.016147910352e-02;
C[14901] = 3.622542625852e-06;
C[14902] = 2.345578463423e-04;
C[14903] = 2.850971467229e-03;
C[14904] = 2.176241094720e+00;
C[14905] = 3.080939215693e-03;
C[14906] = -9.198306160679e-03;
C[14907] = -6.227024956705e+00;
C[14908] = 7.654840006356e-03;
C[14909] = 3.259168998838e-03;
C[14910] = 2.125870961429e+00;
C[14911] = 8.310366206131e-03;
C[14912] = -3.390447572821e-03;
C[14913] = 3.869944384943e-01;
C[14914] = -3.832940465306e-03;
C[14915] = -2.953843230765e-03;
C[14916] = 1.190298811289e-01;
C[14917] = 4.171972304917e-03;
C[14918] = 1.858375180656e-03;
C[14919] = 3.801127542385e-01;
C[14920] = 9.819340093722e-04;
C[14921] = 4.775428588951e-04;
C[14922] = 9.751512829667e-01;
C[14923] = 4.082243294755e-03;
C[14924] = 1.210477088288e-02;
C[14925] = 5.758910161744e-02;
C[14926] = 2.379955806239e-04;
C[14927] = 8.350492944011e-04;
C[14928] = 1.747037394109e-03;
C[14929] = 2.130557748623e+00;
C[14930] = 2.979000172565e-02;
C[14931] = 8.273559026733e-03;
C[14932] = -6.282484341919e+00;
C[14933] = -4.322445365559e-02;
C[14934] = 8.451545204193e-03;
C[14935] = 2.099805297245e+00;
C[14936] = -1.296057749708e-02;
C[14937] = -3.889824211016e-03;
C[14938] = 3.971507567137e-01;
C[14939] = 9.713366060421e-03;
C[14940] = 4.202908255936e-03;
C[14941] = 9.782156420727e-02;
C[14942] = -2.028363168726e-02;
C[14943] = 9.636861442937e-04;
C[14944] = 3.830896785373e-01;
C[14945] = 8.343376034263e-03;
C[14946] = 4.317251112221e-03;
C[14947] = 9.873767896421e-01;
C[14948] = 3.924702981171e-02;
C[14949] = 2.578191648561e-04;
C[14950] = 5.843974053765e-02;
C[14951] = 2.554346996679e-03;
C[14952] = -9.133044676598e-03;
C[14953] = 2.497724238989e-02;
C[14954] = 2.121135791733e+00;
C[14955] = 3.307391769387e-03;
C[14956] = -4.242070532713e-02;
C[14957] = -6.285703137153e+00;
C[14958] = -3.427659339962e-03;
C[14959] = -1.065531740140e-02;
C[14960] = 2.107184378869e+00;
C[14961] = -2.918174068994e-03;
C[14962] = 9.383707693023e-03;
C[14963] = 3.870413041359e-01;
C[14964] = 1.860786319184e-03;
C[14965] = -2.024344427084e-02;
C[14966] = 9.394128862725e-02;
C[14967] = 4.353230195898e-04;
C[14968] = 8.526912598612e-03;
C[14969] = 3.969225870981e-01;
C[14970] = 1.250744192349e-02;
C[14971] = 3.852261560384e-02;
C[14972] = 1.085153259293e+00;
C[14973] = 8.703905721709e-04;
C[14974] = 2.493831335051e-03;
C[14975] = 6.453157248175e-02;
C[14976] = 8.015020918737e-01;
C[14977] = -9.680494492118e-04;
C[14978] = -1.403087054661e-03;
C[14979] = 2.125446409917e+00;
C[14980] = -6.836305500876e-04;
C[14981] = -1.341664310448e-02;
C[14982] = -7.087467643941e+00;
C[14983] = -2.503002829084e-03;
C[14984] = -2.430087305441e-02;
C[14985] = 5.163334383754e-02;
C[14986] = -2.585343244032e-04;
C[14987] = -1.360872168812e-04;
C[14988] = 3.800915266629e-01;
C[14989] = 5.069273565797e-04;
C[14990] = -6.412616863622e-05;
C[14991] = 5.811927919963e-02;
C[14992] = -1.163064619592e-03;
C[14993] = -6.463610914639e-03;
C[14994] = 4.085609713989e+00;
C[14995] = 5.614212320921e-03;
C[14996] = 3.895122970229e-02;
C[14997] = 5.078567407238e-01;
C[14998] = 9.772943116243e-04;
C[14999] = 6.661605034768e-03;
C[15000] = -1.076606771943e-03;
C[15001] = 8.499676914744e-01;
C[15002] = 5.097248576765e-02;
C[15003] = 4.400707715666e-04;
C[15004] = 2.097627777860e+00;
C[15005] = -8.225213853686e-03;
C[15006] = -3.173939984824e-03;
C[15007] = -7.097502642264e+00;
C[15008] = -3.907360159789e-02;
C[15009] = -2.821398380746e-04;
C[15010] = 5.514813906102e-02;
C[15011] = 3.514424669089e-03;
C[15012] = 5.884192376440e-04;
C[15013] = 3.829808025680e-01;
C[15014] = 8.648417775998e-03;
C[15015] = -1.196611477379e-03;
C[15016] = 5.371317719607e-02;
C[15017] = -1.386618324211e-02;
C[15018] = 5.313720426725e-03;
C[15019] = 4.062844494400e+00;
C[15020] = -5.529807316976e-03;
C[15021] = 9.546762701831e-04;
C[15022] = 5.054371342457e-01;
C[15023] = 1.645470929136e-03;
C[15024] = -1.410809994254e-03;
C[15025] = 5.213719133199e-02;
C[15026] = 8.577763300742e-01;
C[15027] = -1.128674150588e-02;
C[15028] = -1.147468258256e-02;
C[15029] = 2.108529312077e+00;
C[15030] = -2.595890105903e-02;
C[15031] = -3.867167030807e-02;
C[15032] = -7.226280101767e+00;
C[15033] = -1.629976614598e-04;
C[15034] = 3.627339748336e-03;
C[15035] = 5.573825275239e-02;
C[15036] = 8.272816224199e-05;
C[15037] = 8.417670779989e-03;
C[15038] = 3.969898337585e-01;
C[15039] = -6.546512314870e-03;
C[15040] = -1.384608667762e-02;
C[15041] = 1.869281884090e-02;
C[15042] = 3.860957503332e-02;
C[15043] = -4.378946173333e-03;
C[15044] = 4.176691651305e+00;
C[15045] = 6.631493136210e-03;
C[15046] = 1.721772915078e-03;
C[15047] = 5.259519963050e-01;
C[15048] = 0.000000000000e+00;
C[15049] = 0.000000000000e+00;
C[15050] = 0.000000000000e+00;
C[15051] = 0.000000000000e+00;
C[15052] = 0.000000000000e+00;
C[15053] = 0.000000000000e+00;
C[15054] = 0.000000000000e+00;
C[15055] = 0.000000000000e+00;
C[15056] = 0.000000000000e+00;
C[15057] = 0.000000000000e+00;
C[15058] = 0.000000000000e+00;
C[15059] = 0.000000000000e+00;
C[15060] = 0.000000000000e+00;
C[15061] = 0.000000000000e+00;
C[15062] = 0.000000000000e+00;
C[15063] = 0.000000000000e+00;
C[15064] = 0.000000000000e+00;
C[15065] = 0.000000000000e+00;
C[15066] = 1.000000000000e+00;
C[15067] = 0.000000000000e+00;
C[15068] = 0.000000000000e+00;
C[15069] = 2.000000000000e-01;
C[15070] = -0.000000000000e+00;
C[15071] = -0.000000000000e+00;
C[15072] = 0.000000000000e+00;
C[15073] = 0.000000000000e+00;
C[15074] = 0.000000000000e+00;
C[15075] = 0.000000000000e+00;
C[15076] = 0.000000000000e+00;
C[15077] = 0.000000000000e+00;
C[15078] = 0.000000000000e+00;
C[15079] = 0.000000000000e+00;
C[15080] = 0.000000000000e+00;
C[15081] = 0.000000000000e+00;
C[15082] = 0.000000000000e+00;
C[15083] = 0.000000000000e+00;
C[15084] = 0.000000000000e+00;
C[15085] = 0.000000000000e+00;
C[15086] = 0.000000000000e+00;
C[15087] = 0.000000000000e+00;
C[15088] = 0.000000000000e+00;
C[15089] = 0.000000000000e+00;
C[15090] = 0.000000000000e+00;
C[15091] = 1.000000000000e+00;
C[15092] = 0.000000000000e+00;
C[15093] = -0.000000000000e+00;
C[15094] = 2.000000000000e-01;
C[15095] = -0.000000000000e+00;
C[15096] = 0.000000000000e+00;
C[15097] = 0.000000000000e+00;
C[15098] = 0.000000000000e+00;
C[15099] = 0.000000000000e+00;
C[15100] = 0.000000000000e+00;
C[15101] = 0.000000000000e+00;
C[15102] = 0.000000000000e+00;
C[15103] = 0.000000000000e+00;
C[15104] = 0.000000000000e+00;
C[15105] = 0.000000000000e+00;
C[15106] = 0.000000000000e+00;
C[15107] = 0.000000000000e+00;
C[15108] = 0.000000000000e+00;
C[15109] = 0.000000000000e+00;
C[15110] = 0.000000000000e+00;
C[15111] = 0.000000000000e+00;
C[15112] = 0.000000000000e+00;
C[15113] = 0.000000000000e+00;
C[15114] = 0.000000000000e+00;
C[15115] = 0.000000000000e+00;
C[15116] = 1.000000000000e+00;
C[15117] = -0.000000000000e+00;
C[15118] = -0.000000000000e+00;
C[15119] = 2.000000000000e-01;
C[15120] = 6.216206773711e-02;
C[15121] = 1.510298557988e-02;
C[15122] = 3.718448029503e-03;
C[15123] = 3.827632752828e-01;
C[15124] = -7.877046349713e-03;
C[15125] = -5.126094915816e-03;
C[15126] = 5.055416756708e-02;
C[15127] = -1.411665420862e-03;
C[15128] = -2.234534060002e-04;
C[15129] = 1.299565098335e-01;
C[15130] = 1.536688321776e-03;
C[15131] = 3.751439005301e-04;
C[15132] = 3.082808798841e-02;
C[15133] = -9.121192871877e-04;
C[15134] = -4.606509994655e-04;
C[15135] = 2.422891010297e-03;
C[15136] = -8.408680759575e-05;
C[15137] = -1.092467396533e-05;
C[15138] = 2.331221838708e-03;
C[15139] = -1.106734972182e-04;
C[15140] = 2.908859284903e-05;
C[15141] = -2.233600766024e-05;
C[15142] = -2.881255295819e-06;
C[15143] = 2.465700583499e-07;
C[15144] = 1.513697390055e-02;
C[15145] = 2.193583866222e-02;
C[15146] = -7.231840671418e-03;
C[15147] = -7.926073069970e-03;
C[15148] = 3.941035395122e-01;
C[15149] = 8.369877400279e-03;
C[15150] = -1.326027864624e-03;
C[15151] = 5.442104182073e-02;
C[15152] = 3.510041939820e-03;
C[15153] = 1.538954209820e-03;
C[15154] = 1.263018500470e-01;
C[15155] = -7.749574197834e-04;
C[15156] = -9.157960398508e-04;
C[15157] = 3.235838451106e-02;
C[15158] = 8.757537367063e-04;
C[15159] = -7.871831433830e-05;
C[15160] = 2.649185748410e-03;
C[15161] = 1.878512129307e-04;
C[15162] = -9.847492321215e-05;
C[15163] = 2.628653958751e-03;
C[15164] = 2.529155491603e-04;
C[15165] = -2.579196577909e-06;
C[15166] = -1.577150377600e-05;
C[15167] = 2.220427252958e-06;
C[15168] = 3.739834810051e-03;
C[15169] = -7.266666541492e-03;
C[15170] = 6.191090446712e-02;
C[15171] = -5.319212322915e-03;
C[15172] = 8.754454676635e-03;
C[15173] = 3.813097882647e-01;
C[15174] = -1.241804318082e-04;
C[15175] = 3.380197283462e-03;
C[15176] = 5.360412906625e-02;
C[15177] = 3.765696858999e-04;
C[15178] = -7.772791444549e-04;
C[15179] = 1.300784520100e-01;
C[15180] = -4.716576595664e-04;
C[15181] = 8.972319622376e-04;
C[15182] = 3.070725693284e-02;
C[15183] = -4.620605426791e-06;
C[15184] = 1.789155173174e-04;
C[15185] = 2.581126318619e-03;
C[15186] = 4.366996726797e-05;
C[15187] = 2.274721675729e-04;
C[15188] = 2.632255522547e-03;
C[15189] = 6.001930101585e-07;
C[15190] = 1.551460490598e-06;
C[15191] = -2.206213954181e-05;
C[15192] = 3.828734894991e-01;
C[15193] = -8.077270671940e-03;
C[15194] = -5.475574581326e-03;
C[15195] = 1.206372778447e-01;
C[15196] = 6.612182587221e-03;
C[15197] = 8.397900386734e-05;
C[15198] = 3.818794196905e-01;
C[15199] = 3.818478509843e-03;
C[15200] = 4.197310264152e-03;
C[15201] = 3.083543560283e-02;
C[15202] = -9.258758799821e-04;
C[15203] = -4.820818101272e-04;
C[15204] = 1.330032624466e-01;
C[15205] = 7.212184147207e-04;
C[15206] = -3.486544273655e-05;
C[15207] = 3.118537705981e-02;
C[15208] = 3.262128139560e-04;
C[15209] = 4.587113958337e-04;
C[15210] = 5.838656151947e-02;
C[15211] = 7.861423277883e-04;
C[15212] = 2.008335162114e-03;
C[15213] = 2.550331412063e-03;
C[15214] = 3.547400434322e-05;
C[15215] = 9.629042680260e-05;
C[15216] = -8.050586631172e-03;
C[15217] = 3.939063093608e-01;
C[15218] = 8.864393692452e-03;
C[15219] = 6.597987148901e-03;
C[15220] = 1.007088198316e-01;
C[15221] = -2.036669407755e-02;
C[15222] = 3.718194398082e-03;
C[15223] = 3.830554642011e-01;
C[15224] = 7.827982119260e-03;
C[15225] = -9.236886392849e-04;
C[15226] = 3.234523583430e-02;
C[15227] = 9.045612299588e-04;
C[15228] = 7.202720521660e-04;
C[15229] = 1.309475526715e-01;
C[15230] = -1.871044079113e-03;
C[15231] = 3.207461094488e-04;
C[15232] = 3.134551643283e-02;
C[15233] = 7.818144564746e-04;
C[15234] = 8.005148447530e-04;
C[15235] = 5.855196228072e-02;
C[15236] = 2.228745231164e-03;
C[15237] = 3.634797167868e-05;
C[15238] = 2.562429483253e-03;
C[15239] = 1.084268916098e-04;
C[15240] = -5.353051015872e-03;
C[15241] = 8.631737653414e-03;
C[15242] = 3.816040211633e-01;
C[15243] = 9.368387739183e-05;
C[15244] = -2.047758084021e-02;
C[15245] = 9.962124397266e-02;
C[15246] = 4.111827698836e-03;
C[15247] = 8.056278186678e-03;
C[15248] = 3.985075344217e-01;
C[15249] = -4.757814061359e-04;
C[15250] = 8.932110869152e-04;
C[15251] = 3.072687245941e-02;
C[15252] = -3.421845116825e-05;
C[15253] = -1.878436529957e-03;
C[15254] = 1.312977035490e-01;
C[15255] = 4.544790957655e-04;
C[15256] = 7.946692203229e-04;
C[15257] = 3.273831291859e-02;
C[15258] = 2.005745362218e-03;
C[15259] = 2.176940681604e-03;
C[15260] = 6.448961460025e-02;
C[15261] = 9.615944836868e-05;
C[15262] = 1.053601735582e-04;
C[15263] = 2.831843996375e-03;
C[15264] = 5.169542186089e-02;
C[15265] = -8.614681949400e-04;
C[15266] = 8.962960352081e-05;
C[15267] = 3.810542978891e-01;
C[15268] = 3.318293635628e-03;
C[15269] = 4.190159354205e-03;
C[15270] = 5.670849596570e-02;
C[15271] = -2.995831402015e-03;
C[15272] = -1.019562667778e-02;
C[15273] = 2.498974629885e-03;
C[15274] = -4.774766969274e-05;
C[15275] = 9.633396928479e-06;
C[15276] = 3.113036893972e-02;
C[15277] = 2.940860586185e-04;
C[15278] = 4.597012061234e-04;
C[15279] = 1.291814916727e-01;
C[15280] = -3.082363917405e-04;
C[15281] = -9.118686763812e-04;
C[15282] = 5.081370645077e-01;
C[15283] = 6.399055083371e-04;
C[15284] = 5.890132365799e-03;
C[15285] = 3.720821252410e-02;
C[15286] = 6.202107145170e-05;
C[15287] = 4.414327596869e-04;
C[15288] = -9.391886676168e-04;
C[15289] = 5.457031351894e-02;
C[15290] = 3.204906891350e-03;
C[15291] = 3.442998638371e-03;
C[15292] = 3.828792359775e-01;
C[15293] = 8.033996876697e-03;
C[15294] = -3.009236515460e-03;
C[15295] = 5.365256713623e-02;
C[15296] = -1.340793468601e-02;
C[15297] = -5.258835737943e-05;
C[15298] = 2.659137194957e-03;
C[15299] = 1.672294911766e-04;
C[15300] = 3.011808225246e-04;
C[15301] = 3.133376788459e-02;
C[15302] = 7.931837996575e-04;
C[15303] = -3.091300659701e-04;
C[15304] = 1.289198480333e-01;
C[15305] = -1.195220950210e-03;
C[15306] = 6.119731129564e-04;
C[15307] = 5.061082915484e-01;
C[15308] = 2.004765497710e-03;
C[15309] = 6.079110152271e-05;
C[15310] = 3.710223350452e-02;
C[15311] = 2.338078107384e-04;
C[15312] = 2.732802747716e-06;
C[15313] = 3.302001638591e-03;
C[15314] = 5.474471710428e-02;
C[15315] = 4.340983373624e-03;
C[15316] = 7.837831905676e-03;
C[15317] = 3.973141572235e-01;
C[15318] = -1.023177068849e-02;
C[15319] = -1.341388346230e-02;
C[15320] = 1.945635806753e-02;
C[15321] = 4.154406617869e-06;
C[15322] = 1.739818595155e-04;
C[15323] = 2.657165521155e-03;
C[15324] = 4.682896031318e-04;
C[15325] = 7.824711089023e-04;
C[15326] = 3.265875443871e-02;
C[15327] = -9.142782770953e-04;
C[15328] = -1.195617535296e-03;
C[15329] = 1.260169553310e-01;
C[15330] = 5.873644561586e-03;
C[15331] = 2.084413258340e-03;
C[15332] = 5.259125614477e-01;
C[15333] = 4.409687423693e-04;
C[15334] = 2.375828668749e-04;
C[15335] = 3.868449860175e-02;
C[15336] = -7.053752885366e+00;
C[15337] = 5.641624970392e-02;
C[15338] = 3.726545388573e-03;
C[15339] = 2.128239534294e+00;
C[15340] = -9.724573421795e-03;
C[15341] = -1.888230205526e-02;
C[15342] = 7.801337587082e-01;
C[15343] = -2.187000509879e-02;
C[15344] = -3.944660403929e-03;
C[15345] = 6.671793657416e-02;
C[15346] = 1.940319360945e-02;
C[15347] = 4.039995267186e-03;
C[15348] = 3.757178581771e-01;
C[15349] = -1.107149104931e-02;
C[15350] = -5.421397243863e-03;
C[15351] = 5.029298798401e-02;
C[15352] = -1.570578819891e-03;
C[15353] = -1.826220306523e-04;
C[15354] = 7.679302695866e-02;
C[15355] = -2.454208014504e-03;
C[15356] = 1.083599519990e-03;
C[15357] = 2.377657827747e-03;
C[15358] = -9.396087390265e-05;
C[15359] = 4.521126312325e-05;
C[15360] = 5.709601611740e-02;
C[15361] = -7.216508182430e+00;
C[15362] = -8.007650589230e-03;
C[15363] = -1.048074411384e-02;
C[15364] = 2.109914102022e+00;
C[15365] = -1.515824264970e-03;
C[15366] = -2.070965352946e-02;
C[15367] = 8.376269925641e-01;
C[15368] = 4.978652436912e-02;
C[15369] = 1.943718193013e-02;
C[15370] = 2.300061678305e-02;
C[15371] = -8.812759720222e-03;
C[15372] = -1.113896320196e-02;
C[15373] = 3.936593474600e-01;
C[15374] = 9.131429133893e-03;
C[15375] = -1.474128767545e-03;
C[15376] = 5.445422988448e-02;
C[15377] = 3.496737176327e-03;
C[15378] = -2.166589029278e-03;
C[15379] = 8.361417293480e-02;
C[15380] = 6.733624056947e-03;
C[15381] = -8.200963561661e-05;
C[15382] = 2.637718776009e-03;
C[15383] = 2.408295110047e-04;
C[15384] = 4.154280999534e-03;
C[15385] = -8.704167990694e-03;
C[15386] = -7.049672256398e+00;
C[15387] = -2.184749318645e-02;
C[15388] = 4.400462959909e-03;
C[15389] = 2.089163573616e+00;
C[15390] = -2.668800885855e-03;
C[15391] = 4.820354304730e-02;
C[15392] = 8.195253393402e-01;
C[15393] = 4.061382047734e-03;
C[15394] = -8.847585590296e-03;
C[15395] = 6.967998530557e-02;
C[15396] = -5.622561166980e-03;
C[15397] = 9.520195104051e-03;
C[15398] = 3.722109995995e-01;
C[15399] = -7.488172400820e-05;
C[15400] = 3.353252468668e-03;
C[15401] = 5.303361594149e-02;
C[15402] = 1.408164764226e-03;
C[15403] = 6.167733113121e-03;
C[15404] = 8.524210351502e-02;
C[15405] = 5.843687240650e-05;
C[15406] = 2.169576071748e-04;
C[15407] = 2.663727744558e-03;
C[15408] = 2.130443818620e+00;
C[15409] = -1.350469615324e-02;
C[15410] = -2.497473835468e-02;
C[15411] = -6.214662922631e+00;
C[15412] = 2.086333783217e-02;
C[15413] = -3.033555600235e-03;
C[15414] = 2.144487745252e+00;
C[15415] = 3.118700797440e-02;
C[15416] = 8.217541417442e-03;
C[15417] = 3.758280723934e-01;
C[15418] = -1.129016080393e-02;
C[15419] = -5.778923425392e-03;
C[15420] = 1.251171095777e-01;
C[15421] = 9.198343083230e-03;
C[15422] = -9.802899407724e-04;
C[15423] = 3.845189305888e-01;
C[15424] = 4.659751677080e-03;
C[15425] = 5.572644783508e-03;
C[15426] = 9.904753566883e-01;
C[15427] = 1.411491342754e-02;
C[15428] = 3.325985378216e-02;
C[15429] = 5.851245797810e-02;
C[15430] = 8.132339441508e-04;
C[15431] = 2.018943371108e-03;
C[15432] = -1.319537905097e-02;
C[15433] = 2.105969498993e+00;
C[15434] = 6.599243276264e-03;
C[15435] = 2.057942906577e-02;
C[15436] = -6.271803628322e+00;
C[15437] = -4.700797855064e-02;
C[15438] = 2.987843887331e-02;
C[15439] = 2.103281569891e+00;
C[15440] = -1.474668656208e-02;
C[15441] = -1.124503133077e-02;
C[15442] = 3.934621173086e-01;
C[15443] = 9.630134119869e-03;
C[15444] = 9.184147644910e-03;
C[15445] = 1.001219917093e-01;
C[15446] = -2.022501499564e-02;
C[15447] = 4.555323461629e-03;
C[15448] = 3.840148930404e-01;
C[15449] = 6.828477834898e-03;
C[15450] = 1.418010162828e-02;
C[15451] = 9.898698091493e-01;
C[15452] = 3.300786468349e-02;
C[15453] = 8.206664697172e-04;
C[15454] = 5.868050580441e-02;
C[15455] = 2.215777689979e-03;
C[15456] = -2.342142405637e-02;
C[15457] = 3.721380797724e-03;
C[15458] = 2.095048231588e+00;
C[15459] = -2.839458129745e-03;
C[15460] = -4.922571380388e-02;
C[15461] = -6.261610569711e+00;
C[15462] = 7.337336005150e-03;
C[15463] = -1.143569620775e-02;
C[15464] = 2.143310834859e+00;
C[15465] = -5.648353343919e-03;
C[15466] = 9.393289387028e-03;
C[15467] = 3.725052324980e-01;
C[15468] = -9.705850672479e-04;
C[15469] = -2.033590175830e-02;
C[15470] = 1.088921223590e-01;
C[15471] = 5.493020501940e-03;
C[15472] = 7.067299369270e-03;
C[15473] = 3.987483205413e-01;
C[15474] = 3.296640255551e-02;
C[15475] = 3.247246588641e-02;
C[15476] = 1.085070652044e+00;
C[15477] = 2.001785203884e-03;
C[15478] = 2.169434403715e-03;
C[15479] = 6.446201219307e-02;
C[15480] = 8.029588445845e-01;
C[15481] = -1.141846013579e-02;
C[15482] = 1.607399820726e-03;
C[15483] = 2.127985309224e+00;
C[15484] = 2.188042362423e-02;
C[15485] = 8.903969112532e-03;
C[15486] = -7.092405219645e+00;
C[15487] = -6.668734029013e-03;
C[15488] = -3.646517961309e-02;
C[15489] = 5.143424227783e-02;
C[15490] = -1.009569097862e-03;
C[15491] = 1.389283113209e-04;
C[15492] = 3.836938087874e-01;
C[15493] = 4.155422699175e-03;
C[15494] = 5.571352157309e-03;
C[15495] = 5.409432284559e-02;
C[15496] = -3.617947072543e-03;
C[15497] = -1.048480682763e-02;
C[15498] = 4.084796192559e+00;
C[15499] = -4.149208960043e-04;
C[15500] = 2.553619774424e-02;
C[15501] = 5.084464755604e-01;
C[15502] = 6.047688234289e-04;
C[15503] = 4.757220867366e-03;
C[15504] = -1.242047003389e-02;
C[15505] = 8.406124265283e-01;
C[15506] = 4.469773520507e-02;
C[15507] = 2.367741054497e-02;
C[15508] = 2.099757005419e+00;
C[15509] = -1.188132240736e-02;
C[15510] = -6.936836297903e-03;
C[15511] = -7.098251556407e+00;
C[15512] = -3.770076755455e-02;
C[15513] = -1.098102066646e-03;
C[15514] = 5.460350158269e-02;
C[15515] = 3.177962076556e-03;
C[15516] = 4.284271805609e-03;
C[15517] = 3.838386648168e-01;
C[15518] = 7.045018059289e-03;
C[15519] = -3.631352185987e-03;
C[15520] = 5.152762310243e-02;
C[15521] = -1.301185174726e-02;
C[15522] = -8.097370541299e-04;
C[15523] = 4.069662241251e+00;
C[15524] = 2.921033443220e-04;
C[15525] = 5.840442443351e-04;
C[15526] = 5.072460759133e-01;
C[15527] = 2.617246286111e-03;
C[15528] = 5.790637710298e-04;
C[15529] = 4.562571834454e-02;
C[15530] = 8.423371001008e-01;
C[15531] = 1.109100360687e-02;
C[15532] = -1.454969083377e-02;
C[15533] = 2.119443290894e+00;
C[15534] = -3.718805982732e-02;
C[15535] = -3.781974308038e-02;
C[15536] = -7.220186814189e+00;
C[15537] = 4.356417809566e-05;
C[15538] = 3.288696875098e-03;
C[15539] = 5.417420397953e-02;
C[15540] = 5.716317892980e-03;
C[15541] = 6.838327621313e-03;
C[15542] = 3.975549433430e-01;
C[15543] = -1.052095083834e-02;
C[15544] = -1.301780052355e-02;
C[15545] = 1.953696094022e-02;
C[15546] = 2.538945849735e-02;
C[15547] = 1.501656340303e-03;
C[15548] = 4.177629834912e+00;
C[15549] = 4.756649482877e-03;
C[15550] = 2.690127616962e-03;
C[15551] = 5.262508928428e-01;
C[15552] = 0.000000000000e+00;
C[15553] = 0.000000000000e+00;
C[15554] = 0.000000000000e+00;
C[15555] = 0.000000000000e+00;
C[15556] = 0.000000000000e+00;
C[15557] = 0.000000000000e+00;
C[15558] = 0.000000000000e+00;
C[15559] = 0.000000000000e+00;
C[15560] = 0.000000000000e+00;
C[15561] = 0.000000000000e+00;
C[15562] = 0.000000000000e+00;
C[15563] = 0.000000000000e+00;
C[15564] = 0.000000000000e+00;
C[15565] = 0.000000000000e+00;
C[15566] = 0.000000000000e+00;
C[15567] = 0.000000000000e+00;
C[15568] = 0.000000000000e+00;
C[15569] = 0.000000000000e+00;
C[15570] = 1.000000000000e+00;
C[15571] = 0.000000000000e+00;
C[15572] = 0.000000000000e+00;
C[15573] = 2.000000000000e-01;
C[15574] = -0.000000000000e+00;
C[15575] = -0.000000000000e+00;
C[15576] = 0.000000000000e+00;
C[15577] = 0.000000000000e+00;
C[15578] = 0.000000000000e+00;
C[15579] = 0.000000000000e+00;
C[15580] = 0.000000000000e+00;
C[15581] = 0.000000000000e+00;
C[15582] = 0.000000000000e+00;
C[15583] = 0.000000000000e+00;
C[15584] = 0.000000000000e+00;
C[15585] = 0.000000000000e+00;
C[15586] = 0.000000000000e+00;
C[15587] = 0.000000000000e+00;
C[15588] = 0.000000000000e+00;
C[15589] = 0.000000000000e+00;
C[15590] = 0.000000000000e+00;
C[15591] = 0.000000000000e+00;
C[15592] = 0.000000000000e+00;
C[15593] = 0.000000000000e+00;
C[15594] = 0.000000000000e+00;
C[15595] = 1.000000000000e+00;
C[15596] = 0.000000000000e+00;
C[15597] = -0.000000000000e+00;
C[15598] = 2.000000000000e-01;
C[15599] = -0.000000000000e+00;
C[15600] = 0.000000000000e+00;
C[15601] = 0.000000000000e+00;
C[15602] = 0.000000000000e+00;
C[15603] = 0.000000000000e+00;
C[15604] = 0.000000000000e+00;
C[15605] = 0.000000000000e+00;
C[15606] = 0.000000000000e+00;
C[15607] = 0.000000000000e+00;
C[15608] = 0.000000000000e+00;
C[15609] = 0.000000000000e+00;
C[15610] = 0.000000000000e+00;
C[15611] = 0.000000000000e+00;
C[15612] = 0.000000000000e+00;
C[15613] = 0.000000000000e+00;
C[15614] = 0.000000000000e+00;
C[15615] = 0.000000000000e+00;
C[15616] = 0.000000000000e+00;
C[15617] = 0.000000000000e+00;
C[15618] = 0.000000000000e+00;
C[15619] = 0.000000000000e+00;
C[15620] = 1.000000000000e+00;
C[15621] = -0.000000000000e+00;
C[15622] = -0.000000000000e+00;
C[15623] = 2.000000000000e-01;
C[15624] = 7.527710712347e-02;
C[15625] = 1.575762745951e-02;
C[15626] = -7.570770482977e-04;
C[15627] = 3.675775098586e-01;
C[15628] = -4.281707799501e-03;
C[15629] = -3.029033550200e-03;
C[15630] = 5.006034665334e-02;
C[15631] = 5.630394897925e-04;
C[15632] = 7.172166036977e-04;
C[15633] = 1.308605738413e-01;
C[15634] = 9.847234535148e-04;
C[15635] = -3.122399858631e-04;
C[15636] = 3.001486953533e-02;
C[15637] = -4.710065917138e-05;
C[15638] = -1.061651744835e-04;
C[15639] = 2.426144229621e-03;
C[15640] = 4.703329795172e-05;
C[15641] = 4.556682673514e-05;
C[15642] = 2.400547963723e-03;
C[15643] = 9.682565587183e-05;
C[15644] = 9.891195440119e-05;
C[15645] = -1.681992523039e-05;
C[15646] = 4.138769933363e-06;
C[15647] = 2.809616290357e-06;
C[15648] = 1.551042087794e-02;
C[15649] = 2.102307081821e-02;
C[15650] = -5.834277702382e-03;
C[15651] = -4.227338427980e-03;
C[15652] = 3.997538336755e-01;
C[15653] = 1.064990564543e-02;
C[15654] = 6.433481845114e-04;
C[15655] = 5.574820629235e-02;
C[15656] = 3.395449069375e-03;
C[15657] = 9.682430147435e-04;
C[15658] = 1.263475910306e-01;
C[15659] = -2.801139250809e-04;
C[15660] = -3.997994204300e-05;
C[15661] = 3.290904888133e-02;
C[15662] = 9.606977345516e-04;
C[15663] = 5.069481918810e-05;
C[15664] = 2.721534948217e-03;
C[15665] = 1.772014614683e-04;
C[15666] = 1.019237992645e-04;
C[15667] = 2.738546969500e-03;
C[15668] = 2.442195420436e-04;
C[15669] = 4.113426433729e-06;
C[15670] = -1.499537044571e-05;
C[15671] = 1.287156470320e-06;
C[15672] = -7.688527611723e-04;
C[15673] = -5.978024101373e-03;
C[15674] = 7.554601422003e-02;
C[15675] = -3.045559842559e-03;
C[15676] = 1.109798455286e-02;
C[15677] = 3.673852167185e-01;
C[15678] = 7.421275319816e-04;
C[15679] = 3.215564259134e-03;
C[15680] = 5.302721757160e-02;
C[15681] = -3.130250333881e-04;
C[15682] = -2.896970183470e-04;
C[15683] = 1.306959258516e-01;
C[15684] = -1.045231489049e-04;
C[15685] = 9.845111005340e-04;
C[15686] = 3.010691809190e-02;
C[15687] = 4.589970378066e-05;
C[15688] = 1.661912423476e-04;
C[15689] = 2.588809257804e-03;
C[15690] = 9.555384747002e-05;
C[15691] = 2.167104131967e-04;
C[15692] = 2.728141852758e-03;
C[15693] = 2.543295458798e-06;
C[15694] = 7.265807605921e-07;
C[15695] = -1.384530643859e-05;
C[15696] = 3.670829313999e-01;
C[15697] = -4.588948702694e-03;
C[15698] = -2.439365120306e-03;
C[15699] = 1.329051418086e-01;
C[15700] = 3.204353919580e-03;
C[15701] = -3.191724244152e-03;
C[15702] = 3.856810054775e-01;
C[15703] = 2.198184659271e-03;
C[15704] = 4.260582168975e-03;
C[15705] = 2.998189763809e-02;
C[15706] = -6.408729369058e-05;
C[15707] = -6.411016742140e-05;
C[15708] = 1.336591639151e-01;
C[15709] = -6.316690532512e-05;
C[15710] = -3.132221996510e-04;
C[15711] = 3.137503171564e-02;
C[15712] = 1.431142443504e-04;
C[15713] = 3.108707815199e-04;
C[15714] = 5.876888523135e-02;
C[15715] = 3.559405174780e-04;
C[15716] = 1.077320916032e-03;
C[15717] = 2.558249729217e-03;
C[15718] = 1.275173694752e-05;
C[15719] = 4.376871842522e-05;
C[15720] = -4.140924696409e-03;
C[15721] = 3.987767958694e-01;
C[15722] = 1.026875620033e-02;
C[15723] = 2.986023901296e-03;
C[15724] = 9.389633328911e-02;
C[15725] = -2.025747359139e-02;
C[15726] = 2.079778612384e-03;
C[15727] = 3.836075859562e-01;
C[15728] = 6.446529583898e-03;
C[15729] = -3.771511896527e-05;
C[15730] = 3.284391302759e-02;
C[15731] = 9.292292103657e-04;
C[15732] = -7.772223987742e-05;
C[15733] = 1.302582102064e-01;
C[15734] = -1.819238821447e-03;
C[15735] = 1.375588799136e-04;
C[15736] = 3.146544530038e-02;
C[15737] = 7.163432373696e-04;
C[15738] = 3.585942628272e-04;
C[15739] = 5.936906224178e-02;
C[15740] = 2.367606691727e-03;
C[15741] = 1.328232960198e-05;
C[15742] = 2.603906996146e-03;
C[15743] = 1.181270609241e-04;
C[15744] = -2.264950316785e-03;
C[15745] = 9.864897854435e-03;
C[15746] = 3.672700492399e-01;
C[15747] = -3.270339367998e-03;
C[15748] = -2.023445122834e-02;
C[15749] = 1.133488208946e-01;
C[15750] = 4.194277391074e-03;
C[15751] = 6.712331562002e-03;
C[15752] = 3.992986191664e-01;
C[15753] = -5.522629225580e-05;
C[15754] = 9.083638818184e-04;
C[15755] = 3.009924025999e-02;
C[15756] = -3.184632079074e-04;
C[15757] = -1.817703997244e-03;
C[15758] = 1.318389411563e-01;
C[15759] = 3.078728129318e-04;
C[15760] = 7.306352616736e-04;
C[15761] = 3.279500683553e-02;
C[15762] = 1.076836982361e-03;
C[15763] = 2.344099910728e-03;
C[15764] = 6.469217983385e-02;
C[15765] = 4.416422685039e-05;
C[15766] = 1.160870183985e-04;
C[15767] = 2.839977784739e-03;
C[15768] = 4.932770127724e-02;
C[15769] = -5.913616302192e-04;
C[15770] = 2.044324763801e-04;
C[15771] = 3.862343975624e-01;
C[15772] = 3.204317431950e-03;
C[15773] = 4.704621850832e-03;
C[15774] = 5.252507610973e-02;
C[15775] = -3.097880700146e-03;
C[15776] = -7.814475923252e-03;
C[15777] = 2.377301204547e-03;
C[15778] = -3.161916846061e-05;
C[15779] = 1.005336674055e-05;
C[15780] = 3.141192452130e-02;
C[15781] = 2.125281345514e-04;
C[15782] = 3.418957769157e-04;
C[15783] = 1.288675999732e-01;
C[15784] = -2.271178233157e-04;
C[15785] = -5.428347576066e-04;
C[15786] = 5.096039186601e-01;
C[15787] = 5.812641990504e-04;
C[15788] = 2.897981999851e-03;
C[15789] = 3.735985247515e-02;
C[15790] = 4.569557290894e-05;
C[15791] = 1.930219190107e-04;
C[15792] = -6.497806980989e-04;
C[15793] = 5.571213355429e-02;
C[15794] = 3.256075017760e-03;
C[15795] = 3.253621371359e-03;
C[15796] = 3.834793946138e-01;
C[15797] = 6.520554414716e-03;
C[15798] = -3.070345294833e-03;
C[15799] = 4.834208911035e-02;
C[15800] = -1.353156556180e-02;
C[15801] = -3.382138124104e-05;
C[15802] = 2.719130099013e-03;
C[15803] = 1.688919595894e-04;
C[15804] = 2.134766918229e-04;
C[15805] = 3.145689921089e-02;
C[15806] = 7.178501185212e-04;
C[15807] = -2.252821296281e-04;
C[15808] = 1.284268626309e-01;
C[15809] = -1.235551943834e-03;
C[15810] = 5.717801181526e-04;
C[15811] = 5.096027865551e-01;
C[15812] = 3.570194221540e-03;
C[15813] = 4.512526278905e-05;
C[15814] = 3.740776208979e-02;
C[15815] = 3.434059758514e-04;
C[15816] = 1.522046209022e-04;
C[15817] = 3.445476902490e-03;
C[15818] = 5.235255692101e-02;
C[15819] = 4.702669101112e-03;
C[15820] = 6.267439267122e-03;
C[15821] = 4.000466903718e-01;
C[15822] = -7.751185850808e-03;
C[15823] = -1.355122940048e-02;
C[15824] = 1.870086143470e-02;
C[15825] = 7.899361215433e-06;
C[15826] = 1.805366503427e-04;
C[15827] = 2.543831881099e-03;
C[15828] = 3.403432436624e-04;
C[15829] = 7.044038829178e-04;
C[15830] = 3.284487824923e-02;
C[15831] = -5.386154194436e-04;
C[15832] = -1.236862866413e-03;
C[15833] = 1.259100426074e-01;
C[15834] = 2.895099081207e-03;
C[15835] = 3.622645100188e-03;
C[15836] = 5.264234247821e-01;
C[15837] = 1.924728150507e-04;
C[15838] = 3.459683791158e-04;
C[15839] = 3.871651661075e-02;
C[15840] = -7.030132489491e+00;
C[15841] = 6.278890554224e-02;
C[15842] = -2.394420920935e-02;
C[15843] = 2.110991230861e+00;
C[15844] = 1.594647253762e-02;
C[15845] = -2.411815758355e-02;
C[15846] = 7.826112476807e-01;
C[15847] = 1.581428359559e-02;
C[15848] = 1.184275260939e-02;
C[15849] = 7.325887040271e-02;
C[15850] = 8.040366453077e-03;
C[15851] = -7.087928520546e-03;
C[15852] = 3.732877700859e-01;
C[15853] = 4.449720486278e-03;
C[15854] = -4.474527005680e-04;
C[15855] = 5.039240599242e-02;
C[15856] = 1.027638646317e-03;
C[15857] = 7.952994440324e-04;
C[15858] = 7.710887386805e-02;
C[15859] = 1.779877728097e-03;
C[15860] = 2.062682016040e-03;
C[15861] = 2.378455619654e-03;
C[15862] = 5.686272306779e-05;
C[15863] = 7.212797996603e-05;
C[15864] = 5.784477391086e-02;
C[15865] = -7.192446300273e+00;
C[15866] = 2.484319101798e-02;
C[15867] = 1.815871111513e-02;
C[15868] = 2.139139958197e+00;
C[15869] = 2.365564687419e-02;
C[15870] = 1.647696734724e-02;
C[15871] = 8.578809862332e-01;
C[15872] = 4.753791046733e-02;
C[15873] = 7.793159871508e-03;
C[15874] = 2.643232977595e-02;
C[15875] = 1.840296468966e-03;
C[15876] = 4.517783569891e-03;
C[15877] = 4.007107952289e-01;
C[15878] = 9.879422906878e-03;
C[15879] = 1.057442750823e-03;
C[15880] = 5.586253417485e-02;
C[15881] = 3.334450135518e-03;
C[15882] = 1.858695087287e-03;
C[15883] = 8.665189013280e-02;
C[15884] = 6.834660726254e-03;
C[15885] = 5.847363822969e-05;
C[15886] = 2.741892122320e-03;
C[15887] = 2.432450828885e-04;
C[15888] = -2.417972346684e-02;
C[15889] = 2.196826303816e-02;
C[15890] = -7.034653218152e+00;
C[15891] = -2.368269229051e-02;
C[15892] = 3.033526929173e-02;
C[15893] = 2.075873951933e+00;
C[15894] = 1.176649159530e-02;
C[15895] = 4.560470425693e-02;
C[15896] = 8.233607182864e-01;
C[15897] = -7.099704233420e-03;
C[15898] = 1.696550069975e-03;
C[15899] = 6.939562160140e-02;
C[15900] = -4.216843808221e-04;
C[15901] = 1.033817929310e-02;
C[15902] = 3.731063527556e-01;
C[15903] = 7.810676314508e-04;
C[15904] = 3.162077791156e-03;
C[15905] = 5.341479525338e-02;
C[15906] = 2.007010557074e-03;
C[15907] = 6.181545400994e-03;
C[15908] = 8.625997255715e-02;
C[15909] = 6.971222079769e-05;
C[15910] = 2.156777489161e-04;
C[15911] = 2.712044507837e-03;
C[15912] = 2.101099661687e+00;
C[15913] = 1.092650562085e-02;
C[15914] = -1.155879784545e-02;
C[15915] = -6.195634776073e+00;
C[15916] = 2.571765493599e-03;
C[15917] = -8.973550106598e-03;
C[15918] = 2.152215210869e+00;
C[15919] = -5.453816475260e-03;
C[15920] = -1.764249854490e-03;
C[15921] = 3.727931916272e-01;
C[15922] = 4.156173295177e-03;
C[15923] = 1.845103414309e-04;
C[15924] = 1.272358180885e-01;
C[15925] = -4.369017665938e-03;
C[15926] = -3.697161463924e-03;
C[15927] = 3.854851069215e-01;
C[15928] = 6.523247000827e-04;
C[15929] = 1.986579115169e-03;
C[15930] = 9.961055139572e-01;
C[15931] = 4.773872243025e-03;
C[15932] = 1.490465911529e-02;
C[15933] = 5.886698102866e-02;
C[15934] = 3.092307780504e-04;
C[15935] = 8.716480410389e-04;
C[15936] = 1.876213459946e-02;
C[15937] = 2.119599202075e+00;
C[15938] = 1.375070224122e-02;
C[15939] = -1.794834872089e-03;
C[15940] = -6.289184253461e+00;
C[15941] = -4.605434495686e-02;
C[15942] = -6.930723708679e-03;
C[15943] = 2.087000013419e+00;
C[15944] = -1.777264958603e-02;
C[15945] = 4.590503589370e-03;
C[15946] = 3.997337574228e-01;
C[15947] = 9.508950940570e-03;
C[15948] = -4.587347684222e-03;
C[15949] = 9.198892060912e-02;
C[15950] = -1.918991838033e-02;
C[15951] = 5.841970118318e-04;
C[15952] = 3.847919238319e-01;
C[15953] = 6.635795085500e-03;
C[15954] = 4.814450581385e-03;
C[15955] = 1.002686290242e+00;
C[15956] = 3.601522698112e-02;
C[15957] = 3.152278381588e-04;
C[15958] = 5.955984629649e-02;
C[15959] = 2.396647231827e-03;
C[15960] = -8.836492915251e-03;
C[15961] = 7.955491054203e-03;
C[15962] = 2.073570602361e+00;
C[15963] = -1.054585258352e-02;
C[15964] = -4.559389769596e-02;
C[15965] = -6.246180562520e+00;
C[15966] = -2.598271962349e-03;
C[15967] = -1.425772897644e-02;
C[15968] = 2.140349576301e+00;
C[15969] = 3.166305328469e-04;
C[15970] = 9.094415115878e-03;
C[15971] = 3.729911852770e-01;
C[15972] = -3.775776587770e-03;
C[15973] = -1.916689601729e-02;
C[15974] = 1.070179647294e-01;
C[15975] = 1.960075556659e-03;
C[15976] = 6.885932690611e-03;
C[15977] = 3.995408463661e-01;
C[15978] = 1.498579984760e-02;
C[15979] = 3.593351904384e-02;
C[15980] = 1.089595252065e+00;
C[15981] = 8.755733501507e-04;
C[15982] = 2.377426137517e-03;
C[15983] = 6.475005572476e-02;
C[15984] = 7.679583401586e-01;
C[15985] = -8.217228947367e-03;
C[15986] = 1.012590483266e-03;
C[15987] = 2.163283052567e+00;
C[15988] = 1.556005268264e-02;
C[15989] = 7.608617232815e-03;
C[15990] = -7.104182721587e+00;
C[15991] = -6.511421889660e-03;
C[15992] = -1.951500556103e-02;
C[15993] = 4.965976061632e-02;
C[15994] = -1.772670639073e-04;
C[15995] = 2.433725758492e-04;
C[15996] = 3.860384990064e-01;
C[15997] = 1.708735831398e-03;
C[15998] = 2.470420016417e-03;
C[15999] = 5.121474191432e-02;
C[16000] = -1.869392090373e-03;
C[16001] = -4.322122908800e-03;
C[16002] = 4.097502555069e+00;
C[16003] = 8.141680151126e-04;
C[16004] = 1.002103998293e-02;
C[16005] = 5.107228250005e-01;
C[16006] = 3.464000887114e-04;
C[16007] = 1.574314658145e-03;
C[16008] = -8.442120162238e-03;
C[16009] = 8.571595314720e-01;
C[16010] = 4.641491942946e-02;
C[16011] = 1.565491776649e-02;
C[16012] = 2.084436186572e+00;
C[16013] = -1.809327192216e-02;
C[16014] = -5.960713783399e-03;
C[16015] = -7.123808728415e+00;
C[16016] = -4.265947570352e-02;
C[16017] = -1.851815415740e-04;
C[16018] = 5.582646143679e-02;
C[16019] = 3.202588549782e-03;
C[16020] = 1.707761412170e-03;
C[16021] = 3.846637324895e-01;
C[16022] = 6.694155543325e-03;
C[16023] = -1.841856685060e-03;
C[16024] = 4.527458161271e-02;
C[16025] = -1.402999297467e-02;
C[16026] = 6.397413443810e-04;
C[16027] = 4.092328761771e+00;
C[16028] = 9.895757437950e-03;
C[16029] = 3.371823004776e-04;
C[16030] = 5.113633787065e-01;
C[16031] = 3.949814089448e-03;
C[16032] = 5.425129534767e-04;
C[16033] = 4.853846712964e-02;
C[16034] = 8.098675052746e-01;
C[16035] = 7.077488788250e-03;
C[16036] = -2.135445592155e-02;
C[16037] = 2.155311000410e+00;
C[16038] = -1.824920411215e-02;
C[16039] = -4.305275247717e-02;
C[16040] = -7.222548823792e+00;
C[16041] = 2.302874612368e-04;
C[16042] = 3.384477968634e-03;
C[16043] = 5.274013460280e-02;
C[16044] = 2.428666047306e-03;
C[16045] = 6.456704768724e-03;
C[16046] = 4.002889175716e-01;
C[16047] = -4.258832836356e-03;
C[16048] = -1.404965681336e-02;
C[16049] = 1.799673579347e-02;
C[16050] = 9.856722443364e-03;
C[16051] = 1.072216024110e-02;
C[16052] = 4.177541314587e+00;
C[16053] = 1.566448710992e-03;
C[16054] = 3.998138763122e-03;
C[16055] = 5.264427830411e-01;
C[16056] = 0.000000000000e+00;
C[16057] = 0.000000000000e+00;
C[16058] = 0.000000000000e+00;
C[16059] = 0.000000000000e+00;
C[16060] = 0.000000000000e+00;
C[16061] = 0.000000000000e+00;
C[16062] = 0.000000000000e+00;
C[16063] = 0.000000000000e+00;
C[16064] = 0.000000000000e+00;
C[16065] = 0.000000000000e+00;
C[16066] = 0.000000000000e+00;
C[16067] = 0.000000000000e+00;
C[16068] = 0.000000000000e+00;
C[16069] = 0.000000000000e+00;
C[16070] = 0.000000000000e+00;
C[16071] = 0.000000000000e+00;
C[16072] = 0.000000000000e+00;
C[16073] = 0.000000000000e+00;
C[16074] = 1.000000000000e+00;
C[16075] = 0.000000000000e+00;
C[16076] = 0.000000000000e+00;
C[16077] = 2.000000000000e-01;
C[16078] = -0.000000000000e+00;
C[16079] = -0.000000000000e+00;
C[16080] = 0.000000000000e+00;
C[16081] = 0.000000000000e+00;
C[16082] = 0.000000000000e+00;
C[16083] = 0.000000000000e+00;
C[16084] = 0.000000000000e+00;
C[16085] = 0.000000000000e+00;
C[16086] = 0.000000000000e+00;
C[16087] = 0.000000000000e+00;
C[16088] = 0.000000000000e+00;
C[16089] = 0.000000000000e+00;
C[16090] = 0.000000000000e+00;
C[16091] = 0.000000000000e+00;
C[16092] = 0.000000000000e+00;
C[16093] = 0.000000000000e+00;
C[16094] = 0.000000000000e+00;
C[16095] = 0.000000000000e+00;
C[16096] = 0.000000000000e+00;
C[16097] = 0.000000000000e+00;
C[16098] = 0.000000000000e+00;
C[16099] = 1.000000000000e+00;
C[16100] = 0.000000000000e+00;
C[16101] = -0.000000000000e+00;
C[16102] = 2.000000000000e-01;
C[16103] = -0.000000000000e+00;
C[16104] = 0.000000000000e+00;
C[16105] = 0.000000000000e+00;
C[16106] = 0.000000000000e+00;
C[16107] = 0.000000000000e+00;
C[16108] = 0.000000000000e+00;
C[16109] = 0.000000000000e+00;
C[16110] = 0.000000000000e+00;
C[16111] = 0.000000000000e+00;
C[16112] = 0.000000000000e+00;
C[16113] = 0.000000000000e+00;
C[16114] = 0.000000000000e+00;
C[16115] = 0.000000000000e+00;
C[16116] = 0.000000000000e+00;
C[16117] = 0.000000000000e+00;
C[16118] = 0.000000000000e+00;
C[16119] = 0.000000000000e+00;
C[16120] = 0.000000000000e+00;
C[16121] = 0.000000000000e+00;
C[16122] = 0.000000000000e+00;
C[16123] = 0.000000000000e+00;
C[16124] = 1.000000000000e+00;
C[16125] = -0.000000000000e+00;
C[16126] = -0.000000000000e+00;
C[16127] = 2.000000000000e-01;
C[16128] = 7.498614526621e-02;
C[16129] = 7.685291347266e-04;
C[16130] = -8.244033447004e-03;
C[16131] = 3.790616344324e-01;
C[16132] = 4.211285984953e-03;
C[16133] = 1.387836731581e-03;
C[16134] = 5.123198599913e-02;
C[16135] = 5.056661784551e-04;
C[16136] = 3.349760918191e-04;
C[16137] = 1.308623081239e-01;
C[16138] = -9.379937240223e-05;
C[16139] = -3.316922046632e-04;
C[16140] = 3.068911218835e-02;
C[16141] = 2.894891280755e-04;
C[16142] = 1.280974861213e-04;
C[16143] = 2.469090873534e-03;
C[16144] = 2.076939398209e-05;
C[16145] = 1.659574971459e-05;
C[16146] = 2.428108650796e-03;
C[16147] = 1.489587999453e-05;
C[16148] = 2.438006158753e-05;
C[16149] = -2.022158349175e-05;
C[16150] = -8.112039905590e-07;
C[16151] = 4.035335911841e-07;
C[16152] = 7.666767209531e-04;
C[16153] = 4.225198015891e-02;
C[16154] = 6.238509481872e-03;
C[16155] = 4.157188116984e-03;
C[16156] = 4.019598911976e-01;
C[16157] = 9.523924844243e-03;
C[16158] = 4.933766135380e-04;
C[16159] = 5.635331251619e-02;
C[16160] = 3.377993797800e-03;
C[16161] = -9.392286665379e-05;
C[16162] = 1.280698725960e-01;
C[16163] = 7.845205960918e-05;
C[16164] = 2.865389421781e-04;
C[16165] = 3.295711802584e-02;
C[16166] = 9.234113044490e-04;
C[16167] = 2.012507817156e-05;
C[16168] = 2.746184480425e-03;
C[16169] = 1.789451533957e-04;
C[16170] = 1.455047935022e-05;
C[16171] = 2.782102991553e-03;
C[16172] = 2.570143582768e-04;
C[16173] = -7.665214159106e-07;
C[16174] = -1.592590795821e-05;
C[16175] = 2.011586304743e-06;
C[16176] = -8.131086186794e-03;
C[16177] = 6.237793410110e-03;
C[16178] = 4.867758763641e-02;
C[16179] = 1.397026837657e-03;
C[16180] = 9.678872938713e-03;
C[16181] = 3.757279112230e-01;
C[16182] = 3.109878562188e-04;
C[16183] = 3.217255982072e-03;
C[16184] = 5.408599006423e-02;
C[16185] = -3.241623873158e-04;
C[16186] = 7.840432149171e-05;
C[16187] = 1.283142025543e-01;
C[16188] = 1.281190816198e-04;
C[16189] = 9.289834805316e-04;
C[16190] = 3.069583096976e-02;
C[16191] = 1.525587993016e-05;
C[16192] = 1.697424217958e-04;
C[16193] = 2.633849733637e-03;
C[16194] = 2.233822320293e-05;
C[16195] = 2.354758844387e-04;
C[16196] = 2.761191679579e-03;
C[16197] = 4.185271111039e-07;
C[16198] = 1.649717542247e-06;
C[16199] = -1.546120931114e-05;
C[16200] = 3.789133970135e-01;
C[16201] = 4.880626019186e-03;
C[16202] = 5.817011084561e-04;
C[16203] = 1.202325297036e-01;
C[16204] = -4.338815405410e-03;
C[16205] = -2.094156313511e-03;
C[16206] = 3.861464179872e-01;
C[16207] = -2.224443862485e-04;
C[16208] = 2.142439071758e-04;
C[16209] = 3.067922969376e-02;
C[16210] = 3.347681356582e-04;
C[16211] = 7.376403300639e-05;
C[16212] = 1.327188596821e-01;
C[16213] = -2.900602145105e-04;
C[16214] = -1.118006570857e-04;
C[16215] = 3.160461343648e-02;
C[16216] = -2.671406833569e-05;
C[16217] = -1.592053835001e-05;
C[16218] = 5.982689739737e-02;
C[16219] = -3.724157550550e-05;
C[16220] = -1.408922463085e-06;
C[16221] = 2.618503571161e-03;
C[16222] = -2.380657753298e-06;
C[16223] = -2.863081374319e-06;
C[16224] = 4.845891063668e-03;
C[16225] = 4.010633518125e-01;
C[16226] = 1.063090207196e-02;
C[16227] = -4.271057829638e-03;
C[16228] = 9.048874680847e-02;
C[16229] = -1.861889494434e-02;
C[16230] = -2.219197275388e-04;
C[16231] = 3.868319619831e-01;
C[16232] = 6.288181873144e-03;
C[16233] = 3.317961333232e-04;
C[16234] = 3.289734873350e-02;
C[16235] = 9.924520894147e-04;
C[16236] = -2.855430427924e-04;
C[16237] = 1.299025807430e-01;
C[16238] = -1.750736337995e-03;
C[16239] = -2.686217828632e-05;
C[16240] = 3.185908527732e-02;
C[16241] = 6.612171888238e-04;
C[16242] = -3.475498909137e-05;
C[16243] = 6.085688382418e-02;
C[16244] = 2.304314760168e-03;
C[16245] = -2.281372162210e-06;
C[16246] = 2.678395931852e-03;
C[16247] = 1.121815734743e-04;
C[16248] = 5.361648164961e-04;
C[16249] = 1.037099954345e-02;
C[16250] = 3.769122828636e-01;
C[16251] = -2.047562448980e-03;
C[16252] = -1.863575129612e-02;
C[16253] = 1.012679801469e-01;
C[16254] = 2.169594435368e-04;
C[16255] = 6.493601592762e-03;
C[16256] = 3.986851678692e-01;
C[16257] = 7.131935844897e-05;
C[16258] = 9.798829510627e-04;
C[16259] = 3.077478907914e-02;
C[16260] = -1.086943994503e-04;
C[16261] = -1.751860094781e-03;
C[16262] = 1.310664008829e-01;
C[16263] = -1.616817830151e-05;
C[16264] = 6.726889013478e-04;
C[16265] = 3.276300637210e-02;
C[16266] = 4.434317958312e-06;
C[16267] = 2.291129159886e-03;
C[16268] = 6.464769860137e-02;
C[16269] = -2.579427365692e-06;
C[16270] = 1.108279801110e-04;
C[16271] = 2.839321460476e-03;
C[16272] = 5.179288331759e-02;
C[16273] = 7.058150325806e-04;
C[16274] = 1.610869309617e-04;
C[16275] = 3.858717640765e-01;
C[16276] = -4.018029033873e-04;
C[16277] = 2.335334670358e-04;
C[16278] = 4.819700778397e-02;
C[16279] = -4.405655559984e-04;
C[16280] = -8.714482204122e-04;
C[16281] = 2.506484028098e-03;
C[16282] = 3.428763944107e-05;
C[16283] = 5.262484913017e-06;
C[16284] = 3.158630317576e-02;
C[16285] = -3.885439000956e-05;
C[16286] = -1.506324340158e-05;
C[16287] = 1.283897086846e-01;
C[16288] = 6.472873025336e-06;
C[16289] = 7.479775156296e-06;
C[16290] = 5.117715701846e-01;
C[16291] = 1.554632313227e-04;
C[16292] = 3.997096728696e-04;
C[16293] = 3.753512310465e-02;
C[16294] = 7.062429837264e-07;
C[16295] = -1.253921675765e-06;
C[16296] = 7.253510947132e-04;
C[16297] = 5.669188431046e-02;
C[16298] = 3.096150814146e-03;
C[16299] = -4.334064719899e-04;
C[16300] = 3.866937254654e-01;
C[16301] = 6.601090261468e-03;
C[16302] = -4.247595775624e-04;
C[16303] = 4.109048842364e-02;
C[16304] = -1.402658680385e-02;
C[16305] = 3.541505506596e-05;
C[16306] = 2.768755933377e-03;
C[16307] = 1.616687439340e-04;
C[16308] = -4.077820738512e-05;
C[16309] = 3.184986950947e-02;
C[16310] = 6.798548125949e-04;
C[16311] = 7.526604921070e-06;
C[16312] = 1.277428978237e-01;
C[16313] = -1.223497421101e-03;
C[16314] = 1.518245898834e-04;
C[16315] = 5.127773160393e-01;
C[16316] = 4.288780795230e-03;
C[16317] = 4.597101679326e-07;
C[16318] = 3.764977996331e-02;
C[16319] = 3.841650509154e-04;
C[16320] = 1.845307428745e-04;
C[16321] = 3.219582923874e-03;
C[16322] = 5.426508825388e-02;
C[16323] = 1.828220124300e-04;
C[16324] = 6.452981857917e-03;
C[16325] = 3.986484777342e-01;
C[16326] = -8.398448481443e-04;
C[16327] = -1.404836418139e-02;
C[16328] = 1.805865122026e-02;
C[16329] = 6.566059784954e-06;
C[16330] = 1.683844284672e-04;
C[16331] = 2.645789612947e-03;
C[16332] = -1.801533133307e-05;
C[16333] = 6.722038544754e-04;
C[16334] = 3.276056036311e-02;
C[16335] = 9.586666640823e-06;
C[16336] = -1.224949246270e-03;
C[16337] = 1.259072334412e-01;
C[16338] = 3.908370162705e-04;
C[16339] = 4.324682139605e-03;
C[16340] = 5.263809749165e-01;
C[16341] = -1.894946692685e-06;
C[16342] = 3.859634326009e-04;
C[16343] = 3.870656042958e-02;
C[16344] = -7.001963272552e+00;
C[16345] = 7.608507254365e-03;
C[16346] = -1.792403286239e-02;
C[16347] = 2.135510958179e+00;
C[16348] = 5.800424783273e-03;
C[16349] = 5.858737933903e-03;
C[16350] = 7.941735993309e-01;
C[16351] = 5.786529000346e-03;
C[16352] = 4.091366788797e-03;
C[16353] = 7.546246570807e-02;
C[16354] = -2.137900237071e-03;
C[16355] = 4.529029163187e-04;
C[16356] = 3.770385940737e-01;
C[16357] = 2.176726588347e-03;
C[16358] = 1.280804881410e-03;
C[16359] = 5.128170386643e-02;
C[16360] = 3.687186066292e-04;
C[16361] = 2.195482387444e-04;
C[16362] = 7.920545985170e-02;
C[16363] = 6.102164099437e-04;
C[16364] = 4.087948842574e-04;
C[16365] = 2.460131326795e-03;
C[16366] = 1.930933429963e-05;
C[16367] = 1.019822463518e-05;
C[16368] = 7.571458978897e-03;
C[16369] = -7.084108380128e+00;
C[16370] = 4.773677017238e-02;
C[16371] = 4.888452912125e-03;
C[16372] = 2.172116366995e+00;
C[16373] = 1.758463570371e-02;
C[16374] = 5.657295654975e-03;
C[16375] = 8.673425931093e-01;
C[16376] = 4.898982253719e-02;
C[16377] = -2.139752650845e-03;
C[16378] = 4.687157975235e-02;
C[16379] = -1.487255635450e-03;
C[16380] = 2.140633554787e-03;
C[16381] = 4.016913806572e-01;
C[16382] = 1.039896852472e-02;
C[16383] = 3.570907999058e-04;
C[16384] = 5.650607389770e-02;
C[16385] = 3.414093814617e-03;
C[16386] = 5.934409050880e-04;
C[16387] = 8.856350354967e-02;
C[16388] = 7.083465638751e-03;
C[16389] = 1.870834176591e-05;
C[16390] = 2.807838188964e-03;
C[16391] = 2.505743577329e-04;
C[16392] = -1.566508765819e-02;
C[16393] = 4.772244873714e-02;
C[16394] = -7.163175108464e+00;
C[16395] = 5.740733664131e-03;
C[16396] = 1.907835688232e-02;
C[16397] = 2.041529322353e+00;
C[16398] = 3.820311536955e-03;
C[16399] = 4.691520994641e-02;
C[16400] = 8.311081712203e-01;
C[16401] = 5.658501765288e-04;
C[16402] = -1.487971707212e-03;
C[16403] = 4.006145621826e-02;
C[16404] = 1.276295129280e-03;
C[16405] = 1.049003017516e-02;
C[16406] = 3.746854535426e-01;
C[16407] = 1.988502030956e-04;
C[16408] = 3.264904727629e-03;
C[16409] = 5.405791541974e-02;
C[16410] = 3.628178659057e-04;
C[16411] = 6.574337962022e-03;
C[16412] = 8.754209237772e-02;
C[16413] = 8.720858742041e-06;
C[16414] = 2.316801560439e-04;
C[16415] = 2.769558887484e-03;
C[16416] = 2.132546209802e+00;
C[16417] = 1.935721095618e-02;
C[16418] = -1.056578091989e-02;
C[16419] = -6.222943390575e+00;
C[16420] = -7.389452668889e-03;
C[16421] = -3.468766827789e-03;
C[16422] = 2.145885573881e+00;
C[16423] = -5.582878518657e-03;
C[16424] = -2.930830103306e-03;
C[16425] = 3.768903566548e-01;
C[16426] = 2.864071456989e-03;
C[16427] = 4.609694000795e-04;
C[16428] = 1.192817712332e-01;
C[16429] = -1.932433204135e-03;
C[16430] = -3.575766618785e-04;
C[16431] = 3.889133376975e-01;
C[16432] = -7.543495769563e-04;
C[16433] = -7.385173126451e-04;
C[16434] = 1.014040977304e+00;
C[16435] = -1.553944982434e-03;
C[16436] = -1.248204331239e-03;
C[16437] = 6.003134308107e-02;
C[16438] = -1.007936706338e-04;
C[16439] = -1.204504812222e-04;
C[16440] = 1.849252635758e-02;
C[16441] = 2.154185579294e+00;
C[16442] = 3.811893954726e-02;
C[16443] = -6.034301153442e-03;
C[16444] = -6.293925016151e+00;
C[16445] = -4.091615062981e-02;
C[16446] = -5.687055065588e-03;
C[16447] = 2.090727597060e+00;
C[16448] = -2.265973955391e-02;
C[16449] = 2.811331667062e-03;
C[16450] = 4.007948412722e-01;
C[16451] = 1.144205930840e-02;
C[16452] = -1.864675628363e-03;
C[16453] = 8.796786697454e-02;
C[16454] = -1.912919397481e-02;
C[16455] = -7.544219011763e-04;
C[16456] = 3.896855045280e-01;
C[16457] = 5.480445927996e-03;
C[16458] = -1.495381299682e-03;
C[16459] = 1.027294738102e+00;
C[16460] = 3.429813740063e-02;
C[16461] = -9.813123816565e-05;
C[16462] = 6.108681521016e-02;
C[16463] = 2.272655512410e-03;
C[16464] = -1.117470036780e-02;
C[16465] = 3.452612968782e-02;
C[16466] = 2.065216755164e+00;
C[16467] = -2.536889537161e-03;
C[16468] = -4.125327766544e-02;
C[16469] = -6.271807197422e+00;
C[16470] = -3.108592856475e-03;
C[16471] = -1.981725758319e-02;
C[16472] = 2.129076423095e+00;
C[16473] = 4.291329663250e-04;
C[16474] = 1.124604322393e-02;
C[16475] = 3.758698251832e-01;
C[16476] = -3.109827973471e-04;
C[16477] = -1.914605032660e-02;
C[16478] = 1.019395576831e-01;
C[16479] = -7.420378588324e-04;
C[16480] = 5.674855484883e-03;
C[16481] = 3.985011577342e-01;
C[16482] = -1.111982976127e-03;
C[16483] = 3.426802651962e-02;
C[16484] = 1.087195402662e+00;
C[16485] = -1.137745570097e-04;
C[16486] = 2.257804000029e-03;
C[16487] = 6.462157670780e-02;
C[16488] = 8.053915457001e-01;
C[16489] = 9.906064035829e-03;
C[16490] = 8.222930318127e-04;
C[16491] = 2.140392495667e+00;
C[16492] = -9.284718582559e-03;
C[16493] = -2.777112386495e-03;
C[16494] = -7.125408425142e+00;
C[16495] = -7.351832276142e-04;
C[16496] = -3.389238217334e-04;
C[16497] = 5.184260118488e-02;
C[16498] = 5.695292189485e-04;
C[16499] = 4.894927783851e-05;
C[16500] = 3.886386837868e-01;
C[16501] = -9.343050770249e-04;
C[16502] = -7.254638353334e-04;
C[16503] = 4.456993756545e-02;
C[16504] = 5.262794608095e-04;
C[16505] = 1.008983259512e-03;
C[16506] = 4.103660489281e+00;
C[16507] = 6.318699188646e-04;
C[16508] = 4.341954476672e-04;
C[16509] = 5.126086061883e-01;
C[16510] = -1.221183937159e-04;
C[16511] = -4.166502478856e-04;
C[16512] = 1.018022732551e-02;
C[16513] = 8.741140289947e-01;
C[16514] = 4.449310658789e-02;
C[16515] = -9.802120233486e-03;
C[16516] = 2.087962866704e+00;
C[16517] = -1.766748420908e-02;
C[16518] = -4.190636588939e-04;
C[16519] = -7.149924159573e+00;
C[16520] = -4.319231595585e-02;
C[16521] = 5.884035228872e-04;
C[16522] = 5.684464569197e-02;
C[16523] = 3.143799559703e-03;
C[16524] = -9.653116626978e-04;
C[16525] = 3.895472680102e-01;
C[16526] = 5.782344153589e-03;
C[16527] = 5.420854392455e-04;
C[16528] = 3.710585741158e-02;
C[16529] = -1.333010847440e-02;
C[16530] = 5.610750430502e-04;
C[16531] = 4.101617728746e+00;
C[16532] = 1.501474656736e-02;
C[16533] = -1.256948752033e-04;
C[16534] = 5.137631246215e-01;
C[16535] = 4.356666502719e-03;
C[16536] = 1.082459809905e-03;
C[16537] = 4.582160505867e-02;
C[16538] = 8.346901350133e-01;
C[16539] = -3.559267998223e-03;
C[16540] = -1.936373985844e-02;
C[16541] = 2.128342620395e+00;
C[16542] = 2.931436236247e-04;
C[16543] = -4.362786350660e-02;
C[16544] = -7.224567178807e+00;
C[16545] = 6.910288979978e-05;
C[16546] = 3.255682940691e-03;
C[16547] = 5.423701360939e-02;
C[16548] = -7.699392073909e-04;
C[16549] = 5.645245912770e-03;
C[16550] = 3.984644675992e-01;
C[16551] = 1.040586631780e-03;
C[16552] = -1.335188585194e-02;
C[16553] = 1.844820568410e-02;
C[16554] = 2.491103602878e-04;
C[16555] = 1.559506587791e-02;
C[16556] = 4.177100912436e+00;
C[16557] = -4.267351397084e-04;
C[16558] = 4.392454084266e-03;
C[16559] = 5.262187734683e-01;
C[16560] = 0.000000000000e+00;
C[16561] = 0.000000000000e+00;
C[16562] = 0.000000000000e+00;
C[16563] = 0.000000000000e+00;
C[16564] = 0.000000000000e+00;
C[16565] = 0.000000000000e+00;
C[16566] = 0.000000000000e+00;
C[16567] = 0.000000000000e+00;
C[16568] = 0.000000000000e+00;
C[16569] = 0.000000000000e+00;
C[16570] = 0.000000000000e+00;
C[16571] = 0.000000000000e+00;
C[16572] = 0.000000000000e+00;
C[16573] = 0.000000000000e+00;
C[16574] = 0.000000000000e+00;
C[16575] = 0.000000000000e+00;
C[16576] = 0.000000000000e+00;
C[16577] = 0.000000000000e+00;
C[16578] = 1.000000000000e+00;
C[16579] = 0.000000000000e+00;
C[16580] = 0.000000000000e+00;
C[16581] = 2.000000000000e-01;
C[16582] = -0.000000000000e+00;
C[16583] = -0.000000000000e+00;
C[16584] = 0.000000000000e+00;
C[16585] = 0.000000000000e+00;
C[16586] = 0.000000000000e+00;
C[16587] = 0.000000000000e+00;
C[16588] = 0.000000000000e+00;
C[16589] = 0.000000000000e+00;
C[16590] = 0.000000000000e+00;
C[16591] = 0.000000000000e+00;
C[16592] = 0.000000000000e+00;
C[16593] = 0.000000000000e+00;
C[16594] = 0.000000000000e+00;
C[16595] = 0.000000000000e+00;
C[16596] = 0.000000000000e+00;
C[16597] = 0.000000000000e+00;
C[16598] = 0.000000000000e+00;
C[16599] = 0.000000000000e+00;
C[16600] = 0.000000000000e+00;
C[16601] = 0.000000000000e+00;
C[16602] = 0.000000000000e+00;
C[16603] = 1.000000000000e+00;
C[16604] = 0.000000000000e+00;
C[16605] = -0.000000000000e+00;
C[16606] = 2.000000000000e-01;
C[16607] = -0.000000000000e+00;
C[16608] = 0.000000000000e+00;
C[16609] = 0.000000000000e+00;
C[16610] = 0.000000000000e+00;
C[16611] = 0.000000000000e+00;
C[16612] = 0.000000000000e+00;
C[16613] = 0.000000000000e+00;
C[16614] = 0.000000000000e+00;
C[16615] = 0.000000000000e+00;
C[16616] = 0.000000000000e+00;
C[16617] = 0.000000000000e+00;
C[16618] = 0.000000000000e+00;
C[16619] = 0.000000000000e+00;
C[16620] = 0.000000000000e+00;
C[16621] = 0.000000000000e+00;
C[16622] = 0.000000000000e+00;
C[16623] = 0.000000000000e+00;
C[16624] = 0.000000000000e+00;
C[16625] = 0.000000000000e+00;
C[16626] = 0.000000000000e+00;
C[16627] = 0.000000000000e+00;
C[16628] = 1.000000000000e+00;
C[16629] = -0.000000000000e+00;
C[16630] = -0.000000000000e+00;
C[16631] = 2.000000000000e-01;
C[16632] = 7.782356913958e-02;
C[16633] = 5.248079031010e-04;
C[16634] = 1.001314342647e-02;
C[16635] = 3.652951178591e-01;
C[16636] = -2.368465378603e-03;
C[16637] = 8.003970084419e-04;
C[16638] = 4.830312650618e-02;
C[16639] = -1.002421296205e-03;
C[16640] = -2.465773281290e-04;
C[16641] = 1.318272686994e-01;
C[16642] = 2.444160465916e-04;
C[16643] = 9.396594148287e-04;
C[16644] = 2.883200895052e-02;
C[16645] = -4.160482452452e-04;
C[16646] = -1.478886068597e-05;
C[16647] = 2.281299717632e-03;
C[16648] = -6.357817041679e-05;
C[16649] = -1.837327083284e-05;
C[16650] = 2.152423998571e-03;
C[16651] = -9.974551152654e-05;
C[16652] = -4.037124678328e-05;
C[16653] = -2.902045253664e-05;
C[16654] = -2.765872393661e-06;
C[16655] = -1.100743184202e-06;
C[16656] = 6.281149109733e-04;
C[16657] = 4.526925788405e-02;
C[16658] = -8.144634968122e-03;
C[16659] = -2.490199743039e-03;
C[16660] = 4.002508701001e-01;
C[16661] = 1.210383459094e-02;
C[16662] = -1.018374408565e-03;
C[16663] = 5.731426091382e-02;
C[16664] = 3.272124598860e-03;
C[16665] = 2.513031804497e-04;
C[16666] = 1.279984180680e-01;
C[16667] = -7.159028757222e-04;
C[16668] = -4.236504753987e-04;
C[16669] = 3.293847461534e-02;
C[16670] = 1.057164096861e-03;
C[16671] = -6.422373397220e-05;
C[16672] = 2.801296650664e-03;
C[16673] = 1.687738720345e-04;
C[16674] = -1.004884099637e-04;
C[16675] = 2.873951223327e-03;
C[16676] = 2.363969222838e-04;
C[16677] = -2.699357203376e-06;
C[16678] = -1.433292164247e-05;
C[16679] = 9.503336950766e-07;
C[16680] = 1.010661002808e-02;
C[16681] = -8.375807695728e-03;
C[16682] = 4.906836711464e-02;
C[16683] = 7.272015710915e-04;
C[16684] = 1.250281492611e-02;
C[16685] = 3.658722837623e-01;
C[16686] = -2.367986268850e-04;
C[16687] = 3.104159588586e-03;
C[16688] = 5.015482981163e-02;
C[16689] = 9.458905216028e-04;
C[16690] = -7.313143908959e-04;
C[16691] = 1.298357688922e-01;
C[16692] = -1.975297561281e-05;
C[16693] = 1.083717804389e-03;
C[16694] = 2.892782083189e-02;
C[16695] = -1.733867113641e-05;
C[16696] = 1.575097745012e-04;
C[16697] = 2.379415275172e-03;
C[16698] = -3.623198120423e-05;
C[16699] = 2.045195958349e-04;
C[16700] = 2.349048100123e-03;
C[16701] = -8.952746562345e-07;
C[16702] = 8.270615599277e-08;
C[16703] = -2.818935702607e-05;
C[16704] = 3.658487271350e-01;
C[16705] = -2.213560571084e-03;
C[16706] = 5.190211371395e-04;
C[16707] = 1.269703075296e-01;
C[16708] = 2.869607117019e-03;
C[16709] = 1.736052958361e-03;
C[16710] = 3.943241162229e-01;
C[16711] = 3.191978794906e-04;
C[16712] = -5.187980762987e-04;
C[16713] = 2.886891623558e-02;
C[16714] = -4.052078639350e-04;
C[16715] = -3.363167120961e-05;
C[16716] = 1.341077846679e-01;
C[16717] = 4.783357856194e-04;
C[16718] = 1.984531554657e-04;
C[16719] = 3.216822935095e-02;
C[16720] = -1.144011070484e-05;
C[16721] = -8.194891056514e-05;
C[16722] = 6.107220693677e-02;
C[16723] = -1.505301591423e-04;
C[16724] = -4.029647965169e-04;
C[16725] = 2.665155879545e-03;
C[16726] = -9.454661559757e-06;
C[16727] = -2.073749718106e-05;
C[16728] = -2.299908770297e-03;
C[16729] = 4.003329289468e-01;
C[16730] = 1.222424891255e-02;
C[16731] = 2.975229704192e-03;
C[16732] = 8.628098300248e-02;
C[16733] = -2.042947400003e-02;
C[16734] = 3.300772755524e-04;
C[16735] = 3.911588286795e-01;
C[16736] = 4.629886249881e-03;
C[16737] = -4.114778046915e-04;
C[16738] = 3.294394520512e-02;
C[16739] = 1.065146736819e-03;
C[16740] = 4.853772914309e-04;
C[16741] = 1.295285371916e-01;
C[16742] = -1.781249840410e-03;
C[16743] = -1.102759009710e-05;
C[16744] = 3.215483522327e-02;
C[16745] = 5.419715244069e-04;
C[16746] = -1.516303280416e-04;
C[16747] = 6.170112288563e-02;
C[16748] = 2.106864154838e-03;
C[16749] = -9.570065690092e-06;
C[16750] = 2.711022512776e-03;
C[16751] = 1.047304079484e-04;
C[16752] = 4.739036575981e-04;
C[16753] = 1.239075886198e-02;
C[16754] = 3.658940398520e-01;
C[16755] = 1.808158245345e-03;
C[16756] = -2.075485243449e-02;
C[16757] = 1.112403605416e-01;
C[16758] = -5.387351060052e-04;
C[16759] = 4.823713781583e-03;
C[16760] = 4.025706533560e-01;
C[16761] = -3.655508407556e-05;
C[16762] = 1.076292381598e-03;
C[16763] = 2.892927123787e-02;
C[16764] = 2.032601745979e-04;
C[16765] = -1.802941736041e-03;
C[16766] = 1.328484504594e-01;
C[16767] = -8.356450638402e-05;
C[16768] = 5.545178916444e-04;
C[16769] = 3.302862047358e-02;
C[16770] = -4.059421984451e-04;
C[16771] = 2.113954051386e-03;
C[16772] = 6.502162578714e-02;
C[16773] = -2.082841831913e-05;
C[16774] = 1.044493330388e-04;
C[16775] = 2.851687786684e-03;
C[16776] = 5.054696701441e-02;
C[16777] = -4.842343494773e-04;
C[16778] = -2.460108418230e-04;
C[16779] = 3.921419639682e-01;
C[16780] = -1.958844977535e-04;
C[16781] = -5.636908535703e-04;
C[16782] = 4.158154362009e-02;
C[16783] = 1.162329338782e-03;
C[16784] = 2.122459377206e-03;
C[16785] = 2.430889084847e-03;
C[16786] = -2.861439669967e-05;
C[16787] = -1.795281879895e-05;
C[16788] = 3.202275253397e-02;
C[16789] = -4.609170831749e-05;
C[16790] = -8.522822288836e-05;
C[16791] = 1.278921679941e-01;
C[16792] = 1.227753378120e-04;
C[16793] = 2.177544300265e-04;
C[16794] = 5.134241319270e-01;
C[16795] = -4.503428827665e-04;
C[16796] = -1.236176804976e-03;
C[16797] = 3.767603084997e-02;
C[16798] = -4.669853254990e-05;
C[16799] = -1.150353789241e-04;
C[16800] = -4.704811487826e-04;
C[16801] = 5.710923371876e-02;
C[16802] = 3.340924094665e-03;
C[16803] = -2.149040822312e-04;
C[16804] = 3.913030589940e-01;
C[16805] = 4.535610332065e-03;
C[16806] = 1.169071186759e-03;
C[16807] = 3.405260065856e-02;
C[16808] = -1.263511268386e-02;
C[16809] = -2.811549392196e-05;
C[16810] = 2.787628170993e-03;
C[16811] = 1.732940749064e-04;
C[16812] = -4.704690815297e-05;
C[16813] = 3.216445057757e-02;
C[16814] = 5.353109950098e-04;
C[16815] = 1.232247943437e-04;
C[16816] = 1.272445086384e-01;
C[16817] = -1.110232425587e-03;
C[16818] = -4.488783852874e-04;
C[16819] = 5.148107977615e-01;
C[16820] = 4.585098022203e-03;
C[16821] = -4.666445014767e-05;
C[16822] = 3.781807582201e-02;
C[16823] = 3.999315620137e-04;
C[16824] = -2.495643619964e-04;
C[16825] = 3.491069536807e-03;
C[16826] = 5.275271714309e-02;
C[16827] = -5.667061935087e-04;
C[16828] = 4.361441854102e-03;
C[16829] = 4.000293024302e-01;
C[16830] = 2.134665832334e-03;
C[16831] = -1.264137975941e-02;
C[16832] = 1.880773490776e-02;
C[16833] = -1.857240642401e-05;
C[16834] = 1.833702012309e-04;
C[16835] = 2.552607763936e-03;
C[16836] = -8.514278504581e-05;
C[16837] = 5.240752313550e-04;
C[16838] = 3.285919707853e-02;
C[16839] = 2.185681937016e-04;
C[16840] = -1.110650230624e-03;
C[16841] = 1.259509442242e-01;
C[16842] = -1.239255935229e-03;
C[16843] = 4.611440963243e-03;
C[16844] = 5.257780613650e-01;
C[16845] = -1.152614482732e-04;
C[16846] = 4.011697420457e-04;
C[16847] = 3.865447247370e-02;
C[16848] = -7.010012487884e+00;
C[16849] = -3.053495912897e-03;
C[16850] = 6.016101041440e-02;
C[16851] = 2.049749807878e+00;
C[16852] = -1.261761886130e-02;
C[16853] = 2.390664466272e-02;
C[16854] = 7.390975961922e-01;
C[16855] = -1.784954470272e-02;
C[16856] = -2.863827933379e-03;
C[16857] = 9.138763898881e-02;
C[16858] = 4.194382303829e-03;
C[16859] = 1.282193443160e-02;
C[16860] = 3.481795847684e-01;
C[16861] = -6.770821914286e-03;
C[16862] = 3.887709717388e-04;
C[16863] = 4.734398776202e-02;
C[16864] = -1.281798376505e-03;
C[16865] = -3.128056287090e-04;
C[16866] = 7.288125619568e-02;
C[16867] = -2.311103408512e-03;
C[16868] = -9.645675178406e-04;
C[16869] = 2.220849143982e-03;
C[16870] = -8.620413831633e-05;
C[16871] = -4.406178743171e-05;
C[16872] = -9.873557554512e-04;
C[16873] = -7.086614353529e+00;
C[16874] = 1.987399682596e-03;
C[16875] = -1.492391107061e-02;
C[16876] = 2.153106926168e+00;
C[16877] = 1.580477802826e-02;
C[16878] = -1.803381186320e-02;
C[16879] = 8.820391486379e-01;
C[16880] = 4.459142287995e-02;
C[16881] = 4.297689311701e-03;
C[16882] = 4.225814657436e-02;
C[16883] = -6.117986810018e-03;
C[16884] = -6.885524112325e-03;
C[16885] = 4.020864369002e-01;
C[16886] = 1.018874063852e-02;
C[16887] = -1.290307708730e-03;
C[16888] = 5.744593961029e-02;
C[16889] = 3.149096013141e-03;
C[16890] = -2.334848203016e-03;
C[16891] = 9.054331939028e-02;
C[16892] = 6.723046922756e-03;
C[16893] = -8.640249476139e-05;
C[16894] = 2.869708991088e-03;
C[16895] = 2.415604368703e-04;
C[16896] = 6.203034244663e-02;
C[16897] = -2.636054869521e-03;
C[16898] = -7.138638975615e+00;
C[16899] = 2.249031688278e-02;
C[16900] = 2.339703191199e-02;
C[16901] = 1.999311445210e+00;
C[16902] = -2.668786552188e-03;
C[16903] = 4.206176660122e-02;
C[16904] = 7.594578903527e-01;
C[16905] = 1.291540103321e-02;
C[16906] = -6.349159537623e-03;
C[16907] = 7.173774568921e-02;
C[16908] = 3.076998975204e-04;
C[16909] = 1.063848795129e-02;
C[16910] = 3.459788145862e-01;
C[16911] = -2.909646376350e-04;
C[16912] = 2.929075176878e-03;
C[16913] = 4.892996415914e-02;
C[16914] = -8.780419840110e-04;
C[16915] = 5.941359244050e-03;
C[16916] = 7.835632383970e-02;
C[16917] = -3.976692126552e-05;
C[16918] = 2.088648409181e-04;
C[16919] = 2.425073773310e-03;
C[16920] = 2.060821993396e+00;
C[16921] = -9.391127631513e-03;
C[16922] = 1.832670820374e-02;
C[16923] = -6.180102656195e+00;
C[16924] = 1.149856207576e-02;
C[16925] = 6.277207072467e-03;
C[16926] = 2.213795703892e+00;
C[16927] = 1.578179140610e-02;
C[16928] = 4.210366708140e-03;
C[16929] = 3.487331940443e-01;
C[16930] = -6.608884940370e-03;
C[16931] = 9.951946356842e-05;
C[16932] = 1.431785420812e-01;
C[16933] = 7.498886929476e-03;
C[16934] = 2.675525718112e-03;
C[16935] = 3.958994669707e-01;
C[16936] = 3.260853462455e-04;
C[16937] = -8.554583935315e-04;
C[16938] = 1.034363087300e+00;
C[16939] = -2.073483443957e-03;
C[16940] = -6.643802306444e-03;
C[16941] = 6.109339039956e-02;
C[16942] = -1.942282695875e-04;
C[16943] = -4.504926610761e-04;
C[16944] = -1.124648669520e-02;
C[16945] = 2.154748103103e+00;
C[16946] = 1.782571164082e-02;
C[16947] = 1.361101381921e-02;
C[16948] = -6.302813451702e+00;
C[16949] = -4.395376231788e-02;
C[16950] = 1.584094420932e-02;
C[16951] = 2.088957769100e+00;
C[16952] = -2.554405371294e-02;
C[16953] = -6.702265305980e-03;
C[16954] = 4.021684957469e-01;
C[16955] = 1.035992193773e-02;
C[16956] = 7.604509516648e-03;
C[16957] = 8.410219738250e-02;
C[16958] = -1.823423310752e-02;
C[16959] = 3.301132786915e-04;
C[16960] = 3.912498038308e-01;
C[16961] = 4.417125858802e-03;
C[16962] = -2.077507480559e-03;
C[16963] = 1.038398820962e+00;
C[16964] = 3.130588729612e-02;
C[16965] = -1.956798017912e-04;
C[16966] = 6.177606682448e-02;
C[16967] = 2.082995213071e-03;
C[16968] = 1.737677764584e-02;
C[16969] = 2.154326344917e-02;
C[16970] = 1.999746567004e+00;
C[16971] = 7.719312812132e-03;
C[16972] = -5.046133100708e-02;
C[16973] = -6.220225038808e+00;
C[16974] = 3.783339865173e-03;
C[16975] = -2.254717953147e-02;
C[16976] = 2.202853142205e+00;
C[16977] = 6.227762089492e-05;
C[16978] = 1.047566490957e-02;
C[16979] = 3.460005706759e-01;
C[16980] = 2.747631005095e-03;
C[16981] = -1.855961154198e-02;
C[16982] = 1.310469793916e-01;
C[16983] = -8.886509465187e-04;
C[16984] = 4.667464266033e-03;
C[16985] = 4.040115152873e-01;
C[16986] = -6.708918127186e-03;
C[16987] = 3.172972488988e-02;
C[16988] = 1.095961239192e+00;
C[16989] = -4.534083512574e-04;
C[16990] = 2.093864482032e-03;
C[16991] = 6.493799526295e-02;
C[16992] = 7.839744063568e-01;
C[16993] = -7.351010681446e-03;
C[16994] = -2.853030850948e-03;
C[16995] = 2.170152658799e+00;
C[16996] = 5.321708743201e-03;
C[16997] = 3.284224913872e-03;
C[16998] = -7.137359613002e+00;
C[16999] = 5.476230382759e-03;
C[17000] = 9.225885849957e-03;
C[17001] = 4.958782827025e-02;
C[17002] = -7.561676496418e-04;
C[17003] = -3.001768525731e-04;
C[17004] = 3.937173147161e-01;
C[17005] = -1.958484946143e-04;
C[17006] = -9.136066940838e-04;
C[17007] = 4.043784206208e-02;
C[17008] = 1.660506725633e-03;
C[17009] = 2.851757098363e-03;
C[17010] = 4.110472444821e+00;
C[17011] = -2.133086806667e-03;
C[17012] = -7.404207852676e-03;
C[17013] = 5.141204120290e-01;
C[17014] = -6.338009068487e-04;
C[17015] = -1.565175835289e-03;
C[17016] = -7.210741754269e-03;
C[17017] = 8.779386047366e-01;
C[17018] = 4.679705672280e-02;
C[17019] = 5.099752171667e-03;
C[17020] = 2.091842375390e+00;
C[17021] = -2.830924852184e-02;
C[17022] = 5.611067342284e-03;
C[17023] = -7.167578155138e+00;
C[17024] = -4.150586276560e-02;
C[17025] = -7.498582290820e-04;
C[17026] = 5.724091241523e-02;
C[17027] = 3.165839682957e-03;
C[17028] = -2.080166154763e-04;
C[17029] = 3.913940341452e-01;
C[17030] = 4.379360816515e-03;
C[17031] = 1.667248573610e-03;
C[17032] = 3.299902843413e-02;
C[17033] = -1.233868237078e-02;
C[17034] = -2.117656971261e-03;
C[17035] = 4.111462096693e+00;
C[17036] = 1.870907754633e-02;
C[17037] = -6.325033089320e-04;
C[17038] = 5.156247893168e-01;
C[17039] = 4.641757273143e-03;
C[17040] = -2.923568610728e-03;
C[17041] = 4.897032163888e-02;
C[17042] = 8.114156369819e-01;
C[17043] = 3.252204363941e-03;
C[17044] = -3.091294162851e-02;
C[17045] = 2.152026123689e+00;
C[17046] = 9.470014952512e-03;
C[17047] = -4.163120427664e-02;
C[17048] = -7.216747287695e+00;
C[17049] = -3.157926625764e-04;
C[17050] = 3.368040951088e-03;
C[17051] = 5.152785149060e-02;
C[17052] = -9.033665107414e-04;
C[17053] = 4.148681463023e-03;
C[17054] = 4.014701643615e-01;
C[17055] = 2.863963553490e-03;
C[17056] = -1.234494944633e-02;
C[17057] = 1.909466373046e-02;
C[17058] = -7.473705220399e-03;
C[17059] = 1.912332571844e-02;
C[17060] = 4.172382767559e+00;
C[17061] = -1.568552286139e-03;
C[17062] = 4.665970852663e-03;
C[17063] = 5.254831971892e-01;
C[17064] = 0.000000000000e+00;
C[17065] = 0.000000000000e+00;
C[17066] = 0.000000000000e+00;
C[17067] = 0.000000000000e+00;
C[17068] = 0.000000000000e+00;
C[17069] = 0.000000000000e+00;
C[17070] = 0.000000000000e+00;
C[17071] = 0.000000000000e+00;
C[17072] = 0.000000000000e+00;
C[17073] = 0.000000000000e+00;
C[17074] = 0.000000000000e+00;
C[17075] = 0.000000000000e+00;
C[17076] = 0.000000000000e+00;
C[17077] = 0.000000000000e+00;
C[17078] = 0.000000000000e+00;
C[17079] = 0.000000000000e+00;
C[17080] = 0.000000000000e+00;
C[17081] = 0.000000000000e+00;
C[17082] = 1.000000000000e+00;
C[17083] = 0.000000000000e+00;
C[17084] = 0.000000000000e+00;
C[17085] = 2.000000000000e-01;
C[17086] = -0.000000000000e+00;
C[17087] = -0.000000000000e+00;
C[17088] = 0.000000000000e+00;
C[17089] = 0.000000000000e+00;
C[17090] = 0.000000000000e+00;
C[17091] = 0.000000000000e+00;
C[17092] = 0.000000000000e+00;
C[17093] = 0.000000000000e+00;
C[17094] = 0.000000000000e+00;
C[17095] = 0.000000000000e+00;
C[17096] = 0.000000000000e+00;
C[17097] = 0.000000000000e+00;
C[17098] = 0.000000000000e+00;
C[17099] = 0.000000000000e+00;
C[17100] = 0.000000000000e+00;
C[17101] = 0.000000000000e+00;
C[17102] = 0.000000000000e+00;
C[17103] = 0.000000000000e+00;
C[17104] = 0.000000000000e+00;
C[17105] = 0.000000000000e+00;
C[17106] = 0.000000000000e+00;
C[17107] = 1.000000000000e+00;
C[17108] = 0.000000000000e+00;
C[17109] = -0.000000000000e+00;
C[17110] = 2.000000000000e-01;
C[17111] = -0.000000000000e+00;
C[17112] = 0.000000000000e+00;
C[17113] = 0.000000000000e+00;
C[17114] = 0.000000000000e+00;
C[17115] = 0.000000000000e+00;
C[17116] = 0.000000000000e+00;
C[17117] = 0.000000000000e+00;
C[17118] = 0.000000000000e+00;
C[17119] = 0.000000000000e+00;
C[17120] = 0.000000000000e+00;
C[17121] = 0.000000000000e+00;
C[17122] = 0.000000000000e+00;
C[17123] = 0.000000000000e+00;
C[17124] = 0.000000000000e+00;
C[17125] = 0.000000000000e+00;
C[17126] = 0.000000000000e+00;
C[17127] = 0.000000000000e+00;
C[17128] = 0.000000000000e+00;
C[17129] = 0.000000000000e+00;
C[17130] = 0.000000000000e+00;
C[17131] = 0.000000000000e+00;
C[17132] = 1.000000000000e+00;
C[17133] = -0.000000000000e+00;
C[17134] = -0.000000000000e+00;
C[17135] = 2.000000000000e-01;
C[17136] = 1.022297027746e-01;
C[17137] = -1.273714146764e-02;
C[17138] = 1.299525473920e-02;
C[17139] = 3.394315291382e-01;
C[17140] = 7.711821798908e-03;
C[17141] = 1.759044027556e-03;
C[17142] = 4.863550183592e-02;
C[17143] = 2.378800641377e-03;
C[17144] = -4.806647879602e-05;
C[17145] = 1.325701976852e-01;
C[17146] = -1.699639835406e-03;
C[17147] = 9.888485147025e-04;
C[17148] = 2.831376449633e-02;
C[17149] = 1.217202697484e-03;
C[17150] = 1.258407341912e-04;
C[17151] = 2.395767629352e-03;
C[17152] = 1.515388305632e-04;
C[17153] = -4.511423503563e-06;
C[17154] = 2.495134698554e-03;
C[17155] = 2.287245527156e-04;
C[17156] = -1.742736233494e-05;
C[17157] = -7.705329927812e-06;
C[17158] = 7.076618285198e-06;
C[17159] = -2.985566039676e-07;
C[17160] = -1.350260199211e-02;
C[17161] = 4.124998067313e-02;
C[17162] = 7.899574104618e-03;
C[17163] = 8.689613455085e-03;
C[17164] = 4.041466049314e-01;
C[17165] = 1.016397186415e-04;
C[17166] = 2.395976830211e-03;
C[17167] = 5.657299076264e-02;
C[17168] = 1.557200653290e-03;
C[17169] = -1.750670537037e-03;
C[17170] = 1.281254968335e-01;
C[17171] = 7.455097653898e-04;
C[17172] = 1.277940513627e-03;
C[17173] = 3.285784421154e-02;
C[17174] = 1.004553914821e-04;
C[17175] = 1.506518591404e-04;
C[17176] = 2.736685317799e-03;
C[17177] = 8.633848708562e-05;
C[17178] = 2.218013614216e-04;
C[17179] = 2.758411475826e-03;
C[17180] = 1.499190495955e-04;
C[17181] = 6.691008352293e-06;
C[17182] = -1.983234500604e-05;
C[17183] = 1.532474048224e-06;
C[17184] = 1.304094400656e-02;
C[17185] = 7.913167092249e-03;
C[17186] = 8.812604029697e-02;
C[17187] = 1.852825976059e-03;
C[17188] = 5.740739824557e-04;
C[17189] = 3.342386303835e-01;
C[17190] = -9.527040778194e-05;
C[17191] = 1.391505610555e-03;
C[17192] = 4.900162122900e-02;
C[17193] = 9.918944658600e-04;
C[17194] = 7.464159645653e-04;
C[17195] = 1.317001834127e-01;
C[17196] = 1.321790971626e-04;
C[17197] = 1.230534353713e-04;
C[17198] = 2.765497175525e-02;
C[17199] = -7.914728112659e-06;
C[17200] = 7.747939772628e-05;
C[17201] = 2.408748544544e-03;
C[17202] = -2.644248507625e-05;
C[17203] = 1.344942020492e-04;
C[17204] = 2.568100197418e-03;
C[17205] = -6.019677054287e-07;
C[17206] = 1.552890646079e-06;
C[17207] = -9.049874374905e-06;
C[17208] = 3.388028755379e-01;
C[17209] = 8.068979990505e-03;
C[17210] = 2.107316232997e-03;
C[17211] = 1.540968125096e-01;
C[17212] = -7.143417577720e-03;
C[17213] = 2.254609293646e-03;
C[17214] = 3.934455281032e-01;
C[17215] = -3.806631193257e-03;
C[17216] = -1.960517362927e-03;
C[17217] = 2.827185425631e-02;
C[17218] = 1.236564949321e-03;
C[17219] = 1.491451142918e-04;
C[17220] = 1.348819361747e-01;
C[17221] = -1.109319384778e-03;
C[17222] = 2.017524011371e-04;
C[17223] = 3.185985860159e-02;
C[17224] = -2.964777789498e-04;
C[17225] = -2.358959433608e-04;
C[17226] = 6.039086798234e-02;
C[17227] = -6.877373475491e-04;
C[17228] = -8.428993018811e-04;
C[17229] = 2.617660922612e-03;
C[17230] = -2.901067359060e-05;
C[17231] = -4.249208595813e-05;
C[17232] = 8.606136664142e-03;
C[17233] = 4.045726752600e-01;
C[17234] = 4.643350569047e-05;
C[17235] = -7.854128528583e-03;
C[17236] = 8.379732160907e-02;
C[17237] = -7.702215692885e-03;
C[17238] = -3.816961798910e-03;
C[17239] = 3.902260728761e-01;
C[17240] = 6.284567423062e-03;
C[17241] = 1.276823688499e-03;
C[17242] = 3.288624890011e-02;
C[17243] = 8.787740358697e-05;
C[17244] = -1.156700114836e-03;
C[17245] = 1.298914701905e-01;
C[17246] = -8.246209603973e-04;
C[17247] = -2.958831783329e-04;
C[17248] = 3.192568488466e-02;
C[17249] = 6.615306274544e-04;
C[17250] = -6.818489612861e-04;
C[17251] = 6.132172875139e-02;
C[17252] = 2.326570868335e-03;
C[17253] = -2.868442719428e-05;
C[17254] = 2.682820897305e-03;
C[17255] = 1.129728068259e-04;
C[17256] = 1.970041636835e-03;
C[17257] = -5.743228176643e-04;
C[17258] = 3.336937321911e-01;
C[17259] = 2.243939890721e-03;
C[17260] = -7.504182475126e-03;
C[17261] = 1.437174966835e-01;
C[17262] = -1.916501093170e-03;
C[17263] = 6.485179876848e-03;
C[17264] = 4.042015287425e-01;
C[17265] = 1.399072414764e-04;
C[17266] = 5.539122239501e-05;
C[17267] = 2.761864520909e-02;
C[17268] = 2.010411076088e-04;
C[17269] = -8.114187458800e-04;
C[17270] = 1.341613237074e-01;
C[17271] = -2.328232519659e-04;
C[17272] = 6.706830687470e-04;
C[17273] = 3.302178930193e-02;
C[17274] = -8.300765874111e-04;
C[17275] = 2.329437928878e-03;
C[17276] = 6.472593144826e-02;
C[17277] = -4.195596194326e-05;
C[17278] = 1.122972131350e-04;
C[17279] = 2.828168443764e-03;
C[17280] = 4.533942223523e-02;
C[17281] = 1.010754474105e-03;
C[17282] = -2.948787686895e-04;
C[17283] = 3.962781057158e-01;
C[17284] = -2.456468933784e-03;
C[17285] = -1.839937507816e-03;
C[17286] = 4.095703809490e-02;
C[17287] = 2.022456671429e-03;
C[17288] = 3.731662206187e-03;
C[17289] = 2.176028989306e-03;
C[17290] = 5.830370206664e-05;
C[17291] = -2.122195217317e-05;
C[17292] = 3.204869710909e-02;
C[17293] = -2.051836539912e-04;
C[17294] = -2.277190129423e-04;
C[17295] = 1.280072174781e-01;
C[17296] = 2.079109592298e-04;
C[17297] = 3.814754346958e-04;
C[17298] = 5.153222063413e-01;
C[17299] = -6.695870340374e-04;
C[17300] = -1.488376783212e-03;
C[17301] = 3.778444603800e-02;
C[17302] = -6.144894115817e-05;
C[17303] = -1.315253007328e-04;
C[17304] = 1.083341154845e-03;
C[17305] = 5.711791991789e-02;
C[17306] = 1.108982381428e-03;
C[17307] = -2.547291813361e-03;
C[17308] = 3.892460545490e-01;
C[17309] = 6.786682238639e-03;
C[17310] = 2.023745491982e-03;
C[17311] = 3.373144217183e-02;
C[17312] = -1.250543944427e-02;
C[17313] = 6.517486479446e-05;
C[17314] = 2.773013928149e-03;
C[17315] = 5.864451578446e-05;
C[17316] = -2.125218202901e-04;
C[17317] = 3.186035032952e-02;
C[17318] = 6.907832261998e-04;
C[17319] = 2.079968806000e-04;
C[17320] = 1.274246789728e-01;
C[17321] = -1.154109321328e-03;
C[17322] = -6.702929392637e-04;
C[17323] = 5.172129286619e-01;
C[17324] = 4.574230282116e-03;
C[17325] = -6.153266557077e-05;
C[17326] = 3.795877303336e-02;
C[17327] = 4.008511580153e-04;
C[17328] = -2.655412489341e-04;
C[17329] = 1.342929874493e-03;
C[17330] = 4.686551683988e-02;
C[17331] = -1.876349430767e-03;
C[17332] = 6.500143791809e-03;
C[17333] = 4.062786908730e-01;
C[17334] = 3.739552519429e-03;
C[17335] = -1.248675782118e-02;
C[17336] = 1.946496983689e-02;
C[17337] = -1.900974151277e-05;
C[17338] = 7.205376849912e-05;
C[17339] = 2.266341585269e-03;
C[17340] = -2.302847478834e-04;
C[17341] = 6.759023853709e-04;
C[17342] = 3.316026677730e-02;
C[17343] = 3.820014555786e-04;
C[17344] = -1.152863879789e-03;
C[17345] = 1.259793833336e-01;
C[17346] = -1.492254603071e-03;
C[17347] = 4.585303776724e-03;
C[17348] = 5.250897316080e-01;
C[17349] = -1.317828521259e-04;
C[17350] = 4.012386590652e-04;
C[17351] = 3.860910124755e-02;
C[17352] = -6.978537059349e+00;
C[17353] = -5.473404209214e-02;
C[17354] = 7.147185788945e-02;
C[17355] = 2.049193507261e+00;
C[17356] = 1.660319644496e-02;
C[17357] = 2.782129490440e-02;
C[17358] = 7.752658832217e-01;
C[17359] = 4.169448702680e-02;
C[17360] = 1.263555347860e-03;
C[17361] = 8.393844697338e-02;
C[17362] = -2.515974488811e-02;
C[17363] = 1.156173583660e-02;
C[17364] = 3.632471100724e-01;
C[17365] = 1.803721686884e-02;
C[17366] = 2.682856186321e-03;
C[17367] = 4.986554976567e-02;
C[17368] = 2.999788609498e-03;
C[17369] = -3.041655813057e-05;
C[17370] = 7.738352146803e-02;
C[17371] = 5.047384468590e-03;
C[17372] = -5.149266975760e-04;
C[17373] = 2.384396487699e-03;
C[17374] = 1.924266024086e-04;
C[17375] = -2.838637408453e-05;
C[17376] = -7.004325258164e-02;
C[17377] = -7.080691862763e+00;
C[17378] = 5.219745411560e-02;
C[17379] = 3.500642470536e-02;
C[17380] = 2.155726305888e+00;
C[17381] = 8.939927374082e-03;
C[17382] = 4.187316675526e-02;
C[17383] = 8.639350995339e-01;
C[17384] = 2.385275259421e-02;
C[17385] = -2.592520541258e-02;
C[17386] = 4.921383181175e-02;
C[17387] = 1.049498247343e-02;
C[17388] = 1.899608064295e-02;
C[17389] = 3.968673898014e-01;
C[17390] = 2.278142423837e-03;
C[17391] = 2.946864017579e-03;
C[17392] = 5.593803194736e-02;
C[17393] = 1.688329001770e-03;
C[17394] = 4.796155884281e-03;
C[17395] = 8.797225079238e-02;
C[17396] = 4.119827983915e-03;
C[17397] = 1.770575337633e-04;
C[17398] = 2.753605853067e-03;
C[17399] = 1.454383708508e-04;
C[17400] = 7.238564323672e-02;
C[17401] = 5.246931386824e-02;
C[17402] = -7.063037737100e+00;
C[17403] = 2.947071521345e-02;
C[17404] = 1.559697377253e-02;
C[17405] = 1.976272038864e+00;
C[17406] = 5.744013136643e-04;
C[17407] = 2.248563364849e-02;
C[17408] = 7.728051745407e-01;
C[17409] = 1.160742510397e-02;
C[17410] = 1.050857546106e-02;
C[17411] = 7.500844548367e-02;
C[17412] = 2.840300400402e-03;
C[17413] = 2.750912817551e-03;
C[17414] = 3.504306412483e-01;
C[17415] = -9.126685386284e-05;
C[17416] = 1.562950177494e-03;
C[17417] = 4.998165983488e-02;
C[17418] = -7.095018589370e-04;
C[17419] = 3.694652670873e-03;
C[17420] = 8.021069245954e-02;
C[17421] = -3.660420167526e-05;
C[17422] = 1.276344393236e-04;
C[17423] = 2.504348878592e-03;
C[17424] = 2.036620435255e+00;
C[17425] = 2.259375541375e-02;
C[17426] = 3.456052035222e-02;
C[17427] = -6.152801041305e+00;
C[17428] = -2.811930475637e-02;
C[17429] = 8.610788513002e-03;
C[17430] = 2.172051563783e+00;
C[17431] = -4.680114339256e-02;
C[17432] = -3.832648706345e-03;
C[17433] = 3.626184564722e-01;
C[17434] = 1.837544717837e-02;
C[17435] = 3.094790657340e-03;
C[17436] = 1.333551748984e-01;
C[17437] = -1.715449408522e-02;
C[17438] = 2.350934828988e-03;
C[17439] = 3.894992740516e-01;
C[17440] = -4.823220818611e-03;
C[17441] = -2.984855440658e-03;
C[17442] = 1.021560846914e+00;
C[17443] = -1.293181053563e-02;
C[17444] = -1.421750285897e-02;
C[17445] = 6.032536921640e-02;
C[17446] = -6.786996197763e-04;
C[17447] = -8.998564636700e-04;
C[17448] = 3.448949374964e-02;
C[17449] = 2.164247712460e+00;
C[17450] = 5.044164237228e-03;
C[17451] = -4.233352377363e-02;
C[17452] = -6.296364645428e+00;
C[17453] = -2.291636181374e-02;
C[17454] = -4.683330413656e-02;
C[17455] = 2.095212673431e+00;
C[17456] = -2.688346672950e-03;
C[17457] = 1.893153173407e-02;
C[17458] = 3.972934601300e-01;
C[17459] = 2.223272340786e-03;
C[17460] = -1.786520503609e-02;
C[17461] = 9.427443917338e-02;
C[17462] = -1.025194989672e-02;
C[17463] = -4.769424421810e-03;
C[17464] = 3.876417969368e-01;
C[17465] = 6.838680804078e-03;
C[17466] = -1.267479623520e-02;
C[17467] = 1.030085316626e+00;
C[17468] = 3.691726951386e-02;
C[17469] = -6.633521426660e-04;
C[17470] = 6.112695451340e-02;
C[17471] = 2.371482734426e-03;
C[17472] = 3.204124708999e-02;
C[17473] = -4.579323352033e-03;
C[17474] = 1.965374075017e+00;
C[17475] = 8.397400454509e-03;
C[17476] = -1.895569745856e-02;
C[17477] = -6.175547110761e+00;
C[17478] = -3.110688669268e-03;
C[17479] = -6.440573239618e-04;
C[17480] = 2.193598458427e+00;
C[17481] = 2.893853795600e-03;
C[17482] = 1.602179887531e-03;
C[17483] = 3.498857430560e-01;
C[17484] = 2.340265426063e-03;
C[17485] = -1.005391667896e-02;
C[17486] = 1.278058367469e-01;
C[17487] = -2.921821995647e-03;
C[17488] = 6.986045130425e-03;
C[17489] = 4.031665419014e-01;
C[17490] = -1.393513816858e-02;
C[17491] = 3.720286642364e-02;
C[17492] = 1.092733529149e+00;
C[17493] = -8.861632339545e-04;
C[17494] = 2.379089512061e-03;
C[17495] = 6.477749748497e-02;
C[17496] = 7.093442912079e-01;
C[17497] = 1.416871963314e-02;
C[17498] = -3.417765904487e-03;
C[17499] = 2.228703116035e+00;
C[17500] = -1.962344683405e-02;
C[17501] = -1.579416962174e-03;
C[17502] = -7.136961515341e+00;
C[17503] = 3.259521880704e-03;
C[17504] = 1.283095454281e-02;
C[17505] = 4.656947016498e-02;
C[17506] = 1.561641661473e-03;
C[17507] = -2.908752147704e-04;
C[17508] = 3.923318516642e-01;
C[17509] = -3.408931556685e-03;
C[17510] = -2.845258410292e-03;
C[17511] = 4.338987711178e-02;
C[17512] = 2.353493578732e-03;
C[17513] = 4.755873561664e-03;
C[17514] = 4.126841593033e+00;
C[17515] = -1.915898513465e-03;
C[17516] = -4.811139791690e-03;
C[17517] = 5.153668920577e-01;
C[17518] = -7.008923400855e-04;
C[17519] = -1.511642641295e-03;
C[17520] = 1.578529729617e-02;
C[17521] = 8.748336826389e-01;
C[17522] = 1.683516906595e-02;
C[17523] = -2.161435579463e-02;
C[17524] = 2.075612306889e+00;
C[17525] = 5.385989911862e-03;
C[17526] = 3.285298291773e-03;
C[17527] = -7.168186565571e+00;
C[17528] = -4.005277064893e-02;
C[17529] = 1.704329122966e-03;
C[17530] = 5.648296110261e-02;
C[17531] = 1.280426948367e-03;
C[17532] = -3.563881438715e-03;
C[17533] = 3.866617786097e-01;
C[17534] = 7.287547492216e-03;
C[17535] = 2.354782399285e-03;
C[17536] = 3.688303291267e-02;
C[17537] = -1.324070448452e-02;
C[17538] = -1.934649905507e-03;
C[17539] = 4.133156017830e+00;
C[17540] = 1.706319557185e-02;
C[17541] = -7.023077341437e-04;
C[17542] = 5.173403943016e-01;
C[17543] = 4.612231345550e-03;
C[17544] = -3.085940054901e-03;
C[17545] = 1.956733701826e-02;
C[17546] = 7.300830867583e-01;
C[17547] = -2.149290063137e-03;
C[17548] = 1.623180701974e-03;
C[17549] = 2.235141701037e+00;
C[17550] = 1.298876080766e-02;
C[17551] = -3.967913818707e-02;
C[17552] = -7.210126484570e+00;
C[17553] = -2.478913282686e-04;
C[17554] = 1.474058222973e-03;
C[17555] = 4.784555544576e-02;
C[17556] = -2.900687508498e-03;
C[17557] = 7.054257172825e-03;
C[17558] = 4.052437040319e-01;
C[17559] = 4.763763874907e-03;
C[17560] = -1.322202286143e-02;
C[17561] = 1.944759786316e-02;
C[17562] = -4.886073338694e-03;
C[17563] = 1.720928441236e-02;
C[17564] = 4.168358559581e+00;
C[17565] = -1.516919392591e-03;
C[17566] = 4.619308692782e-03;
C[17567] = 5.249933011764e-01;
C[17568] = 0.000000000000e+00;
C[17569] = 0.000000000000e+00;
C[17570] = 0.000000000000e+00;
C[17571] = 0.000000000000e+00;
C[17572] = 0.000000000000e+00;
C[17573] = 0.000000000000e+00;
C[17574] = 0.000000000000e+00;
C[17575] = 0.000000000000e+00;
C[17576] = 0.000000000000e+00;
C[17577] = 0.000000000000e+00;
C[17578] = 0.000000000000e+00;
C[17579] = 0.000000000000e+00;
C[17580] = 0.000000000000e+00;
C[17581] = 0.000000000000e+00;
C[17582] = 0.000000000000e+00;
C[17583] = 0.000000000000e+00;
C[17584] = 0.000000000000e+00;
C[17585] = 0.000000000000e+00;
C[17586] = 1.000000000000e+00;
C[17587] = 0.000000000000e+00;
C[17588] = 0.000000000000e+00;
C[17589] = 2.000000000000e-01;
C[17590] = -0.000000000000e+00;
C[17591] = -0.000000000000e+00;
C[17592] = 0.000000000000e+00;
C[17593] = 0.000000000000e+00;
C[17594] = 0.000000000000e+00;
C[17595] = 0.000000000000e+00;
C[17596] = 0.000000000000e+00;
C[17597] = 0.000000000000e+00;
C[17598] = 0.000000000000e+00;
C[17599] = 0.000000000000e+00;
C[17600] = 0.000000000000e+00;
C[17601] = 0.000000000000e+00;
C[17602] = 0.000000000000e+00;
C[17603] = 0.000000000000e+00;
C[17604] = 0.000000000000e+00;
C[17605] = 0.000000000000e+00;
C[17606] = 0.000000000000e+00;
C[17607] = 0.000000000000e+00;
C[17608] = 0.000000000000e+00;
C[17609] = 0.000000000000e+00;
C[17610] = 0.000000000000e+00;
C[17611] = 1.000000000000e+00;
C[17612] = 0.000000000000e+00;
C[17613] = -0.000000000000e+00;
C[17614] = 2.000000000000e-01;
C[17615] = -0.000000000000e+00;
C[17616] = 0.000000000000e+00;
C[17617] = 0.000000000000e+00;
C[17618] = 0.000000000000e+00;
C[17619] = 0.000000000000e+00;
C[17620] = 0.000000000000e+00;
C[17621] = 0.000000000000e+00;
C[17622] = 0.000000000000e+00;
C[17623] = 0.000000000000e+00;
C[17624] = 0.000000000000e+00;
C[17625] = 0.000000000000e+00;
C[17626] = 0.000000000000e+00;
C[17627] = 0.000000000000e+00;
C[17628] = 0.000000000000e+00;
C[17629] = 0.000000000000e+00;
C[17630] = 0.000000000000e+00;
C[17631] = 0.000000000000e+00;
C[17632] = 0.000000000000e+00;
C[17633] = 0.000000000000e+00;
C[17634] = 0.000000000000e+00;
C[17635] = 0.000000000000e+00;
C[17636] = 1.000000000000e+00;
C[17637] = -0.000000000000e+00;
C[17638] = -0.000000000000e+00;
C[17639] = 2.000000000000e-01;
C[17640] = 8.052602522086e-02;
C[17641] = -2.014682545268e-02;
C[17642] = 3.469446436711e-03;
C[17643] = 3.787983260832e-01;
C[17644] = 1.446545970271e-02;
C[17645] = 4.187177768296e-03;
C[17646] = 5.116879415509e-02;
C[17647] = 1.465923251663e-03;
C[17648] = 1.378290619910e-04;
C[17649] = 1.313287902581e-01;
C[17650] = -1.344088421950e-03;
C[17651] = 1.395035838808e-05;
C[17652] = 3.076977766610e-02;
C[17653] = 1.069259516789e-03;
C[17654] = 3.412977446248e-04;
C[17655] = 2.468949388846e-03;
C[17656] = 6.417595249050e-05;
C[17657] = 6.619415275056e-06;
C[17658] = 2.460370095149e-03;
C[17659] = 4.369104404351e-05;
C[17660] = -4.908728737965e-06;
C[17661] = -1.797531240066e-05;
C[17662] = -1.959383287031e-06;
C[17663] = 1.202838032521e-08;
C[17664] = -1.986045385516e-02;
C[17665] = 4.428577210173e-02;
C[17666] = 1.271432970875e-02;
C[17667] = 1.433852926097e-02;
C[17668] = 3.960792719415e-01;
C[17669] = 7.775992664929e-03;
C[17670] = 1.379734399377e-03;
C[17671] = 5.521853178386e-02;
C[17672] = 3.234100550668e-03;
C[17673] = -1.324996982116e-03;
C[17674] = 1.276706209818e-01;
C[17675] = 8.348512302789e-04;
C[17676] = 1.059792703433e-03;
C[17677] = 3.254132236206e-02;
C[17678] = 8.302472971007e-04;
C[17679] = 6.005241147242e-05;
C[17680] = 2.691789569333e-03;
C[17681] = 1.767023697301e-04;
C[17682] = 3.745116956478e-05;
C[17683] = 2.763188277349e-03;
C[17684] = 2.664096090411e-04;
C[17685] = -1.948005130589e-06;
C[17686] = -1.335133818690e-05;
C[17687] = 3.107921042431e-06;
C[17688] = 3.463812824616e-03;
C[17689] = 1.286300706469e-02;
C[17690] = 6.535410893864e-02;
C[17691] = 4.301470539222e-03;
C[17692] = 8.106058330921e-03;
C[17693] = 3.681415809996e-01;
C[17694] = 5.568407444700e-05;
C[17695] = 3.096173970187e-03;
C[17696] = 5.364429173745e-02;
C[17697] = 1.357478424840e-05;
C[17698] = 8.447630540081e-04;
C[17699] = 1.300235382745e-01;
C[17700] = 3.451072876094e-04;
C[17701] = 8.460148201165e-04;
C[17702] = 3.033249299885e-02;
C[17703] = 2.282293528335e-06;
C[17704] = 1.683133259368e-04;
C[17705] = 2.628882253070e-03;
C[17706] = -1.268085333817e-05;
C[17707] = 2.490292465583e-04;
C[17708] = 2.783976197254e-03;
C[17709] = -5.547921498050e-08;
C[17710] = 2.827707613879e-06;
C[17711] = -1.200530945998e-05;
C[17712] = 3.786759272226e-01;
C[17713] = 1.387515630094e-02;
C[17714] = 4.704733026361e-03;
C[17715] = 1.233224080442e-01;
C[17716] = -1.309142315173e-02;
C[17717] = -2.871152082032e-04;
C[17718] = 3.841024232123e-01;
C[17719] = -3.149094943107e-03;
C[17720] = -3.365411183190e-03;
C[17721] = 3.076161774205e-02;
C[17722] = 1.028901172764e-03;
C[17723] = 3.719914534186e-04;
C[17724] = 1.330532115665e-01;
C[17725] = -9.795874371986e-04;
C[17726] = -1.556251429896e-04;
C[17727] = 3.123275527557e-02;
C[17728] = -1.517471360145e-04;
C[17729] = -1.789101685147e-04;
C[17730] = 5.922012267193e-02;
C[17731] = -2.906934787782e-04;
C[17732] = -5.205305891210e-04;
C[17733] = 2.576040018830e-03;
C[17734] = -4.768295563715e-06;
C[17735] = -1.526725988555e-05;
C[17736] = 1.368621411353e-02;
C[17737] = 3.971084534254e-01;
C[17738] = 7.975064916559e-03;
C[17739] = -1.300556540835e-02;
C[17740] = 9.758582581267e-02;
C[17741] = -1.606835464403e-02;
C[17742] = -3.031706630221e-03;
C[17743] = 3.844632755995e-01;
C[17744] = 6.687950369499e-03;
C[17745] = 1.017309810844e-03;
C[17746] = 3.260993446099e-02;
C[17747] = 8.372819258257e-04;
C[17748] = -9.738635876395e-04;
C[17749] = 1.305866896793e-01;
C[17750] = -1.698709808379e-03;
C[17751] = -1.454073862335e-04;
C[17752] = 3.150304451212e-02;
C[17753] = 7.704904088632e-04;
C[17754] = -2.895923998939e-04;
C[17755] = 6.041037144554e-02;
C[17756] = 2.660744302654e-03;
C[17757] = -5.207631476301e-06;
C[17758] = 2.645880529367e-03;
C[17759] = 1.333803826983e-04;
C[17760] = 4.411092444827e-03;
C[17761] = 7.449143821512e-03;
C[17762] = 3.673508527141e-01;
C[17763] = -2.198905121743e-04;
C[17764] = -1.593781459469e-02;
C[17765] = 1.115389574062e-01;
C[17766] = -3.210497900602e-03;
C[17767] = 6.886697259612e-03;
C[17768] = 3.994239346690e-01;
C[17769] = 3.562253897269e-04;
C[17770] = 8.084573742062e-04;
C[17771] = 3.027977777981e-02;
C[17772] = -1.511434965877e-04;
C[17773] = -1.690007138423e-03;
C[17774] = 1.316227424641e-01;
C[17775] = -1.701200719322e-04;
C[17776] = 7.817717190746e-04;
C[17777] = 3.275443811787e-02;
C[17778] = -5.211481054894e-04;
C[17779] = 2.665209902101e-03;
C[17780] = 6.437364008398e-02;
C[17781] = -1.608715532064e-05;
C[17782] = 1.326744295721e-04;
C[17783] = 2.822331297810e-03;
C[17784] = 5.084099939109e-02;
C[17785] = 2.262001971138e-03;
C[17786] = 4.512682435001e-04;
C[17787] = 3.844623879680e-01;
C[17788] = -3.824004363329e-03;
C[17789] = -3.198507299526e-03;
C[17790] = 4.783809996680e-02;
C[17791] = 2.259437056667e-03;
C[17792] = 4.857991697664e-03;
C[17793] = 2.447096404580e-03;
C[17794] = 1.188702495898e-04;
C[17795] = 2.865457146521e-05;
C[17796] = 3.125675292596e-02;
C[17797] = -1.982272351074e-04;
C[17798] = -1.693206985271e-04;
C[17799] = 1.286982972611e-01;
C[17800] = 1.450225108925e-04;
C[17801] = 3.161815717139e-04;
C[17802] = 5.145806926881e-01;
C[17803] = -8.844487536039e-04;
C[17804] = -2.080278531665e-03;
C[17805] = 3.761553871667e-02;
C[17806] = -6.881102555416e-05;
C[17807] = -1.740323745334e-04;
C[17808] = 2.329141391340e-03;
C[17809] = 5.516090375144e-02;
C[17810] = 2.766036782769e-03;
C[17811] = -3.867170863768e-03;
C[17812] = 3.850190489218e-01;
C[17813] = 7.352013309867e-03;
C[17814] = 2.227613405663e-03;
C[17815] = 4.128892405506e-02;
C[17816] = -1.452456791300e-02;
C[17817] = 1.217238284689e-04;
C[17818] = 2.687947700505e-03;
C[17819] = 1.463041801090e-04;
C[17820] = -1.996188640586e-04;
C[17821] = 3.154009606694e-02;
C[17822] = 8.127927890916e-04;
C[17823] = 1.429009341589e-04;
C[17824] = 1.280595093078e-01;
C[17825] = -1.354069405629e-03;
C[17826] = -8.790309563251e-04;
C[17827] = 5.159171624976e-01;
C[17828] = 4.366653166591e-03;
C[17829] = -6.836691610793e-05;
C[17830] = 3.773191996082e-02;
C[17831] = 3.920397159695e-04;
C[17832] = 5.465275459097e-04;
C[17833] = 2.916604744241e-03;
C[17834] = 5.237679438498e-02;
C[17835] = -3.250468562140e-03;
C[17836] = 7.181843320707e-03;
C[17837] = 4.004077047976e-01;
C[17838] = 4.793544986539e-03;
C[17839] = -1.453262384469e-02;
C[17840] = 1.924259455534e-02;
C[17841] = 3.386598086964e-05;
C[17842] = 1.555359826350e-04;
C[17843] = 2.544382429572e-03;
C[17844] = -1.712473271114e-04;
C[17845] = 8.034166056104e-04;
C[17846] = 3.282002279311e-02;
C[17847] = 3.118851243056e-04;
C[17848] = -1.354606467741e-03;
C[17849] = 1.259917614830e-01;
C[17850] = -2.070137019039e-03;
C[17851] = 4.374835374749e-03;
C[17852] = 5.254155362987e-01;
C[17853] = -1.730945334919e-04;
C[17854] = 3.925894652723e-04;
C[17855] = 3.865589105784e-02;
C[17856] = -6.965882564251e+00;
C[17857] = -4.403099712198e-02;
C[17858] = 1.540965887221e-02;
C[17859] = 2.154755310733e+00;
C[17860] = 2.808844709888e-02;
C[17861] = 1.615208220880e-02;
C[17862] = 7.957840558820e-01;
C[17863] = 1.594003523786e-02;
C[17864] = 2.168090187177e-03;
C[17865] = 8.165760899053e-02;
C[17866] = -9.574885595288e-03;
C[17867] = -2.165006548167e-03;
C[17868] = 3.802067816674e-01;
C[17869] = 8.532545243269e-03;
C[17870] = 3.790220830297e-03;
C[17871] = 5.115319909264e-02;
C[17872] = 1.168004193896e-03;
C[17873] = 2.753314172948e-04;
C[17874] = 7.918786823419e-02;
C[17875] = 1.881171309413e-03;
C[17876] = 3.035820861416e-04;
C[17877] = 2.443052656062e-03;
C[17878] = 7.428967893300e-05;
C[17879] = 2.576758836410e-05;
C[17880] = -3.830356517171e-02;
C[17881] = -7.123376100046e+00;
C[17882] = 8.920413656125e-02;
C[17883] = 2.490894441858e-02;
C[17884] = 2.135233202627e+00;
C[17885] = 3.773620794144e-02;
C[17886] = 1.487431483802e-02;
C[17887] = 8.514526895350e-01;
C[17888] = 4.995055633963e-02;
C[17889] = -9.288513997774e-03;
C[17890] = 3.423490753036e-02;
C[17891] = 9.979196452888e-03;
C[17892] = 8.363463448923e-03;
C[17893] = 3.974391159695e-01;
C[17894] = 1.123575085871e-02;
C[17895] = 1.107024009859e-03;
C[17896] = 5.536156289433e-02;
C[17897] = 3.524151997767e-03;
C[17898] = 1.730957428879e-03;
C[17899] = 8.733651936799e-02;
C[17900] = 7.367259466284e-03;
C[17901] = 6.984274252818e-05;
C[17902] = 2.768826017782e-03;
C[17903] = 2.648183231929e-04;
C[17904] = 1.529698663030e-02;
C[17905] = 9.217768368001e-02;
C[17906] = -7.070558771682e+00;
C[17907] = 1.705232514032e-02;
C[17908] = 4.185965584192e-02;
C[17909] = 2.079446670889e+00;
C[17910] = 1.334723846213e-03;
C[17911] = 4.840231698119e-02;
C[17912] = 8.341544074382e-01;
C[17913] = -2.170640160262e-03;
C[17914] = 1.012787380883e-02;
C[17915] = 6.374607177201e-02;
C[17916] = 3.871917718033e-03;
C[17917] = 1.156121732275e-02;
C[17918] = 3.768744809414e-01;
C[17919] = 2.108047755516e-04;
C[17920] = 3.386867951490e-03;
C[17921] = 5.412114803567e-02;
C[17922] = 9.588172740138e-05;
C[17923] = 6.926332106426e-03;
C[17924] = 8.716921186450e-02;
C[17925] = 1.694747517587e-05;
C[17926] = 2.462107464163e-04;
C[17927] = 2.739918414831e-03;
C[17928] = 2.152307333521e+00;
C[17929] = 1.564148521802e-02;
C[17930] = 2.511757488309e-02;
C[17931] = -6.210872165546e+00;
C[17932] = -2.717644955503e-02;
C[17933] = -2.424078293250e-03;
C[17934] = 2.129765869200e+00;
C[17935] = -1.997492629989e-02;
C[17936] = -9.097273265688e-03;
C[17937] = 3.800843828067e-01;
C[17938] = 7.900090488895e-03;
C[17939] = 4.275180205171e-03;
C[17940] = 1.236836789167e-01;
C[17941] = -8.283469648112e-03;
C[17942] = -2.964584639129e-03;
C[17943] = 3.826570618311e-01;
C[17944] = -1.456868170729e-03;
C[17945] = -7.809628644666e-04;
C[17946] = 1.000519587534e+00;
C[17947] = -4.810057640781e-03;
C[17948] = -6.653429418531e-03;
C[17949] = 5.900194879344e-02;
C[17950] = -2.243994214789e-04;
C[17951] = -3.711803833961e-04;
C[17952] = 1.250353531534e-02;
C[17953] = 2.155816832305e+00;
C[17954] = 3.923978755468e-02;
C[17955] = -2.545929468730e-02;
C[17956] = -6.278691412525e+00;
C[17957] = -4.165376944987e-02;
C[17958] = -1.825363398365e-02;
C[17959] = 2.082675925099e+00;
C[17960] = -1.934688252987e-02;
C[17961] = 7.753299654092e-03;
C[17962] = 3.984682974534e-01;
C[17963] = 1.143022390839e-02;
C[17964] = -8.197611904725e-03;
C[17965] = 9.556854027111e-02;
C[17966] = -2.063924613710e-02;
C[17967] = -1.357449271062e-03;
C[17968] = 3.846279340903e-01;
C[17969] = 7.596696622843e-03;
C[17970] = -4.796145745177e-03;
C[17971] = 1.017758008238e+00;
C[17972] = 4.158530794957e-02;
C[17973] = -2.310237940026e-04;
C[17974] = 6.025854166405e-02;
C[17975] = 2.738279521668e-03;
C[17976] = 2.063037573942e-02;
C[17977] = 3.119923107309e-02;
C[17978] = 2.063632105178e+00;
C[17979] = -1.079584372672e-03;
C[17980] = -3.904296846305e-02;
C[17981] = -6.253914128262e+00;
C[17982] = -6.731240792688e-03;
C[17983] = -1.652856110763e-02;
C[17984] = 2.133365536822e+00;
C[17985] = 4.014135506828e-03;
C[17986] = 1.090890201529e-02;
C[17987] = 3.760837526559e-01;
C[17988] = -2.897359943100e-03;
C[17989] = -2.050870608776e-02;
C[17990] = 1.039882833260e-01;
C[17991] = -6.366560571950e-04;
C[17992] = 7.801938398123e-03;
C[17993] = 3.982171201861e-01;
C[17994] = -6.668854644591e-03;
C[17995] = 4.176794226120e-02;
C[17996] = 1.083335769860e+00;
C[17997] = -3.823940965793e-04;
C[17998] = 2.730683693262e-03;
C[17999] = 6.440026732385e-02;
C[18000] = 7.892281606021e-01;
C[18001] = 3.251966627323e-02;
C[18002] = 9.246407227274e-03;
C[18003] = 2.136965164315e+00;
C[18004] = -3.409958864582e-02;
C[18005] = -6.491428771169e-03;
C[18006] = -7.120645489960e+00;
C[18007] = 3.039768701541e-03;
C[18008] = 1.277355856422e-02;
C[18009] = 5.082540432864e-02;
C[18010] = 1.989291581620e-03;
C[18011] = 6.063889446046e-04;
C[18012] = 3.830170265869e-01;
C[18013] = -2.149747004171e-03;
C[18014] = -6.246654561191e-04;
C[18015] = 5.141814592465e-02;
C[18016] = 8.330565242846e-04;
C[18017] = 2.295728070521e-03;
C[18018] = 4.120467248113e+00;
C[18019] = -5.604507425432e-03;
C[18020] = -1.465289984334e-02;
C[18021] = 5.124452250472e-01;
C[18022] = -8.022891391926e-04;
C[18023] = -2.234104015404e-03;
C[18024] = 3.320439803139e-02;
C[18025] = 8.503001288864e-01;
C[18026] = 4.179957323283e-02;
C[18027] = -3.433644471312e-02;
C[18028] = 2.093791391544e+00;
C[18029] = -7.222240102526e-03;
C[18030] = 2.403295681451e-03;
C[18031] = -7.144594171314e+00;
C[18032] = -4.678759265220e-02;
C[18033] = 2.031222333573e-03;
C[18034] = 5.530393486190e-02;
C[18035] = 3.056730764072e-03;
C[18036] = -2.174944091391e-03;
C[18037] = 3.851837074126e-01;
C[18038] = 8.267254448379e-03;
C[18039] = 8.012328732801e-04;
C[18040] = 4.359498401370e-02;
C[18041] = -1.567867835362e-02;
C[18042] = -5.458066572224e-03;
C[18043] = 4.116779702742e+00;
C[18044] = 1.117535691942e-02;
C[18045] = -7.907089710668e-04;
C[18046] = 5.132989571893e-01;
C[18047] = 4.281756430496e-03;
C[18048] = 1.034205986555e-02;
C[18049] = 4.360064021110e-02;
C[18050] = 8.088044603887e-01;
C[18051] = -6.798420844685e-03;
C[18052] = -9.469023505695e-03;
C[18053] = 2.153040939393e+00;
C[18054] = 1.148462434174e-02;
C[18055] = -4.694871128594e-02;
C[18056] = -7.218696421866e+00;
C[18057] = 6.840299012136e-04;
C[18058] = 3.206656191340e-03;
C[18059] = 5.285365068320e-02;
C[18060] = -6.660202434165e-04;
C[18061] = 8.090589574051e-03;
C[18062] = 3.992008903147e-01;
C[18063] = 2.231281359397e-03;
C[18064] = -1.568673428530e-02;
C[18065] = 1.934620697954e-02;
C[18066] = -1.437735264454e-02;
C[18067] = 1.136514286614e-02;
C[18068] = 4.174517519587e+00;
C[18069] = -2.212804618505e-03;
C[18070] = 4.302246668307e-03;
C[18071] = 5.259486707924e-01;
C[18072] = 0.000000000000e+00;
C[18073] = 0.000000000000e+00;
C[18074] = 0.000000000000e+00;
C[18075] = 0.000000000000e+00;
C[18076] = 0.000000000000e+00;
C[18077] = 0.000000000000e+00;
C[18078] = 0.000000000000e+00;
C[18079] = 0.000000000000e+00;
C[18080] = 0.000000000000e+00;
C[18081] = 0.000000000000e+00;
C[18082] = 0.000000000000e+00;
C[18083] = 0.000000000000e+00;
C[18084] = 0.000000000000e+00;
C[18085] = 0.000000000000e+00;
C[18086] = 0.000000000000e+00;
C[18087] = 0.000000000000e+00;
C[18088] = 0.000000000000e+00;
C[18089] = 0.000000000000e+00;
C[18090] = 1.000000000000e+00;
C[18091] = 0.000000000000e+00;
C[18092] = 0.000000000000e+00;
C[18093] = 2.000000000000e-01;
C[18094] = -0.000000000000e+00;
C[18095] = -0.000000000000e+00;
C[18096] = 0.000000000000e+00;
C[18097] = 0.000000000000e+00;
C[18098] = 0.000000000000e+00;
C[18099] = 0.000000000000e+00;
C[18100] = 0.000000000000e+00;
C[18101] = 0.000000000000e+00;
C[18102] = 0.000000000000e+00;
C[18103] = 0.000000000000e+00;
C[18104] = 0.000000000000e+00;
C[18105] = 0.000000000000e+00;
C[18106] = 0.000000000000e+00;
C[18107] = 0.000000000000e+00;
C[18108] = 0.000000000000e+00;
C[18109] = 0.000000000000e+00;
C[18110] = 0.000000000000e+00;
C[18111] = 0.000000000000e+00;
C[18112] = 0.000000000000e+00;
C[18113] = 0.000000000000e+00;
C[18114] = 0.000000000000e+00;
C[18115] = 1.000000000000e+00;
C[18116] = 0.000000000000e+00;
C[18117] = -0.000000000000e+00;
C[18118] = 2.000000000000e-01;
C[18119] = -0.000000000000e+00;
C[18120] = 0.000000000000e+00;
C[18121] = 0.000000000000e+00;
C[18122] = 0.000000000000e+00;
C[18123] = 0.000000000000e+00;
C[18124] = 0.000000000000e+00;
C[18125] = 0.000000000000e+00;
C[18126] = 0.000000000000e+00;
C[18127] = 0.000000000000e+00;
C[18128] = 0.000000000000e+00;
C[18129] = 0.000000000000e+00;
C[18130] = 0.000000000000e+00;
C[18131] = 0.000000000000e+00;
C[18132] = 0.000000000000e+00;
C[18133] = 0.000000000000e+00;
C[18134] = 0.000000000000e+00;
C[18135] = 0.000000000000e+00;
C[18136] = 0.000000000000e+00;
C[18137] = 0.000000000000e+00;
C[18138] = 0.000000000000e+00;
C[18139] = 0.000000000000e+00;
C[18140] = 1.000000000000e+00;
C[18141] = -0.000000000000e+00;
C[18142] = -0.000000000000e+00;
C[18143] = 2.000000000000e-01;
C[18144] = 7.800166175851e-02;
C[18145] = 2.156036411085e-03;
C[18146] = -2.527211775814e-03;
C[18147] = 3.788167757485e-01;
C[18148] = 1.879702373064e-03;
C[18149] = 2.273134858991e-04;
C[18150] = 5.160432997131e-02;
C[18151] = -1.360401440945e-04;
C[18152] = 6.520366717668e-05;
C[18153] = 1.307902892415e-01;
C[18154] = 4.815820293427e-04;
C[18155] = -7.095102440502e-05;
C[18156] = 3.086701173682e-02;
C[18157] = -9.113009889225e-05;
C[18158] = -1.000604916480e-04;
C[18159] = 2.496091892892e-03;
C[18160] = -2.164010865344e-05;
C[18161] = -3.211023686410e-06;
C[18162] = 2.463375942721e-03;
C[18163] = -5.755900833065e-05;
C[18164] = -3.722378787927e-05;
C[18165] = -1.604798173373e-05;
C[18166] = -3.346257288039e-06;
C[18167] = -3.135228721887e-06;
C[18168] = 2.346581524704e-03;
C[18169] = 2.538564187775e-02;
C[18170] = 2.859568885127e-03;
C[18171] = 1.676150671522e-03;
C[18172] = 3.958705479184e-01;
C[18173] = 1.195573786645e-02;
C[18174] = -1.140917632014e-04;
C[18175] = 5.577051531739e-02;
C[18176] = 3.720194432848e-03;
C[18177] = 4.942850369173e-04;
C[18178] = 1.264293546752e-01;
C[18179] = -5.627229535520e-05;
C[18180] = -1.039474478304e-04;
C[18181] = 3.250946353793e-02;
C[18182] = 1.070373163666e-03;
C[18183] = -1.950020951933e-05;
C[18184] = 2.716524487769e-03;
C[18185] = 1.916195528597e-04;
C[18186] = -5.449403033083e-05;
C[18187] = 2.735479249470e-03;
C[18188] = 2.648817806567e-04;
C[18189] = -3.244258331994e-06;
C[18190] = -1.280206467090e-05;
C[18191] = 2.189349402687e-06;
C[18192] = -2.492828678026e-03;
C[18193] = 2.926469374845e-03;
C[18194] = 6.410873593880e-02;
C[18195] = 1.134410556287e-04;
C[18196] = 1.229078046612e-02;
C[18197] = 3.823830431924e-01;
C[18198] = 8.888125178697e-05;
C[18199] = 3.631827984527e-03;
C[18200] = 5.493216928656e-02;
C[18201] = -6.865881788580e-05;
C[18202] = -5.181226270728e-05;
C[18203] = 1.297975686141e-01;
C[18204] = -1.078497946069e-04;
C[18205] = 1.086967591629e-03;
C[18206] = 3.130382123829e-02;
C[18207] = -9.685994329526e-07;
C[18208] = 1.856877411224e-04;
C[18209] = 2.679157471631e-03;
C[18210] = -3.260000615912e-05;
C[18211] = 2.454070616615e-04;
C[18212] = 2.819786119628e-03;
C[18213] = -2.914838726152e-06;
C[18214] = 1.963642615647e-06;
C[18215] = -1.517303576191e-05;
C[18216] = 3.790336511057e-01;
C[18217] = 1.627941055278e-03;
C[18218] = 1.288114734297e-04;
C[18219] = 1.214013347286e-01;
C[18220] = -2.675953844653e-03;
C[18221] = -2.737987536186e-03;
C[18222] = 3.855931347663e-01;
C[18223] = 1.725542925524e-03;
C[18224] = 2.333758710710e-03;
C[18225] = 3.088147009397e-02;
C[18226] = -1.071614222467e-04;
C[18227] = -1.068251000868e-04;
C[18228] = 1.326526215994e-01;
C[18229] = -2.269345876263e-05;
C[18230] = -1.355052209883e-04;
C[18231] = 3.149231367484e-02;
C[18232] = 1.752730206032e-04;
C[18233] = 2.339869864088e-04;
C[18234] = 5.890753761418e-02;
C[18235] = 3.358280480188e-04;
C[18236] = 3.444318909831e-04;
C[18237] = 2.572328714150e-03;
C[18238] = 1.725910911078e-05;
C[18239] = 1.923437459204e-05;
C[18240] = 1.538074478168e-03;
C[18241] = 3.961690551688e-01;
C[18242] = 1.299973269366e-02;
C[18243] = -2.557946966677e-03;
C[18244] = 9.239230516360e-02;
C[18245] = -2.211778819451e-02;
C[18246] = 1.716774304191e-03;
C[18247] = 3.884225216732e-01;
C[18248] = 7.034634723839e-03;
C[18249] = -1.139052918853e-04;
C[18250] = 3.252936402129e-02;
C[18251] = 1.134231073466e-03;
C[18252] = -1.482633356424e-05;
C[18253] = 1.301896671968e-01;
C[18254] = -1.984470201252e-03;
C[18255] = 1.740285846190e-04;
C[18256] = 3.188281202900e-02;
C[18257] = 7.408854560145e-04;
C[18258] = 3.218192174341e-04;
C[18259] = 6.001876853450e-02;
C[18260] = 2.403280996031e-03;
C[18261] = 1.671615865854e-05;
C[18262] = 2.632093908642e-03;
C[18263] = 1.172560902305e-04;
C[18264] = 9.350483393806e-05;
C[18265] = 1.262392032225e-02;
C[18266] = 3.823021995563e-01;
C[18267] = -2.655043275420e-03;
C[18268] = -2.209157411100e-02;
C[18269] = 1.000739311673e-01;
C[18270] = 2.354825480373e-03;
C[18271] = 7.224345586937e-03;
C[18272] = 3.961550669213e-01;
C[18273] = -1.089810684454e-04;
C[18274] = 1.114918660720e-03;
C[18275] = 3.129843166255e-02;
C[18276] = -1.299756036040e-04;
C[18277] = -1.982722595685e-03;
C[18278] = 1.307983640529e-01;
C[18279] = 2.341792515780e-04;
C[18280] = 7.510501507067e-04;
C[18281] = 3.257523059044e-02;
C[18282] = 3.060397294528e-04;
C[18283] = 2.352328774221e-03;
C[18284] = 6.439975776083e-02;
C[18285] = 1.738677497599e-05;
C[18286] = 1.138386199175e-04;
C[18287] = 2.830125344455e-03;
C[18288] = 5.155871405617e-02;
C[18289] = 5.773730771706e-04;
C[18290] = 4.990903282759e-04;
C[18291] = 3.863532743778e-01;
C[18292] = 9.889307923387e-04;
C[18293] = 1.580455097424e-03;
C[18294] = 5.355256992290e-02;
C[18295] = -1.055335201248e-03;
C[18296] = -1.160912325648e-03;
C[18297] = 2.493050831882e-03;
C[18298] = 2.659744650547e-05;
C[18299] = 2.637867233298e-05;
C[18300] = 3.154298964894e-02;
C[18301] = 1.255056838288e-04;
C[18302] = 1.825545593814e-04;
C[18303] = 1.290375888353e-01;
C[18304] = -1.366019397140e-04;
C[18305] = -2.679644133948e-04;
C[18306] = 5.062129061161e-01;
C[18307] = -5.187300351594e-04;
C[18308] = -9.581232355268e-04;
C[18309] = 3.694562691917e-02;
C[18310] = -1.609129427408e-05;
C[18311] = 5.885894632365e-05;
C[18312] = 5.576454161581e-04;
C[18313] = 5.615066266288e-02;
C[18314] = 3.736894114403e-03;
C[18315] = 1.023524634312e-03;
C[18316] = 3.888151708082e-01;
C[18317] = 7.372381689703e-03;
C[18318] = -1.078288290165e-03;
C[18319] = 4.622100907183e-02;
C[18320] = -1.415556535475e-02;
C[18321] = 2.460559536341e-05;
C[18322] = 2.741867644136e-03;
C[18323] = 1.926921497808e-04;
C[18324] = 1.284718011890e-04;
C[18325] = 3.190898863800e-02;
C[18326] = 7.609192242244e-04;
C[18327] = -1.381321456418e-04;
C[18328] = 1.284359537111e-01;
C[18329] = -1.162720014950e-03;
C[18330] = -5.049639745573e-04;
C[18331] = 5.059956714268e-01;
C[18332] = 2.966104440951e-03;
C[18333] = -1.543970210914e-05;
C[18334] = 3.692969621653e-02;
C[18335] = 2.101292715304e-04;
C[18336] = 4.957174558257e-04;
C[18337] = 3.826515527271e-03;
C[18338] = 5.464780074893e-02;
C[18339] = 1.581123702374e-03;
C[18340] = 7.164702344128e-03;
C[18341] = 3.963886450047e-01;
C[18342] = -1.211193352572e-03;
C[18343] = -1.411537205244e-02;
C[18344] = 1.885573973753e-02;
C[18345] = 2.548989555686e-05;
C[18346] = 1.987076258213e-04;
C[18347] = 2.660199569121e-03;
C[18348] = 1.838113191864e-04;
C[18349] = 7.495566307005e-04;
C[18350] = 3.259080246267e-02;
C[18351] = -2.713164818564e-04;
C[18352] = -1.160040461463e-03;
C[18353] = 1.259496412718e-01;
C[18354] = -9.025597288785e-04;
C[18355] = 3.018294646491e-03;
C[18356] = 5.275328357614e-01;
C[18357] = 6.194148713717e-05;
C[18358] = 2.124036020613e-04;
C[18359] = 3.881414440165e-02;
C[18360] = -6.999110396530e+00;
C[18361] = 2.487531803283e-02;
C[18362] = -9.128895510340e-03;
C[18363] = 2.142143812680e+00;
C[18364] = 7.025342364592e-03;
C[18365] = -8.936304214452e-03;
C[18366] = 8.035075385394e-01;
C[18367] = -5.661953138501e-03;
C[18368] = -1.721093753474e-03;
C[18369] = 7.119676330335e-02;
C[18370] = 9.197473749164e-03;
C[18371] = 5.110592862873e-04;
C[18372] = 3.814053382857e-01;
C[18373] = -3.347824158235e-03;
C[18374] = -2.868879817143e-03;
C[18375] = 5.186947181743e-02;
C[18376] = -4.429970514567e-04;
C[18377] = -1.006591282535e-04;
C[18378] = 7.937277882587e-02;
C[18379] = -8.772140818786e-04;
C[18380] = -4.847406986104e-04;
C[18381] = 2.476618239290e-03;
C[18382] = -3.735314829116e-05;
C[18383] = -2.405693449324e-05;
C[18384] = 2.868622030522e-02;
C[18385] = -7.203038296088e+00;
C[18386] = 4.516182672442e-02;
C[18387] = 3.243938985084e-03;
C[18388] = 2.104615403988e+00;
C[18389] = 1.993859115362e-02;
C[18390] = -5.217421667633e-03;
C[18391] = 8.564745028715e-01;
C[18392] = 5.225092053200e-02;
C[18393] = 9.388018862783e-03;
C[18394] = 2.299889855408e-02;
C[18395] = -9.519261806965e-04;
C[18396] = -3.554906901206e-03;
C[18397] = 3.947259188562e-01;
C[18398] = 1.100369230239e-02;
C[18399] = -3.932522582366e-04;
C[18400] = 5.578499041126e-02;
C[18401] = 3.631031691288e-03;
C[18402] = -7.905772897963e-04;
C[18403] = 8.649593426221e-02;
C[18404] = 7.377044712062e-03;
C[18405] = -3.287170053872e-05;
C[18406] = 2.757745434014e-03;
C[18407] = 2.668898843321e-04;
C[18408] = -8.441233554573e-03;
C[18409] = 4.649983651879e-02;
C[18410] = -7.043754740318e+00;
C[18411] = -1.123781126364e-02;
C[18412] = 2.437155696242e-02;
C[18413] = 2.115373490507e+00;
C[18414] = -1.178672478930e-03;
C[18415] = 5.160711925044e-02;
C[18416] = 8.478110430394e-01;
C[18417] = 5.454423840757e-04;
C[18418] = -8.850256909777e-04;
C[18419] = 6.225898698893e-02;
C[18420] = -3.014500892742e-03;
C[18421] = 1.130965923981e-02;
C[18422] = 3.846913330619e-01;
C[18423] = -5.009504687837e-05;
C[18424] = 3.557423497909e-03;
C[18425] = 5.507692683613e-02;
C[18426] = -3.896899330950e-04;
C[18427] = 6.906791756628e-03;
C[18428] = 8.919019556613e-02;
C[18429] = -1.980351027710e-05;
C[18430] = 2.476606760197e-04;
C[18431] = 2.811804553060e-03;
C[18432] = 2.146481319824e+00;
C[18433] = 2.279746660196e-03;
C[18434] = -1.093040290762e-02;
C[18435] = -6.223711787367e+00;
C[18436] = -4.669595996401e-04;
C[18437] = -4.618544506106e-03;
C[18438] = 2.142654256180e+00;
C[18439] = 1.330508618619e-02;
C[18440] = 1.213834369541e-02;
C[18441] = 3.816222136429e-01;
C[18442] = -3.603116517451e-03;
C[18443] = -2.999130474941e-03;
C[18444] = 1.167369781020e-01;
C[18445] = 2.230289655416e-03;
C[18446] = -2.523717637328e-04;
C[18447] = 3.873501356796e-01;
C[18448] = 2.432917135214e-03;
C[18449] = 2.904325594381e-03;
C[18450] = 9.971163723557e-01;
C[18451] = 5.839076358489e-03;
C[18452] = 6.735238711477e-03;
C[18453] = 5.881590346456e-02;
C[18454] = 3.639365917991e-04;
C[18455] = 5.653417723417e-04;
C[18456] = 1.927844666700e-04;
C[18457] = 2.110585548997e+00;
C[18458] = 3.855060151332e-02;
C[18459] = 1.893177959876e-03;
C[18460] = -6.289960487015e+00;
C[18461] = -4.659764617671e-02;
C[18462] = 1.303147593192e-02;
C[18463] = 2.107853510060e+00;
C[18464] = -2.054795379414e-02;
C[18465] = -3.689452053131e-03;
C[18466] = 3.950244261066e-01;
C[18467] = 1.201861146735e-02;
C[18468] = 2.348296533392e-03;
C[18469] = 9.207033963857e-02;
C[18470] = -2.066938754781e-02;
C[18471] = 2.389947946366e-03;
C[18472] = 3.898881692892e-01;
C[18473] = 6.456566683606e-03;
C[18474] = 5.611675322136e-03;
C[18475] = 1.011256400557e+00;
C[18476] = 3.502424269174e-02;
C[18477] = 3.566480917256e-04;
C[18478] = 5.981374239245e-02;
C[18479] = 2.234478555004e-03;
C[18480] = -1.161247725367e-02;
C[18481] = 3.330224026976e-02;
C[18482] = 2.113756617785e+00;
C[18483] = -2.959659290786e-03;
C[18484] = -4.607336450647e-02;
C[18485] = -6.274046093774e+00;
C[18486] = 1.224401213289e-02;
C[18487] = -1.804349516493e-02;
C[18488] = 2.110737651444e+00;
C[18489] = -3.002688469104e-03;
C[18490] = 1.167187475819e-02;
C[18491] = 3.846104894258e-01;
C[18492] = -1.694275029668e-04;
C[18493] = -2.064317346430e-02;
C[18494] = 9.762361030594e-02;
C[18495] = 2.886903565730e-03;
C[18496] = 6.643103117123e-03;
C[18497] = 3.961432476782e-01;
C[18498] = 6.146730385363e-03;
C[18499] = 3.438390256991e-02;
C[18500] = 1.083316036303e+00;
C[18501] = 5.350890609331e-04;
C[18502] = 2.168833913560e-03;
C[18503] = 6.444918174955e-02;
C[18504] = 8.025952202364e-01;
C[18505] = 8.611875139807e-03;
C[18506] = 7.025509050849e-03;
C[18507] = 2.157857048410e+00;
C[18508] = -1.525394305120e-03;
C[18509] = -3.243395526100e-03;
C[18510] = -7.087120613986e+00;
C[18511] = -1.774290079927e-03;
C[18512] = -6.948470024762e-03;
C[18513] = 5.182385590228e-02;
C[18514] = 2.982125821353e-04;
C[18515] = 3.601140296106e-04;
C[18516] = 3.881102752911e-01;
C[18517] = 1.662104434514e-03;
C[18518] = 2.112533182781e-03;
C[18519] = 5.472449918312e-02;
C[18520] = -1.802640627999e-03;
C[18521] = -4.678367619541e-03;
C[18522] = 4.051128053150e+00;
C[18523] = -4.694569651997e-03;
C[18524] = 2.435137163675e-03;
C[18525] = 5.030048349757e-01;
C[18526] = -1.141209821247e-04;
C[18527] = 2.192714668039e-03;
C[18528] = 8.211758066552e-03;
C[18529] = 8.640774497814e-01;
C[18530] = 5.370844184795e-02;
C[18531] = -7.352796380555e-04;
C[18532] = 2.115706492760e+00;
C[18533] = -1.508277310961e-02;
C[18534] = -2.233351858269e-03;
C[18535] = -7.104960690564e+00;
C[18536] = -3.692191075912e-02;
C[18537] = 2.506885087960e-04;
C[18538] = 5.616513775676e-02;
C[18539] = 3.662489627784e-03;
C[18540] = 1.730898844002e-03;
C[18541] = 3.902808184242e-01;
C[18542] = 6.791139219889e-03;
C[18543] = -1.825593716916e-03;
C[18544] = 4.883401911175e-02;
C[18545] = -1.154689789409e-02;
C[18546] = -4.503301535630e-03;
C[18547] = 4.036203060585e+00;
C[18548] = -3.819354361933e-03;
C[18549] = -1.083107222960e-04;
C[18550] = 5.019228420344e-01;
C[18551] = 1.023904243644e-03;
C[18552] = 6.889182019505e-03;
C[18553] = 5.437734242047e-02;
C[18554] = 8.421236722866e-01;
C[18555] = -2.914356471298e-03;
C[18556] = -1.794660138836e-02;
C[18557] = 2.115409213113e+00;
C[18558] = -7.954090563241e-03;
C[18559] = -3.611804471286e-02;
C[18560] = -7.229354280132e+00;
C[18561] = 3.298546603955e-04;
C[18562] = 3.737352785711e-03;
C[18563] = 5.479255829849e-02;
C[18564] = 2.151690586046e-03;
C[18565] = 6.586634303895e-03;
C[18566] = 3.963768257617e-01;
C[18567] = -4.728648646465e-03;
C[18568] = -1.150670459178e-02;
C[18569] = 1.818013369050e-02;
C[18570] = 3.357149505252e-03;
C[18571] = -3.023247362820e-03;
C[18572] = 4.189743468137e+00;
C[18573] = 2.240256962718e-03;
C[18574] = 1.089875588554e-03;
C[18575] = 5.280528305258e-01;
C[18576] = 0.000000000000e+00;
C[18577] = 0.000000000000e+00;
C[18578] = 0.000000000000e+00;
C[18579] = 0.000000000000e+00;
C[18580] = 0.000000000000e+00;
C[18581] = 0.000000000000e+00;
C[18582] = 0.000000000000e+00;
C[18583] = 0.000000000000e+00;
C[18584] = 0.000000000000e+00;
C[18585] = 0.000000000000e+00;
C[18586] = 0.000000000000e+00;
C[18587] = 0.000000000000e+00;
C[18588] = 0.000000000000e+00;
C[18589] = 0.000000000000e+00;
C[18590] = 0.000000000000e+00;
C[18591] = 0.000000000000e+00;
C[18592] = 0.000000000000e+00;
C[18593] = 0.000000000000e+00;
C[18594] = 1.000000000000e+00;
C[18595] = 0.000000000000e+00;
C[18596] = 0.000000000000e+00;
C[18597] = 2.000000000000e-01;
C[18598] = -0.000000000000e+00;
C[18599] = -0.000000000000e+00;
C[18600] = 0.000000000000e+00;
C[18601] = 0.000000000000e+00;
C[18602] = 0.000000000000e+00;
C[18603] = 0.000000000000e+00;
C[18604] = 0.000000000000e+00;
C[18605] = 0.000000000000e+00;
C[18606] = 0.000000000000e+00;
C[18607] = 0.000000000000e+00;
C[18608] = 0.000000000000e+00;
C[18609] = 0.000000000000e+00;
C[18610] = 0.000000000000e+00;
C[18611] = 0.000000000000e+00;
C[18612] = 0.000000000000e+00;
C[18613] = 0.000000000000e+00;
C[18614] = 0.000000000000e+00;
C[18615] = 0.000000000000e+00;
C[18616] = 0.000000000000e+00;
C[18617] = 0.000000000000e+00;
C[18618] = 0.000000000000e+00;
C[18619] = 1.000000000000e+00;
C[18620] = 0.000000000000e+00;
C[18621] = -0.000000000000e+00;
C[18622] = 2.000000000000e-01;
C[18623] = -0.000000000000e+00;
C[18624] = 0.000000000000e+00;
C[18625] = 0.000000000000e+00;
C[18626] = 0.000000000000e+00;
C[18627] = 0.000000000000e+00;
C[18628] = 0.000000000000e+00;
C[18629] = 0.000000000000e+00;
C[18630] = 0.000000000000e+00;
C[18631] = 0.000000000000e+00;
C[18632] = 0.000000000000e+00;
C[18633] = 0.000000000000e+00;
C[18634] = 0.000000000000e+00;
C[18635] = 0.000000000000e+00;
C[18636] = 0.000000000000e+00;
C[18637] = 0.000000000000e+00;
C[18638] = 0.000000000000e+00;
C[18639] = 0.000000000000e+00;
C[18640] = 0.000000000000e+00;
C[18641] = 0.000000000000e+00;
C[18642] = 0.000000000000e+00;
C[18643] = 0.000000000000e+00;
C[18644] = 1.000000000000e+00;
C[18645] = -0.000000000000e+00;
C[18646] = -0.000000000000e+00;
C[18647] = 2.000000000000e-01;
C[18648] = 6.469394078925e-02;
C[18649] = 1.259142703831e-02;
C[18650] = 1.762798689493e-03;
C[18651] = 3.832955554736e-01;
C[18652] = -5.027966075722e-03;
C[18653] = -3.765892216343e-03;
C[18654] = 5.231237550569e-02;
C[18655] = -7.147618130741e-04;
C[18656] = -2.524440003428e-04;
C[18657] = 1.298636453819e-01;
C[18658] = 1.141780277740e-03;
C[18659] = 1.637028459162e-04;
C[18660] = 3.119476481465e-02;
C[18661] = -4.970794805847e-04;
C[18662] = -3.073001519648e-04;
C[18663] = 2.526367925082e-03;
C[18664] = -3.984376710377e-05;
C[18665] = -1.034207131606e-05;
C[18666] = 2.444776961526e-03;
C[18667] = -4.943974960725e-05;
C[18668] = 3.346417656216e-05;
C[18669] = -2.034489915814e-05;
C[18670] = -1.138195581285e-06;
C[18671] = -2.532911763898e-07;
C[18672] = 1.260233594761e-02;
C[18673] = 2.135933188784e-02;
C[18674] = -3.595373118835e-03;
C[18675] = -5.054506921627e-03;
C[18676] = 3.947783886494e-01;
C[18677] = 9.052455383692e-03;
C[18678] = -6.683051477768e-04;
C[18679] = 5.591530230487e-02;
C[18680] = 3.230508774217e-03;
C[18681] = 1.142507538360e-03;
C[18682] = 1.261490455789e-01;
C[18683] = -4.314961959086e-04;
C[18684] = -4.985885503942e-04;
C[18685] = 3.253314633982e-02;
C[18686] = 8.454925135649e-04;
C[18687] = -3.724464557973e-05;
C[18688] = 2.727050353863e-03;
C[18689] = 1.694255564388e-04;
C[18690] = -4.299483488510e-05;
C[18691] = 2.687342417206e-03;
C[18692] = 2.094740757651e-04;
C[18693] = -1.091917419887e-06;
C[18694] = -1.572039332775e-05;
C[18695] = 1.117374865380e-06;
C[18696] = 1.764139961798e-03;
C[18697] = -3.603882885250e-03;
C[18698] = 6.097179783610e-02;
C[18699] = -3.879179289023e-03;
C[18700] = 9.358619755804e-03;
C[18701] = 3.846976587756e-01;
C[18702] = -2.030368552951e-04;
C[18703] = 3.155486823428e-03;
C[18704] = 5.513964064583e-02;
C[18705] = 1.637922640698e-04;
C[18706] = -4.320635136696e-04;
C[18707] = 1.296119114964e-01;
C[18708] = -3.132618081315e-04;
C[18709] = 8.622219398752e-04;
C[18710] = 3.138112573202e-02;
C[18711] = -7.288684200905e-06;
C[18712] = 1.635368025140e-04;
C[18713] = 2.682431532723e-03;
C[18714] = 4.177337910872e-05;
C[18715] = 1.868699047376e-04;
C[18716] = 2.805825485173e-03;
C[18717] = -1.274914843964e-07;
C[18718] = 4.807449426657e-07;
C[18719] = -1.587084550328e-05;
C[18720] = 3.839641881115e-01;
C[18721] = -5.107991604622e-03;
C[18722] = -3.749050347964e-03;
C[18723] = 1.151317614339e-01;
C[18724] = 4.774569370747e-03;
C[18725] = 1.178733060727e-03;
C[18726] = 3.861762175896e-01;
C[18727] = 2.191959694698e-03;
C[18728] = 1.663856834781e-03;
C[18729] = 3.123934032384e-02;
C[18730] = -5.021541959272e-04;
C[18731] = -3.045865453943e-04;
C[18732] = 1.324610200502e-01;
C[18733] = 4.668957490752e-04;
C[18734] = 1.475267713530e-04;
C[18735] = 3.131193487100e-02;
C[18736] = 1.267339961613e-04;
C[18737] = 1.187683222424e-04;
C[18738] = 5.786577276324e-02;
C[18739] = 2.231705943633e-04;
C[18740] = 1.219208356227e-03;
C[18741] = 2.511803825728e-03;
C[18742] = 6.054542014198e-06;
C[18743] = 5.180995575834e-05;
C[18744] = -5.070954290764e-03;
C[18745] = 3.952752197009e-01;
C[18746] = 9.493741125261e-03;
C[18747] = 4.784253468599e-03;
C[18748] = 9.349215668446e-02;
C[18749] = -1.924619004436e-02;
C[18750] = 2.099067290411e-03;
C[18751] = 3.887927169555e-01;
C[18752] = 7.157780331307e-03;
C[18753] = -4.999453615875e-04;
C[18754] = 3.256626840992e-02;
C[18755] = 8.712300311723e-04;
C[18756] = 4.675413555986e-04;
C[18757] = 1.304523330479e-01;
C[18758] = -1.764355017766e-03;
C[18759] = 1.216862338818e-04;
C[18760] = 3.162049845860e-02;
C[18761] = 7.594480131030e-04;
C[18762] = 2.467013255944e-04;
C[18763] = 5.828721858150e-02;
C[18764] = 1.928333322480e-03;
C[18765] = 7.526437869221e-06;
C[18766] = 2.537136642001e-03;
C[18767] = 9.765688104172e-05;
C[18768] = -3.636621837296e-03;
C[18769] = 9.218069986385e-03;
C[18770] = 3.846360264245e-01;
C[18771] = 1.200146985371e-03;
C[18772] = -1.921506438986e-02;
C[18773] = 9.629334270627e-02;
C[18774] = 1.531959901264e-03;
C[18775] = 7.321128726046e-03;
C[18776] = 3.972225576877e-01;
C[18777] = -2.986821266950e-04;
C[18778] = 8.565334870777e-04;
C[18779] = 3.137701690862e-02;
C[18780] = 1.489543663293e-04;
C[18781] = -1.762279974132e-03;
C[18782] = 1.305454668486e-01;
C[18783] = 1.108823262736e-04;
C[18784] = 7.678358279721e-04;
C[18785] = 3.273553429167e-02;
C[18786] = 1.256671562307e-03;
C[18787] = 1.844857128315e-03;
C[18788] = 6.467926429815e-02;
C[18789] = 5.453747905909e-05;
C[18790] = 9.332893810948e-05;
C[18791] = 2.847307187322e-03;
C[18792] = 5.187808601291e-02;
C[18793] = -8.167667658733e-04;
C[18794] = -4.185313612873e-04;
C[18795] = 3.857299058148e-01;
C[18796] = 2.050582317028e-03;
C[18797] = 1.499293914956e-03;
C[18798] = 5.465510944317e-02;
C[18799] = -1.585209476321e-03;
C[18800] = -8.766030364053e-03;
C[18801] = 2.497415292230e-03;
C[18802] = -4.714208678617e-05;
C[18803] = -2.165498460038e-05;
C[18804] = 3.128218075267e-02;
C[18805] = 1.184539023229e-04;
C[18806] = 1.087045938531e-04;
C[18807] = 1.291292009815e-01;
C[18808] = -1.308028161483e-04;
C[18809] = -7.215798470036e-04;
C[18810] = 5.053751873958e-01;
C[18811] = 4.539642712671e-04;
C[18812] = 7.698453130110e-03;
C[18813] = 3.711211850542e-02;
C[18814] = 5.987308082780e-05;
C[18815] = 6.347064471112e-04;
C[18816] = -8.564227277256e-04;
C[18817] = 5.516738661285e-02;
C[18818] = 3.259720052263e-03;
C[18819] = 2.114681307102e-03;
C[18820] = 3.886030017022e-01;
C[18821] = 7.160209310883e-03;
C[18822] = -1.575695954824e-03;
C[18823] = 5.120964962370e-02;
C[18824] = -1.049391749043e-02;
C[18825] = -4.928782808053e-05;
C[18826] = 2.677189307728e-03;
C[18827] = 1.704856844363e-04;
C[18828] = 1.215821036549e-04;
C[18829] = 3.160785077504e-02;
C[18830] = 7.571078669612e-04;
C[18831] = -1.301685813819e-04;
C[18832] = 1.288600720704e-01;
C[18833] = -1.016323623121e-03;
C[18834] = 4.211971577553e-04;
C[18835] = 5.022476790556e-01;
C[18836] = -6.608330165875e-05;
C[18837] = 5.820354595713e-05;
C[18838] = 3.687271259926e-02;
C[18839] = 8.771115925422e-05;
C[18840] = -4.756558161144e-04;
C[18841] = 3.353444808368e-03;
C[18842] = 5.545129264011e-02;
C[18843] = 1.570843167075e-03;
C[18844] = 6.911962008228e-03;
C[18845] = 3.972651419610e-01;
C[18846] = -8.740943911040e-03;
C[18847] = -1.047933287894e-02;
C[18848] = 1.885094259800e-02;
C[18849] = -2.522285903416e-05;
C[18850] = 1.776212920488e-04;
C[18851] = 2.703208332342e-03;
C[18852] = 1.125674110620e-04;
C[18853] = 7.430601248977e-04;
C[18854] = 3.273837324322e-02;
C[18855] = -7.199074168028e-04;
C[18856] = -1.015351315688e-03;
C[18857] = 1.259468217819e-01;
C[18858] = 7.667413990236e-03;
C[18859] = 4.653213652906e-05;
C[18860] = 5.258044159092e-01;
C[18861] = 6.328877554502e-04;
C[18862] = 9.300181884577e-05;
C[18863] = 3.862638399424e-02;
C[18864] = -7.044684422472e+00;
C[18865] = 4.955475650543e-02;
C[18866] = -2.939353354692e-03;
C[18867] = 2.136683066139e+00;
C[18868] = 2.763584010249e-03;
C[18869] = -1.317303967463e-02;
C[18870] = 8.106726039280e-01;
C[18871] = -9.236227219568e-03;
C[18872] = -3.877239799604e-03;
C[18873] = 6.293333868323e-02;
C[18874] = 1.355146525328e-02;
C[18875] = 1.315748429286e-03;
C[18876] = 3.832681561080e-01;
C[18877] = -4.743714006969e-03;
C[18878] = -3.215223341580e-03;
C[18879] = 5.211506009356e-02;
C[18880] = -7.450823151896e-04;
C[18881] = -2.872840178759e-04;
C[18882] = 7.864308318797e-02;
C[18883] = -1.306643871387e-03;
C[18884] = 8.214491349159e-04;
C[18885] = 2.409696400463e-03;
C[18886] = -5.888194915939e-05;
C[18887] = 1.810936333285e-05;
C[18888] = 4.977293469134e-02;
C[18889] = -7.218036588994e+00;
C[18890] = 1.220763463591e-02;
C[18891] = 2.460723306285e-03;
C[18892] = 2.103205447986e+00;
C[18893] = 5.799686007086e-03;
C[18894] = -8.733523835831e-03;
C[18895] = 8.602161849182e-01;
C[18896] = 4.528659230956e-02;
C[18897] = 1.356237416258e-02;
C[18898] = 2.041914042296e-02;
C[18899] = -4.223545079082e-03;
C[18900] = -4.771044093502e-03;
C[18901] = 3.962638450299e-01;
C[18902] = 8.299245488976e-03;
C[18903] = -7.083098922388e-04;
C[18904] = 5.572414520293e-02;
C[18905] = 3.218087186579e-03;
C[18906] = -1.150566318915e-03;
C[18907] = 8.442847464581e-02;
C[18908] = 5.902497818598e-03;
C[18909] = -5.312744931474e-05;
C[18910] = 2.637322536924e-03;
C[18911] = 2.146251204656e-04;
C[18912] = -2.912527908605e-03;
C[18913] = 1.203743930761e-02;
C[18914] = -7.052333228074e+00;
C[18915] = -1.474351916951e-02;
C[18916] = 1.030349298926e-02;
C[18917] = 2.111609842210e+00;
C[18918] = -3.422557482450e-03;
C[18919] = 4.457576053126e-02;
C[18920] = 8.481934381105e-01;
C[18921] = 1.317089701590e-03;
C[18922] = -4.232054845497e-03;
C[18923] = 6.111088355265e-02;
C[18924] = -3.325581837889e-03;
C[18925] = 8.617759030803e-03;
C[18926] = 3.836783012285e-01;
C[18927] = -2.427388295504e-04;
C[18928] = 3.121711005215e-03;
C[18929] = 5.515910321409e-02;
C[18930] = 1.049406524893e-03;
C[18931] = 5.361839085508e-03;
C[18932] = 8.923107514607e-02;
C[18933] = 2.841290674457e-05;
C[18934] = 1.902783393799e-04;
C[18935] = 2.824402194897e-03;
C[18936] = 2.150055718896e+00;
C[18937] = 1.391029646382e-03;
C[18938] = -1.214094034834e-02;
C[18939] = -6.230341202193e+00;
C[18940] = 1.300476560023e-02;
C[18941] = 1.232324211974e-03;
C[18942] = 2.138644514623e+00;
C[18943] = 1.517749371161e-02;
C[18944] = -3.719883820280e-03;
C[18945] = 3.839367887458e-01;
C[18946] = -4.824528776497e-03;
C[18947] = -3.195452896830e-03;
C[18948] = 1.187325594238e-01;
C[18949] = 5.433558016723e-03;
C[18950] = 1.861095288680e-03;
C[18951] = 3.824900102135e-01;
C[18952] = 1.328793680103e-03;
C[18953] = 4.325127994790e-04;
C[18954] = 9.779103532846e-01;
C[18955] = 3.734067814595e-03;
C[18956] = 1.921871276732e-02;
C[18957] = 5.786874178096e-02;
C[18958] = 2.115125260692e-04;
C[18959] = 1.198192811672e-03;
C[18960] = 1.903819709408e-03;
C[18961] = 2.113142069016e+00;
C[18962] = 1.300592037839e-02;
C[18963] = 1.319844755727e-02;
C[18964] = -6.284524017205e+00;
C[18965] = -4.222240056500e-02;
C[18966] = 1.385115504226e-02;
C[18967] = 2.110295995983e+00;
C[18968] = -8.718344492854e-03;
C[18969] = -4.786702222011e-03;
C[18970] = 3.967606760814e-01;
C[18971] = 8.752880400260e-03;
C[18972] = 5.443242114575e-03;
C[18973] = 9.658623645916e-02;
C[18974] = -1.884536179712e-02;
C[18975] = 1.241906065582e-03;
C[18976] = 3.842404135309e-01;
C[18977] = 7.368017523535e-03;
C[18978] = 4.049659191285e-03;
C[18979] = 9.815217068056e-01;
C[18980] = 2.922277867579e-02;
C[18981] = 2.390494370634e-04;
C[18982] = 5.821791565466e-02;
C[18983] = 1.998278577013e-03;
C[18984] = -1.058763209369e-02;
C[18985] = 9.111978060938e-03;
C[18986] = 2.110377195188e+00;
C[18987] = 1.660602704848e-03;
C[18988] = -4.159988747491e-02;
C[18989] = -6.281003041862e+00;
C[18990] = -5.823104741345e-03;
C[18991] = -6.691201201309e-03;
C[18992] = 2.117097941311e+00;
C[18993] = -3.085952962533e-03;
C[18994] = 8.464860091668e-03;
C[18995] = 3.836166688774e-01;
C[18996] = 1.882509213323e-03;
C[18997] = -1.881423614261e-02;
C[18998] = 9.588245814168e-02;
C[18999] = 2.865909793490e-04;
C[19000] = 7.535867948254e-03;
C[19001] = 3.986421856576e-01;
C[19002] = 1.986256509266e-02;
C[19003] = 2.824738106277e-02;
C[19004] = 1.087942829092e+00;
C[19005] = 1.253410507383e-03;
C[19006] = 1.933651425319e-03;
C[19007] = 6.467729478889e-02;
C[19008] = 8.019868140724e-01;
C[19009] = -1.170275619776e-02;
C[19010] = -7.732447602294e-03;
C[19011] = 2.129718279126e+00;
C[19012] = 1.288145557459e-02;
C[19013] = -6.476424467490e-03;
C[19014] = -7.089873762550e+00;
C[19015] = -3.384162819194e-03;
C[19016] = -4.027423386659e-02;
C[19017] = 5.168077060078e-02;
C[19018] = -8.567715103353e-04;
C[19019] = -4.582333355426e-04;
C[19020] = 3.820436984387e-01;
C[19021] = 1.193421092198e-03;
C[19022] = 2.539249930417e-04;
C[19023] = 5.527065935013e-02;
C[19024] = -1.253286527426e-03;
C[19025] = -8.342153349454e-03;
C[19026] = 4.081296439924e+00;
C[19027] = 5.048575511490e-03;
C[19028] = 5.477242626006e-02;
C[19029] = 5.086320427048e-01;
C[19030] = 1.012951964124e-03;
C[19031] = 8.550647513521e-03;
C[19032] = -1.206944551260e-02;
C[19033] = 8.452578710777e-01;
C[19034] = 4.666042510795e-02;
C[19035] = 1.363192595968e-02;
C[19036] = 2.106501690917e+00;
C[19037] = -9.909589504571e-03;
C[19038] = -3.193892389261e-03;
C[19039] = -7.090480575537e+00;
C[19040] = -2.809726616321e-02;
C[19041] = -8.867432298410e-04;
C[19042] = 5.497622951090e-02;
C[19043] = 3.225944234049e-03;
C[19044] = 1.251515292507e-03;
C[19045] = 3.840506982775e-01;
C[19046] = 7.374948533091e-03;
C[19047] = -1.243773005930e-03;
C[19048] = 5.329308674800e-02;
C[19049] = -1.170084451237e-02;
C[19050] = 4.513356445117e-03;
C[19051] = 4.050093349203e+00;
C[19052] = -1.231866902336e-02;
C[19053] = 9.771599301002e-04;
C[19054] = 5.048768193365e-01;
C[19055] = 9.679256242428e-04;
C[19056] = -8.341476115036e-03;
C[19057] = 4.774531299258e-02;
C[19058] = 8.544264779961e-01;
C[19059] = -5.580157174408e-03;
C[19060] = -1.363471095445e-02;
C[19061] = 2.117949626776e+00;
C[19062] = -3.977250480634e-02;
C[19063] = -2.780557393353e-02;
C[19064] = -7.222384366943e+00;
C[19065] = -5.104958336475e-04;
C[19066] = 3.341023220729e-03;
C[19067] = 5.547075520837e-02;
C[19068] = 3.394991317726e-04;
C[19069] = 7.122199200456e-03;
C[19070] = 3.986847699308e-01;
C[19071] = -8.317066896442e-03;
C[19072] = -1.168625990089e-02;
C[19073] = 1.859320190956e-02;
C[19074] = 5.417502201466e-02;
C[19075] = -1.061795520179e-02;
C[19076] = 4.166930881733e+00;
C[19077] = 8.499970674880e-03;
C[19078] = 1.063676432372e-03;
C[19079] = 5.246382226708e-01;
C[19080] = 0.000000000000e+00;
C[19081] = 0.000000000000e+00;
C[19082] = 0.000000000000e+00;
C[19083] = 0.000000000000e+00;
C[19084] = 0.000000000000e+00;
C[19085] = 0.000000000000e+00;
C[19086] = 0.000000000000e+00;
C[19087] = 0.000000000000e+00;
C[19088] = 0.000000000000e+00;
C[19089] = 0.000000000000e+00;
C[19090] = 0.000000000000e+00;
C[19091] = 0.000000000000e+00;
C[19092] = 0.000000000000e+00;
C[19093] = 0.000000000000e+00;
C[19094] = 0.000000000000e+00;
C[19095] = 0.000000000000e+00;
C[19096] = 0.000000000000e+00;
C[19097] = 0.000000000000e+00;
C[19098] = 1.000000000000e+00;
C[19099] = 0.000000000000e+00;
C[19100] = 0.000000000000e+00;
C[19101] = 2.000000000000e-01;
C[19102] = -0.000000000000e+00;
C[19103] = -0.000000000000e+00;
C[19104] = 0.000000000000e+00;
C[19105] = 0.000000000000e+00;
C[19106] = 0.000000000000e+00;
C[19107] = 0.000000000000e+00;
C[19108] = 0.000000000000e+00;
C[19109] = 0.000000000000e+00;
C[19110] = 0.000000000000e+00;
C[19111] = 0.000000000000e+00;
C[19112] = 0.000000000000e+00;
C[19113] = 0.000000000000e+00;
C[19114] = 0.000000000000e+00;
C[19115] = 0.000000000000e+00;
C[19116] = 0.000000000000e+00;
C[19117] = 0.000000000000e+00;
C[19118] = 0.000000000000e+00;
C[19119] = 0.000000000000e+00;
C[19120] = 0.000000000000e+00;
C[19121] = 0.000000000000e+00;
C[19122] = 0.000000000000e+00;
C[19123] = 1.000000000000e+00;
C[19124] = 0.000000000000e+00;
C[19125] = -0.000000000000e+00;
C[19126] = 2.000000000000e-01;
C[19127] = -0.000000000000e+00;
C[19128] = 0.000000000000e+00;
C[19129] = 0.000000000000e+00;
C[19130] = 0.000000000000e+00;
C[19131] = 0.000000000000e+00;
C[19132] = 0.000000000000e+00;
C[19133] = 0.000000000000e+00;
C[19134] = 0.000000000000e+00;
C[19135] = 0.000000000000e+00;
C[19136] = 0.000000000000e+00;
C[19137] = 0.000000000000e+00;
C[19138] = 0.000000000000e+00;
C[19139] = 0.000000000000e+00;
C[19140] = 0.000000000000e+00;
C[19141] = 0.000000000000e+00;
C[19142] = 0.000000000000e+00;
C[19143] = 0.000000000000e+00;
C[19144] = 0.000000000000e+00;
C[19145] = 0.000000000000e+00;
C[19146] = 0.000000000000e+00;
C[19147] = 0.000000000000e+00;
C[19148] = 1.000000000000e+00;
C[19149] = -0.000000000000e+00;
C[19150] = -0.000000000000e+00;
C[19151] = 2.000000000000e-01;
C[19152] = 6.474949596874e-02;
C[19153] = 1.383095771410e-02;
C[19154] = -2.038746800272e-04;
C[19155] = 3.813299975378e-01;
C[19156] = -3.094153064255e-03;
C[19157] = -2.205274124560e-03;
C[19158] = 4.988328617606e-02;
C[19159] = -5.418839914439e-04;
C[19160] = -3.653524068175e-04;
C[19161] = 1.302845408695e-01;
C[19162] = 1.147299129399e-03;
C[19163] = -7.651167721875e-05;
C[19164] = 3.071109737981e-02;
C[19165] = -2.904177331587e-04;
C[19166] = -1.601148321854e-04;
C[19167] = 2.389543677967e-03;
C[19168] = -2.791786448403e-05;
C[19169] = -1.472575477800e-05;
C[19170] = 2.311611317658e-03;
C[19171] = -2.210838396425e-05;
C[19172] = 3.501305365172e-05;
C[19173] = -2.259812654687e-05;
C[19174] = -8.269334061048e-08;
C[19175] = 1.591470003516e-06;
C[19176] = 1.376128820718e-02;
C[19177] = 2.053690100807e-02;
C[19178] = -3.771665866990e-03;
C[19179] = -3.100271370063e-03;
C[19180] = 3.994438977771e-01;
C[19181] = 8.837990633209e-03;
C[19182] = -5.342194790170e-04;
C[19183] = 5.463283185757e-02;
C[19184] = 3.373501843190e-03;
C[19185] = 1.142654495604e-03;
C[19186] = 1.261257473953e-01;
C[19187] = -3.419339502064e-04;
C[19188] = -2.903673462411e-04;
C[19189] = 3.286978127564e-02;
C[19190] = 8.860940942840e-04;
C[19191] = -2.713461599251e-05;
C[19192] = 2.663682421905e-03;
C[19193] = 1.810254768285e-04;
C[19194] = -1.531594582767e-05;
C[19195] = 2.667930317634e-03;
C[19196] = 2.505603835125e-04;
C[19197] = 1.244302566189e-07;
C[19198] = -1.569999080345e-05;
C[19199] = 2.135797063116e-06;
C[19200] = -2.039753142717e-04;
C[19201] = -3.833041412745e-03;
C[19202] = 6.335111446646e-02;
C[19203] = -2.283347084494e-03;
C[19204] = 9.284173822058e-03;
C[19205] = 3.806432010280e-01;
C[19206] = -3.393421693659e-04;
C[19207] = 3.186006480412e-03;
C[19208] = 5.438228040245e-02;
C[19209] = -7.651838616838e-05;
C[19210] = -3.460256532568e-04;
C[19211] = 1.301002192061e-01;
C[19212] = -1.640595971724e-04;
C[19213] = 9.114267873300e-04;
C[19214] = 3.080990953559e-02;
C[19215] = -1.293327844903e-05;
C[19216] = 1.688091782982e-04;
C[19217] = 2.629848033511e-03;
C[19218] = 4.439327393042e-05;
C[19219] = 2.217846466918e-04;
C[19220] = 2.702599469469e-03;
C[19221] = 1.816788135615e-06;
C[19222] = 1.342649213953e-06;
C[19223] = -2.007047221737e-05;
C[19224] = 3.812898598973e-01;
C[19225] = -3.353890233215e-03;
C[19226] = -2.369377825118e-03;
C[19227] = 1.259919696986e-01;
C[19228] = 4.688826866061e-03;
C[19229] = 1.997149653967e-03;
C[19230] = 3.788109885068e-01;
C[19231] = 3.471847966564e-04;
C[19232] = -7.466146758973e-04;
C[19233] = 3.070842153711e-02;
C[19234] = -3.072752704512e-04;
C[19235] = -1.697949798806e-04;
C[19236] = 1.335473493635e-01;
C[19237] = 3.342803020389e-04;
C[19238] = 7.337402463635e-05;
C[19239] = 3.080892447469e-02;
C[19240] = 4.698323703206e-05;
C[19241] = 3.878675083803e-05;
C[19242] = 5.790447209179e-02;
C[19243] = 3.197668632218e-04;
C[19244] = 1.118022073696e-03;
C[19245] = 2.520075610671e-03;
C[19246] = 1.533104326690e-05;
C[19247] = 5.580330829565e-05;
C[19248] = -3.209740864230e-03;
C[19249] = 3.990791584559e-01;
C[19250] = 8.784019654289e-03;
C[19251] = 4.610482709903e-03;
C[19252] = 1.002644318005e-01;
C[19253] = -1.888672167520e-02;
C[19254] = 3.190800525582e-04;
C[19255] = 3.780653479937e-01;
C[19256] = 6.784327166049e-03;
C[19257] = -2.981235864904e-04;
C[19258] = 3.284546532089e-02;
C[19259] = 8.780831761455e-04;
C[19260] = 3.290573582951e-04;
C[19261] = 1.308609578727e-01;
C[19262] = -1.808561691551e-03;
C[19263] = 4.575982523986e-05;
C[19264] = 3.091681493832e-02;
C[19265] = 7.594514803508e-04;
C[19266] = 3.440098118383e-04;
C[19267] = 5.815406674942e-02;
C[19268] = 2.353997432390e-03;
C[19269] = 1.652177537425e-05;
C[19270] = 2.543924051240e-03;
C[19271] = 1.187181992368e-04;
C[19272] = -2.258458029617e-03;
C[19273] = 8.439321031178e-03;
C[19274] = 3.807734374027e-01;
C[19275] = 1.930889539440e-03;
C[19276] = -1.897734861333e-02;
C[19277] = 9.748664216104e-02;
C[19278] = -8.117242510572e-04;
C[19279] = 7.058854566692e-03;
C[19280] = 4.009145526697e-01;
C[19281] = -1.636604258559e-04;
C[19282] = 8.595161208153e-04;
C[19283] = 3.081859196057e-02;
C[19284] = 6.895668366788e-05;
C[19285] = -1.814603487427e-03;
C[19286] = 1.309662110852e-01;
C[19287] = 3.653612079602e-05;
C[19288] = 7.755677439160e-04;
C[19289] = 3.294207740512e-02;
C[19290] = 1.182783411565e-03;
C[19291] = 2.344986635486e-03;
C[19292] = 6.479504112453e-02;
C[19293] = 5.876042574939e-05;
C[19294] = 1.174853901959e-04;
C[19295] = 2.845866027506e-03;
C[19296] = 5.059188883499e-02;
C[19297] = -5.921054287099e-04;
C[19298] = -1.649258006696e-04;
C[19299] = 3.780294367864e-01;
C[19300] = 5.018663732436e-04;
C[19301] = -4.554635957697e-04;
C[19302] = 5.682694976244e-02;
C[19303] = -1.231326056194e-03;
C[19304] = -5.585379136474e-03;
C[19305] = 2.436783855229e-03;
C[19306] = -3.099367930538e-05;
C[19307] = -1.305520535940e-06;
C[19308] = 3.075682102666e-02;
C[19309] = 5.794557995222e-05;
C[19310] = 6.028683114852e-05;
C[19311] = 1.293105134492e-01;
C[19312] = -1.267659759341e-04;
C[19313] = -4.571594661704e-04;
C[19314] = 5.122127420687e-01;
C[19315] = 1.413934023997e-03;
C[19316] = 6.185298288386e-03;
C[19317] = 3.751452323006e-02;
C[19318] = 9.941071780833e-05;
C[19319] = 3.975127759455e-04;
C[19320] = -6.107676166535e-04;
C[19321] = 5.454895470751e-02;
C[19322] = 3.139025117820e-03;
C[19323] = 5.638456524802e-04;
C[19324] = 3.779092855750e-01;
C[19325] = 7.026153247629e-03;
C[19326] = -1.234056839789e-03;
C[19327] = 5.429181421006e-02;
C[19328] = -1.312442148534e-02;
C[19329] = -3.251010616467e-05;
C[19330] = 2.658090611901e-03;
C[19331] = 1.656770874587e-04;
C[19332] = 6.142729408697e-05;
C[19333] = 3.090641077708e-02;
C[19334] = 7.733876559785e-04;
C[19335] = -1.269480281737e-04;
C[19336] = 1.289877454429e-01;
C[19337] = -1.240358411454e-03;
C[19338] = 1.379691858714e-03;
C[19339] = 5.104741522432e-01;
C[19340] = 2.796308859025e-03;
C[19341] = 9.782271635104e-05;
C[19342] = 3.746179682453e-02;
C[19343] = 2.976930740595e-04;
C[19344] = -2.015853994883e-04;
C[19345] = 3.306266614797e-03;
C[19346] = 5.501190741740e-02;
C[19347] = -3.073402058038e-04;
C[19348] = 6.830923941703e-03;
C[19349] = 4.000868677376e-01;
C[19350] = -5.606753237573e-03;
C[19351] = -1.316826822702e-02;
C[19352] = 1.741496947788e-02;
C[19353] = -3.807954289386e-06;
C[19354] = 1.765431282689e-04;
C[19355] = 2.671823167841e-03;
C[19356] = 6.807171551093e-05;
C[19357] = 7.625579320611e-04;
C[19358] = 3.288689840965e-02;
C[19359] = -4.585844062437e-04;
C[19360] = -1.243281527566e-03;
C[19361] = 1.258243901807e-01;
C[19362] = 6.103429153317e-03;
C[19363] = 2.837655148351e-03;
C[19364] = 5.248966060741e-01;
C[19365] = 3.938544425813e-04;
C[19366] = 2.998404718838e-04;
C[19367] = 3.863307207356e-02;
C[19368] = -7.039379359739e+00;
C[19369] = 5.990503299062e-02;
C[19370] = -1.227213999658e-02;
C[19371] = 2.139891713843e+00;
C[19372] = 1.356757203465e-02;
C[19373] = -9.152399344719e-03;
C[19374] = 7.712881313586e-01;
C[19375] = -5.850892163955e-03;
C[19376] = -5.218327706471e-03;
C[19377] = 7.164936410483e-02;
C[19378] = 1.308547057095e-02;
C[19379] = -2.144535075534e-03;
C[19380] = 3.755380184344e-01;
C[19381] = -1.848226063268e-03;
C[19382] = -1.624910443815e-03;
C[19383] = 4.956299702683e-02;
C[19384] = -4.294939842369e-04;
C[19385] = -2.224622000641e-04;
C[19386] = 7.593678915644e-02;
C[19387] = -4.686634739114e-04;
C[19388] = 1.080172784768e-03;
C[19389] = 2.332047527066e-03;
C[19390] = -1.945439431324e-05;
C[19391] = 5.337865321503e-05;
C[19392] = 5.851164285208e-02;
C[19393] = -7.206741918367e+00;
C[19394] = 1.869185459876e-02;
C[19395] = 1.368910311832e-02;
C[19396] = 2.138708985597e+00;
C[19397] = 1.392175876644e-02;
C[19398] = -5.886265423273e-03;
C[19399] = 8.417541679093e-01;
C[19400] = 4.843303534687e-02;
C[19401] = 1.301580106402e-02;
C[19402] = 2.152858567360e-02;
C[19403] = -1.841348012473e-03;
C[19404] = -1.879008301255e-03;
C[19405] = 4.004107470017e-01;
C[19406] = 9.679118636223e-03;
C[19407] = -4.201858522825e-04;
C[19408] = 5.472980205780e-02;
C[19409] = 3.453464973705e-03;
C[19410] = -3.106289166115e-04;
C[19411] = 8.465104952880e-02;
C[19412] = 6.907021888272e-03;
C[19413] = -1.234747149786e-05;
C[19414] = 2.667700699618e-03;
C[19415] = 2.510875187638e-04;
C[19416] = -1.227415268147e-02;
C[19417] = 1.746434368365e-02;
C[19418] = -7.044385578799e+00;
C[19419] = -1.018710044085e-02;
C[19420] = 2.072157164638e-02;
C[19421] = 2.091632938209e+00;
C[19422] = -5.041109777162e-03;
C[19423] = 4.614160175976e-02;
C[19424] = 8.335400654076e-01;
C[19425] = -2.144635709779e-03;
C[19426] = -1.902723558228e-03;
C[19427] = 6.902756440146e-02;
C[19428] = -1.705661801885e-03;
C[19429] = 1.015908197838e-02;
C[19430] = 3.750405976893e-01;
C[19431] = -1.972585951933e-04;
C[19432] = 3.250623415338e-03;
C[19433] = 5.399389363327e-02;
C[19434] = 1.294967781589e-03;
C[19435] = 6.253075307813e-03;
C[19436] = 8.674514955568e-02;
C[19437] = 6.232821686564e-05;
C[19438] = 2.226967099814e-04;
C[19439] = 2.715839905488e-03;
C[19440] = 2.139088961033e+00;
C[19441] = 8.616725855288e-03;
C[19442] = -1.190771525332e-02;
C[19443] = -6.199568726696e+00;
C[19444] = 1.133894400851e-02;
C[19445] = 7.994801594948e-04;
C[19446] = 2.132673268978e+00;
C[19447] = 5.119596049253e-03;
C[19448] = -8.851238228395e-03;
C[19449] = 3.754978807939e-01;
C[19450] = -2.132627164407e-03;
C[19451] = -1.791692542508e-03;
C[19452] = 1.316504357916e-01;
C[19453] = 2.726594056565e-03;
C[19454] = -4.984571887830e-04;
C[19455] = 3.792852016787e-01;
C[19456] = 1.006006632543e-03;
C[19457] = 1.078691967463e-03;
C[19458] = 9.818487854057e-01;
C[19459] = 6.059331871412e-03;
C[19460] = 1.900828936276e-02;
C[19461] = 5.794065128382e-02;
C[19462] = 3.483119853597e-04;
C[19463] = 1.090160920458e-03;
C[19464] = 1.125581603515e-02;
C[19465] = 2.131414199172e+00;
C[19466] = 1.071848829102e-02;
C[19467] = 9.772060885343e-03;
C[19468] = -6.274159058481e+00;
C[19469] = -4.423457760565e-02;
C[19470] = 4.877653691300e-03;
C[19471] = 2.075542638072e+00;
C[19472] = -1.610323482079e-02;
C[19473] = -1.963813863243e-03;
C[19474] = 4.000460076805e-01;
C[19475] = 9.658927810611e-03;
C[19476] = 2.648249900407e-03;
C[19477] = 9.867745508904e-02;
C[19478] = -2.026369297558e-02;
C[19479] = 9.822221385541e-04;
C[19480] = 3.785884391714e-01;
C[19481] = 7.346685711465e-03;
C[19482] = 6.334446741674e-03;
C[19483] = 9.834178677485e-01;
C[19484] = 3.637041958846e-02;
C[19485] = 3.643714224022e-04;
C[19486] = 5.834295749188e-02;
C[19487] = 2.441734814282e-03;
C[19488] = -1.021607744586e-02;
C[19489] = 5.948366725812e-03;
C[19490] = 2.094237665704e+00;
C[19491] = -5.257221310460e-04;
C[19492] = -4.604711636827e-02;
C[19493] = -6.269865423691e+00;
C[19494] = -9.341443121519e-03;
C[19495] = -1.205714285246e-02;
C[19496] = 2.142369339141e+00;
C[19497] = -1.678094348872e-03;
C[19498] = 9.280449034192e-03;
C[19499] = 3.751708340641e-01;
C[19500] = -5.647173033101e-04;
C[19501] = -2.035431991371e-02;
C[19502] = 1.038465512977e-01;
C[19503] = 1.038437343972e-03;
C[19504] = 7.637790348443e-03;
C[19505] = 4.004757230648e-01;
C[19506] = 1.976833927794e-02;
C[19507] = 3.646907436296e-02;
C[19508] = 1.089344513872e+00;
C[19509] = 1.131785926583e-03;
C[19510] = 2.435042470281e-03;
C[19511] = 6.478003045668e-02;
C[19512] = 7.854601845371e-01;
C[19513] = -7.043984417132e-03;
C[19514] = -1.552782403236e-03;
C[19515] = 2.117042234569e+00;
C[19516] = 8.533380105007e-03;
C[19517] = -2.216230015769e-03;
C[19518] = -7.100718094948e+00;
C[19519] = -5.942776940840e-03;
C[19520] = -2.607045887381e-02;
C[19521] = 5.027159968575e-02;
C[19522] = -4.780718019754e-04;
C[19523] = -2.284222649702e-05;
C[19524] = 3.785036499582e-01;
C[19525] = 1.165008459239e-03;
C[19526] = 1.394697999260e-03;
C[19527] = 5.600211289242e-02;
C[19528] = -1.701088853075e-03;
C[19529] = -5.302965478887e-03;
C[19530] = 4.123230374842e+00;
C[19531] = 7.065187576684e-03;
C[19532] = 2.940845552799e-02;
C[19533] = 5.129365149677e-01;
C[19534] = 1.095427999013e-03;
C[19535] = 3.913443420051e-03;
C[19536] = -7.228564668147e-03;
C[19537] = 8.400766249080e-01;
C[19538] = 4.520197450792e-02;
C[19539] = 9.452813165728e-03;
C[19540] = 2.072421389700e+00;
C[19541] = -1.271116923371e-02;
C[19542] = -5.997392612736e-03;
C[19543] = -7.111495459541e+00;
C[19544] = -4.141192658237e-02;
C[19545] = -4.983776094465e-04;
C[19546] = 5.464592490773e-02;
C[19547] = 3.203642052747e-03;
C[19548] = 1.222667488367e-03;
C[19549] = 3.784323767528e-01;
C[19550] = 7.605089029380e-03;
C[19551] = -1.703819636670e-03;
C[19552] = 5.136339734534e-02;
C[19553] = -1.445815742979e-02;
C[19554] = 6.549883096121e-03;
C[19555] = 4.111123900748e+00;
C[19556] = 4.742617499044e-03;
C[19557] = 1.068750365510e-03;
C[19558] = 5.127721976143e-01;
C[19559] = 3.479680333280e-03;
C[19560] = -1.942987559887e-03;
C[19561] = 4.708833077900e-02;
C[19562] = 8.461326057066e-01;
C[19563] = -6.574882652487e-05;
C[19564] = -1.517129930771e-02;
C[19565] = 2.125815640499e+00;
C[19566] = -2.649794089579e-02;
C[19567] = -4.228886141588e-02;
C[19568] = -7.222339956269e+00;
C[19569] = -5.869519273496e-05;
C[19570] = 3.386229745312e-03;
C[19571] = 5.462352064822e-02;
C[19572] = 1.517966437556e-03;
C[19573] = 7.393282487119e-03;
C[19574] = 3.996480381327e-01;
C[19575] = -5.324339579986e-03;
C[19576] = -1.450200417146e-02;
C[19577] = 1.758350371844e-02;
C[19578] = 2.825931745560e-02;
C[19579] = 5.432579103194e-03;
C[19580] = 4.168762546849e+00;
C[19581] = 3.855042751685e-03;
C[19582] = 3.518874459709e-03;
C[19583] = 5.256080677824e-01;
C[19584] = 0.000000000000e+00;
C[19585] = 0.000000000000e+00;
C[19586] = 0.000000000000e+00;
C[19587] = 0.000000000000e+00;
C[19588] = 0.000000000000e+00;
C[19589] = 0.000000000000e+00;
C[19590] = 0.000000000000e+00;
C[19591] = 0.000000000000e+00;
C[19592] = 0.000000000000e+00;
C[19593] = 0.000000000000e+00;
C[19594] = 0.000000000000e+00;
C[19595] = 0.000000000000e+00;
C[19596] = 0.000000000000e+00;
C[19597] = 0.000000000000e+00;
C[19598] = 0.000000000000e+00;
C[19599] = 0.000000000000e+00;
C[19600] = 0.000000000000e+00;
C[19601] = 0.000000000000e+00;
C[19602] = 1.000000000000e+00;
C[19603] = 0.000000000000e+00;
C[19604] = 0.000000000000e+00;
C[19605] = 2.000000000000e-01;
C[19606] = -0.000000000000e+00;
C[19607] = -0.000000000000e+00;
C[19608] = 0.000000000000e+00;
C[19609] = 0.000000000000e+00;
C[19610] = 0.000000000000e+00;
C[19611] = 0.000000000000e+00;
C[19612] = 0.000000000000e+00;
C[19613] = 0.000000000000e+00;
C[19614] = 0.000000000000e+00;
C[19615] = 0.000000000000e+00;
C[19616] = 0.000000000000e+00;
C[19617] = 0.000000000000e+00;
C[19618] = 0.000000000000e+00;
C[19619] = 0.000000000000e+00;
C[19620] = 0.000000000000e+00;
C[19621] = 0.000000000000e+00;
C[19622] = 0.000000000000e+00;
C[19623] = 0.000000000000e+00;
C[19624] = 0.000000000000e+00;
C[19625] = 0.000000000000e+00;
C[19626] = 0.000000000000e+00;
C[19627] = 1.000000000000e+00;
C[19628] = 0.000000000000e+00;
C[19629] = -0.000000000000e+00;
C[19630] = 2.000000000000e-01;
C[19631] = -0.000000000000e+00;
C[19632] = 0.000000000000e+00;
C[19633] = 0.000000000000e+00;
C[19634] = 0.000000000000e+00;
C[19635] = 0.000000000000e+00;
C[19636] = 0.000000000000e+00;
C[19637] = 0.000000000000e+00;
C[19638] = 0.000000000000e+00;
C[19639] = 0.000000000000e+00;
C[19640] = 0.000000000000e+00;
C[19641] = 0.000000000000e+00;
C[19642] = 0.000000000000e+00;
C[19643] = 0.000000000000e+00;
C[19644] = 0.000000000000e+00;
C[19645] = 0.000000000000e+00;
C[19646] = 0.000000000000e+00;
C[19647] = 0.000000000000e+00;
C[19648] = 0.000000000000e+00;
C[19649] = 0.000000000000e+00;
C[19650] = 0.000000000000e+00;
C[19651] = 0.000000000000e+00;
C[19652] = 1.000000000000e+00;
C[19653] = -0.000000000000e+00;
C[19654] = -0.000000000000e+00;
C[19655] = 2.000000000000e-01;
C[19656] = 8.579760335192e-02;
C[19657] = 1.051427100028e-02;
C[19658] = -4.421351268275e-03;
C[19659] = 3.656860977750e-01;
C[19660] = -2.067483380900e-03;
C[19661] = -1.740792484862e-03;
C[19662] = 4.810832977423e-02;
C[19663] = 2.065890166258e-04;
C[19664] = 4.325800149345e-04;
C[19665] = 1.323593913601e-01;
C[19666] = 5.687563500691e-04;
C[19667] = -3.416954264455e-04;
C[19668] = 2.919871469581e-02;
C[19669] = -8.739188421670e-05;
C[19670] = -7.694601747760e-05;
C[19671] = 2.293071961106e-03;
C[19672] = 1.561491703559e-05;
C[19673] = 2.556588801945e-05;
C[19674] = 2.201108707850e-03;
C[19675] = 3.262132480190e-05;
C[19676] = 4.960397471938e-05;
C[19677] = -2.428998972652e-05;
C[19678] = 1.407526458617e-06;
C[19679] = 1.653992243614e-06;
C[19680] = 1.029732758627e-02;
C[19681] = 2.732244864555e-02;
C[19682] = 3.506854735314e-03;
C[19683] = -2.085088839060e-03;
C[19684] = 4.018482137328e-01;
C[19685] = 1.167601136393e-02;
C[19686] = 2.798448896051e-04;
C[19687] = 5.620204025163e-02;
C[19688] = 3.716345200901e-03;
C[19689] = 5.542934558018e-04;
C[19690] = 1.270019092796e-01;
C[19691] = 2.735532601497e-04;
C[19692] = -8.496273917356e-05;
C[19693] = 3.305862302457e-02;
C[19694] = 1.064582790297e-03;
C[19695] = 1.964997800138e-05;
C[19696] = 2.746415912217e-03;
C[19697] = 1.931185189712e-04;
C[19698] = 3.902689368581e-05;
C[19699] = 2.800877720666e-03;
C[19700] = 2.659277869670e-04;
C[19701] = 1.578900254608e-06;
C[19702] = -1.410823109965e-05;
C[19703] = 1.426385620092e-06;
C[19704] = -4.399037716468e-03;
C[19705] = 3.404910205974e-03;
C[19706] = 7.454410196245e-02;
C[19707] = -1.764182828294e-03;
C[19708] = 1.230812996294e-02;
C[19709] = 3.639732583793e-01;
C[19710] = 4.551669002655e-04;
C[19711] = 3.475328369290e-03;
C[19712] = 5.089626817595e-02;
C[19713] = -3.402078563250e-04;
C[19714] = 2.667569581938e-04;
C[19715] = 1.308819786989e-01;
C[19716] = -7.710351500495e-05;
C[19717] = 1.098251694322e-03;
C[19718] = 2.919914672343e-02;
C[19719] = 2.657124067049e-05;
C[19720] = 1.780602848847e-04;
C[19721] = 2.443434528812e-03;
C[19722] = 4.981053671370e-05;
C[19723] = 2.302928557855e-04;
C[19724] = 2.481716628057e-03;
C[19725] = 1.644137117373e-06;
C[19726] = 5.926456674999e-07;
C[19727] = -2.236402513942e-05;
C[19728] = 3.648088996311e-01;
C[19729] = -1.703529146213e-03;
C[19730] = -1.816835407553e-03;
C[19731] = 1.358878447903e-01;
C[19732] = 7.094718981068e-04;
C[19733] = -2.605260202082e-03;
C[19734] = 3.871364240788e-01;
C[19735] = 1.833808935330e-03;
C[19736] = 3.083672205856e-03;
C[19737] = 2.914023481955e-02;
C[19738] = -5.952542631714e-05;
C[19739] = -8.061368695553e-05;
C[19740] = 1.343124106381e-01;
C[19741] = -6.490393463211e-05;
C[19742] = -2.354800066100e-04;
C[19743] = 3.168898019662e-02;
C[19744] = 1.539505930952e-04;
C[19745] = 2.539108514825e-04;
C[19746] = 6.003129828980e-02;
C[19747] = 3.655511744679e-04;
C[19748] = 6.893455693255e-04;
C[19749] = 2.627605972525e-03;
C[19750] = 1.612298214443e-05;
C[19751] = 2.981876109435e-05;
C[19752] = -1.393826621093e-03;
C[19753] = 3.999217381673e-01;
C[19754] = 1.201980760003e-02;
C[19755] = 6.103234011448e-04;
C[19756] = 9.160997713794e-02;
C[19757] = -2.095599770240e-02;
C[19758] = 1.757243993725e-03;
C[19759] = 3.843504439282e-01;
C[19760] = 6.777072179617e-03;
C[19761] = -4.248143356289e-05;
C[19762] = 3.293019132021e-02;
C[19763] = 1.079030203461e-03;
C[19764] = -7.151383442957e-05;
C[19765] = 1.298723171282e-01;
C[19766] = -1.875365327906e-03;
C[19767] = 1.500406222395e-04;
C[19768] = 3.173417733846e-02;
C[19769] = 6.859470338062e-04;
C[19770] = 3.579483869523e-04;
C[19771] = 6.073750350681e-02;
C[19772] = 2.311144089568e-03;
C[19773] = 1.591245164402e-05;
C[19774] = 2.678826123615e-03;
C[19775] = 1.113895792905e-04;
C[19776] = -1.717701259114e-03;
C[19777] = 1.152295287917e-02;
C[19778] = 3.652244893238e-01;
C[19779] = -2.645438777393e-03;
C[19780] = -2.112710131055e-02;
C[19781] = 1.157268884131e-01;
C[19782] = 3.074275093531e-03;
C[19783] = 7.098345605805e-03;
C[19784] = 4.010699256947e-01;
C[19785] = -7.540660242775e-05;
C[19786] = 1.054378891313e-03;
C[19787] = 2.928256211972e-02;
C[19788] = -2.381585782974e-04;
C[19789] = -1.886772235116e-03;
C[19790] = 1.327081355551e-01;
C[19791] = 2.537173122740e-04;
C[19792] = 7.044509675521e-04;
C[19793] = 3.288424719735e-02;
C[19794] = 6.756665233640e-04;
C[19795] = 2.309976587563e-03;
C[19796] = 6.478562689472e-02;
C[19797] = 2.921906627953e-05;
C[19798] = 1.102629048356e-04;
C[19799] = 2.839739330631e-03;
C[19800] = 4.955353735153e-02;
C[19801] = -4.527322023655e-05;
C[19802] = 2.705270089659e-04;
C[19803] = 3.862376996980e-01;
C[19804] = 2.198546083510e-03;
C[19805] = 3.345648612318e-03;
C[19806] = 4.990488324526e-02;
C[19807] = -2.062832114099e-03;
C[19808] = -4.084442302237e-03;
C[19809] = 2.389419132926e-03;
C[19810] = -2.024562654729e-06;
C[19811] = 1.426191458385e-05;
C[19812] = 3.162906523790e-02;
C[19813] = 1.794607615585e-04;
C[19814] = 2.718088801932e-04;
C[19815] = 1.284604217196e-01;
C[19816] = -1.623615793381e-04;
C[19817] = -2.741412193634e-04;
C[19818] = 5.120255090443e-01;
C[19819] = -2.050081112792e-05;
C[19820] = 4.173573709041e-04;
C[19821] = 3.753708934102e-02;
C[19822] = -1.308888612181e-05;
C[19823] = -1.240646838426e-05;
C[19824] = -9.413386789191e-05;
C[19825] = 5.659111277150e-02;
C[19826] = 3.313593844383e-03;
C[19827] = 2.247058190192e-03;
C[19828] = 3.843921349341e-01;
C[19829] = 6.997859701537e-03;
C[19830] = -2.050890260456e-03;
C[19831] = 4.318467904853e-02;
C[19832] = -1.451215321530e-02;
C[19833] = -4.433275265587e-06;
C[19834] = 2.772354080209e-03;
C[19835] = 1.672779832242e-04;
C[19836] = 1.815005434194e-04;
C[19837] = 3.173695673885e-02;
C[19838] = 6.977519072675e-04;
C[19839] = -1.615654557618e-04;
C[19840] = 1.278545594787e-01;
C[19841] = -1.240685768255e-03;
C[19842] = -2.072206636404e-05;
C[19843] = 5.129298939781e-01;
C[19844] = 4.117321063253e-03;
C[19845] = -1.313804379379e-05;
C[19846] = 3.764240834632e-02;
C[19847] = 3.735272482383e-04;
C[19848] = 2.428596715009e-04;
C[19849] = 3.484961613245e-03;
C[19850] = 5.206821968148e-02;
C[19851] = 3.354427077607e-03;
C[19852] = 6.783038358913e-03;
C[19853] = 4.000604372188e-01;
C[19854] = -4.066991106535e-03;
C[19855] = -1.454166133197e-02;
C[19856] = 1.896133126788e-02;
C[19857] = 1.291786512389e-05;
C[19858] = 1.776929464608e-04;
C[19859] = 2.521564629181e-03;
C[19860] = 2.719611762659e-04;
C[19861] = 6.863447790927e-04;
C[19862] = 3.281694796563e-02;
C[19863] = -2.729778063166e-04;
C[19864] = -1.242652976034e-03;
C[19865] = 1.259690370205e-01;
C[19866] = 4.233019711286e-04;
C[19867] = 4.153047743808e-03;
C[19868] = 5.265055038754e-01;
C[19869] = -1.229243599977e-05;
C[19870] = 3.753230092113e-04;
C[19871] = 3.871504798478e-02;
C[19872] = -6.961607983132e+00;
C[19873] = 3.625409956112e-02;
C[19874] = -2.600793749086e-02;
C[19875] = 2.099045801011e+00;
C[19876] = -1.551122412743e-04;
C[19877] = -1.562925118361e-02;
C[19878] = 7.450031049917e-01;
C[19879] = 5.552366443032e-03;
C[19880] = 6.351535300934e-03;
C[19881] = 9.728302497446e-02;
C[19882] = 3.205481500511e-03;
C[19883] = -4.138981909423e-03;
C[19884] = 3.579321834940e-01;
C[19885] = 5.338571337500e-04;
C[19886] = -8.323808683487e-04;
C[19887] = 4.790171897312e-02;
C[19888] = 3.756261254490e-04;
C[19889] = 4.389333177057e-04;
C[19890] = 7.370374706981e-02;
C[19891] = 7.112386980380e-04;
C[19892] = 1.031597275210e-03;
C[19893] = 2.266729114098e-03;
C[19894] = 2.576397337009e-05;
C[19895] = 4.032861960003e-05;
C[19896] = 3.191523128093e-02;
C[19897] = -7.147076524483e+00;
C[19898] = 6.476165350455e-02;
C[19899] = 5.774110898214e-04;
C[19900] = 2.162389846531e+00;
C[19901] = 3.601168858786e-02;
C[19902] = 6.417092545635e-03;
C[19903] = 8.651332691869e-01;
C[19904] = 5.347050684100e-02;
C[19905] = 2.988538086501e-03;
C[19906] = 3.624396864811e-02;
C[19907] = 6.247672524871e-03;
C[19908] = 5.515146925254e-04;
C[19909] = 4.028414113390e-01;
C[19910] = 1.196540418629e-02;
C[19911] = 4.325881368510e-04;
C[19912] = 5.653152011180e-02;
C[19913] = 3.633597772037e-03;
C[19914] = 8.264102074775e-04;
C[19915] = 8.891438668314e-02;
C[19916] = 7.291926947183e-03;
C[19917] = 2.998747826853e-05;
C[19918] = 2.838786549897e-03;
C[19919] = 2.490381788609e-04;
C[19920] = -2.556166645471e-02;
C[19921] = 6.272276291776e-02;
C[19922] = -7.056456358212e+00;
C[19923] = -1.557645457732e-02;
C[19924] = 4.538931831037e-02;
C[19925] = 2.024437081673e+00;
C[19926] = 6.522215179137e-03;
C[19927] = 5.046603890823e-02;
C[19928] = 7.807683372436e-01;
C[19929] = -4.116668357616e-03;
C[19930] = 6.145727995531e-03;
C[19931] = 7.266504813550e-02;
C[19932] = -8.338686367126e-04;
C[19933] = 1.258675102322e-02;
C[19934] = 3.553892251822e-01;
C[19935] = 4.495256279869e-04;
C[19936] = 3.396368413983e-03;
C[19937] = 5.041429538504e-02;
C[19938] = 1.032078473641e-03;
C[19939] = 6.508746792852e-03;
C[19940] = 8.101402109621e-02;
C[19941] = 4.044873721829e-05;
C[19942] = 2.178800297097e-04;
C[19943] = 2.519722283476e-03;
C[19944] = 2.081501838134e+00;
C[19945] = 8.208604946749e-03;
C[19946] = -1.662950616250e-02;
C[19947] = -6.172289657005e+00;
C[19948] = 7.571478374281e-04;
C[19949] = -6.830988466489e-03;
C[19950] = 2.188489309018e+00;
C[19951] = 1.720928755206e-03;
C[19952] = 3.244897846618e-03;
C[19953] = 3.570549853501e-01;
C[19954] = 9.330743853717e-04;
C[19955] = -8.865212159712e-04;
C[19956] = 1.388743908382e-01;
C[19957] = -1.990123567241e-03;
C[19958] = -2.573132019654e-03;
C[19959] = 3.922428923279e-01;
C[19960] = 1.497681261875e-03;
C[19961] = 2.301122970819e-03;
C[19962] = 1.020693251820e+00;
C[19963] = 5.513178044970e-03;
C[19964] = 9.752194961006e-03;
C[19965] = 6.030555019157e-02;
C[19966] = 3.287549551868e-04;
C[19967] = 5.425013619536e-04;
C[19968] = 1.331802295487e-02;
C[19969] = 2.123860335221e+00;
C[19970] = 3.962287105202e-02;
C[19971] = -1.225822101812e-03;
C[19972] = -6.296529204362e+00;
C[19973] = -4.543801288316e-02;
C[19974] = 7.265265910069e-04;
C[19975] = 2.083140993803e+00;
C[19976] = -2.286457069400e-02;
C[19977] = 1.207513893557e-03;
C[19978] = 4.009149357735e-01;
C[19979] = 1.229842866031e-02;
C[19980] = -2.089272064203e-03;
C[19981] = 8.606775180273e-02;
C[19982] = -1.963754425065e-02;
C[19983] = 1.439104253727e-03;
C[19984] = 3.892230660965e-01;
C[19985] = 5.419552510006e-03;
C[19986] = 5.353046327148e-03;
C[19987] = 1.026761906061e+00;
C[19988] = 3.378098304873e-02;
C[19989] = 3.194386091151e-04;
C[19990] = 6.106397982856e-02;
C[19991] = 2.241609108701e-03;
C[19992] = -1.516742666866e-02;
C[19993] = 3.295051889258e-02;
C[19994] = 2.049461700562e+00;
C[19995] = -7.634559972702e-03;
C[19996] = -4.886008504619e-02;
C[19997] = -6.220840393225e+00;
C[19998] = 3.291085545265e-03;
C[19999] = -1.819026187892e-02;
C[20000] = 2.178871414533e+00;
C[20001] = -8.092896426010e-04;
C[20002] = 1.181234570152e-02;
C[20003] = 3.566404561267e-01;
C[20004] = -2.613310594964e-03;
C[20005] = -1.980864785880e-02;
C[20006] = 1.237131298075e-01;
C[20007] = 2.309934694562e-03;
C[20008] = 5.744494318552e-03;
C[20009] = 4.015366163959e-01;
C[20010] = 9.522233527295e-03;
C[20011] = 3.403207660219e-02;
C[20012] = 1.091775829534e+00;
C[20013] = 5.239529128504e-04;
C[20014] = 2.236102055587e-03;
C[20015] = 6.474682073927e-02;
C[20016] = 7.739072565377e-01;
C[20017] = -8.526965119741e-05;
C[20018] = 2.829417353145e-03;
C[20019] = 2.170514821403e+00;
C[20020] = 9.552568386708e-03;
C[20021] = 8.718555921021e-03;
C[20022] = -7.120166228238e+00;
C[20023] = -2.812808423845e-03;
C[20024] = -5.219622076003e-03;
C[20025] = 4.934692655042e-02;
C[20026] = 1.074700270094e-04;
C[20027] = 2.648857366872e-04;
C[20028] = 3.913441679471e-01;
C[20029] = 1.880406343512e-03;
C[20030] = 2.581308213350e-03;
C[20031] = 4.462580259795e-02;
C[20032] = -1.367939416371e-03;
C[20033] = -1.740539824863e-03;
C[20034] = 4.103330068439e+00;
C[20035] = -5.204933222299e-03;
C[20036] = -7.862464289318e-03;
C[20037] = 5.124867685270e-01;
C[20038] = -5.736191888167e-04;
C[20039] = -1.164799675316e-03;
C[20040] = -4.620912473216e-04;
C[20041] = 8.729147195843e-01;
C[20042] = 4.723134841009e-02;
C[20043] = 9.985913852461e-03;
C[20044] = 2.083974813922e+00;
C[20045] = -2.019997996428e-02;
C[20046] = -2.573971350979e-03;
C[20047] = -7.147920927209e+00;
C[20048] = -4.321735812321e-02;
C[20049] = 7.490324093129e-05;
C[20050] = 5.692059263167e-02;
C[20051] = 3.234633889076e-03;
C[20052] = 1.910930516738e-03;
C[20053] = 3.892647571024e-01;
C[20054] = 5.644008414284e-03;
C[20055] = -1.355997562727e-03;
C[20056] = 3.737844871197e-02;
C[20057] = -1.319255539692e-02;
C[20058] = -5.208012755995e-03;
C[20059] = 4.100826369677e+00;
C[20060] = 1.415472447258e-02;
C[20061] = -5.705830856843e-04;
C[20062] = 5.135336430590e-01;
C[20063] = 4.224246011521e-03;
C[20064] = 2.557128432264e-03;
C[20065] = 4.884283508787e-02;
C[20066] = 8.042073673542e-01;
C[20067] = 8.659995281636e-03;
C[20068] = -2.274524710806e-02;
C[20069] = 2.158681645015e+00;
C[20070] = -4.870598161961e-03;
C[20071] = -4.380752045671e-02;
C[20072] = -7.219072982703e+00;
C[20073] = 2.492129742722e-04;
C[20074] = 3.402214184381e-03;
C[20075] = 5.158624689057e-02;
C[20076] = 2.571877842569e-03;
C[20077] = 5.425518689303e-03;
C[20078] = 4.005271279201e-01;
C[20079] = -1.723088629161e-03;
C[20080] = -1.322206351359e-02;
C[20081] = 1.914787696129e-02;
C[20082] = -7.787039932875e-03;
C[20083] = 1.474344383455e-02;
C[20084] = 4.179505989902e+00;
C[20085] = -1.153965847741e-03;
C[20086] = 4.261398740784e-03;
C[20087] = 5.264268330588e-01;
C[20088] = 0.000000000000e+00;
C[20089] = 0.000000000000e+00;
C[20090] = 0.000000000000e+00;
C[20091] = 0.000000000000e+00;
C[20092] = 0.000000000000e+00;
C[20093] = 0.000000000000e+00;
C[20094] = 0.000000000000e+00;
C[20095] = 0.000000000000e+00;
C[20096] = 0.000000000000e+00;
C[20097] = 0.000000000000e+00;
C[20098] = 0.000000000000e+00;
C[20099] = 0.000000000000e+00;
C[20100] = 0.000000000000e+00;
C[20101] = 0.000000000000e+00;
C[20102] = 0.000000000000e+00;
C[20103] = 0.000000000000e+00;
C[20104] = 0.000000000000e+00;
C[20105] = 0.000000000000e+00;
C[20106] = 1.000000000000e+00;
C[20107] = 0.000000000000e+00;
C[20108] = 0.000000000000e+00;
C[20109] = 2.000000000000e-01;
C[20110] = -0.000000000000e+00;
C[20111] = -0.000000000000e+00;
C[20112] = 0.000000000000e+00;
C[20113] = 0.000000000000e+00;
C[20114] = 0.000000000000e+00;
C[20115] = 0.000000000000e+00;
C[20116] = 0.000000000000e+00;
C[20117] = 0.000000000000e+00;
C[20118] = 0.000000000000e+00;
C[20119] = 0.000000000000e+00;
C[20120] = 0.000000000000e+00;
C[20121] = 0.000000000000e+00;
C[20122] = 0.000000000000e+00;
C[20123] = 0.000000000000e+00;
C[20124] = 0.000000000000e+00;
C[20125] = 0.000000000000e+00;
C[20126] = 0.000000000000e+00;
C[20127] = 0.000000000000e+00;
C[20128] = 0.000000000000e+00;
C[20129] = 0.000000000000e+00;
C[20130] = 0.000000000000e+00;
C[20131] = 1.000000000000e+00;
C[20132] = 0.000000000000e+00;
C[20133] = -0.000000000000e+00;
C[20134] = 2.000000000000e-01;
C[20135] = -0.000000000000e+00;
C[20136] = 0.000000000000e+00;
C[20137] = 0.000000000000e+00;
C[20138] = 0.000000000000e+00;
C[20139] = 0.000000000000e+00;
C[20140] = 0.000000000000e+00;
C[20141] = 0.000000000000e+00;
C[20142] = 0.000000000000e+00;
C[20143] = 0.000000000000e+00;
C[20144] = 0.000000000000e+00;
C[20145] = 0.000000000000e+00;
C[20146] = 0.000000000000e+00;
C[20147] = 0.000000000000e+00;
C[20148] = 0.000000000000e+00;
C[20149] = 0.000000000000e+00;
C[20150] = 0.000000000000e+00;
C[20151] = 0.000000000000e+00;
C[20152] = 0.000000000000e+00;
C[20153] = 0.000000000000e+00;
C[20154] = 0.000000000000e+00;
C[20155] = 0.000000000000e+00;
C[20156] = 1.000000000000e+00;
C[20157] = -0.000000000000e+00;
C[20158] = -0.000000000000e+00;
C[20159] = 2.000000000000e-01;
C[20160] = 9.040148457510e-02;
C[20161] = -5.743357855086e-03;
C[20162] = 1.458612615264e-03;
C[20163] = 3.608409158453e-01;
C[20164] = 4.852281059119e-03;
C[20165] = 1.847591970807e-03;
C[20166] = 5.045394080685e-02;
C[20167] = 1.255554907323e-03;
C[20168] = 6.037672374945e-04;
C[20169] = 1.316539249861e-01;
C[20170] = -6.350909106931e-04;
C[20171] = 2.189449538388e-04;
C[20172] = 2.956540495349e-02;
C[20173] = 5.669264623904e-04;
C[20174] = 2.012097304166e-04;
C[20175] = 2.454884314756e-03;
C[20176] = 7.131738593121e-05;
C[20177] = 3.171042182645e-05;
C[20178] = 2.477967272842e-03;
C[20179] = 9.178572196991e-05;
C[20180] = 2.979167714615e-05;
C[20181] = -1.479139693308e-05;
C[20182] = 1.911382247668e-06;
C[20183] = 4.801820207479e-07;
C[20184] = -5.877232997658e-03;
C[20185] = 4.540910463453e-02;
C[20186] = -1.228407309180e-03;
C[20187] = 4.973935130713e-03;
C[20188] = 4.020075695422e-01;
C[20189] = 1.041010187131e-02;
C[20190] = 1.262455491396e-03;
C[20191] = 5.719265386100e-02;
C[20192] = 3.055498001073e-03;
C[20193] = -6.440159201980e-04;
C[20194] = 1.281492223033e-01;
C[20195] = -4.299765562383e-04;
C[20196] = 5.752767360971e-04;
C[20197] = 3.294031180848e-02;
C[20198] = 9.328021711039e-04;
C[20199] = 7.103487437169e-05;
C[20200] = 2.785748649000e-03;
C[20201] = 1.588062384336e-04;
C[20202] = 8.799818796477e-05;
C[20203] = 2.833335813152e-03;
C[20204] = 2.251608622651e-04;
C[20205] = 1.774502870977e-06;
C[20206] = -1.630928859565e-05;
C[20207] = 1.043161014134e-06;
C[20208] = 1.488942567438e-03;
C[20209] = -1.177253579681e-03;
C[20210] = 5.662056969800e-02;
C[20211] = 1.907003968716e-03;
C[20212] = 1.050627892320e-02;
C[20213] = 3.578166862312e-01;
C[20214] = 5.756320615812e-04;
C[20215] = 2.889314362000e-03;
C[20216] = 5.213643565062e-02;
C[20217] = 2.209669506505e-04;
C[20218] = -4.265663076050e-04;
C[20219] = 1.290761757886e-01;
C[20220] = 2.046324165633e-04;
C[20221] = 9.353307507054e-04;
C[20222] = 2.948411749006e-02;
C[20223] = 2.961397307775e-05;
C[20224] = 1.498595541269e-04;
C[20225] = 2.555449755835e-03;
C[20226] = 2.272706856957e-05;
C[20227] = 2.064994679946e-04;
C[20228] = 2.713587740732e-03;
C[20229] = 2.822941275075e-07;
C[20230] = 9.183528563255e-07;
C[20231] = -1.118505741138e-05;
C[20232] = 3.612640336849e-01;
C[20233] = 4.996418920330e-03;
C[20234] = 1.652834474588e-03;
C[20235] = 1.322200444101e-01;
C[20236] = -5.757436803766e-03;
C[20237] = -2.243401912304e-03;
C[20238] = 3.923408628225e-01;
C[20239] = -3.488723020684e-04;
C[20240] = 8.544896488495e-04;
C[20241] = 2.959361280946e-02;
C[20242] = 5.767756554050e-04;
C[20243] = 1.876877836215e-04;
C[20244] = 1.334227131093e-01;
C[20245] = -5.791578514915e-04;
C[20246] = -1.677380994202e-04;
C[20247] = 3.199246752016e-02;
C[20248] = -5.593230557517e-05;
C[20249] = -5.454537352852e-06;
C[20250] = 6.067978920303e-02;
C[20251] = -2.418545286644e-04;
C[20252] = -3.651640122599e-04;
C[20253] = 2.646961162078e-03;
C[20254] = -1.270046147681e-05;
C[20255] = -2.172099900637e-05;
C[20256] = 5.100726015967e-03;
C[20257] = 4.021738815023e-01;
C[20258] = 1.141852248596e-02;
C[20259] = -5.852788616495e-03;
C[20260] = 8.484295924345e-02;
C[20261] = -1.859840223722e-02;
C[20262] = -3.643713816795e-04;
C[20263] = 3.912996277844e-01;
C[20264] = 4.760336676469e-03;
C[20265] = 5.834894595136e-04;
C[20266] = 3.295139927248e-02;
C[20267] = 9.961469882230e-04;
C[20268] = -5.855146390068e-04;
C[20269] = 1.295571923073e-01;
C[20270] = -1.665007851332e-03;
C[20271] = -5.646713465934e-05;
C[20272] = 3.215352615868e-02;
C[20273] = 5.447160203973e-04;
C[20274] = -2.460055866365e-04;
C[20275] = 6.160773406575e-02;
C[20276] = 2.088961509293e-03;
C[20277] = -1.282920300921e-05;
C[20278] = 2.705940097191e-03;
C[20279] = 1.033618637398e-04;
C[20280] = 1.647596896893e-03;
C[20281] = 1.112683303084e-02;
C[20282] = 3.577820951717e-01;
C[20283] = -2.250698536853e-03;
C[20284] = -1.851586540584e-02;
C[20285] = 1.189359258274e-01;
C[20286] = 8.679815760163e-04;
C[20287] = 4.950422886819e-03;
C[20288] = 4.006394928211e-01;
C[20289] = 1.878767254890e-04;
C[20290] = 9.805842484055e-04;
C[20291] = 2.948181141942e-02;
C[20292] = -1.682245410568e-04;
C[20293] = -1.659505395906e-03;
C[20294] = 1.322584040054e-01;
C[20295] = -4.805469359525e-06;
C[20296] = 5.544614514282e-04;
C[20297] = 3.287099342050e-02;
C[20298] = -3.692066788039e-04;
C[20299] = 2.081401501397e-03;
C[20300] = 6.471858883586e-02;
C[20301] = -2.182268939140e-05;
C[20302] = 1.022724764167e-04;
C[20303] = 2.838464647847e-03;
C[20304] = 4.941674788362e-02;
C[20305] = 7.910064525032e-04;
C[20306] = 1.813437984414e-04;
C[20307] = 3.934217099719e-01;
C[20308] = -4.957476382667e-05;
C[20309] = 9.756951313239e-04;
C[20310] = 4.183786841205e-02;
C[20311] = -9.663865715374e-05;
C[20312] = 6.451415539003e-04;
C[20313] = 2.385738119874e-03;
C[20314] = 3.960493844552e-05;
C[20315] = 3.328088868424e-06;
C[20316] = 3.206452399678e-02;
C[20317] = -3.548069346915e-05;
C[20318] = 2.375434327653e-06;
C[20319] = 1.279262947533e-01;
C[20320] = 5.142088168019e-05;
C[20321] = 1.339727026423e-04;
C[20322] = 5.130898300377e-01;
C[20323] = -6.897913362172e-04;
C[20324] = -1.792956391218e-03;
C[20325] = 3.764335840996e-02;
C[20326] = -5.552902734134e-05;
C[20327] = -1.405981743714e-04;
C[20328] = 8.101801665333e-04;
C[20329] = 5.736121920977e-02;
C[20330] = 3.036858174675e-03;
C[20331] = -9.594276974084e-05;
C[20332] = 3.910456500134e-01;
C[20333] = 4.846695532958e-03;
C[20334] = -8.074663637383e-05;
C[20335] = 3.457514732311e-02;
C[20336] = -1.250215088317e-02;
C[20337] = 4.162573654520e-05;
C[20338] = 2.796986338919e-03;
C[20339] = 1.596958083053e-04;
C[20340] = -3.907033675333e-05;
C[20341] = 3.213659430728e-02;
C[20342] = 5.475462945041e-04;
C[20343] = 5.248034973218e-05;
C[20344] = 1.273066842523e-01;
C[20345] = -1.104567388434e-03;
C[20346] = -6.826163741958e-04;
C[20347] = 5.143058780624e-01;
C[20348] = 4.527169988555e-03;
C[20349] = -5.530525238578e-05;
C[20350] = 3.777514378154e-02;
C[20351] = 3.989596544418e-04;
C[20352] = 2.017760447922e-04;
C[20353] = 3.206252370151e-03;
C[20354] = 5.115185910964e-02;
C[20355] = 9.089396593101e-04;
C[20356] = 4.637931503157e-03;
C[20357] = 4.017440916901e-01;
C[20358] = 6.776307443525e-04;
C[20359] = -1.249941422923e-02;
C[20360] = 1.876207395169e-02;
C[20361] = 4.911008979626e-06;
C[20362] = 1.688565297054e-04;
C[20363] = 2.489811319770e-03;
C[20364] = -1.824536655475e-06;
C[20365] = 5.365556755098e-04;
C[20366] = 3.294463334510e-02;
C[20367] = 1.361386486725e-04;
C[20368] = -1.104384944838e-03;
C[20369] = 1.259274030123e-01;
C[20370] = -1.783824423404e-03;
C[20371] = 4.554849452337e-03;
C[20372] = 5.257614044086e-01;
C[20373] = -1.404251741320e-04;
C[20374] = 4.002300896210e-04;
C[20375] = 3.865399251730e-02;
C[20376] = -6.992982783966e+00;
C[20377] = -1.599827346476e-02;
C[20378] = 2.276986371759e-02;
C[20379] = 2.084276389072e+00;
C[20380] = 1.028828030946e-02;
C[20381] = 1.511335781787e-02;
C[20382] = 7.907245625861e-01;
C[20383] = 1.970493194443e-02;
C[20384] = 1.002482994434e-02;
C[20385] = 7.647816288230e-02;
C[20386] = -8.069636970619e-03;
C[20387] = 4.448387360077e-03;
C[20388] = 3.695599817758e-01;
C[20389] = 7.254948528270e-03;
C[20390] = 3.117494800744e-03;
C[20391] = 5.101298535293e-02;
C[20392] = 1.335996693999e-03;
C[20393] = 5.647661310898e-04;
C[20394] = 7.907532692391e-02;
C[20395] = 2.073994420640e-03;
C[20396] = 4.695311912398e-04;
C[20397] = 2.452221145588e-03;
C[20398] = 7.195985767421e-05;
C[20399] = 5.349488707415e-06;
C[20400] = -1.867577631621e-02;
C[20401] = -7.077038889831e+00;
C[20402] = 1.401674736508e-02;
C[20403] = 1.277654083069e-02;
C[20404] = 2.156573122774e+00;
C[20405] = 1.197848811997e-02;
C[20406] = 1.968900247602e-02;
C[20407] = 8.779693114032e-01;
C[20408] = 4.232344492592e-02;
C[20409] = -8.203512113191e-03;
C[20410] = 4.577274877591e-02;
C[20411] = -6.419622748081e-03;
C[20412] = 7.390863576481e-03;
C[20413] = 4.006256454656e-01;
C[20414] = 9.283417811937e-03;
C[20415] = 1.313692481475e-03;
C[20416] = 5.710979499548e-02;
C[20417] = 2.992441219019e-03;
C[20418] = 1.947385980477e-03;
C[20419] = 8.978061071617e-02;
C[20420] = 6.409134125552e-03;
C[20421] = 6.553103256627e-05;
C[20422] = 2.837738063579e-03;
C[20423] = 2.300858907004e-04;
C[20424] = 2.337646276109e-02;
C[20425] = 1.503982195507e-02;
C[20426] = -7.165547697711e+00;
C[20427] = 1.609933958619e-02;
C[20428] = 1.218904388709e-02;
C[20429] = 2.010255978491e+00;
C[20430] = 9.552388788837e-03;
C[20431] = 4.039112337369e-02;
C[20432] = 8.102136821400e-01;
C[20433] = 4.478717312251e-03;
C[20434] = -6.368469018581e-03;
C[20435] = 4.719976622622e-02;
C[20436] = 3.174210503587e-03;
C[20437] = 9.317020204302e-03;
C[20438] = 3.658711984495e-01;
C[20439] = 5.220438981829e-04;
C[20440] = 2.849799549676e-03;
C[20441] = 5.265861276176e-02;
C[20442] = 3.083047934280e-04;
C[20443] = 5.925189204580e-03;
C[20444] = 8.511030696833e-02;
C[20445] = -1.103581789274e-06;
C[20446] = 2.119936761586e-04;
C[20447] = 2.688365136069e-03;
C[20448] = 2.092738745864e+00;
C[20449] = 1.322621662303e-02;
C[20450] = 1.101594970363e-02;
C[20451] = -6.200940409903e+00;
C[20452] = -1.443515700071e-02;
C[20453] = -2.914735812826e-03;
C[20454] = 2.164212337024e+00;
C[20455] = -1.609029680318e-02;
C[20456] = -1.504921746581e-03;
C[20457] = 3.699830996154e-01;
C[20458] = 7.413347366098e-03;
C[20459] = 2.920041009459e-03;
C[20460] = 1.228381904500e-01;
C[20461] = -6.819847963688e-03;
C[20462] = -1.308108005138e-03;
C[20463] = 3.921495924321e-01;
C[20464] = -1.758150418871e-03;
C[20465] = -1.045758508858e-03;
C[20466] = 1.026877953815e+00;
C[20467] = -5.679889601745e-03;
C[20468] = -7.645860760379e-03;
C[20469] = 6.076816792150e-02;
C[20470] = -3.134768127386e-04;
C[20471] = -4.765245147831e-04;
C[20472] = 1.525717944643e-02;
C[20473] = 2.159899361975e+00;
C[20474] = 3.043391514237e-02;
C[20475] = -1.634219325529e-02;
C[20476] = -6.301898969487e+00;
C[20477] = -4.271741250750e-02;
C[20478] = -1.630879766421e-02;
C[20479] = 2.094043854115e+00;
C[20480] = -2.366975354935e-02;
C[20481] = 7.503393485118e-03;
C[20482] = 4.007919574257e-01;
C[20483] = 1.022926376707e-02;
C[20484] = -6.915199776417e-03;
C[20485] = 8.617428893689e-02;
C[20486] = -1.752207001278e-02;
C[20487] = -1.747292283170e-03;
C[20488] = 3.912891597399e-01;
C[20489] = 4.515198437082e-03;
C[20490] = -5.641499859014e-03;
C[20491] = 1.036704668416e+00;
C[20492] = 3.126374495961e-02;
C[20493] = -3.114193926825e-04;
C[20494] = 6.165001533435e-02;
C[20495] = 2.078791264744e-03;
C[20496] = 1.111345633958e-02;
C[20497] = 2.631311131046e-02;
C[20498] = 2.009564157301e+00;
C[20499] = -3.060668303803e-03;
C[20500] = -4.106667587973e-02;
C[20501] = -6.238453214062e+00;
C[20502] = -1.448991514781e-03;
C[20503] = -2.142119137279e-02;
C[20504] = 2.151727189176e+00;
C[20505] = 2.917499726830e-03;
C[20506] = 1.000014897146e-02;
C[20507] = 3.658366073900e-01;
C[20508] = -1.315404629687e-03;
C[20509] = -1.743953318139e-02;
C[20510] = 1.103352733436e-01;
C[20511] = -1.024785879090e-03;
C[20512] = 4.676150932272e-03;
C[20513] = 4.005970544904e-01;
C[20514] = -7.576601951800e-03;
C[20515] = 3.136801976018e-02;
C[20516] = 1.090404747997e+00;
C[20517] = -4.738376068173e-04;
C[20518] = 2.074247951701e-03;
C[20519] = 6.475508062477e-02;
C[20520] = 7.699807041216e-01;
C[20521] = 1.026002169817e-02;
C[20522] = 1.666623526041e-03;
C[20523] = 2.185829280011e+00;
C[20524] = -1.001286530715e-02;
C[20525] = 7.052795913722e-04;
C[20526] = -7.137234998126e+00;
C[20527] = 1.756638680823e-03;
C[20528] = 8.038399227427e-03;
C[20529] = 4.997579242971e-02;
C[20530] = 8.422434425823e-04;
C[20531] = 1.277556350431e-04;
C[20532] = 3.932304395814e-01;
C[20533] = -1.432495665317e-03;
C[20534] = -9.170723237820e-04;
C[20535] = 4.087236011191e-02;
C[20536] = 1.247462876406e-03;
C[20537] = 2.581035510926e-03;
C[20538] = 4.107465293470e+00;
C[20539] = -3.675329374245e-03;
C[20540] = -1.029562486235e-02;
C[20541] = 5.136086808232e-01;
C[20542] = -7.288705280534e-04;
C[20543] = -1.802113664704e-03;
C[20544] = 1.079743712862e-02;
C[20545] = 8.813406183787e-01;
C[20546] = 4.334199962718e-02;
C[20547] = -1.103170615663e-02;
C[20548] = 2.088964298695e+00;
C[20549] = -2.349573845002e-02;
C[20550] = 2.074479096421e-03;
C[20551] = -7.164433833570e+00;
C[20552] = -4.117044313132e-02;
C[20553] = 8.906219532090e-04;
C[20554] = 5.727836034426e-02;
C[20555] = 2.997343362351e-03;
C[20556] = -1.505220886543e-03;
C[20557] = 3.910351819689e-01;
C[20558] = 4.572423578410e-03;
C[20559] = 1.263354897186e-03;
C[20560] = 3.392961548396e-02;
C[20561] = -1.233298087629e-02;
C[20562] = -3.595255136670e-03;
C[20563] = 4.108472389444e+00;
C[20564] = 1.881898017380e-02;
C[20565] = -7.248128297842e-04;
C[20566] = 5.150575741084e-01;
C[20567] = 4.684879806658e-03;
C[20568] = 1.985006090296e-03;
C[20569] = 4.533853230747e-02;
C[20570] = 7.905221513205e-01;
C[20571] = -4.159215373687e-04;
C[20572] = -2.611785701560e-02;
C[20573] = 2.173819166556e+00;
C[20574] = 8.688183036470e-03;
C[20575] = -4.111571005257e-02;
C[20576] = -7.218974511964e+00;
C[20577] = 1.627749383875e-04;
C[20578] = 3.143195588097e-03;
C[20579] = 5.167403622079e-02;
C[20580] = -9.913084983977e-04;
C[20581] = 4.392793263770e-03;
C[20582] = 4.017016533594e-01;
C[20583] = 2.613524701378e-03;
C[20584] = -1.233024422235e-02;
C[20585] = 1.852101883093e-02;
C[20586] = -1.023363264940e-02;
C[20587] = 1.924005300249e-02;
C[20588] = 4.171188635302e+00;
C[20589] = -1.799847962120e-03;
C[20590] = 4.708889206144e-03;
C[20591] = 5.254398719094e-01;
C[20592] = 0.000000000000e+00;
C[20593] = 0.000000000000e+00;
C[20594] = 0.000000000000e+00;
C[20595] = 0.000000000000e+00;
C[20596] = 0.000000000000e+00;
C[20597] = 0.000000000000e+00;
C[20598] = 0.000000000000e+00;
C[20599] = 0.000000000000e+00;
C[20600] = 0.000000000000e+00;
C[20601] = 0.000000000000e+00;
C[20602] = 0.000000000000e+00;
C[20603] = 0.000000000000e+00;
C[20604] = 0.000000000000e+00;
C[20605] = 0.000000000000e+00;
C[20606] = 0.000000000000e+00;
C[20607] = 0.000000000000e+00;
C[20608] = 0.000000000000e+00;
C[20609] = 0.000000000000e+00;
C[20610] = 1.000000000000e+00;
C[20611] = 0.000000000000e+00;
C[20612] = 0.000000000000e+00;
C[20613] = 2.000000000000e-01;
C[20614] = -0.000000000000e+00;
C[20615] = -0.000000000000e+00;
C[20616] = 0.000000000000e+00;
C[20617] = 0.000000000000e+00;
C[20618] = 0.000000000000e+00;
C[20619] = 0.000000000000e+00;
C[20620] = 0.000000000000e+00;
C[20621] = 0.000000000000e+00;
C[20622] = 0.000000000000e+00;
C[20623] = 0.000000000000e+00;
C[20624] = 0.000000000000e+00;
C[20625] = 0.000000000000e+00;
C[20626] = 0.000000000000e+00;
C[20627] = 0.000000000000e+00;
C[20628] = 0.000000000000e+00;
C[20629] = 0.000000000000e+00;
C[20630] = 0.000000000000e+00;
C[20631] = 0.000000000000e+00;
C[20632] = 0.000000000000e+00;
C[20633] = 0.000000000000e+00;
C[20634] = 0.000000000000e+00;
C[20635] = 1.000000000000e+00;
C[20636] = 0.000000000000e+00;
C[20637] = -0.000000000000e+00;
C[20638] = 2.000000000000e-01;
C[20639] = -0.000000000000e+00;
C[20640] = 0.000000000000e+00;
C[20641] = 0.000000000000e+00;
C[20642] = 0.000000000000e+00;
C[20643] = 0.000000000000e+00;
C[20644] = 0.000000000000e+00;
C[20645] = 0.000000000000e+00;
C[20646] = 0.000000000000e+00;
C[20647] = 0.000000000000e+00;
C[20648] = 0.000000000000e+00;
C[20649] = 0.000000000000e+00;
C[20650] = 0.000000000000e+00;
C[20651] = 0.000000000000e+00;
C[20652] = 0.000000000000e+00;
C[20653] = 0.000000000000e+00;
C[20654] = 0.000000000000e+00;
C[20655] = 0.000000000000e+00;
C[20656] = 0.000000000000e+00;
C[20657] = 0.000000000000e+00;
C[20658] = 0.000000000000e+00;
C[20659] = 0.000000000000e+00;
C[20660] = 1.000000000000e+00;
C[20661] = -0.000000000000e+00;
C[20662] = -0.000000000000e+00;
C[20663] = 2.000000000000e-01;
C[20664] = 6.822007109964e-02;
C[20665] = -8.591177693666e-03;
C[20666] = 7.554790472190e-03;
C[20667] = 3.732677509601e-01;
C[20668] = 7.827253518847e-03;
C[20669] = 4.577743132364e-03;
C[20670] = 5.114595740399e-02;
C[20671] = 1.064606596598e-03;
C[20672] = 1.506220929881e-04;
C[20673] = 1.304806770373e-01;
C[20674] = -8.692673087722e-04;
C[20675] = 6.727032340895e-04;
C[20676] = 3.015891938883e-02;
C[20677] = 7.343335949303e-04;
C[20678] = 3.458457303844e-04;
C[20679] = 2.461227155667e-03;
C[20680] = 5.662971856833e-05;
C[20681] = 3.822117793939e-06;
C[20682] = 2.432589833030e-03;
C[20683] = 6.388655965148e-05;
C[20684] = -1.268472466028e-05;
C[20685] = -2.063320933553e-05;
C[20686] = 7.490789033070e-07;
C[20687] = -7.641771056955e-07;
C[20688] = -8.577414265426e-03;
C[20689] = 4.527814701844e-02;
C[20690] = -7.826277302999e-03;
C[20691] = 7.873964925612e-03;
C[20692] = 4.002223394906e-01;
C[20693] = 9.049828264549e-03;
C[20694] = 9.973046815623e-04;
C[20695] = 5.678533413457e-02;
C[20696] = 3.001364256044e-03;
C[20697] = -8.683497468895e-04;
C[20698] = 1.281832342699e-01;
C[20699] = -6.845409831543e-04;
C[20700] = 7.372318311846e-04;
C[20701] = 3.283144259561e-02;
C[20702] = 9.170626404006e-04;
C[20703] = 5.259845502156e-05;
C[20704] = 2.767179567435e-03;
C[20705] = 1.626378867091e-04;
C[20706] = 5.630369876898e-05;
C[20707] = 2.831364097035e-03;
C[20708] = 2.459594208918e-04;
C[20709] = 5.916972499142e-07;
C[20710] = -1.560509103980e-05;
C[20711] = 2.530050317380e-06;
C[20712] = 7.571825072598e-03;
C[20713] = -7.845944210096e-03;
C[20714] = 4.525545802060e-02;
C[20715] = 4.626880894711e-03;
C[20716] = 9.177259685464e-03;
C[20717] = 3.702615045645e-01;
C[20718] = 9.683090423217e-05;
C[20719] = 2.880192129450e-03;
C[20720] = 5.289440694793e-02;
C[20721] = 6.738388741166e-04;
C[20722] = -6.858521102941e-04;
C[20723] = 1.287277532126e-01;
C[20724] = 3.489605078434e-04;
C[20725] = 9.252951201989e-04;
C[20726] = 3.008578766810e-02;
C[20727] = 5.993350634701e-07;
C[20728] = 1.553899349926e-04;
C[20729] = 2.564706018169e-03;
C[20730] = -1.903398135363e-05;
C[20731] = 2.278666895649e-04;
C[20732] = 2.656694495309e-03;
C[20733] = -8.842077024785e-07;
C[20734] = 2.187921580557e-06;
C[20735] = -1.759080792783e-05;
C[20736] = 3.734978565390e-01;
C[20737] = 8.161843565290e-03;
C[20738] = 4.473555354783e-03;
C[20739] = 1.186495456980e-01;
C[20740] = -6.075902023687e-03;
C[20741] = 3.818384583266e-04;
C[20742] = 3.917346507151e-01;
C[20743] = -3.021394331794e-03;
C[20744] = -3.088224779432e-03;
C[20745] = 3.017425976076e-02;
C[20746] = 7.564237404964e-04;
C[20747] = 3.387388051816e-04;
C[20748] = 1.328377357048e-01;
C[20749] = -5.410475319639e-04;
C[20750] = 7.755218332362e-05;
C[20751] = 3.194309289212e-02;
C[20752] = -2.616997197098e-04;
C[20753] = -3.311693855041e-04;
C[20754] = 6.075021083118e-02;
C[20755] = -6.412133176116e-04;
C[20756] = -9.982999246965e-04;
C[20757] = 2.649905128547e-03;
C[20758] = -2.919165711432e-05;
C[20759] = -4.959728495126e-05;
C[20760] = 8.089439206704e-03;
C[20761] = 4.002711319309e-01;
C[20762] = 8.741485039833e-03;
C[20763] = -6.044714286985e-03;
C[20764] = 8.794503383804e-02;
C[20765] = -1.691588870139e-02;
C[20766] = -2.963627220543e-03;
C[20767] = 3.900390008788e-01;
C[20768] = 4.323302159289e-03;
C[20769] = 7.518126407874e-04;
C[20770] = 3.283469542496e-02;
C[20771] = 8.962434771568e-04;
C[20772] = -5.389683495171e-04;
C[20773] = 1.298553551692e-01;
C[20774] = -1.619607694394e-03;
C[20775] = -2.583703195769e-04;
C[20776] = 3.197328101594e-02;
C[20777] = 5.291011757115e-04;
C[20778] = -6.286382797713e-04;
C[20779] = 6.140974858301e-02;
C[20780] = 2.126022948199e-03;
C[20781] = -2.872498729512e-05;
C[20782] = 2.691429703135e-03;
C[20783] = 1.063978676071e-04;
C[20784] = 4.434905016348e-03;
C[20785] = 8.679897114543e-03;
C[20786] = 3.700376944611e-01;
C[20787] = 3.993207661461e-04;
C[20788] = -1.692783897458e-02;
C[20789] = 1.065583947174e-01;
C[20790] = -3.058306733464e-03;
C[20791] = 4.460105413961e-03;
C[20792] = 4.002772207294e-01;
C[20793] = 3.363231893167e-04;
C[20794] = 8.924005637336e-04;
C[20795] = 3.007086699454e-02;
C[20796] = 7.871767051158e-05;
C[20797] = -1.620404379273e-03;
C[20798] = 1.317134525090e-01;
C[20799] = -3.295243151389e-04;
C[20800] = 5.371044952677e-04;
C[20801] = 3.283834987989e-02;
C[20802] = -9.773750852546e-04;
C[20803] = 2.131201973029e-03;
C[20804] = 6.456657724644e-02;
C[20805] = -4.863385856776e-05;
C[20806] = 1.061315570158e-04;
C[20807] = 2.831675323844e-03;
C[20808] = 5.152962373060e-02;
C[20809] = 9.016285009527e-04;
C[20810] = -7.558469916451e-05;
C[20811] = 3.912238259862e-01;
C[20812] = -2.929739301859e-03;
C[20813] = -3.075916514838e-03;
C[20814] = 4.053255672870e-02;
C[20815] = 2.339428613778e-03;
C[20816] = 4.219679865593e-03;
C[20817] = 2.486804910775e-03;
C[20818] = 4.622004298092e-05;
C[20819] = -1.089503849631e-05;
C[20820] = 3.190903791019e-02;
C[20821] = -2.561111249981e-04;
C[20822] = -3.306983005639e-04;
C[20823] = 1.279039441736e-01;
C[20824] = 2.486744939390e-04;
C[20825] = 4.335921608684e-04;
C[20826] = 5.143120294996e-01;
C[20827] = -3.591363883771e-04;
C[20828] = -9.884502715657e-04;
C[20829] = 3.771916228026e-02;
C[20830] = -3.810475953468e-05;
C[20831] = -9.219865673161e-05;
C[20832] = 9.723988472720e-04;
C[20833] = 5.660174546935e-02;
C[20834] = 2.752650149553e-03;
C[20835] = -3.006102212641e-03;
C[20836] = 3.899853152171e-01;
C[20837] = 4.581139755542e-03;
C[20838] = 2.347800454373e-03;
C[20839] = 3.432309393405e-02;
C[20840] = -1.235372268829e-02;
C[20841] = 5.048253527992e-05;
C[20842] = 2.754940323087e-03;
C[20843] = 1.468871363328e-04;
C[20844] = -2.606802450996e-04;
C[20845] = 3.196970197183e-02;
C[20846] = 5.451734513730e-04;
C[20847] = 2.492326166454e-04;
C[20848] = 1.273848739441e-01;
C[20849] = -1.113885385227e-03;
C[20850] = -3.641096960173e-04;
C[20851] = 5.163983607135e-01;
C[20852] = 4.837997710292e-03;
C[20853] = -3.842168930049e-05;
C[20854] = 3.790532943153e-02;
C[20855] = 4.179613418036e-04;
C[20856] = -2.123677469301e-05;
C[20857] = 2.874117193515e-03;
C[20858] = 5.322215496512e-02;
C[20859] = -3.133810568170e-03;
C[20860] = 4.446253112526e-03;
C[20861] = 3.998141078517e-01;
C[20862] = 4.235059665349e-03;
C[20863] = -1.235329911553e-02;
C[20864] = 1.893032719781e-02;
C[20865] = -7.635140051468e-06;
C[20866] = 1.541547492072e-04;
C[20867] = 2.586555885981e-03;
C[20868] = -3.342084380867e-04;
C[20869] = 5.372979059273e-04;
C[20870] = 3.280747568804e-02;
C[20871] = 4.346174808521e-04;
C[20872] = -1.113857147043e-03;
C[20873] = 1.259907714989e-01;
C[20874] = -1.003072344378e-03;
C[20875] = 4.851530717340e-03;
C[20876] = 5.253380758074e-01;
C[20877] = -9.305123143267e-05;
C[20878] = 4.185862344635e-04;
C[20879] = 3.862976727448e-02;
C[20880] = -7.046857729734e+00;
C[20881] = -2.464550413898e-02;
C[20882] = 5.814789215055e-02;
C[20883] = 2.091428990441e+00;
C[20884] = 1.616660803833e-02;
C[20885] = 3.719954768547e-02;
C[20886] = 7.899046755631e-01;
C[20887] = 1.516009429761e-02;
C[20888] = 3.349961079727e-03;
C[20889] = 7.160295433027e-02;
C[20890] = -1.043261428163e-02;
C[20891] = 9.781478553105e-03;
C[20892] = 3.694159380390e-01;
C[20893] = 8.024587478266e-03;
C[20894] = 4.771528615684e-03;
C[20895] = 5.086678618910e-02;
C[20896] = 1.047924811501e-03;
C[20897] = 9.186471231212e-05;
C[20898] = 7.891996159435e-02;
C[20899] = 1.611362419643e-03;
C[20900] = -3.490321210720e-04;
C[20901] = 2.438216859234e-03;
C[20902] = 5.782279759317e-05;
C[20903] = -2.554293953707e-05;
C[20904] = -2.437023557419e-02;
C[20905] = -7.079628201875e+00;
C[20906] = -3.532679782494e-03;
C[20907] = 1.684351710666e-02;
C[20908] = 2.157188560862e+00;
C[20909] = 1.471404629905e-02;
C[20910] = 1.425641265633e-02;
C[20911] = 8.729452212324e-01;
C[20912] = 4.296169561045e-02;
C[20913] = -1.041885085339e-02;
C[20914] = 4.641132492075e-02;
C[20915] = -6.106609981285e-03;
C[20916] = 8.089532385082e-03;
C[20917] = 4.002353472925e-01;
C[20918] = 1.014905058006e-02;
C[20919] = 9.731998690878e-04;
C[20920] = 5.669340907455e-02;
C[20921] = 3.073090831327e-03;
C[20922] = 1.419986487269e-03;
C[20923] = 8.936154002338e-02;
C[20924] = 6.661020757415e-03;
C[20925] = 4.991688677294e-05;
C[20926] = 2.816410127594e-03;
C[20927] = 2.426909737748e-04;
C[20928] = 5.848858415870e-02;
C[20929] = -3.926017924442e-03;
C[20930] = -7.163185431821e+00;
C[20931] = 3.798119582616e-02;
C[20932] = 1.688420570498e-02;
C[20933] = 2.033825774602e+00;
C[20934] = 2.635427761941e-03;
C[20935] = 4.135580652305e-02;
C[20936] = 8.119601726320e-01;
C[20937] = 9.798513153512e-03;
C[20938] = -6.126276888383e-03;
C[20939] = 5.175391077574e-02;
C[20940] = 4.833300470592e-03;
C[20941] = 1.030899821697e-02;
C[20942] = 3.669680304032e-01;
C[20943] = 3.337911336773e-05;
C[20944] = 2.942146345306e-03;
C[20945] = 5.263964092998e-02;
C[20946] = -5.050685446972e-04;
C[20947] = 6.203593852997e-03;
C[20948] = 8.474340061887e-02;
C[20949] = -3.171676913411e-05;
C[20950] = 2.243934166320e-04;
C[20951] = 2.660746462483e-03;
C[20952] = 2.096031102019e+00;
C[20953] = 2.260108990020e-02;
C[20954] = 3.491468502761e-02;
C[20955] = -6.220312249174e+00;
C[20956] = -1.485458719106e-02;
C[20957] = 4.451248044145e-03;
C[20958] = 2.161414936890e+00;
C[20959] = -2.180638467475e-02;
C[20960] = -9.482443074886e-03;
C[20961] = 3.696460436179e-01;
C[20962] = 8.377411024759e-03;
C[20963] = 4.679974930664e-03;
C[20964] = 1.234176851207e-01;
C[20965] = -5.724958516093e-03;
C[20966] = 1.471782513695e-03;
C[20967] = 3.910512613437e-01;
C[20968] = -3.661261667667e-03;
C[20969] = -4.182521065222e-03;
C[20970] = 1.026201828809e+00;
C[20971] = -1.152092685357e-02;
C[20972] = -1.688093944074e-02;
C[20973] = 6.069093625643e-02;
C[20974] = -6.745524945394e-04;
C[20975] = -1.056580657481e-03;
C[20976] = 2.141032179546e-02;
C[20977] = 2.158164409667e+00;
C[20978] = 8.168712792350e-03;
C[20979] = -1.423083245702e-02;
C[20980] = -6.295699759465e+00;
C[20981] = -4.081807151182e-02;
C[20982] = -2.108393219206e-02;
C[20983] = 2.088974856955e+00;
C[20984] = -2.552652737191e-02;
C[20985] = 8.286773166122e-03;
C[20986] = 4.002841397327e-01;
C[20987] = 9.873223571337e-03;
C[20988] = -5.693770779391e-03;
C[20989] = 8.944875224880e-02;
C[20990] = -1.820145337834e-02;
C[20991] = -3.598078677573e-03;
C[20992] = 3.886737146831e-01;
C[20993] = 4.597936514559e-03;
C[20994] = -1.123186855944e-02;
C[20995] = 1.032716325371e+00;
C[20996] = 3.238037516697e-02;
C[20997] = -6.594623179537e-04;
C[20998] = 6.134491749122e-02;
C[20999] = 2.150397146638e-03;
C[21000] = 3.434278536514e-02;
C[21001] = 7.315423298939e-03;
C[21002] = 2.029349572533e+00;
C[21003] = 4.800894200535e-03;
C[21004] = -4.105707697572e-02;
C[21005] = -6.254326218034e+00;
C[21006] = -9.210956936789e-03;
C[21007] = -2.367189001975e-02;
C[21008] = 2.152345032380e+00;
C[21009] = 4.628690499667e-03;
C[21010] = 9.779119430059e-03;
C[21011] = 3.667442202998e-01;
C[21012] = 1.489264821515e-03;
C[21013] = -1.821340365153e-02;
C[21014] = 1.107555742603e-01;
C[21015] = -4.152113292226e-03;
C[21016] = 4.744203270602e-03;
C[21017] = 3.998250747345e-01;
C[21018] = -1.646128920539e-02;
C[21019] = 3.264025080833e-02;
C[21020] = 1.086471551445e+00;
C[21021] = -1.031850897163e-03;
C[21022] = 2.156360619491e-03;
C[21023] = 6.451196625474e-02;
C[21024] = 7.975780020953e-01;
C[21025] = 1.234288904414e-02;
C[21026] = -8.128843059930e-04;
C[21027] = 2.151198442313e+00;
C[21028] = -2.040617381840e-02;
C[21029] = -9.563152564281e-03;
C[21030] = -7.141920547089e+00;
C[21031] = 4.607287802132e-03;
C[21032] = 1.178310594544e-02;
C[21033] = 5.125045251571e-02;
C[21034] = 8.775236884782e-04;
C[21035] = -1.390364900289e-04;
C[21036] = 3.905404366148e-01;
C[21037] = -3.564190758889e-03;
C[21038] = -4.169723073600e-03;
C[21039] = 4.141829027170e-02;
C[21040] = 2.944004398676e-03;
C[21041] = 5.242830348055e-03;
C[21042] = 4.116083691767e+00;
C[21043] = 1.946513630506e-03;
C[21044] = 8.353133168823e-04;
C[21045] = 5.144237508109e-01;
C[21046] = -3.078495152601e-04;
C[21047] = -8.625097699124e-04;
C[21048] = 1.331593931109e-02;
C[21049] = 8.692734479278e-01;
C[21050] = 3.880496692510e-02;
C[21051] = -2.150054229169e-02;
C[21052] = 2.087901143720e+00;
C[21053] = -2.125120318814e-02;
C[21054] = 4.774724614041e-03;
C[21055] = -7.167349270516e+00;
C[21056] = -4.224671562857e-02;
C[21057] = 9.557170621746e-04;
C[21058] = 5.650982040932e-02;
C[21059] = 2.814604365409e-03;
C[21060] = -3.645969548514e-03;
C[21061] = 3.886200290214e-01;
C[21062] = 4.865237612183e-03;
C[21063] = 2.952376239272e-03;
C[21064] = 3.555110024665e-02;
C[21065] = -1.274029153390e-02;
C[21066] = 1.852763735866e-03;
C[21067] = 4.124922269760e+00;
C[21068] = 1.994936870297e-02;
C[21069] = -3.149848842632e-04;
C[21070] = 5.166225847778e-01;
C[21071] = 4.867279602552e-03;
C[21072] = -8.721627389461e-05;
C[21073] = 4.041675435987e-02;
C[21074] = 8.185151329758e-01;
C[21075] = -1.039415884965e-02;
C[21076] = -2.306750830718e-02;
C[21077] = 2.143082774826e+00;
C[21078] = 1.209070194055e-02;
C[21079] = -4.223824417336e-02;
C[21080] = -7.217919088368e+00;
C[21081] = -7.999415536896e-05;
C[21082] = 2.945843768797e-03;
C[21083] = 5.296738894717e-02;
C[21084] = -4.228106853960e-03;
C[21085] = 4.720887467795e-03;
C[21086] = 3.993619618568e-01;
C[21087] = 5.258210147811e-03;
C[21088] = -1.273986796114e-02;
C[21089] = 1.969049688959e-02;
C[21090] = 5.738705627264e-04;
C[21091] = 2.016488821591e-02;
C[21092] = 4.171410401986e+00;
C[21093] = -8.810577135814e-04;
C[21094] = 4.880100354794e-03;
C[21095] = 5.253157718734e-01;
C[21096] = 0.000000000000e+00;
C[21097] = 0.000000000000e+00;
C[21098] = 0.000000000000e+00;
C[21099] = 0.000000000000e+00;
C[21100] = 0.000000000000e+00;
C[21101] = 0.000000000000e+00;
C[21102] = 0.000000000000e+00;
C[21103] = 0.000000000000e+00;
C[21104] = 0.000000000000e+00;
C[21105] = 0.000000000000e+00;
C[21106] = 0.000000000000e+00;
C[21107] = 0.000000000000e+00;
C[21108] = 0.000000000000e+00;
C[21109] = 0.000000000000e+00;
C[21110] = 0.000000000000e+00;
C[21111] = 0.000000000000e+00;
C[21112] = 0.000000000000e+00;
C[21113] = 0.000000000000e+00;
C[21114] = 1.000000000000e+00;
C[21115] = 0.000000000000e+00;
C[21116] = 0.000000000000e+00;
C[21117] = 2.000000000000e-01;
C[21118] = -0.000000000000e+00;
C[21119] = -0.000000000000e+00;
C[21120] = 0.000000000000e+00;
C[21121] = 0.000000000000e+00;
C[21122] = 0.000000000000e+00;
C[21123] = 0.000000000000e+00;
C[21124] = 0.000000000000e+00;
C[21125] = 0.000000000000e+00;
C[21126] = 0.000000000000e+00;
C[21127] = 0.000000000000e+00;
C[21128] = 0.000000000000e+00;
C[21129] = 0.000000000000e+00;
C[21130] = 0.000000000000e+00;
C[21131] = 0.000000000000e+00;
C[21132] = 0.000000000000e+00;
C[21133] = 0.000000000000e+00;
C[21134] = 0.000000000000e+00;
C[21135] = 0.000000000000e+00;
C[21136] = 0.000000000000e+00;
C[21137] = 0.000000000000e+00;
C[21138] = 0.000000000000e+00;
C[21139] = 1.000000000000e+00;
C[21140] = 0.000000000000e+00;
C[21141] = -0.000000000000e+00;
C[21142] = 2.000000000000e-01;
C[21143] = -0.000000000000e+00;
C[21144] = 0.000000000000e+00;
C[21145] = 0.000000000000e+00;
C[21146] = 0.000000000000e+00;
C[21147] = 0.000000000000e+00;
C[21148] = 0.000000000000e+00;
C[21149] = 0.000000000000e+00;
C[21150] = 0.000000000000e+00;
C[21151] = 0.000000000000e+00;
C[21152] = 0.000000000000e+00;
C[21153] = 0.000000000000e+00;
C[21154] = 0.000000000000e+00;
C[21155] = 0.000000000000e+00;
C[21156] = 0.000000000000e+00;
C[21157] = 0.000000000000e+00;
C[21158] = 0.000000000000e+00;
C[21159] = 0.000000000000e+00;
C[21160] = 0.000000000000e+00;
C[21161] = 0.000000000000e+00;
C[21162] = 0.000000000000e+00;
C[21163] = 0.000000000000e+00;
C[21164] = 1.000000000000e+00;
C[21165] = -0.000000000000e+00;
C[21166] = -0.000000000000e+00;
C[21167] = 2.000000000000e-01;
C[21168] = 8.849288809986e-02;
C[21169] = -1.481979808971e-02;
C[21170] = 1.091121285175e-02;
C[21171] = 3.565769170572e-01;
C[21172] = 5.781783517023e-03;
C[21173] = 3.454758479040e-03;
C[21174] = 4.721389421752e-02;
C[21175] = 1.076351837753e-04;
C[21176] = -2.985409677549e-04;
C[21177] = 1.327969599184e-01;
C[21178] = -1.123032637851e-03;
C[21179] = 7.415900174079e-04;
C[21180] = 2.835248082373e-02;
C[21181] = 3.218796020183e-04;
C[21182] = 1.675931522167e-04;
C[21183] = 2.240835504020e-03;
C[21184] = -4.486018727942e-06;
C[21185] = -2.231068082122e-05;
C[21186] = 2.147945429444e-03;
C[21187] = -3.059308253566e-05;
C[21188] = -4.925421613156e-05;
C[21189] = -2.506289119834e-05;
C[21190] = -2.068340776709e-06;
C[21191] = -1.396637502895e-06;
C[21192] = -1.449568564555e-02;
C[21193] = 4.223778292490e-02;
C[21194] = -5.583895869548e-04;
C[21195] = 5.742403344892e-03;
C[21196] = 3.999634723126e-01;
C[21197] = 1.306201848624e-02;
C[21198] = 6.578623934395e-06;
C[21199] = 5.656039153975e-02;
C[21200] = 3.815245303224e-03;
C[21201] = -1.101425141574e-03;
C[21202] = 1.274022748745e-01;
C[21203] = -9.425581410442e-05;
C[21204] = 3.167742387776e-04;
C[21205] = 3.288907359318e-02;
C[21206] = 1.227869168663e-03;
C[21207] = -9.166436395120e-06;
C[21208] = 2.760440214950e-03;
C[21209] = 2.027150559236e-04;
C[21210] = -3.632556411867e-05;
C[21211] = 2.838255122852e-03;
C[21212] = 2.897525606032e-04;
C[21213] = -1.974695166724e-06;
C[21214] = -1.365571206054e-05;
C[21215] = 2.418708840876e-06;
C[21216] = 1.093185417436e-02;
C[21217] = -6.253284617093e-04;
C[21218] = 7.420351961772e-02;
C[21219] = 3.514052144893e-03;
C[21220] = 1.374240059048e-02;
C[21221] = 3.562753736211e-01;
C[21222] = -3.380764428581e-04;
C[21223] = 3.607033850224e-03;
C[21224] = 4.954546876160e-02;
C[21225] = 7.429661055822e-04;
C[21226] = -9.871840575471e-05;
C[21227] = 1.317757234128e-01;
C[21228] = 1.692446115468e-04;
C[21229] = 1.269072569334e-03;
C[21230] = 2.848723383235e-02;
C[21231] = -2.381738911178e-05;
C[21232] = 1.885503787477e-04;
C[21233] = 2.372125863431e-03;
C[21234] = -4.940533440225e-05;
C[21235] = 2.538402198524e-04;
C[21236] = 2.388300802229e-03;
C[21237] = -1.229073546586e-06;
C[21238] = 1.484533402914e-06;
C[21239] = -2.327293530763e-05;
C[21240] = 3.563858537099e-01;
C[21241] = 6.157180496669e-03;
C[21242] = 3.768886025401e-03;
C[21243] = 1.402289487454e-01;
C[21244] = -4.828651748401e-03;
C[21245] = 1.498153162712e-03;
C[21246] = 3.920129470137e-01;
C[21247] = -2.490412903634e-03;
C[21248] = -3.202104123815e-03;
C[21249] = 2.833974326725e-02;
C[21250] = 3.444260488961e-04;
C[21251] = 1.862335369140e-04;
C[21252] = 1.350961545417e-01;
C[21253] = -2.624982360340e-04;
C[21254] = 8.195974790736e-05;
C[21255] = 3.177293328171e-02;
C[21256] = -1.262565598834e-04;
C[21257] = -1.783567108383e-04;
C[21258] = 6.011975704487e-02;
C[21259] = -2.591381447420e-04;
C[21260] = -4.907894912077e-04;
C[21261] = 2.607783267687e-03;
C[21262] = -5.500984930498e-06;
C[21263] = -1.501478989066e-05;
C[21264] = 5.719280839478e-03;
C[21265] = 4.013690204960e-01;
C[21266] = 1.304130848315e-02;
C[21267] = -4.629670631386e-03;
C[21268] = 8.936592037802e-02;
C[21269] = -2.156461060932e-02;
C[21270] = -2.343748410980e-03;
C[21271] = 3.875445108548e-01;
C[21272] = 5.335549649581e-03;
C[21273] = 3.177127568486e-04;
C[21274] = 3.298277680541e-02;
C[21275] = 1.222333095513e-03;
C[21276] = -2.492328282331e-04;
C[21277] = 1.298872602754e-01;
C[21278] = -2.074113728223e-03;
C[21279] = -1.188763559228e-04;
C[21280] = 3.178612974507e-02;
C[21281] = 6.837391040237e-04;
C[21282] = -2.550166461577e-04;
C[21283] = 6.102104431670e-02;
C[21284] = 2.531290704470e-03;
C[21285] = -5.930038162957e-06;
C[21286] = 2.675241225983e-03;
C[21287] = 1.295730430381e-04;
C[21288] = 3.605553949374e-03;
C[21289] = 1.263327312246e-02;
C[21290] = 3.552705700585e-01;
C[21291] = 1.534623377511e-03;
C[21292] = -2.165555190675e-02;
C[21293] = 1.244899349026e-01;
C[21294] = -3.108687548277e-03;
C[21295] = 5.619567874688e-03;
C[21296] = 4.024420779819e-01;
C[21297] = 1.776461835723e-04;
C[21298] = 1.199286144411e-03;
C[21299] = 2.842024692818e-02;
C[21300] = 8.439109556059e-05;
C[21301] = -2.080176481385e-03;
C[21302] = 1.335376286998e-01;
C[21303] = -1.731807596012e-04;
C[21304] = 7.010801552207e-04;
C[21305] = 3.293652957392e-02;
C[21306] = -4.898955100781e-04;
C[21307] = 2.546167612832e-03;
C[21308] = 6.465346074931e-02;
C[21309] = -1.560699583565e-05;
C[21310] = 1.291842513892e-04;
C[21311] = 2.829582276354e-03;
C[21312] = 4.807819213035e-02;
C[21313] = 9.387312176634e-04;
C[21314] = 2.069451144704e-04;
C[21315] = 3.909074314078e-01;
C[21316] = -3.023518317853e-03;
C[21317] = -3.221056352972e-03;
C[21318] = 4.442525638344e-02;
C[21319] = 2.704840929720e-03;
C[21320] = 4.813489126392e-03;
C[21321] = 2.298455364875e-03;
C[21322] = 5.297706985348e-05;
C[21323] = 1.251738137679e-05;
C[21324] = 3.169923224132e-02;
C[21325] = -1.641943497143e-04;
C[21326] = -1.806720132475e-04;
C[21327] = 1.283634673665e-01;
C[21328] = 1.782318596026e-04;
C[21329] = 3.289003746506e-04;
C[21330] = 5.144036819788e-01;
C[21331] = -7.206404737621e-04;
C[21332] = -1.725305092648e-03;
C[21333] = 3.765789944035e-02;
C[21334] = -6.652732507529e-05;
C[21335] = -1.594634992573e-04;
C[21336] = 9.928899864046e-04;
C[21337] = 5.652079224953e-02;
C[21338] = 3.595400840787e-03;
C[21339] = -3.055824837204e-03;
C[21340] = 3.878674468322e-01;
C[21341] = 5.732316035335e-03;
C[21342] = 2.678693956820e-03;
C[21343] = 3.712193524686e-02;
C[21344] = -1.396764426048e-02;
C[21345] = 5.453096811401e-05;
C[21346] = 2.757800262268e-03;
C[21347] = 1.877748447852e-04;
C[21348] = -1.639506887880e-04;
C[21349] = 3.180765881023e-02;
C[21350] = 7.085966992638e-04;
C[21351] = 1.764887280760e-04;
C[21352] = 1.276669634110e-01;
C[21353] = -1.301387137338e-03;
C[21354] = -7.186090443664e-04;
C[21355] = 5.158274396627e-01;
C[21356] = 4.492235278848e-03;
C[21357] = -6.624580111373e-05;
C[21358] = 3.778795687591e-02;
C[21359] = 4.006942838784e-04;
C[21360] = 2.624120121878e-04;
C[21361] = 3.827430123356e-03;
C[21362] = 5.050337114655e-02;
C[21363] = -3.223126348639e-03;
C[21364] = 5.461439404059e-03;
C[21365] = 4.012883832125e-01;
C[21366] = 4.759550867299e-03;
C[21367] = -1.397285914806e-02;
C[21368] = 2.023453571082e-02;
C[21369] = 1.508618450830e-05;
C[21370] = 2.035273772657e-04;
C[21371] = 2.435986022428e-03;
C[21372] = -1.797581924932e-04;
C[21373] = 6.921317543222e-04;
C[21374] = 3.285961658930e-02;
C[21375] = 3.253044907111e-04;
C[21376] = -1.301734796510e-03;
C[21377] = 1.260705546693e-01;
C[21378] = -1.725782121541e-03;
C[21379] = 4.510410556486e-03;
C[21380] = 5.255293227786e-01;
C[21381] = -1.590676712761e-04;
C[21382] = 4.016286221786e-04;
C[21383] = 3.864740180649e-02;
C[21384] = -6.962064815853e+00;
C[21385] = -6.297489979911e-02;
C[21386] = 5.791313134232e-02;
C[21387] = 2.056898567753e+00;
C[21388] = -6.315556911702e-03;
C[21389] = 2.411913977636e-02;
C[21390] = 7.298037080720e-01;
C[21391] = -3.529653320954e-03;
C[21392] = -4.640086487598e-03;
C[21393] = 1.033086558773e-01;
C[21394] = -1.173114383921e-02;
C[21395] = 7.788343360104e-03;
C[21396] = 3.475355828185e-01;
C[21397] = 2.139371700478e-04;
C[21398] = 1.555312373829e-03;
C[21399] = 4.654449570528e-02;
C[21400] = -1.375808455214e-04;
C[21401] = -2.700274211950e-04;
C[21402] = 7.143824597328e-02;
C[21403] = -2.064417922532e-04;
C[21404] = -6.737806805331e-04;
C[21405] = 2.158836234832e-03;
C[21406] = 4.234382118735e-06;
C[21407] = -1.611210957776e-05;
C[21408] = -5.649265091599e-02;
C[21409] = -7.124068878705e+00;
C[21410] = 4.618927999648e-02;
C[21411] = -8.246880783944e-03;
C[21412] = 2.133622291741e+00;
C[21413] = 3.873425365368e-02;
C[21414] = -4.580200631149e-03;
C[21415] = 8.697380769155e-01;
C[21416] = 5.544434132483e-02;
C[21417] = -1.140703139505e-02;
C[21418] = 3.086978931806e-02;
C[21419] = 1.730245679714e-03;
C[21420] = 1.815484289893e-04;
C[21421] = 4.005280017282e-01;
C[21422] = 1.393473334268e-02;
C[21423] = -1.845365293371e-04;
C[21424] = 5.661775145687e-02;
C[21425] = 3.939406843614e-03;
C[21426] = -3.003716228672e-04;
C[21427] = 8.945269930964e-02;
C[21428] = 8.175719722250e-03;
C[21429] = 3.239634429307e-06;
C[21430] = 2.839849910422e-03;
C[21431] = 2.970971503682e-04;
C[21432] = 5.832595779460e-02;
C[21433] = 4.485050250140e-02;
C[21434] = -7.026746261339e+00;
C[21435] = 2.461928175959e-02;
C[21436] = 5.013359463738e-02;
C[21437] = 2.025023889968e+00;
C[21438] = -4.883746786945e-03;
C[21439] = 5.293047638963e-02;
C[21440] = 7.616428945236e-01;
C[21441] = 7.808984682718e-03;
C[21442] = 1.663306804959e-03;
C[21443] = 9.286119788330e-02;
C[21444] = 1.597070163519e-03;
C[21445] = 1.474759306959e-02;
C[21446] = 3.481359611696e-01;
C[21447] = -2.821806060644e-04;
C[21448] = 3.688794480893e-03;
C[21449] = 4.899254227025e-02;
C[21450] = -7.054454921825e-04;
C[21451] = 7.294089658864e-03;
C[21452] = 7.822920084827e-02;
C[21453] = -1.720998236697e-05;
C[21454] = 2.577845238150e-04;
C[21455] = 2.410473154098e-03;
C[21456] = 2.053077300808e+00;
C[21457] = 4.866225159428e-05;
C[21458] = 2.971595936975e-02;
C[21459] = -6.145530687565e+00;
C[21460] = -1.144322818563e-02;
C[21461] = 4.122227171099e-03;
C[21462] = 2.208916384687e+00;
C[21463] = -4.975366234164e-03;
C[21464] = -2.725128161395e-03;
C[21465] = 3.473445194712e-01;
C[21466] = 5.963255807662e-04;
C[21467] = 1.851904044027e-03;
C[21468] = 1.519920962515e-01;
C[21469] = -1.117867812317e-03;
C[21470] = 4.280136784412e-04;
C[21471] = 3.903383057311e-01;
C[21472] = -5.548800208342e-04;
C[21473] = -5.537127119847e-04;
C[21474] = 1.017349863354e+00;
C[21475] = -3.261228401463e-03;
C[21476] = -5.946806572208e-03;
C[21477] = 5.988376393943e-02;
C[21478] = -1.862530553391e-04;
C[21479] = -3.663263845863e-04;
C[21480] = -7.565610462590e-03;
C[21481] = 2.161733255410e+00;
C[21482] = 3.611175249090e-02;
C[21483] = -7.463605845338e-03;
C[21484] = -6.294741555652e+00;
C[21485] = -5.413925948737e-02;
C[21486] = -2.898731770101e-03;
C[21487] = 2.080240734040e+00;
C[21488] = -2.920131804606e-02;
C[21489] = 1.514344925034e-04;
C[21490] = 4.019335499117e-01;
C[21491] = 1.404650096227e-02;
C[21492] = -9.188866953023e-04;
C[21493] = 8.872984808603e-02;
C[21494] = -2.355726168678e-02;
C[21495] = -4.536174611059e-04;
C[21496] = 3.869684449751e-01;
C[21497] = 6.434441145981e-03;
C[21498] = -3.241445420344e-03;
C[21499] = 1.026530570643e+00;
C[21500] = 3.895485174787e-02;
C[21501] = -1.945007329838e-04;
C[21502] = 6.089806088321e-02;
C[21503] = 2.601243254520e-03;
C[21504] = 2.713504918305e-02;
C[21505] = 3.015934637810e-02;
C[21506] = 2.004927818715e+00;
C[21507] = 4.851631467069e-03;
C[21508] = -5.595808543612e-02;
C[21509] = -6.192397886389e+00;
C[21510] = -1.287983205960e-03;
C[21511] = -2.504218743699e-02;
C[21512] = 2.202321979744e+00;
C[21513] = 1.706107844163e-03;
C[21514] = 1.350598797890e-02;
C[21515] = 3.471311576069e-01;
C[21516] = 4.644838932397e-04;
C[21517] = -2.364820298422e-02;
C[21518] = 1.334361133350e-01;
C[21519] = -4.782494590428e-04;
C[21520] = 6.770138787475e-03;
C[21521] = 4.027254138897e-01;
C[21522] = -6.020624536292e-03;
C[21523] = 3.951425418625e-02;
C[21524] = 1.090670271892e+00;
C[21525] = -3.779573580071e-04;
C[21526] = 2.612089411829e-03;
C[21527] = 6.461981494081e-02;
C[21528] = 7.470896663286e-01;
C[21529] = 1.406285124343e-02;
C[21530] = 6.016684359625e-03;
C[21531] = 2.186806072570e+00;
C[21532] = -1.649412990756e-02;
C[21533] = -3.535359299864e-03;
C[21534] = -7.124012610638e+00;
C[21535] = 5.275490000904e-03;
C[21536] = 1.277749715161e-02;
C[21537] = 4.740879361811e-02;
C[21538] = 7.476160643920e-04;
C[21539] = 2.628409512641e-04;
C[21540] = 3.892327901252e-01;
C[21541] = -1.133387367979e-03;
C[21542] = -5.906182637380e-04;
C[21543] = 4.790991690324e-02;
C[21544] = 1.202779938093e-03;
C[21545] = 2.598372090363e-03;
C[21546] = 4.119890825762e+00;
C[21547] = -4.998532210769e-03;
C[21548] = -1.290917640946e-02;
C[21549] = 5.133395138665e-01;
C[21550] = -8.565003180871e-04;
C[21551] = -2.172345176455e-03;
C[21552] = 1.417544273163e-02;
C[21553] = 8.689460911111e-01;
C[21554] = 5.269781620089e-02;
C[21555] = -1.628360490556e-02;
C[21556] = 2.086699453586e+00;
C[21557] = -2.278722422405e-02;
C[21558] = 4.752550542916e-03;
C[21559] = -7.156819142479e+00;
C[21560] = -4.668941270141e-02;
C[21561] = 7.476739571079e-04;
C[21562] = 5.657815216665e-02;
C[21563] = 3.677161471456e-03;
C[21564] = -1.120291954404e-03;
C[21565] = 3.872913809524e-01;
C[21566] = 6.882886948121e-03;
C[21567] = 1.176632965194e-03;
C[21568] = 3.909605680772e-02;
C[21569] = -1.517446429557e-02;
C[21570] = -4.905354413241e-03;
C[21571] = 4.116067326276e+00;
C[21572] = 1.314600102647e-02;
C[21573] = -8.477983872240e-04;
C[21574] = 5.143730197619e-01;
C[21575] = 4.459362058730e-03;
C[21576] = 6.578973111258e-03;
C[21577] = 5.568803772748e-02;
C[21578] = 7.808009422226e-01;
C[21579] = -3.145572657866e-03;
C[21580] = -2.668352295650e-02;
C[21581] = 2.179248084357e+00;
C[21582] = 1.169873196976e-02;
C[21583] = -4.679371045306e-02;
C[21584] = -7.211924210288e+00;
C[21585] = 2.909255587477e-04;
C[21586] = 3.951591663747e-03;
C[21587] = 4.995044465520e-02;
C[21588] = -5.747349368083e-04;
C[21589] = 6.560330900459e-03;
C[21590] = 4.015717191203e-01;
C[21591] = 2.544433831270e-03;
C[21592] = -1.517967918315e-02;
C[21593] = 2.038161032231e-02;
C[21594] = -1.279874744492e-02;
C[21595] = 1.347197613733e-02;
C[21596] = 4.174992088757e+00;
C[21597] = -2.160215200557e-03;
C[21598] = 4.483989825938e-03;
C[21599] = 5.257259513059e-01;
C[21600] = 0.000000000000e+00;
C[21601] = 0.000000000000e+00;
C[21602] = 0.000000000000e+00;
C[21603] = 0.000000000000e+00;
C[21604] = 0.000000000000e+00;
C[21605] = 0.000000000000e+00;
C[21606] = 0.000000000000e+00;
C[21607] = 0.000000000000e+00;
C[21608] = 0.000000000000e+00;
C[21609] = 0.000000000000e+00;
C[21610] = 0.000000000000e+00;
C[21611] = 0.000000000000e+00;
C[21612] = 0.000000000000e+00;
C[21613] = 0.000000000000e+00;
C[21614] = 0.000000000000e+00;
C[21615] = 0.000000000000e+00;
C[21616] = 0.000000000000e+00;
C[21617] = 0.000000000000e+00;
C[21618] = 1.000000000000e+00;
C[21619] = 0.000000000000e+00;
C[21620] = 0.000000000000e+00;
C[21621] = 2.000000000000e-01;
C[21622] = -0.000000000000e+00;
C[21623] = -0.000000000000e+00;
C[21624] = 0.000000000000e+00;
C[21625] = 0.000000000000e+00;
C[21626] = 0.000000000000e+00;
C[21627] = 0.000000000000e+00;
C[21628] = 0.000000000000e+00;
C[21629] = 0.000000000000e+00;
C[21630] = 0.000000000000e+00;
C[21631] = 0.000000000000e+00;
C[21632] = 0.000000000000e+00;
C[21633] = 0.000000000000e+00;
C[21634] = 0.000000000000e+00;
C[21635] = 0.000000000000e+00;
C[21636] = 0.000000000000e+00;
C[21637] = 0.000000000000e+00;
C[21638] = 0.000000000000e+00;
C[21639] = 0.000000000000e+00;
C[21640] = 0.000000000000e+00;
C[21641] = 0.000000000000e+00;
C[21642] = 0.000000000000e+00;
C[21643] = 1.000000000000e+00;
C[21644] = 0.000000000000e+00;
C[21645] = -0.000000000000e+00;
C[21646] = 2.000000000000e-01;
C[21647] = -0.000000000000e+00;
C[21648] = 0.000000000000e+00;
C[21649] = 0.000000000000e+00;
C[21650] = 0.000000000000e+00;
C[21651] = 0.000000000000e+00;
C[21652] = 0.000000000000e+00;
C[21653] = 0.000000000000e+00;
C[21654] = 0.000000000000e+00;
C[21655] = 0.000000000000e+00;
C[21656] = 0.000000000000e+00;
C[21657] = 0.000000000000e+00;
C[21658] = 0.000000000000e+00;
C[21659] = 0.000000000000e+00;
C[21660] = 0.000000000000e+00;
C[21661] = 0.000000000000e+00;
C[21662] = 0.000000000000e+00;
C[21663] = 0.000000000000e+00;
C[21664] = 0.000000000000e+00;
C[21665] = 0.000000000000e+00;
C[21666] = 0.000000000000e+00;
C[21667] = 0.000000000000e+00;
C[21668] = 1.000000000000e+00;
C[21669] = -0.000000000000e+00;
C[21670] = -0.000000000000e+00;
C[21671] = 2.000000000000e-01;
C[21672] = 9.540459273516e-02;
C[21673] = -4.013895808326e-03;
C[21674] = 6.379885269675e-04;
C[21675] = 3.606661163629e-01;
C[21676] = 1.594749151069e-03;
C[21677] = 6.901768250981e-05;
C[21678] = 5.006749746683e-02;
C[21679] = 5.691451647325e-04;
C[21680] = 4.194133213252e-04;
C[21681] = 1.318506353184e-01;
C[21682] = -5.631148312377e-05;
C[21683] = -6.972221979644e-05;
C[21684] = 2.980642770856e-02;
C[21685] = 1.888050694785e-04;
C[21686] = 3.672647228757e-05;
C[21687] = 2.446803772521e-03;
C[21688] = 3.219109522261e-05;
C[21689] = 2.405444056847e-05;
C[21690] = 2.486771537601e-03;
C[21691] = 3.676898528785e-05;
C[21692] = 1.646435309729e-05;
C[21693] = -1.044588311382e-05;
C[21694] = 6.009208310188e-07;
C[21695] = 1.032530211150e-07;
C[21696] = -3.886892951802e-03;
C[21697] = 1.997644994142e-02;
C[21698] = -1.595058932823e-03;
C[21699] = 1.572053814044e-03;
C[21700] = 3.981912284378e-01;
C[21701] = 1.289492745462e-02;
C[21702] = 5.886954117255e-04;
C[21703] = 5.605563024046e-02;
C[21704] = 3.777837250868e-03;
C[21705] = -4.784462602221e-05;
C[21706] = 1.260288440641e-01;
C[21707] = -3.106798795445e-04;
C[21708] = 1.868252717934e-04;
C[21709] = 3.274645853032e-02;
C[21710] = 1.108495797134e-03;
C[21711] = 3.310015720415e-05;
C[21712] = 2.731879445209e-03;
C[21713] = 1.934768421223e-04;
C[21714] = 3.480555206478e-05;
C[21715] = 2.776185637350e-03;
C[21716] = 2.644274142384e-04;
C[21717] = 4.957950710091e-07;
C[21718] = -1.309563047270e-05;
C[21719] = 1.062873984914e-06;
C[21720] = 6.312081149379e-04;
C[21721] = -1.501918456911e-03;
C[21722] = 8.706925449378e-02;
C[21723] = 7.485837767621e-05;
C[21724] = 1.334829793075e-02;
C[21725] = 3.622551210845e-01;
C[21726] = 4.175985628723e-04;
C[21727] = 3.603114187973e-03;
C[21728] = 5.322769584510e-02;
C[21729] = -7.017424726509e-05;
C[21730] = -3.044705144837e-04;
C[21731] = 1.312166343670e-01;
C[21732] = 3.659772196904e-05;
C[21733] = 1.128361914372e-03;
C[21734] = 3.009049027675e-02;
C[21735] = 2.377443341094e-05;
C[21736] = 1.832788708177e-04;
C[21737] = 2.624558913281e-03;
C[21738] = 1.199148989553e-05;
C[21739] = 2.428535048136e-04;
C[21740] = 2.819571215986e-03;
C[21741] = -3.673866549826e-08;
C[21742] = 9.514498128587e-07;
C[21743] = -8.545581500143e-06;
C[21744] = 3.608800525039e-01;
C[21745] = 5.200241994158e-04;
C[21746] = 2.209634509836e-04;
C[21747] = 1.404443577536e-01;
C[21748] = -2.504727355780e-03;
C[21749] = -2.300564943943e-03;
C[21750] = 3.860445925637e-01;
C[21751] = 9.412211114315e-04;
C[21752] = 1.956861373073e-03;
C[21753] = 2.982069011796e-02;
C[21754] = 1.166899641515e-04;
C[21755] = 4.633806018953e-05;
C[21756] = 1.339173505244e-01;
C[21757] = -2.848385498762e-04;
C[21758] = -2.852287484455e-04;
C[21759] = 3.133229936912e-02;
C[21760] = 1.164425462221e-04;
C[21761] = 2.254988101437e-04;
C[21762] = 5.878768511814e-02;
C[21763] = 1.724532168484e-04;
C[21764] = 2.200461000281e-04;
C[21765] = 2.552811428853e-03;
C[21766] = 1.059403496072e-05;
C[21767] = 1.502412109232e-05;
C[21768] = 5.842491914129e-04;
C[21769] = 3.986457874223e-01;
C[21770] = 1.398965580699e-02;
C[21771] = -2.584878522294e-03;
C[21772] = 9.075931469963e-02;
C[21773] = -2.373944091801e-02;
C[21774] = 9.417376966853e-04;
C[21775] = 3.866562337273e-01;
C[21776] = 6.925641202653e-03;
C[21777] = 1.214384055014e-04;
C[21778] = 3.277676246262e-02;
C[21779] = 1.171119106122e-03;
C[21780] = -2.901819609771e-04;
C[21781] = 1.301881021477e-01;
C[21782] = -2.081799504694e-03;
C[21783] = 1.168556515283e-04;
C[21784] = 3.160348608866e-02;
C[21785] = 7.692400885991e-04;
C[21786] = 1.556030913760e-04;
C[21787] = 5.985443841594e-02;
C[21788] = 2.628029622367e-03;
C[21789] = 9.804199873507e-06;
C[21790] = 2.615522172379e-03;
C[21791] = 1.302366815570e-04;
C[21792] = 2.095868469386e-04;
C[21793] = 1.327681285016e-02;
C[21794] = 3.620591237347e-01;
C[21795] = -2.334730086188e-03;
C[21796] = -2.351514316724e-02;
C[21797] = 1.201340640279e-01;
C[21798] = 2.013655707585e-03;
C[21799] = 7.165458520939e-03;
C[21800] = 3.983911859540e-01;
C[21801] = 4.609774991615e-05;
C[21802] = 1.133954823503e-03;
C[21803] = 3.007742378676e-02;
C[21804] = -2.875064245951e-04;
C[21805] = -2.066846321309e-03;
C[21806] = 1.319705008493e-01;
C[21807] = 2.288520614011e-04;
C[21808] = 7.806610656683e-04;
C[21809] = 3.268836015737e-02;
C[21810] = 1.811125731353e-04;
C[21811] = 2.604839278556e-03;
C[21812] = 6.438041601136e-02;
C[21813] = 1.291024315176e-05;
C[21814] = 1.279547109897e-04;
C[21815] = 2.824106422284e-03;
C[21816] = 4.803253130558e-02;
C[21817] = 4.522578500963e-04;
C[21818] = 4.943404058164e-04;
C[21819] = 3.883483464605e-01;
C[21820] = 9.559513531015e-04;
C[21821] = 1.942852501361e-03;
C[21822] = 5.140751526698e-02;
C[21823] = -5.051946545212e-04;
C[21824] = 1.347932565293e-04;
C[21825] = 2.311139361771e-03;
C[21826] = 2.400431976220e-05;
C[21827] = 2.889055627388e-05;
C[21828] = 3.148588296224e-02;
C[21829] = 1.178032286227e-04;
C[21830] = 2.241318476528e-04;
C[21831] = 1.289849031745e-01;
C[21832] = -9.543542778677e-05;
C[21833] = -1.124834317077e-04;
C[21834] = 5.100713595072e-01;
C[21835] = -9.804682266508e-04;
C[21836] = -2.592475657277e-03;
C[21837] = 3.723765441924e-02;
C[21838] = -5.063380038074e-05;
C[21839] = -1.401865299673e-04;
C[21840] = 4.041716345429e-04;
C[21841] = 5.595869999821e-02;
C[21842] = 3.864670059011e-03;
C[21843] = 1.001982214774e-03;
C[21844] = 3.869357864680e-01;
C[21845] = 7.042691789801e-03;
C[21846] = -5.255973996728e-04;
C[21847] = 4.349543335464e-02;
C[21848] = -1.527969930555e-02;
C[21849] = 2.119285987663e-05;
C[21850] = 2.725417429058e-03;
C[21851] = 2.007159288868e-04;
C[21852] = 1.204932864449e-04;
C[21853] = 3.162212293804e-02;
C[21854] = 7.724766169258e-04;
C[21855] = -9.679561079687e-05;
C[21856] = 1.283561821549e-01;
C[21857] = -1.339977298066e-03;
C[21858] = -9.604758185059e-04;
C[21859] = 5.107915077327e-01;
C[21860] = 4.251716280099e-03;
C[21861] = -4.962392745829e-05;
C[21862] = 3.731312655788e-02;
C[21863] = 3.667191035727e-04;
C[21864] = 4.851524054710e-04;
C[21865] = 4.087911880537e-03;
C[21866] = 5.115969667002e-02;
C[21867] = 1.933431197863e-03;
C[21868] = 6.687080464635e-03;
C[21869] = 4.004156985456e-01;
C[21870] = 9.726542015800e-05;
C[21871] = -1.522641287155e-02;
C[21872] = 1.967719227030e-02;
C[21873] = 2.843704617819e-05;
C[21874] = 2.141484841001e-04;
C[21875] = 2.486692301609e-03;
C[21876] = 2.239367984630e-04;
C[21877] = 7.533360393979e-04;
C[21878] = 3.282332766348e-02;
C[21879] = -1.149852874658e-04;
C[21880] = -1.336424869133e-03;
C[21881] = 1.259694296059e-01;
C[21882] = -2.539611214498e-03;
C[21883] = 4.285054060684e-03;
C[21884] = 5.263652126480e-01;
C[21885] = -1.372941082253e-04;
C[21886] = 3.681145185722e-04;
C[21887] = 3.873527076914e-02;
C[21888] = -6.972408583250e+00;
C[21889] = 2.406515638261e-03;
C[21890] = -2.765312164536e-03;
C[21891] = 2.116172944136e+00;
C[21892] = 1.380364278817e-02;
C[21893] = -5.279323831884e-03;
C[21894] = 7.911961535614e-01;
C[21895] = 9.174001469990e-03;
C[21896] = 6.587410687929e-03;
C[21897] = 7.678087474984e-02;
C[21898] = 3.090434807242e-03;
C[21899] = -2.219047204542e-03;
C[21900] = 3.766119461310e-01;
C[21901] = 2.764761771161e-03;
C[21902] = 3.604132292589e-04;
C[21903] = 5.093666616726e-02;
C[21904] = 6.312856609891e-04;
C[21905] = 5.227492306213e-04;
C[21906] = 7.828802963684e-02;
C[21907] = 9.005992034835e-04;
C[21908] = 5.656952598864e-04;
C[21909] = 2.426920728527e-03;
C[21910] = 3.121846810785e-05;
C[21911] = 2.523770803281e-05;
C[21912] = 4.946572768730e-03;
C[21913] = -7.216496048823e+00;
C[21914] = 2.952914108719e-02;
C[21915] = 1.293570781106e-02;
C[21916] = 2.108545120013e+00;
C[21917] = 1.802063283946e-02;
C[21918] = 9.276131450746e-03;
C[21919] = 8.598346384032e-01;
C[21920] = 5.172985596718e-02;
C[21921] = 3.217437663766e-03;
C[21922] = 1.950069475261e-02;
C[21923] = -2.618882426951e-03;
C[21924] = 2.676155604352e-03;
C[21925] = 3.974591364977e-01;
C[21926] = 1.060365914254e-02;
C[21927] = 6.462373085308e-04;
C[21928] = 5.587195780984e-02;
C[21929] = 3.643714865945e-03;
C[21930] = 8.692127830327e-04;
C[21931] = 8.717800248919e-02;
C[21932] = 7.599797165125e-03;
C[21933] = 3.049528803053e-05;
C[21934] = 2.762783473668e-03;
C[21935] = 2.742431900733e-04;
C[21936] = -2.900920405130e-03;
C[21937] = 3.139195060543e-02;
C[21938] = -7.001233997181e+00;
C[21939] = -5.365540100278e-03;
C[21940] = 2.301661706493e-02;
C[21941] = 2.088513431718e+00;
C[21942] = 6.556570562026e-03;
C[21943] = 5.035845617390e-02;
C[21944] = 8.353986212184e-01;
C[21945] = -2.225827616572e-03;
C[21946] = -2.525741951038e-03;
C[21947] = 7.051449970081e-02;
C[21948] = 3.480950794551e-04;
C[21949] = 1.096716745310e-02;
C[21950] = 3.787235807817e-01;
C[21951] = 5.205754237963e-04;
C[21952] = 3.506878930319e-03;
C[21953] = 5.426739077672e-02;
C[21954] = 4.638563219036e-04;
C[21955] = 7.053506824310e-03;
C[21956] = 8.766588562473e-02;
C[21957] = 2.105615497344e-05;
C[21958] = 2.509691146294e-04;
C[21959] = 2.762367028822e-03;
C[21960] = 2.120451666955e+00;
C[21961] = -8.104884481505e-03;
C[21962] = -2.443438634131e-03;
C[21963] = -6.185985318553e+00;
C[21964] = -6.218811337304e-03;
C[21965] = -7.054588681641e-03;
C[21966] = 2.143321777768e+00;
C[21967] = -8.566934146582e-04;
C[21968] = 6.084017673262e-03;
C[21969] = 3.768258822720e-01;
C[21970] = 1.624125989724e-03;
C[21971] = 4.942001527625e-04;
C[21972] = 1.255036314975e-01;
C[21973] = -3.633348119736e-03;
C[21974] = -3.885704317824e-03;
C[21975] = 3.842574313022e-01;
C[21976] = 1.350770619814e-03;
C[21977] = 2.950905949615e-03;
C[21978] = 9.946345065998e-01;
C[21979] = 2.923753623022e-03;
C[21980] = 4.388548126914e-03;
C[21981] = 5.859609222817e-02;
C[21982] = 2.303578491915e-04;
C[21983] = 3.655611857862e-04;
C[21984] = -6.406356404958e-03;
C[21985] = 2.117636299704e+00;
C[21986] = 3.584377458981e-02;
C[21987] = -7.821834667574e-03;
C[21988] = -6.290426554801e+00;
C[21989] = -5.430779617974e-02;
C[21990] = -6.880525694329e-04;
C[21991] = 2.089285141187e+00;
C[21992] = -2.139193684964e-02;
C[21993] = 1.754261811504e-03;
C[21994] = 3.979136954822e-01;
C[21995] = 1.160852532935e-02;
C[21996] = -3.713499286249e-03;
C[21997] = 9.364095793399e-02;
C[21998] = -2.144230300426e-02;
C[21999] = 1.354243048057e-03;
C[22000] = 3.847923483720e-01;
C[22001] = 6.953511200014e-03;
C[22002] = 2.662141577809e-03;
C[22003] = 1.006345490314e+00;
C[22004] = 3.939777229939e-02;
C[22005] = 2.148424429636e-04;
C[22006] = 5.951790320898e-02;
C[22007] = 2.579566172531e-03;
C[22008] = -2.467940543307e-03;
C[22009] = 2.565834075026e-02;
C[22010] = 2.084593484721e+00;
C[22011] = -7.737891526534e-03;
C[22012] = -4.982184116422e-02;
C[22013] = -6.241508573925e+00;
C[22014] = 7.051883267660e-03;
C[22015] = -1.884289091608e-02;
C[22016] = 2.127400437345e+00;
C[22017] = 5.009823936877e-04;
C[22018] = 1.098554453808e-02;
C[22019] = 3.785275834318e-01;
C[22020] = -3.919869460069e-03;
C[22021] = -2.121800525349e-02;
C[22022] = 1.031753813634e-01;
C[22023] = 3.009910002872e-03;
C[22024] = 7.150647820977e-03;
C[22025] = 3.978775505470e-01;
C[22026] = 3.792307173017e-03;
C[22027] = 3.929713087350e-02;
C[22028] = 1.084865984391e+00;
C[22029] = 3.256306289988e-04;
C[22030] = 2.550749658012e-03;
C[22031] = 6.452310475768e-02;
C[22032] = 7.504968303365e-01;
C[22033] = 6.547380218162e-03;
C[22034] = 8.091407420908e-03;
C[22035] = 2.189396855704e+00;
C[22036] = -4.037794411089e-04;
C[22037] = 5.635819143169e-03;
C[22038] = -7.096619636933e+00;
C[22039] = -3.360060021607e-04;
C[22040] = 3.493566447051e-03;
C[22041] = 4.890170000602e-02;
C[22042] = 5.097997469016e-04;
C[22043] = 5.973172667404e-04;
C[22044] = 3.865611851990e-01;
C[22045] = 1.368456704473e-03;
C[22046] = 2.939106796647e-03;
C[22047] = 5.518257242746e-02;
C[22048] = -1.432715216409e-03;
C[22049] = -2.164066739671e-03;
C[22050] = 4.084507499711e+00;
C[22051] = -7.686985348895e-03;
C[22052] = -1.785054114279e-02;
C[22053] = 5.070687550176e-01;
C[22054] = -5.378313218511e-04;
C[22055] = -1.392567636469e-03;
C[22056] = 5.874530866198e-03;
C[22057] = 8.578960335581e-01;
C[22058] = 5.558957359465e-02;
C[22059] = 3.585286521956e-04;
C[22060] = 2.094876196001e+00;
C[22061] = -2.129822553883e-02;
C[22062] = -7.440609051924e-04;
C[22063] = -7.121639283102e+00;
C[22064] = -4.714008331300e-02;
C[22065] = 4.663121307995e-04;
C[22066] = 5.577502756759e-02;
C[22067] = 3.768434801357e-03;
C[22068] = 1.411531723157e-03;
C[22069] = 3.850719011126e-01;
C[22070] = 7.027881089839e-03;
C[22071] = -1.453117961560e-03;
C[22072] = 4.885225753602e-02;
C[22073] = -1.464475673197e-02;
C[22074] = -7.368858129537e-03;
C[22075] = 4.079630733622e+00;
C[22076] = 9.824003282767e-03;
C[22077] = -5.195899282814e-04;
C[22078] = 5.074869858893e-01;
C[22079] = 3.747268346336e-03;
C[22080] = 7.902192370846e-03;
C[22081] = 5.793134856055e-02;
C[22082] = 7.940386377169e-01;
C[22083] = 5.615414169061e-03;
C[22084] = -2.616315161000e-02;
C[22085] = 2.167890689178e+00;
C[22086] = 2.743009719625e-03;
C[22087] = -4.607435463287e-02;
C[22088] = -7.221924481915e+00;
C[22089] = 5.884883147672e-04;
C[22090] = 3.953789495613e-03;
C[22091] = 5.219939160165e-02;
C[22092] = 2.927475774405e-03;
C[22093] = 6.714950461996e-03;
C[22094] = 3.999020631386e-01;
C[22095] = -2.201594576042e-03;
C[22096] = -1.459147029797e-02;
C[22097] = 1.824976772029e-02;
C[22098] = -1.695400945762e-02;
C[22099] = 1.031508339991e-02;
C[22100] = 4.181471719657e+00;
C[22101] = -1.338471098768e-03;
C[22102] = 3.786692288683e-03;
C[22103] = 5.271072353741e-01;
C[22104] = 0.000000000000e+00;
C[22105] = 0.000000000000e+00;
C[22106] = 0.000000000000e+00;
C[22107] = 0.000000000000e+00;
C[22108] = 0.000000000000e+00;
C[22109] = 0.000000000000e+00;
C[22110] = 0.000000000000e+00;
C[22111] = 0.000000000000e+00;
C[22112] = 0.000000000000e+00;
C[22113] = 0.000000000000e+00;
C[22114] = 0.000000000000e+00;
C[22115] = 0.000000000000e+00;
C[22116] = 0.000000000000e+00;
C[22117] = 0.000000000000e+00;
C[22118] = 0.000000000000e+00;
C[22119] = 0.000000000000e+00;
C[22120] = 0.000000000000e+00;
C[22121] = 0.000000000000e+00;
C[22122] = 1.000000000000e+00;
C[22123] = 0.000000000000e+00;
C[22124] = 0.000000000000e+00;
C[22125] = 2.000000000000e-01;
C[22126] = -0.000000000000e+00;
C[22127] = -0.000000000000e+00;
C[22128] = 0.000000000000e+00;
C[22129] = 0.000000000000e+00;
C[22130] = 0.000000000000e+00;
C[22131] = 0.000000000000e+00;
C[22132] = 0.000000000000e+00;
C[22133] = 0.000000000000e+00;
C[22134] = 0.000000000000e+00;
C[22135] = 0.000000000000e+00;
C[22136] = 0.000000000000e+00;
C[22137] = 0.000000000000e+00;
C[22138] = 0.000000000000e+00;
C[22139] = 0.000000000000e+00;
C[22140] = 0.000000000000e+00;
C[22141] = 0.000000000000e+00;
C[22142] = 0.000000000000e+00;
C[22143] = 0.000000000000e+00;
C[22144] = 0.000000000000e+00;
C[22145] = 0.000000000000e+00;
C[22146] = 0.000000000000e+00;
C[22147] = 1.000000000000e+00;
C[22148] = 0.000000000000e+00;
C[22149] = -0.000000000000e+00;
C[22150] = 2.000000000000e-01;
C[22151] = -0.000000000000e+00;
C[22152] = 0.000000000000e+00;
C[22153] = 0.000000000000e+00;
C[22154] = 0.000000000000e+00;
C[22155] = 0.000000000000e+00;
C[22156] = 0.000000000000e+00;
C[22157] = 0.000000000000e+00;
C[22158] = 0.000000000000e+00;
C[22159] = 0.000000000000e+00;
C[22160] = 0.000000000000e+00;
C[22161] = 0.000000000000e+00;
C[22162] = 0.000000000000e+00;
C[22163] = 0.000000000000e+00;
C[22164] = 0.000000000000e+00;
C[22165] = 0.000000000000e+00;
C[22166] = 0.000000000000e+00;
C[22167] = 0.000000000000e+00;
C[22168] = 0.000000000000e+00;
C[22169] = 0.000000000000e+00;
C[22170] = 0.000000000000e+00;
C[22171] = 0.000000000000e+00;
C[22172] = 1.000000000000e+00;
C[22173] = -0.000000000000e+00;
C[22174] = -0.000000000000e+00;
C[22175] = 2.000000000000e-01;
C[22176] = 6.306431181450e-02;
C[22177] = 9.272071074981e-03;
C[22178] = -3.635105824237e-03;
C[22179] = 3.844676029136e-01;
C[22180] = 2.726176484580e-03;
C[22181] = -7.590809686565e-05;
C[22182] = 5.195999821620e-02;
C[22183] = 2.956307114342e-04;
C[22184] = 3.596887602332e-04;
C[22185] = 1.296338808627e-01;
C[22186] = 8.448336017549e-04;
C[22187] = -2.710976519913e-04;
C[22188] = 3.135795830249e-02;
C[22189] = 1.097972289089e-04;
C[22190] = -1.779351340498e-05;
C[22191] = 2.513573997576e-03;
C[22192] = 7.574811811515e-06;
C[22193] = 1.873365209273e-05;
C[22194] = 2.447759315242e-03;
C[22195] = -4.584162174224e-06;
C[22196] = 2.943540691569e-05;
C[22197] = -1.819424643023e-05;
C[22198] = -1.572852359746e-06;
C[22199] = -1.677072730434e-06;
C[22200] = 9.333141543934e-03;
C[22201] = 2.192732034042e-02;
C[22202] = -2.052271093389e-03;
C[22203] = 2.589833243638e-03;
C[22204] = 3.986311182142e-01;
C[22205] = 8.601922415921e-03;
C[22206] = 3.053965986020e-04;
C[22207] = 5.538264841681e-02;
C[22208] = 3.297117333063e-03;
C[22209] = 8.489049663517e-04;
C[22210] = 1.262536323862e-01;
C[22211] = -2.678657894311e-04;
C[22212] = 1.022896675364e-04;
C[22213] = 3.276949745945e-02;
C[22214] = 8.182614665628e-04;
C[22215] = 8.603110739596e-06;
C[22216] = 2.697183658340e-03;
C[22217] = 1.730460120975e-04;
C[22218] = -1.589918462033e-06;
C[22219] = 2.674916210325e-03;
C[22220] = 2.395782287783e-04;
C[22221] = -1.555474478566e-06;
C[22222] = -1.462047676813e-05;
C[22223] = 2.452654342635e-06;
C[22224] = -3.624719612629e-03;
C[22225] = -2.039769987056e-03;
C[22226] = 6.021971330152e-02;
C[22227] = -1.358594304087e-04;
C[22228] = 8.867618751833e-03;
C[22229] = 3.863931616607e-01;
C[22230] = 3.633950249616e-04;
C[22231] = 3.190696385056e-03;
C[22232] = 5.590672951683e-02;
C[22233] = -2.704052378841e-04;
C[22234] = -2.670323823422e-04;
C[22235] = 1.293661693649e-01;
C[22236] = -2.139576405933e-05;
C[22237] = 8.317894274291e-04;
C[22238] = 3.167146145843e-02;
C[22239] = 1.939944373563e-05;
C[22240] = 1.661535245459e-04;
C[22241] = 2.729577529270e-03;
C[22242] = 3.129343442994e-05;
C[22243] = 2.177221423323e-04;
C[22244] = 2.886482286837e-03;
C[22245] = -1.661976669761e-06;
C[22246] = 2.052352868181e-06;
C[22247] = -1.394503706056e-05;
C[22248] = 3.847010900201e-01;
C[22249] = 2.666613342444e-03;
C[22250] = 3.015706193774e-05;
C[22251] = 1.182885409150e-01;
C[22252] = -2.437118117545e-03;
C[22253] = -3.093453757561e-03;
C[22254] = 3.820758209766e-01;
C[22255] = 1.077280207467e-03;
C[22256] = 1.927499076335e-03;
C[22257] = 3.137352410959e-02;
C[22258] = 1.074083407902e-04;
C[22259] = -1.032799790290e-05;
C[22260] = 1.325119893187e-01;
C[22261] = -1.307573009871e-04;
C[22262] = -2.002125302381e-04;
C[22263] = 3.110357820765e-02;
C[22264] = 8.867949255591e-05;
C[22265] = 1.606747607531e-04;
C[22266] = 5.760656778968e-02;
C[22267] = 2.314074668561e-04;
C[22268] = 7.483951221103e-04;
C[22269] = 2.509235630331e-03;
C[22270] = 1.017351157717e-05;
C[22271] = 3.364067864040e-05;
C[22272] = 2.720697774101e-03;
C[22273] = 3.984955859627e-01;
C[22274] = 8.906924307047e-03;
C[22275] = -2.391342364692e-03;
C[22276] = 9.541816268813e-02;
C[22277] = -1.918817570525e-02;
C[22278] = 1.041401026983e-03;
C[22279] = 3.837700226510e-01;
C[22280] = 7.459640988061e-03;
C[22281] = 1.094319815436e-04;
C[22282] = 3.276046197602e-02;
C[22283] = 8.344097977767e-04;
C[22284] = -1.277055841302e-04;
C[22285] = 1.304849573369e-01;
C[22286] = -1.769131030941e-03;
C[22287] = 8.634042601574e-05;
C[22288] = 3.139231353334e-02;
C[22289] = 7.907017774060e-04;
C[22290] = 2.395841504119e-04;
C[22291] = 5.837438442066e-02;
C[22292] = 2.338220838341e-03;
C[22293] = 1.084471927143e-05;
C[22294] = 2.553241524494e-03;
C[22295] = 1.156182967183e-04;
C[22296] = 4.722478853416e-05;
C[22297] = 8.601459954497e-03;
C[22298] = 3.865545726456e-01;
C[22299] = -3.047226431581e-03;
C[22300] = -1.916659515216e-02;
C[22301] = 9.382097998414e-02;
C[22302] = 1.897900226008e-03;
C[22303] = 7.674024900851e-03;
C[22304] = 3.971365743459e-01;
C[22305] = -9.584654378324e-06;
C[22306] = 8.182306358012e-04;
C[22307] = 3.168222219075e-02;
C[22308] = -1.971307085060e-04;
C[22309] = -1.767692327401e-03;
C[22310] = 1.302414537387e-01;
C[22311] = 1.581610298255e-04;
C[22312] = 8.022181110904e-04;
C[22313] = 3.269795939710e-02;
C[22314] = 7.534314227657e-04;
C[22315] = 2.273113241652e-03;
C[22316] = 6.474436397610e-02;
C[22317] = 3.437218354798e-05;
C[22318] = 1.116297942419e-04;
C[22319] = 2.847896663765e-03;
C[22320] = 5.177582904133e-02;
C[22321] = 4.622197699354e-04;
C[22322] = 3.336885380810e-04;
C[22323] = 3.821566547090e-01;
C[22324] = 8.066785679042e-04;
C[22325] = 1.400464439003e-03;
C[22326] = 5.894761972259e-02;
C[22327] = -1.567291402501e-03;
C[22328] = -5.566883339205e-03;
C[22329] = 2.501296052584e-03;
C[22330] = 1.905798882849e-05;
C[22331] = 1.741901127693e-05;
C[22332] = 3.110896712315e-02;
C[22333] = 7.069226207712e-05;
C[22334] = 1.249986440252e-04;
C[22335] = 1.294415553919e-01;
C[22336] = -1.264528602880e-04;
C[22337] = -5.669506122002e-04;
C[22338] = 5.047315094291e-01;
C[22339] = 3.520923224038e-04;
C[22340] = 3.797128990607e-03;
C[22341] = 3.696740234942e-02;
C[22342] = 3.684971325433e-05;
C[22343] = 4.246880791766e-04;
C[22344] = 4.617211280768e-04;
C[22345] = 5.543037205228e-02;
C[22346] = 3.247859313287e-03;
C[22347] = 8.290032708177e-04;
C[22348] = 3.838272031088e-01;
C[22349] = 7.812084122054e-03;
C[22350] = -1.574944085731e-03;
C[22351] = 5.435217007062e-02;
C[22352] = -1.294154271947e-02;
C[22353] = 1.864750625435e-05;
C[22354] = 2.700365234039e-03;
C[22355] = 1.699643864280e-04;
C[22356] = 7.212769677929e-05;
C[22357] = 3.139612556386e-02;
C[22358] = 8.114220591706e-04;
C[22359] = -1.269630391700e-04;
C[22360] = 1.290750938039e-01;
C[22361] = -1.123337625659e-03;
C[22362] = 3.448172712562e-04;
C[22363] = 5.035902160604e-01;
C[22364] = 1.743867919459e-03;
C[22365] = 3.645650658787e-05;
C[22366] = 3.684350353886e-02;
C[22367] = 1.402430150247e-04;
C[22368] = 3.445451725134e-04;
C[22369] = 3.358391752062e-03;
C[22370] = 5.570617648890e-02;
C[22371] = 1.399491438586e-03;
C[22372] = 7.587228006172e-03;
C[22373] = 3.975177647245e-01;
C[22374] = -5.590239943135e-03;
C[22375] = -1.292689825900e-02;
C[22376] = 1.704862339427e-02;
C[22377] = 1.772407957807e-05;
C[22378] = 1.771309733640e-04;
C[22379] = 2.716207327408e-03;
C[22380] = 1.254742515698e-04;
C[22381] = 7.992075786134e-04;
C[22382] = 3.272337208901e-02;
C[22383] = -5.685077191288e-04;
C[22384] = -1.122361328294e-03;
C[22385] = 1.258438904462e-01;
C[22386] = 3.814156701336e-03;
C[22387] = 1.819594122404e-03;
C[22388] = 5.270959779795e-01;
C[22389] = 4.255534830976e-04;
C[22390] = 1.437763517256e-04;
C[22391] = 3.873165164149e-02;
C[22392] = -7.054340865050e+00;
C[22393] = 5.726197755631e-02;
C[22394] = -1.813282200247e-02;
C[22395] = 2.147782206159e+00;
C[22396] = 2.288482137595e-02;
C[22397] = -1.128090212274e-02;
C[22398] = 8.070716538116e-01;
C[22399] = 3.817124351487e-03;
C[22400] = 3.701826645527e-03;
C[22401] = 5.932391443626e-02;
C[22402] = 1.146242295352e-02;
C[22403] = -2.992239945218e-03;
C[22404] = 3.859548725822e-01;
C[22405] = 9.935966151678e-04;
C[22406] = -9.909780176166e-04;
C[22407] = 5.205486893113e-02;
C[22408] = 1.657039537761e-04;
C[22409] = 2.567248620832e-04;
C[22410] = 7.885763330329e-02;
C[22411] = 2.083476493744e-04;
C[22412] = 9.401491612924e-04;
C[22413] = 2.442554597276e-03;
C[22414] = -2.813345010078e-06;
C[22415] = 2.120167504177e-05;
C[22416] = 5.848338693536e-02;
C[22417] = -7.203207394553e+00;
C[22418] = 2.082514029250e-02;
C[22419] = 2.074093577450e-02;
C[22420] = 2.126532107114e+00;
C[22421] = 8.628334429841e-03;
C[22422] = 3.890067734534e-03;
C[22423] = 8.520444640653e-01;
C[22424] = 4.644689450839e-02;
C[22425] = 1.152349342247e-02;
C[22426] = 2.293150107955e-02;
C[22427] = -1.912534973621e-03;
C[22428] = 8.642825249933e-04;
C[22429] = 3.984824606987e-01;
C[22430] = 8.338375056850e-03;
C[22431] = 1.782778556033e-04;
C[22432] = 5.537989689951e-02;
C[22433] = 3.280687349112e-03;
C[22434] = 2.788174828169e-04;
C[22435] = 8.471775591698e-02;
C[22436] = 6.578327966106e-03;
C[22437] = 1.116812078464e-07;
C[22438] = 2.677488161963e-03;
C[22439] = 2.425712102060e-04;
C[22440] = -1.792509777030e-02;
C[22441] = 2.107516241917e-02;
C[22442] = -7.055656230291e+00;
C[22443] = -1.231107036132e-02;
C[22444] = 1.216226760473e-02;
C[22445] = 2.117589102035e+00;
C[22446] = 3.768311577728e-03;
C[22447] = 4.544928656668e-02;
C[22448] = 8.612398774545e-01;
C[22449] = -2.981853733610e-03;
C[22450] = -1.900033867287e-03;
C[22451] = 5.672659197621e-02;
C[22452] = -1.057455044084e-03;
C[22453] = 8.592089488803e-03;
C[22454] = 3.882624779719e-01;
C[22455] = 2.659124593061e-04;
C[22456] = 3.181380853052e-03;
C[22457] = 5.603396188957e-02;
C[22458] = 9.882412037357e-04;
C[22459] = 6.054516312244e-03;
C[22460] = 9.097326613718e-02;
C[22461] = 2.329357793495e-05;
C[22462] = 2.207245216137e-04;
C[22463] = 2.882796012077e-03;
C[22464] = 2.152451948288e+00;
C[22465] = 2.227653775063e-02;
C[22466] = -8.990740514390e-03;
C[22467] = -6.229131760298e+00;
C[22468] = -1.386459794289e-03;
C[22469] = -6.874661249771e-03;
C[22470] = 2.128260772785e+00;
C[22471] = 2.412702861232e-03;
C[22472] = -1.712918686180e-03;
C[22473] = 3.861883596887e-01;
C[22474] = 9.410626237999e-04;
C[22475] = -8.914385517373e-04;
C[22476] = 1.166757947731e-01;
C[22477] = -2.551999708642e-04;
C[22478] = -1.353698488525e-03;
C[22479] = 3.819067392730e-01;
C[22480] = 7.743975221836e-04;
C[22481] = 9.957990137317e-04;
C[22482] = 9.756903946638e-01;
C[22483] = 3.315529482931e-03;
C[22484] = 1.208638145479e-02;
C[22485] = 5.767234237540e-02;
C[22486] = 2.085404459380e-04;
C[22487] = 8.911114504829e-04;
C[22488] = 2.277524716638e-02;
C[22489] = 2.123821462084e+00;
C[22490] = 1.294837870901e-02;
C[22491] = -4.709447372218e-04;
C[22492] = -6.282882926047e+00;
C[22493] = -4.165488525576e-02;
C[22494] = 1.793669807364e-03;
C[22495] = 2.098777220079e+00;
C[22496] = -1.196129830745e-02;
C[22497] = 9.881179046891e-04;
C[22498] = 3.983469284472e-01;
C[22499] = 8.631395044017e-03;
C[22500] = -2.094242180108e-04;
C[22501] = 9.542212231418e-02;
C[22502] = -1.896424374625e-02;
C[22503] = 7.267446607959e-04;
C[22504] = 3.840137895070e-01;
C[22505] = 7.599503678933e-03;
C[22506] = 3.457356973165e-03;
C[22507] = 9.852989136345e-01;
C[22508] = 3.504480971507e-02;
C[22509] = 2.264257955290e-04;
C[22510] = 5.833935024138e-02;
C[22511] = 2.253450032844e-03;
C[22512] = -8.818244414744e-03;
C[22513] = 8.619085201361e-03;
C[22514] = 2.120817321733e+00;
C[22515] = -5.950114730159e-03;
C[22516] = -4.122327419381e-02;
C[22517] = -6.288730375291e+00;
C[22518] = -2.437414054607e-03;
C[22519] = -9.081472670616e-03;
C[22520] = 2.103475722836e+00;
C[22521] = -8.678451322168e-04;
C[22522] = 8.337912595426e-03;
C[22523] = 3.884238889568e-01;
C[22524] = -1.307471162544e-03;
C[22525] = -1.894266319315e-02;
C[22526] = 9.175940310446e-02;
C[22527] = 9.370141711936e-04;
C[22528] = 7.810925549984e-03;
C[22529] = 3.971813919114e-01;
C[22530] = 1.227942638553e-02;
C[22531] = 3.422852770520e-02;
C[22532] = 1.087640186251e+00;
C[22533] = 9.184073290278e-04;
C[22534] = 2.183269611342e-03;
C[22535] = 6.468676960783e-02;
C[22536] = 8.033882703142e-01;
C[22537] = 7.026531161202e-03;
C[22538] = 3.174181840117e-03;
C[22539] = 2.129877447433e+00;
C[22540] = -2.900779374221e-03;
C[22541] = -1.238612979471e-02;
C[22542] = -7.077720971119e+00;
C[22543] = -4.064026254439e-03;
C[22544] = -2.865752556085e-02;
C[22545] = 5.187069975626e-02;
C[22546] = 3.351010269367e-04;
C[22547] = 2.362059724256e-04;
C[22548] = 3.819875730053e-01;
C[22549] = 4.920222017166e-04;
C[22550] = 4.395783841885e-04;
C[22551] = 5.803542337325e-02;
C[22552] = -1.235251781299e-03;
C[22553] = -7.690416143498e-03;
C[22554] = 4.066979170185e+00;
C[22555] = 1.766168907003e-03;
C[22556] = 3.691161341417e-02;
C[22557] = 5.057042689781e-01;
C[22558] = 4.774535726071e-04;
C[22559] = 6.982262612029e-03;
C[22560] = 7.138932684339e-03;
C[22561] = 8.529989367747e-01;
C[22562] = 4.659254513132e-02;
C[22563] = -2.552835871755e-03;
C[22564] = 2.099920829236e+00;
C[22565] = -6.320288246564e-03;
C[22566] = -4.217079919041e-03;
C[22567] = -7.085642666989e+00;
C[22568] = -3.308598480472e-02;
C[22569] = 3.317943704187e-04;
C[22570] = 5.542762053498e-02;
C[22571] = 3.238543781284e-03;
C[22572] = 5.261205855343e-04;
C[22573] = 3.840709699648e-01;
C[22574] = 7.948984771187e-03;
C[22575] = -1.242904464529e-03;
C[22576] = 5.477352820889e-02;
C[22577] = -1.172606062721e-02;
C[22578] = 1.619636456641e-03;
C[22579] = 4.044139879392e+00;
C[22580] = -1.074237225336e-02;
C[22581] = 4.606783338339e-04;
C[22582] = 5.029297181227e-01;
C[22583] = 4.042747966936e-04;
C[22584] = 3.398954891130e-03;
C[22585] = 4.767238288837e-02;
C[22586] = 8.572288168959e-01;
C[22587] = -1.227307144115e-02;
C[22588] = -9.409557945220e-03;
C[22589] = 2.111099530409e+00;
C[22590] = -2.912465763944e-02;
C[22591] = -3.279309559523e-02;
C[22592] = -7.229155347650e+00;
C[22593] = 2.415812743634e-04;
C[22594] = 3.341961768111e-03;
C[22595] = 5.583340886164e-02;
C[22596] = 4.677913759830e-04;
C[22597] = 7.727090697044e-03;
C[22598] = 3.975625822900e-01;
C[22599] = -7.713772747427e-03;
C[22600] = -1.171141616673e-02;
C[22601] = 1.788549568644e-02;
C[22602] = 3.714072847031e-02;
C[22603] = -9.576908998228e-03;
C[22604] = 4.177191518681e+00;
C[22605] = 6.976767521269e-03;
C[22606] = 4.852306797151e-04;
C[22607] = 5.260708382919e-01;
C[22608] = 0.000000000000e+00;
C[22609] = 0.000000000000e+00;
C[22610] = 0.000000000000e+00;
C[22611] = 0.000000000000e+00;
C[22612] = 0.000000000000e+00;
C[22613] = 0.000000000000e+00;
C[22614] = 0.000000000000e+00;
C[22615] = 0.000000000000e+00;
C[22616] = 0.000000000000e+00;
C[22617] = 0.000000000000e+00;
C[22618] = 0.000000000000e+00;
C[22619] = 0.000000000000e+00;
C[22620] = 0.000000000000e+00;
C[22621] = 0.000000000000e+00;
C[22622] = 0.000000000000e+00;
C[22623] = 0.000000000000e+00;
C[22624] = 0.000000000000e+00;
C[22625] = 0.000000000000e+00;
C[22626] = 1.000000000000e+00;
C[22627] = 0.000000000000e+00;
C[22628] = 0.000000000000e+00;
C[22629] = 2.000000000000e-01;
C[22630] = -0.000000000000e+00;
C[22631] = -0.000000000000e+00;
C[22632] = 0.000000000000e+00;
C[22633] = 0.000000000000e+00;
C[22634] = 0.000000000000e+00;
C[22635] = 0.000000000000e+00;
C[22636] = 0.000000000000e+00;
C[22637] = 0.000000000000e+00;
C[22638] = 0.000000000000e+00;
C[22639] = 0.000000000000e+00;
C[22640] = 0.000000000000e+00;
C[22641] = 0.000000000000e+00;
C[22642] = 0.000000000000e+00;
C[22643] = 0.000000000000e+00;
C[22644] = 0.000000000000e+00;
C[22645] = 0.000000000000e+00;
C[22646] = 0.000000000000e+00;
C[22647] = 0.000000000000e+00;
C[22648] = 0.000000000000e+00;
C[22649] = 0.000000000000e+00;
C[22650] = 0.000000000000e+00;
C[22651] = 1.000000000000e+00;
C[22652] = 0.000000000000e+00;
C[22653] = -0.000000000000e+00;
C[22654] = 2.000000000000e-01;
C[22655] = -0.000000000000e+00;
C[22656] = 0.000000000000e+00;
C[22657] = 0.000000000000e+00;
C[22658] = 0.000000000000e+00;
C[22659] = 0.000000000000e+00;
C[22660] = 0.000000000000e+00;
C[22661] = 0.000000000000e+00;
C[22662] = 0.000000000000e+00;
C[22663] = 0.000000000000e+00;
C[22664] = 0.000000000000e+00;
C[22665] = 0.000000000000e+00;
C[22666] = 0.000000000000e+00;
C[22667] = 0.000000000000e+00;
C[22668] = 0.000000000000e+00;
C[22669] = 0.000000000000e+00;
C[22670] = 0.000000000000e+00;
C[22671] = 0.000000000000e+00;
C[22672] = 0.000000000000e+00;
C[22673] = 0.000000000000e+00;
C[22674] = 0.000000000000e+00;
C[22675] = 0.000000000000e+00;
C[22676] = 1.000000000000e+00;
C[22677] = -0.000000000000e+00;
C[22678] = -0.000000000000e+00;
C[22679] = 2.000000000000e-01;
C[22680] = 6.162562283158e-02;
C[22681] = 1.325963045412e-02;
C[22682] = -5.746236745263e-04;
C[22683] = 3.842712304793e-01;
C[22684] = -3.107935242090e-03;
C[22685] = -2.782799074670e-03;
C[22686] = 5.175463652531e-02;
C[22687] = -1.224246949873e-03;
C[22688] = -5.556403368833e-04;
C[22689] = 1.298382986826e-01;
C[22690] = 1.344867948946e-03;
C[22691] = 1.183417459924e-04;
C[22692] = 3.118610174531e-02;
C[22693] = -6.209143788698e-04;
C[22694] = -3.555380684409e-04;
C[22695] = 2.496317784704e-03;
C[22696] = -8.211013591314e-05;
C[22697] = -3.373696224077e-05;
C[22698] = 2.432912482751e-03;
C[22699] = -1.267925018150e-04;
C[22700] = 4.487716599568e-07;
C[22701] = -2.010788775731e-05;
C[22702] = -4.547910921377e-06;
C[22703] = -9.401045194527e-07;
C[22704] = 1.336460174169e-02;
C[22705] = 2.309068006212e-02;
C[22706] = -2.165562671038e-03;
C[22707] = -3.281365678307e-03;
C[22708] = 3.964022544289e-01;
C[22709] = 9.211521892770e-03;
C[22710] = -1.183161610758e-03;
C[22711] = 5.509798991603e-02;
C[22712] = 3.414825583174e-03;
C[22713] = 1.351866034784e-03;
C[22714] = 1.264452472341e-01;
C[22715] = -2.441389059483e-04;
C[22716] = -6.314255952022e-04;
C[22717] = 3.244177470134e-02;
C[22718] = 8.781320143095e-04;
C[22719] = -7.880541223820e-05;
C[22720] = 2.677388158100e-03;
C[22721] = 1.790374105697e-04;
C[22722] = -1.168123572615e-04;
C[22723] = 2.633694609424e-03;
C[22724] = 2.216758933651e-04;
C[22725] = -4.242109380685e-06;
C[22726] = -1.771124691748e-05;
C[22727] = 1.009746385264e-06;
C[22728] = -5.307658074480e-04;
C[22729] = -2.185339050441e-03;
C[22730] = 5.793056043452e-02;
C[22731] = -2.977396675787e-03;
C[22732] = 9.542658493753e-03;
C[22733] = 3.858600482872e-01;
C[22734] = -4.927595801900e-04;
C[22735] = 3.315318015378e-03;
C[22736] = 5.471454323814e-02;
C[22737] = 1.212656037976e-04;
C[22738] = -2.454573312419e-04;
C[22739] = 1.296546737921e-01;
C[22740] = -3.674040057622e-04;
C[22741] = 8.962922179802e-04;
C[22742] = 3.122071561245e-02;
C[22743] = -2.894439406911e-05;
C[22744] = 1.722717884121e-04;
C[22745] = 2.644311475772e-03;
C[22746] = 1.533046489714e-05;
C[22747] = 2.015398763097e-04;
C[22748] = 2.714656133380e-03;
C[22749] = -4.806851833577e-07;
C[22750] = 5.480284741651e-07;
C[22751] = -2.045622629872e-05;
C[22752] = 3.838972252117e-01;
C[22753] = -2.876632219980e-03;
C[22754] = -3.005274037684e-03;
C[22755] = 1.169065362416e-01;
C[22756] = 3.328013517440e-03;
C[22757] = 5.688941548397e-04;
C[22758] = 3.834711461715e-01;
C[22759] = 2.190028900877e-03;
C[22760] = 1.279344521666e-03;
C[22761] = 3.116116806081e-02;
C[22762] = -6.044433646471e-04;
C[22763] = -3.692624965553e-04;
C[22764] = 1.324855857193e-01;
C[22765] = 5.621494274959e-04;
C[22766] = 1.538331573980e-04;
C[22767] = 3.131203683745e-02;
C[22768] = 1.841683192443e-04;
C[22769] = 1.687540390987e-04;
C[22770] = 5.861924291417e-02;
C[22771] = 4.497069410719e-04;
C[22772] = 1.535502772098e-03;
C[22773] = 2.560489587050e-03;
C[22774] = 2.014595516818e-05;
C[22775] = 7.159612200874e-05;
C[22776] = -2.945167674855e-03;
C[22777] = 3.963491705903e-01;
C[22778] = 9.381144527920e-03;
C[22779] = 3.432562351444e-03;
C[22780] = 9.487547901661e-02;
C[22781] = -1.969113837480e-02;
C[22782] = 2.115038260809e-03;
C[22783] = 3.864728191076e-01;
C[22784] = 7.674932762651e-03;
C[22785] = -6.100632077208e-04;
C[22786] = 3.243823577877e-02;
C[22787] = 8.855246202580e-04;
C[22788] = 5.691193497628e-04;
C[22789] = 1.305129050966e-01;
C[22790] = -1.809174011165e-03;
C[22791] = 1.795915415109e-04;
C[22792] = 3.169093927185e-02;
C[22793] = 7.879451900196e-04;
C[22794] = 4.764956682926e-04;
C[22795] = 5.912354424403e-02;
C[22796] = 2.048634503249e-03;
C[22797] = 2.157440616404e-05;
C[22798] = 2.592589757433e-03;
C[22799] = 1.021516150750e-04;
C[22800] = -2.945479124484e-03;
C[22801] = 9.113489850109e-03;
C[22802] = 3.859900165570e-01;
C[22803] = 6.557703220307e-04;
C[22804] = -1.975556341296e-02;
C[22805] = 9.470717357425e-02;
C[22806] = 1.178398249676e-03;
C[22807] = 7.898486419460e-03;
C[22808] = 3.980600203484e-01;
C[22809] = -3.663834050951e-04;
C[22810] = 8.715965447988e-04;
C[22811] = 3.122938016378e-02;
C[22812] = 1.596249018774e-04;
C[22813] = -1.813469013710e-03;
C[22814] = 1.307543166125e-01;
C[22815] = 1.628000470109e-04;
C[22816] = 8.002407238018e-04;
C[22817] = 3.272494522155e-02;
C[22818] = 1.565327739894e-03;
C[22819] = 1.979344400262e-03;
C[22820] = 6.442607803634e-02;
C[22821] = 7.310889904724e-05;
C[22822] = 9.823613394671e-05;
C[22823] = 2.830808405219e-03;
C[22824] = 5.214219139146e-02;
C[22825] = -3.527594169171e-04;
C[22826] = -8.625601134410e-05;
C[22827] = 3.830781445055e-01;
C[22828] = 1.338967235133e-03;
C[22829] = 1.022564391278e-03;
C[22830] = 5.414612321715e-02;
C[22831] = -1.477884796791e-03;
C[22832] = -8.663190552049e-03;
C[22833] = 2.522154775781e-03;
C[22834] = -2.344526598216e-05;
C[22835] = -1.844156146052e-06;
C[22836] = 3.128583672639e-02;
C[22837] = 1.278534731325e-04;
C[22838] = 1.524111231177e-04;
C[22839] = 1.289455514644e-01;
C[22840] = -1.641760971400e-04;
C[22841] = -7.040971795870e-04;
C[22842] = 5.081896608771e-01;
C[22843] = 5.984403584150e-04;
C[22844] = 7.703541674876e-03;
C[22845] = 3.726307598704e-02;
C[22846] = 6.168197850242e-05;
C[22847] = 5.529149329526e-04;
C[22848] = -3.873165711328e-04;
C[22849] = 5.567585883421e-02;
C[22850] = 3.282159744841e-03;
C[22851] = 1.417575524138e-03;
C[22852] = 3.857890530133e-01;
C[22853] = 7.701006134016e-03;
C[22854] = -1.474510149787e-03;
C[22855] = 5.107901740722e-02;
C[22856] = -1.162972144025e-02;
C[22857] = -2.631477733046e-05;
C[22858] = 2.715912752645e-03;
C[22859] = 1.700612370430e-04;
C[22860] = 1.326714274617e-04;
C[22861] = 3.164535486557e-02;
C[22862] = 7.870753714388e-04;
C[22863] = -1.639511206730e-04;
C[22864] = 1.285966915378e-01;
C[22865] = -1.097025857187e-03;
C[22866] = 5.590712497569e-04;
C[22867] = 5.046356620892e-01;
C[22868] = 5.077078137043e-04;
C[22869] = 5.979382881039e-05;
C[22870] = 3.705734656631e-02;
C[22871] = 1.377921774692e-04;
C[22872] = -1.237839100508e-04;
C[22873] = 3.369827994149e-03;
C[22874] = 5.556132770755e-02;
C[22875] = 1.126709720847e-03;
C[22876] = 7.526192901577e-03;
C[22877] = 3.971109741561e-01;
C[22878] = -8.670712417728e-03;
C[22879] = -1.166166594347e-02;
C[22880] = 1.922570762649e-02;
C[22881] = -4.946533785270e-06;
C[22882] = 1.760375713013e-04;
C[22883] = 2.700763773733e-03;
C[22884] = 1.585783857108e-04;
C[22885] = 7.780291992813e-04;
C[22886] = 3.266167547539e-02;
C[22887] = -7.045986372989e-04;
C[22888] = -1.099155490735e-03;
C[22889] = 1.259971971242e-01;
C[22890] = 7.654723096136e-03;
C[22891] = 6.042952322547e-04;
C[22892] = 5.254785159717e-01;
C[22893] = 5.507049486021e-04;
C[22894] = 1.425389720628e-04;
C[22895] = 3.865642921115e-02;
C[22896] = -7.047835948528e+00;
C[22897] = 5.666126046814e-02;
C[22898] = -8.720138956134e-03;
C[22899] = 2.144065202995e+00;
C[22900] = -3.928258572789e-04;
C[22901] = -1.436294448303e-02;
C[22902] = 8.012264463059e-01;
C[22903] = -2.141159139508e-02;
C[22904] = -1.088461985883e-02;
C[22905] = 6.528462091881e-02;
C[22906] = 1.745013631854e-02;
C[22907] = 2.374735286324e-03;
C[22908] = 3.822638162041e-01;
C[22909] = -9.282284870046e-03;
C[22910] = -5.339718609621e-03;
C[22911] = 5.171600054455e-02;
C[22912] = -1.566418401815e-03;
C[22913] = -7.007880895765e-04;
C[22914] = 7.903780151451e-02;
C[22915] = -2.558668602195e-03;
C[22916] = 4.146782345005e-04;
C[22917] = 2.450880247199e-03;
C[22918] = -1.008475052519e-04;
C[22919] = 1.744421722037e-05;
C[22920] = 5.876068621955e-02;
C[22921] = -7.204750373647e+00;
C[22922] = 2.479257611290e-02;
C[22923] = -3.396508953591e-03;
C[22924] = 2.109556463383e+00;
C[22925] = 1.331381951024e-02;
C[22926] = -2.074506808887e-02;
C[22927] = 8.454462212049e-01;
C[22928] = 4.845227684906e-02;
C[22929] = 1.755510760611e-02;
C[22930] = 2.549757304308e-02;
C[22931] = -1.070969166870e-03;
C[22932] = -9.438669003201e-03;
C[22933] = 3.930526425439e-01;
C[22934] = 9.206029013216e-03;
C[22935] = -1.499150441984e-03;
C[22936] = 5.500547763234e-02;
C[22937] = 3.381075221815e-03;
C[22938] = -2.311642814721e-03;
C[22939] = 8.425441242336e-02;
C[22940] = 6.163622130652e-03;
C[22941] = -8.987445489931e-05;
C[22942] = 2.658249973811e-03;
C[22943] = 2.175397000048e-04;
C[22944] = -7.842981614568e-03;
C[22945] = 2.439704852483e-02;
C[22946] = -7.056471355142e+00;
C[22947] = -1.768830866276e-02;
C[22948] = 1.828316690270e-02;
C[22949] = 2.105003657079e+00;
C[22950] = -9.940568011959e-03;
C[22951] = 4.743029547169e-02;
C[22952] = 8.371657987534e-01;
C[22953] = 2.418593153402e-03;
C[22954] = -1.090745546274e-03;
C[22955] = 6.473149173168e-02;
C[22956] = -5.559196470940e-03;
C[22957] = 9.548225670753e-03;
C[22958] = 3.789311638571e-01;
C[22959] = -6.133316457076e-04;
C[22960] = 3.281084694132e-03;
C[22961] = 5.426500101734e-02;
C[22962] = 7.454853185038e-04;
C[22963] = 5.698668793413e-03;
C[22964] = 8.725818463265e-02;
C[22965] = 3.141810161727e-05;
C[22966] = 1.977819619583e-04;
C[22967] = 2.736673547709e-03;
C[22968] = 2.136585097644e+00;
C[22969] = 4.698160212935e-03;
C[22970] = -1.824585590069e-02;
C[22971] = -6.230098649527e+00;
C[22972] = 1.478304382126e-02;
C[22973] = 5.398239166086e-04;
C[22974] = 2.134844697961e+00;
C[22975] = 2.548904313713e-02;
C[22976] = 4.039355471635e-04;
C[22977] = 3.818898109366e-01;
C[22978] = -9.033935544875e-03;
C[22979] = -5.587073832836e-03;
C[22980] = 1.174652681756e-01;
C[22981] = 8.935360898563e-03;
C[22982] = 2.572715976624e-03;
C[22983] = 3.849615698862e-01;
C[22984] = 2.935122567572e-03;
C[22985] = 2.071787029846e-03;
C[22986] = 9.933445761513e-01;
C[22987] = 8.375897316824e-03;
C[22988] = 2.515332160150e-02;
C[22989] = 5.878257845716e-02;
C[22990] = 4.694817185802e-04;
C[22991] = 1.472419308951e-03;
C[22992] = 2.862525487414e-03;
C[22993] = 2.108494786609e+00;
C[22994] = 1.505288758604e-02;
C[22995] = 1.687402050134e-02;
C[22996] = -6.282921684550e+00;
C[22997] = -4.260439823081e-02;
C[22998] = 2.433635905529e-02;
C[22999] = 2.114112057528e+00;
C[23000] = -1.035518499463e-02;
C[23001] = -9.119517302811e-03;
C[23002] = 3.929995587052e-01;
C[23003] = 9.386711704920e-03;
C[23004] = 9.039909732567e-03;
C[23005] = 9.652117727922e-02;
C[23006] = -1.935411007662e-02;
C[23007] = 2.840526164537e-03;
C[23008] = 3.880588481332e-01;
C[23009] = 7.344045417741e-03;
C[23010] = 8.665723451531e-03;
C[23011] = 9.994160576364e-01;
C[23012] = 3.080967998829e-02;
C[23013] = 4.904617175817e-04;
C[23014] = 5.935786120324e-02;
C[23015] = 2.105708134279e-03;
C[23016] = -1.761654547932e-02;
C[23017] = 1.135317865703e-02;
C[23018] = 2.107603022476e+00;
C[23019] = 2.277347260428e-03;
C[23020] = -4.389289899408e-02;
C[23021] = -6.276009690268e+00;
C[23022] = -1.102989774474e-03;
C[23023] = -7.158775475006e-03;
C[23024] = 2.129495580829e+00;
C[23025] = -5.502398659435e-03;
C[23026] = 9.107996970556e-03;
C[23027] = 3.790611321269e-01;
C[23028] = 2.659592143815e-03;
C[23029] = -1.941853511478e-02;
C[23030] = 1.019785126586e-01;
C[23031] = 1.952125419444e-03;
C[23032] = 7.563249596427e-03;
C[23033] = 3.980889074070e-01;
C[23034] = 2.541057222637e-02;
C[23035] = 2.996436542696e-02;
C[23036] = 1.083322228009e+00;
C[23037] = 1.492681286613e-03;
C[23038] = 2.043385098900e-03;
C[23039] = 6.440837157472e-02;
C[23040] = 8.089775436288e-01;
C[23041] = -4.137024212060e-03;
C[23042] = -1.810496635042e-03;
C[23043] = 2.126984664642e+00;
C[23044] = 8.814938541756e-03;
C[23045] = -4.219666942445e-03;
C[23046] = -7.101603996173e+00;
C[23047] = -4.062609184793e-03;
C[23048] = -3.678802840252e-02;
C[23049] = 5.210355541070e-02;
C[23050] = -6.687482481436e-04;
C[23051] = -2.068280768617e-04;
C[23052] = 3.845685682203e-01;
C[23053] = 2.064455138860e-03;
C[23054] = 1.796291561045e-03;
C[23055] = 5.132463966005e-02;
C[23056] = -2.075352875993e-03;
C[23057] = -7.873795555085e-03;
C[23058] = 4.087433183891e+00;
C[23059] = 3.141654048547e-03;
C[23060] = 4.264875052642e-02;
C[23061] = 5.095794120792e-01;
C[23062] = 8.492615051323e-04;
C[23063] = 6.286843490077e-03;
C[23064] = -4.672983820275e-03;
C[23065] = 8.570035995684e-01;
C[23066] = 4.676713006096e-02;
C[23067] = 1.003997560235e-02;
C[23068] = 2.100436735642e+00;
C[23069] = -1.110838118389e-02;
C[23070] = -3.995116244706e-03;
C[23071] = -7.104876914915e+00;
C[23072] = -3.113028891326e-02;
C[23073] = -7.294880230748e-04;
C[23074] = 5.558334655052e-02;
C[23075] = 3.247926423595e-03;
C[23076] = 2.162669190833e-03;
C[23077] = 3.873750820389e-01;
C[23078] = 7.365769310983e-03;
C[23079] = -2.071978228989e-03;
C[23080] = 4.719635235471e-02;
C[23081] = -1.239679646545e-02;
C[23082] = 2.540384015512e-03;
C[23083] = 4.058747777274e+00;
C[23084] = -8.124416749419e-03;
C[23085] = 8.145421581940e-04;
C[23086] = 5.070606874327e-01;
C[23087] = 1.643515379479e-03;
C[23088] = -2.247491322180e-03;
C[23089] = 4.755232506856e-02;
C[23090] = 8.541014881417e-01;
C[23091] = -2.648760469215e-03;
C[23092] = -1.332998221613e-02;
C[23093] = 2.110514656981e+00;
C[23094] = -3.693846571609e-02;
C[23095] = -3.176917897752e-02;
C[23096] = -7.221396843603e+00;
C[23097] = -2.689316627440e-04;
C[23098] = 3.336077632790e-03;
C[23099] = 5.511178548675e-02;
C[23100] = 1.919152229027e-03;
C[23101] = 7.195305556666e-03;
C[23102] = 3.971398612146e-01;
C[23103] = -7.881317420763e-03;
C[23104] = -1.242874096866e-02;
C[23105] = 1.929177861750e-02;
C[23106] = 4.195144507297e-02;
C[23107] = -6.623338724214e-03;
C[23108] = 4.174405415952e+00;
C[23109] = 6.248043747678e-03;
C[23110] = 1.732779272159e-03;
C[23111] = 5.259003760425e-01;
C[23112] = 0.000000000000e+00;
C[23113] = 0.000000000000e+00;
C[23114] = 0.000000000000e+00;
C[23115] = 0.000000000000e+00;
C[23116] = 0.000000000000e+00;
C[23117] = 0.000000000000e+00;
C[23118] = 0.000000000000e+00;
C[23119] = 0.000000000000e+00;
C[23120] = 0.000000000000e+00;
C[23121] = 0.000000000000e+00;
C[23122] = 0.000000000000e+00;
C[23123] = 0.000000000000e+00;
C[23124] = 0.000000000000e+00;
C[23125] = 0.000000000000e+00;
C[23126] = 0.000000000000e+00;
C[23127] = 0.000000000000e+00;
C[23128] = 0.000000000000e+00;
C[23129] = 0.000000000000e+00;
C[23130] = 1.000000000000e+00;
C[23131] = 0.000000000000e+00;
C[23132] = 0.000000000000e+00;
C[23133] = 2.000000000000e-01;
C[23134] = -0.000000000000e+00;
C[23135] = -0.000000000000e+00;
C[23136] = 0.000000000000e+00;
C[23137] = 0.000000000000e+00;
C[23138] = 0.000000000000e+00;
C[23139] = 0.000000000000e+00;
C[23140] = 0.000000000000e+00;
C[23141] = 0.000000000000e+00;
C[23142] = 0.000000000000e+00;
C[23143] = 0.000000000000e+00;
C[23144] = 0.000000000000e+00;
C[23145] = 0.000000000000e+00;
C[23146] = 0.000000000000e+00;
C[23147] = 0.000000000000e+00;
C[23148] = 0.000000000000e+00;
C[23149] = 0.000000000000e+00;
C[23150] = 0.000000000000e+00;
C[23151] = 0.000000000000e+00;
C[23152] = 0.000000000000e+00;
C[23153] = 0.000000000000e+00;
C[23154] = 0.000000000000e+00;
C[23155] = 1.000000000000e+00;
C[23156] = 0.000000000000e+00;
C[23157] = -0.000000000000e+00;
C[23158] = 2.000000000000e-01;
C[23159] = -0.000000000000e+00;
C[23160] = 0.000000000000e+00;
C[23161] = 0.000000000000e+00;
C[23162] = 0.000000000000e+00;
C[23163] = 0.000000000000e+00;
C[23164] = 0.000000000000e+00;
C[23165] = 0.000000000000e+00;
C[23166] = 0.000000000000e+00;
C[23167] = 0.000000000000e+00;
C[23168] = 0.000000000000e+00;
C[23169] = 0.000000000000e+00;
C[23170] = 0.000000000000e+00;
C[23171] = 0.000000000000e+00;
C[23172] = 0.000000000000e+00;
C[23173] = 0.000000000000e+00;
C[23174] = 0.000000000000e+00;
C[23175] = 0.000000000000e+00;
C[23176] = 0.000000000000e+00;
C[23177] = 0.000000000000e+00;
C[23178] = 0.000000000000e+00;
C[23179] = 0.000000000000e+00;
C[23180] = 1.000000000000e+00;
C[23181] = -0.000000000000e+00;
C[23182] = -0.000000000000e+00;
C[23183] = 2.000000000000e-01;
C[23184] = 7.302332359630e-02;
C[23185] = 2.134546036326e-02;
C[23186] = 2.273061033322e-03;
C[23187] = 3.783636628124e-01;
C[23188] = -1.569832507573e-02;
C[23189] = -5.398357415859e-03;
C[23190] = 4.974864310712e-02;
C[23191] = -1.925462961536e-03;
C[23192] = -1.640632128162e-04;
C[23193] = 1.314455662224e-01;
C[23194] = 1.672510319436e-03;
C[23195] = 6.928725932965e-05;
C[23196] = 3.009633059645e-02;
C[23197] = -1.266417272354e-03;
C[23198] = -3.277243401590e-04;
C[23199] = 2.361978139792e-03;
C[23200] = -9.208055904654e-05;
C[23201] = -2.542752788612e-07;
C[23202] = 2.238845220058e-03;
C[23203] = -8.175001008097e-05;
C[23204] = 4.263059192454e-05;
C[23205] = -2.747370979087e-05;
C[23206] = 1.126105028316e-06;
C[23207] = 2.549386989638e-06;
C[23208] = 2.095991418543e-02;
C[23209] = 2.965764233990e-02;
C[23210] = 4.764898385956e-03;
C[23211] = -1.533460132959e-02;
C[23212] = 3.920216369496e-01;
C[23213] = 8.116362667631e-03;
C[23214] = -1.811634831230e-03;
C[23215] = 5.584261790357e-02;
C[23216] = 2.747736609687e-03;
C[23217] = 1.646807240915e-03;
C[23218] = 1.268362265938e-01;
C[23219] = 6.019476273295e-04;
C[23220] = -1.242890221768e-03;
C[23221] = 3.253263983173e-02;
C[23222] = 6.389146195025e-04;
C[23223] = -8.674829728814e-05;
C[23224] = 2.742538183639e-03;
C[23225] = 1.382804520500e-04;
C[23226] = -7.401379170175e-05;
C[23227] = 2.805263443832e-03;
C[23228] = 1.837832962125e-04;
C[23229] = 1.107281625947e-06;
C[23230] = -1.206153331726e-05;
C[23231] = -2.896169226893e-07;
C[23232] = 2.327069238015e-03;
C[23233] = 4.558442886311e-03;
C[23234] = 7.496983010287e-02;
C[23235] = -5.603365779543e-03;
C[23236] = 8.729200553590e-03;
C[23237] = 3.673148496321e-01;
C[23238] = -9.130578499261e-05;
C[23239] = 2.607083922357e-03;
C[23240] = 5.129337891348e-02;
C[23241] = 7.288780630924e-05;
C[23242] = 5.881839273532e-04;
C[23243] = 1.311745425605e-01;
C[23244] = -3.380103062304e-04;
C[23245] = 6.759984260764e-04;
C[23246] = 2.943730505805e-02;
C[23247] = 3.613969584116e-06;
C[23248] = 1.279448098905e-04;
C[23249] = 2.460652234800e-03;
C[23250] = 4.934753059873e-05;
C[23251] = 1.558606090968e-04;
C[23252] = 2.492926377829e-03;
C[23253] = 2.578940570454e-06;
C[23254] = -1.156474101506e-06;
C[23255] = -2.296945755214e-05;
C[23256] = 3.781683837141e-01;
C[23257] = -1.551442128699e-02;
C[23258] = -5.448620901921e-03;
C[23259] = 1.215274077219e-01;
C[23260] = 1.482557132700e-02;
C[23261] = 2.175105529723e-03;
C[23262] = 3.872401150458e-01;
C[23263] = 3.459147445011e-03;
C[23264] = 2.443985145450e-03;
C[23265] = 3.008331198989e-02;
C[23266] = -1.254878218928e-03;
C[23267] = -3.276939810556e-04;
C[23268] = 1.335123031684e-01;
C[23269] = 1.174761991071e-03;
C[23270] = 9.629338397950e-05;
C[23271] = 3.152925408292e-02;
C[23272] = 2.054400295379e-04;
C[23273] = 1.884468159556e-04;
C[23274] = 5.939110044520e-02;
C[23275] = 4.326049659619e-04;
C[23276] = 8.804864305165e-04;
C[23277] = 2.587458797330e-03;
C[23278] = 1.306739846187e-05;
C[23279] = 3.701141241460e-05;
C[23280] = -1.517202972617e-02;
C[23281] = 3.910590943610e-01;
C[23282] = 8.011171958498e-03;
C[23283] = 1.450391341380e-02;
C[23284] = 9.843515111371e-02;
C[23285] = -1.705386766293e-02;
C[23286] = 3.328545271102e-03;
C[23287] = 3.867967559544e-01;
C[23288] = 7.391008249887e-03;
C[23289] = -1.231330915717e-03;
C[23290] = 3.246847032583e-02;
C[23291] = 6.281298530703e-04;
C[23292] = 1.153318130191e-03;
C[23293] = 1.304415513141e-01;
C[23294] = -1.466812040342e-03;
C[23295] = 1.991319969127e-04;
C[23296] = 3.164427212310e-02;
C[23297] = 7.595390334548e-04;
C[23298] = 4.369595748646e-04;
C[23299] = 5.989662983773e-02;
C[23300] = 2.387822661631e-03;
C[23301] = 1.364407538995e-05;
C[23302] = 2.621161596704e-03;
C[23303] = 1.163823908537e-04;
C[23304] = -5.246818848649e-03;
C[23305] = 7.739403988651e-03;
C[23306] = 3.677512826524e-01;
C[23307] = 2.159757842874e-03;
C[23308] = -1.725498416222e-02;
C[23309] = 1.129763857115e-01;
C[23310] = 2.329017150964e-03;
C[23311] = 7.589839793399e-03;
C[23312] = 4.010470028500e-01;
C[23313] = -3.176217690117e-04;
C[23314] = 6.137840409038e-04;
C[23315] = 2.946640059274e-02;
C[23316] = 9.527020485622e-05;
C[23317] = -1.480219806961e-03;
C[23318] = 1.324641748591e-01;
C[23319] = 1.826907126630e-04;
C[23320] = 7.715703530898e-04;
C[23321] = 3.293155102414e-02;
C[23322] = 8.865670937264e-04;
C[23323] = 2.369889806171e-03;
C[23324] = 6.492460011307e-02;
C[23325] = 3.769157655858e-05;
C[23326] = 1.151031283918e-04;
C[23327] = 2.849586260637e-03;
C[23328] = 5.133811778825e-02;
C[23329] = -2.407416567045e-03;
C[23330] = -5.154926442294e-04;
C[23331] = 3.853666983060e-01;
C[23332] = 3.935938953763e-03;
C[23333] = 2.929027827288e-03;
C[23334] = 4.912920121603e-02;
C[23335] = -2.082181752735e-03;
C[23336] = -5.460407180107e-03;
C[23337] = 2.467943118534e-03;
C[23338] = -1.264670796758e-04;
C[23339] = -2.466515436500e-05;
C[23340] = 3.140435963359e-02;
C[23341] = 2.396249090901e-04;
C[23342] = 2.226914244179e-04;
C[23343] = 1.285941448122e-01;
C[23344] = -1.598394114378e-04;
C[23345] = -3.757204647775e-04;
C[23346] = 5.118288926102e-01;
C[23347] = 7.446912294859e-04;
C[23348] = 3.054156457955e-03;
C[23349] = 3.755178211330e-02;
C[23350] = 4.860112950085e-05;
C[23351] = 1.779868114733e-04;
C[23352] = -2.485917341981e-03;
C[23353] = 5.463502264426e-02;
C[23354] = 2.783753744746e-03;
C[23355] = 4.010651122585e-03;
C[23356] = 3.874306389260e-01;
C[23357] = 7.229412273839e-03;
C[23358] = -2.068614481974e-03;
C[23359] = 4.337283933270e-02;
C[23360] = -1.357157948540e-02;
C[23361] = -1.294441844095e-04;
C[23362] = 2.662031833018e-03;
C[23363] = 1.397227980498e-04;
C[23364] = 2.422069413762e-04;
C[23365] = 3.168653098788e-02;
C[23366] = 7.475418517858e-04;
C[23367] = -1.589349267204e-04;
C[23368] = 1.280447951316e-01;
C[23369] = -1.235063794230e-03;
C[23370] = 7.293399169096e-04;
C[23371] = 5.117748949195e-01;
C[23372] = 3.474976515756e-03;
C[23373] = 4.783700108699e-05;
C[23374] = 3.761769871578e-02;
C[23375] = 3.444578486060e-04;
C[23376] = -6.169058488319e-04;
C[23377] = 2.902498089564e-03;
C[23378] = 5.258990294590e-02;
C[23379] = 3.027016662334e-03;
C[23380] = 7.033302649687e-03;
C[23381] = 3.999392718020e-01;
C[23382] = -5.424531936883e-03;
C[23383] = -1.357575220113e-02;
C[23384] = 1.884571363106e-02;
C[23385] = -3.044378434657e-05;
C[23386] = 1.485978840419e-04;
C[23387] = 2.547087170295e-03;
C[23388] = 2.273155837479e-04;
C[23389] = 7.356919934415e-04;
C[23390] = 3.285770228760e-02;
C[23391] = -3.733287818959e-04;
C[23392] = -1.235341975279e-03;
C[23393] = 1.259218208165e-01;
C[23394] = 3.034655172104e-03;
C[23395] = 3.524511401799e-03;
C[23396] = 5.261827295645e-01;
C[23397] = 1.768979406271e-04;
C[23398] = 3.467530196204e-04;
C[23399] = 3.869259505869e-02;
C[23400] = -6.985119358859e+00;
C[23401] = 5.774761221717e-02;
C[23402] = -8.238437863285e-03;
C[23403] = 2.120839718728e+00;
C[23404] = -2.925689487710e-02;
C[23405] = -1.828420672193e-02;
C[23406] = 7.637601975502e-01;
C[23407] = -2.356972871278e-02;
C[23408] = -2.913129778934e-04;
C[23409] = 9.021337692862e-02;
C[23410] = 1.582607144734e-02;
C[23411] = -1.418542302716e-03;
C[23412] = 3.648871706657e-01;
C[23413] = -1.140504824664e-02;
C[23414] = -2.324563646380e-03;
C[23415] = 4.893809108576e-02;
C[23416] = -1.674281593729e-03;
C[23417] = -1.139070329240e-05;
C[23418] = 7.483318429780e-02;
C[23419] = -2.646805844482e-03;
C[23420] = 6.722726056985e-04;
C[23421] = 2.279418732508e-03;
C[23422] = -9.827955462451e-05;
C[23423] = 2.707561383272e-05;
C[23424] = 5.003668866063e-02;
C[23425] = -7.176990604306e+00;
C[23426] = 6.856230795529e-02;
C[23427] = -2.218807867009e-02;
C[23428] = 2.128666850916e+00;
C[23429] = 2.528126121858e-02;
C[23430] = -2.222350641128e-02;
C[23431] = 8.645826553178e-01;
C[23432] = 3.826544664383e-02;
C[23433] = 1.544052526951e-02;
C[23434] = 2.860084924965e-02;
C[23435] = 1.177605619111e-02;
C[23436] = -1.106234833624e-02;
C[23437] = 4.002214355499e-01;
C[23438] = 6.221694243997e-03;
C[23439] = -1.603492120556e-03;
C[23440] = 5.624445770395e-02;
C[23441] = 2.620489367528e-03;
C[23442] = -2.490877464479e-03;
C[23443] = 8.753986221998e-02;
C[23444] = 5.417837060890e-03;
C[23445] = -9.467664318281e-05;
C[23446] = 2.761142952699e-03;
C[23447] = 1.844753223512e-04;
C[23448] = -7.158273769408e-03;
C[23449] = 6.443319796239e-02;
C[23450] = -7.029070758104e+00;
C[23451] = -2.110020564046e-02;
C[23452] = 3.553128487394e-02;
C[23453] = 2.048988405459e+00;
C[23454] = 3.154344403851e-04;
C[23455] = 3.642076505700e-02;
C[23456] = 7.865458772288e-01;
C[23457] = -1.364534098022e-03;
C[23458] = 1.156960069146e-02;
C[23459] = 7.991630390008e-02;
C[23460] = -2.511939095221e-03;
C[23461] = 6.920177427837e-03;
C[23462] = 3.581759562329e-01;
C[23463] = 4.321851962169e-05;
C[23464] = 2.438753220049e-03;
C[23465] = 5.078823227526e-02;
C[23466] = 8.546023690190e-04;
C[23467] = 4.786130689938e-03;
C[23468] = 8.155999619098e-02;
C[23469] = 3.506301340751e-05;
C[23470] = 1.565606766641e-04;
C[23471] = 2.536634291060e-03;
C[23472] = 2.116934136762e+00;
C[23473] = -2.578447781808e-02;
C[23474] = -1.800530808803e-02;
C[23475] = -6.197518952075e+00;
C[23476] = 3.627478946311e-02;
C[23477] = 1.382082072279e-03;
C[23478] = 2.173156751700e+00;
C[23479] = 2.816489540171e-02;
C[23480] = 1.823386437847e-03;
C[23481] = 3.646918915674e-01;
C[23482] = -1.124216829364e-02;
C[23483] = -2.357194217599e-03;
C[23484] = 1.355507255082e-01;
C[23485] = 1.098076848757e-02;
C[23486] = -1.622149409307e-04;
C[23487] = 3.875976111459e-01;
C[23488] = 2.711147398386e-03;
C[23489] = 1.640854108480e-03;
C[23490] = 1.005813633966e+00;
C[23491] = 7.932659732802e-03;
C[23492] = 1.333341988625e-02;
C[23493] = 5.941713646273e-02;
C[23494] = 3.930742133525e-04;
C[23495] = 7.324113635887e-04;
C[23496] = -1.873098788597e-02;
C[23497] = 2.109415999144e+00;
C[23498] = 2.117071297210e-02;
C[23499] = 2.984163119911e-02;
C[23500] = -6.284495518177e+00;
C[23501] = -3.453351267249e-02;
C[23502] = 2.651511172238e-02;
C[23503] = 2.088032329766e+00;
C[23504] = -6.601253068560e-03;
C[23505] = -1.087875289708e-02;
C[23506] = 3.992588929613e-01;
C[23507] = 6.202148832744e-03;
C[23508] = 1.065911057437e-02;
C[23509] = 9.142950728918e-02;
C[23510] = -1.465164615517e-02;
C[23511] = 2.624233342865e-03;
C[23512] = 3.855021361493e-01;
C[23513] = 7.240453681239e-03;
C[23514] = 7.944337186844e-03;
C[23515] = 1.010191771400e+00;
C[23516] = 3.655850282338e-02;
C[23517] = 3.972855658417e-04;
C[23518] = 6.005330415489e-02;
C[23519] = 2.402259875194e-03;
C[23520] = -1.525343537772e-02;
C[23521] = 1.774208763897e-02;
C[23522] = 2.057717065864e+00;
C[23523] = 1.075128335294e-03;
C[23524] = -3.855584265824e-02;
C[23525] = -6.228892995608e+00;
C[23526] = 3.999937714135e-04;
C[23527] = -3.643600477256e-03;
C[23528] = 2.176773477689e+00;
C[23529] = -2.173025079169e-03;
C[23530] = 5.844735565016e-03;
C[23531] = 3.586123892532e-01;
C[23532] = -1.775626277799e-04;
C[23533] = -1.485276265446e-02;
C[23534] = 1.210475784301e-01;
C[23535] = 1.551466912499e-03;
C[23536] = 7.478518063393e-03;
C[23537] = 4.023330057960e-01;
C[23538] = 1.334200174714e-02;
C[23539] = 3.659943213457e-02;
C[23540] = 1.094328261537e+00;
C[23541] = 7.337499960054e-04;
C[23542] = 2.396213582099e-03;
C[23543] = 6.492753703685e-02;
C[23544] = 7.955496911727e-01;
C[23545] = -3.413914112756e-02;
C[23546] = -8.168302744350e-03;
C[23547] = 2.135688416904e+00;
C[23548] = 3.866298537561e-02;
C[23549] = 1.240020729788e-02;
C[23550] = -7.118564013394e+00;
C[23551] = -1.959177008692e-03;
C[23552] = -1.460761199426e-02;
C[23553] = 5.052756576689e-02;
C[23554] = -2.199273856370e-03;
C[23555] = -3.809683396151e-04;
C[23556] = 3.857241944061e-01;
C[23557] = 3.231627025527e-03;
C[23558] = 2.151477588822e-03;
C[23559] = 4.818276080077e-02;
C[23560] = -1.240695853014e-03;
C[23561] = -3.015884500651e-03;
C[23562] = 4.113490101862e+00;
C[23563] = 2.045743080912e-03;
C[23564] = 1.017849845667e-02;
C[23565] = 5.133487879560e-01;
C[23566] = 3.558879464584e-04;
C[23567] = 1.226222391280e-03;
C[23568] = -3.477881632168e-02;
C[23569] = 8.404307501316e-01;
C[23570] = 3.995416150480e-02;
C[23571] = 3.919496895319e-02;
C[23572] = 2.100709989198e+00;
C[23573] = -1.085215086845e-02;
C[23574] = -1.687831593464e-03;
C[23575] = -7.139577808607e+00;
C[23576] = -4.124652878176e-02;
C[23577] = -2.234735974174e-03;
C[23578] = 5.503686244464e-02;
C[23579] = 2.615423042439e-03;
C[23580] = 3.262651075961e-03;
C[23581] = 3.861360191209e-01;
C[23582] = 7.118090543834e-03;
C[23583] = -1.227128582253e-03;
C[23584] = 4.155120939269e-02;
C[23585] = -1.388085588990e-02;
C[23586] = 1.810642136901e-03;
C[23587] = 4.109557005561e+00;
C[23588] = 1.060048479794e-02;
C[23589] = 3.449462991527e-04;
C[23590] = 5.143992024081e-01;
C[23591] = 4.082592529604e-03;
C[23592] = -9.348165698205e-03;
C[23593] = 4.136067624137e-02;
C[23594] = 8.124763578773e-01;
C[23595] = 1.348401677553e-02;
C[23596] = -1.375536507255e-02;
C[23597] = 2.154618856728e+00;
C[23598] = -1.389010712977e-02;
C[23599] = -4.132998309637e-02;
C[23600] = -7.219584879832e+00;
C[23601] = -4.642333393080e-04;
C[23602] = 2.775250847405e-03;
C[23603] = 5.208475630768e-02;
C[23604] = 2.223885625364e-03;
C[23605] = 6.882748081040e-03;
C[23606] = 4.012252747480e-01;
C[23607] = -2.980009257427e-03;
C[23608] = -1.388502860563e-02;
C[23609] = 1.828504404284e-02;
C[23610] = 9.832889817406e-03;
C[23611] = 1.133350235997e-02;
C[23612] = 4.175238283278e+00;
C[23613] = 1.210562044939e-03;
C[23614] = 4.121403724947e-03;
C[23615] = 5.260576363910e-01;
C[23616] = 0.000000000000e+00;
C[23617] = 0.000000000000e+00;
C[23618] = 0.000000000000e+00;
C[23619] = 0.000000000000e+00;
C[23620] = 0.000000000000e+00;
C[23621] = 0.000000000000e+00;
C[23622] = 0.000000000000e+00;
C[23623] = 0.000000000000e+00;
C[23624] = 0.000000000000e+00;
C[23625] = 0.000000000000e+00;
C[23626] = 0.000000000000e+00;
C[23627] = 0.000000000000e+00;
C[23628] = 0.000000000000e+00;
C[23629] = 0.000000000000e+00;
C[23630] = 0.000000000000e+00;
C[23631] = 0.000000000000e+00;
C[23632] = 0.000000000000e+00;
C[23633] = 0.000000000000e+00;
C[23634] = 1.000000000000e+00;
C[23635] = 0.000000000000e+00;
C[23636] = 0.000000000000e+00;
C[23637] = 2.000000000000e-01;
C[23638] = -0.000000000000e+00;
C[23639] = -0.000000000000e+00;
C[23640] = 0.000000000000e+00;
C[23641] = 0.000000000000e+00;
C[23642] = 0.000000000000e+00;
C[23643] = 0.000000000000e+00;
C[23644] = 0.000000000000e+00;
C[23645] = 0.000000000000e+00;
C[23646] = 0.000000000000e+00;
C[23647] = 0.000000000000e+00;
C[23648] = 0.000000000000e+00;
C[23649] = 0.000000000000e+00;
C[23650] = 0.000000000000e+00;
C[23651] = 0.000000000000e+00;
C[23652] = 0.000000000000e+00;
C[23653] = 0.000000000000e+00;
C[23654] = 0.000000000000e+00;
C[23655] = 0.000000000000e+00;
C[23656] = 0.000000000000e+00;
C[23657] = 0.000000000000e+00;
C[23658] = 0.000000000000e+00;
C[23659] = 1.000000000000e+00;
C[23660] = 0.000000000000e+00;
C[23661] = -0.000000000000e+00;
C[23662] = 2.000000000000e-01;
C[23663] = -0.000000000000e+00;
C[23664] = 0.000000000000e+00;
C[23665] = 0.000000000000e+00;
C[23666] = 0.000000000000e+00;
C[23667] = 0.000000000000e+00;
C[23668] = 0.000000000000e+00;
C[23669] = 0.000000000000e+00;
C[23670] = 0.000000000000e+00;
C[23671] = 0.000000000000e+00;
C[23672] = 0.000000000000e+00;
C[23673] = 0.000000000000e+00;
C[23674] = 0.000000000000e+00;
C[23675] = 0.000000000000e+00;
C[23676] = 0.000000000000e+00;
C[23677] = 0.000000000000e+00;
C[23678] = 0.000000000000e+00;
C[23679] = 0.000000000000e+00;
C[23680] = 0.000000000000e+00;
C[23681] = 0.000000000000e+00;
C[23682] = 0.000000000000e+00;
C[23683] = 0.000000000000e+00;
C[23684] = 1.000000000000e+00;
C[23685] = -0.000000000000e+00;
C[23686] = -0.000000000000e+00;
C[23687] = 2.000000000000e-01;
C[23688] = 8.899965252033e-02;
C[23689] = -4.932385098629e-03;
C[23690] = -3.666641120732e-03;
C[23691] = 3.674260334885e-01;
C[23692] = 6.790703945366e-03;
C[23693] = 9.264621865218e-04;
C[23694] = 5.060825320136e-02;
C[23695] = 1.811057244576e-03;
C[23696] = 4.992540112636e-04;
C[23697] = 1.315063927869e-01;
C[23698] = -8.204265808112e-04;
C[23699] = -2.662477434400e-04;
C[23700] = 3.016256091576e-02;
C[23701] = 8.905110320996e-04;
C[23702] = 1.856089900996e-04;
C[23703] = 2.465532094404e-03;
C[23704] = 1.090432751562e-04;
C[23705] = 3.120416275016e-05;
C[23706] = 2.500325068868e-03;
C[23707] = 1.589339990508e-04;
C[23708] = 5.252407987488e-05;
C[23709] = -1.418698395557e-05;
C[23710] = 4.217830560995e-06;
C[23711] = 1.916298354543e-06;
C[23712] = -5.231741806350e-03;
C[23713] = 4.087944871615e-02;
C[23714] = 1.046025820725e-02;
C[23715] = 7.062768025136e-03;
C[23716] = 4.044774346289e-01;
C[23717] = 5.617920789653e-03;
C[23718] = 1.807630676494e-03;
C[23719] = 5.595033413297e-02;
C[23720] = 2.585327860181e-03;
C[23721] = -8.403836946593e-04;
C[23722] = 1.281336219901e-01;
C[23723] = 4.494041752236e-04;
C[23724] = 9.096806159501e-04;
C[23725] = 3.296122575902e-02;
C[23726] = 5.419030478850e-04;
C[23727] = 1.079765575855e-04;
C[23728] = 2.713473052141e-03;
C[23729] = 1.358572835214e-04;
C[23730] = 1.545322984359e-04;
C[23731] = 2.742717777608e-03;
C[23732] = 2.016824270178e-04;
C[23733] = 4.030785564288e-06;
C[23734] = -1.839966118392e-05;
C[23735] = 1.251259615831e-06;
C[23736] = -3.624581691759e-03;
C[23737] = 1.047498062865e-02;
C[23738] = 6.104853284294e-02;
C[23739] = 1.040605666763e-03;
C[23740] = 5.864448226124e-03;
C[23741] = 3.602814851134e-01;
C[23742] = 4.397145390507e-04;
C[23743] = 2.398510064710e-03;
C[23744] = 5.242224762621e-02;
C[23745] = -2.634437815085e-04;
C[23746] = 4.503856699832e-04;
C[23747] = 1.292461361554e-01;
C[23748] = 1.918285884106e-04;
C[23749] = 5.505732767594e-04;
C[23750] = 2.960802204738e-02;
C[23751] = 2.698645006187e-05;
C[23752] = 1.258192502541e-04;
C[23753] = 2.564123285334e-03;
C[23754] = 4.182925926036e-05;
C[23755] = 1.837691603879e-04;
C[23756] = 2.709837684246e-03;
C[23757] = 1.574686030113e-06;
C[23758] = 1.116327970375e-06;
C[23759] = -1.246310349378e-05;
C[23760] = 3.673816911426e-01;
C[23761] = 7.002231874565e-03;
C[23762] = 7.703606250071e-04;
C[23763] = 1.325183183064e-01;
C[23764] = -6.847649788482e-03;
C[23765] = -1.158270064776e-03;
C[23766] = 3.859724273750e-01;
C[23767] = -1.693904385793e-03;
C[23768] = -2.426867663973e-04;
C[23769] = 3.015960475936e-02;
C[23770] = 9.056448725788e-04;
C[23771] = 1.738122522935e-04;
C[23772] = 1.333072868875e-01;
C[23773] = -8.685811706900e-04;
C[23774] = -1.440232174534e-04;
C[23775] = 3.153522606473e-02;
C[23776] = -1.286790004620e-04;
C[23777] = -6.112962572370e-05;
C[23778] = 6.000006618991e-02;
C[23779] = -2.459067999314e-04;
C[23780] = -2.024822844425e-04;
C[23781] = 2.621899353263e-03;
C[23782] = -1.017722494693e-05;
C[23783] = -1.222824800216e-05;
C[23784] = 7.282432950858e-03;
C[23785] = 4.037345700865e-01;
C[23786] = 6.957998275803e-03;
C[23787] = -7.103188519176e-03;
C[23788] = 8.926199212493e-02;
C[23789] = -1.447042255960e-02;
C[23790] = -1.697982448834e-03;
C[23791] = 3.855086546964e-01;
C[23792] = 6.631512661224e-03;
C[23793] = 9.232929657991e-04;
C[23794] = 3.291170145620e-02;
C[23795] = 6.234766134047e-04;
C[23796] = -8.856170860696e-04;
C[23797] = 1.300270397159e-01;
C[23798] = -1.356282526983e-03;
C[23799] = -1.281575442062e-04;
C[23800] = 3.174054580094e-02;
C[23801] = 6.773110546986e-04;
C[23802] = -2.413070882470e-04;
C[23803] = 6.108150790663e-02;
C[23804] = 2.331918851605e-03;
C[23805] = -9.942781683750e-06;
C[23806] = 2.687361446384e-03;
C[23807] = 1.124308523017e-04;
C[23808] = 7.122228578204e-04;
C[23809] = 6.475847078304e-03;
C[23810] = 3.610887015591e-01;
C[23811] = -1.192781162895e-03;
C[23812] = -1.437904550701e-02;
C[23813] = 1.157771866024e-01;
C[23814] = -2.019024483852e-04;
C[23815] = 6.853200000446e-03;
C[23816] = 4.012932557836e-01;
C[23817] = 1.713263681862e-04;
C[23818] = 5.990981337951e-04;
C[23819] = 2.966183647709e-02;
C[23820] = -1.463239573279e-04;
C[23821] = -1.350190723477e-03;
C[23822] = 1.320812998708e-01;
C[23823] = -5.847025575125e-05;
C[23824] = 6.887989208195e-04;
C[23825] = 3.290949371403e-02;
C[23826] = -1.915672579495e-04;
C[23827] = 2.335745775413e-03;
C[23828] = 6.476518208527e-02;
C[23829] = -1.178052294860e-05;
C[23830] = 1.117645339937e-04;
C[23831] = 2.839353474365e-03;
C[23832] = 4.965196158768e-02;
C[23833] = 1.025613115542e-03;
C[23834] = 4.737058437077e-05;
C[23835] = 3.872308436385e-01;
C[23836] = -9.224571633861e-04;
C[23837] = 1.334154548645e-04;
C[23838] = 4.625903369867e-02;
C[23839] = -4.293341870025e-05;
C[23840] = 8.575877810849e-05;
C[23841] = 2.401779320159e-03;
C[23842] = 5.584205352201e-05;
C[23843] = 8.301864165408e-07;
C[23844] = 3.161912048229e-02;
C[23845] = -7.645585850971e-05;
C[23846] = -3.611572886794e-05;
C[23847] = 1.282613491671e-01;
C[23848] = 4.504263448818e-05;
C[23849] = 9.178448068883e-05;
C[23850] = 5.146125052016e-01;
C[23851] = -1.100356287640e-04;
C[23852] = -2.903737682824e-04;
C[23853] = 3.773564517176e-02;
C[23854] = -2.458414002976e-05;
C[23855] = -5.766111814997e-05;
C[23856] = 1.057512996978e-03;
C[23857] = 5.684972618885e-02;
C[23858] = 2.344460113507e-03;
C[23859] = -9.671859469095e-04;
C[23860] = 3.848505622717e-01;
C[23861] = 6.951703088070e-03;
C[23862] = -3.019389872301e-05;
C[23863] = 3.923503595950e-02;
C[23864] = -1.371618028428e-02;
C[23865] = 5.880699198302e-05;
C[23866] = 2.773432522534e-03;
C[23867] = 1.222159201738e-04;
C[23868] = -8.023110453644e-05;
C[23869] = 3.169667297263e-02;
C[23870] = 6.953657933277e-04;
C[23871] = 4.589193581999e-05;
C[23872] = 1.276473727290e-01;
C[23873] = -1.205171383085e-03;
C[23874] = -1.120427768220e-04;
C[23875] = 5.163257392771e-01;
C[23876] = 4.431161193098e-03;
C[23877] = -2.475333898915e-05;
C[23878] = 3.789463764488e-02;
C[23879] = 3.906081382462e-04;
C[23880] = 7.535183327219e-05;
C[23881] = 2.520705655722e-03;
C[23882] = 5.165549934157e-02;
C[23883] = 7.001751953377e-05;
C[23884] = 6.757530697215e-03;
C[23885] = 4.021521676113e-01;
C[23886] = 1.184218370832e-04;
C[23887] = -1.373294447967e-02;
C[23888] = 1.801261073423e-02;
C[23889] = 2.944017550737e-06;
C[23890] = 1.315491365574e-04;
C[23891] = 2.513006733025e-03;
C[23892] = -4.028267332829e-05;
C[23893] = 6.857122570980e-04;
C[23894] = 3.296675450254e-02;
C[23895] = 9.396201795381e-05;
C[23896] = -1.206288996111e-03;
C[23897] = 1.258872737667e-01;
C[23898] = -2.953109983698e-04;
C[23899] = 4.445078704490e-03;
C[23900] = 5.256504339854e-01;
C[23901] = -5.809293264071e-05;
C[23902] = 3.913493363509e-04;
C[23903] = 3.865239352909e-02;
C[23904] = -6.979691569913e+00;
C[23905] = -1.240625630446e-02;
C[23906] = -1.063511553333e-02;
C[23907] = 2.122773260701e+00;
C[23908] = 1.952897832835e-02;
C[23909] = 5.098095319976e-03;
C[23910] = 7.942533314241e-01;
C[23911] = 3.034562392078e-02;
C[23912] = 8.687608074208e-03;
C[23913] = 7.581543189031e-02;
C[23914] = -1.240301584047e-02;
C[23915] = -2.353301640634e-03;
C[23916] = 3.773286214247e-01;
C[23917] = 1.248535825368e-02;
C[23918] = 3.054177527235e-03;
C[23919] = 5.133384691043e-02;
C[23920] = 2.141668437402e-03;
C[23921] = 5.627023893564e-04;
C[23922] = 7.987497639040e-02;
C[23923] = 3.662675210271e-03;
C[23924] = 9.483666176539e-04;
C[23925] = 2.487692799186e-03;
C[23926] = 1.361195110615e-04;
C[23927] = 3.282634942775e-05;
C[23928] = -1.839339045889e-02;
C[23929] = -7.079506888948e+00;
C[23930] = 5.444019717741e-02;
C[23931] = 2.521507326122e-02;
C[23932] = 2.175053719479e+00;
C[23933] = 9.238981038297e-03;
C[23934] = 3.015659939507e-02;
C[23935] = 8.577001319851e-01;
C[23936] = 3.742302327213e-02;
C[23937] = -1.270237254819e-02;
C[23938] = 4.982586515599e-02;
C[23939] = 2.162079249513e-03;
C[23940] = 1.279167046109e-02;
C[23941] = 3.995030403360e-01;
C[23942] = 6.122047666819e-03;
C[23943] = 2.101667914570e-03;
C[23944] = 5.579893306083e-02;
C[23945] = 2.605145293569e-03;
C[23946] = 3.519873388053e-03;
C[23947] = 8.799388272209e-02;
C[23948] = 5.681502408283e-03;
C[23949] = 1.285063666479e-04;
C[23950] = 2.782794742227e-03;
C[23951] = 1.997844298468e-04;
C[23952] = -9.793926953883e-03;
C[23953] = 5.473464560532e-02;
C[23954] = -7.149190501464e+00;
C[23955] = 6.929667661059e-03;
C[23956] = 1.116592616155e-02;
C[23957] = 2.012912356576e+00;
C[23958] = 7.726087977964e-03;
C[23959] = 3.528630660014e-02;
C[23960] = 8.131577728504e-01;
C[23961] = -2.311242211661e-03;
C[23962] = 2.176801670908e-03;
C[23963] = 4.726149016801e-02;
C[23964] = 3.167862264934e-03;
C[23965] = 6.255929069728e-03;
C[23966] = 3.663421850878e-01;
C[23967] = 4.843056758243e-04;
C[23968] = 2.452278729979e-03;
C[23969] = 5.281515350895e-02;
C[23970] = 7.213024722905e-04;
C[23971] = 5.240926571850e-03;
C[23972] = 8.528026354073e-02;
C[23973] = 2.361131615330e-05;
C[23974] = 1.835295103385e-04;
C[23975] = 2.689309864079e-03;
C[23976] = 2.121886413782e+00;
C[23977] = 2.400435024981e-02;
C[23978] = 1.524766825934e-03;
C[23979] = -6.204766408484e+00;
C[23980] = -2.066998457245e-02;
C[23981] = -3.116622340526e-03;
C[23982] = 2.138148384990e+00;
C[23983] = -3.015514474737e-02;
C[23984] = -7.701037854464e-03;
C[23985] = 3.772842790788e-01;
C[23986] = 1.273113431052e-02;
C[23987] = 2.897617223178e-03;
C[23988] = 1.199763255456e-01;
C[23989] = -1.198715217371e-02;
C[23990] = -1.941472806366e-03;
C[23991] = 3.876947916734e-01;
C[23992] = -2.647779126977e-03;
C[23993] = -1.389073509731e-03;
C[23994] = 1.017167286729e+00;
C[23995] = -5.714224005236e-03;
C[23996] = -4.621489165135e-03;
C[23997] = 6.020064643421e-02;
C[23998] = -2.893013935064e-04;
C[23999] = -3.151152464008e-04;
C[24000] = 2.936355843820e-02;
C[24001] = 2.160196428632e+00;
C[24002] = 3.303692715513e-02;
C[24003] = -2.578075918633e-02;
C[24004] = -6.291872106849e+00;
C[24005] = -3.383465919305e-02;
C[24006] = -3.011044178478e-02;
C[24007] = 2.088086549262e+00;
C[24008] = -1.285592571490e-02;
C[24009] = 1.297708725917e-02;
C[24010] = 3.987601757936e-01;
C[24011] = 7.349479119407e-03;
C[24012] = -1.224269090441e-02;
C[24013] = 9.182066173658e-02;
C[24014] = -1.488041290873e-02;
C[24015] = -2.614022792460e-03;
C[24016] = 3.882886671317e-01;
C[24017] = 6.099652730419e-03;
C[24018] = -5.557671244314e-03;
C[24019] = 1.030652637389e+00;
C[24020] = 3.544651137439e-02;
C[24021] = -2.817300629999e-04;
C[24022] = 6.123516761471e-02;
C[24023] = 2.307693279209e-03;
C[24024] = 8.133087459493e-04;
C[24025] = 2.639750681132e-02;
C[24026] = 2.029056685491e+00;
C[24027] = -3.806844302892e-03;
C[24028] = -3.200711814123e-02;
C[24029] = -6.243971543556e+00;
C[24030] = -7.082310746553e-03;
C[24031] = -1.008725446959e-02;
C[24032] = 2.151395503810e+00;
C[24033] = 2.839938198534e-03;
C[24034] = 6.979973955470e-03;
C[24035] = 3.671494015335e-01;
C[24036] = -1.975983904484e-03;
C[24037] = -1.478903585614e-02;
C[24038] = 1.095840412439e-01;
C[24039] = -1.327884081993e-03;
C[24040] = 6.283129147179e-03;
C[24041] = 4.006092893602e-01;
C[24042] = -4.354395088512e-03;
C[24043] = 3.565668898024e-02;
C[24044] = 1.090368883446e+00;
C[24045] = -3.046759855518e-04;
C[24046] = 2.308571730663e-03;
C[24047] = 6.475893494818e-02;
C[24048] = 7.751274991505e-01;
C[24049] = 1.451624817602e-02;
C[24050] = -1.207911156338e-04;
C[24051] = 2.163316710259e+00;
C[24052] = -1.459993607582e-02;
C[24053] = -3.759526815599e-04;
C[24054] = -7.134993674364e+00;
C[24055] = -5.693319544850e-04;
C[24056] = 3.426674673185e-03;
C[24057] = 5.037755529676e-02;
C[24058] = 1.319650353617e-03;
C[24059] = 9.196172114442e-05;
C[24060] = 3.889532079369e-01;
C[24061] = -1.838497507012e-03;
C[24062] = -9.925661787437e-04;
C[24063] = 4.330170468648e-02;
C[24064] = 9.100771514905e-04;
C[24065] = 1.989263500205e-03;
C[24066] = 4.122484385356e+00;
C[24067] = -1.497287140098e-03;
C[24068] = -3.821136683470e-03;
C[24069] = 5.150384371159e-01;
C[24070] = -4.944722479193e-04;
C[24071] = -1.136798420397e-03;
C[24072] = 1.527473896883e-02;
C[24073] = 8.756879731028e-01;
C[24074] = 3.420530757606e-02;
C[24075] = -1.562077596970e-02;
C[24076] = 2.074924700769e+00;
C[24077] = -8.117192717107e-03;
C[24078] = -3.145415549403e-04;
C[24079] = -7.161825544307e+00;
C[24080] = -4.193449882934e-02;
C[24081] = 1.388124189804e-03;
C[24082] = 5.669832511671e-02;
C[24083] = 2.398228778775e-03;
C[24084] = -1.921060688094e-03;
C[24085] = 3.876305747071e-01;
C[24086] = 6.381632234803e-03;
C[24087] = 9.228166714678e-04;
C[24088] = 3.625738233236e-02;
C[24089] = -1.318846869784e-02;
C[24090] = -1.541497542151e-03;
C[24091] = 4.125766790579e+00;
C[24092] = 1.586035331741e-02;
C[24093] = -4.960324348986e-04;
C[24094] = 5.167271386983e-01;
C[24095] = 4.415245958571e-03;
C[24096] = 2.095645143796e-04;
C[24097] = 3.613057918295e-02;
C[24098] = 7.978228071577e-01;
C[24099] = -1.446952135842e-03;
C[24100] = -1.033556499508e-02;
C[24101] = 2.168573740365e+00;
C[24102] = 4.079935852679e-03;
C[24103] = -4.226978273705e-02;
C[24104] = -7.219498265719e+00;
C[24105] = 1.388002113650e-04;
C[24106] = 2.540523089110e-03;
C[24107] = 5.204840522431e-02;
C[24108] = -1.076369223800e-03;
C[24109] = 6.225670766410e-03;
C[24110] = 4.014682011879e-01;
C[24111] = 2.021926559180e-03;
C[24112] = -1.320523289323e-02;
C[24113] = 1.843298021766e-02;
C[24114] = -3.936638149329e-03;
C[24115] = 1.610846396869e-02;
C[24116] = 4.171005574535e+00;
C[24117] = -1.142230577795e-03;
C[24118] = 4.430623959059e-03;
C[24119] = 5.254609322777e-01;
C[24120] = 0.000000000000e+00;
C[24121] = 0.000000000000e+00;
C[24122] = 0.000000000000e+00;
C[24123] = 0.000000000000e+00;
C[24124] = 0.000000000000e+00;
C[24125] = 0.000000000000e+00;
C[24126] = 0.000000000000e+00;
C[24127] = 0.000000000000e+00;
C[24128] = 0.000000000000e+00;
C[24129] = 0.000000000000e+00;
C[24130] = 0.000000000000e+00;
C[24131] = 0.000000000000e+00;
C[24132] = 0.000000000000e+00;
C[24133] = 0.000000000000e+00;
C[24134] = 0.000000000000e+00;
C[24135] = 0.000000000000e+00;
C[24136] = 0.000000000000e+00;
C[24137] = 0.000000000000e+00;
C[24138] = 1.000000000000e+00;
C[24139] = 0.000000000000e+00;
C[24140] = 0.000000000000e+00;
C[24141] = 2.000000000000e-01;
C[24142] = -0.000000000000e+00;
C[24143] = -0.000000000000e+00;
C[24144] = 0.000000000000e+00;
C[24145] = 0.000000000000e+00;
C[24146] = 0.000000000000e+00;
C[24147] = 0.000000000000e+00;
C[24148] = 0.000000000000e+00;
C[24149] = 0.000000000000e+00;
C[24150] = 0.000000000000e+00;
C[24151] = 0.000000000000e+00;
C[24152] = 0.000000000000e+00;
C[24153] = 0.000000000000e+00;
C[24154] = 0.000000000000e+00;
C[24155] = 0.000000000000e+00;
C[24156] = 0.000000000000e+00;
C[24157] = 0.000000000000e+00;
C[24158] = 0.000000000000e+00;
C[24159] = 0.000000000000e+00;
C[24160] = 0.000000000000e+00;
C[24161] = 0.000000000000e+00;
C[24162] = 0.000000000000e+00;
C[24163] = 1.000000000000e+00;
C[24164] = 0.000000000000e+00;
C[24165] = -0.000000000000e+00;
C[24166] = 2.000000000000e-01;
C[24167] = -0.000000000000e+00;
C[24168] = 0.000000000000e+00;
C[24169] = 0.000000000000e+00;
C[24170] = 0.000000000000e+00;
C[24171] = 0.000000000000e+00;
C[24172] = 0.000000000000e+00;
C[24173] = 0.000000000000e+00;
C[24174] = 0.000000000000e+00;
C[24175] = 0.000000000000e+00;
C[24176] = 0.000000000000e+00;
C[24177] = 0.000000000000e+00;
C[24178] = 0.000000000000e+00;
C[24179] = 0.000000000000e+00;
C[24180] = 0.000000000000e+00;
C[24181] = 0.000000000000e+00;
C[24182] = 0.000000000000e+00;
C[24183] = 0.000000000000e+00;
C[24184] = 0.000000000000e+00;
C[24185] = 0.000000000000e+00;
C[24186] = 0.000000000000e+00;
C[24187] = 0.000000000000e+00;
C[24188] = 1.000000000000e+00;
C[24189] = -0.000000000000e+00;
C[24190] = -0.000000000000e+00;
C[24191] = 2.000000000000e-01;
C[24192] = 7.073411938090e-02;
C[24193] = -1.401004495778e-02;
C[24194] = 1.454183685696e-03;
C[24195] = 3.754984888459e-01;
C[24196] = 1.335371137879e-02;
C[24197] = 5.601629169205e-03;
C[24198] = 5.071683617818e-02;
C[24199] = 1.912058469480e-03;
C[24200] = 7.428253717322e-04;
C[24201] = 1.308194811362e-01;
C[24202] = -1.213345329295e-03;
C[24203] = 2.082259447988e-04;
C[24204] = 3.009064519135e-02;
C[24205] = 1.134695119428e-03;
C[24206] = 4.733871277710e-04;
C[24207] = 2.425493230628e-03;
C[24208] = 9.537765493238e-05;
C[24209] = 3.649540232429e-05;
C[24210] = 2.366751205955e-03;
C[24211] = 9.814088416198e-05;
C[24212] = 2.975139913308e-05;
C[24213] = -2.347436587782e-05;
C[24214] = 1.099652121512e-07;
C[24215] = 2.329036332975e-07;
C[24216] = -1.386997065177e-02;
C[24217] = 5.075879150557e-02;
C[24218] = -5.794456857611e-03;
C[24219] = 1.326468220985e-02;
C[24220] = 3.959683304464e-01;
C[24221] = 8.242560043541e-03;
C[24222] = 1.835088887673e-03;
C[24223] = 5.627524431519e-02;
C[24224] = 3.374894335126e-03;
C[24225] = -1.204007042228e-03;
C[24226] = 1.285746481969e-01;
C[24227] = -7.850181120228e-04;
C[24228] = 1.127864382475e-03;
C[24229] = 3.246472599443e-02;
C[24230] = 1.004604153344e-03;
C[24231] = 9.118358245150e-05;
C[24232] = 2.743591819037e-03;
C[24233] = 1.905141059353e-04;
C[24234] = 8.972205092814e-05;
C[24235] = 2.812842006261e-03;
C[24236] = 2.947935773878e-04;
C[24237] = 9.513769933816e-09;
C[24238] = -1.472422151255e-05;
C[24239] = 4.602966205788e-06;
C[24240] = 1.551745520104e-03;
C[24241] = -5.796106551123e-03;
C[24242] = 4.540174785226e-02;
C[24243] = 5.592735480023e-03;
C[24244] = 8.303617799504e-03;
C[24245] = 3.669780949737e-01;
C[24246] = 6.618431237312e-04;
C[24247] = 3.235671346749e-03;
C[24248] = 5.229969379719e-02;
C[24249] = 2.147300670927e-04;
C[24250] = -7.851280915902e-04;
C[24251] = 1.287624327187e-01;
C[24252] = 4.719423606141e-04;
C[24253] = 1.007921977280e-03;
C[24254] = 2.977370529345e-02;
C[24255] = 3.201811104585e-05;
C[24256] = 1.823773198357e-04;
C[24257] = 2.533336437871e-03;
C[24258] = 2.020409486055e-05;
C[24259] = 2.746967420818e-04;
C[24260] = 2.619458607229e-03;
C[24261] = 1.134998817892e-07;
C[24262] = 4.211990302494e-06;
C[24263] = -1.790378496615e-05;
C[24264] = 3.758148324932e-01;
C[24265] = 1.350055231051e-02;
C[24266] = 5.213596686522e-03;
C[24267] = 1.174478049122e-01;
C[24268] = -1.279182734169e-02;
C[24269] = -3.345761908934e-03;
C[24270] = 3.914112304663e-01;
C[24271] = -2.548076627463e-03;
C[24272] = -1.488018911159e-03;
C[24273] = 3.011173476783e-02;
C[24274] = 1.143589055852e-03;
C[24275] = 4.466664410474e-04;
C[24276] = 1.328739167348e-01;
C[24277] = -1.075488220471e-03;
C[24278] = -3.178645542625e-04;
C[24279] = 3.201441722079e-02;
C[24280] = -1.552277152165e-04;
C[24281] = -1.143394053822e-04;
C[24282] = 6.105970633555e-02;
C[24283] = -3.702647709713e-04;
C[24284] = -5.301248726083e-04;
C[24285] = 2.670495096404e-03;
C[24286] = -1.229855585640e-05;
C[24287] = -2.270877477702e-05;
C[24288] = 1.329354813099e-02;
C[24289] = 3.959079542440e-01;
C[24290] = 8.275694210221e-03;
C[24291] = -1.263960689744e-02;
C[24292] = 9.130008556240e-02;
C[24293] = -1.677471431591e-02;
C[24294] = -2.461992202934e-03;
C[24295] = 3.914272634658e-01;
C[24296] = 4.564522572283e-03;
C[24297] = 1.130684236241e-03;
C[24298] = 3.246070091426e-02;
C[24299] = 1.006060404661e-03;
C[24300] = -1.065340190854e-03;
C[24301] = 1.300027574558e-01;
C[24302] = -1.736151390715e-03;
C[24303] = -1.506595128417e-04;
C[24304] = 3.220845290624e-02;
C[24305] = 5.218378075247e-04;
C[24306] = -3.686094587987e-04;
C[24307] = 6.200777874016e-02;
C[24308] = 2.041532760429e-03;
C[24309] = -1.254875836443e-05;
C[24310] = 2.726315970019e-03;
C[24311] = 1.007355219829e-04;
C[24312] = 5.093783834022e-03;
C[24313] = 8.183804083245e-03;
C[24314] = 3.668441531873e-01;
C[24315] = -3.265452491525e-03;
C[24316] = -1.674797535425e-02;
C[24317] = 1.095338190076e-01;
C[24318] = -1.398128089846e-03;
C[24319] = 4.734785181578e-03;
C[24320] = 4.007124819221e-01;
C[24321] = 4.395307720922e-04;
C[24322] = 1.000687089324e-03;
C[24323] = 2.976477584102e-02;
C[24324] = -3.125105931019e-04;
C[24325] = -1.734368793271e-03;
C[24326] = 1.319911083425e-01;
C[24327] = -1.095441582170e-04;
C[24328] = 5.317487641975e-04;
C[24329] = 3.287933570472e-02;
C[24330] = -5.274324914751e-04;
C[24331] = 2.041743033162e-03;
C[24332] = 6.466288475096e-02;
C[24333] = -2.293348381344e-05;
C[24334] = 1.001054201322e-04;
C[24335] = 2.836223072797e-03;
C[24336] = 5.209113346466e-02;
C[24337] = 2.137760383386e-03;
C[24338] = 6.391667978038e-04;
C[24339] = 3.903181738457e-01;
C[24340] = -2.728373384171e-03;
C[24341] = -1.346089399060e-03;
C[24342] = 4.084903431846e-02;
C[24343] = 1.412672142593e-03;
C[24344] = 2.954522309788e-03;
C[24345] = 2.517113049727e-03;
C[24346] = 1.113616821657e-04;
C[24347] = 3.050635598402e-05;
C[24348] = 3.194154677942e-02;
C[24349] = -1.684182582575e-04;
C[24350] = -1.060749121646e-04;
C[24351] = 1.278555746257e-01;
C[24352] = 1.397561248469e-04;
C[24353] = 2.716029942159e-04;
C[24354] = 5.143426011235e-01;
C[24355] = -9.470090752059e-04;
C[24356] = -2.257240861791e-03;
C[24357] = 3.770538718953e-02;
C[24358] = -8.329690671991e-05;
C[24359] = -1.979554477908e-04;
C[24360] = 2.195371827988e-03;
C[24361] = 5.640946362551e-02;
C[24362] = 2.629247850135e-03;
C[24363] = -2.796672427995e-03;
C[24364] = 3.915982159007e-01;
C[24365] = 5.268874052803e-03;
C[24366] = 1.415821714063e-03;
C[24367] = 3.377871311970e-02;
C[24368] = -1.258072999709e-02;
C[24369] = 1.142652121662e-04;
C[24370] = 2.752539773059e-03;
C[24371] = 1.419490867281e-04;
C[24372] = -1.718007685854e-04;
C[24373] = 3.221984973524e-02;
C[24374] = 5.673546889458e-04;
C[24375] = 1.399660962782e-04;
C[24376] = 1.272233822900e-01;
C[24377] = -1.106713445821e-03;
C[24378] = -9.411839729293e-04;
C[24379] = 5.155299943654e-01;
C[24380] = 4.525788456166e-03;
C[24381] = -8.302756016165e-05;
C[24382] = 3.781872768915e-02;
C[24383] = 3.948526084875e-04;
C[24384] = 7.087451147749e-04;
C[24385] = 2.739180355837e-03;
C[24386] = 5.276660979093e-02;
C[24387] = -1.428600552634e-03;
C[24388] = 5.148887322174e-03;
C[24389] = 4.001986143387e-01;
C[24390] = 2.960456853728e-03;
C[24391] = -1.258761995825e-02;
C[24392] = 1.898968453325e-02;
C[24393] = 3.422338519380e-05;
C[24394] = 1.481331739827e-04;
C[24395] = 2.564464170786e-03;
C[24396] = -1.103781814805e-04;
C[24397] = 5.607954575175e-04;
C[24398] = 3.284507786583e-02;
C[24399] = 2.719986304785e-04;
C[24400] = -1.107172776565e-03;
C[24401] = 1.259756778793e-01;
C[24402] = -2.252369335595e-03;
C[24403] = 4.545515359602e-03;
C[24404] = 5.253618690932e-01;
C[24405] = -1.977677068287e-04;
C[24406] = 3.958403811195e-04;
C[24407] = 3.862991695970e-02;
C[24408] = -7.026257182646e+00;
C[24409] = -3.168409879570e-02;
C[24410] = 3.287193442755e-02;
C[24411] = 2.092460106119e+00;
C[24412] = 2.753952100061e-02;
C[24413] = 3.127212031011e-02;
C[24414] = 7.797564281131e-01;
C[24415] = 2.537754337410e-02;
C[24416] = 1.139386519816e-02;
C[24417] = 7.723835783405e-02;
C[24418] = -1.232244344583e-02;
C[24419] = 4.934318116919e-03;
C[24420] = 3.659172032714e-01;
C[24421] = 1.115610466053e-02;
C[24422] = 5.781639628496e-03;
C[24423] = 5.023488940408e-02;
C[24424] = 1.781768640276e-03;
C[24425] = 7.353136023121e-04;
C[24426] = 7.803877532079e-02;
C[24427] = 2.869180849514e-03;
C[24428] = 8.304770753489e-04;
C[24429] = 2.414779315703e-03;
C[24430] = 1.063037361862e-04;
C[24431] = 2.752981342719e-05;
C[24432] = -2.888261267548e-02;
C[24433] = -7.069877728859e+00;
C[24434] = -6.119414569575e-03;
C[24435] = 2.538299344056e-02;
C[24436] = 2.145254151769e+00;
C[24437] = 1.411493161481e-02;
C[24438] = 2.441822661075e-02;
C[24439] = 8.663887452120e-01;
C[24440] = 5.108076864942e-02;
C[24441] = -1.218236913982e-02;
C[24442] = 4.940899050866e-02;
C[24443] = -1.031441625199e-02;
C[24444] = 1.106342836924e-02;
C[24445] = 3.963230074926e-01;
C[24446] = 1.277093314423e-02;
C[24447] = 1.717355033978e-03;
C[24448] = 5.640058151647e-02;
C[24449] = 3.628867700444e-03;
C[24450] = 2.679207941922e-03;
C[24451] = 8.913246014387e-02;
C[24452] = 7.535477650125e-03;
C[24453] = 9.975918316705e-05;
C[24454] = 2.826779653787e-03;
C[24455] = 2.738237598489e-04;
C[24456] = 3.482317111570e-02;
C[24457] = -6.152408439805e-03;
C[24458] = -7.173513385371e+00;
C[24459] = 3.070409510621e-02;
C[24460] = 1.487686552167e-02;
C[24461] = 2.018288265035e+00;
C[24462] = 1.039338389550e-02;
C[24463] = 4.913736192329e-02;
C[24464] = 8.027604964014e-01;
C[24465] = 5.031879951327e-03;
C[24466] = -1.031606594550e-02;
C[24467] = 5.161268083063e-02;
C[24468] = 5.766671808174e-03;
C[24469] = 1.283730491778e-02;
C[24470] = 3.630039146355e-01;
C[24471] = 6.638525815764e-04;
C[24472] = 3.485558012646e-03;
C[24473] = 5.204004005156e-02;
C[24474] = 6.047318443975e-04;
C[24475] = 7.052404904803e-03;
C[24476] = 8.379323998464e-02;
C[24477] = 1.927296348697e-05;
C[24478] = 2.555967898675e-04;
C[24479] = 2.630698772785e-03;
C[24480] = 2.098786979064e+00;
C[24481] = 3.010039545362e-02;
C[24482] = 2.312131923621e-02;
C[24483] = -6.217513103759e+00;
C[24484] = -2.965930248516e-02;
C[24485] = -6.645885506828e-03;
C[24486] = 2.170420777662e+00;
C[24487] = -2.622836278786e-02;
C[24488] = -8.484826668324e-03;
C[24489] = 3.662335469187e-01;
C[24490] = 1.129929846989e-02;
C[24491] = 5.387533014674e-03;
C[24492] = 1.255296562005e-01;
C[24493] = -1.064448375198e-02;
C[24494] = -3.427818404301e-03;
C[24495] = 3.932074802535e-01;
C[24496] = -2.387900462903e-03;
C[24497] = -1.344618188198e-03;
C[24498] = 1.032751575462e+00;
C[24499] = -7.388392872206e-03;
C[24500] = -9.079153764076e-03;
C[24501] = 6.107027098983e-02;
C[24502] = -3.878329828259e-04;
C[24503] = -5.539490370667e-04;
C[24504] = 2.633625604455e-02;
C[24505] = 2.144046627720e+00;
C[24506] = 1.431839373601e-02;
C[24507] = -2.661489360006e-02;
C[24508] = -6.290817218972e+00;
C[24509] = -4.458415509233e-02;
C[24510] = -2.510749564116e-02;
C[24511] = 2.101152048101e+00;
C[24512] = -3.202844224230e-02;
C[24513] = 1.109594141273e-02;
C[24514] = 3.962626312901e-01;
C[24515] = 1.280938132849e-02;
C[24516] = -1.049226330773e-02;
C[24517] = 8.936724895571e-02;
C[24518] = -2.091167751732e-02;
C[24519] = -2.313824823722e-03;
C[24520] = 3.928755820107e-01;
C[24521] = 3.931792345355e-03;
C[24522] = -7.305762930828e-03;
C[24523] = 1.044440086445e+00;
C[24524] = 3.034387611756e-02;
C[24525] = -3.891519762946e-04;
C[24526] = 6.205045470161e-02;
C[24527] = 2.041966026938e-03;
C[24528] = 2.111521360646e-02;
C[24529] = 1.293981240888e-02;
C[24530] = 2.015609429306e+00;
C[24531] = -5.039697158660e-03;
C[24532] = -4.404937585915e-02;
C[24533] = -6.246124141902e+00;
C[24534] = -7.342000885771e-03;
C[24535] = -2.956091582663e-02;
C[24536] = 2.160644958836e+00;
C[24537] = 5.273794293313e-03;
C[24538] = 1.271217718393e-02;
C[24539] = 3.628699728490e-01;
C[24540] = -3.347508986893e-03;
C[24541] = -2.088493855566e-02;
C[24542] = 1.138800717160e-01;
C[24543] = -1.264875569951e-03;
C[24544] = 4.107492004339e-03;
C[24545] = 4.007459402336e-01;
C[24546] = -8.959907115252e-03;
C[24547] = 3.049355686101e-02;
C[24548] = 1.088722813813e+00;
C[24549] = -5.535821939228e-04;
C[24550] = 2.037339033603e-03;
C[24551] = 6.462821955647e-02;
C[24552] = 8.072423738427e-01;
C[24553] = 3.047165652502e-02;
C[24554] = 9.939857376949e-03;
C[24555] = 2.148559645251e+00;
C[24556] = -3.043511926589e-02;
C[24557] = -6.301227070056e-03;
C[24558] = -7.143011497547e+00;
C[24559] = 2.933500424842e-03;
C[24560] = 1.212904921428e-02;
C[24561] = 5.160918669056e-02;
C[24562] = 2.020026529692e-03;
C[24563] = 6.411762556490e-04;
C[24564] = 3.921144236330e-01;
C[24565] = -2.580206004959e-03;
C[24566] = -1.212836879165e-03;
C[24567] = 4.006169169256e-02;
C[24568] = 1.579066906492e-03;
C[24569] = 3.288119036795e-03;
C[24570] = 4.111060684602e+00;
C[24571] = -6.153216288634e-03;
C[24572] = -1.583574332831e-02;
C[24573] = 5.139131959464e-01;
C[24574] = -1.121740149573e-03;
C[24575] = -2.731418475721e-03;
C[24576] = 3.104381054426e-02;
C[24577] = 8.690731314185e-01;
C[24578] = 3.700889199102e-02;
C[24579] = -3.120027879850e-02;
C[24580] = 2.104571096801e+00;
C[24581] = -1.887913840214e-02;
C[24582] = 2.996491854235e-03;
C[24583] = -7.167186121139e+00;
C[24584] = -4.106918660907e-02;
C[24585] = 2.065081998784e-03;
C[24586] = 5.653480082679e-02;
C[24587] = 2.879134516033e-03;
C[24588] = -2.636496263435e-03;
C[24589] = 3.930465344457e-01;
C[24590] = 4.641580875563e-03;
C[24591] = 1.582216477962e-03;
C[24592] = 3.287652460390e-02;
C[24593] = -1.229378134699e-02;
C[24594] = -6.058491135892e-03;
C[24595] = 4.108646979367e+00;
C[24596] = 1.822970304612e-02;
C[24597] = -1.114891964098e-03;
C[24598] = 5.148597691355e-01;
C[24599] = 4.568793669790e-03;
C[24600] = 1.071226005902e-02;
C[24601] = 3.836648906363e-02;
C[24602] = 8.120988162761e-01;
C[24603] = -7.296459497814e-03;
C[24604] = -2.034114724449e-02;
C[24605] = 2.150367607168e+00;
C[24606] = 1.224774009308e-02;
C[24607] = -4.120698583229e-02;
C[24608] = -7.217641465925e+00;
C[24609] = 7.012333453548e-04;
C[24610] = 2.993153721155e-03;
C[24611] = 5.250695604530e-02;
C[24612] = -1.285199829672e-03;
C[24613] = 4.516157095246e-03;
C[24614] = 4.002320726503e-01;
C[24615] = 3.294053580735e-03;
C[24616] = -1.230067130815e-02;
C[24617] = 1.934777532509e-02;
C[24618] = -1.576139974899e-02;
C[24619] = 1.856058676508e-02;
C[24620] = 4.170639925039e+00;
C[24621] = -2.725629895478e-03;
C[24622] = 4.591182509474e-03;
C[24623] = 5.252695375076e-01;
C[24624] = 0.000000000000e+00;
C[24625] = 0.000000000000e+00;
C[24626] = 0.000000000000e+00;
C[24627] = 0.000000000000e+00;
C[24628] = 0.000000000000e+00;
C[24629] = 0.000000000000e+00;
C[24630] = 0.000000000000e+00;
C[24631] = 0.000000000000e+00;
C[24632] = 0.000000000000e+00;
C[24633] = 0.000000000000e+00;
C[24634] = 0.000000000000e+00;
C[24635] = 0.000000000000e+00;
C[24636] = 0.000000000000e+00;
C[24637] = 0.000000000000e+00;
C[24638] = 0.000000000000e+00;
C[24639] = 0.000000000000e+00;
C[24640] = 0.000000000000e+00;
C[24641] = 0.000000000000e+00;
C[24642] = 1.000000000000e+00;
C[24643] = 0.000000000000e+00;
C[24644] = 0.000000000000e+00;
C[24645] = 2.000000000000e-01;
C[24646] = -0.000000000000e+00;
C[24647] = -0.000000000000e+00;
C[24648] = 0.000000000000e+00;
C[24649] = 0.000000000000e+00;
C[24650] = 0.000000000000e+00;
C[24651] = 0.000000000000e+00;
C[24652] = 0.000000000000e+00;
C[24653] = 0.000000000000e+00;
C[24654] = 0.000000000000e+00;
C[24655] = 0.000000000000e+00;
C[24656] = 0.000000000000e+00;
C[24657] = 0.000000000000e+00;
C[24658] = 0.000000000000e+00;
C[24659] = 0.000000000000e+00;
C[24660] = 0.000000000000e+00;
C[24661] = 0.000000000000e+00;
C[24662] = 0.000000000000e+00;
C[24663] = 0.000000000000e+00;
C[24664] = 0.000000000000e+00;
C[24665] = 0.000000000000e+00;
C[24666] = 0.000000000000e+00;
C[24667] = 1.000000000000e+00;
C[24668] = 0.000000000000e+00;
C[24669] = -0.000000000000e+00;
C[24670] = 2.000000000000e-01;
C[24671] = -0.000000000000e+00;
C[24672] = 0.000000000000e+00;
C[24673] = 0.000000000000e+00;
C[24674] = 0.000000000000e+00;
C[24675] = 0.000000000000e+00;
C[24676] = 0.000000000000e+00;
C[24677] = 0.000000000000e+00;
C[24678] = 0.000000000000e+00;
C[24679] = 0.000000000000e+00;
C[24680] = 0.000000000000e+00;
C[24681] = 0.000000000000e+00;
C[24682] = 0.000000000000e+00;
C[24683] = 0.000000000000e+00;
C[24684] = 0.000000000000e+00;
C[24685] = 0.000000000000e+00;
C[24686] = 0.000000000000e+00;
C[24687] = 0.000000000000e+00;
C[24688] = 0.000000000000e+00;
C[24689] = 0.000000000000e+00;
C[24690] = 0.000000000000e+00;
C[24691] = 0.000000000000e+00;
C[24692] = 1.000000000000e+00;
C[24693] = -0.000000000000e+00;
C[24694] = -0.000000000000e+00;
C[24695] = 2.000000000000e-01;
C[24696] = 8.342026803583e-02;
C[24697] = -6.452727937887e-03;
C[24698] = 1.220650286890e-02;
C[24699] = 3.575947967352e-01;
C[24700] = 1.436234647275e-03;
C[24701] = 2.696908339344e-03;
C[24702] = 4.958747469907e-02;
C[24703] = -3.281077710185e-04;
C[24704] = -8.937933410966e-05;
C[24705] = 1.315788126863e-01;
C[24706] = -5.199302697930e-04;
C[24707] = 1.075447905215e-03;
C[24708] = 2.911422312503e-02;
C[24709] = -4.564999698187e-05;
C[24710] = 9.544800426254e-05;
C[24711] = 2.398994607449e-03;
C[24712] = -2.737213430911e-05;
C[24713] = -1.429176575788e-05;
C[24714] = 2.398213475579e-03;
C[24715] = -6.096759509101e-05;
C[24716] = -5.574226120376e-05;
C[24717] = -1.674612709337e-05;
C[24718] = -2.454626314062e-06;
C[24719] = -2.357899545820e-06;
C[24720] = -6.329141112828e-03;
C[24721] = 4.670806042795e-02;
C[24722] = -9.170801986705e-03;
C[24723] = 1.346084618401e-03;
C[24724] = 3.987017093195e-01;
C[24725] = 1.617143246454e-02;
C[24726] = -3.580891211838e-04;
C[24727] = 5.714423843636e-02;
C[24728] = 3.794463465532e-03;
C[24729] = -5.116911481224e-04;
C[24730] = 1.281306143233e-01;
C[24731] = -6.089633663446e-04;
C[24732] = -5.147434421011e-05;
C[24733] = 3.278459071933e-02;
C[24734] = 1.376505857952e-03;
C[24735] = -2.866966819082e-05;
C[24736] = 2.788509760085e-03;
C[24737] = 1.943649406062e-04;
C[24738] = -6.408437816433e-05;
C[24739] = 2.854697905234e-03;
C[24740] = 2.630845807701e-04;
C[24741] = -2.419525700100e-06;
C[24742] = -1.445453866476e-05;
C[24743] = 6.585087942259e-07;
C[24744] = 1.226493524426e-02;
C[24745] = -9.271222374078e-03;
C[24746] = 6.156570465312e-02;
C[24747] = 2.668440555987e-03;
C[24748] = 1.647732222695e-02;
C[24749] = 3.607820162625e-01;
C[24750] = -1.003143153685e-04;
C[24751] = 3.655887863295e-03;
C[24752] = 5.188731976663e-02;
C[24753] = 1.079343396905e-03;
C[24754] = -6.156580588361e-04;
C[24755] = 1.300456376332e-01;
C[24756] = 9.289703276583e-05;
C[24757] = 1.395799150131e-03;
C[24758] = 2.947421109143e-02;
C[24759] = -1.438196575741e-05;
C[24760] = 1.857989333201e-04;
C[24761] = 2.526419992336e-03;
C[24762] = -5.475445181873e-05;
C[24763] = 2.406790373709e-04;
C[24764] = 2.640425714949e-03;
C[24765] = -2.197052293962e-06;
C[24766] = 3.089119969924e-07;
C[24767] = -1.502612850032e-05;
C[24768] = 3.576731881996e-01;
C[24769] = 2.118623082202e-03;
C[24770] = 2.818085137056e-03;
C[24771] = 1.345905534084e-01;
C[24772] = -1.520811291997e-03;
C[24773] = 3.055156066531e-04;
C[24774] = 3.934941612419e-01;
C[24775] = -4.945915811701e-04;
C[24776] = -6.550805128386e-04;
C[24777] = 2.911944922266e-02;
C[24778] = 2.822004329882e-08;
C[24779] = 1.028733381705e-04;
C[24780] = 1.339857577821e-01;
C[24781] = 9.094837737158e-05;
C[24782] = 1.398260587852e-04;
C[24783] = 3.192729118217e-02;
C[24784] = -7.299711959505e-05;
C[24785] = -1.210994633637e-04;
C[24786] = 6.038741096855e-02;
C[24787] = -3.558925262555e-04;
C[24788] = -7.296174159705e-04;
C[24789] = 2.622098452343e-03;
C[24790] = -1.793875878355e-05;
C[24791] = -3.683212482615e-05;
C[24792] = 1.936687456355e-03;
C[24793] = 3.992423666816e-01;
C[24794] = 1.570187517530e-02;
C[24795] = -1.332258583063e-03;
C[24796] = 8.730923568038e-02;
C[24797] = -2.378755925013e-02;
C[24798] = -4.558683174572e-04;
C[24799] = 3.919413162496e-01;
C[24800] = 4.084823526599e-03;
C[24801] = -1.228647637658e-05;
C[24802] = 3.282063454347e-02;
C[24803] = 1.344102680021e-03;
C[24804] = 1.035185579672e-04;
C[24805] = 1.297884659281e-01;
C[24806] = -2.112203117778e-03;
C[24807] = -7.147542880260e-05;
C[24808] = 3.206407867505e-02;
C[24809] = 5.721150590984e-04;
C[24810] = -3.604121420264e-04;
C[24811] = 6.128586177669e-02;
C[24812] = 2.286825690387e-03;
C[24813] = -1.829368107854e-05;
C[24814] = 2.681089092416e-03;
C[24815] = 1.181934881000e-04;
C[24816] = 2.738526006528e-03;
C[24817] = 1.559291132792e-02;
C[24818] = 3.598562676569e-01;
C[24819] = 3.949482976014e-04;
C[24820] = -2.379554571109e-02;
C[24821] = 1.180249014735e-01;
C[24822] = -6.437483529268e-04;
C[24823] = 4.257980883475e-03;
C[24824] = 4.007895771556e-01;
C[24825] = 9.822251540809e-05;
C[24826] = 1.337937782177e-03;
C[24827] = 2.941249451772e-02;
C[24828] = 1.457882381818e-04;
C[24829] = -2.112735548508e-03;
C[24830] = 1.324295260842e-01;
C[24831] = -1.212702105171e-04;
C[24832] = 5.818759870159e-04;
C[24833] = 3.284929024495e-02;
C[24834] = -7.366581295962e-04;
C[24835] = 2.283587128906e-03;
C[24836] = 6.453631113050e-02;
C[24837] = -3.716256671023e-05;
C[24838] = 1.172904407021e-04;
C[24839] = 2.828122794260e-03;
C[24840] = 4.867558782803e-02;
C[24841] = 1.861653453888e-04;
C[24842] = -2.325420667034e-05;
C[24843] = 3.940850813046e-01;
C[24844] = -1.034644958041e-03;
C[24845] = -9.320261199438e-04;
C[24846] = 4.157951549077e-02;
C[24847] = 1.799151047851e-03;
C[24848] = 3.515771405566e-03;
C[24849] = 2.338202149380e-03;
C[24850] = 7.613962914022e-06;
C[24851] = -9.244625177531e-06;
C[24852] = 3.196668585301e-02;
C[24853] = -1.100605381748e-04;
C[24854] = -1.404887283182e-04;
C[24855] = 1.281020129800e-01;
C[24856] = 1.764873605609e-04;
C[24857] = 3.291208139720e-04;
C[24858] = 5.134120139305e-01;
C[24859] = -9.597359240165e-04;
C[24860] = -2.460792047634e-03;
C[24861] = 3.761216441131e-02;
C[24862] = -7.119727034340e-05;
C[24863] = -1.785850191510e-04;
C[24864] = 2.341330943228e-04;
C[24865] = 5.659318383698e-02;
C[24866] = 3.992961180809e-03;
C[24867] = -1.093959672732e-03;
C[24868] = 3.921466780314e-01;
C[24869] = 3.946350453845e-03;
C[24870] = 1.804654398827e-03;
C[24871] = 3.406303512177e-02;
C[24872] = -1.283243255877e-02;
C[24873] = 1.011059004698e-05;
C[24874] = 2.751772786794e-03;
C[24875] = 2.082704878211e-04;
C[24876] = -1.129549923659e-04;
C[24877] = 3.207776946050e-02;
C[24878] = 5.611006250406e-04;
C[24879] = 1.768542506259e-04;
C[24880] = 1.274444231581e-01;
C[24881] = -1.192884137374e-03;
C[24882] = -9.505982691206e-04;
C[24883] = 5.145333893000e-01;
C[24884] = 4.669586553289e-03;
C[24885] = -7.081069531639e-05;
C[24886] = 3.774290778081e-02;
C[24887] = 4.180879038300e-04;
C[24888] = 7.523414514511e-06;
C[24889] = 4.186667737557e-03;
C[24890] = 5.137550965269e-02;
C[24891] = -9.738765637966e-04;
C[24892] = 3.710494639868e-03;
C[24893] = 4.011432506400e-01;
C[24894] = 3.520877232029e-03;
C[24895] = -1.282057251011e-02;
C[24896] = 1.959954846874e-02;
C[24897] = -7.831582516269e-06;
C[24898] = 2.205118920746e-04;
C[24899] = 2.492299318073e-03;
C[24900] = -1.423525334276e-04;
C[24901] = 5.471597999830e-04;
C[24902] = 3.287286847724e-02;
C[24903] = 3.294612024029e-04;
C[24904] = -1.192093467464e-03;
C[24905] = 1.260165464145e-01;
C[24906] = -2.451582593566e-03;
C[24907] = 4.694973135050e-03;
C[24908] = 5.253122379163e-01;
C[24909] = -1.782132539946e-04;
C[24910] = 4.192267962920e-04;
C[24911] = 3.863060369023e-02;
C[24912] = -7.015613817150e+00;
C[24913] = -3.659813961773e-02;
C[24914] = 7.521350659191e-02;
C[24915] = 2.065143518905e+00;
C[24916] = -1.548035464986e-02;
C[24917] = 2.895881049683e-02;
C[24918] = 7.742905974362e-01;
C[24919] = -8.887365676391e-03;
C[24920] = -1.283644315698e-03;
C[24921] = 8.044839787972e-02;
C[24922] = -6.344557824136e-03;
C[24923] = 1.433616646664e-02;
C[24924] = 3.623920308472e-01;
C[24925] = -3.214440627210e-03;
C[24926] = 1.156511668090e-03;
C[24927] = 4.970782832198e-02;
C[24928] = -6.244123384759e-04;
C[24929] = -2.497313525531e-04;
C[24930] = 7.674589614854e-02;
C[24931] = -1.288692213104e-03;
C[24932] = -1.220021097707e-03;
C[24933] = 2.356347167020e-03;
C[24934] = -4.676439053666e-05;
C[24935] = -5.907847862408e-05;
C[24936] = -3.412640311654e-02;
C[24937] = -7.084389144464e+00;
C[24938] = 2.005026114895e-02;
C[24939] = -1.725966427422e-02;
C[24940] = 2.147576095173e+00;
C[24941] = 3.831838328766e-02;
C[24942] = -9.099653921189e-03;
C[24943] = 8.791821065070e-01;
C[24944] = 5.199222341870e-02;
C[24945] = -6.220970999076e-03;
C[24946] = 4.361161640764e-02;
C[24947] = -1.705254726048e-03;
C[24948] = -3.268702966284e-03;
C[24949] = 4.003420638659e-01;
C[24950] = 1.372040320147e-02;
C[24951] = -6.405296818543e-04;
C[24952] = 5.705441233269e-02;
C[24953] = 3.689718331848e-03;
C[24954] = -1.380955488619e-03;
C[24955] = 8.958333408166e-02;
C[24956] = 7.705526982078e-03;
C[24957] = -5.016589141357e-05;
C[24958] = 2.820706101907e-03;
C[24959] = 2.782571013530e-04;
C[24960] = 7.638215409905e-02;
C[24961] = 1.804185340149e-02;
C[24962] = -7.102294594112e+00;
C[24963] = 2.835112704240e-02;
C[24964] = 4.378852709003e-02;
C[24965] = 2.037007606930e+00;
C[24966] = -1.281304346913e-03;
C[24967] = 5.040998966504e-02;
C[24968] = 8.036349241520e-01;
C[24969] = 1.439459884200e-02;
C[24970] = -1.805675113421e-03;
C[24971] = 6.586074173598e-02;
C[24972] = 1.129456762513e-03;
C[24973] = 1.409941356444e-02;
C[24974] = 3.641941298262e-01;
C[24975] = -2.470585866745e-04;
C[24976] = 3.533114553282e-03;
C[24977] = 5.204613831890e-02;
C[24978] = -1.222005664486e-03;
C[24979] = 7.088702911947e-03;
C[24980] = 8.369383495081e-02;
C[24981] = -5.859641742981e-05;
C[24982] = 2.517578986702e-04;
C[24983] = 2.621098364530e-03;
C[24984] = 2.066711348193e+00;
C[24985] = -1.808894998200e-03;
C[24986] = 3.134401866379e-02;
C[24987] = -6.183959488616e+00;
C[24988] = -4.157448275065e-04;
C[24989] = 5.515727916154e-03;
C[24990] = 2.178506760985e+00;
C[24991] = 5.855298737885e-03;
C[24992] = 4.925181382397e-03;
C[24993] = 3.624704223116e-01;
C[24994] = -2.496164502483e-03;
C[24995] = 1.279101343582e-03;
C[24996] = 1.328606363261e-01;
C[24997] = 3.374913420573e-03;
C[24998] = 2.888531660444e-03;
C[24999] = 3.904824198214e-01;
C[25000] = -7.747869253926e-04;
C[25001] = -1.544828679262e-03;
C[25002] = 1.020054027672e+00;
C[25003] = -5.964640616156e-03;
C[25004] = -1.220348771884e-02;
C[25005] = 6.019653375536e-02;
C[25006] = -3.873414069698e-04;
C[25007] = -7.580090959186e-04;
C[25008] = -5.471298468269e-03;
C[25009] = 2.158389242415e+00;
C[25010] = 2.827958605703e-02;
C[25011] = 3.355309351173e-03;
C[25012] = -6.295572049928e+00;
C[25013] = -5.414295033280e-02;
C[25014] = 6.178828082459e-03;
C[25015] = 2.092113578303e+00;
C[25016] = -3.079230556734e-02;
C[25017] = -2.713987818131e-03;
C[25018] = 4.008827212280e-01;
C[25019] = 1.332396651279e-02;
C[25020] = 3.563466129507e-03;
C[25021] = 8.875527955285e-02;
C[25022] = -2.206869528127e-02;
C[25023] = -7.546439332922e-04;
C[25024] = 3.888829497322e-01;
C[25025] = 5.179880988635e-03;
C[25026] = -5.974861313405e-03;
C[25027] = 1.029183949328e+00;
C[25028] = 3.484920483680e-02;
C[25029] = -3.892196034033e-04;
C[25030] = 6.109521248036e-02;
C[25031] = 2.346858718084e-03;
C[25032] = 2.979116384050e-02;
C[25033] = 2.674796055530e-02;
C[25034] = 2.018492634818e+00;
C[25035] = 7.304381735120e-03;
C[25036] = -5.430267955186e-02;
C[25037] = -6.229268325832e+00;
C[25038] = 4.854967212435e-03;
C[25039] = -2.862504199175e-02;
C[25040] = 2.161492101045e+00;
C[25041] = 1.198129335273e-03;
C[25042] = 1.314188206485e-02;
C[25043] = 3.632683812206e-01;
C[25044] = 2.977964351393e-03;
C[25045] = -2.207668174222e-02;
C[25046] = 1.152810651557e-01;
C[25047] = -1.554083965591e-03;
C[25048] = 5.367325914183e-03;
C[25049] = 4.004135923078e-01;
C[25050] = -1.227061675884e-02;
C[25051] = 3.508808030049e-02;
C[25052] = 1.087305511493e+00;
C[25053] = -7.601554712342e-04;
C[25054] = 2.348078752533e-03;
C[25055] = 6.454245408074e-02;
C[25056] = 7.560528600155e-01;
C[25057] = 1.785435410264e-03;
C[25058] = 2.598978270515e-04;
C[25059] = 2.190325162239e+00;
C[25060] = -5.396704729215e-03;
C[25061] = -9.105881279043e-04;
C[25062] = -7.130284287288e+00;
C[25063] = 6.571436659465e-03;
C[25064] = 1.414528277820e-02;
C[25065] = 4.879594145095e-02;
C[25066] = -9.627521528165e-05;
C[25067] = -1.699984779762e-04;
C[25068] = 3.910733398840e-01;
C[25069] = -1.333420573876e-03;
C[25070] = -1.842361732608e-03;
C[25071] = 4.510845828993e-02;
C[25072] = 2.170526390989e-03;
C[25073] = 4.011365667909e-03;
C[25074] = 4.109956610931e+00;
C[25075] = -3.014036327757e-03;
C[25076] = -1.050239405959e-02;
C[25077] = 5.127200912760e-01;
C[25078] = -7.060572960455e-04;
C[25079] = -1.886976084132e-03;
C[25080] = 2.357451630436e-03;
C[25081] = 8.681610145195e-01;
C[25082] = 5.715145601532e-02;
C[25083] = -6.132063093359e-03;
C[25084] = 2.096220813940e+00;
C[25085] = -3.485765058436e-02;
C[25086] = 6.681503678992e-03;
C[25087] = -7.159215479754e+00;
C[25088] = -4.533462350897e-02;
C[25089] = -6.217147313451e-05;
C[25090] = 5.650335773332e-02;
C[25091] = 3.870187870796e-03;
C[25092] = -1.374155016955e-03;
C[25093] = 3.890883115141e-01;
C[25094] = 5.055695484553e-03;
C[25095] = 2.176029741966e-03;
C[25096] = 3.754439605602e-02;
C[25097] = -1.404360507309e-02;
C[25098] = -2.885069682977e-03;
C[25099] = 4.109843424735e+00;
C[25100] = 1.781895352834e-02;
C[25101] = -7.000039582524e-04;
C[25102] = 5.141775021670e-01;
C[25103] = 4.916217423667e-03;
C[25104] = 6.544106567858e-04;
C[25105] = 5.983630885920e-02;
C[25106] = 7.933987218731e-01;
C[25107] = -1.450739636763e-03;
C[25108] = -3.827888330196e-02;
C[25109] = 2.168565570732e+00;
C[25110] = 1.424739930746e-02;
C[25111] = -4.509742253574e-02;
C[25112] = -7.215606992393e+00;
C[25113] = -1.528286039289e-04;
C[25114] = 4.081922603873e-03;
C[25115] = 5.153432820496e-02;
C[25116] = -1.863624730220e-03;
C[25117] = 4.805552101904e-03;
C[25118] = 4.007672657922e-01;
C[25119] = 4.016471494372e-03;
C[25120] = -1.403174502443e-02;
C[25121] = 1.969091366115e-02;
C[25122] = -1.037810447974e-02;
C[25123] = 1.820941604282e-02;
C[25124] = 4.171646021177e+00;
C[25125] = -1.882649606516e-03;
C[25126] = 4.940260023033e-03;
C[25127] = 5.253977470827e-01;
C[25128] = 0.000000000000e+00;
C[25129] = 0.000000000000e+00;
C[25130] = 0.000000000000e+00;
C[25131] = 0.000000000000e+00;
C[25132] = 0.000000000000e+00;
C[25133] = 0.000000000000e+00;
C[25134] = 0.000000000000e+00;
C[25135] = 0.000000000000e+00;
C[25136] = 0.000000000000e+00;
C[25137] = 0.000000000000e+00;
C[25138] = 0.000000000000e+00;
C[25139] = 0.000000000000e+00;
C[25140] = 0.000000000000e+00;
C[25141] = 0.000000000000e+00;
C[25142] = 0.000000000000e+00;
C[25143] = 0.000000000000e+00;
C[25144] = 0.000000000000e+00;
C[25145] = 0.000000000000e+00;
C[25146] = 1.000000000000e+00;
C[25147] = 0.000000000000e+00;
C[25148] = 0.000000000000e+00;
C[25149] = 2.000000000000e-01;
C[25150] = -0.000000000000e+00;
C[25151] = -0.000000000000e+00;
C[25152] = 0.000000000000e+00;
C[25153] = 0.000000000000e+00;
C[25154] = 0.000000000000e+00;
C[25155] = 0.000000000000e+00;
C[25156] = 0.000000000000e+00;
C[25157] = 0.000000000000e+00;
C[25158] = 0.000000000000e+00;
C[25159] = 0.000000000000e+00;
C[25160] = 0.000000000000e+00;
C[25161] = 0.000000000000e+00;
C[25162] = 0.000000000000e+00;
C[25163] = 0.000000000000e+00;
C[25164] = 0.000000000000e+00;
C[25165] = 0.000000000000e+00;
C[25166] = 0.000000000000e+00;
C[25167] = 0.000000000000e+00;
C[25168] = 0.000000000000e+00;
C[25169] = 0.000000000000e+00;
C[25170] = 0.000000000000e+00;
C[25171] = 1.000000000000e+00;
C[25172] = 0.000000000000e+00;
C[25173] = -0.000000000000e+00;
C[25174] = 2.000000000000e-01;
C[25175] = -0.000000000000e+00;
C[25176] = 0.000000000000e+00;
C[25177] = 0.000000000000e+00;
C[25178] = 0.000000000000e+00;
C[25179] = 0.000000000000e+00;
C[25180] = 0.000000000000e+00;
C[25181] = 0.000000000000e+00;
C[25182] = 0.000000000000e+00;
C[25183] = 0.000000000000e+00;
C[25184] = 0.000000000000e+00;
C[25185] = 0.000000000000e+00;
C[25186] = 0.000000000000e+00;
C[25187] = 0.000000000000e+00;
C[25188] = 0.000000000000e+00;
C[25189] = 0.000000000000e+00;
C[25190] = 0.000000000000e+00;
C[25191] = 0.000000000000e+00;
C[25192] = 0.000000000000e+00;
C[25193] = 0.000000000000e+00;
C[25194] = 0.000000000000e+00;
C[25195] = 0.000000000000e+00;
C[25196] = 1.000000000000e+00;
C[25197] = -0.000000000000e+00;
C[25198] = -0.000000000000e+00;
C[25199] = 2.000000000000e-01;
c[0] = 2.164410317640e-11;
c[1] = 1.598277066250e-12;
c[2] = -1.136174487826e-11;
c[3] = -5.734168695426e-11;
c[4] = 5.952316417535e-11;
c[5] = 2.404344778828e-11;
c[6] = 3.876202137043e-11;
c[7] = -5.985856255109e-11;
c[8] = -4.257261210228e-12;
c[9] = 4.969877842598e-10;
c[10] = 1.393610782330e-10;
c[11] = -3.579241347751e-10;
c[12] = -1.532882154542e-09;
c[13] = 1.521573256280e-09;
c[14] = 6.777897132437e-10;
c[15] = 9.940421819010e-10;
c[16] = -1.498506985165e-09;
c[17] = -3.830669115246e-11;
c[18] = 0.000000000000e+00;
c[19] = 0.000000000000e+00;
c[20] = 0.000000000000e+00;
c[21] = -2.464446149930e-08;
c[22] = -5.253341317513e-09;
c[23] = 1.506400273965e-08;
c[24] = 2.678360613850e-08;
c[25] = 1.056672083166e-09;
c[26] = -1.552886165490e-08;
c[27] = 6.681035724654e-09;
c[28] = -9.762013419845e-10;
c[29] = -4.275606202420e-09;
c[30] = -3.367169785773e-07;
c[31] = -1.120741495164e-07;
c[32] = 1.986067952042e-07;
c[33] = 3.743663989919e-07;
c[34] = 2.777647845065e-08;
c[35] = -1.972176297826e-07;
c[36] = 1.466662511440e-07;
c[37] = -1.439624508492e-08;
c[38] = -8.508972282328e-08;
c[39] = 0.000000000000e+00;
c[40] = 0.000000000000e+00;
c[41] = 0.000000000000e+00;
c[42] = -1.387136963626e-08;
c[43] = -3.199053799641e-10;
c[44] = 1.597836835066e-08;
c[45] = 1.485251860389e-08;
c[46] = -1.304949759362e-09;
c[47] = -1.680257288861e-08;
c[48] = -4.262023511892e-10;
c[49] = 1.886378053761e-09;
c[50] = -3.649235358338e-10;
c[51] = -3.533937225075e-07;
c[52] = 3.798737391314e-08;
c[53] = 3.613979065453e-07;
c[54] = 3.339964433913e-07;
c[55] = 9.955669399986e-09;
c[56] = -3.812041000639e-07;
c[57] = -3.225397349205e-08;
c[58] = 5.067889585764e-08;
c[59] = 3.072498921952e-08;
c[60] = 0.000000000000e+00;
c[61] = 0.000000000000e+00;
c[62] = 0.000000000000e+00;
c[63] = -6.466548410217e-07;
c[64] = 1.023403491035e-06;
c[65] = 5.813965145984e-07;
c[66] = 5.200257409133e-07;
c[67] = -8.260760050449e-07;
c[68] = -6.200102734000e-07;
c[69] = 1.736821038101e-07;
c[70] = -9.284681057320e-08;
c[71] = -1.329077366496e-07;
c[72] = 3.039760811951e-07;
c[73] = 1.099533865578e-05;
c[74] = 1.611077305230e-06;
c[75] = -1.024730769683e-06;
c[76] = -9.191514648510e-06;
c[77] = -2.798567806162e-06;
c[78] = 1.989676933656e-06;
c[79] = -1.560384221166e-06;
c[80] = -1.917554589059e-06;
c[81] = 0.000000000000e+00;
c[82] = 0.000000000000e+00;
c[83] = 0.000000000000e+00;
c[84] = -4.762114753643e-07;
c[85] = 2.633779891126e-07;
c[86] = 3.506103742890e-07;
c[87] = 4.479922735279e-07;
c[88] = -2.729419034675e-07;
c[89] = -2.481857694359e-07;
c[90] = 1.385173550861e-08;
c[91] = -6.980074296337e-10;
c[92] = 3.700308837118e-09;
c[93] = -7.741953077456e-06;
c[94] = 5.025534858039e-06;
c[95] = 8.241436230905e-06;
c[96] = 7.316956227243e-06;
c[97] = -5.451129726386e-06;
c[98] = -5.282745800361e-06;
c[99] = 5.792113138559e-07;
c[100] = -2.508123728706e-07;
c[101] = -3.842885647476e-07;
c[102] = 0.000000000000e+00;
c[103] = 0.000000000000e+00;
c[104] = 0.000000000000e+00;
c[105] = 3.638514797788e-08;
c[106] = -1.330182997084e-08;
c[107] = -1.699131968458e-08;
c[108] = -3.928562736233e-08;
c[109] = 2.488053252314e-08;
c[110] = 2.426909018194e-08;
c[111] = -1.009812650166e-10;
c[112] = -3.465004727587e-09;
c[113] = 1.262792564960e-09;
c[114] = 1.039646746914e-06;
c[115] = -5.230795685485e-07;
c[116] = -4.020353492340e-07;
c[117] = -1.095265152684e-06;
c[118] = 8.198305684326e-07;
c[119] = 5.577783140920e-07;
c[120] = -6.318149892159e-08;
c[121] = -2.895043543782e-08;
c[122] = 5.527165952701e-08;
c[123] = 0.000000000000e+00;
c[124] = 0.000000000000e+00;
c[125] = 0.000000000000e+00;
c[126] = -1.222346389107e-07;
c[127] = 2.161322903982e-08;
c[128] = 1.979078310121e-07;
c[129] = 8.185234512403e-08;
c[130] = -1.053718271526e-07;
c[131] = -7.882627933720e-08;
c[132] = -1.700668593452e-08;
c[133] = 3.304118867398e-09;
c[134] = 7.761364051451e-09;
c[135] = -4.302713023652e-06;
c[136] = 1.788263557345e-06;
c[137] = 6.489935739484e-06;
c[138] = 2.408812767141e-06;
c[139] = -2.091857271402e-06;
c[140] = -2.719721825750e-06;
c[141] = -2.395575044445e-07;
c[142] = -2.635151102881e-08;
c[143] = -2.682423555500e-08;
c[144] = 0.000000000000e+00;
c[145] = 0.000000000000e+00;
c[146] = 0.000000000000e+00;
c[147] = -2.747609120779e-08;
c[148] = 2.509261809447e-08;
c[149] = 6.508079273410e-08;
c[150] = 1.504414218201e-08;
c[151] = 1.696483908908e-08;
c[152] = -5.705190900263e-08;
c[153] = 3.437598757039e-08;
c[154] = -4.877366333034e-08;
c[155] = -2.028105033913e-08;
c[156] = -8.739284649284e-07;
c[157] = 9.441716619651e-07;
c[158] = 1.438068026771e-06;
c[159] = 1.863134011515e-07;
c[160] = 5.074649810366e-07;
c[161] = -1.181793120364e-06;
c[162] = 8.451768553908e-07;
c[163] = -1.036411959476e-06;
c[164] = -4.790437604929e-07;
c[165] = 0.000000000000e+00;
c[166] = 0.000000000000e+00;
c[167] = 0.000000000000e+00;
c[168] = -8.092981010344e-08;
c[169] = 2.337953064480e-08;
c[170] = 3.482067380267e-08;
c[171] = 1.220457984674e-07;
c[172] = -5.093275223800e-08;
c[173] = -6.048873663067e-08;
c[174] = -3.540112233913e-08;
c[175] = 2.602558513143e-08;
c[176] = 2.198472065507e-08;
c[177] = -1.580348391927e-06;
c[178] = 4.497050993257e-07;
c[179] = 6.227649677060e-07;
c[180] = 2.355370239737e-06;
c[181] = -8.729872471580e-07;
c[182] = -1.044624214686e-06;
c[183] = -4.335867442862e-07;
c[184] = 2.530683481106e-07;
c[185] = 1.979597133994e-07;
c[186] = 0.000000000000e+00;
c[187] = 0.000000000000e+00;
c[188] = 0.000000000000e+00;
c[189] = 5.046755513494e-09;
c[190] = -8.727058964375e-09;
c[191] = 8.193733191941e-10;
c[192] = -3.161469241952e-08;
c[193] = 2.539378085586e-08;
c[194] = 1.498279394996e-08;
c[195] = 3.145447302622e-08;
c[196] = -2.358348272224e-08;
c[197] = -1.867405891032e-08;
c[198] = 1.426471454291e-07;
c[199] = -1.805877039018e-07;
c[200] = -4.665654307345e-09;
c[201] = -5.662197346545e-07;
c[202] = 4.473113324188e-07;
c[203] = 2.640301014489e-07;
c[204] = 6.234061670218e-07;
c[205] = -4.761952978605e-07;
c[206] = -3.535606292182e-07;
c[207] = 0.000000000000e+00;
c[208] = 0.000000000000e+00;
c[209] = 0.000000000000e+00;
c[210] = -5.662436766696e-08;
c[211] = 6.186284648435e-08;
c[212] = 1.531636983909e-08;
c[213] = 5.404937325482e-08;
c[214] = -4.213543119969e-08;
c[215] = -2.942387872240e-08;
c[216] = -6.768513000077e-09;
c[217] = 3.567477202537e-09;
c[218] = 1.518715153637e-10;
c[219] = -1.194499839152e-06;
c[220] = 1.215050598580e-06;
c[221] = 2.738053613838e-07;
c[222] = 1.149242871201e-06;
c[223] = -7.853830663596e-07;
c[224] = -5.405627800936e-07;
c[225] = -1.110349507893e-07;
c[226] = 5.672773661924e-08;
c[227] = -7.204407181938e-09;
c[228] = 0.000000000000e+00;
c[229] = 0.000000000000e+00;
c[230] = 0.000000000000e+00;
c[231] = 1.325924731121e-08;
c[232] = -1.816273148547e-08;
c[233] = -7.609351482252e-09;
c[234] = -1.593047954662e-08;
c[235] = 2.144718536101e-08;
c[236] = 1.552419476303e-08;
c[237] = -2.656673742107e-09;
c[238] = -1.695281248715e-09;
c[239] = -2.160775780524e-09;
c[240] = 3.180211655129e-07;
c[241] = -5.992509927721e-07;
c[242] = -2.133999723597e-07;
c[243] = -4.525996690763e-07;
c[244] = 6.401107724430e-07;
c[245] = 4.743456004097e-07;
c[246] = -4.582589596680e-08;
c[247] = 5.477558406852e-10;
c[248] = -2.545352462135e-08;
c[249] = 0.000000000000e+00;
c[250] = 0.000000000000e+00;
c[251] = 0.000000000000e+00;
c[252] = 2.328326947909e-08;
c[253] = -2.963412017165e-08;
c[254] = -1.963307705299e-08;
c[255] = -3.170439843991e-08;
c[256] = 3.466067127755e-08;
c[257] = 2.754873673627e-08;
c[258] = 1.165084556609e-09;
c[259] = 2.005913657399e-09;
c[260] = 7.225338660710e-10;
c[261] = 2.764052442483e-07;
c[262] = -2.833358642151e-07;
c[263] = -2.365428607431e-07;
c[264] = -4.760701725059e-07;
c[265] = 4.033431570960e-07;
c[266] = 2.776067395960e-07;
c[267] = 1.323645308915e-07;
c[268] = -3.946362936436e-08;
c[269] = 1.517732839407e-08;
c[270] = 0.000000000000e+00;
c[271] = 0.000000000000e+00;
c[272] = 0.000000000000e+00;
c[273] = -1.127520752875e-07;
c[274] = 4.302367898057e-08;
c[275] = 6.131644691987e-08;
c[276] = 1.297707193648e-07;
c[277] = -6.184391010411e-08;
c[278] = -1.219542502262e-07;
c[279] = 1.301852509128e-08;
c[280] = 9.861683025747e-09;
c[281] = 1.928007714591e-08;
c[282] = -3.849770216280e-06;
c[283] = 2.716631543942e-06;
c[284] = 2.908303410765e-06;
c[285] = 4.541887768550e-06;
c[286] = -3.299840446447e-06;
c[287] = -4.354757459524e-06;
c[288] = 3.941246731198e-07;
c[289] = -1.364554991223e-08;
c[290] = 1.171599545713e-07;
c[291] = 0.000000000000e+00;
c[292] = 0.000000000000e+00;
c[293] = 0.000000000000e+00;
c[294] = 7.602612900337e-09;
c[295] = -2.469288723495e-08;
c[296] = 5.984076845156e-10;
c[297] = -3.299666678780e-08;
c[298] = 1.932782645042e-08;
c[299] = 2.333613979832e-08;
c[300] = 5.239469756668e-08;
c[301] = -3.077333809998e-08;
c[302] = -4.167694023227e-08;
c[303] = 1.760333715461e-07;
c[304] = -9.247438045312e-07;
c[305] = 6.005324282299e-07;
c[306] = -1.358863927714e-06;
c[307] = 4.249967174852e-07;
c[308] = 6.799637413124e-07;
c[309] = 1.437361486611e-06;
c[310] = -6.763358503248e-07;
c[311] = -9.102969172892e-07;
c[312] = 0.000000000000e+00;
c[313] = 0.000000000000e+00;
c[314] = 0.000000000000e+00;
c[315] = -7.317752792613e-09;
c[316] = 1.159668155948e-08;
c[317] = 4.910228584842e-09;
c[318] = 8.073561374999e-09;
c[319] = -7.871802765180e-09;
c[320] = -7.480970912188e-09;
c[321] = -3.092715805675e-09;
c[322] = 4.826475663577e-09;
c[323] = 2.598832038458e-09;
c[324] = -2.040238459289e-07;
c[325] = 6.833950161322e-08;
c[326] = 2.056255135052e-07;
c[327] = 1.889431368030e-07;
c[328] = -7.057119477683e-08;
c[329] = -1.821380151701e-07;
c[330] = 7.329476714446e-08;
c[331] = 3.486073968384e-08;
c[332] = -1.344731648256e-08;
c[333] = 0.000000000000e+00;
c[334] = 0.000000000000e+00;
c[335] = 0.000000000000e+00;
c[336] = 1.596204449172e-08;
c[337] = -6.845119138177e-09;
c[338] = -2.333084503370e-09;
c[339] = -2.044254979183e-08;
c[340] = 3.144995597282e-09;
c[341] = 4.558328658977e-09;
c[342] = 1.998621407262e-09;
c[343] = -2.658499087538e-09;
c[344] = -4.671728337513e-09;
c[345] = 3.303967782697e-07;
c[346] = -1.491732297687e-07;
c[347] = -4.613814752474e-08;
c[348] = -4.590702673557e-07;
c[349] = 6.601265178929e-08;
c[350] = 9.397584421755e-08;
c[351] = 4.353413374236e-08;
c[352] = -7.199622309439e-08;
c[353] = -1.099350412925e-07;
c[354] = 0.000000000000e+00;
c[355] = 0.000000000000e+00;
c[356] = 0.000000000000e+00;
c[357] = 1.379571046289e-08;
c[358] = -1.639645880180e-08;
c[359] = 6.786491202337e-09;
c[360] = -3.389738406634e-09;
c[361] = -2.213111804927e-09;
c[362] = -4.970354128275e-09;
c[363] = -1.507957425595e-09;
c[364] = 6.320741707810e-10;
c[365] = -4.116967600165e-10;
c[366] = 2.697500308679e-07;
c[367] = -1.862591700874e-07;
c[368] = 7.598366036365e-08;
c[369] = -3.242862312014e-08;
c[370] = -9.205471540596e-08;
c[371] = -9.627504971021e-08;
c[372] = -5.033639216556e-08;
c[373] = 2.473690674520e-08;
c[374] = 1.525037451877e-08;
c[375] = 0.000000000000e+00;
c[376] = 0.000000000000e+00;
c[377] = 0.000000000000e+00;
c[378] = -7.022160960768e-07;
c[379] = 6.165111125223e-07;
c[380] = 7.340906147402e-07;
c[381] = 7.080761027789e-07;
c[382] = -7.136735251656e-07;
c[383] = -6.430497230814e-07;
c[384] = -2.491359321621e-08;
c[385] = -2.540194299350e-08;
c[386] = 1.776829600120e-08;
c[387] = -1.864730159096e-05;
c[388] = 1.225703658553e-05;
c[389] = 1.583616478232e-05;
c[390] = 1.860255362118e-05;
c[391] = -1.250420623777e-05;
c[392] = -1.648193346837e-05;
c[393] = 9.320356659237e-07;
c[394] = -1.217313696678e-06;
c[395] = -4.420441358510e-07;
c[396] = 0.000000000000e+00;
c[397] = 0.000000000000e+00;
c[398] = 0.000000000000e+00;
c[399] = 7.588044746987e-07;
c[400] = -1.398422757828e-06;
c[401] = 5.255023009521e-07;
c[402] = -7.842804027602e-07;
c[403] = 1.414174246150e-06;
c[404] = -3.548775461271e-07;
c[405] = -2.330196667844e-07;
c[406] = 1.321106284369e-07;
c[407] = 1.184887127026e-07;
c[408] = -5.660126993462e-06;
c[409] = -4.144985814719e-06;
c[410] = 2.328648888211e-05;
c[411] = 4.910291625304e-06;
c[412] = 4.300449059524e-06;
c[413] = -2.101832169865e-05;
c[414] = -2.783667738759e-06;
c[415] = 1.703459031610e-06;
c[416] = 6.661485976167e-07;
c[417] = 0.000000000000e+00;
c[418] = 0.000000000000e+00;
c[419] = 0.000000000000e+00;
c[420] = -8.646260601930e-08;
c[421] = 2.856302349363e-07;
c[422] = 2.176717069924e-07;
c[423] = 9.341033463395e-08;
c[424] = -3.101365393388e-07;
c[425] = -2.269419200618e-07;
c[426] = 1.345534848818e-08;
c[427] = 2.605717608350e-08;
c[428] = -1.669778670887e-08;
c[429] = -1.484185950029e-06;
c[430] = 8.836212409991e-06;
c[431] = 4.555315968657e-06;
c[432] = 1.524961540600e-06;
c[433] = -9.354558903163e-06;
c[434] = -4.688343026360e-06;
c[435] = 4.185370279242e-07;
c[436] = -5.668309110973e-08;
c[437] = -5.936319186217e-07;
c[438] = 0.000000000000e+00;
c[439] = 0.000000000000e+00;
c[440] = 0.000000000000e+00;
c[441] = 1.606657495934e-08;
c[442] = -2.413620858732e-08;
c[443] = 2.053445186112e-08;
c[444] = -1.092387225121e-08;
c[445] = -1.013517647586e-08;
c[446] = -9.879148499259e-09;
c[447] = 1.845541691692e-09;
c[448] = -3.762073985669e-10;
c[449] = -2.794146192198e-09;
c[450] = 6.489495878448e-07;
c[451] = -8.933777813080e-07;
c[452] = 4.629692027525e-07;
c[453] = -6.755584716078e-07;
c[454] = 4.757142202383e-08;
c[455] = 2.507447083300e-08;
c[456] = -1.278100925095e-08;
c[457] = 3.717224528188e-09;
c[458] = -3.400945880472e-08;
c[459] = 0.000000000000e+00;
c[460] = 0.000000000000e+00;
c[461] = 0.000000000000e+00;
c[462] = 1.654113096494e-08;
c[463] = -1.858626907936e-08;
c[464] = -9.832143293975e-09;
c[465] = -1.012338513018e-08;
c[466] = 2.218128847264e-08;
c[467] = 1.605130597881e-09;
c[468] = 7.495483012043e-09;
c[469] = -1.963632323410e-08;
c[470] = -1.189336351626e-09;
c[471] = 2.628973154728e-07;
c[472] = -1.193481603545e-07;
c[473] = -1.711824347828e-07;
c[474] = -2.862511618851e-07;
c[475] = 5.809890183395e-07;
c[476] = 2.062897519473e-08;
c[477] = 2.504595344455e-07;
c[478] = -5.967756462888e-07;
c[479] = -3.303692402978e-08;
c[480] = 0.000000000000e+00;
c[481] = 0.000000000000e+00;
c[482] = 0.000000000000e+00;
c[483] = -4.741319864499e-09;
c[484] = -2.091482487199e-09;
c[485] = 3.187020966577e-09;
c[486] = -6.950773956405e-09;
c[487] = 8.970154619092e-10;
c[488] = 1.647908667857e-08;
c[489] = 7.904199988973e-09;
c[490] = 1.151827633272e-09;
c[491] = -1.712407199372e-08;
c[492] = -7.438316718122e-08;
c[493] = -3.259858161186e-09;
c[494] = 5.000066072292e-08;
c[495] = 1.174711179441e-07;
c[496] = 5.861136909147e-09;
c[497] = 2.023771774606e-07;
c[498] = -8.328489109477e-08;
c[499] = 6.843485528263e-09;
c[500] = -2.043949191288e-07;
c[501] = 0.000000000000e+00;
c[502] = 0.000000000000e+00;
c[503] = 0.000000000000e+00;
c[504] = 1.657047482562e-08;
c[505] = -1.732073992566e-08;
c[506] = 3.923085134527e-09;
c[507] = -2.528558817882e-08;
c[508] = 4.021998262216e-08;
c[509] = 8.876436918825e-09;
c[510] = 2.247622586660e-08;
c[511] = -4.163365363574e-08;
c[512] = -9.467695072374e-09;
c[513] = 4.442178865860e-07;
c[514] = -4.070722660066e-07;
c[515] = -1.299016123557e-07;
c[516] = -9.103837697033e-07;
c[517] = 1.264927045996e-06;
c[518] = 2.277862338751e-07;
c[519] = 8.950936695451e-07;
c[520] = -1.353478946475e-06;
c[521] = -3.161695446563e-07;
c[522] = 0.000000000000e+00;
c[523] = 0.000000000000e+00;
c[524] = 0.000000000000e+00;
c[525] = -9.006859119642e-09;
c[526] = 9.613974463751e-09;
c[527] = 3.432587547536e-12;
c[528] = 7.444869026818e-09;
c[529] = -7.129017931540e-09;
c[530] = 2.312707914065e-09;
c[531] = 1.666395341160e-09;
c[532] = -1.409516281470e-09;
c[533] = -3.444575846778e-09;
c[534] = -3.127656903468e-07;
c[535] = 3.897322155000e-07;
c[536] = -1.093427213217e-07;
c[537] = 2.869355523183e-07;
c[538] = -1.315074936725e-07;
c[539] = -1.626895773654e-08;
c[540] = -3.422640992845e-08;
c[541] = -1.507303715176e-08;
c[542] = -4.547739562710e-08;
c[543] = 0.000000000000e+00;
c[544] = 0.000000000000e+00;
c[545] = 0.000000000000e+00;
c[546] = 1.201144855800e-08;
c[547] = -1.469872315529e-08;
c[548] = -2.075360738818e-08;
c[549] = -1.169797975153e-08;
c[550] = 1.255363318853e-08;
c[551] = 1.695811324698e-08;
c[552] = 2.065625615000e-09;
c[553] = -1.138823590985e-09;
c[554] = -2.406249366427e-10;
c[555] = 2.255081044122e-07;
c[556] = -3.967920226833e-07;
c[557] = -4.391199285303e-07;
c[558] = -3.080207480188e-07;
c[559] = 3.761221215592e-07;
c[560] = 4.150588349683e-07;
c[561] = 4.276905152256e-08;
c[562] = -5.490163879074e-11;
c[563] = 1.517193948253e-08;
c[564] = 0.000000000000e+00;
c[565] = 0.000000000000e+00;
c[566] = 0.000000000000e+00;
c[567] = -4.788480875706e-08;
c[568] = 6.000798827288e-08;
c[569] = 6.042148248397e-08;
c[570] = 4.846668183456e-08;
c[571] = -4.876106829421e-08;
c[572] = -4.990806221072e-08;
c[573] = 1.350741185568e-10;
c[574] = -7.333611495852e-09;
c[575] = -4.404708903172e-09;
c[576] = -2.414585444921e-07;
c[577] = 3.454928073632e-07;
c[578] = 3.463874150289e-07;
c[579] = 2.109179081389e-07;
c[580] = -2.023046512534e-07;
c[581] = -1.975884531547e-07;
c[582] = 4.954403021173e-08;
c[583] = -1.248788401575e-07;
c[584] = -9.405220979808e-08;
c[585] = 0.000000000000e+00;
c[586] = 0.000000000000e+00;
c[587] = 0.000000000000e+00;
c[588] = -1.936440836570e-07;
c[589] = 1.175293731226e-07;
c[590] = 3.546960015977e-07;
c[591] = 1.925286259585e-07;
c[592] = -6.343523206720e-08;
c[593] = -3.917634301964e-07;
c[594] = -4.174150183833e-08;
c[595] = 4.010394694109e-08;
c[596] = 7.314974825867e-08;
c[597] = -7.020467416630e-06;
c[598] = 6.950782316162e-06;
c[599] = 1.268515871822e-05;
c[600] = 8.447291306446e-06;
c[601] = -6.071216796411e-06;
c[602] = -1.534915547396e-05;
c[603] = -3.911738801765e-07;
c[604] = 3.253311933182e-07;
c[605] = 2.890915342668e-07;
c[606] = 0.000000000000e+00;
c[607] = 0.000000000000e+00;
c[608] = 0.000000000000e+00;
c[609] = -5.322929969021e-08;
c[610] = 4.574749462583e-08;
c[611] = 2.005701837948e-08;
c[612] = 2.939851559414e-08;
c[613] = -1.623518497418e-08;
c[614] = -1.434168868386e-08;
c[615] = 9.599543410177e-09;
c[616] = -4.294519406756e-09;
c[617] = -2.416378486192e-09;
c[618] = -6.027003480735e-07;
c[619] = 5.733519632400e-07;
c[620] = 2.824736095031e-07;
c[621] = 1.472117652224e-07;
c[622] = -2.287284373015e-07;
c[623] = -1.463548984226e-07;
c[624] = 2.853083945764e-07;
c[625] = -8.381581428907e-08;
c[626] = -8.380143745601e-08;
c[627] = 0.000000000000e+00;
c[628] = 0.000000000000e+00;
c[629] = 0.000000000000e+00;
c[630] = 3.429828573931e-08;
c[631] = 8.249486649348e-09;
c[632] = -1.325143800246e-09;
c[633] = -2.314197863895e-08;
c[634] = -3.039754670198e-09;
c[635] = -7.731896373153e-10;
c[636] = -4.261244412884e-09;
c[637] = -2.920045094612e-09;
c[638] = -4.743156145626e-09;
c[639] = 7.488393420907e-07;
c[640] = 2.982214279967e-07;
c[641] = 1.224767280927e-07;
c[642] = -5.360319674796e-07;
c[643] = -2.328398659435e-07;
c[644] = -1.452226161547e-07;
c[645] = -1.928221560199e-07;
c[646] = -9.030713493674e-08;
c[647] = -1.441168565042e-07;
c[648] = 0.000000000000e+00;
c[649] = 0.000000000000e+00;
c[650] = 0.000000000000e+00;
c[651] = -1.364537275395e-07;
c[652] = 8.133039042235e-08;
c[653] = 5.248686768544e-08;
c[654] = 1.091661848418e-07;
c[655] = -5.889573806428e-08;
c[656] = -2.608172916529e-08;
c[657] = 2.294002268699e-08;
c[658] = -7.269569834989e-09;
c[659] = -8.418452268621e-09;
c[660] = -1.914814361514e-06;
c[661] = 1.625499484792e-06;
c[662] = 9.331601034113e-07;
c[663] = 1.347891059433e-06;
c[664] = -1.001212386220e-06;
c[665] = -3.870965976649e-07;
c[666] = 3.920273130009e-07;
c[667] = -2.039982184288e-07;
c[668] = -1.865298306925e-07;
c[669] = 0.000000000000e+00;
c[670] = 0.000000000000e+00;
c[671] = 0.000000000000e+00;
c[672] = 3.002819498610e-08;
c[673] = -8.247528504590e-08;
c[674] = 4.214217852461e-08;
c[675] = 1.839371482903e-08;
c[676] = 5.348363418811e-09;
c[677] = -8.629111813985e-09;
c[678] = 4.634481826082e-10;
c[679] = 2.077747862650e-09;
c[680] = 6.643612882051e-10;
c[681] = 6.914992189966e-07;
c[682] = -1.655517752441e-06;
c[683] = 5.280959294307e-07;
c[684] = 3.275407529646e-07;
c[685] = -2.683211264287e-08;
c[686] = -2.663785825252e-08;
c[687] = 3.295639500189e-08;
c[688] = -2.222564671195e-08;
c[689] = -7.267215607065e-09;
c[690] = 0.000000000000e+00;
c[691] = 0.000000000000e+00;
c[692] = 0.000000000000e+00;
c[693] = 9.363616101932e-08;
c[694] = 1.231947555080e-08;
c[695] = -5.948822601187e-08;
c[696] = -1.008663098323e-07;
c[697] = -3.218855404441e-09;
c[698] = 4.194344560560e-08;
c[699] = 7.640350824234e-09;
c[700] = 1.549748218643e-09;
c[701] = 2.722961467416e-09;
c[702] = 2.656580729177e-06;
c[703] = 4.642632637220e-07;
c[704] = -8.133583828673e-07;
c[705] = -2.824037497129e-06;
c[706] = -5.110674641262e-07;
c[707] = 6.603664520766e-07;
c[708] = -1.923690370886e-07;
c[709] = -3.750192206775e-08;
c[710] = 1.413672888473e-07;
c[711] = 0.000000000000e+00;
c[712] = 0.000000000000e+00;
c[713] = 0.000000000000e+00;
c[714] = -8.725586466141e-07;
c[715] = 6.594910917090e-07;
c[716] = 3.693970986873e-07;
c[717] = 9.152613260066e-07;
c[718] = -5.890429818356e-07;
c[719] = -3.286745217934e-07;
c[720] = 1.487623653984e-07;
c[721] = -3.233440515338e-08;
c[722] = -4.406764289588e-08;
c[723] = -4.049268873096e-06;
c[724] = 9.845364528638e-06;
c[725] = 4.664647592745e-06;
c[726] = 4.163065282170e-06;
c[727] = -9.094926096687e-06;
c[728] = -3.795463129563e-06;
c[729] = 2.267054212801e-06;
c[730] = -8.266732727247e-07;
c[731] = -6.700343633259e-07;
c[732] = 0.000000000000e+00;
c[733] = 0.000000000000e+00;
c[734] = 0.000000000000e+00;
c[735] = -2.304765042704e-08;
c[736] = 1.441266178626e-07;
c[737] = 2.762166033121e-07;
c[738] = 8.100389886634e-09;
c[739] = -1.409432341948e-07;
c[740] = -3.293824734074e-07;
c[741] = 3.101255928450e-08;
c[742] = -1.024942353212e-10;
c[743] = 2.668300824737e-08;
c[744] = -5.160967583295e-07;
c[745] = 2.718436086568e-06;
c[746] = 7.940096630410e-06;
c[747] = 3.262959888950e-07;
c[748] = -2.743991580512e-06;
c[749] = -9.677668790342e-06;
c[750] = 7.160074155799e-07;
c[751] = -1.192770284764e-07;
c[752] = -3.783092061838e-08;
c[753] = 0.000000000000e+00;
c[754] = 0.000000000000e+00;
c[755] = 0.000000000000e+00;
c[756] = -1.806040886287e-09;
c[757] = 1.826845552610e-08;
c[758] = 1.018721443313e-08;
c[759] = 7.937251411683e-10;
c[760] = -1.966445284385e-08;
c[761] = -3.260782421854e-09;
c[762] = -1.213769917907e-08;
c[763] = 2.543469768135e-09;
c[764] = 4.957054211552e-10;
c[765] = -1.529016180335e-07;
c[766] = 1.059605740206e-06;
c[767] = 2.389293737304e-07;
c[768] = 1.786750810640e-07;
c[769] = -8.154596362608e-07;
c[770] = -2.694894906119e-07;
c[771] = -3.002556711040e-07;
c[772] = -3.691184979893e-08;
c[773] = -6.382900308299e-08;
c[774] = 0.000000000000e+00;
c[775] = 0.000000000000e+00;
c[776] = 0.000000000000e+00;
c[777] = 6.664111568133e-09;
c[778] = -3.728393871416e-09;
c[779] = -9.669507416987e-10;
c[780] = -3.786815583329e-09;
c[781] = 4.284534060872e-09;
c[782] = 4.233298434020e-10;
c[783] = -9.124923872061e-10;
c[784] = -2.100146168083e-09;
c[785] = 1.154133899561e-09;
c[786] = 1.217853908553e-07;
c[787] = -5.259539503921e-08;
c[788] = -2.484960520688e-08;
c[789] = -3.391264358221e-08;
c[790] = 6.879499503398e-08;
c[791] = 1.275626437813e-08;
c[792] = -3.002977222444e-08;
c[793] = -4.865154279443e-08;
c[794] = -2.219142913873e-08;
c[795] = 0.000000000000e+00;
c[796] = 0.000000000000e+00;
c[797] = 0.000000000000e+00;
c[798] = -6.927155848935e-09;
c[799] = 3.319568397941e-09;
c[800] = 5.727309543291e-09;
c[801] = 8.843192456087e-09;
c[802] = 2.873762339206e-09;
c[803] = -5.854891904633e-09;
c[804] = -2.495061296459e-09;
c[805] = -2.763949513707e-09;
c[806] = 1.309901576096e-09;
c[807] = -1.678560778151e-07;
c[808] = 5.326111507387e-08;
c[809] = 1.532933178172e-07;
c[810] = 2.051235172829e-07;
c[811] = -3.680038940246e-09;
c[812] = -1.543064227505e-07;
c[813] = -4.520160645782e-08;
c[814] = 3.216047483878e-09;
c[815] = 9.206724382871e-09;
c[816] = 0.000000000000e+00;
c[817] = 0.000000000000e+00;
c[818] = 0.000000000000e+00;
c[819] = 5.888947871013e-09;
c[820] = -2.074130805996e-08;
c[821] = 2.216879235739e-09;
c[822] = 1.600011229064e-08;
c[823] = 3.501660528205e-08;
c[824] = -1.752584299908e-09;
c[825] = -1.260776444356e-08;
c[826] = -2.607201299298e-08;
c[827] = 9.973483122705e-09;
c[828] = 1.100751086502e-06;
c[829] = -6.423622649354e-07;
c[830] = -4.611776908137e-07;
c[831] = -2.579833664473e-07;
c[832] = 1.045401645361e-06;
c[833] = 4.115494788293e-07;
c[834] = -5.326465675193e-07;
c[835] = -8.734745902128e-07;
c[836] = 1.079560851824e-07;
c[837] = 0.000000000000e+00;
c[838] = 0.000000000000e+00;
c[839] = 0.000000000000e+00;
c[840] = 1.025582154757e-07;
c[841] = 1.300590191100e-10;
c[842] = -2.068033988500e-08;
c[843] = -8.423125580359e-08;
c[844] = -2.060700554640e-09;
c[845] = 5.062828500879e-08;
c[846] = -1.631114882050e-09;
c[847] = -1.707008756568e-09;
c[848] = -7.949264385632e-09;
c[849] = 2.158045405487e-06;
c[850] = 3.133382516785e-07;
c[851] = -6.156746650277e-07;
c[852] = -1.798599614133e-06;
c[853] = -5.000460457927e-07;
c[854] = 7.899885825502e-07;
c[855] = 1.327931248296e-07;
c[856] = -6.524292939236e-08;
c[857] = -2.343630734480e-07;
c[858] = 0.000000000000e+00;
c[859] = 0.000000000000e+00;
c[860] = 0.000000000000e+00;
c[861] = -3.384330954370e-09;
c[862] = -3.018808023003e-09;
c[863] = -8.244674942759e-09;
c[864] = 5.731473989856e-09;
c[865] = -9.245604282171e-11;
c[866] = 8.994285205510e-09;
c[867] = 1.465544213763e-09;
c[868] = -6.791275319884e-10;
c[869] = 2.099102003328e-09;
c[870] = 2.528734066587e-08;
c[871] = -5.047265272395e-09;
c[872] = 2.475386651213e-08;
c[873] = 4.329164249306e-08;
c[874] = 2.747329969077e-08;
c[875] = 1.318436237963e-08;
c[876] = -1.330133267419e-08;
c[877] = -2.576070767812e-08;
c[878] = 1.898556423185e-08;
c[879] = 0.000000000000e+00;
c[880] = 0.000000000000e+00;
c[881] = 0.000000000000e+00;
c[882] = 8.016749653006e-08;
c[883] = 3.138376086786e-08;
c[884] = 1.152537868476e-07;
c[885] = -9.586530441696e-08;
c[886] = -2.094967571642e-08;
c[887] = -8.052382591273e-08;
c[888] = 1.076948680234e-08;
c[889] = -1.420359085902e-08;
c[890] = -3.584171071047e-09;
c[891] = 2.538660376028e-06;
c[892] = 8.112494465085e-07;
c[893] = 4.126478955091e-06;
c[894] = -2.843187855484e-06;
c[895] = -1.665304543153e-07;
c[896] = -2.771336287788e-06;
c[897] = 1.553683987021e-07;
c[898] = -3.132816489559e-07;
c[899] = -2.329255456290e-07;
c[900] = 0.000000000000e+00;
c[901] = 0.000000000000e+00;
c[902] = 0.000000000000e+00;
c[903] = -3.123978429831e-07;
c[904] = 1.499793272908e-08;
c[905] = 5.728098784186e-08;
c[906] = 2.644903885907e-07;
c[907] = 5.667292113376e-08;
c[908] = -3.849738733352e-08;
c[909] = 1.089472512206e-09;
c[910] = 8.399957063254e-09;
c[911] = 2.801054943546e-09;
c[912] = -5.737519796600e-06;
c[913] = 1.501429440887e-06;
c[914] = 9.837259373757e-07;
c[915] = 5.114392106709e-06;
c[916] = 5.875086042106e-07;
c[917] = -1.124025148624e-06;
c[918] = 4.040513592596e-07;
c[919] = 5.002988823222e-08;
c[920] = -2.869735602484e-07;
c[921] = 0.000000000000e+00;
c[922] = 0.000000000000e+00;
c[923] = 0.000000000000e+00;
c[924] = 2.359200976243e-08;
c[925] = -8.681227903207e-09;
c[926] = 6.338789593219e-08;
c[927] = -1.254723483446e-08;
c[928] = -1.332364107931e-09;
c[929] = -5.967071507740e-08;
c[930] = -1.197157500910e-08;
c[931] = 2.969627876936e-09;
c[932] = 2.069854176945e-09;
c[933] = -6.575411291010e-07;
c[934] = 7.747606181607e-08;
c[935] = 9.139478004822e-07;
c[936] = 5.971338546962e-07;
c[937] = -1.057413423489e-08;
c[938] = -8.779322770991e-07;
c[939] = -1.331176005537e-07;
c[940] = 2.524278808202e-08;
c[941] = 1.012365680175e-08;
c[942] = 0.000000000000e+00;
c[943] = 0.000000000000e+00;
c[944] = 0.000000000000e+00;
c[945] = 2.455136943746e-08;
c[946] = 1.488705209107e-08;
c[947] = 1.829910029061e-08;
c[948] = -2.009945754278e-08;
c[949] = -1.520486858642e-08;
c[950] = -1.538310598348e-08;
c[951] = -1.072527039403e-09;
c[952] = -1.472481470088e-09;
c[953] = -4.654476248867e-10;
c[954] = 6.660447325890e-07;
c[955] = 4.574124804479e-07;
c[956] = 4.105014491107e-07;
c[957] = -6.094339199958e-07;
c[958] = -4.547805241284e-07;
c[959] = -3.776384187582e-07;
c[960] = -5.146256032429e-08;
c[961] = -8.663376593887e-08;
c[962] = -1.444379793547e-08;
c[963] = 0.000000000000e+00;
c[964] = 0.000000000000e+00;
c[965] = 0.000000000000e+00;
c[966] = 1.519887966872e-08;
c[967] = -1.200809010560e-08;
c[968] = 1.523685333460e-08;
c[969] = -9.524677324624e-09;
c[970] = 5.413474446492e-09;
c[971] = -5.454602713950e-09;
c[972] = -5.328165131324e-09;
c[973] = -2.774751539647e-09;
c[974] = -1.021940893020e-08;
c[975] = 7.498990647892e-08;
c[976] = -7.782845261595e-07;
c[977] = -9.580988697966e-07;
c[978] = -7.517584510452e-08;
c[979] = 5.593174842167e-07;
c[980] = 1.157097981763e-06;
c[981] = -5.601988151138e-08;
c[982] = 4.335692582735e-09;
c[983] = -3.865104114542e-08;
c[984] = 0.000000000000e+00;
c[985] = 0.000000000000e+00;
c[986] = 0.000000000000e+00;
c[987] = 3.041531062009e-07;
c[988] = 2.334470915466e-07;
c[989] = 5.264108227099e-08;
c[990] = -2.954713711201e-07;
c[991] = -2.000855391926e-07;
c[992] = -5.028161104148e-08;
c[993] = 2.422677050640e-08;
c[994] = 1.966259599584e-08;
c[995] = -1.806305618968e-08;
c[996] = 7.393245754495e-06;
c[997] = 6.227923604940e-06;
c[998] = 1.265676111639e-06;
c[999] = -7.168743991870e-06;
c[1000] = -6.281182978629e-06;
c[1001] = -1.005252342423e-06;
c[1002] = 9.332983874621e-08;
c[1003] = -8.630037939827e-08;
c[1004] = -5.179524146248e-07;
c[1005] = 0.000000000000e+00;
c[1006] = 0.000000000000e+00;
c[1007] = 0.000000000000e+00;
c[1008] = -7.358574344307e-08;
c[1009] = -5.626652137269e-08;
c[1010] = -3.518907049083e-08;
c[1011] = 7.175080998789e-08;
c[1012] = 4.182540214170e-08;
c[1013] = 3.834109968404e-08;
c[1014] = -1.674589356071e-09;
c[1015] = 7.437158666690e-10;
c[1016] = 1.008518049628e-08;
c[1017] = 2.881494276519e-08;
c[1018] = -2.458287695628e-07;
c[1019] = -4.375412028246e-07;
c[1020] = -8.342865953637e-09;
c[1021] = 1.821439621352e-07;
c[1022] = 3.422473369152e-07;
c[1023] = 1.004458181075e-07;
c[1024] = -1.464978610644e-08;
c[1025] = 1.321047346514e-07;
c[1026] = 0.000000000000e+00;
c[1027] = 0.000000000000e+00;
c[1028] = 0.000000000000e+00;
c[1029] = 7.218527184327e-08;
c[1030] = 6.429460330803e-08;
c[1031] = 5.727166407787e-08;
c[1032] = -8.935170425772e-08;
c[1033] = -7.785814787331e-08;
c[1034] = -2.744376648156e-08;
c[1035] = 2.983162494807e-09;
c[1036] = 1.357785994749e-09;
c[1037] = 1.986376119145e-09;
c[1038] = 1.021548654112e-06;
c[1039] = 1.097174741727e-06;
c[1040] = 1.211785279343e-06;
c[1041] = -1.307242264548e-06;
c[1042] = -1.364140252758e-06;
c[1043] = -5.739580326569e-07;
c[1044] = 5.461648178162e-08;
c[1045] = -6.021247667132e-09;
c[1046] = 1.005580474445e-08;
c[1047] = 0.000000000000e+00;
c[1048] = 0.000000000000e+00;
c[1049] = 0.000000000000e+00;
dLow[0] = -9.999999999999e+11;
dLow[1] = -3.959849010350e-01;
dLow[2] = -9.999999999999e+11;
dLow[3] = -9.999999999999e+11;
dLow[4] = -6.386149249081e-01;
dLow[5] = -1.000000000000e+12;
dLow[6] = -9.999999999999e+11;
dLow[7] = -8.376744723341e-01;
dLow[8] = -1.000000000000e+12;
dLow[9] = -9.999999999995e+11;
dLow[10] = -1.000000000001e+12;
dLow[11] = -1.000000000000e+12;
dLow[12] = -9.999999999993e+11;
dLow[13] = -1.000000000001e+12;
dLow[14] = -1.000000000001e+12;
dLow[15] = -9.999999999990e+11;
dLow[16] = -1.000000000001e+12;
dLow[17] = -1.000000000001e+12;
dLow[18] = -1.000000000000e+12;
dLow[19] = -1.010000000000e+00;
dLow[20] = -1.000000000001e+12;
dLow[21] = -1.000000000000e+00;
dLow[22] = -1.000000000000e+00;
dLow[23] = -1.000000000000e+00;
dLow[24] = -9.999999999999e+11;
dLow[25] = -4.121858641525e-01;
dLow[26] = -9.999999999999e+11;
dLow[27] = -9.999999999998e+11;
dLow[28] = -7.502624587364e-01;
dLow[29] = -1.000000000000e+12;
dLow[30] = -9.999999999998e+11;
dLow[31] = -9.770690459981e-01;
dLow[32] = -1.000000000000e+12;
dLow[33] = -1.000000000000e+12;
dLow[34] = -9.999999999997e+11;
dLow[35] = -9.999999999998e+11;
dLow[36] = -9.999999999997e+11;
dLow[37] = -1.000000000000e+12;
dLow[38] = -1.000000000000e+12;
dLow[39] = -9.999999999997e+11;
dLow[40] = -1.000000000000e+12;
dLow[41] = -1.000000000000e+12;
dLow[42] = -1.000000000000e+12;
dLow[43] = -1.010000000000e+00;
dLow[44] = -1.000000000001e+12;
dLow[45] = -1.000000000000e+00;
dLow[46] = -1.000000000000e+00;
dLow[47] = -1.000000000000e+00;
dLow[48] = -9.999999999999e+11;
dLow[49] = -3.483250465375e-01;
dLow[50] = -9.999999999999e+11;
dLow[51] = -9.999999999998e+11;
dLow[52] = -7.254866340027e-01;
dLow[53] = -1.000000000000e+12;
dLow[54] = -9.999999999999e+11;
dLow[55] = -9.122002577152e-01;
dLow[56] = -1.000000000000e+12;
dLow[57] = -1.000000000000e+12;
dLow[58] = -9.999999999998e+11;
dLow[59] = -9.999999999998e+11;
dLow[60] = -1.000000000001e+12;
dLow[61] = -9.999999999993e+11;
dLow[62] = -9.999999999993e+11;
dLow[63] = -1.000000000001e+12;
dLow[64] = -9.999999999993e+11;
dLow[65] = -9.999999999992e+11;
dLow[66] = -1.000000000000e+12;
dLow[67] = -1.010000000000e+00;
dLow[68] = -1.000000000001e+12;
dLow[69] = -1.000000000000e+00;
dLow[70] = -1.000000000000e+00;
dLow[71] = -1.000000000000e+00;
dLow[72] = -1.000000000000e+12;
dLow[73] = -3.028021659371e-01;
dLow[74] = -9.999999999998e+11;
dLow[75] = -1.000000000000e+12;
dLow[76] = -5.069351753052e-01;
dLow[77] = -9.999999999999e+11;
dLow[78] = -1.000000000000e+12;
dLow[79] = -7.413186919335e-01;
dLow[80] = -1.000000000000e+12;
dLow[81] = -1.000000000000e+12;
dLow[82] = -9.999999999996e+11;
dLow[83] = -9.999999999996e+11;
dLow[84] = -1.000000000001e+12;
dLow[85] = -9.999999999988e+11;
dLow[86] = -9.999999999988e+11;
dLow[87] = -1.000000000001e+12;
dLow[88] = -9.999999999992e+11;
dLow[89] = -9.999999999993e+11;
dLow[90] = -1.000000000000e+12;
dLow[91] = -1.010000000000e+00;
dLow[92] = -1.000000000001e+12;
dLow[93] = -1.000000000000e+00;
dLow[94] = -1.000000000000e+00;
dLow[95] = -1.000000000000e+00;
dLow[96] = -1.000000000000e+12;
dLow[97] = -1.662154841108e-01;
dLow[98] = -9.999999999997e+11;
dLow[99] = -1.000000000000e+12;
dLow[100] = -3.276384833722e-01;
dLow[101] = -9.999999999997e+11;
dLow[102] = -1.000000000000e+12;
dLow[103] = -6.180010261015e-01;
dLow[104] = -1.000000000000e+12;
dLow[105] = -1.000000000001e+12;
dLow[106] = -9.999999999991e+11;
dLow[107] = -9.999999999991e+11;
dLow[108] = -1.000000000001e+12;
dLow[109] = -9.999999999995e+11;
dLow[110] = -9.999999999995e+11;
dLow[111] = -1.000000000000e+12;
dLow[112] = -9.999999999997e+11;
dLow[113] = -9.999999999997e+11;
dLow[114] = -1.000000000000e+12;
dLow[115] = -1.010000000000e+00;
dLow[116] = -1.000000000001e+12;
dLow[117] = -1.000000000000e+00;
dLow[118] = -1.000000000000e+00;
dLow[119] = -1.000000000000e+00;
dLow[120] = -1.000000000000e+12;
dLow[121] = -3.958270891046e-02;
dLow[122] = -9.999999999996e+11;
dLow[123] = -1.000000000000e+12;
dLow[124] = -3.094122245580e-01;
dLow[125] = -9.999999999997e+11;
dLow[126] = -1.000000000000e+12;
dLow[127] = -6.240348689535e-01;
dLow[128] = -1.000000000000e+12;
dLow[129] = -1.000000000000e+12;
dLow[130] = -9.999999999999e+11;
dLow[131] = -9.999999999999e+11;
dLow[132] = -9.999999999997e+11;
dLow[133] = -1.000000000000e+12;
dLow[134] = -1.000000000000e+12;
dLow[135] = -9.999999999998e+11;
dLow[136] = -1.000000000000e+12;
dLow[137] = -1.000000000000e+12;
dLow[138] = -1.000000000000e+12;
dLow[139] = -1.010000000000e+00;
dLow[140] = -1.000000000001e+12;
dLow[141] = -1.000000000000e+00;
dLow[142] = -1.000000000000e+00;
dLow[143] = -1.000000000000e+00;
dLow[144] = -1.000000000000e+12;
dLow[145] = -1.470744385981e-01;
dLow[146] = -9.999999999997e+11;
dLow[147] = -1.000000000000e+12;
dLow[148] = -3.895349806907e-01;
dLow[149] = -9.999999999998e+11;
dLow[150] = -1.000000000000e+12;
dLow[151] = -7.190465749857e-01;
dLow[152] = -1.000000000000e+12;
dLow[153] = -9.999999999991e+11;
dLow[154] = -1.000000000001e+12;
dLow[155] = -1.000000000001e+12;
dLow[156] = -9.999999999994e+11;
dLow[157] = -1.000000000001e+12;
dLow[158] = -1.000000000001e+12;
dLow[159] = -9.999999999995e+11;
dLow[160] = -1.000000000000e+12;
dLow[161] = -1.000000000000e+12;
dLow[162] = -1.000000000000e+12;
dLow[163] = -1.010000000000e+00;
dLow[164] = -1.000000000001e+12;
dLow[165] = -1.000000000000e+00;
dLow[166] = -1.000000000000e+00;
dLow[167] = -1.000000000000e+00;
dLow[168] = -9.999999999999e+11;
dLow[169] = -3.601310562953e-01;
dLow[170] = -9.999999999999e+11;
dLow[171] = -1.000000000000e+12;
dLow[172] = -5.826534265310e-01;
dLow[173] = -9.999999999999e+11;
dLow[174] = -1.000000000000e+12;
dLow[175] = -7.940651252917e-01;
dLow[176] = -1.000000000000e+12;
dLow[177] = -9.999999999990e+11;
dLow[178] = -1.000000000001e+12;
dLow[179] = -1.000000000001e+12;
dLow[180] = -9.999999999990e+11;
dLow[181] = -1.000000000001e+12;
dLow[182] = -1.000000000001e+12;
dLow[183] = -9.999999999995e+11;
dLow[184] = -1.000000000000e+12;
dLow[185] = -1.000000000000e+12;
dLow[186] = -1.000000000000e+12;
dLow[187] = -1.010000000000e+00;
dLow[188] = -1.000000000001e+12;
dLow[189] = -1.000000000000e+00;
dLow[190] = -1.000000000000e+00;
dLow[191] = -1.000000000000e+00;
dLow[192] = -9.999999999998e+11;
dLow[193] = -4.363772944133e-01;
dLow[194] = -9.999999999999e+11;
dLow[195] = -9.999999999998e+11;
dLow[196] = -7.648642470127e-01;
dLow[197] = -1.000000000000e+12;
dLow[198] = -9.999999999999e+11;
dLow[199] = -8.796978649239e-01;
dLow[200] = -1.000000000000e+12;
dLow[201] = -1.000000000000e+12;
dLow[202] = -1.000000000000e+12;
dLow[203] = -9.999999999999e+11;
dLow[204] = -9.999999999993e+11;
dLow[205] = -1.000000000000e+12;
dLow[206] = -1.000000000000e+12;
dLow[207] = -9.999999999996e+11;
dLow[208] = -1.000000000001e+12;
dLow[209] = -1.000000000000e+12;
dLow[210] = -1.000000000000e+12;
dLow[211] = -1.010000000000e+00;
dLow[212] = -1.000000000001e+12;
dLow[213] = -1.000000000000e+00;
dLow[214] = -1.000000000000e+00;
dLow[215] = -1.000000000000e+00;
dLow[216] = -9.999999999999e+11;
dLow[217] = -3.848651031161e-01;
dLow[218] = -9.999999999999e+11;
dLow[219] = -9.999999999998e+11;
dLow[220] = -7.277811641008e-01;
dLow[221] = -1.000000000000e+12;
dLow[222] = -9.999999999998e+11;
dLow[223] = -9.615050403425e-01;
dLow[224] = -1.000000000000e+12;
dLow[225] = -1.000000000000e+12;
dLow[226] = -9.999999999996e+11;
dLow[227] = -9.999999999996e+11;
dLow[228] = -1.000000000001e+12;
dLow[229] = -9.999999999993e+11;
dLow[230] = -9.999999999993e+11;
dLow[231] = -9.999999999999e+11;
dLow[232] = -1.000000000000e+12;
dLow[233] = -9.999999999999e+11;
dLow[234] = -1.000000000000e+12;
dLow[235] = -1.010000000000e+00;
dLow[236] = -1.000000000001e+12;
dLow[237] = -1.000000000000e+00;
dLow[238] = -1.000000000000e+00;
dLow[239] = -1.000000000000e+00;
dLow[240] = -1.000000000000e+12;
dLow[241] = -2.875965081530e-01;
dLow[242] = -9.999999999998e+11;
dLow[243] = -9.999999999999e+11;
dLow[244] = -5.609255426012e-01;
dLow[245] = -9.999999999999e+11;
dLow[246] = -9.999999999999e+11;
dLow[247] = -8.477531942148e-01;
dLow[248] = -1.000000000000e+12;
dLow[249] = -1.000000000000e+12;
dLow[250] = -9.999999999995e+11;
dLow[251] = -9.999999999995e+11;
dLow[252] = -1.000000000001e+12;
dLow[253] = -9.999999999992e+11;
dLow[254] = -9.999999999991e+11;
dLow[255] = -1.000000000001e+12;
dLow[256] = -9.999999999989e+11;
dLow[257] = -9.999999999989e+11;
dLow[258] = -1.000000000000e+12;
dLow[259] = -1.010000000000e+00;
dLow[260] = -1.000000000001e+12;
dLow[261] = -1.000000000000e+00;
dLow[262] = -1.000000000000e+00;
dLow[263] = -1.000000000000e+00;
dLow[264] = -1.000000000000e+12;
dLow[265] = -1.838803796863e-01;
dLow[266] = -9.999999999997e+11;
dLow[267] = -1.000000000000e+12;
dLow[268] = -4.058334354211e-01;
dLow[269] = -9.999999999997e+11;
dLow[270] = -1.000000000000e+12;
dLow[271] = -6.161049030665e-01;
dLow[272] = -1.000000000000e+12;
dLow[273] = -1.000000000001e+12;
dLow[274] = -9.999999999996e+11;
dLow[275] = -9.999999999995e+11;
dLow[276] = -1.000000000001e+12;
dLow[277] = -9.999999999993e+11;
dLow[278] = -9.999999999993e+11;
dLow[279] = -1.000000000001e+12;
dLow[280] = -9.999999999992e+11;
dLow[281] = -9.999999999993e+11;
dLow[282] = -1.000000000000e+12;
dLow[283] = -1.010000000000e+00;
dLow[284] = -1.000000000001e+12;
dLow[285] = -1.000000000000e+00;
dLow[286] = -1.000000000000e+00;
dLow[287] = -1.000000000000e+00;
dLow[288] = -1.000000000000e+12;
dLow[289] = -1.172092650693e-01;
dLow[290] = -9.999999999996e+11;
dLow[291] = -1.000000000000e+12;
dLow[292] = -2.829083664202e-01;
dLow[293] = -9.999999999996e+11;
dLow[294] = -1.000000000000e+12;
dLow[295] = -5.821510996785e-01;
dLow[296] = -1.000000000000e+12;
dLow[297] = -1.000000000000e+12;
dLow[298] = -9.999999999998e+11;
dLow[299] = -9.999999999998e+11;
dLow[300] = -1.000000000000e+12;
dLow[301] = -9.999999999997e+11;
dLow[302] = -9.999999999998e+11;
dLow[303] = -9.999999999999e+11;
dLow[304] = -1.000000000000e+12;
dLow[305] = -1.000000000000e+12;
dLow[306] = -1.000000000000e+12;
dLow[307] = -1.010000000000e+00;
dLow[308] = -1.000000000001e+12;
dLow[309] = -1.000000000000e+00;
dLow[310] = -1.000000000000e+00;
dLow[311] = -1.000000000000e+00;
dLow[312] = -1.000000000000e+12;
dLow[313] = -1.101972227457e-01;
dLow[314] = -9.999999999996e+11;
dLow[315] = -1.000000000000e+12;
dLow[316] = -3.373090547632e-01;
dLow[317] = -9.999999999997e+11;
dLow[318] = -1.000000000000e+12;
dLow[319] = -7.070927592871e-01;
dLow[320] = -1.000000000000e+12;
dLow[321] = -9.999999999998e+11;
dLow[322] = -1.000000000000e+12;
dLow[323] = -1.000000000000e+12;
dLow[324] = -9.999999999995e+11;
dLow[325] = -1.000000000001e+12;
dLow[326] = -1.000000000001e+12;
dLow[327] = -9.999999999992e+11;
dLow[328] = -1.000000000001e+12;
dLow[329] = -1.000000000001e+12;
dLow[330] = -1.000000000000e+12;
dLow[331] = -1.010000000000e+00;
dLow[332] = -1.000000000001e+12;
dLow[333] = -1.000000000000e+00;
dLow[334] = -1.000000000000e+00;
dLow[335] = -1.000000000000e+00;
dLow[336] = -1.000000000000e+12;
dLow[337] = -2.354177295463e-01;
dLow[338] = -9.999999999998e+11;
dLow[339] = -1.000000000000e+12;
dLow[340] = -5.616073459155e-01;
dLow[341] = -9.999999999999e+11;
dLow[342] = -1.000000000000e+12;
dLow[343] = -7.890979150333e-01;
dLow[344] = -1.000000000000e+12;
dLow[345] = -9.999999999993e+11;
dLow[346] = -1.000000000001e+12;
dLow[347] = -1.000000000001e+12;
dLow[348] = -9.999999999987e+11;
dLow[349] = -1.000000000001e+12;
dLow[350] = -1.000000000001e+12;
dLow[351] = -9.999999999997e+11;
dLow[352] = -1.000000000000e+12;
dLow[353] = -1.000000000000e+12;
dLow[354] = -1.000000000000e+12;
dLow[355] = -1.010000000000e+00;
dLow[356] = -1.000000000001e+12;
dLow[357] = -1.000000000000e+00;
dLow[358] = -1.000000000000e+00;
dLow[359] = -1.000000000000e+00;
dLow[360] = -9.999999999999e+11;
dLow[361] = -4.479137845165e-01;
dLow[362] = -1.000000000000e+12;
dLow[363] = -9.999999999998e+11;
dLow[364] = -7.073774762171e-01;
dLow[365] = -1.000000000000e+12;
dLow[366] = -9.999999999999e+11;
dLow[367] = -8.605606909239e-01;
dLow[368] = -1.000000000000e+12;
dLow[369] = -9.999999999991e+11;
dLow[370] = -1.000000000001e+12;
dLow[371] = -1.000000000001e+12;
dLow[372] = -9.999999999994e+11;
dLow[373] = -1.000000000000e+12;
dLow[374] = -1.000000000000e+12;
dLow[375] = -9.999999999997e+11;
dLow[376] = -1.000000000000e+12;
dLow[377] = -1.000000000000e+12;
dLow[378] = -1.000000000000e+12;
dLow[379] = -1.010000000000e+00;
dLow[380] = -1.000000000001e+12;
dLow[381] = -1.000000000000e+00;
dLow[382] = -1.000000000000e+00;
dLow[383] = -1.000000000000e+00;
dLow[384] = -9.999999999998e+11;
dLow[385] = -4.614858567287e-01;
dLow[386] = -1.000000000000e+12;
dLow[387] = -9.999999999998e+11;
dLow[388] = -7.337152239654e-01;
dLow[389] = -1.000000000000e+12;
dLow[390] = -9.999999999999e+11;
dLow[391] = -9.219700881791e-01;
dLow[392] = -1.000000000000e+12;
dLow[393] = -1.000000000000e+12;
dLow[394] = -9.999999999993e+11;
dLow[395] = -9.999999999994e+11;
dLow[396] = -1.000000000000e+12;
dLow[397] = -9.999999999999e+11;
dLow[398] = -9.999999999998e+11;
dLow[399] = -9.999999999997e+11;
dLow[400] = -1.000000000000e+12;
dLow[401] = -1.000000000000e+12;
dLow[402] = -1.000000000000e+12;
dLow[403] = -1.010000000000e+00;
dLow[404] = -1.000000000001e+12;
dLow[405] = -1.000000000000e+00;
dLow[406] = -1.000000000000e+00;
dLow[407] = -1.000000000000e+00;
dLow[408] = -9.999999999999e+11;
dLow[409] = -2.759645406140e-01;
dLow[410] = -9.999999999998e+11;
dLow[411] = -9.999999999999e+11;
dLow[412] = -6.516664430436e-01;
dLow[413] = -1.000000000000e+12;
dLow[414] = -9.999999999999e+11;
dLow[415] = -8.771372045598e-01;
dLow[416] = -1.000000000000e+12;
dLow[417] = -1.000000000001e+12;
dLow[418] = -9.999999999991e+11;
dLow[419] = -9.999999999992e+11;
dLow[420] = -1.000000000001e+12;
dLow[421] = -9.999999999992e+11;
dLow[422] = -9.999999999992e+11;
dLow[423] = -1.000000000000e+12;
dLow[424] = -9.999999999995e+11;
dLow[425] = -9.999999999993e+11;
dLow[426] = -1.000000000000e+12;
dLow[427] = -1.010000000000e+00;
dLow[428] = -1.000000000001e+12;
dLow[429] = -1.000000000000e+00;
dLow[430] = -1.000000000000e+00;
dLow[431] = -1.000000000000e+00;
dLow[432] = -1.000000000000e+12;
dLow[433] = -1.826162324133e-01;
dLow[434] = -9.999999999997e+11;
dLow[435] = -1.000000000000e+12;
dLow[436] = -4.265775349734e-01;
dLow[437] = -9.999999999998e+11;
dLow[438] = -1.000000000000e+12;
dLow[439] = -7.273531312226e-01;
dLow[440] = -1.000000000000e+12;
dLow[441] = -1.000000000001e+12;
dLow[442] = -9.999999999999e+11;
dLow[443] = -9.999999999998e+11;
dLow[444] = -1.000000000001e+12;
dLow[445] = -9.999999999988e+11;
dLow[446] = -9.999999999988e+11;
dLow[447] = -1.000000000001e+12;
dLow[448] = -9.999999999991e+11;
dLow[449] = -9.999999999993e+11;
dLow[450] = -1.000000000000e+12;
dLow[451] = -1.010000000000e+00;
dLow[452] = -1.000000000001e+12;
dLow[453] = -1.000000000000e+00;
dLow[454] = -1.000000000000e+00;
dLow[455] = -1.000000000000e+00;
dLow[456] = -1.000000000000e+12;
dLow[457] = -1.590850284975e-01;
dLow[458] = -9.999999999997e+11;
dLow[459] = -1.000000000000e+12;
dLow[460] = -2.805851940069e-01;
dLow[461] = -9.999999999996e+11;
dLow[462] = -1.000000000000e+12;
dLow[463] = -5.752092471259e-01;
dLow[464] = -1.000000000000e+12;
dLow[465] = -9.999999999999e+11;
dLow[466] = -9.999999999998e+11;
dLow[467] = -9.999999999999e+11;
dLow[468] = -1.000000000001e+12;
dLow[469] = -9.999999999999e+11;
dLow[470] = -9.999999999998e+11;
dLow[471] = -1.000000000000e+12;
dLow[472] = -9.999999999995e+11;
dLow[473] = -9.999999999997e+11;
dLow[474] = -1.000000000000e+12;
dLow[475] = -1.010000000000e+00;
dLow[476] = -1.000000000001e+12;
dLow[477] = -1.000000000000e+00;
dLow[478] = -1.000000000000e+00;
dLow[479] = -1.000000000000e+00;
dLow[480] = -1.000000000000e+12;
dLow[481] = -1.160674810867e-01;
dLow[482] = -9.999999999996e+11;
dLow[483] = -1.000000000000e+12;
dLow[484] = -3.420354108234e-01;
dLow[485] = -9.999999999997e+11;
dLow[486] = -1.000000000000e+12;
dLow[487] = -5.978581303118e-01;
dLow[488] = -1.000000000000e+12;
dLow[489] = -1.000000000000e+12;
dLow[490] = -9.999999999999e+11;
dLow[491] = -1.000000000000e+12;
dLow[492] = -9.999999999995e+11;
dLow[493] = -1.000000000001e+12;
dLow[494] = -1.000000000001e+12;
dLow[495] = -9.999999999997e+11;
dLow[496] = -1.000000000001e+12;
dLow[497] = -1.000000000001e+12;
dLow[498] = -1.000000000000e+12;
dLow[499] = -1.010000000000e+00;
dLow[500] = -1.000000000001e+12;
dLow[501] = -1.000000000000e+00;
dLow[502] = -1.000000000000e+00;
dLow[503] = -1.000000000000e+00;
dLow[504] = -1.000000000000e+12;
dLow[505] = -1.880351238226e-01;
dLow[506] = -9.999999999997e+11;
dLow[507] = -1.000000000000e+12;
dLow[508] = -4.803034753734e-01;
dLow[509] = -9.999999999999e+11;
dLow[510] = -1.000000000000e+12;
dLow[511] = -7.948742351340e-01;
dLow[512] = -1.000000000000e+12;
dLow[513] = -9.999999999997e+11;
dLow[514] = -1.000000000001e+12;
dLow[515] = -1.000000000001e+12;
dLow[516] = -9.999999999990e+11;
dLow[517] = -1.000000000001e+12;
dLow[518] = -1.000000000001e+12;
dLow[519] = -9.999999999991e+11;
dLow[520] = -1.000000000001e+12;
dLow[521] = -1.000000000001e+12;
dLow[522] = -1.000000000000e+12;
dLow[523] = -1.010000000000e+00;
dLow[524] = -1.000000000001e+12;
dLow[525] = -1.000000000000e+00;
dLow[526] = -1.000000000000e+00;
dLow[527] = -1.000000000000e+00;
dLow[528] = -1.000000000000e+12;
dLow[529] = -3.724416092051e-01;
dLow[530] = -9.999999999999e+11;
dLow[531] = -9.999999999999e+11;
dLow[532] = -6.548645112370e-01;
dLow[533] = -1.000000000000e+12;
dLow[534] = -9.999999999999e+11;
dLow[535] = -9.013477510110e-01;
dLow[536] = -1.000000000000e+12;
dLow[537] = -9.999999999990e+11;
dLow[538] = -1.000000000001e+12;
dLow[539] = -1.000000000001e+12;
dLow[540] = -9.999999999993e+11;
dLow[541] = -1.000000000001e+12;
dLow[542] = -1.000000000001e+12;
dLow[543] = -9.999999999993e+11;
dLow[544] = -1.000000000000e+12;
dLow[545] = -1.000000000000e+12;
dLow[546] = -1.000000000000e+12;
dLow[547] = -1.010000000000e+00;
dLow[548] = -1.000000000001e+12;
dLow[549] = -1.000000000000e+00;
dLow[550] = -1.000000000000e+00;
dLow[551] = -1.000000000000e+00;
dLow[552] = -9.999999999998e+11;
dLow[553] = -4.689063789091e-01;
dLow[554] = -1.000000000000e+12;
dLow[555] = -9.999999999998e+11;
dLow[556] = -7.778747903160e-01;
dLow[557] = -1.000000000000e+12;
dLow[558] = -9.999999999998e+11;
dLow[559] = -8.692566523857e-01;
dLow[560] = -1.000000000000e+12;
dLow[561] = -9.999999999996e+11;
dLow[562] = -1.000000000000e+12;
dLow[563] = -9.999999999999e+11;
dLow[564] = -9.999999999996e+11;
dLow[565] = -1.000000000000e+12;
dLow[566] = -1.000000000000e+12;
dLow[567] = -1.000000000000e+12;
dLow[568] = -9.999999999998e+11;
dLow[569] = -9.999999999998e+11;
dLow[570] = -1.000000000000e+12;
dLow[571] = -1.010000000000e+00;
dLow[572] = -1.000000000001e+12;
dLow[573] = -1.000000000000e+00;
dLow[574] = -1.000000000000e+00;
dLow[575] = -1.000000000000e+00;
dLow[576] = -9.999999999998e+11;
dLow[577] = -3.820378617879e-01;
dLow[578] = -9.999999999999e+11;
dLow[579] = -9.999999999998e+11;
dLow[580] = -6.768563741150e-01;
dLow[581] = -1.000000000000e+12;
dLow[582] = -9.999999999999e+11;
dLow[583] = -8.668078641543e-01;
dLow[584] = -1.000000000000e+12;
dLow[585] = -1.000000000001e+12;
dLow[586] = -9.999999999992e+11;
dLow[587] = -9.999999999992e+11;
dLow[588] = -1.000000000000e+12;
dLow[589] = -9.999999999990e+11;
dLow[590] = -9.999999999990e+11;
dLow[591] = -1.000000000000e+12;
dLow[592] = -1.000000000000e+12;
dLow[593] = -9.999999999998e+11;
dLow[594] = -1.000000000000e+12;
dLow[595] = -1.010000000000e+00;
dLow[596] = -1.000000000001e+12;
dLow[597] = -1.000000000000e+00;
dLow[598] = -1.000000000000e+00;
dLow[599] = -1.000000000000e+00;
dLow[600] = -1.000000000000e+12;
dLow[601] = -1.999388284532e-01;
dLow[602] = -9.999999999997e+11;
dLow[603] = -1.000000000000e+12;
dLow[604] = -4.648863484942e-01;
dLow[605] = -9.999999999998e+11;
dLow[606] = -1.000000000000e+12;
dLow[607] = -8.102838444591e-01;
dLow[608] = -1.000000000000e+12;
dLow[609] = -1.000000000001e+12;
dLow[610] = -9.999999999991e+11;
dLow[611] = -9.999999999992e+11;
dLow[612] = -1.000000000001e+12;
dLow[613] = -9.999999999991e+11;
dLow[614] = -9.999999999990e+11;
dLow[615] = -1.000000000000e+12;
dLow[616] = -9.999999999993e+11;
dLow[617] = -9.999999999994e+11;
dLow[618] = -1.000000000000e+12;
dLow[619] = -1.010000000000e+00;
dLow[620] = -1.000000000001e+12;
dLow[621] = -1.000000000000e+00;
dLow[622] = -1.000000000000e+00;
dLow[623] = -1.000000000000e+00;
dLow[624] = -1.000000000000e+12;
dLow[625] = -9.505498415426e-02;
dLow[626] = -9.999999999996e+11;
dLow[627] = -1.000000000000e+12;
dLow[628] = -3.415898172234e-01;
dLow[629] = -9.999999999997e+11;
dLow[630] = -1.000000000000e+12;
dLow[631] = -6.219722539733e-01;
dLow[632] = -1.000000000000e+12;
dLow[633] = -1.000000000000e+12;
dLow[634] = -9.999999999999e+11;
dLow[635] = -9.999999999999e+11;
dLow[636] = -1.000000000001e+12;
dLow[637] = -9.999999999996e+11;
dLow[638] = -9.999999999997e+11;
dLow[639] = -1.000000000001e+12;
dLow[640] = -9.999999999991e+11;
dLow[641] = -9.999999999993e+11;
dLow[642] = -1.000000000000e+12;
dLow[643] = -1.010000000000e+00;
dLow[644] = -1.000000000001e+12;
dLow[645] = -1.000000000000e+00;
dLow[646] = -1.000000000000e+00;
dLow[647] = -1.000000000000e+00;
dLow[648] = -1.000000000000e+12;
dLow[649] = -1.433700443716e-01;
dLow[650] = -9.999999999997e+11;
dLow[651] = -1.000000000000e+12;
dLow[652] = -3.021641124706e-01;
dLow[653] = -9.999999999997e+11;
dLow[654] = -1.000000000000e+12;
dLow[655] = -5.583285305989e-01;
dLow[656] = -1.000000000000e+12;
dLow[657] = -9.999999999998e+11;
dLow[658] = -1.000000000000e+12;
dLow[659] = -1.000000000000e+12;
dLow[660] = -9.999999999998e+11;
dLow[661] = -1.000000000000e+12;
dLow[662] = -1.000000000000e+12;
dLow[663] = -1.000000000000e+12;
dLow[664] = -1.000000000000e+12;
dLow[665] = -1.000000000000e+12;
dLow[666] = -1.000000000000e+12;
dLow[667] = -1.010000000000e+00;
dLow[668] = -1.000000000001e+12;
dLow[669] = -1.000000000000e+00;
dLow[670] = -1.000000000000e+00;
dLow[671] = -1.000000000000e+00;
dLow[672] = -1.000000000000e+12;
dLow[673] = -2.141372205670e-01;
dLow[674] = -9.999999999997e+11;
dLow[675] = -1.000000000000e+12;
dLow[676] = -4.071809593655e-01;
dLow[677] = -9.999999999998e+11;
dLow[678] = -1.000000000000e+12;
dLow[679] = -7.185347670973e-01;
dLow[680] = -1.000000000000e+12;
dLow[681] = -9.999999999994e+11;
dLow[682] = -1.000000000000e+12;
dLow[683] = -1.000000000000e+12;
dLow[684] = -9.999999999995e+11;
dLow[685] = -1.000000000001e+12;
dLow[686] = -1.000000000001e+12;
dLow[687] = -9.999999999992e+11;
dLow[688] = -1.000000000001e+12;
dLow[689] = -1.000000000001e+12;
dLow[690] = -1.000000000000e+12;
dLow[691] = -1.010000000000e+00;
dLow[692] = -1.000000000001e+12;
dLow[693] = -1.000000000000e+00;
dLow[694] = -1.000000000000e+00;
dLow[695] = -1.000000000000e+00;
dLow[696] = -1.000000000000e+12;
dLow[697] = -2.795501500680e-01;
dLow[698] = -9.999999999998e+11;
dLow[699] = -1.000000000000e+12;
dLow[700] = -6.456593641208e-01;
dLow[701] = -1.000000000000e+12;
dLow[702] = -9.999999999999e+11;
dLow[703] = -8.816721262188e-01;
dLow[704] = -1.000000000000e+12;
dLow[705] = -9.999999999997e+11;
dLow[706] = -1.000000000001e+12;
dLow[707] = -1.000000000000e+12;
dLow[708] = -9.999999999990e+11;
dLow[709] = -1.000000000001e+12;
dLow[710] = -1.000000000001e+12;
dLow[711] = -9.999999999990e+11;
dLow[712] = -1.000000000001e+12;
dLow[713] = -1.000000000001e+12;
dLow[714] = -1.000000000000e+12;
dLow[715] = -1.010000000000e+00;
dLow[716] = -1.000000000001e+12;
dLow[717] = -1.000000000000e+00;
dLow[718] = -1.000000000000e+00;
dLow[719] = -1.000000000000e+00;
dLow[720] = -9.999999999999e+11;
dLow[721] = -4.288047628250e-01;
dLow[722] = -9.999999999999e+11;
dLow[723] = -9.999999999998e+11;
dLow[724] = -7.586814266311e-01;
dLow[725] = -1.000000000000e+12;
dLow[726] = -9.999999999998e+11;
dLow[727] = -9.264810286873e-01;
dLow[728] = -1.000000000000e+12;
dLow[729] = -9.999999999996e+11;
dLow[730] = -1.000000000001e+12;
dLow[731] = -1.000000000001e+12;
dLow[732] = -9.999999999992e+11;
dLow[733] = -1.000000000000e+12;
dLow[734] = -1.000000000000e+12;
dLow[735] = -9.999999999999e+11;
dLow[736] = -9.999999999999e+11;
dLow[737] = -9.999999999999e+11;
dLow[738] = -1.000000000000e+12;
dLow[739] = -1.010000000000e+00;
dLow[740] = -1.000000000001e+12;
dLow[741] = -1.000000000000e+00;
dLow[742] = -1.000000000000e+00;
dLow[743] = -1.000000000000e+00;
dLow[744] = -9.999999999998e+11;
dLow[745] = -4.678993367663e-01;
dLow[746] = -1.000000000000e+12;
dLow[747] = -9.999999999998e+11;
dLow[748] = -6.923523007671e-01;
dLow[749] = -1.000000000000e+12;
dLow[750] = -9.999999999999e+11;
dLow[751] = -8.725490404897e-01;
dLow[752] = -1.000000000000e+12;
dLow[753] = -9.999999999997e+11;
dLow[754] = -9.999999999995e+11;
dLow[755] = -9.999999999996e+11;
dLow[756] = -1.000000000000e+12;
dLow[757] = -9.999999999995e+11;
dLow[758] = -9.999999999993e+11;
dLow[759] = -1.000000000000e+12;
dLow[760] = -9.999999999996e+11;
dLow[761] = -9.999999999995e+11;
dLow[762] = -1.000000000000e+12;
dLow[763] = -1.010000000000e+00;
dLow[764] = -1.000000000001e+12;
dLow[765] = -1.000000000000e+00;
dLow[766] = -1.000000000000e+00;
dLow[767] = -1.000000000000e+00;
dLow[768] = -9.999999999999e+11;
dLow[769] = -2.574076400724e-01;
dLow[770] = -9.999999999998e+11;
dLow[771] = -9.999999999999e+11;
dLow[772] = -5.598650210198e-01;
dLow[773] = -9.999999999999e+11;
dLow[774] = -1.000000000000e+12;
dLow[775] = -7.760814817360e-01;
dLow[776] = -1.000000000000e+12;
dLow[777] = -1.000000000001e+12;
dLow[778] = -9.999999999987e+11;
dLow[779] = -9.999999999988e+11;
dLow[780] = -1.000000000001e+12;
dLow[781] = -9.999999999992e+11;
dLow[782] = -9.999999999992e+11;
dLow[783] = -1.000000000000e+12;
dLow[784] = -9.999999999995e+11;
dLow[785] = -9.999999999995e+11;
dLow[786] = -1.000000000000e+12;
dLow[787] = -1.010000000000e+00;
dLow[788] = -1.000000000001e+12;
dLow[789] = -1.000000000000e+00;
dLow[790] = -1.000000000000e+00;
dLow[791] = -1.000000000000e+00;
dLow[792] = -1.000000000000e+12;
dLow[793] = -8.584375823923e-02;
dLow[794] = -9.999999999996e+11;
dLow[795] = -1.000000000000e+12;
dLow[796] = -3.607602524124e-01;
dLow[797] = -9.999999999997e+11;
dLow[798] = -1.000000000000e+12;
dLow[799] = -6.955584298081e-01;
dLow[800] = -1.000000000000e+12;
dLow[801] = -1.000000000001e+12;
dLow[802] = -9.999999999997e+11;
dLow[803] = -9.999999999996e+11;
dLow[804] = -1.000000000001e+12;
dLow[805] = -9.999999999990e+11;
dLow[806] = -9.999999999992e+11;
dLow[807] = -1.000000000000e+12;
dLow[808] = -9.999999999996e+11;
dLow[809] = -9.999999999997e+11;
dLow[810] = -1.000000000000e+12;
dLow[811] = -1.010000000000e+00;
dLow[812] = -1.000000000001e+12;
dLow[813] = -1.000000000000e+00;
dLow[814] = -1.000000000000e+00;
dLow[815] = -1.000000000000e+00;
dLow[816] = -1.000000000000e+12;
dLow[817] = -1.097759667638e-01;
dLow[818] = -9.999999999996e+11;
dLow[819] = -1.000000000000e+12;
dLow[820] = -2.595367349029e-01;
dLow[821] = -9.999999999996e+11;
dLow[822] = -1.000000000000e+12;
dLow[823] = -6.108059213761e-01;
dLow[824] = -1.000000000000e+12;
dLow[825] = -1.000000000000e+12;
dLow[826] = -1.000000000000e+12;
dLow[827] = -1.000000000001e+12;
dLow[828] = -1.000000000000e+12;
dLow[829] = -1.000000000000e+12;
dLow[830] = -1.000000000000e+12;
dLow[831] = -1.000000000000e+12;
dLow[832] = -9.999999999996e+11;
dLow[833] = -9.999999999998e+11;
dLow[834] = -1.000000000000e+12;
dLow[835] = -1.010000000000e+00;
dLow[836] = -1.000000000001e+12;
dLow[837] = -1.000000000000e+00;
dLow[838] = -1.000000000000e+00;
dLow[839] = -1.000000000000e+00;
dLow[840] = -1.000000000000e+12;
dLow[841] = -1.900712411346e-01;
dLow[842] = -9.999999999997e+11;
dLow[843] = -1.000000000000e+12;
dLow[844] = -3.969784810935e-01;
dLow[845] = -9.999999999998e+11;
dLow[846] = -1.000000000000e+12;
dLow[847] = -6.096791239478e-01;
dLow[848] = -1.000000000000e+12;
dLow[849] = -9.999999999992e+11;
dLow[850] = -1.000000000000e+12;
dLow[851] = -1.000000000001e+12;
dLow[852] = -9.999999999996e+11;
dLow[853] = -1.000000000001e+12;
dLow[854] = -1.000000000001e+12;
dLow[855] = -9.999999999999e+11;
dLow[856] = -1.000000000001e+12;
dLow[857] = -1.000000000000e+12;
dLow[858] = -1.000000000000e+12;
dLow[859] = -1.010000000000e+00;
dLow[860] = -1.000000000001e+12;
dLow[861] = -1.000000000000e+00;
dLow[862] = -1.000000000000e+00;
dLow[863] = -1.000000000000e+00;
dLow[864] = -1.000000000000e+12;
dLow[865] = -2.895915735152e-01;
dLow[866] = -9.999999999998e+11;
dLow[867] = -1.000000000000e+12;
dLow[868] = -5.820412158276e-01;
dLow[869] = -1.000000000000e+12;
dLow[870] = -1.000000000000e+12;
dLow[871] = -8.115850826710e-01;
dLow[872] = -1.000000000000e+12;
dLow[873] = -9.999999999996e+11;
dLow[874] = -1.000000000001e+12;
dLow[875] = -1.000000000000e+12;
dLow[876] = -9.999999999988e+11;
dLow[877] = -1.000000000001e+12;
dLow[878] = -1.000000000001e+12;
dLow[879] = -9.999999999994e+11;
dLow[880] = -1.000000000001e+12;
dLow[881] = -1.000000000001e+12;
dLow[882] = -1.000000000000e+12;
dLow[883] = -1.010000000000e+00;
dLow[884] = -1.000000000001e+12;
dLow[885] = -1.000000000000e+00;
dLow[886] = -1.000000000000e+00;
dLow[887] = -1.000000000000e+00;
dLow[888] = -9.999999999999e+11;
dLow[889] = -3.941125823101e-01;
dLow[890] = -9.999999999999e+11;
dLow[891] = -9.999999999998e+11;
dLow[892] = -7.111715382190e-01;
dLow[893] = -1.000000000000e+12;
dLow[894] = -9.999999999999e+11;
dLow[895] = -9.826872278596e-01;
dLow[896] = -1.000000000000e+12;
dLow[897] = -9.999999999997e+11;
dLow[898] = -1.000000000000e+12;
dLow[899] = -1.000000000000e+12;
dLow[900] = -9.999999999993e+11;
dLow[901] = -1.000000000001e+12;
dLow[902] = -1.000000000000e+12;
dLow[903] = -9.999999999991e+11;
dLow[904] = -1.000000000000e+12;
dLow[905] = -1.000000000000e+12;
dLow[906] = -1.000000000000e+12;
dLow[907] = -1.010000000000e+00;
dLow[908] = -1.000000000001e+12;
dLow[909] = -1.000000000000e+00;
dLow[910] = -1.000000000000e+00;
dLow[911] = -1.000000000000e+00;
dLow[912] = -9.999999999999e+11;
dLow[913] = -4.248663668234e-01;
dLow[914] = -9.999999999999e+11;
dLow[915] = -9.999999999998e+11;
dLow[916] = -7.649337416279e-01;
dLow[917] = -1.000000000000e+12;
dLow[918] = -9.999999999998e+11;
dLow[919] = -9.059873530427e-01;
dLow[920] = -1.000000000000e+12;
dLow[921] = -9.999999999998e+11;
dLow[922] = -9.999999999999e+11;
dLow[923] = -9.999999999999e+11;
dLow[924] = -1.000000000000e+12;
dLow[925] = -9.999999999998e+11;
dLow[926] = -9.999999999996e+11;
dLow[927] = -9.999999999999e+11;
dLow[928] = -9.999999999992e+11;
dLow[929] = -9.999999999992e+11;
dLow[930] = -1.000000000000e+12;
dLow[931] = -1.010000000000e+00;
dLow[932] = -1.000000000001e+12;
dLow[933] = -1.000000000000e+00;
dLow[934] = -1.000000000000e+00;
dLow[935] = -1.000000000000e+00;
dLow[936] = -9.999999999999e+11;
dLow[937] = -3.541261581976e-01;
dLow[938] = -9.999999999998e+11;
dLow[939] = -9.999999999998e+11;
dLow[940] = -6.109360209392e-01;
dLow[941] = -9.999999999999e+11;
dLow[942] = -9.999999999999e+11;
dLow[943] = -7.740220541765e-01;
dLow[944] = -1.000000000000e+12;
dLow[945] = -1.000000000000e+12;
dLow[946] = -9.999999999994e+11;
dLow[947] = -9.999999999993e+11;
dLow[948] = -1.000000000001e+12;
dLow[949] = -9.999999999988e+11;
dLow[950] = -9.999999999989e+11;
dLow[951] = -1.000000000001e+12;
dLow[952] = -9.999999999996e+11;
dLow[953] = -9.999999999994e+11;
dLow[954] = -1.000000000000e+12;
dLow[955] = -1.010000000000e+00;
dLow[956] = -1.000000000001e+12;
dLow[957] = -1.000000000000e+00;
dLow[958] = -1.000000000000e+00;
dLow[959] = -1.000000000000e+00;
dLow[960] = -1.000000000000e+12;
dLow[961] = -1.776394152020e-01;
dLow[962] = -9.999999999997e+11;
dLow[963] = -1.000000000000e+12;
dLow[964] = -3.735500773957e-01;
dLow[965] = -9.999999999997e+11;
dLow[966] = -1.000000000000e+12;
dLow[967] = -7.132848302410e-01;
dLow[968] = -1.000000000000e+12;
dLow[969] = -1.000000000001e+12;
dLow[970] = -9.999999999990e+11;
dLow[971] = -9.999999999991e+11;
dLow[972] = -1.000000000001e+12;
dLow[973] = -9.999999999991e+11;
dLow[974] = -9.999999999992e+11;
dLow[975] = -1.000000000001e+12;
dLow[976] = -9.999999999997e+11;
dLow[977] = -9.999999999998e+11;
dLow[978] = -1.000000000000e+12;
dLow[979] = -1.010000000000e+00;
dLow[980] = -1.000000000001e+12;
dLow[981] = -1.000000000000e+00;
dLow[982] = -1.000000000000e+00;
dLow[983] = -1.000000000000e+00;
dLow[984] = -1.000000000000e+12;
dLow[985] = -3.383177149579e-02;
dLow[986] = -9.999999999996e+11;
dLow[987] = -1.000000000000e+12;
dLow[988] = -2.998847405269e-01;
dLow[989] = -9.999999999997e+11;
dLow[990] = -1.000000000000e+12;
dLow[991] = -6.348697049761e-01;
dLow[992] = -1.000000000000e+12;
dLow[993] = -1.000000000001e+12;
dLow[994] = -9.999999999998e+11;
dLow[995] = -9.999999999999e+11;
dLow[996] = -1.000000000001e+12;
dLow[997] = -1.000000000000e+12;
dLow[998] = -1.000000000000e+12;
dLow[999] = -9.999999999999e+11;
dLow[1000] = -9.999999999996e+11;
dLow[1001] = -9.999999999998e+11;
dLow[1002] = -1.000000000000e+12;
dLow[1003] = -1.010000000000e+00;
dLow[1004] = -1.000000000001e+12;
dLow[1005] = -1.000000000000e+00;
dLow[1006] = -1.000000000000e+00;
dLow[1007] = -1.000000000000e+00;
dLow[1008] = -1.000000000000e+12;
dLow[1009] = -1.283731083811e-01;
dLow[1010] = -9.999999999997e+11;
dLow[1011] = -1.000000000000e+12;
dLow[1012] = -3.471182345050e-01;
dLow[1013] = -9.999999999997e+11;
dLow[1014] = -1.000000000000e+12;
dLow[1015] = -6.154814314061e-01;
dLow[1016] = -1.000000000000e+12;
dLow[1017] = -9.999999999999e+11;
dLow[1018] = -1.000000000001e+12;
dLow[1019] = -1.000000000001e+12;
dLow[1020] = -9.999999999995e+11;
dLow[1021] = -1.000000000000e+12;
dLow[1022] = -1.000000000001e+12;
dLow[1023] = -1.000000000000e+12;
dLow[1024] = -1.000000000000e+12;
dLow[1025] = -1.000000000000e+12;
dLow[1026] = -1.000000000000e+12;
dLow[1027] = -1.010000000000e+00;
dLow[1028] = -1.000000000001e+12;
dLow[1029] = -1.000000000000e+00;
dLow[1030] = -1.000000000000e+00;
dLow[1031] = -1.000000000000e+00;
dLow[1032] = -1.000000000000e+12;
dLow[1033] = -3.169197318183e-01;
dLow[1034] = -9.999999999999e+11;
dLow[1035] = -1.000000000000e+12;
dLow[1036] = -4.891766407178e-01;
dLow[1037] = -9.999999999999e+11;
dLow[1038] = -1.000000000000e+12;
dLow[1039] = -7.416488537625e-01;
dLow[1040] = -1.000000000000e+12;
dLow[1041] = -9.999999999991e+11;
dLow[1042] = -1.000000000001e+12;
dLow[1043] = -1.000000000001e+12;
dLow[1044] = -9.999999999991e+11;
dLow[1045] = -1.000000000001e+12;
dLow[1046] = -1.000000000001e+12;
dLow[1047] = -9.999999999997e+11;
dLow[1048] = -1.000000000001e+12;
dLow[1049] = -1.000000000001e+12;
dLow[1050] = -1.000000000000e+12;
dLow[1051] = -1.010000000000e+00;
dLow[1052] = -1.000000000001e+12;
dLow[1053] = -1.000000000000e+00;
dLow[1054] = -1.000000000000e+00;
dLow[1055] = -1.000000000000e+00;
dLow[1056] = -9.999999999999e+11;
dLow[1057] = -3.708979066378e-01;
dLow[1058] = -9.999999999999e+11;
dLow[1059] = -9.999999999999e+11;
dLow[1060] = -7.159166897590e-01;
dLow[1061] = -1.000000000000e+12;
dLow[1062] = -1.000000000000e+12;
dLow[1063] = -9.028456613990e-01;
dLow[1064] = -1.000000000000e+12;
dLow[1065] = -9.999999999992e+11;
dLow[1066] = -1.000000000000e+12;
dLow[1067] = -9.999999999999e+11;
dLow[1068] = -9.999999999993e+11;
dLow[1069] = -1.000000000001e+12;
dLow[1070] = -1.000000000001e+12;
dLow[1071] = -9.999999999991e+11;
dLow[1072] = -1.000000000001e+12;
dLow[1073] = -1.000000000001e+12;
dLow[1074] = -1.000000000000e+12;
dLow[1075] = -1.010000000000e+00;
dLow[1076] = -1.000000000001e+12;
dLow[1077] = -1.000000000000e+00;
dLow[1078] = -1.000000000000e+00;
dLow[1079] = -1.000000000000e+00;
dLow[1080] = -9.999999999998e+11;
dLow[1081] = -3.895454168729e-01;
dLow[1082] = -9.999999999999e+11;
dLow[1083] = -9.999999999998e+11;
dLow[1084] = -7.788428308749e-01;
dLow[1085] = -1.000000000000e+12;
dLow[1086] = -9.999999999998e+11;
dLow[1087] = -9.706435062045e-01;
dLow[1088] = -1.000000000000e+12;
dLow[1089] = -1.000000000000e+12;
dLow[1090] = -1.000000000000e+12;
dLow[1091] = -9.999999999999e+11;
dLow[1092] = -9.999999999995e+11;
dLow[1093] = -9.999999999996e+11;
dLow[1094] = -9.999999999996e+11;
dLow[1095] = -9.999999999996e+11;
dLow[1096] = -9.999999999999e+11;
dLow[1097] = -9.999999999998e+11;
dLow[1098] = -1.000000000000e+12;
dLow[1099] = -1.010000000000e+00;
dLow[1100] = -1.000000000001e+12;
dLow[1101] = -1.000000000000e+00;
dLow[1102] = -1.000000000000e+00;
dLow[1103] = -1.000000000000e+00;
dLow[1104] = -9.999999999999e+11;
dLow[1105] = -3.947487882668e-01;
dLow[1106] = -9.999999999999e+11;
dLow[1107] = -9.999999999998e+11;
dLow[1108] = -6.289632449356e-01;
dLow[1109] = -1.000000000000e+12;
dLow[1110] = -9.999999999998e+11;
dLow[1111] = -8.662649442909e-01;
dLow[1112] = -1.000000000000e+12;
dLow[1113] = -1.000000000000e+12;
dLow[1114] = -9.999999999997e+11;
dLow[1115] = -9.999999999997e+11;
dLow[1116] = -1.000000000000e+12;
dLow[1117] = -9.999999999991e+11;
dLow[1118] = -9.999999999990e+11;
dLow[1119] = -1.000000000000e+12;
dLow[1120] = -9.999999999991e+11;
dLow[1121] = -9.999999999991e+11;
dLow[1122] = -1.000000000000e+12;
dLow[1123] = -1.010000000000e+00;
dLow[1124] = -1.000000000001e+12;
dLow[1125] = -1.000000000000e+00;
dLow[1126] = -1.000000000000e+00;
dLow[1127] = -1.000000000000e+00;
dLow[1128] = -1.000000000000e+12;
dLow[1129] = -2.441327588019e-01;
dLow[1130] = -9.999999999997e+11;
dLow[1131] = -9.999999999999e+11;
dLow[1132] = -4.640737618958e-01;
dLow[1133] = -9.999999999998e+11;
dLow[1134] = -1.000000000000e+12;
dLow[1135] = -6.740359401262e-01;
dLow[1136] = -1.000000000000e+12;
dLow[1137] = -1.000000000000e+12;
dLow[1138] = -9.999999999989e+11;
dLow[1139] = -9.999999999992e+11;
dLow[1140] = -1.000000000001e+12;
dLow[1141] = -9.999999999993e+11;
dLow[1142] = -9.999999999992e+11;
dLow[1143] = -1.000000000001e+12;
dLow[1144] = -9.999999999992e+11;
dLow[1145] = -9.999999999994e+11;
dLow[1146] = -1.000000000000e+12;
dLow[1147] = -1.010000000000e+00;
dLow[1148] = -1.000000000001e+12;
dLow[1149] = -1.000000000000e+00;
dLow[1150] = -1.000000000000e+00;
dLow[1151] = -1.000000000000e+00;
dLow[1152] = -1.000000000000e+12;
dLow[1153] = -6.952692800220e-02;
dLow[1154] = -9.999999999996e+11;
dLow[1155] = -1.000000000000e+12;
dLow[1156] = -3.185629322454e-01;
dLow[1157] = -9.999999999997e+11;
dLow[1158] = -1.000000000000e+12;
dLow[1159] = -6.193675376126e-01;
dLow[1160] = -1.000000000000e+12;
dLow[1161] = -1.000000000001e+12;
dLow[1162] = -9.999999999995e+11;
dLow[1163] = -9.999999999995e+11;
dLow[1164] = -1.000000000001e+12;
dLow[1165] = -9.999999999993e+11;
dLow[1166] = -9.999999999997e+11;
dLow[1167] = -1.000000000001e+12;
dLow[1168] = -1.000000000000e+12;
dLow[1169] = -1.000000000000e+12;
dLow[1170] = -1.000000000000e+12;
dLow[1171] = -1.010000000000e+00;
dLow[1172] = -1.000000000001e+12;
dLow[1173] = -1.000000000000e+00;
dLow[1174] = -1.000000000000e+00;
dLow[1175] = -1.000000000000e+00;
dLow[1176] = -1.000000000000e+12;
dLow[1177] = -7.990658822914e-02;
dLow[1178] = -9.999999999996e+11;
dLow[1179] = -1.000000000000e+12;
dLow[1180] = -2.747276187230e-01;
dLow[1181] = -9.999999999997e+11;
dLow[1182] = -1.000000000000e+12;
dLow[1183] = -6.732033723223e-01;
dLow[1184] = -1.000000000000e+12;
dLow[1185] = -1.000000000001e+12;
dLow[1186] = -1.000000000001e+12;
dLow[1187] = -1.000000000001e+12;
dLow[1188] = -9.999999999999e+11;
dLow[1189] = -1.000000000000e+12;
dLow[1190] = -1.000000000001e+12;
dLow[1191] = -9.999999999998e+11;
dLow[1192] = -1.000000000000e+12;
dLow[1193] = -1.000000000000e+12;
dLow[1194] = -1.000000000000e+12;
dLow[1195] = -1.010000000000e+00;
dLow[1196] = -1.000000000001e+12;
dLow[1197] = -1.000000000000e+00;
dLow[1198] = -1.000000000000e+00;
dLow[1199] = -1.000000000000e+00;
dLow[1200] = -1.000000000000e+12;
dLow[1201] = -2.252616945075e-01;
dLow[1202] = -9.999999999998e+11;
dLow[1203] = -1.000000000000e+12;
dLow[1204] = -4.643507734679e-01;
dLow[1205] = -9.999999999999e+11;
dLow[1206] = -1.000000000000e+12;
dLow[1207] = -6.975107904621e-01;
dLow[1208] = -1.000000000000e+12;
dLow[1209] = -9.999999999993e+11;
dLow[1210] = -1.000000000001e+12;
dLow[1211] = -1.000000000001e+12;
dLow[1212] = -9.999999999995e+11;
dLow[1213] = -1.000000000001e+12;
dLow[1214] = -1.000000000001e+12;
dLow[1215] = -9.999999999994e+11;
dLow[1216] = -1.000000000000e+12;
dLow[1217] = -1.000000000000e+12;
dLow[1218] = -1.000000000000e+12;
dLow[1219] = -1.010000000000e+00;
dLow[1220] = -1.000000000001e+12;
dUpp[0] = 1.000000000000e+12;
dUpp[1] = 9.999999999996e+11;
dUpp[2] = 1.000000000000e+12;
dUpp[3] = 1.000000000000e+12;
dUpp[4] = 9.999999999994e+11;
dUpp[5] = 1.000000000000e+12;
dUpp[6] = 1.000000000000e+12;
dUpp[7] = 9.999999999992e+11;
dUpp[8] = 9.999999999996e+11;
dUpp[9] = 1.000000000001e+12;
dUpp[10] = 9.999999999995e+11;
dUpp[11] = 9.999999999995e+11;
dUpp[12] = 1.000000000001e+12;
dUpp[13] = 9.999999999993e+11;
dUpp[14] = 9.999999999994e+11;
dUpp[15] = 1.000000000001e+12;
dUpp[16] = 9.999999999990e+11;
dUpp[17] = 9.999999999990e+11;
dUpp[18] = 1.000000000000e+12;
dUpp[19] = 9.999999999990e+11;
dUpp[20] = 9.999999999990e+11;
dUpp[21] = 1.000000000000e+00;
dUpp[22] = 1.000000000000e+00;
dUpp[23] = 1.000000000000e+00;
dUpp[24] = 1.000000000000e+12;
dUpp[25] = 9.999999999996e+11;
dUpp[26] = 1.000000000000e+12;
dUpp[27] = 1.000000000000e+12;
dUpp[28] = 9.999999999993e+11;
dUpp[29] = 9.999999999999e+11;
dUpp[30] = 1.000000000000e+12;
dUpp[31] = 9.999999999990e+11;
dUpp[32] = 9.999999999995e+11;
dUpp[33] = 9.999999999998e+11;
dUpp[34] = 1.000000000000e+12;
dUpp[35] = 1.000000000000e+12;
dUpp[36] = 1.000000000000e+12;
dUpp[37] = 9.999999999996e+11;
dUpp[38] = 9.999999999997e+11;
dUpp[39] = 1.000000000000e+12;
dUpp[40] = 9.999999999998e+11;
dUpp[41] = 9.999999999999e+11;
dUpp[42] = 1.000000000000e+12;
dUpp[43] = 9.999999999990e+11;
dUpp[44] = 9.999999999990e+11;
dUpp[45] = 1.000000000000e+00;
dUpp[46] = 1.000000000000e+00;
dUpp[47] = 1.000000000000e+00;
dUpp[48] = 1.000000000000e+12;
dUpp[49] = 9.999999999997e+11;
dUpp[50] = 1.000000000000e+12;
dUpp[51] = 1.000000000000e+12;
dUpp[52] = 9.999999999993e+11;
dUpp[53] = 9.999999999999e+11;
dUpp[54] = 1.000000000000e+12;
dUpp[55] = 9.999999999991e+11;
dUpp[56] = 9.999999999996e+11;
dUpp[57] = 9.999999999997e+11;
dUpp[58] = 1.000000000000e+12;
dUpp[59] = 1.000000000000e+12;
dUpp[60] = 9.999999999995e+11;
dUpp[61] = 1.000000000001e+12;
dUpp[62] = 1.000000000001e+12;
dUpp[63] = 9.999999999992e+11;
dUpp[64] = 1.000000000001e+12;
dUpp[65] = 1.000000000001e+12;
dUpp[66] = 1.000000000000e+12;
dUpp[67] = 9.999999999990e+11;
dUpp[68] = 9.999999999990e+11;
dUpp[69] = 1.000000000000e+00;
dUpp[70] = 1.000000000000e+00;
dUpp[71] = 1.000000000000e+00;
dUpp[72] = 1.000000000000e+12;
dUpp[73] = 9.999999999997e+11;
dUpp[74] = 1.000000000000e+12;
dUpp[75] = 1.000000000000e+12;
dUpp[76] = 9.999999999995e+11;
dUpp[77] = 1.000000000000e+12;
dUpp[78] = 1.000000000000e+12;
dUpp[79] = 9.999999999993e+11;
dUpp[80] = 9.999999999997e+11;
dUpp[81] = 9.999999999997e+11;
dUpp[82] = 1.000000000000e+12;
dUpp[83] = 1.000000000000e+12;
dUpp[84] = 9.999999999988e+11;
dUpp[85] = 1.000000000001e+12;
dUpp[86] = 1.000000000001e+12;
dUpp[87] = 9.999999999992e+11;
dUpp[88] = 1.000000000001e+12;
dUpp[89] = 1.000000000001e+12;
dUpp[90] = 1.000000000000e+12;
dUpp[91] = 9.999999999990e+11;
dUpp[92] = 9.999999999990e+11;
dUpp[93] = 1.000000000000e+00;
dUpp[94] = 1.000000000000e+00;
dUpp[95] = 1.000000000000e+00;
dUpp[96] = 9.999999999999e+11;
dUpp[97] = 9.999999999998e+11;
dUpp[98] = 1.000000000000e+12;
dUpp[99] = 9.999999999998e+11;
dUpp[100] = 9.999999999997e+11;
dUpp[101] = 1.000000000000e+12;
dUpp[102] = 9.999999999999e+11;
dUpp[103] = 9.999999999994e+11;
dUpp[104] = 9.999999999998e+11;
dUpp[105] = 9.999999999992e+11;
dUpp[106] = 1.000000000001e+12;
dUpp[107] = 1.000000000001e+12;
dUpp[108] = 9.999999999993e+11;
dUpp[109] = 1.000000000000e+12;
dUpp[110] = 1.000000000000e+12;
dUpp[111] = 9.999999999997e+11;
dUpp[112] = 1.000000000000e+12;
dUpp[113] = 1.000000000000e+12;
dUpp[114] = 1.000000000000e+12;
dUpp[115] = 9.999999999990e+11;
dUpp[116] = 9.999999999990e+11;
dUpp[117] = 1.000000000000e+00;
dUpp[118] = 1.000000000000e+00;
dUpp[119] = 1.000000000000e+00;
dUpp[120] = 9.999999999998e+11;
dUpp[121] = 1.000000000000e+12;
dUpp[122] = 1.000000000000e+12;
dUpp[123] = 9.999999999998e+11;
dUpp[124] = 9.999999999997e+11;
dUpp[125] = 1.000000000000e+12;
dUpp[126] = 9.999999999999e+11;
dUpp[127] = 9.999999999994e+11;
dUpp[128] = 9.999999999998e+11;
dUpp[129] = 9.999999999997e+11;
dUpp[130] = 1.000000000000e+12;
dUpp[131] = 1.000000000000e+12;
dUpp[132] = 1.000000000000e+12;
dUpp[133] = 9.999999999998e+11;
dUpp[134] = 9.999999999997e+11;
dUpp[135] = 1.000000000000e+12;
dUpp[136] = 9.999999999996e+11;
dUpp[137] = 9.999999999998e+11;
dUpp[138] = 1.000000000000e+12;
dUpp[139] = 9.999999999990e+11;
dUpp[140] = 9.999999999990e+11;
dUpp[141] = 1.000000000000e+00;
dUpp[142] = 1.000000000000e+00;
dUpp[143] = 1.000000000000e+00;
dUpp[144] = 9.999999999999e+11;
dUpp[145] = 9.999999999999e+11;
dUpp[146] = 1.000000000000e+12;
dUpp[147] = 9.999999999999e+11;
dUpp[148] = 9.999999999996e+11;
dUpp[149] = 1.000000000000e+12;
dUpp[150] = 9.999999999999e+11;
dUpp[151] = 9.999999999993e+11;
dUpp[152] = 9.999999999998e+11;
dUpp[153] = 1.000000000001e+12;
dUpp[154] = 9.999999999989e+11;
dUpp[155] = 9.999999999989e+11;
dUpp[156] = 1.000000000001e+12;
dUpp[157] = 9.999999999993e+11;
dUpp[158] = 9.999999999993e+11;
dUpp[159] = 1.000000000000e+12;
dUpp[160] = 9.999999999995e+11;
dUpp[161] = 9.999999999995e+11;
dUpp[162] = 1.000000000000e+12;
dUpp[163] = 9.999999999990e+11;
dUpp[164] = 9.999999999990e+11;
dUpp[165] = 1.000000000000e+00;
dUpp[166] = 1.000000000000e+00;
dUpp[167] = 1.000000000000e+00;
dUpp[168] = 1.000000000000e+12;
dUpp[169] = 9.999999999996e+11;
dUpp[170] = 1.000000000000e+12;
dUpp[171] = 1.000000000000e+12;
dUpp[172] = 9.999999999994e+11;
dUpp[173] = 1.000000000000e+12;
dUpp[174] = 1.000000000000e+12;
dUpp[175] = 9.999999999992e+11;
dUpp[176] = 9.999999999997e+11;
dUpp[177] = 1.000000000001e+12;
dUpp[178] = 9.999999999992e+11;
dUpp[179] = 9.999999999992e+11;
dUpp[180] = 1.000000000001e+12;
dUpp[181] = 9.999999999988e+11;
dUpp[182] = 9.999999999989e+11;
dUpp[183] = 1.000000000000e+12;
dUpp[184] = 9.999999999997e+11;
dUpp[185] = 9.999999999996e+11;
dUpp[186] = 1.000000000000e+12;
dUpp[187] = 9.999999999990e+11;
dUpp[188] = 9.999999999990e+11;
dUpp[189] = 1.000000000000e+00;
dUpp[190] = 1.000000000000e+00;
dUpp[191] = 1.000000000000e+00;
dUpp[192] = 1.000000000000e+12;
dUpp[193] = 9.999999999996e+11;
dUpp[194] = 1.000000000000e+12;
dUpp[195] = 1.000000000000e+12;
dUpp[196] = 9.999999999992e+11;
dUpp[197] = 9.999999999999e+11;
dUpp[198] = 1.000000000000e+12;
dUpp[199] = 9.999999999991e+11;
dUpp[200] = 9.999999999996e+11;
dUpp[201] = 1.000000000000e+12;
dUpp[202] = 1.000000000000e+12;
dUpp[203] = 1.000000000000e+12;
dUpp[204] = 1.000000000001e+12;
dUpp[205] = 9.999999999996e+11;
dUpp[206] = 9.999999999996e+11;
dUpp[207] = 1.000000000000e+12;
dUpp[208] = 9.999999999995e+11;
dUpp[209] = 9.999999999995e+11;
dUpp[210] = 1.000000000000e+12;
dUpp[211] = 9.999999999990e+11;
dUpp[212] = 9.999999999990e+11;
dUpp[213] = 1.000000000000e+00;
dUpp[214] = 1.000000000000e+00;
dUpp[215] = 1.000000000000e+00;
dUpp[216] = 1.000000000000e+12;
dUpp[217] = 9.999999999996e+11;
dUpp[218] = 1.000000000000e+12;
dUpp[219] = 1.000000000000e+12;
dUpp[220] = 9.999999999993e+11;
dUpp[221] = 9.999999999999e+11;
dUpp[222] = 1.000000000000e+12;
dUpp[223] = 9.999999999990e+11;
dUpp[224] = 9.999999999995e+11;
dUpp[225] = 9.999999999996e+11;
dUpp[226] = 1.000000000000e+12;
dUpp[227] = 1.000000000000e+12;
dUpp[228] = 9.999999999995e+11;
dUpp[229] = 1.000000000001e+12;
dUpp[230] = 1.000000000001e+12;
dUpp[231] = 1.000000000000e+12;
dUpp[232] = 9.999999999999e+11;
dUpp[233] = 1.000000000000e+12;
dUpp[234] = 1.000000000000e+12;
dUpp[235] = 9.999999999990e+11;
dUpp[236] = 9.999999999990e+11;
dUpp[237] = 1.000000000000e+00;
dUpp[238] = 1.000000000000e+00;
dUpp[239] = 1.000000000000e+00;
dUpp[240] = 1.000000000000e+12;
dUpp[241] = 9.999999999997e+11;
dUpp[242] = 1.000000000000e+12;
dUpp[243] = 1.000000000000e+12;
dUpp[244] = 9.999999999994e+11;
dUpp[245] = 1.000000000000e+12;
dUpp[246] = 1.000000000000e+12;
dUpp[247] = 9.999999999992e+11;
dUpp[248] = 9.999999999997e+11;
dUpp[249] = 9.999999999996e+11;
dUpp[250] = 1.000000000001e+12;
dUpp[251] = 1.000000000000e+12;
dUpp[252] = 9.999999999989e+11;
dUpp[253] = 1.000000000001e+12;
dUpp[254] = 1.000000000001e+12;
dUpp[255] = 9.999999999992e+11;
dUpp[256] = 1.000000000001e+12;
dUpp[257] = 1.000000000001e+12;
dUpp[258] = 1.000000000000e+12;
dUpp[259] = 9.999999999990e+11;
dUpp[260] = 9.999999999990e+11;
dUpp[261] = 1.000000000000e+00;
dUpp[262] = 1.000000000000e+00;
dUpp[263] = 1.000000000000e+00;
dUpp[264] = 9.999999999999e+11;
dUpp[265] = 9.999999999998e+11;
dUpp[266] = 1.000000000000e+12;
dUpp[267] = 9.999999999999e+11;
dUpp[268] = 9.999999999996e+11;
dUpp[269] = 1.000000000000e+12;
dUpp[270] = 9.999999999999e+11;
dUpp[271] = 9.999999999994e+11;
dUpp[272] = 9.999999999999e+11;
dUpp[273] = 9.999999999995e+11;
dUpp[274] = 1.000000000000e+12;
dUpp[275] = 1.000000000001e+12;
dUpp[276] = 9.999999999993e+11;
dUpp[277] = 1.000000000001e+12;
dUpp[278] = 1.000000000001e+12;
dUpp[279] = 9.999999999989e+11;
dUpp[280] = 1.000000000001e+12;
dUpp[281] = 1.000000000001e+12;
dUpp[282] = 1.000000000000e+12;
dUpp[283] = 9.999999999990e+11;
dUpp[284] = 9.999999999990e+11;
dUpp[285] = 1.000000000000e+00;
dUpp[286] = 1.000000000000e+00;
dUpp[287] = 1.000000000000e+00;
dUpp[288] = 9.999999999999e+11;
dUpp[289] = 9.999999999999e+11;
dUpp[290] = 1.000000000000e+12;
dUpp[291] = 9.999999999998e+11;
dUpp[292] = 9.999999999997e+11;
dUpp[293] = 1.000000000000e+12;
dUpp[294] = 9.999999999998e+11;
dUpp[295] = 9.999999999994e+11;
dUpp[296] = 9.999999999999e+11;
dUpp[297] = 9.999999999997e+11;
dUpp[298] = 1.000000000000e+12;
dUpp[299] = 1.000000000000e+12;
dUpp[300] = 9.999999999997e+11;
dUpp[301] = 1.000000000000e+12;
dUpp[302] = 1.000000000000e+12;
dUpp[303] = 1.000000000000e+12;
dUpp[304] = 9.999999999996e+11;
dUpp[305] = 9.999999999996e+11;
dUpp[306] = 1.000000000000e+12;
dUpp[307] = 9.999999999990e+11;
dUpp[308] = 9.999999999990e+11;
dUpp[309] = 1.000000000000e+00;
dUpp[310] = 1.000000000000e+00;
dUpp[311] = 1.000000000000e+00;
dUpp[312] = 9.999999999999e+11;
dUpp[313] = 9.999999999999e+11;
dUpp[314] = 1.000000000000e+12;
dUpp[315] = 9.999999999998e+11;
dUpp[316] = 9.999999999997e+11;
dUpp[317] = 1.000000000000e+12;
dUpp[318] = 9.999999999999e+11;
dUpp[319] = 9.999999999993e+11;
dUpp[320] = 9.999999999998e+11;
dUpp[321] = 1.000000000000e+12;
dUpp[322] = 9.999999999998e+11;
dUpp[323] = 9.999999999996e+11;
dUpp[324] = 1.000000000000e+12;
dUpp[325] = 9.999999999991e+11;
dUpp[326] = 9.999999999991e+11;
dUpp[327] = 1.000000000001e+12;
dUpp[328] = 9.999999999994e+11;
dUpp[329] = 9.999999999995e+11;
dUpp[330] = 1.000000000000e+12;
dUpp[331] = 9.999999999990e+11;
dUpp[332] = 9.999999999990e+11;
dUpp[333] = 1.000000000000e+00;
dUpp[334] = 1.000000000000e+00;
dUpp[335] = 1.000000000000e+00;
dUpp[336] = 1.000000000000e+12;
dUpp[337] = 9.999999999998e+11;
dUpp[338] = 1.000000000000e+12;
dUpp[339] = 1.000000000000e+12;
dUpp[340] = 9.999999999994e+11;
dUpp[341] = 1.000000000000e+12;
dUpp[342] = 1.000000000000e+12;
dUpp[343] = 9.999999999992e+11;
dUpp[344] = 9.999999999997e+11;
dUpp[345] = 1.000000000001e+12;
dUpp[346] = 9.999999999990e+11;
dUpp[347] = 9.999999999991e+11;
dUpp[348] = 1.000000000001e+12;
dUpp[349] = 9.999999999989e+11;
dUpp[350] = 9.999999999989e+11;
dUpp[351] = 1.000000000000e+12;
dUpp[352] = 9.999999999997e+11;
dUpp[353] = 9.999999999997e+11;
dUpp[354] = 1.000000000000e+12;
dUpp[355] = 9.999999999990e+11;
dUpp[356] = 9.999999999990e+11;
dUpp[357] = 1.000000000000e+00;
dUpp[358] = 1.000000000000e+00;
dUpp[359] = 1.000000000000e+00;
dUpp[360] = 1.000000000000e+12;
dUpp[361] = 9.999999999996e+11;
dUpp[362] = 1.000000000000e+12;
dUpp[363] = 1.000000000000e+12;
dUpp[364] = 9.999999999993e+11;
dUpp[365] = 9.999999999999e+11;
dUpp[366] = 1.000000000000e+12;
dUpp[367] = 9.999999999991e+11;
dUpp[368] = 9.999999999996e+11;
dUpp[369] = 1.000000000001e+12;
dUpp[370] = 9.999999999992e+11;
dUpp[371] = 9.999999999994e+11;
dUpp[372] = 1.000000000001e+12;
dUpp[373] = 9.999999999996e+11;
dUpp[374] = 9.999999999997e+11;
dUpp[375] = 1.000000000000e+12;
dUpp[376] = 9.999999999996e+11;
dUpp[377] = 9.999999999995e+11;
dUpp[378] = 1.000000000000e+12;
dUpp[379] = 9.999999999990e+11;
dUpp[380] = 9.999999999990e+11;
dUpp[381] = 1.000000000000e+00;
dUpp[382] = 1.000000000000e+00;
dUpp[383] = 1.000000000000e+00;
dUpp[384] = 1.000000000000e+12;
dUpp[385] = 9.999999999995e+11;
dUpp[386] = 1.000000000000e+12;
dUpp[387] = 1.000000000000e+12;
dUpp[388] = 9.999999999993e+11;
dUpp[389] = 9.999999999999e+11;
dUpp[390] = 1.000000000000e+12;
dUpp[391] = 9.999999999991e+11;
dUpp[392] = 9.999999999995e+11;
dUpp[393] = 9.999999999999e+11;
dUpp[394] = 1.000000000001e+12;
dUpp[395] = 1.000000000001e+12;
dUpp[396] = 9.999999999998e+11;
dUpp[397] = 1.000000000000e+12;
dUpp[398] = 1.000000000000e+12;
dUpp[399] = 1.000000000000e+12;
dUpp[400] = 9.999999999999e+11;
dUpp[401] = 1.000000000000e+12;
dUpp[402] = 1.000000000000e+12;
dUpp[403] = 9.999999999990e+11;
dUpp[404] = 9.999999999990e+11;
dUpp[405] = 1.000000000000e+00;
dUpp[406] = 1.000000000000e+00;
dUpp[407] = 1.000000000000e+00;
dUpp[408] = 1.000000000000e+12;
dUpp[409] = 9.999999999997e+11;
dUpp[410] = 1.000000000000e+12;
dUpp[411] = 1.000000000000e+12;
dUpp[412] = 9.999999999994e+11;
dUpp[413] = 1.000000000000e+12;
dUpp[414] = 1.000000000000e+12;
dUpp[415] = 9.999999999991e+11;
dUpp[416] = 9.999999999996e+11;
dUpp[417] = 9.999999999990e+11;
dUpp[418] = 1.000000000001e+12;
dUpp[419] = 1.000000000001e+12;
dUpp[420] = 9.999999999994e+11;
dUpp[421] = 1.000000000001e+12;
dUpp[422] = 1.000000000001e+12;
dUpp[423] = 9.999999999996e+11;
dUpp[424] = 1.000000000001e+12;
dUpp[425] = 1.000000000001e+12;
dUpp[426] = 1.000000000000e+12;
dUpp[427] = 9.999999999990e+11;
dUpp[428] = 9.999999999990e+11;
dUpp[429] = 1.000000000000e+00;
dUpp[430] = 1.000000000000e+00;
dUpp[431] = 1.000000000000e+00;
dUpp[432] = 9.999999999999e+11;
dUpp[433] = 9.999999999998e+11;
dUpp[434] = 1.000000000000e+12;
dUpp[435] = 1.000000000000e+12;
dUpp[436] = 9.999999999996e+11;
dUpp[437] = 1.000000000000e+12;
dUpp[438] = 1.000000000000e+12;
dUpp[439] = 9.999999999993e+11;
dUpp[440] = 9.999999999998e+11;
dUpp[441] = 9.999999999995e+11;
dUpp[442] = 1.000000000000e+12;
dUpp[443] = 1.000000000000e+12;
dUpp[444] = 9.999999999990e+11;
dUpp[445] = 1.000000000001e+12;
dUpp[446] = 1.000000000001e+12;
dUpp[447] = 9.999999999991e+11;
dUpp[448] = 1.000000000001e+12;
dUpp[449] = 1.000000000001e+12;
dUpp[450] = 1.000000000000e+12;
dUpp[451] = 9.999999999990e+11;
dUpp[452] = 9.999999999990e+11;
dUpp[453] = 1.000000000000e+00;
dUpp[454] = 1.000000000000e+00;
dUpp[455] = 1.000000000000e+00;
dUpp[456] = 9.999999999999e+11;
dUpp[457] = 9.999999999999e+11;
dUpp[458] = 1.000000000000e+12;
dUpp[459] = 9.999999999998e+11;
dUpp[460] = 9.999999999997e+11;
dUpp[461] = 1.000000000000e+12;
dUpp[462] = 9.999999999998e+11;
dUpp[463] = 9.999999999994e+11;
dUpp[464] = 9.999999999999e+11;
dUpp[465] = 1.000000000000e+12;
dUpp[466] = 1.000000000000e+12;
dUpp[467] = 1.000000000000e+12;
dUpp[468] = 9.999999999993e+11;
dUpp[469] = 1.000000000000e+12;
dUpp[470] = 1.000000000000e+12;
dUpp[471] = 9.999999999995e+11;
dUpp[472] = 1.000000000000e+12;
dUpp[473] = 1.000000000000e+12;
dUpp[474] = 1.000000000000e+12;
dUpp[475] = 9.999999999990e+11;
dUpp[476] = 9.999999999990e+11;
dUpp[477] = 1.000000000000e+00;
dUpp[478] = 1.000000000000e+00;
dUpp[479] = 1.000000000000e+00;
dUpp[480] = 9.999999999999e+11;
dUpp[481] = 9.999999999999e+11;
dUpp[482] = 1.000000000000e+12;
dUpp[483] = 9.999999999998e+11;
dUpp[484] = 9.999999999997e+11;
dUpp[485] = 1.000000000000e+12;
dUpp[486] = 9.999999999998e+11;
dUpp[487] = 9.999999999994e+11;
dUpp[488] = 9.999999999999e+11;
dUpp[489] = 9.999999999999e+11;
dUpp[490] = 1.000000000000e+12;
dUpp[491] = 1.000000000000e+12;
dUpp[492] = 1.000000000001e+12;
dUpp[493] = 9.999999999994e+11;
dUpp[494] = 9.999999999992e+11;
dUpp[495] = 1.000000000000e+12;
dUpp[496] = 9.999999999993e+11;
dUpp[497] = 9.999999999994e+11;
dUpp[498] = 1.000000000000e+12;
dUpp[499] = 9.999999999990e+11;
dUpp[500] = 9.999999999990e+11;
dUpp[501] = 1.000000000000e+00;
dUpp[502] = 1.000000000000e+00;
dUpp[503] = 1.000000000000e+00;
dUpp[504] = 9.999999999999e+11;
dUpp[505] = 9.999999999998e+11;
dUpp[506] = 1.000000000000e+12;
dUpp[507] = 9.999999999999e+11;
dUpp[508] = 9.999999999995e+11;
dUpp[509] = 1.000000000000e+12;
dUpp[510] = 9.999999999999e+11;
dUpp[511] = 9.999999999992e+11;
dUpp[512] = 9.999999999997e+11;
dUpp[513] = 1.000000000000e+12;
dUpp[514] = 9.999999999992e+11;
dUpp[515] = 9.999999999992e+11;
dUpp[516] = 1.000000000001e+12;
dUpp[517] = 9.999999999992e+11;
dUpp[518] = 9.999999999991e+11;
dUpp[519] = 1.000000000001e+12;
dUpp[520] = 9.999999999990e+11;
dUpp[521] = 9.999999999991e+11;
dUpp[522] = 1.000000000000e+12;
dUpp[523] = 9.999999999990e+11;
dUpp[524] = 9.999999999990e+11;
dUpp[525] = 1.000000000000e+00;
dUpp[526] = 1.000000000000e+00;
dUpp[527] = 1.000000000000e+00;
dUpp[528] = 1.000000000000e+12;
dUpp[529] = 9.999999999996e+11;
dUpp[530] = 1.000000000000e+12;
dUpp[531] = 1.000000000000e+12;
dUpp[532] = 9.999999999994e+11;
dUpp[533] = 1.000000000000e+12;
dUpp[534] = 1.000000000000e+12;
dUpp[535] = 9.999999999991e+11;
dUpp[536] = 9.999999999996e+11;
dUpp[537] = 1.000000000001e+12;
dUpp[538] = 9.999999999991e+11;
dUpp[539] = 9.999999999992e+11;
dUpp[540] = 1.000000000001e+12;
dUpp[541] = 9.999999999991e+11;
dUpp[542] = 9.999999999993e+11;
dUpp[543] = 1.000000000001e+12;
dUpp[544] = 1.000000000000e+12;
dUpp[545] = 9.999999999998e+11;
dUpp[546] = 1.000000000000e+12;
dUpp[547] = 9.999999999990e+11;
dUpp[548] = 9.999999999990e+11;
dUpp[549] = 1.000000000000e+00;
dUpp[550] = 1.000000000000e+00;
dUpp[551] = 1.000000000000e+00;
dUpp[552] = 1.000000000000e+12;
dUpp[553] = 9.999999999995e+11;
dUpp[554] = 1.000000000000e+12;
dUpp[555] = 1.000000000000e+12;
dUpp[556] = 9.999999999992e+11;
dUpp[557] = 9.999999999999e+11;
dUpp[558] = 1.000000000000e+12;
dUpp[559] = 9.999999999991e+11;
dUpp[560] = 9.999999999996e+11;
dUpp[561] = 1.000000000000e+12;
dUpp[562] = 1.000000000000e+12;
dUpp[563] = 1.000000000000e+12;
dUpp[564] = 1.000000000000e+12;
dUpp[565] = 9.999999999999e+11;
dUpp[566] = 9.999999999999e+11;
dUpp[567] = 9.999999999998e+11;
dUpp[568] = 1.000000000000e+12;
dUpp[569] = 1.000000000000e+12;
dUpp[570] = 1.000000000000e+12;
dUpp[571] = 9.999999999990e+11;
dUpp[572] = 9.999999999990e+11;
dUpp[573] = 1.000000000000e+00;
dUpp[574] = 1.000000000000e+00;
dUpp[575] = 1.000000000000e+00;
dUpp[576] = 1.000000000000e+12;
dUpp[577] = 9.999999999996e+11;
dUpp[578] = 1.000000000000e+12;
dUpp[579] = 1.000000000000e+12;
dUpp[580] = 9.999999999993e+11;
dUpp[581] = 1.000000000000e+12;
dUpp[582] = 1.000000000000e+12;
dUpp[583] = 9.999999999991e+11;
dUpp[584] = 9.999999999996e+11;
dUpp[585] = 9.999999999993e+11;
dUpp[586] = 1.000000000001e+12;
dUpp[587] = 1.000000000001e+12;
dUpp[588] = 9.999999999996e+11;
dUpp[589] = 1.000000000001e+12;
dUpp[590] = 1.000000000001e+12;
dUpp[591] = 9.999999999996e+11;
dUpp[592] = 1.000000000000e+12;
dUpp[593] = 1.000000000000e+12;
dUpp[594] = 1.000000000000e+12;
dUpp[595] = 9.999999999990e+11;
dUpp[596] = 9.999999999990e+11;
dUpp[597] = 1.000000000000e+00;
dUpp[598] = 1.000000000000e+00;
dUpp[599] = 1.000000000000e+00;
dUpp[600] = 1.000000000000e+12;
dUpp[601] = 9.999999999998e+11;
dUpp[602] = 1.000000000000e+12;
dUpp[603] = 1.000000000000e+12;
dUpp[604] = 9.999999999995e+11;
dUpp[605] = 1.000000000000e+12;
dUpp[606] = 1.000000000000e+12;
dUpp[607] = 9.999999999992e+11;
dUpp[608] = 9.999999999997e+11;
dUpp[609] = 9.999999999991e+11;
dUpp[610] = 1.000000000001e+12;
dUpp[611] = 1.000000000001e+12;
dUpp[612] = 9.999999999987e+11;
dUpp[613] = 1.000000000001e+12;
dUpp[614] = 1.000000000001e+12;
dUpp[615] = 9.999999999998e+11;
dUpp[616] = 1.000000000001e+12;
dUpp[617] = 1.000000000001e+12;
dUpp[618] = 1.000000000000e+12;
dUpp[619] = 9.999999999990e+11;
dUpp[620] = 9.999999999990e+11;
dUpp[621] = 1.000000000000e+00;
dUpp[622] = 1.000000000000e+00;
dUpp[623] = 1.000000000000e+00;
dUpp[624] = 9.999999999999e+11;
dUpp[625] = 9.999999999999e+11;
dUpp[626] = 1.000000000000e+12;
dUpp[627] = 9.999999999998e+11;
dUpp[628] = 9.999999999997e+11;
dUpp[629] = 1.000000000000e+12;
dUpp[630] = 9.999999999999e+11;
dUpp[631] = 9.999999999994e+11;
dUpp[632] = 9.999999999999e+11;
dUpp[633] = 9.999999999996e+11;
dUpp[634] = 1.000000000000e+12;
dUpp[635] = 1.000000000000e+12;
dUpp[636] = 9.999999999993e+11;
dUpp[637] = 1.000000000000e+12;
dUpp[638] = 1.000000000000e+12;
dUpp[639] = 9.999999999993e+11;
dUpp[640] = 1.000000000001e+12;
dUpp[641] = 1.000000000001e+12;
dUpp[642] = 1.000000000000e+12;
dUpp[643] = 9.999999999990e+11;
dUpp[644] = 9.999999999990e+11;
dUpp[645] = 1.000000000000e+00;
dUpp[646] = 1.000000000000e+00;
dUpp[647] = 1.000000000000e+00;
dUpp[648] = 9.999999999998e+11;
dUpp[649] = 9.999999999999e+11;
dUpp[650] = 1.000000000000e+12;
dUpp[651] = 9.999999999998e+11;
dUpp[652] = 9.999999999997e+11;
dUpp[653] = 1.000000000000e+12;
dUpp[654] = 9.999999999998e+11;
dUpp[655] = 9.999999999995e+11;
dUpp[656] = 9.999999999999e+11;
dUpp[657] = 1.000000000000e+12;
dUpp[658] = 9.999999999996e+11;
dUpp[659] = 9.999999999996e+11;
dUpp[660] = 1.000000000000e+12;
dUpp[661] = 9.999999999999e+11;
dUpp[662] = 9.999999999998e+11;
dUpp[663] = 9.999999999995e+11;
dUpp[664] = 9.999999999996e+11;
dUpp[665] = 9.999999999997e+11;
dUpp[666] = 1.000000000000e+12;
dUpp[667] = 9.999999999990e+11;
dUpp[668] = 9.999999999990e+11;
dUpp[669] = 1.000000000000e+00;
dUpp[670] = 1.000000000000e+00;
dUpp[671] = 1.000000000000e+00;
dUpp[672] = 9.999999999999e+11;
dUpp[673] = 9.999999999998e+11;
dUpp[674] = 1.000000000000e+12;
dUpp[675] = 9.999999999999e+11;
dUpp[676] = 9.999999999996e+11;
dUpp[677] = 1.000000000000e+12;
dUpp[678] = 9.999999999998e+11;
dUpp[679] = 9.999999999993e+11;
dUpp[680] = 9.999999999998e+11;
dUpp[681] = 1.000000000001e+12;
dUpp[682] = 9.999999999998e+11;
dUpp[683] = 9.999999999996e+11;
dUpp[684] = 1.000000000001e+12;
dUpp[685] = 9.999999999990e+11;
dUpp[686] = 9.999999999990e+11;
dUpp[687] = 1.000000000001e+12;
dUpp[688] = 9.999999999990e+11;
dUpp[689] = 9.999999999990e+11;
dUpp[690] = 1.000000000000e+12;
dUpp[691] = 9.999999999990e+11;
dUpp[692] = 9.999999999990e+11;
dUpp[693] = 1.000000000000e+00;
dUpp[694] = 1.000000000000e+00;
dUpp[695] = 1.000000000000e+00;
dUpp[696] = 1.000000000000e+12;
dUpp[697] = 9.999999999997e+11;
dUpp[698] = 1.000000000000e+12;
dUpp[699] = 1.000000000000e+12;
dUpp[700] = 9.999999999994e+11;
dUpp[701] = 1.000000000000e+12;
dUpp[702] = 1.000000000000e+12;
dUpp[703] = 9.999999999991e+11;
dUpp[704] = 9.999999999996e+11;
dUpp[705] = 1.000000000000e+12;
dUpp[706] = 9.999999999995e+11;
dUpp[707] = 9.999999999995e+11;
dUpp[708] = 1.000000000001e+12;
dUpp[709] = 9.999999999989e+11;
dUpp[710] = 9.999999999990e+11;
dUpp[711] = 1.000000000001e+12;
dUpp[712] = 9.999999999995e+11;
dUpp[713] = 9.999999999995e+11;
dUpp[714] = 1.000000000000e+12;
dUpp[715] = 9.999999999990e+11;
dUpp[716] = 9.999999999990e+11;
dUpp[717] = 1.000000000000e+00;
dUpp[718] = 1.000000000000e+00;
dUpp[719] = 1.000000000000e+00;
dUpp[720] = 1.000000000000e+12;
dUpp[721] = 9.999999999996e+11;
dUpp[722] = 1.000000000000e+12;
dUpp[723] = 1.000000000000e+12;
dUpp[724] = 9.999999999993e+11;
dUpp[725] = 9.999999999999e+11;
dUpp[726] = 1.000000000000e+12;
dUpp[727] = 9.999999999991e+11;
dUpp[728] = 9.999999999996e+11;
dUpp[729] = 1.000000000000e+12;
dUpp[730] = 9.999999999993e+11;
dUpp[731] = 9.999999999995e+11;
dUpp[732] = 1.000000000001e+12;
dUpp[733] = 1.000000000000e+12;
dUpp[734] = 1.000000000000e+12;
dUpp[735] = 1.000000000000e+12;
dUpp[736] = 1.000000000000e+12;
dUpp[737] = 1.000000000000e+12;
dUpp[738] = 1.000000000000e+12;
dUpp[739] = 9.999999999990e+11;
dUpp[740] = 9.999999999990e+11;
dUpp[741] = 1.000000000000e+00;
dUpp[742] = 1.000000000000e+00;
dUpp[743] = 1.000000000000e+00;
dUpp[744] = 1.000000000000e+12;
dUpp[745] = 9.999999999995e+11;
dUpp[746] = 1.000000000000e+12;
dUpp[747] = 1.000000000000e+12;
dUpp[748] = 9.999999999993e+11;
dUpp[749] = 1.000000000000e+12;
dUpp[750] = 1.000000000000e+12;
dUpp[751] = 9.999999999991e+11;
dUpp[752] = 9.999999999996e+11;
dUpp[753] = 1.000000000000e+12;
dUpp[754] = 1.000000000001e+12;
dUpp[755] = 1.000000000000e+12;
dUpp[756] = 9.999999999995e+11;
dUpp[757] = 1.000000000001e+12;
dUpp[758] = 1.000000000001e+12;
dUpp[759] = 9.999999999996e+11;
dUpp[760] = 1.000000000000e+12;
dUpp[761] = 1.000000000000e+12;
dUpp[762] = 1.000000000000e+12;
dUpp[763] = 9.999999999990e+11;
dUpp[764] = 9.999999999990e+11;
dUpp[765] = 1.000000000000e+00;
dUpp[766] = 1.000000000000e+00;
dUpp[767] = 1.000000000000e+00;
dUpp[768] = 1.000000000000e+12;
dUpp[769] = 9.999999999998e+11;
dUpp[770] = 1.000000000000e+12;
dUpp[771] = 1.000000000000e+12;
dUpp[772] = 9.999999999995e+11;
dUpp[773] = 1.000000000000e+12;
dUpp[774] = 1.000000000000e+12;
dUpp[775] = 9.999999999992e+11;
dUpp[776] = 9.999999999997e+11;
dUpp[777] = 9.999999999993e+11;
dUpp[778] = 1.000000000001e+12;
dUpp[779] = 1.000000000001e+12;
dUpp[780] = 9.999999999990e+11;
dUpp[781] = 1.000000000001e+12;
dUpp[782] = 1.000000000001e+12;
dUpp[783] = 9.999999999995e+11;
dUpp[784] = 1.000000000000e+12;
dUpp[785] = 1.000000000001e+12;
dUpp[786] = 1.000000000000e+12;
dUpp[787] = 9.999999999990e+11;
dUpp[788] = 9.999999999990e+11;
dUpp[789] = 1.000000000000e+00;
dUpp[790] = 1.000000000000e+00;
dUpp[791] = 1.000000000000e+00;
dUpp[792] = 9.999999999999e+11;
dUpp[793] = 9.999999999999e+11;
dUpp[794] = 1.000000000000e+12;
dUpp[795] = 9.999999999999e+11;
dUpp[796] = 9.999999999996e+11;
dUpp[797] = 1.000000000000e+12;
dUpp[798] = 9.999999999999e+11;
dUpp[799] = 9.999999999993e+11;
dUpp[800] = 9.999999999998e+11;
dUpp[801] = 9.999999999989e+11;
dUpp[802] = 1.000000000000e+12;
dUpp[803] = 1.000000000000e+12;
dUpp[804] = 9.999999999993e+11;
dUpp[805] = 1.000000000001e+12;
dUpp[806] = 1.000000000001e+12;
dUpp[807] = 9.999999999995e+11;
dUpp[808] = 1.000000000000e+12;
dUpp[809] = 1.000000000000e+12;
dUpp[810] = 1.000000000000e+12;
dUpp[811] = 9.999999999990e+11;
dUpp[812] = 9.999999999990e+11;
dUpp[813] = 1.000000000000e+00;
dUpp[814] = 1.000000000000e+00;
dUpp[815] = 1.000000000000e+00;
dUpp[816] = 9.999999999998e+11;
dUpp[817] = 9.999999999999e+11;
dUpp[818] = 1.000000000000e+12;
dUpp[819] = 9.999999999998e+11;
dUpp[820] = 9.999999999998e+11;
dUpp[821] = 1.000000000000e+12;
dUpp[822] = 9.999999999999e+11;
dUpp[823] = 9.999999999994e+11;
dUpp[824] = 9.999999999999e+11;
dUpp[825] = 1.000000000000e+12;
dUpp[826] = 9.999999999996e+11;
dUpp[827] = 9.999999999995e+11;
dUpp[828] = 9.999999999996e+11;
dUpp[829] = 9.999999999999e+11;
dUpp[830] = 9.999999999999e+11;
dUpp[831] = 9.999999999998e+11;
dUpp[832] = 1.000000000000e+12;
dUpp[833] = 1.000000000000e+12;
dUpp[834] = 1.000000000000e+12;
dUpp[835] = 9.999999999990e+11;
dUpp[836] = 9.999999999990e+11;
dUpp[837] = 1.000000000000e+00;
dUpp[838] = 1.000000000000e+00;
dUpp[839] = 1.000000000000e+00;
dUpp[840] = 9.999999999999e+11;
dUpp[841] = 9.999999999998e+11;
dUpp[842] = 1.000000000000e+12;
dUpp[843] = 9.999999999998e+11;
dUpp[844] = 9.999999999996e+11;
dUpp[845] = 1.000000000000e+12;
dUpp[846] = 9.999999999999e+11;
dUpp[847] = 9.999999999994e+11;
dUpp[848] = 9.999999999998e+11;
dUpp[849] = 1.000000000001e+12;
dUpp[850] = 9.999999999996e+11;
dUpp[851] = 9.999999999995e+11;
dUpp[852] = 1.000000000000e+12;
dUpp[853] = 9.999999999990e+11;
dUpp[854] = 9.999999999989e+11;
dUpp[855] = 1.000000000000e+12;
dUpp[856] = 9.999999999995e+11;
dUpp[857] = 9.999999999995e+11;
dUpp[858] = 1.000000000000e+12;
dUpp[859] = 9.999999999990e+11;
dUpp[860] = 9.999999999990e+11;
dUpp[861] = 1.000000000000e+00;
dUpp[862] = 1.000000000000e+00;
dUpp[863] = 1.000000000000e+00;
dUpp[864] = 1.000000000000e+12;
dUpp[865] = 9.999999999997e+11;
dUpp[866] = 1.000000000000e+12;
dUpp[867] = 1.000000000000e+12;
dUpp[868] = 9.999999999994e+11;
dUpp[869] = 1.000000000000e+12;
dUpp[870] = 9.999999999999e+11;
dUpp[871] = 9.999999999992e+11;
dUpp[872] = 9.999999999997e+11;
dUpp[873] = 1.000000000000e+12;
dUpp[874] = 9.999999999994e+11;
dUpp[875] = 9.999999999996e+11;
dUpp[876] = 1.000000000001e+12;
dUpp[877] = 9.999999999992e+11;
dUpp[878] = 9.999999999991e+11;
dUpp[879] = 1.000000000001e+12;
dUpp[880] = 9.999999999987e+11;
dUpp[881] = 9.999999999989e+11;
dUpp[882] = 1.000000000000e+12;
dUpp[883] = 9.999999999990e+11;
dUpp[884] = 9.999999999990e+11;
dUpp[885] = 1.000000000000e+00;
dUpp[886] = 1.000000000000e+00;
dUpp[887] = 1.000000000000e+00;
dUpp[888] = 1.000000000000e+12;
dUpp[889] = 9.999999999996e+11;
dUpp[890] = 1.000000000000e+12;
dUpp[891] = 1.000000000000e+12;
dUpp[892] = 9.999999999993e+11;
dUpp[893] = 9.999999999999e+11;
dUpp[894] = 1.000000000000e+12;
dUpp[895] = 9.999999999990e+11;
dUpp[896] = 9.999999999995e+11;
dUpp[897] = 1.000000000000e+12;
dUpp[898] = 9.999999999996e+11;
dUpp[899] = 9.999999999997e+11;
dUpp[900] = 1.000000000001e+12;
dUpp[901] = 9.999999999994e+11;
dUpp[902] = 9.999999999997e+11;
dUpp[903] = 1.000000000001e+12;
dUpp[904] = 9.999999999998e+11;
dUpp[905] = 9.999999999997e+11;
dUpp[906] = 1.000000000000e+12;
dUpp[907] = 9.999999999990e+11;
dUpp[908] = 9.999999999990e+11;
dUpp[909] = 1.000000000000e+00;
dUpp[910] = 1.000000000000e+00;
dUpp[911] = 1.000000000000e+00;
dUpp[912] = 1.000000000000e+12;
dUpp[913] = 9.999999999996e+11;
dUpp[914] = 1.000000000000e+12;
dUpp[915] = 1.000000000000e+12;
dUpp[916] = 9.999999999992e+11;
dUpp[917] = 9.999999999999e+11;
dUpp[918] = 1.000000000000e+12;
dUpp[919] = 9.999999999991e+11;
dUpp[920] = 9.999999999996e+11;
dUpp[921] = 1.000000000000e+12;
dUpp[922] = 1.000000000000e+12;
dUpp[923] = 1.000000000000e+12;
dUpp[924] = 9.999999999999e+11;
dUpp[925] = 1.000000000000e+12;
dUpp[926] = 1.000000000000e+12;
dUpp[927] = 1.000000000000e+12;
dUpp[928] = 1.000000000001e+12;
dUpp[929] = 1.000000000001e+12;
dUpp[930] = 1.000000000000e+12;
dUpp[931] = 9.999999999990e+11;
dUpp[932] = 9.999999999990e+11;
dUpp[933] = 1.000000000000e+00;
dUpp[934] = 1.000000000000e+00;
dUpp[935] = 1.000000000000e+00;
dUpp[936] = 1.000000000000e+12;
dUpp[937] = 9.999999999997e+11;
dUpp[938] = 1.000000000000e+12;
dUpp[939] = 1.000000000000e+12;
dUpp[940] = 9.999999999994e+11;
dUpp[941] = 1.000000000000e+12;
dUpp[942] = 1.000000000000e+12;
dUpp[943] = 9.999999999992e+11;
dUpp[944] = 9.999999999997e+11;
dUpp[945] = 9.999999999997e+11;
dUpp[946] = 1.000000000001e+12;
dUpp[947] = 1.000000000001e+12;
dUpp[948] = 9.999999999995e+11;
dUpp[949] = 1.000000000001e+12;
dUpp[950] = 1.000000000001e+12;
dUpp[951] = 9.999999999990e+11;
dUpp[952] = 1.000000000000e+12;
dUpp[953] = 1.000000000001e+12;
dUpp[954] = 1.000000000000e+12;
dUpp[955] = 9.999999999990e+11;
dUpp[956] = 9.999999999990e+11;
dUpp[957] = 1.000000000000e+00;
dUpp[958] = 1.000000000000e+00;
dUpp[959] = 1.000000000000e+00;
dUpp[960] = 1.000000000000e+12;
dUpp[961] = 9.999999999998e+11;
dUpp[962] = 1.000000000000e+12;
dUpp[963] = 1.000000000000e+12;
dUpp[964] = 9.999999999996e+11;
dUpp[965] = 1.000000000000e+12;
dUpp[966] = 9.999999999999e+11;
dUpp[967] = 9.999999999993e+11;
dUpp[968] = 9.999999999998e+11;
dUpp[969] = 9.999999999992e+11;
dUpp[970] = 1.000000000001e+12;
dUpp[971] = 1.000000000001e+12;
dUpp[972] = 9.999999999990e+11;
dUpp[973] = 1.000000000001e+12;
dUpp[974] = 1.000000000001e+12;
dUpp[975] = 9.999999999994e+11;
dUpp[976] = 1.000000000000e+12;
dUpp[977] = 1.000000000000e+12;
dUpp[978] = 1.000000000000e+12;
dUpp[979] = 9.999999999990e+11;
dUpp[980] = 9.999999999990e+11;
dUpp[981] = 1.000000000000e+00;
dUpp[982] = 1.000000000000e+00;
dUpp[983] = 1.000000000000e+00;
dUpp[984] = 9.999999999999e+11;
dUpp[985] = 1.000000000000e+12;
dUpp[986] = 1.000000000000e+12;
dUpp[987] = 9.999999999998e+11;
dUpp[988] = 9.999999999997e+11;
dUpp[989] = 1.000000000000e+12;
dUpp[990] = 9.999999999999e+11;
dUpp[991] = 9.999999999994e+11;
dUpp[992] = 9.999999999998e+11;
dUpp[993] = 9.999999999994e+11;
dUpp[994] = 1.000000000000e+12;
dUpp[995] = 1.000000000000e+12;
dUpp[996] = 9.999999999992e+11;
dUpp[997] = 1.000000000000e+12;
dUpp[998] = 9.999999999999e+11;
dUpp[999] = 1.000000000000e+12;
dUpp[1000] = 1.000000000000e+12;
dUpp[1001] = 1.000000000000e+12;
dUpp[1002] = 1.000000000000e+12;
dUpp[1003] = 9.999999999990e+11;
dUpp[1004] = 9.999999999990e+11;
dUpp[1005] = 1.000000000000e+00;
dUpp[1006] = 1.000000000000e+00;
dUpp[1007] = 1.000000000000e+00;
dUpp[1008] = 9.999999999998e+11;
dUpp[1009] = 9.999999999999e+11;
dUpp[1010] = 1.000000000000e+12;
dUpp[1011] = 9.999999999998e+11;
dUpp[1012] = 9.999999999997e+11;
dUpp[1013] = 1.000000000000e+12;
dUpp[1014] = 9.999999999999e+11;
dUpp[1015] = 9.999999999994e+11;
dUpp[1016] = 9.999999999998e+11;
dUpp[1017] = 1.000000000000e+12;
dUpp[1018] = 9.999999999990e+11;
dUpp[1019] = 9.999999999990e+11;
dUpp[1020] = 1.000000000000e+12;
dUpp[1021] = 9.999999999996e+11;
dUpp[1022] = 9.999999999993e+11;
dUpp[1023] = 9.999999999999e+11;
dUpp[1024] = 9.999999999997e+11;
dUpp[1025] = 9.999999999998e+11;
dUpp[1026] = 1.000000000000e+12;
dUpp[1027] = 9.999999999990e+11;
dUpp[1028] = 9.999999999990e+11;
dUpp[1029] = 1.000000000000e+00;
dUpp[1030] = 1.000000000000e+00;
dUpp[1031] = 1.000000000000e+00;
dUpp[1032] = 9.999999999999e+11;
dUpp[1033] = 9.999999999997e+11;
dUpp[1034] = 1.000000000000e+12;
dUpp[1035] = 9.999999999999e+11;
dUpp[1036] = 9.999999999995e+11;
dUpp[1037] = 1.000000000000e+12;
dUpp[1038] = 9.999999999999e+11;
dUpp[1039] = 9.999999999993e+11;
dUpp[1040] = 9.999999999997e+11;
dUpp[1041] = 1.000000000001e+12;
dUpp[1042] = 9.999999999993e+11;
dUpp[1043] = 9.999999999993e+11;
dUpp[1044] = 1.000000000001e+12;
dUpp[1045] = 9.999999999990e+11;
dUpp[1046] = 9.999999999991e+11;
dUpp[1047] = 1.000000000000e+12;
dUpp[1048] = 9.999999999991e+11;
dUpp[1049] = 9.999999999992e+11;
dUpp[1050] = 1.000000000000e+12;
dUpp[1051] = 9.999999999990e+11;
dUpp[1052] = 9.999999999990e+11;
dUpp[1053] = 1.000000000000e+00;
dUpp[1054] = 1.000000000000e+00;
dUpp[1055] = 1.000000000000e+00;
dUpp[1056] = 1.000000000000e+12;
dUpp[1057] = 9.999999999996e+11;
dUpp[1058] = 1.000000000000e+12;
dUpp[1059] = 1.000000000000e+12;
dUpp[1060] = 9.999999999993e+11;
dUpp[1061] = 9.999999999999e+11;
dUpp[1062] = 1.000000000000e+12;
dUpp[1063] = 9.999999999991e+11;
dUpp[1064] = 9.999999999996e+11;
dUpp[1065] = 1.000000000001e+12;
dUpp[1066] = 1.000000000000e+12;
dUpp[1067] = 1.000000000000e+12;
dUpp[1068] = 1.000000000001e+12;
dUpp[1069] = 9.999999999990e+11;
dUpp[1070] = 9.999999999993e+11;
dUpp[1071] = 1.000000000001e+12;
dUpp[1072] = 9.999999999993e+11;
dUpp[1073] = 9.999999999994e+11;
dUpp[1074] = 1.000000000000e+12;
dUpp[1075] = 9.999999999990e+11;
dUpp[1076] = 9.999999999990e+11;
dUpp[1077] = 1.000000000000e+00;
dUpp[1078] = 1.000000000000e+00;
dUpp[1079] = 1.000000000000e+00;
dUpp[1080] = 1.000000000000e+12;
dUpp[1081] = 9.999999999996e+11;
dUpp[1082] = 1.000000000000e+12;
dUpp[1083] = 1.000000000000e+12;
dUpp[1084] = 9.999999999992e+11;
dUpp[1085] = 9.999999999999e+11;
dUpp[1086] = 1.000000000000e+12;
dUpp[1087] = 9.999999999990e+11;
dUpp[1088] = 9.999999999995e+11;
dUpp[1089] = 9.999999999998e+11;
dUpp[1090] = 9.999999999998e+11;
dUpp[1091] = 1.000000000000e+12;
dUpp[1092] = 1.000000000001e+12;
dUpp[1093] = 1.000000000000e+12;
dUpp[1094] = 1.000000000000e+12;
dUpp[1095] = 1.000000000000e+12;
dUpp[1096] = 1.000000000000e+12;
dUpp[1097] = 1.000000000000e+12;
dUpp[1098] = 1.000000000000e+12;
dUpp[1099] = 9.999999999990e+11;
dUpp[1100] = 9.999999999990e+11;
dUpp[1101] = 1.000000000000e+00;
dUpp[1102] = 1.000000000000e+00;
dUpp[1103] = 1.000000000000e+00;
dUpp[1104] = 1.000000000000e+12;
dUpp[1105] = 9.999999999996e+11;
dUpp[1106] = 1.000000000000e+12;
dUpp[1107] = 1.000000000000e+12;
dUpp[1108] = 9.999999999994e+11;
dUpp[1109] = 1.000000000000e+12;
dUpp[1110] = 1.000000000000e+12;
dUpp[1111] = 9.999999999991e+11;
dUpp[1112] = 9.999999999997e+11;
dUpp[1113] = 9.999999999997e+11;
dUpp[1114] = 1.000000000000e+12;
dUpp[1115] = 1.000000000000e+12;
dUpp[1116] = 9.999999999996e+11;
dUpp[1117] = 1.000000000001e+12;
dUpp[1118] = 1.000000000001e+12;
dUpp[1119] = 9.999999999995e+11;
dUpp[1120] = 1.000000000001e+12;
dUpp[1121] = 1.000000000001e+12;
dUpp[1122] = 1.000000000000e+12;
dUpp[1123] = 9.999999999990e+11;
dUpp[1124] = 9.999999999990e+11;
dUpp[1125] = 1.000000000000e+00;
dUpp[1126] = 1.000000000000e+00;
dUpp[1127] = 1.000000000000e+00;
dUpp[1128] = 1.000000000000e+12;
dUpp[1129] = 9.999999999998e+11;
dUpp[1130] = 1.000000000000e+12;
dUpp[1131] = 1.000000000000e+12;
dUpp[1132] = 9.999999999995e+11;
dUpp[1133] = 1.000000000000e+12;
dUpp[1134] = 1.000000000000e+12;
dUpp[1135] = 9.999999999993e+11;
dUpp[1136] = 9.999999999998e+11;
dUpp[1137] = 9.999999999998e+11;
dUpp[1138] = 1.000000000001e+12;
dUpp[1139] = 1.000000000001e+12;
dUpp[1140] = 9.999999999988e+11;
dUpp[1141] = 1.000000000001e+12;
dUpp[1142] = 1.000000000001e+12;
dUpp[1143] = 9.999999999992e+11;
dUpp[1144] = 1.000000000001e+12;
dUpp[1145] = 1.000000000001e+12;
dUpp[1146] = 1.000000000000e+12;
dUpp[1147] = 9.999999999990e+11;
dUpp[1148] = 9.999999999990e+11;
dUpp[1149] = 1.000000000000e+00;
dUpp[1150] = 1.000000000000e+00;
dUpp[1151] = 1.000000000000e+00;
dUpp[1152] = 1.000000000000e+12;
dUpp[1153] = 9.999999999999e+11;
dUpp[1154] = 1.000000000000e+12;
dUpp[1155] = 9.999999999998e+11;
dUpp[1156] = 9.999999999997e+11;
dUpp[1157] = 1.000000000000e+12;
dUpp[1158] = 9.999999999999e+11;
dUpp[1159] = 9.999999999994e+11;
dUpp[1160] = 9.999999999999e+11;
dUpp[1161] = 9.999999999993e+11;
dUpp[1162] = 1.000000000000e+12;
dUpp[1163] = 1.000000000000e+12;
dUpp[1164] = 9.999999999993e+11;
dUpp[1165] = 1.000000000001e+12;
dUpp[1166] = 1.000000000000e+12;
dUpp[1167] = 9.999999999994e+11;
dUpp[1168] = 9.999999999998e+11;
dUpp[1169] = 9.999999999998e+11;
dUpp[1170] = 1.000000000000e+12;
dUpp[1171] = 9.999999999990e+11;
dUpp[1172] = 9.999999999990e+11;
dUpp[1173] = 1.000000000000e+00;
dUpp[1174] = 1.000000000000e+00;
dUpp[1175] = 1.000000000000e+00;
dUpp[1176] = 9.999999999998e+11;
dUpp[1177] = 9.999999999999e+11;
dUpp[1178] = 1.000000000000e+12;
dUpp[1179] = 9.999999999998e+11;
dUpp[1180] = 9.999999999997e+11;
dUpp[1181] = 1.000000000000e+12;
dUpp[1182] = 9.999999999998e+11;
dUpp[1183] = 9.999999999993e+11;
dUpp[1184] = 9.999999999998e+11;
dUpp[1185] = 9.999999999995e+11;
dUpp[1186] = 9.999999999995e+11;
dUpp[1187] = 9.999999999994e+11;
dUpp[1188] = 1.000000000000e+12;
dUpp[1189] = 9.999999999996e+11;
dUpp[1190] = 9.999999999995e+11;
dUpp[1191] = 1.000000000000e+12;
dUpp[1192] = 9.999999999998e+11;
dUpp[1193] = 9.999999999997e+11;
dUpp[1194] = 1.000000000000e+12;
dUpp[1195] = 9.999999999990e+11;
dUpp[1196] = 9.999999999990e+11;
dUpp[1197] = 1.000000000000e+00;
dUpp[1198] = 1.000000000000e+00;
dUpp[1199] = 1.000000000000e+00;
dUpp[1200] = 9.999999999998e+11;
dUpp[1201] = 9.999999999998e+11;
dUpp[1202] = 1.000000000000e+12;
dUpp[1203] = 9.999999999998e+11;
dUpp[1204] = 9.999999999995e+11;
dUpp[1205] = 1.000000000000e+12;
dUpp[1206] = 9.999999999999e+11;
dUpp[1207] = 9.999999999993e+11;
dUpp[1208] = 9.999999999997e+11;
dUpp[1209] = 1.000000000001e+12;
dUpp[1210] = 9.999999999991e+11;
dUpp[1211] = 9.999999999990e+11;
dUpp[1212] = 1.000000000000e+12;
dUpp[1213] = 9.999999999987e+11;
dUpp[1214] = 9.999999999989e+11;
dUpp[1215] = 1.000000000001e+12;
dUpp[1216] = 9.999999999998e+11;
dUpp[1217] = 9.999999999997e+11;
dUpp[1218] = 1.000000000000e+12;
dUpp[1219] = 9.999999999990e+11;
dUpp[1220] = 9.999999999990e+11;

return;
}





/*
 *	end of file
 */
